;;; -*- Mode:Common-Lisp; Package:CONDITIONS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (c) 1989 Texas Instruments Incorporated.  All Rights Reserved.

;;; Common LISP CONDITION System 

;;; This file contains functions and macros which follow the Common LISP
;;; standard on defining and making conditions.

;;; The Explorer version of the Common LISP Condition System exists in its
;;; own package, CONDITIONS, aka CLEH.

;;; This is not a portable version of the Common LISP condition/error
;;; system.  Rather it is an implementation which depends on and interfaces
;;; with the existing Explorer error system.  It allows both the Common LISP
;;; error system and the Explorer error system to co-exist.  The Explorer
;;; error handler has been changed to accommodate features in the Common LISP
;;; error system.

;;;Record of Changes:
;;; 03/09/89 clm  fixed two problems in DEFINE-CONDITION.  1) made SLOT-SPECS argument
;;;               optional as per Common Lisp Condition System, Revision #18. 2) renamed
;;;               agrument STREAM used in DEFMETHOD to REPORT-STREAM.  This was causing
;;;               a name conflict with STREAM-ERROR instance variable.


;;; Making and defining conditions

;;; Note that:
;;; MAKE-CONDITION is built on MAKE-INSTANCE.
;;; DEFINE-CONDITION is built on DEFFLAVOR. 

;;; For a portable Common LISP condition system, MAKE-CONDITION and
;;; DEFINE-CONDITION would need to be rewritten to not depend on the
;;; Explorer flavor system.

(DEFUN condition-type-p (type &aux flavor-name depends-on)
  (COND ((AND (SETQ flavor-name (GET type 'si:flavor))
	      (OR (EQ type 'condition)
		  (MEMBER 'condition 
			  (SETQ depends-ON (SI:FLAVOR-DEPENDS-ON flavor-name)) :test #'eq)
		  ))
	 t)
	(depends-on 
	 (DOLIST (f depends-on)
	   (IF (condition-type-p f) (RETURN t))))))

(DEFUN get-local-slots (type)
  (UNLESS (TYPEP type 'si:flavor)
    (SETQ type (GET type 'si:flavor)))
  (IF type
      (MAPCAR #'(LAMBDA (slot)
		  (IF (ATOM slot) slot (CAR slot)))
	      (SI:FLAVOR-LOCAL-INSTANCE-VARIABLES type))))

(DEFUN get-parents (type)
  (LET ((flavor-type (GET type 'si:flavor)))
    (COND
     ((EQ type 'condition) (LIST type))
     ((NULL flavor-type) nil)
     (t
      (CONS type
       (MAPCAN #'(LAMBDA (pt) (get-parents pt))
	       (SI:FLAVOR-DEPENDS-ON flavor-type)))))))

(DEFUN get-all-slots (type slots)
  (LET ((type-flavor (GET type 'si:flavor)))
    (COND ((NULL type-flavor) slots)
	  ((EQ type 'condition)
	   (NUNION slots (get-local-slots type-flavor)))
	    (T
	     (REMOVE-DUPLICATES
	      (MAPCAN #'(LAMBDA (parent)
			  (get-local-slots (GET parent 'si:flavor)))
		      (get-parents type)))))))

(DEFMACRO report-function (condition-type)
  `(GET ,condition-type 'report-function))

(DEFUN make-condition (type &rest slot-initializations)
  "  MAKE-CONDITION creates and returns a condition object of the given
TYPE using the SLOT-INITIALIZATIONS as a specification of the initial
value of the slots.  The TYPE must have been defined with
DEFINE-CONDITION.
   The SLOT-INITIALIZATIONS are given as alternating keyword/value
pairs.  Each SLOT-INTIALIZATION has the form:  {:slot-name value}.  The
keyword part of each SLOT-INTIALIZATION is the printname (in the keyword
package) of any of the defined slots of TYPE.  Each designated slot will
be initialized to a specified value.
"
  (COND ((condition-type-p type)
	 (APPLY #'MAKE-INSTANCE type slot-initializations))
	(:else
	 (ERROR 'simple-type-error
		:DATUM type
		:EXPECTED-TYPE '(SATISFIES condition)
		:FORMAT-STRING "Not a condition type: ~S"
		:FORMAT-ARGUMENTS (LIST type)))))

;;; 05/08/89 clm - redefined so that this now uses CLOS.
(DEFMACRO define-condition (name (parent-types) &optional slot-specs &rest options)
  (LET (class-options
	report-function )
    (DO ((o options (CDR o)))
	((NULL o))
      (LET ((option (CAR o)))
	(CASE (CAR option)
	  (:documentation (setq class-options (list option)))
	  
	  (:REPORT (SETQ report-function (IF (STRINGP (CADR option))
					     `(LAMBDA (condition stream)
						(DECLARE (IGNORE condition))
						(WRITE-STRING ,(CADR option) stream))
					     (CADR option))))
	  )))
    ;; Finally, the expansion ...
    (IF report-function (SETF (report-function name) `#',report-function))
    (push (list :metaclass 'ticlos:flavor-class) class-options)
    `(PROGN 
	    (TICLOS:DEFCLASS ,name
		      (,parent-types)
		      ,slot-specs
	      ,@class-options)
	    ,(IF report-function
		 `(ticlos:DEFMETHOD ticlos:print-object ((X ,name) report-stream)
		    (if *print-escape*
			(ticlos:call-next-method)
			(,report-function X report-stream))
		    X
		    )))))
						
