;;; -*- Mode:Common-Lisp; Package:CONDITIONS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (c) 1989 Texas Instruments Incorporated.  All Rights Reserved.

;;; Common LISP CONDITION System 

;;; This file contains functions and macros which follow the Common LISP
;;; standard on making and establishing condition handlers

;;; The Explorer version of the Common LISP Condition System exists in its
;;; own package, CONDITIONS, aka CLEH.

;;; This is not a portable version of the Common LISP condition/error
;;; system.  Rather it is an implementation which depends on and interfaces
;;; with the existing Explorer error system.  It allows both the Common LISP
;;; error system and the Explorer error system to co-exist.  The Explorer
;;; error handler as been changed to accommodate features in the Common LISP
;;; error system.


;;; Macros for establishing handlers

;;; Note that:

;;; HANDLER-BIND is built on EH:CONDITION-BIND.
;;; For a portable Common LISP condition system, HANDLER-BIND would need to
;;; be rewritten to not depend on Explorer error system for establishing
;;; condition handlers.

;;; HANDLER-CASE is built on HANDLER-BIND and is portable.
;;; IGNORE-ERRORS is built on HANDLER-CASE and is portable.


(DEFMACRO handler-bind (handler-bindings &body forms)
  "  HANDLER-BIND executes FORMS in the dynamic context where the given
handler bindings are in effect
  The HANDLER-BINDINGS have the form:  
     ({(CONDITION-TYPESPEC HANDLER-FUNCTION)}*)
  The CONDITION-TYPESPEC may be any type specifier.  If during the
execution of the FORMS a condition is signalled for which (TYPEP
<condition-object> 'CONDITION-TYPESPEC) is true, and there is no
intervening handler for conditions of that type then the
HANDLER-FUNCTION for this handler binding is invoked.  The
HANDLER-FUNCTION should evaluate to a function of one argument, the
signalled condition.
  The HANDLER-FUNCTION is executed in the dynamic context of when the
condition is signaled.  It can look at the state of the condition in
determining whether to accept or decline this condition.  The
HANDLER-FUNCTION usually performs some non-local transfer of control
such as a GO to a tag in a TAGBODY, a RETURN from a BLOCK, or a THROW to
a CATCH.  If it returns, it essentially declines the condition and other
handlers will be sought to handle this condition.
  If no condition is signalled and the FORMS are executed to completion,
then HANDLER-BIND returns the values of the last form.
  If more than one handler binding is specified, the bindings are
searched sequentially from top to bottom for a CONDITION-TYPESPEC which
matches thrsignalled condition.  If the same CONDITION-TYPESPEC
is used in two handler bindings, the earlier of the two bindings will be
selected.
 If an appropriate CONDITION-TYPESPEC is found, the associated
HANDLER-FUNCTION is executed in a context where none of the previous
handler bindings are visible.
"
  (UNLESS (EVERY #'(LAMBDA (h) (AND (LISTP h) (= (LENGTH h) 2))) handler-bindings)
    (ERROR "Ill-formed handler bindings."))
  (LET ((typespec-handler-bindings
	 (MAPCAR #'(LAMBDA (h)  
		     `((typespec ,(first h)) ,(second h)))
		     handler-bindings)))
    `(EH:CONDITION-BIND-IF t ,typespec-handler-bindings . ,forms)))

(DEFMACRO handler-case (expression &rest cases)
  "  HANDLER-CASE executes the EXPRESSION in the dynamic context where
condition handlers are established to transfer control to the
appropriate HANDLER-CASE clause.
  The CASES have the form: {(CONDITION-TYPESPEC ([VAR]) {FORM}*)}*
  The CONDITION-TYPESPEC may be any type specifier.  If during the
execution of the EXPRESSION a condition is signalled for which (TYPEP
<condition-object> 'CONDITION-TYPESPEC) is true, and there is no
intervening handler for conditions of that type, then control is
transferred to the FORMS of the relevant clause unwinding the dynamic
state appropriately in the process.
  The VAR is any symbol name and is bound to the condition which was
signalled.  Any FORM in a clause can reference VAR, the condition
object.  If VAR is not needed, it may be omitted by using ().  If a
clause is selected, HANDLER-CASE returns the value of the last form in
the clause.  If there are no forms in the selected clause, then
HANDLER-CASE returns NIL.
  A condition handler is established for each HANDLER-CASE clause.  When
a condition is signalled, the handlers are searched sequentially from
top to bottom for a CONDITION-TYPESPEC match.  If the same
CONDITION-TYPESPEC is used in two cases, the earlier of the two cases
with the desired typespec will be selected.
  If no condition is signalled and the EXPRESSION runs to completion,
then the values resulting from the EXPRESSION are returned by
HANDLER-CASE.
  The CONDITION-TYPESPEC can also be the symbol :NO-ERROR in the last
clause.  The FORMS in the :NO-ERROR clause will be executed if the
EXPRESSION returns normally.  In this case, there can be more than one
VAR argument.  Each VAR is bound to the values returned by the
EXPRESSION.  Any FORM in the :no-error clause can reference the values
of the EXPRESSION.  The values of the last FORM in the :NO-ERROR clause
are returned by HANDLER-CASE.
"
  (LET ((no-error-clause (ASSOC ':no-error cases)))
    (IF no-error-clause
	(LET ((normal-return (MAKE-SYMBOL "NORMAL-RETURN"))
	      (error-return  (MAKE-SYMBOL "ERROR-RETURN")))
	  `(BLOCK ,error-return
	     (MULTIPLE-VALUE-CALL #'(LAMBDA ,@(CDR no-error-clause))
	       (BLOCK ,normal-return
		 (RETURN-FROM ,error-return
		   (HANDLER-CASE (return-from ,normal-return ,expression)
		     ,@(REMOVE no-error-clause cases)))))))
	(LET ((tag (GENSYM))
	      (var (GENSYM))
	      (ANNOTATED-CASES (MAPCAR #'(LAMBDA (case) (CONS (GENSYM) case))
				       cases)))
	  `(BLOCK ,tag
	     (LET ((,var nil))
	       ,var				;ignorable
	       (TAGBODY
		 (HANDLER-BIND ,(MAPCAR #'(LAMBDA (annotated-case)
					    (LIST (CADR annotated-case)
						  `#'(LAMBDA (temp)
						       ,@(IF (CADDR annotated-case)
							     `((SETQ ,var temp)))
						       temp
						       (GO ,(CAR annotated-case)))))
					annotated-cases)
			       (RETURN-FROM ,tag ,expression))
		 ,@(MAPCAN #'(LAMBDA (annotated-case)
			       (LIST (CAR annotated-case)
				     (LET ((body (CDDDR annotated-case)))
				       `(RETURN-FROM ,tag
					  ,(COND ((CADDR annotated-case)
						  `(LET ((,(CAADDR annotated-case)
							  ,var))
						     ,@body))
						 ((NOT (CDR body))
						  (CAR body))
						 (T
						  `(PROGN ,@body)))))))
			   annotated-cases)))))))
  )

(DEFMACRO ignore-errors (&rest forms)
  "  IGNORE-ERRORS executes FORMS with a handler established to transfer
control back to IGNORE-ERRORS if a condition of type ERROR is signalled.
If no such condition is signalled, IGNORE-ERRORS returns the values of
the last form.  If an error is signalled, IGNORE-ERRORS returns two
values:  NIL and the error condition which was signalled.
"
  `(handler-case (PROGN ,@forms)
     (global:error (condition) (VALUES nil condition))))

