;;; -*- Mode:Common-Lisp; Package:CONDITIONS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (c) 1989 Texas Instruments Incorporated.  All Rights Reserved.

;;; Pre-defined Common LISP condition types

;;; Record of changes
;;; 05/08/89  clm  -  Changed implementation of DEFINE-CONDITION to use CLOS.
;;;                   This required changes to parameter lists and removed
;;;                   :conc-name option.

;; Note that EH:WARNING and EH:SERIOUS-CONDITION are subtypes of EH:CONDTION.

(DEFINE-CONDITION warning (eh:warning) ()
  (:documentation
"All types of warnings should inherit from this type.  This is a subtype
of CONDITION."))

(DEFINE-CONDITION serious-condition (eh:serious-condition) ()
  (:documentation
"All serious conditions (conditions serious enough to require
interactive intervention if not handled) should inherit from this type.
This is a subtype of CONDITION."))

(DEFINE-CONDITION error (eh:error) ()
 (:documentation
"All types of error conditions inherit from this condition.  This is a
subtype of SERIOUS-CONDITION."))

(DEFINE-CONDITION type-error (error)
  ((datum :initarg :datum :reader type-error-datum)
   (expected-type :initarg :expected-type :reader type-error-expected-type))
  (:documentation
"Errors in the transfer of data in a program should inherit from this
type.  This is a subtype of ERROR.  For example, conditions to be
signalled by CHECK-TYPE should inherit from this type.  The init
keywords :DATUM and :EXPECTED-TYPE are supported to initialize the
slots, which can be accessed using TYPE-ERROR-DATUM and
TYPE-ERROR-EXPECTED-TYPE."))

(DEFINE-CONDITION simple-condition (condition)
  ((format-string :initarg :format-string :reader internal-simple-condition-format-string)
   (format-arguments :initarg :format-arguments :reader internal-simple-condition-format-arguments
		     :initform '()) 
   (format-args :initarg :format-args :reader internal-simple-condition-format-args
		:initform nil))
  (:report simple-condition-printer)
  (:documentation
"Conditions signalled by SIGNAL when given a format string as a first
argument are of this type.  This is a subtype of CONDITION.  The init
keywords :FORMAT-STRING and :FORMAT-ARGUMENTS are supported to
initialize the slots, which can be accessed using
SIMPLE-CONDITION-FORMAT-STRING and SIMPLE-CONDITION-FORMAT-ARGUMENTS.
If :FORMAT-ARGUMENTS is not supplied to MAKE-CONDITION, the
format-arguments slot defaults to NIL."))

(DEFINE-CONDITION simple-warning (warning)
  ((format-string :initarg :format-string :reader internal-simple-warning-format-string)
   (format-arguments :initarg :format-arguments :reader internal-simple-warning-format-arguments
		     :initform '())
   (format-args :initarg :format-args :reader internal-simple-warning-format-args
		:initform nil))
  (:report simple-condition-printer)
  (:documentation
"Conditions signalled by WARN when given a format string as a first
argument are of this type.  This is a subtype of WARNING.  The init
keywords :FORMAT-STRING and :FORMAT-ARGUMENTS are supported to
initialize the slots, which can be accessed using
SIMPLE-CONDITION-FORMAT-STRING and SIMPLE-CONDITION-FORMAT-ARGUMENTS.
If :FORMAT-ARGUMENTS is not supplied to MAKE-CONDITION, the
format-arguments slot defaults to NIL."))

(DEFINE-CONDITION simple-error (error)
  ((format-string :initarg :format-string :reader internal-simple-error-format-string)
   (format-arguments :initarg :format-arguments :reader internal-simple-error-format-arguments 
		     :initform '())
   (format-args :initarg :format-args :reader internal-simple-error-format-args
		:initform nil))
  (:report simple-condition-printer)
  (:documentation
"Conditions signalled by ERROR and CERROR when given a format string as
a first argument are of this type.  This is a subtype of ERROR.  The
init keywords :FORMAT-STRING and :FORMAT-ARGUMENTS are supported to
initialize the slots, which can be accessed using
SIMPLE-CONDITION-FORMAT-STRING and SIMPLE-CONDITION-FORMAT-ARGUMENTS.
If :FORMAT-ARGUMENTS is not supplied to MAKE-CONDITION, the
format-arguments slot defaults to NIL."))

(DEFINE-CONDITION simple-type-error (type-error)
  ((format-string :initarg :format-string :reader internal-simple-type-error-format-string)
   (format-arguments :initarg :format-arguments :reader internal-simple-type-error-format-arguments 
		     :initform '())
   (format-args :initarg :format-args :reader internal-simple-type-error-format-args
		:initform nil))
  (:report simple-condition-printer)
  (:documentation 
"Conditions signalled by facilities similar to CHECK-TYPE may want to
use this type.  The init keywords :FORMAT-STRING and :FORMAT-ARGUMENTS
are supported to initialize the slots, which can be accessed using
SIMPLE-CONDITION-FORMAT-STRING and SIMPLE-CONDITION-FORMAT-ARGUMENTS.
If :FORMAT-ARGUMENTS is not supplied to MAKE-CONDITION, the
format-arguments slot defaults to NIL."))

(DEFUN simple-condition-format-string (condition)
  "Accesses the format-string slot of a given condition, which must be
of type SIMPLE-CONDITION, SIMPLE-WARNING, SIMPLE-ERROR, or
SIMPLE-TYPE-ERROR.
"
  (etypecase condition
    (simple-condition  (internal-simple-condition-format-string  condition))
    (simple-warning    (internal-simple-warning-format-string    condition))
    (simple-type-error (internal-simple-type-error-format-string condition))
    (simple-error      (internal-simple-error-format-string      condition))))

(DEFUN simple-condition-format-arguments (condition)
  "Accesses the format-arguments slot of a given condition, which must
be of type SIMPLE-CONDITION, SIMPLE-WARNING, SIMPLE-ERROR, or
SIMPLE-TYPE-ERROR.
"
  (etypecase condition
    (simple-condition  (internal-simple-condition-format-arguments  condition))
    (simple-warning    (internal-simple-warning-format-arguments    condition))
    (simple-type-error (internal-simple-type-error-format-arguments condition))
    (simple-error      (internal-simple-error-format-arguments      condition))))

(DEFUN simple-condition-format-args (condition)
  "Accesses the format-arguments slot of a given condition, which must
be of type SIMPLE-CONDITION, SIMPLE-WARNING, SIMPLE-ERROR, or
SIMPLE-TYPE-ERROR.
"
  (etypecase condition
    (simple-condition  (internal-simple-condition-format-args  condition))
    (simple-warning    (internal-simple-warning-format-args    condition))
    (simple-type-error (internal-simple-type-error-format-args condition))
    (simple-error      (internal-simple-error-format-args      condition))))

(DEFUN simple-condition-printer (condition stream)
  "Report function used by conditions of type SIMPLE-CONDITION,
SIMPLE-WARNING, SIMPLE-ERROR, or SIMPLE-TYPE-ERROR.
"
  (apply #'format stream (simple-condition-format-string    condition)
                      	 (or (simple-condition-format-arguments condition)
			     (simple-condition-format-args condition))
			 ))


(DEFINE-CONDITION storage-condition (serious-condition) ()
  (:documentation 
"Conditions which relate to storage overflow should inherit from this
type.  This is a subtype of SERIOUS-CONDITION."))

(DEFINE-CONDITION stack-overflow    (storage-condition) ()
  (:documentation "This is a subtype of storage-condition"))

(DEFINE-CONDITION storage-exhausted (storage-condition) ()
  (:documentation "This is a subtype of storage-condition"))

(DEFINE-CONDITION case-failure (type-error)
  ((name :initarg :name :reader case-failure-name)
   (test-object :initarg :test-object :reader case-failure-test-object)
   (possibilities :initarg :possibilities :reader case-failure-possibilities))
  (:report
    (lambda (condition stream)
      (let ((cname (case-failure-name condition)))
	(format stream "~A failure: The value of ~S is ~S which ~A: ~{~S~^, ~}."
		(case-failure-name condition)
		(case-failure-test-object condition)
		(type-error-datum condition)
		(if (or (eq cname 'ctypecase) (eq cname 'etypecase))
		    "does not satisfy type specifier(s)"
		    "is not one of")
 	      (case-failure-possibilities condition)))))
  (:documentation
"Conditions of this type are signalled by ECASE, CCASE, ETYPECASE and
CTYPECASE when no clause is satisfied.  This is a subtype of TYPE-ERROR."))

(DEFINE-CONDITION program-error (error) ()
  (:documentation
"Errors which relate to incorrect program syntax which are statically
detectable should inherit from this type.  This is a subtype of ERROR."))

(DEFINE-CONDITION control-error (error) ()
  (:documentation
"Errors in the dynamic transfer of control in a program should inherit
from this type.  This is a subtype of ERROR."))

(DEFINE-CONDITION stream-error (error)
  ((stream :initarg :stream :reader stream-error-stream))
  (:documentation
"Errors which occur during input from, output to, or closing of a
stream should inherit from this type.  This is a subtype of ERROR.  The
init keyword :STREAM is supported to initialize the slot, which can be
accessed using STREAM-ERROR-STREAM."))

(DEFINE-CONDITION end-of-file (stream-error) ()
  (:documentation
"The error which results when a read operation is done on a stream which
has no more tokens should inherit from this type.  This is a subtype of
STREAM-ERROR."))

(DEFINE-CONDITION file-error (error)
  ((pathname :initarg :pathname :reader file-error-pathname))
  (:documentation
"Errors which occur during an attempt to open or close a file, or during
some low-level transaction with a file system should inherit from this
type.  This is a subtype of ERROR.  The init keyword :PATHNAME is
supported to initialize the slot, which can be accessed using
FILE-ERROR-PATHNAME."))

(DEFINE-CONDITION package-error (error)
  ((package-name :initarg :package-name :reader package-error-package-name))
  (:documentation
 "Errors which occur during operations on packages should inherit from
this type.  This is a subtype of ERROR.  The init keyword :PACKAGE is
supported to initialize the slot, which can be accessed using
PACKAGE-ERROR-PACKAGE-NAME."))


(DEFINE-CONDITION cell-error (error)
  ((name :initarg :name :reader cell-error-name))
  (:documentation
"Errors which occur while accessing a location should inherit from this
type.  This is a subtype of ERROR.  The init keyword :NAME is supported
to initialize the slot, which can be accessed using CELL-ERROR-NAME."))

(DEFINE-CONDITION unbound-variable (cell-error) ()
  (:report (lambda (condition stream)
	     (format stream "The variable ~S is unbound."
		     (cell-error-name condition))))
  (:documentation
"The error which results from trying to access the value of an unbound
variable should inherit from this type.  This is a subtype of
CELL-ERROR."))
  
(DEFINE-CONDITION undefined-function (cell-error) ()
  (:report (lambda (condition stream)
	     (format stream "The function ~S is undefined."
		     (cell-error-name condition))))
  (:documentation
"The error which results from trying to access the value of an undefined
function should inherit from this type.  This is a subtype of CELL-ERROR."))

(DEFINE-CONDITION arithmetic-error (error)
  ((operation :initarg :operation :reader arithmetic-error-operation)
   (operands :initarg :operands :reader arithmetic-error-operands))
  (:documentation
"Errors which occur while doing arithmetic type operations should
inherit from this type.  This is a subtype of ERROR.  The init keywords
:OPERATION and :OPERANDS are supported to initialize the slots, which
can be accessed using ARITHMETIC-ERROR-OPERATION and
ARITHMETIC-ERROR-OPERANDS."))

(DEFINE-CONDITION division-by-zero         (arithmetic-error) ()
  (:documentation
"Errors which occur because of division by 0 should inherit from this
type.  This is a subtype of ARITHMETIC-ERROR."))

(DEFINE-CONDITION floating-point-overflow  (arithmetic-error) ()
   (:documentation
"Errors which occur because of floating point overflow should inherit
from this type.  This is a subtype of ARITHMETIC-ERROR."))

(DEFINE-CONDITION floating-point-underflow (arithmetic-error) ()
  (:documentation
"Errors which occur because of floating point underflow should inherit
from this type.  This is a subtype of ARITHMETIC-ERROR."))

(DEFINE-CONDITION abort-failure (control-error) ()
  (:report "Abort failed.")
  (:documentation 
"Condition signalled by restart function ABORT when no restart of type
ABORT is active.  This is a subtype of CONTROL-ERROR."))

