;;; -*- Mode:Common-Lisp; Package:CONDITIONS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (c) 1989 Texas Instruments Incorporated.  All Rights Reserved.

;;; Common LISP CONDITION System 

;;; This file contains functions and macros which follow the Common LISP
;;; standard on making and establishing restarts, finding and invoking
;;; active restarts.

;;; The Explorer version of the Common LISP Condition System exists in its
;;; own package, CONDITIONS, aka CLEH.

;;; This is not a portable version of the Common LISP condition/error
;;; system.  Rather it is an implementation which depends on and interfaces
;;; with the existing Explorer error system.  It allows both the Common LISP
;;; error system and the Explorer error system to co-exist.  The Explorer
;;; error handler as been changed to accommodate features in the Common LISP
;;; error system.


;;; Unique Ids  - used by restarts

(DEFVAR *unique-id-table* (MAKE-HASH-TABLE))
(DEFVAR *unique-id-count* -1)

(DEFUN unique-id (obj)
  "Generates a unique integer ID for its argument."
  (OR (GETHASH obj *unique-id-table*)
      (SETF (GETHASH obj *unique-id-table*) (INCF *unique-id-count*))))

(EVAL-WHEN (eval compile load)

;;; Miscellaneous utility - used by restarts

(DEFUN parse-keyword-pairs (list keys)
  (DO ((l list (CDDR l))
       (k '() (LIST* (CADR l) (CAR l) k)))
      ((OR (NULL l) (NOT (MEMBER (CAR l) keys)))
       (VALUES (NREVERSE k) l))))

(DEFMACRO with-keyword-pairs ((names expression &optional keywords-var) &body forms)
  "Macro used by RESTART-CASE to parse keyword-value pairs."
  (LET ((temp (MEMBER '&rest names)))
    (UNLESS (= (LENGTH temp) 2) (error "&REST keyword is ~:[missing~;misplaced~]." temp))
    (LET ((key-vars (LDIFF names temp))
          (key-var (OR keywords-var (GENSYM)))
          (rest-var (CADR temp)))
      (LET ((keywords (MAPCAR #'(LAMBDA (x) (INTERN (STRING x) (FIND-PACKAGE "KEYWORD")))
			      key-vars)))
        `(MULTIPLE-VALUE-BIND (,key-var ,rest-var)
             (parse-keyword-pairs ,expression ',keywords)
           (LET ,(MAPCAR #'(LAMBDA (var keyword) `(,var (GETF ,key-var ,keyword)))
                                 key-vars keywords)
             ,@forms))))))

);NEHW-LAVE

;(DEFUN read-evaluated-form (&optional (prompt "~&Type a form to be evaluated and used instead: ") &rest args)
;  (APPLY #'FORMAT *query-io* prompt args)
;  (EVAL (READ *query-io*)))


(DEFUN read-evaluated-form (&optional (prompt "~&Type a form to be evaluated and used instead: ") &rest args)
  ;; new and improved to use special window debugger read function if 
  ;; if window debugger is active
  (apply #'format *query-io* prompt args)
  (if eh:*window-debugger-old-window*		;window debugger has its own read function
      (apply 'eh:window-read-object :eval-read prompt args)
      (eval (read *query-io*))))



;;; Making restarts

(DEFSTRUCT (restart (:print-function restart-print))
  name
  function
  report-function
  interactive-function
  condition-names	;; for converting Explorer resume handler to a restart
  predicate)		;; for converting Explorer resume handler to a restart

(DEFUN restart-report (restart stream)
  (FUNCALL (OR (restart-report-function restart)
               (LET ((NAME (restart-name restart)))
		 #'(LAMBDA (stream)
		     (IF name
			 (format stream "~S" name)
			 (format stream "~S" restart)))))
           stream))

(DEFUN restart-print (restart stream depth)
  (DECLARE (ignore depth))
  (IF *print-escape*
      (FORMAT stream "#<~S.~D>" (TYPE-OF restart) (UNIQUE-ID restart))
      (restart-report restart stream)))


;;; Macros for establishing restarts

;;; Note that:

;;; RESTART-BIND is built on EH:CONDITION-RESUME.
;;; For a portable Common LISP condition system, RESTART-BIND would need to
;;; be rewritten to not depend on the Explorer error system for establishing
;;; resume handlers.

;;; RESTART-CASE is built on RESTART-BIND and is portable.
;;; WITH-SIMPLE-RESTART is built on RESTART-CASE and is portable.

(DEFMACRO restart-bind (restart-bindings &body forms)
  "  RESTART-BIND executes FORMS in a dynamic context where the given
restart bindings are in effect.
  The RESTART-BINDINGS have the form:
    ({(NAME FUNCTION {KEYWORD VALUE}*)}*)
  The NAME may be NIL to indicate an anonymous restart, or some other
symbol to indicate a named restart.  There can be more than one restart
binding with the same restart name.  In this case, the first binding
with that name will be found by the FIND-RESTART or INVOKE-RESTART.  The
other clauses with the same restart name are accessible using
COMPUTE-RESTARTS.
  The FUNCTION should evaluate to a function which if executed will
perform the restart.  This function may either do a non-local transfer
of control or it may return normally.  The function may take whatever
arguments are appropriate.  This function will be invoked only if the
functions, INVOKE-RESTART or INVOKE-RESTART-INTERACTIVELY, are called.
  Any form being executed in FORMS may invoke a restart by calling
INVOKE-RESTART with the desired restart NAME.  If a condition is
signalled while executing a FORM, a restart might also be invoked either
explicitly from a handler (by calling INVOKE-RESTART) or implicitly from
the debugger (which calls INVOKE-RESTART-INTERACTIVELY) after a restart
command is selected by a user.  If a restart is invoked, the FUNCTION of
the restart is executed.  If FORMS finishes executing and no restart is
invoked, then the values resulting from the last FORM are returned by
RESTART-BIND.
  The valid KEYWORD/VALUE pairs in RESTART-BIND are:
  :INTERACTIVE-FUNCTION form
  The :INTERACTIVE-FUNCTION option is used to define a function which
prompts for arguments to be used when invoking a restart.
  The FORM will be evaluated in the current lexical environment and
should return a function of no arguments which constructs and returns a
list of arguments to be used by INVOKE-RESTART-INTERACTIVELY when
invoking this restart.  The function usually uses *QUERY-IO* to prompt
interactively for the values of arguments to return.  The debugger uses
INVOKE-RESTART-INTERACTIVELY to perform a selected restart after
  :REPORT-FUNCTION form
  The :REPORT-FUNCTION option is used to define a function which reports
the purpose of the restart.
  The FORM will be evaluated in the current lexical environment and
should return a function of one argument, a stream, which prints on the
stream a summary of the action that this restart will take.  This
function is called whenever the restart is printed while *PRINT-ESCAPE*
is NIL.  This is how the debugger prints the restart report messages for
all active restarts when describing ways of continuing execution.
"
  `(EH:CONDITION-RESUME
    (LIST ,@(MAPCAR #'(LAMBDA (binding)
			`(make-restart
			  :NAME     ',(CAR binding)
			  :FUNCTION ,(CADR binding)
			  ,@(CDDR binding)))
		    restart-bindings))
     ,@forms))

(DEFMACRO restart-case (expression &body clauses)
  "
  RESTART-CASE executes EXPRESSION in a dynamic context where restarts
have been established to transfer control to an appropriate RESTART-CASE
clause.
  The CLAUSES must have the form:
   {(NAME ARGLIST {KEYWORD VALUE}* {FORM}*)}*
  The NAME may be NIL to indicate an anonymous restart, or some other
symbol to indicate a named restart.  There can be more than one
RESTART-CASE clause with the same restart name.  In this case, the first
clause with that name will be found by the FIND-RESTART or
INVOKE-RESTART.  The other clauses with the same restart name are
accessible using COMPUTE-RESTARTS.
  The ARGLIST is a normal LAMBDA arglist to be bound during the
execution of its corresponding FORMS.  These arguments are 
used to pass to the FORMS of a RESTART-CASE clause any data gathered 
by INVOKE-RESTART or INVOKE-RESTART-INTERACTIVELY.
  By default, INVOKE-RESTART-INTERACTIVELY will pass no arguments to the
FORMS of a RESTART-CASE clause.  Usually all arguments in ARGLIST are
optional in order to accomodate interactive restarting.  However, the
arguments need not be optional if the :INTERACTIVE keyword has been used
to inform INVOKE-RESTART-INTERACTIVELY about how to compute a proper
argument list.
  While executing EXPRESSION, any form in EXPRESSION may transfer
control to one of the RESTART-CASE clauses by calling INVOKE-RESTART
with the desired restart NAME.  If a condition is signalled while
executing EXPRESSION, a restart might also be invoked either explicitly
from a handler (by calling INVOKE-RESTART) or implicitly from the
debugger (which calls INVOKE-RESTART-INTERACTIVELY) after a restart
command is selected by a user.
  If a restart is invoked, the FORMS of the selected clause will be
executed.  RESTART-CASE returns the values of the last form of the
clause.  If there are no forms in a selected clause, RESTART-CASE
returns NIL.  If EXPRESSION finishes executing and no restart is
invoked, then the values resulting from the EXPRESSION are returned by
RESTART-CASE.
  The valid KEYWORD/VALUE pairs in RESTART-CASE are:
  :INTERACTIVE exp
  The :INTERACTIVE option provides a way of defining a function which
prompts for arguments to be used when invoking a restart.
   EXP must be a suitable argument for the special form, FUNCTION.  The
expression (FUNCTION EXP) will be evaluated in the current lexical
environment and should return a function with no arguments.  This
function is called by INVOKE-RESTART-INTERACTIVELY prior to invoking the
restart.  It returns a list of arguments which are passed on to the
restart function.  This function will be called in the dynamic
environment available prior to any restart invocation.  This function
usually does user interaction on the stream, *QUERY-IO*, prompting for
the data to return and use when invoking the restart.
  If a restart is invoked interactively but no :INTERACTIVE option was
supplied, the empty list is passed on to the restart function being
invoked.
  :REPORT exp 
  The :REPORT option provides a way a reporting the purpose of the
restart.
  If EXP is a literal string, it is a shorthand for
  (LAMBDA (STREAM) (WRITE-STRING exp STREAM))
  If EXP is not a literal string, it must be a suitable argument for the
special form, FUNCTION.  The expression (FUNCTION exp) will be evaluated
in the current lexical environment and should return a function of one
argument, a stream.  This function prints on the stream a description of
the restart.  This function is called whenever the restart is printed
while *PRINT-ESCAPE* is NIL.  This is how the debugger prints the
restart report messages for all active restarts when describing ways of
continuing execution.  
  If a named restart is asked to report but no :REPORT option was
supplied, the name of the restart is used in generating the 
report message of the restart.
"

  (FLET ((transform-keywords (&key report interactive)
	   (LET ((result '()))
	     (WHEN report
	       (SETQ result (LIST* (IF (STRINGP report)
				       `#'(LAMBDA (stream)
					    (WRITE-STRING ,report stream))
				       `#',report)
				   :REPORT-FUNCTION 
				   result)))
	     (WHEN interactive
	       (SETQ result (LIST* `#',interactive
				   :INTERACTIVE-FUNCTION
				   result)))
	     (NREVERSE result))))
    (LET ((block-tag (GENSYM))
	  (temp-var  (GENSYM))
	  (data
	    (MAPCAR #'(LAMBDA (clause)
			(with-keyword-pairs ((report interactive &rest forms)
					     (CDDR clause))
			  (LIST (CAR clause)			   ;Name=0
				(GENSYM)			   ;Tag=1
				(TRANSFORM-KEYWORDS :REPORT report ;Keywords=2
						    :INTERACTIVE interactive)
				(CADR clause)			   ;BVL=3
				forms)))			   ;Body=4
		    clauses)))
      `(BLOCK ,block-tag
	 (LET ((,temp-var nil))
	   (TAGBODY
	     (restart-bind
	       ,(MAPCAR #'(LAMBDA (datum)
			    (LET ((name (NTH 0 datum))
				  (tag  (NTH 1 datum))
				  (keys (NTH 2 datum)))
			      `(,name #'(LAMBDA (&rest temp)
					  #+LISPM (SETQ temp (COPY-LIST temp))
					  (SETQ ,temp-var temp)
					  (GO ,tag))
				,@keys)))
			data)
	       (RETURN-FROM ,block-tag ,expression))
	     ,@(MAPCAN #'(LAMBDA (datum)
			   (LET ((tag  (NTH 1 datum))
				 (bvl  (NTH 3 datum))
				 (body (NTH 4 datum)))
			     (LIST tag
				   `(RETURN-FROM ,block-tag
				      (APPLY #'(LAMBDA ,bvl ,@body)
					     ,temp-var)))))
		       data)))))))


(DEFMACRO with-simple-restart ((name format-string &rest format-arguments)
			       &body forms)
  "  WITH-SIMPLE-RESTART executes FORMS with a restart established to
transfer control back to WITH-SIMPLE-RESTART.
  The NAME may be NIL to indicate an anonymous restart, or some other
symbol to indicate a named restart.  FORMAT-STRING and FORMAT-ARGUMENTS
are used to form the report message for this restart.
  If NAME is invoked while executing FORMS, control is transferred back
to WITH-SIMPLE-RESTART and WITH-SIMPLE-RESTART immediately returns two
values:  NIL and T.  If RESTART-NAME is not invoked while executing
FORMS, WITH-SIMPLE-RESTART returns the values of the last form in FORMS.
"
  `(restart-case (PROGN ,@forms)
     (,name ()
        :REPORT (LAMBDA (stream)
		  (FORMAT stream ,format-string ,@format-arguments))
      (VALUES nil t))))


;;; Finding restarts

;;; Note that:
;;; COMPUTE-RESTART depends on the variable, EH:*CONDITION-RESUME-HANDLERS*.
;;; FIND-RESTART calls EH:FIND-RESUME-HANDLER.

;;; For a portable Common LISP condition system, COMPUTE-RESTART and
;;; FIND-RESTART would need to be rewritten to not depend on the Explorer
;;; error system for finding resume handlers.

;;; INVOKE-RESTART calls FIND-RESTART and is portable
;;; INVOKE-RESTART-INTERACTIVELY calls FIND-RESTART and is portable.

(DEFUN compute-restarts (&aux restarts)
  "  COMPUTE-RESTARTS uses the dynamic state of the program to compute a
list of the restarts which are currently active.
  Each restart represents a function which can be called to perform some
form of recovery action, usually a transfer of control to an outer point
in the running program.  The list of restart returned are implemented as
defstruct objects and are ordered so that the innermost (more-recently
established) restarts are nearer the head of the list.
  COMPUTE-RESTARTS returns all anonymous and named restarts which are
active.  COMPUTE-RESTARTS is used to find restarts which FIND-RESTART
cannot.  FIND-RESTART will never find anonymous restarts and will only
return the first restart with the desired restart name.
"
  (DOLIST (rh eh:*condition-resume-handlers*)
    (COND ((AND (CONSP rh) (TYPEP (CAR rh) 'restart))
	   (DOLIST (restart rh)
	     (PUSH restart restarts)))
	  ((CONSP rh)
	   (push (convert-resume-handler-to-restart rh) restarts))))
  (reverse restarts))

(DEFUN convert-resume-handler-to-restart (handler)
  "Converts Explorer resume handler:
 (CONDITION-NAMES PROCEED-TYPE PREDICATE (FORMAT-STRING FORMAT-ARGS...)
  FUNCTION EXTRA-ARGS...)
to a Common LISP restart"
  (MAKE-RESTART
   :NAME (COND ((CONSP (SECOND handler))  
		;; Either use "ABORT" as restart name if resume handler is for abort condition
		;; or use NIL (anonymous restart)
		(COND ((AND (CONSP (first handler)) (MEMBER 'sys:abort (first handler)))
		       'ABORT)
		      ((EQ (first handler) 'sys:abort)
		       'ABORT)
		      (:else
		       nil)))
	       (:else   ;; use proceed-type as restart name
		(SECOND handler)))
   :FUNCTION #'(LAMBDA (condition) (FUNCALL (FIFTH handler) condition (NTHCDR 5 handler)))
   :REPORT-FUNCTION #'(LAMBDA (stream) (APPLY #'FORMAT stream (FOURTH handler)))
   :PREDICATE (THIRD handler)
   :CONDITION-NAMES (FIRST handler)))

(DEFUN find-restart (restart)
  "  FIND-RESTART searches for the particular RESTART in the current
dynamic environment.
  If RESTART is a symbol, then the most recently established restart
with that name is returned.  FIND-RESTART returns NIL if no such restart
is found.
  If RESTART is a restart object and this restart is active, then it is
returned.  If the RESTART is not active, FIND-RESTART returns NIL.
  It is an error for RESTART to be the symbol NIL.  Anonymous restarts
(restarts with the name of NIL) can only be found by using
COMPUTE-RESTARTS.
"
  (DO ()
      ((NOT (NULL restart)))
    (restart-case (error "Restart name should be non-NIL.")
      (CONTINUE (value)
	  :REPORT "Supply a named restart."
	  :INTERACTIVE (LAMBDA () 
			 (LIST (read-evaluated-form "Type a form to be evaluated and used as the restart: ")))
	(setq restart value))))
  (LET ((r (EH:FIND-RESUME-HANDLER 'condition restart)))
    (IF (CONSP r)
	(convert-resume-handler-to-restart r)
	r)))
  
(DEFUN invoke-restart (restart &rest arguments)
  "  INVOKE-RESTART calls the function associated with the given RESTART,
passing any given ARGUMENTS.  The RESTART must be a restart object or
the non-NIL name of a restart which is active in the current dynamic
context.  An error is signalled if the ARGUMENTS are not valid for this
RESTART.
"
  (DO ()
      ((NOT (NULL restart)))
    (restart-case (error "Restart name should be non-NIL.")
      (CONTINUE (value)
		:REPORT "Supply a named restart."
		:INTERACTIVE 
		(LAMBDA () 
		  (LIST (read-evaluated-form "Type a form to be evaluated and used as the restart: ")))
		(setq restart value))))

  (DO ((real-restart (find-restart restart) (find-restart restart)))
      ((NOT (NULL real-restart))
       (APPLY (restart-function real-restart) arguments))
    (restart-case (error "Restart ~S is not active." restart)
      (CONTINUE (value)
	  :REPORT "Supply an active restart."
	  :INTERACTIVE 
	  (LAMBDA () 
	    (LIST (read-evaluated-form "Type a form to be evaluated and used as the restart: ")))
	(setq restart value)))))

(DEFUN invoke-restart-interactively (restart)
  "  INVOKE-RESTART-INTERACTIVELY first calls the interactive function
associated with the given RESTART to get the necessary arguments.  It
then calls the function associated with the given RESTART, passing
any of the arguments genereated interactively.
  The RESTART must be a restart object or the non-NIL name of a restart
which is active in the current dynamic context.  An error is signalled
if any arguments aquired by the interactive-function are not valid for
this restart.
  INVOKE-RESTART-INTERACTIVELY will prompt for arguments by executing
the code provided in the :INTERACTIVE keyword to RESTART-CASE or
:INTERACTIVE-FUNCTION keyword to RESTART-BIND.
  If no :INTERACTIVE option has been supplied in the corresponding
RESTART-BIND or RESTART-CASE, then the argument list of nil
will be used when invoking the restart function.
  This operation is used internally by the debugger.  It may also be
used in a condition handler or in implementing other portable,
interactive debugging tools.
"
  (DO ()
      ((NOT (NULL restart)))
    (restart-case (error "Restart name should be non-NIL.")
      (CONTINUE (value)
		:REPORT "Supply a named restart."
		:INTERACTIVE 
		(LAMBDA () 
		  (LIST (read-evaluated-form "Type a form to be evaluated and used as the restart: ")))
		(setq restart value))))

  (DO ((real-restart (find-restart restart) (find-restart restart)))
      ((NOT (NULL real-restart))
       (APPLY (restart-function real-restart) 
	      (LET ((interactive-function (restart-interactive-function real-restart)))
		(IF interactive-function
		    (FUNCALL interactive-function)
		    '()))))
    (restart-case (error "Restart ~S is not active." restart)
      (CONTINUE (value)
	  :REPORT "Supply an active restart."
	  :INTERACTIVE 
	  (LAMBDA () 
	    (LIST (read-evaluated-form "Type a form to be evaluated and used as the restart: ")))
	(setq restart value)))))

;;; Restart Predicates used by Explorer error handler

(defun restartp (type restart)
  "Returns non-nil, if restart of specified type"
  (and (typep restart 'restart)
       (eq (intern (restart-name restart) :keyword) 
	   (intern type :keyword))))

(defun restartp-or-eq (restart type)
  "Returns non-NIL, if a restart of specified type
or if restart simply EQs type."
  (or (restartp type restart)
      (eq type restart)))


;;; Pre-defined named restarts

(DEFUN abort ()
  "  The ABORT function transfers control to the restart named ABORT.  If
no such restart exists, ABORT signals an error of type CONTROL-ERROR.
The purpose of the ABORT restart is generally to allow return to the
innermost command loop level.
"
  (invoke-restart 'abort)
  (error 'abort-failure))

(DEFUN continue ()
  "  The CONTINUE function transfers control to the restart named
CONTINUE.  If no such restart exists, CONTINUE returns NIL.  The
CONTINUE restart is a general purpose restart which provides a single
and obvious way to continue.  This restart is set up in CERROR.
"
  (invoke-restart 'continue))

(DEFUN muffle-warning ()
  "  The MUFFLE-WARNING function transfers control to the restart named
MUFFLE-WARNING.  If no such restart exists, MUFFLE-WARNING signals an
error of type CONTROL-ERROR.  The MUFFLE-WARNING restart is set up by
WARN so that handlers of warning conditions can cancel the warning
message.
"
  (invoke-restart 'muffle-warning))

(DEFUN store-value (value)
  "  The STORE-VALUE function transfers control (and one value) to the
restart named STORE-VALUE.  If no such restart exists, STORE-VALUE
returns NIL.  The STORE-VALUE restart is used by handlers trying to
recover from errors of types such as CELL-ERROR or TYPE-ERROR, where the
handler may wish to supply a replacement datum to be stored permanently.
"
  (invoke-restart 'store-value value))

(DEFUN use-value (value)
  "  The USE-VALUE function transfers control (and one value) to the
restart named USE-VALUE.  If no such restart exists, USE-VALUE returns
NIL.  The USE-VALUE restart is generally used by handlers trying to
recover from errors of types such as CELL-ERROR, where the handler may
wish to supply a replacement datum for one-time use.
"
  (invoke-restart 'use-value value))
