;;; -*- Mode:Common-Lisp; Package:CONDITIONS; Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (c) 1989 Texas Instruments Incorporated.  All Rights Reserved.

;;; Common LISP CONDITION System 

;;; This file contains functions and macros which follow the Common LISP
;;; standard on signalling conditions and invoking the debugger.

;;; The Explorer version of the Common LISP Condition System exists in its
;;; own package, CONDITIONS, aka CLEH.

;;; This is not a portable version of the Common LISP condition/error
;;; system.  Rather it is an implementation which depends on and interfaces
;;; with the existing Explorer error system.  It allows both the Common LISP
;;; error system and the Explorer error system to co-exist.  The Explorer
;;; error handler as been changed to accommodate features in the Common LISP
;;; error system.

;;;Record of changes:
;;; 01/18/89   clm   Changed the error message in CHECK-TYPE to be less confusing.


;;; Invoking the Debugger

;;; Note that:
;;; INVOKE-DEBUGGER calls STANDARD-DEBUGGER.
;;; STANDARD-DEBUGGER calls EH:INVOKE-DEBUGGER.
;;; For a portable Common LISP condition system, STANDARD-DEBUGGER would need to be
;;; rewritten to not depend on the Explorer error system for invoking a debugger.

(DEFVAR *debugger-hook* nil
  "  This variable should hold either NIL or a function of a two arguments,
a condition and the value of *DEBUGGER-HOOK*.  This function may either
handle the condition (transfer control) or return normally (allowing the
standard debugger to run).
")

(DEFUN invoke-debugger (&OPTIONAL (datum "Debug") &rest arguments)
  "  INVOKE-DEBUGGER invokes the standard debugger passing a condition.
It first creates a condition which is constructed from the
INVOKE-DEBUGGER arguments.  When the debugger is entered, the error
message for the condition will be displayed and a description of any
active restarts will also be displayed.  The standard debugger will
never directly return.  Return can occur only if a user selects a
debugger command which will transfer control, such as invoking a
restart.
  If DATUM is a condition type, then ARGUMENTS should be keyword/value
pairs and the condition is created by doing:
 (APPLY #'MAKE-CONDITION DATUM ARGUMENTS).
  If DATUM is a condition object, then this condition is passed directly
to the debugger In this case, it is an error for ARGUMENTS is non-NIL.
  If DATUM is a format string, then the condition is created by doing:
  (MAKE-CONDITION 'SIMPLE-ERROR :FORMAT-STRING DATUM
                                :FORMAT-ARGUMENTS ARGUMENTS).
  If DATUM is omitted, then the condition is created with a
format string of \"Debug\".
  If the variable *DEBUGGER-HOOK* is bound to a function, it will be
called before entering the standard debugger.  This hook function is
called with two arguments, the condition being handled and the value of
*DEBUGGER-HOOK*.  If this hook function returns, the standard debugger
will be entered.
  INVOKE-DEBUGGER is called by signalling functions, ERROR and CERROR,
when no handlers are found for CONDITION.
"
  (LET ((condition (coerce-to-condition datum arguments 'SIMPLE-CONDITION 'DEBUG)))
    (WHEN *debugger-hook*
      (LET ((hook *debugger-hook*)
	    (*debugger-hook* nil))
	(FUNCALL hook condition hook)))
    (standard-debugger condition)))

(DEFUN standard-debugger (condition)
  (LET ((eh:*error-depth* (1+ eh:*error-depth*)))
    (DECLARE (SPECIAL eh:*error-depth*))
    (LET ((restart-request (FUNCALL eh:%error-handler-stack-group condition)))
      (APPLY #'EH:INVOKE-RESUME-HANDLER condition restart-request))
    nil))

;;; Breakpoint utility
;;; Note that the function BREAK calls INVOKE-DEBUGGER and WITH-SIMPLE-RESTART.
;;; BREAK is portable.

(DEFUN break (&optional (format-string "Break.") &rest format-arguments)
  "    BREAK enters the debugger without allowing any possibility of
interception by condition handlers.  The debugger will print a message
using FORMAT-STRING and FORMAT-ARGUMENTS.  It is possible to continue
with program execution and return from BREAK by invoking the CONTINUE
restart.  This restart is available as a resume command in the debugger.
If the CONTINUE restart is invoked, BREAK returns NIL.
  If no FORMAT-STRING is supplied, the string \"Break\" is used instead.
  BREAK should be used as a way of inserting temporary debugging
breakpoints in a program, not as a way of signalling errors.
"
  (with-simple-restart (CONTINUE "Return from BREAK.")
    (invoke-debugger (make-condition 'simple-condition
				     :FORMAT-STRING format-string
				     :FORMAT-ARGUMENTS format-arguments)))
  nil)


;;; Signalling functions

;;; Note that:

;;; SIGNAL calls EH:INVOKE-HANDLERS.
;;; For a portable Common LISP condition system, SIGNAL would need to be
;;; rewritten to not depend on the Explorer error system for invoking
;;; condition handlers.

;;; ERROR calls SIGNAL and INVOKE-DEBUGGER; and is portable.
;;; CERROR calls SIGNAL, INVOKE-DEBUGGER and WITH-SIMPLE-RESTART; and is portable.


(DEFUN coerce-to-condition (datum arguments default-type function-name)
  "  Internal routine used in SIGNAL, ERROR, CERROR, BREAK, and WARN for
parsing the hairy argument conventions into a single argument that's
directly usable by all the other routines.
"
  (SETQ arguments (COPY-LIST arguments))
  (COND ((TYPEP datum 'condition)
	 (IF arguments
	     (cerror "Ignore the additional arguments."
		     'simple-type-error
		     :DATUM arguments
		     :EXPECTED-TYPE 'null
		     :FORMAT-STRING "You may not supply additional arguments ~
				     when giving ~S to ~S."
		     :FORMAT-ARGUMENTS (LIST datum function-name)))
	 datum)
        ((SYMBOLP datum)                  ;roughly, (SUBTYPEP datum 'CONDITION)
         (APPLY #'make-condition datum arguments))
        ((STRINGP datum)
	 (MAKE-CONDITION default-type
                         :FORMAT-STRING datum
                         :FORMAT-ARGUMENTS arguments))
        (T
         (error 'simple-type-error
		:DATUM datum
		:EXPECTED-TYPE '(OR symbol string)
		:FORMAT-STRING "Bad argument to ~S: ~S"
		:FORMAT-ARGUMENTS (LIST function-name datum)))))

(DEFVAR *break-on-signals* nil
  "  This flag is used for debugging programs that signal conditions.  When
(TYPEP condition *BREAK-ON-SIGNALS*) is true, then calls to SIGNAL will
enter the debugger prior to signalling that condition.  The CONTINUE
restart may be used to continue with the normal signalling process.
This is also true for other signalling functions and macros such as
ERROR, CERROR, ASSERT and CHECK-TYPE.

  By default the value of this variable is NIL.  IF *BREAK-ON-SIGNALS*
is non-NIL, it should be set to a valid condition type specifier.
")

(DEFUN signal (datum &rest arguments)
  "  SIGNAL will signal a condition.  It first creates a condition
object which is constructed from the SIGNAL arguments.  It then searches
for a condition handler to invoke on this condition.  If no handler is
found to handle this condition, SIGNAL returns NIL.  The debugger is
never invoked.
  If DATUM is a condition type, then ARGUMENTS should be keyword/value
pairs and the condition to be signalled is created by doing:  
 (APPLY #'MAKE-CONDITION DATUM ARGUMENTS).
  If DATUM is a condition object, then this condition is used directly
for signalling.  In this case, it is an error for ARGUMENTS to be
non-NIL.
  If DATUM is a format string, then the condition to be signalled is
created by doing: (MAKE-CONDITION 'SIMPLE-CONDITION
               	                  :FORMAT-STRING DATUM
                                  :FORMAT-ARGUMENTS ARGUMENTS).

"
  (LET ((condition (coerce-to-condition datum arguments 'simple-condition 'signal))
        (eh:*condition-handlers* eh:*condition-handlers*))
    (IF (TYPEP condition *break-on-signals*)
	(BREAK "~A ~%Break entered because of *BREAK-ON-SIGNALS*."
	       condition))
    (LET ((restart-request (EH:INVOKE-HANDLERS condition)))
      (IF restart-request 
	(APPLY #'EH:INVOKE-RESUME-HANDLER condition restart-request)))
    nil))

(DEFUN error (datum &rest arguments)
  "  ERROR signals a condition.  It first creates a condition object
which is constructed from the DATUM.  It then searches for a
condition handler to invoke on this condition.  If no handler is found
to handle this condition, the debugger will be entered and the error
message for this condition will be displayed.  A call to ERROR cannot
directly return.  There is no option to simply continue.  The only exit
from this function can come by non-local transfer of control in a
handler or by using one of the interactive commands in the debugger.
  If DATUM is a condition type, then ARGUMENTS should be keyword/value
pairs and the condition to be signalled is created by doing:
 (APPLY #'MAKE-CONDITION DATUM ARGUMENTS).
  If DATUM is a condition object, then this condition is used directly
for signalling.  In this case, it is an error for ARGUMENTS to be non-NIL.
  If DATUM is a format string, then the condition to be signalled is
created by doing: (MAKE-CONDITION 'SIMPLE-ERROR 
                                  :FORMAT-STRING DATUM 
                                  :FORMAT-ARGUMENTS ARGUMENTS).
"
  (LET ((condition (coerce-to-condition datum arguments 'simple-error 'error)))
    (signal condition)
    (invoke-debugger condition)))

(DEFUN cerror (continue-format-string datum &rest arguments)
  "  CERROR signals a condition which is continuable.  It first creates
a condition object which is constructed from the arguments after the
CONTINUE-FORMAT-STRING.  It then searches for a condition handler to
invoke on this condition.  If no handler is found to handle this
condition, the debugger will be entered and the error message for this
condition will be displayed.  It is possible to continue with program
execution and return from CERROR by invoking the CONTINUE restart.  The
CONTINUE restart can be invoked directly by a handler or interactively
by selecting the appropiate resume command in the debugger.  If the
CONTINUE restart is invoked, the value returned by CERROR is NIL.
  The CONTINUE-FORMAT-STRING and the ARGUMENTS are used to form the
report message for the CONTINUE restart.  The debugger will display this
message when describing a resume command.
  If DATUM is a condition type, then ARGUMENTS should be keyword/value
pairs and the condition to be signalled is created by doing:  
 (APPLY #'MAKE-CONDITION DATUM ARGUMENTS).
  If DATUM is a condition object, then this condition is used directly
for signalling.  In this case, ARGUMENTS will be used only with
the CONTINUE-FORMAT-STRING.
  If DATUM is a format string, then the condition to be signalled is
created by doing:  (MAKE-CONDITION 'SIMPLE-ERROR 
                                   :FORMAT-STRING DATUM
                                   :FORMAT-ARGUMENTS ARGUMENTS).
  Note that if DATUM is a string, then the ARGUMENTS will be used to
construct both an error message and a continue message string.  Also, if
DATUM is a condition type then the ARGUMENTS will be used not only to
initialize the condition but also to construct the continue message
string.  In both of these cases, some care may be taken to set up the
CONTINUE-FORMAT-STRING correctly.  The FORMAT option, ~*, may be
particularly useful in this situation.
"
  (LET ((condition (coerce-to-condition datum 
					(UNLESS (TYPEP datum 'condition) arguments)
					'simple-error 'cerror)))
    (with-simple-restart (CONTINUE "~A" (APPLY #'FORMAT nil continue-format-string arguments))
      (PROGN
       (signal condition)
       (invoke-debugger condition))))
  nil)



;;;  ACCUMULATE-CASES

(EVAL-WHEN (eval compile load)

(DEFUN accumulate-cases (macro-name cases list-is-atom-p)
  "Internal function used by ECASE, CCASE, ETYPECASE, and CTYPECASE.
"
  (DO ((l '())
       (c cases (CDR c)))
      ((NULL c) (NREVERSE l))
    (LET ((keys (CAAR c)))
      (COND ((ATOM keys)
	     (COND ((NULL keys))
		   ((MEMBER keys '(otherwise t))
		    (error "OTHERWISE (or T) clause is not allowed in ~S expressions."
			   macro-name))
		   (t (PUSH keys l))))
	    (list-is-atom-p
	     (PUSH keys l))
	    (t
	     (DOLIST (key keys) (PUSH key l)))))))

);NEHW-LAVE


;;; Error case and typecase 

;;; Note that:

;;; ECASE is built on ERROR and is portable.
;;; CCASE is built on ERROR and RESTART-CASE; and is portable.
;;; ETYPECASE is built on ERROR and is portable.
;;; CTYPECASE is built on ERROR and RESTART-CASE; and is portable.


(DEFMACRO ecase (test-object &rest cases)
  "  The name of this macro stands for error-checking TYPECASE.  ETYPECASE
is the same as TYPECASE, except that no OTHERWISE or T clause is
permitted.  ETYPECASE uses TYPECASE with an OTHERWISE clause containing
a call to ERROR.
  Each CASE clause has the form: (TEST-FORM FORMS).
  ECASE executes the FORMS of the first clause whose TEST-FORM matches the
value of TEST-OBJECT.  ECASE returns the values of the last form of the
clause which TEST-OBJECT satisfies.  If no clause is satisfied, ECASE
signals an error of type TYPE-ERROR.  If no handler handles this
condition, the debugger is entered and the error message displayed will
list the TEST-FORMs which did not match TEST-OBJECT.  There is no way to 
continue from this error condition.
"
  (LET ((test-forms (accumulate-cases 'ecase cases nil))
	(var (GENSYM)))
    `(LET ((,var ,test-object))
       (CASE ,var
	 ,@cases
	 (otherwise
	   (error 'case-failure :NAME 'ecase
		   		:TEST-OBJECT ',test-object
		  		:DATUM ,var
				:EXPECTED-TYPE '(MEMBER ,@test-forms)
				:POSSIBILITIES ',test-forms))))))

(DEFMACRO ccase (test-object &rest cases)
  "  The name of this macro stands for continuable CASE.  CCASE is the same
as CASE, except that no explicit OTHERWISE or T clause is permitted.
CCASE uses CASE with an OTHERWISE clause containing a call to ERROR and
a STORE-VALUE restart.
  Each CASE clause has the form: (TEST-FORM FORMS).
  CCASE executes the FORMS of the first clause whose TEST-FORM matches the
value of TEST-OBJECT.  CCASE returns the values of the last form of the
clause which TEST-OBJECT satisfies.  If no clause is satisfied, CCASE
signals an error of type TYPE-ERROR.  If no handler handles this
condition, the debugger is entered and the error message displayed will
list the TEST-FORMS which did not match TEST-OBJECT.  This error may be 
continued using the STORE-VALUE restart.
  If an error condition is signalled, CCASE can return only if the
STORE-VALUE restart is invoked, either explicitly from a handler (by
calling STORE-VALUE) or implicitly as a command in the debugger (with
the new value prompted for interactively).
  The TEST-OBJECT must be a generalized variable reference acceptable to
SETF.  If the STORE-VALUE restart is invoked, CCASE will set TEST-OBJECT to
the new value and start over, signalling another error if no clause is
found with a TEST-FORM matching the value of TEST-OBJECT.  
"
  (LET ((test-forms (accumulate-cases 'ccase cases nil))
	(tag1 (GENSYM))
	(tag2 (GENSYM)))
    `(BLOCK ,tag1
       (TAGBODY ,tag2
	 (RETURN-FROM ,tag1
	   (CASE ,test-object
	     ,@cases
	     (otherwise
	       (restart-case (error 'case-failure
				    :NAME 'ccase
				    :TEST-OBJECT ',test-object
				    :DATUM ,test-object
				    :EXPECTED-TYPE '(MEMBER ,@test-forms)
				    :POSSIBILITIES ',test-forms)
		 (STORE-VALUE (value)
		     :REPORT (LAMBDA (stream)
			       (FORMAT stream "Supply a new value for ~S."
				       ',test-object))
		     :INTERACTIVE (LAMBDA () (LIST (read-evaluated-form)))
		   (SETF ,test-object value)
		   (GO ,tag2))))))))))

(DEFMACRO etypecase (test-object &rest cases)
  "  The name of this macro stands for error-checking TYPECASE.  ETYPECASE
is the same as TYPECASE, except that no OTHERWISE or T clause is
permitted.  ETYPECASE uses TYPECASE with an OTHERWISE clause containing
a call to ERROR.
  Each CASE form has the form: (TYPESPEC FORMS).
  ETYPECASE executes the FORMS of the first clause where (TYPEP TEST-OBJECT
TYPESPEC) is true.  ETYPECASE returns the values of the last form of the
clause with the desired TYPESPEC.  If no clause is satisfied, ETYPECASE
signals an error condition of type TYPE-ERROR.  If no handler handles
this condition, the debugger is entered and the error message displayed
will list the TYPESPECS which TEST-OBJECT did not satisfy.  There is no way
to continue from this error condition.
"
  (LET ((types (accumulate-cases 'etypecase cases t))
	(var (GENSYM)))
    `(LET ((,var ,test-object))
       (TYPECASE ,var
	 ,@cases
	 (otherwise
	   (error 'case-failure :NAME 'etypecase
		  		:TEST-OBJECT ',test-object
		  		:DATUM ,var
				:EXPECTED-TYPE '(OR ,@types)
				:POSSIBILITIES ',types))))))

(DEFMACRO ctypecase (test-object &rest cases)
  "  The name of this macro stands for continuable TYPECASE.  CTYPECASE is
the same as TYPECASE, except that no otherwise or T clause is permitted.
CTYPECASE uses TYPECASE with an OTHERWISE clause containing a call to
ERROR and a STORE-VALUE restart.
  Each CASE form has the form: (TYPESPEC FORMS).
  CTYPECASE executes the FORMS of the first clause where (TYPEP TEST-OBJECT
TYPESPEC) is true.  CTYPECASE returns the values of the last form of the
clause with the desired TYPESPEC.  If no clause is satisfied, CTYPECASE
signals an error condition of type TYPE-ERROR.  If no handler handlers
this condition, the debugger is entered and the error message displayed
will list the TYPESPECS which TEST-OBJECT did not satisfy.  This error
condition may be continued using the STORE-VALUE restart.
  If an error condition is signalled, CTYPECASE can return only if the
STORE-VALUE restart is invoked, either explicitly from a handler (by
calling STORE-VALUE) or implicitly as a command in the debugger (with
the new value prompted for interactively).
  The TEST-OBJECT must be a generalized variable reference acceptable to
SETF.  If the STORE-VALUE restart is invoked, CTYPECASE will set
TEST-OBJECT to the new value and start over, signalling another error if no
clause is found with a desired TYPESPEC.  value.
"
  (LET ((types (accumulate-cases 'ctypecase cases t))
	(tag1 (GENSYM))
	(tag2 (GENSYM)))
    `(BLOCK ,tag1
       (TAGBODY ,tag2
	 (RETURN-FROM ,tag1
	   (TYPECASE ,test-object
	     ,@cases
	     (OTHERWISE
	       (restart-case (error 'case-failure
				    :NAME 'ctypecase
				    :TEST-OBJECT ',test-object
				    :DATUM ,test-object
				    :EXPECTED-TYPE '(OR ,@types)
				    :POSSIBILITIES ',types)
		 (STORE-VALUE (value)
		     :REPORT (LAMBDA (stream)
			       (FORMAT stream "Supply a new value for ~S."
				       ',test-object))
		     :INTERACTIVE (LAMBDA () (LIST (read-evaluated-form)))
		   (SETF ,test-object value)
		   (GO ,tag2))))))))))


;;; Assert

;;; Note that the macro ASSERT is built on ERROR and RESTART-CASE.
;;; ASSERT is portable.

(DEFUN assert-report (names stream)
  (FORMAT stream "Retry assertion")
  (IF names
      (FORMAT stream " with new value~P for ~{~S~^, ~}."
	      (LENGTH names) names)
      (FORMAT stream ".")))

(DEFUN assert-prompt (name value)
  (COND ((Y-OR-N-P "The old value of ~S is ~S.~
		  ~%Do you want to supply a new value? "
		   name value)
	 (FORMAT *query-io* "~&Type a form to be evaluated: ")
	 (FLET ((read-it () (EVAL (READ *query-io*))))
	   (IF (SYMBOLP name) ;Help user debug lexical variables
	       (PROGV (LIST name) (LIST value) (read-it))
	       (read-it))))
	(t value)))

(DEFUN simple-assertion-failure (assertion)
  (error 'simple-error  
;	 :DATUM assertion   ;; if use SIMPLE-TYPE-ERROR
;	 :EXPECTED-TYPE nil			
	 :FORMAT-STRING "The assertion ~S failed."
	 :FORMAT-ARGUMENTS (LIST assertion)))

(DEFMACRO assert (test-form &optional places datum &rest arguments)
  "  ASSERT signals a continuable condition if the value of the TEST-FORM
is NIL.  The condition signalled is constructed from the arguments after
the TEST-FORM and PLACES.  ASSERT returns nil if TEST-FORM is non-NIL.
  The TEST-FORM is any form.  PLACES should be a list of variables on
which TEST-FORM can depend on.  TEST-FORM will usually check the values
of all or some of the variables in PLACES.  Each variable in PLACES is a
generalized-variable reference acceptable to SETF.  PLACES can also be
omitted or NIL.
  If a condition is signalled, ASSERT can return only if the CONTINUE
restart is invoked, either explicitly from a handler (by calling
CONTINUE) or implicitly as a command in the debugger.  The CONTINUE
restart allows the user to alter the values of of the variables listed
in PLACES.  If the CONTINUE restart is invoked, ASSERT will
interactively set the variables in PLACES to their new values and start
over; re-evaluating TEST-FORM and signalling another error if TEST-FORM
is still NIL.  Note that if the PLACES argument is NIL, the CONTINUE
restart will simply transfer control to where ASSERT re-evaluates
TEST-FORM.
  DATUM and ARGUMENTS are used to create an error condition and are
evaluated only if an error is to be signalled, and re-evaluated if the
error is to be signalled again.
  If DATUM is a condition type, then ARGUMENTS should be keyword/value
pairs and the condition used for signalling is the result of doing:
 (APPLY #'MAKE-CONDITION datum arguments).
  If DATUM is a condition object, then this condition is used directly
for signalling.  In this case, it is an error for ARGUMENTS to be
non-NIL.
  If DATUM is a format string, then the condition used for signalling is
the result of doing: (MAKE-CONDITION 'SIMPLE-ERROR :FORMAT-STRING datum
        	                                   :FORMAT-ARGUMENTS arguments).
  If DATUM and ARGUMENTS are omitted, then a condition of type
SIMPLE-ERROR is constructed using the TEST-FORM in the error report
message:   (MAKE-CONDITION 'SIMPLE-ERROR :FORMAT-STRING \"The assertion ~S failed.\"
		                         :FORMAT-ARGUMENTS '(TEST-FORM))

"
  (LET ((tag (GENSYM)))
    `(TAGBODY ,tag
       (UNLESS ,test-form
	 (restart-case ,(IF datum
			    `(error ,datum ,@arguments)
			    `(simple-assertion-failure ',test-form))
	   (CONTINUE ()
		 :REPORT (LAMBDA (stream) (assert-report ',places stream))
	     ,@(MAPCAR #'(LAMBDA (place)
			   `(SETF ,place (assert-prompt ',place ,place)))
		       places)
             (GO ,tag)))))))


;;; Type Checking

;;; Note that the macro CHECK-TYPE is built on ERROR and RESTART-CASE.
;;; CHECK-TYPE is portable.

(DEFMACRO check-type (place typespec &optional type-string)
  "  CHECK-TYPE signals a condition of type TYPE-ERROR if (TYPEP PLACE
TYPESPEC) is true.  CHECK-TYPE returns NIL if the contents of PLACE is
of the desired TYPESPEC.
  If a condition is signalled, handlers of this condition can use the
functions TYPE-ERROR-OBJECT and TYPE-ERROR-EXPECTED-TYPE to access the
contents of PLACE and the TYPESPEC, respectively.  If no handler is
found to handle this condition, the debugger is entered and the error
message displayed will mention PLACE, its contents, and the desired
TYPESPEC.
  If a condition is signalled, CHECK-TYPE can return only if the
STORE-VALUE restart is invoked, either explicitly from a handler (by
calling STORE-VALUE) or implicitly as a command in the debugger (with
the new value prompted for interactively).  If the STORE-VALUE restart
is invoked, CHECK-TYPE will set PLACE to the new value and start over,
checking the type of the new value of PLACE and signalling another error
if it is still not the desired type.

  The PLACE must be a generalized-variable reference acceptable to SETF.
The TYPESPEC must be a type specifier and it is not evaluated.  The
TYPE-STRING is evaluated and should be an English description of the
type (such as \"an integer\").  If the TYPE-STRING is not supplied, it is
generated from the TYPESPEC.

"
  (LET ((tag1 (GENSYM))
	(tag2 (GENSYM)))
    `(BLOCK ,tag1
       (TAGBODY ,tag2
	 (IF (TYPEP ,place ',typespec) (RETURN-FROM ,tag1 nil))
	 (restart-case (error 'type-error
			      :FORMAT-STRING "The value of ~S is ~S, ~
				     which is not of type ~A."             ;; clm 01/18/89
			      :FORMAT-ARGS (LIST ',place ,place 
						 ,(IF type-string 
						      type-string
						      `',typespec))
			      :DATUM ,place 
			      :EXPECTED-TYPE ',typespec)

	   (STORE-VALUE (value)
	       :REPORT (LAMBDA (stream)
			 (FORMAT stream "Supply a new value for ~S."
				 ',place))
	       :INTERACTIVE (LAMBDA () (LIST (read-evaluated-form)))
	     (SETF ,place value)
	     (GO ,tag2)))))))


;;; Warning

;;; Note that the function WARN calls BREAK and RESTART-CASE.
;;; WARN is portable.

(DEFVAR *break-on-warnings* nil
  "  If *BREAK-ON-WARNINGS* is true, then the function WARN will invoke the
debugger before signalling the WARNING condition.  If the CONTINUE
restart is used, WARN continues normally with the warning process.
  This variable is intended primarily for use when the user is debugging
programs that issue warnings.  The default value of *BREAK-ON-WARNINGS*
is NIL.
  Note that this flag is still supported but is considered obsolete.
*BREAK-ON-SIGNALS* should be used instead.
")

(DEFUN warn (datum &rest arguments)
  "  WARN will signal a condition of type WARNING.  It first searches
for a handler to invoke for this condition.  If the warning condition is
not handled, WARN will report a warning message for this condition to
*ERROR-OUTPUT* and returns NIL.  The debugger is never invoked.
  While the WARNING condition is being signalled, the MUFFLE-WARNING
restart is established for use by a handler to immediately cause WARN to
return NIL and not report a warning.
  If DATUM is a condition type, then ARGUMENTS should be keyword/value
pairs and the condition to be signalled is created by doing
a (APPLY #'MAKE-CONDITION DATUM ARGUMENTS).  The resulting condition
must be of type WARNING or an error of type TYPE-ERROR is signalled.
  If DATUM is a condition object, then that condition is used directly
for signalling.  In this case, it is an error for ARGUMENTS to be
non-NIL.  If the condition is not of type WARNING an error of type
TYPE-ERROR is signalled.
  If DATUM is a format string, then the condition to be
signalled is created by doing: (MAKE-CONDITION 'SIMPLE-WARNING 
        	                               :FORMAT-STRING DATUM
        	                               :FORMAT-ARGUMENTS ARGUMENTS).
"
  (LET ((condition (coerce-to-condition datum arguments 'SIMPLE-WARNING 'WARN)))
    (check-type condition WARNING "a warning condition")
    (IF *break-on-warnings*
	(BREAK "~A ~%Break entered because of *BREAK-ON-WARNINGS*."
	       condition))
    (restart-case (signal condition)
      (MUFFLE-WARNING ()
	  :REPORT "Skip warning."
	(RETURN-FROM warn nil)))
    (FORMAT *error-output* "~&Warning: ~A~%" condition)
    nil))


;;; The Explorer Error Handler needs to skip these functions:

(DEFPROP invoke-debugger t :error-reporter)
(DEFPROP standard-debugger t :error-reporter)
(DEFPROP break t :error-reporter)
(DEFPROP signal t :error-reporter)
(DEFPROP cerror t :error-reporter)
(DEFPROP error t :error-reporter)
(DEFPROP simple-assertion-failure t :error-reporter)
