;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

(defgeneric make-instances-obsolete (class)
  (:documentation 
 "This generic function is called when a class has been redefined such that the
list of slot names is different.  It constructs a new class description and 
marks the old one obsolete so that the next time an attempt is made to access 
one of the slots in an old instance, UPDATE-INSTANCE-FOR-REDEFINED-CLASS will 
be called to update the instance."))

;; 11/30/88 DNG - Add code to remove MAKE-INSTANCE and SHARED-INITIALIZE methods.
;; 12/20/88 DNG - Changed from a flavor method to a CLOS method.
(defmethod make-instances-obsolete ((self standard-class))
  (declare (unspecial self))
  (let* ((new-class-description
	   (or (and (boundp 'temp-class-description)
		    temp-class-description)
	       (sys:copy-object (class-description self) )))
	 (obsolete-class-description (class-description self))
	 (obs-mapped-supers (class-mapped-supers self ))
	 (obs-mapped-slots (class-mapped-slot-names self )))
    (declare (special temp-class-description))
    ;;make sure that we don't share the plist with the obsolete class description
    (setf (class-description-plist new-class-description) nil)
    (obsolete-mapping-table self)
    (uncache-accessors self)
    (setf (class-description self) new-class-description)
    (set-obsolete-class-description obsolete-class-description )
    (setf (class-description-class-object new-class-description ) self )
    ;;transfer information from obsolete to new class-description
    (setf (class-mapped-supers self) obs-mapped-supers )
    (setf (class-mapped-slot-names self) obs-mapped-slots )
    ;;clear mapping tables slot so we don't pick them by mistake.
    (setf (class-mapping-tables self) nil)
    ;; Make sure that we don't call an obsolete method that was automatically 
    ;; generated from the old class description.
    (let ((old-method (find-method #'make-instance nil (list `(eql ,self)) nil)))
       (when (and old-method (getf (method-plist old-method) 'computed-method))
	 (remove-method #'make-instance old-method)))
    (let ((old-method (find-method #'shared-initialize nil (list self *t-class*) nil)))
       (when (and old-method (getf (method-plist old-method) 'computed-method))
	 (remove-method #'shared-initialize old-method)))
    nil))

(p-defgeneric-options  uncache-accessors(class ))

(defmethod uncache-accessors ((class standard-class))
  (let ((class-desc (class-description class)))
    (flet ((clean (key mloc mtl ht keylist)
	       (declare (ignore mloc mtl keylist))
	       (when (eq key class-desc)
		 (remhash key ht))))
      (dolist (slot-description (class-direct-slots class ))
	(let ((writers (slot-writers slot-description))
	      (readers (slot-readers slot-description)))
	  (dolist (reader readers )
	    (map-all-hash-entries (get-generic-function-object reader)
				  #'clean))
	  (dolist (writer writers )
	    (map-all-hash-entries (get-generic-function-object writer)
				  #'clean)))))))





;;Runtime function
(defun restructure-instance (instance class)
  ;; 12/01/88 DNG - Added calls to COMPOSE-CLASS and FINALIZE-INHERITANCE.
  ;;  3/17/89 GK/DNG - Fix JOIN-INTO call to use LIST instead of LIST* so that PLIST is built properly.
  (unless (class-composed-p class)
    (compose-class class))
  (unless (internal-class-finalized-p class)
    (finalize-inheritance-internal class :slots t :methods t :initargs t ))
  (without-interrupts
    (let* ((new-instance (allocate-instance class))
	   (old-class-description (sys:%class-description instance))
	   (old-slots (class-description-instance-slot-names old-class-description ))  
	   (old-class-slots (class-description-class-slot-alist old-class-description ))
	   (instance (sys:follow-structure-forwarding instance))
	   (old (class-description-all-slots old-class-description ))
	   (new (class-all-slots class)))
      (do ((sl (class-description-instance-slot-names (class-description class)) (cdr sl))
	   (i 1 (1+ i)))
	  ((null sl))
	(let ((old-pos (position (car sl) (the list old-slots) :test #'eq)))
	  (if old-pos
	      (sys:%blt-typed (sys:%make-pointer-offset sys:dtp-locative instance (1+ old-pos))
			      (sys:%make-pointer-offset sys:dtp-locative new-instance i) 1 1))
	  (let ((old-loc (cdr (assoc (car sl) old-class-slots :test #'eq))))
	    (when old-loc
	      (sys:%blt-typed old-loc 
			      (sys:%make-pointer-offset sys:dtp-locative new-instance i) 1 1)))))
      (let ((acons (assoc (class-description class)
			  (class-description-update-instance-cache-alist old-class-description )
			  :test #'eq))
	    added discarded plist)
	(if acons
	    (progn
	      (setf added (cadr acons)
		    discarded (caddr acons))
	      (setf plist (collect-body
			    (dolist (sn discarded)
			      (let ((loc (sys:locate-in-instance  instance sn)))
				(when (slot-location-boundp loc)
				  (join (list* sn (contents loc)))))))))
			      
	    (progn 
	      (multiple-value-setq (added discarded plist)
		(multiple-values-collect-body (added discarded plist)
		  (dolist (s old)
		    (let ((sn (slot-name s)))
		      (unless (member sn new :key #'slot-name :test #'eq)
			(collect-into discarded sn)
			;;locate-in-instance does not call out on obsolete instances
			(let ((loc (sys:locate-in-instance  instance sn)))
			  (when (slot-location-boundp loc)
			    (join-into plist (list sn (contents loc))))))))
		  (dolist (s new)
		    (let ((sn (slot-name s)))
		      (unless (member sn old :key #'slot-name :test #'eq)
			(collect-into added sn))))))
	      ;;Once computed, these list should be cached.
	      (setf (class-description-update-instance-cache-alist old-class-description )
		    (cons (cons (class-description class) (list added discarded))
			  (class-description-update-instance-cache-alist old-class-description )))))
	(sys:structure-forward instance
			       new-instance)
	(sys:%p-store-cdr-code instance '#.(ldb sys:%%q-cdr-code
						(dpb 1 sys:%%CLOS-INSTANCE-HEADER-FLAG 0)))
	(update-instance-for-redefined-class  instance added discarded plist)))))




(defgeneric update-instance-for-redefined-class (instance added-slots discarded-slots property-list
							  &rest initargs)
  (:documentation 
 "This is called by the system when an attempt is made to access a slot in an 
instance whose class has been changed to have a different set of slots since 
this instance was created.  It initializes the new version of the instance by 
invoking SHARED-INITIALIZE to initialize any added slots.  Users would not 
usually call this directly, but can add methods to perform additional action."))

(defmethod update-instance-for-redefined-class
	   ((instance standard-object) added-slots discarded-slots property-list &rest initargs)
  (declare (ignore discarded-slots property-list ))
  (let ((instance-slots (class-description-instance-slot-names (sys:%class-description instance))))
    ;;filter out class slots.
    (apply #'shared-initialize instance
	   (collect-body
	     (dolist (s added-slots)
	       (when (member s instance-slots :test #'eq)
		 (collect s))))
	    initargs)))



(defgeneric change-class (instance new-class)
  (:documentation "Change the class of an instance to a new class.
It destructively modifies and returns the instance."))

(defmethod change-class ((instance standard-object ) (new-class standard-class))
  ;;We could use a resource of objects if this conses too much.
  (let ((previous (sys:copy-object instance)))
    (unless (class-composed-p new-class)
      (compose-class new-class))
    (without-interrupts
      (let* ((new-instance (allocate-instance new-class))
	     (old-class-description (sys:%class-description instance))
	     (old-slots (class-description-instance-slot-names old-class-description ))  
	     (old-class-slots (class-description-class-slot-alist old-class-description ))
	     (instance (sys:follow-structure-forwarding instance)))
	(do ((sl (class-description-instance-slot-names (class-description new-class)) (cdr sl))
	     (i 1 (1+ i)))
	    ((null sl))
	  (let ((old-pos (position (car sl) (the list old-slots) :test #'eq)))
	    (if old-pos
		(sys:%blt-typed (sys:%make-pointer-offset sys:dtp-locative instance (1+ old-pos))
				(sys:%make-pointer-offset sys:dtp-locative new-instance i) 1 1))
	    (let ((old-loc (cdr (assoc (car sl) old-class-slots :test #'eq))))
	      (when old-loc
		(sys:%blt-typed old-loc 
				(sys:%make-pointer-offset sys:dtp-locative new-instance i) 1 1)))))
	(sys:structure-forward instance new-instance)
	(sys:%p-store-cdr-code instance #.(ldb sys:%%q-cdr-code
					       (dpb 1 sys:%%CLOS-INSTANCE-HEADER-FLAG 0)))))
    (update-instance-for-different-class  previous instance)
    instance))

(defmethod change-class ((instance class) (new-class class-class))
  ;; 12/07/88 DNG - Original.
  (let ((previous (sys:copy-object instance)))
    (unless (class-composed-p new-class)
      (compose-class new-class))
    (without-interrupts
      (let* ((new-instance (allocate-instance new-class))
	     (old-flavor (sys:%class-description instance))
	     (old-class-description (class-description (class-description-class-object old-flavor)))
	     (old-slots (class-description-instance-slot-names old-flavor))
	     (old-class-slots (class-description-class-slot-alist old-class-description ))
	     (instance (sys:follow-structure-forwarding instance)))
	(do ((sl (class-description-instance-slot-names (class-description new-class)) (cdr sl))
	     (i 1 (1+ i)))
	    ((null sl))
	  (let ((old-pos (position (car sl) (the list old-slots) :test #'eq)))
	    (if old-pos
		(sys:%blt-typed (sys:%make-pointer-offset sys:dtp-locative instance (1+ old-pos))
				(sys:%make-pointer-offset sys:dtp-locative new-instance i) 1 1))
	    (let ((old-loc (cdr (assoc (car sl) old-class-slots :test #'eq))))
	      (when old-loc
		(sys:%blt-typed old-loc 
				(sys:%make-pointer-offset sys:dtp-locative new-instance i) 1 1)))))
	(sys:structure-forward instance new-instance)
	))
    (update-instance-for-different-class previous instance)
    instance))

(defmethod change-class ((instance t) (new-class symbol))
  (change-class instance (class-named new-class)))


(defgeneric update-instance-for-different-class (previous current &rest initargs))

(defmethod update-instance-for-different-class  ((previous  standard-object) (current  standard-object) &rest initargs)
  (without-interrupts
    (let* ((new-slots (class-description-instance-slot-names (sys:%class-description current)))
	   (old-slots (class-description-instance-slot-names (sys:%class-description  previous)))
	   (added-slots (collect-body
			  (dolist (sl new-slots)
			    (unless (member sl old-slots :test #'eq)
			      (collect sl))))))
      (apply #'shared-initialize current added-slots  initargs))))

;;;Microcode support function, when the microcode uses the class-description and
;;; it is obsolete, this function is called and is expected to return the new class-description.
;;; It should restructure the instance.

(si:%add-to-support-vector 'obsolete-class-description
			   sys:%SVC-CLOS-OBSOLETE-CLASS-DESCRIPTION t)

(defun obsolete-class-description (instance old-class-description)
  (restructure-instance instance (class-description-class-object old-class-description ))
  (sys:%class-description instance))

;;;03/28/89 clm - moved this before ADD-CLASS-NAMED - that was using it before it was defined
;;;03/28/89 clm - added the following; a strange error "VALID-SUPERCLASS-P is not a generic function"
;;;was occurring which this appears to fix.
(unless (ticlos:generic-function-p 'valid-superclass-p)
  (fmakunbound 'valid-superclass-p))

;; Added 11/23/88
;;  1/27/89 correct doc string
(defgeneric valid-superclass-p (class-prototype superclass)
  (:documentation "Tests whether SUPERCLASS is valid for a class of CLASS-PROTOTYPE's metaclass.
Returns true if the inheritance is permissible, or NIL if an error should be signalled."))

;;  1/30/89 DNG - Added reference to VALIDATE-SUPERCLASS to the error message.
(defmethod valid-superclass-p ((class-prototype class) (superclass class))
  (let ((metaclass (class-of class-prototype))
	(supermeta (class-of superclass)))
    (or (eq metaclass supermeta)
	(error "~S is not a valid super class for metaclass ~S.~A" superclass metaclass
	       (let ((this-package '#,(symbol-package 'standard-class)))
		 (if (and (eq (symbol-package (class-name metaclass)) this-package)
			  (eq (symbol-package (class-name supermeta)) this-package))
		     ""
		   "
  If you want this to be permitted, write a method for VALIDATE-SUPERCLASS."))))))

(defmethod valid-superclass-p ((class-prototype class-class) (superclass flavor-class)) t)

(defmethod valid-superclass-p ((class-prototype hybrid-class) (superclass flavor-class)) t)
(defmethod valid-superclass-p ((class-prototype hybrid-class) (superclass standard-class)) t)
(defmethod valid-superclass-p ((class-prototype standard-class) (superclass class-class))
  ;; Can't accept as is, but can promote metaclass to class-class.
  (values nil (class-of superclass)))

;;  2/25/89 DNG - Added the following two methods to facilitate building 
;;		classes on top of the standard types STREAM or PATHNAME.
(defmethod valid-superclass-p ((class-prototype standard-class) (superclass flavor-class))
  ;; These two standard Common Lisp types are implemented here as flavors.
  (if (and (or (subclassp superclass '#,(find-class 'stream))
	       (subclassp superclass '#,(find-class 'pathname)))
	   (eq (class-of class-prototype) '#,(find-class 'standard-class)))
      ;; Can't accept as is, but can promote metaclass to hybrid-class.
      (values nil '#,(find-class 'hybrid-class))
    (call-next-method)))

(defmethod valid-superclass-p ((class-prototype standard-class) (superclass hybrid-class))
  (if (and (or (subclassp superclass '#,(find-class 'stream))
	       (subclassp superclass '#,(find-class 'pathname)))
	   (eq (class-of class-prototype) '#,(find-class 'standard-class)))
      ;; Can't accept as is, but can promote metaclass to hybrid-class.
      (values nil '#,(find-class 'hybrid-class))
    nil))


;; From 88-003 page 3-37	added 9/27/88 by DNG
(defgeneric CLOS:ADD-NAMED-CLASS (class-prototype &key name 
				 superclass-names slot-specifications direct-options environment)
  (:documentation "Create and install a new class or update an existing class.")
  (declare (values class-object)))

;;  9/27/88 DNG - Added environment argument.
;; 10/21/88 DNG - Redesigned using CLOS:CLASS-FOR-REDEFINITION and CREATE-CLASS .
;; 10/28/88 DNG - Remove :ENVIRONMENT argument from  REINITIALIZE-INSTANCE.
;; 11/01/88 DNG - Added call to DEFFLAVOR1 when defining a metaclass.
;; 11/17/88 DNG - Use valid-superclass-p; move metaclass definition support to separate method.
;; 12/02/88 DNG - Changed from a flavor method to a CLOS method.
;; 12/05/88 DNG - Add promotion of metaclass when needed.
;;  5/07/89 DNG - Add use of *ALL-STANDARD-CLASS-NAMES* .
(defmethod clos:ADD-NAMED-CLASS ((class-prototype standard-class) &rest args &key name 
				 superclass-names slot-specifications direct-options environment)
  ;; call FIND-CLASS first in order to get an error if the class-name argument is not a symbol.
  (let* ((class-name name)
	 (old-class-object (find-class class-name nil environment))
	 (metaclass (class-of class-prototype))
	 (direct-supers superclass-names)
	 (direct-slots (mapcar #'(lambda (slot)
				     (if (typep slot 'slot-description)
					  slot
				       (make-slot-description metaclass slot class-name)))
				 slot-specifications))
	 (class-options direct-options)
	 (promote-metaclass nil))
    (dolist (super-name direct-supers)
      (let ((super-class (class-named super-name t environment)))
	 (unless (null super-class)
	   (multiple-value-bind (ok promote)
	       (valid-superclass-p class-prototype super-class)
	     (cond (ok)
		   ((and promote
			 (or (null promote-metaclass)
			     (eq promote promote-metaclass)))
		    (setq promote-metaclass promote))
		   (t (non-fatal-error :probable-error "Class ~S can't inherit from class ~S."
			    class-name super-name)) )))))
    ;; A temporary hack to enable users to define metaclasses even though 
    ;; metaclasses are not really implemented as standard objects.
    (when promote-metaclass
      (return-from clos:add-named-class
	(apply #'clos:add-named-class (class-prototype (class-named promote-metaclass nil environment))
	       args)))
    (when (null direct-supers)
      (setq direct-supers (class-default-supers class-prototype)))
    (when (or (compiler:environment-remote-p environment)
	      (sys:RECORD-SOURCE-FILE-NAME class-name 'defclass))
      (let ((class-object
	      (if (and old-class-object		; There is a class of the same name already.
		       (or (null environment)
			   (compiler:same-environment-p (class-environment old-class-object)
							environment)))
		  ;; Try to re-use the existing class.
		  (reinitialize-instance 
		    (clos:class-for-redefinition class-prototype old-class-object
						 :direct-superclasses direct-supers
						 :direct-slots direct-slots
						 :direct-options class-options
						 :environment environment)
		    :direct-superclasses direct-supers :direct-slots direct-slots
		    :direct-options class-options)
		;; Else create a new class object.
		(make-instance metaclass
				 :name class-name
				 :direct-superclasses direct-supers
				 :direct-slots direct-slots
				 :direct-options class-options :environment environment))))
	;; Update the slot-class-pointer to point the the class-object.
	(LET ((sys:%inhibit-read-only t))
	  (dolist (ds direct-slots)
	    (when (eq (car-safe (slot-allocation ds)) :class)
	      (setf (cadr (slot-allocation ds)) class-object))))
	(unless (eq class-object old-class-object)
	  ;; Delete any old type definitions having the same name.
	  (compiler:remprop-from-environment class-name 'sys:defstruct-description environment)
	  (unless (typep class-prototype 'hybrid-class)
	    (compiler:remprop-from-environment class-name 'sys:flavor environment))
	  (compiler:remprop-from-environment class-name 'sys:type-expander environment)
	  ;; Install the definition.
	  (setf (class-named class-name t environment) class-object)
	  (pushnew class-name *all-standard-class-names* :test #'eq)
	  (when old-class-object
	    ;; Now, make sure that we don't find somebody that depends on old-class-object 
	    ;; in the current environment.
	    (update-subclass-cpl old-class-object environment)
	    (setf (class-composed-p class-object) nil)
	    ))
	class-object))))

(DEFPROP defclass "Class" sys::DEFINITION-TYPE-NAME) ; Used in RECORD-SOURCE-FILE-NAME -- DNG 5/8/89

(defmethod clos:ADD-NAMED-CLASS ((class-prototype built-in-class) &rest args &key name 
				 superclass-names slot-specifications direct-options environment)
  (send class-prototype :add-named-class name superclass-names slot-specifications direct-options environment))

(defmethod class-default-supers ((class standard-class))
  '(standard-object))

(defmethod class-default-supers ((class class-class))
  (error "Null inheritance not valid for metaclass ~S."
	 (class-proper-name class)))

;; 12/20/88 DNG - Moved ALLOCATE-INSTANCE methods to file MAKE-INSTANCE because of bootstrapping problems.

;; From 88-003 page 3-39	added 10/17/88 by DNG
(defgeneric CLOS:CLASS-FOR-REDEFINITION (prototype-instance old-class &key
					 direct-superclasses direct-slots direct-options environment)
  (:documentation "Return the class object to be used for the new definition of an existing class.")
  (declare (values class-object)))

;; 12/07/88 DNG - Added CHANGE-CLASS.
(defmethod CLOS:CLASS-FOR-REDEFINITION ((prototype-instance standard-class)
					(old-class standard-class) &rest ignore)
  (unless (eq (class-of old-class) (class-of prototype-instance))
    (change-class old-class (class-of prototype-instance)))
  old-class)

(defmethod CLOS:CLASS-FOR-REDEFINITION ((prototype-instance structure-class)
					(old-class structure-class) &rest ignore)
  old-class)


;; Default method for when the metaclass doesn't match
;; Added 10/21/88 by DNG
;; 12/02/88 - use MAKE-INSTANCE instead of CREATE-CLASS.
(defmethod CLOS:CLASS-FOR-REDEFINITION ((prototype-instance class) (old-class class) &key
					direct-superclasses direct-slots direct-options environment)
  (let ((name (class-name old-class)))
    (cerror "Discard the old definition of ~S and proceed with installation of the new one."
	    "Class ~S was defined as a ~S, but is being redefined as a ~S."
	    name (type-of old-class) (type-of prototype-instance))
    (setf (find-class name nil environment) nil)
    (make-instance (class-of prototype-instance)
		     :name name
		     :direct-superclasses direct-superclasses
		     :direct-slots direct-slots
		     :direct-options direct-options :environment environment)))


;; Added 10/28/88 by DNG but not used yet.
(defmethod initialize-instance ((class-object standard-class) &key 
				direct-superclasses direct-slots direct-options documentation)
  (send class-object :update-class direct-superclasses direct-slots
	(if (null documentation)
	    direct-options
	  (list* :documentation documentation direct-options)))
  class-object)

;;;  4/ 7/89 GBK - Add REMOVE-ACCESSOR-METHODS
(defmethod reinitialize-instance ((class-object standard-class) &key 
				  direct-superclasses direct-slots direct-options documentation)
  (remove-accessor-methods class-object)
  (send class-object :update-class direct-superclasses direct-slots
	(if (null documentation)
	    direct-options
	  (list* :documentation documentation direct-options)))
  class-object)

;;; This method is called from (METHOD REINITIALIZE-INSTANCE (STANDARD-CLASS))
;;; to remove all accessor methods from redefined class.  Eventually it should
;;; be called from (METHOD SHARED-INITIALIZE (STANDARD-CLASS)).
;;;  4/ 6/89 GBK - Original

(DEFMETHOD remove-accessor-methods ((class standard-class))
  "This method removes all methods for accessor functions"
  (LET ((accessor-methods (specializer-direct-methods class)))
    (DOLIST (method accessor-methods)
      (WHEN (TYPEP method 'standard-accessor-method)
	(remove-method (*method-generic-function method) method)))))

(defmethod reinitialize-instance ((class-object structure-class) &key 
				  direct-superclasses direct-slots direct-options documentation)
  (declare (ignore direct-options documentation))
  (let ((class-descript (class-description class-object )))
    (setf (class-description-direct-supers class-descript)
	  direct-superclasses)
    (setf (class-description-direct-slots class-descript)
	  direct-slots))
  class-object)

(unless (generic-function-p 'compiler:deleted-from-environment)
  (fmakunbound 'compiler:deleted-from-environment))

(defmethod compiler:deleted-from-environment ((class class) environment)
  ;; Called at the end of COMPILE-FILE to remove pointers to classes defined 
  ;; in the temporary environment.
  ;; 10/26/88 DNG - Original.
  ;;  3/15/89 DNG - Remove any methods specialized on the class.
  (when (compiler:same-environment-p (class-environment class) environment)
    ;; This class is going away.
    (dolist (super (class-direct-supers class))
      (let ((super-class (class-named super t environment)))
	(unless (or (null super-class)
		    (compiler:same-environment-p (class-environment super-class)
						 environment))
	  ;; Remove pointers to this class from the super-class.
	  (remove-direct-subclass super-class class nil) )))
    (dolist (method (specializer-direct-methods class))	; but need to also include combined methods
      (remove-method (*method-generic-function method) method))
    )
  (values))

(defmethod compiler:deleted-from-environment :after ((class hybrid-class) environment)
  ;;  3/15/89 DNG - Original.
  (declare (ignore environment))
  (with-slots (flavor) class
    (let ((flavor-name (si:flavor-name flavor)))
      (when (null (sys:get-flavor flavor-name nil))
	;; When removing a compile-time flavor that does not have a definition 
	;; in the resident environment, need to remove any pointers to its name 
	;; from base flavors inherited from the resident environment.
	(dolist (f (or (si:flavor-depends-on-all flavor)
		       (si:flavor-depends-on flavor)
		       (si:flavor-includes flavor)))
	  (let ((fs (sys:get-flavor f nil)))
	    (when (and fs (member flavor-name (si:flavor-depended-on-by fs) :test #'eq))
	      (setf (si:flavor-depended-on-by fs)
		    (delete flavor-name (the list (si:flavor-depended-on-by fs)) :test #'eq :count 1))))))))
  (values))


(defmethod clos:add-named-class ((class-prototype hybrid-class) &key name 
				 superclass-names environment &allow-other-keys)
  (let* ((class-name name)
	 (super-flavors nil)
	 (class-object (call-next-method)))
    (declare (list super-flavors) (type hybrid-class class-object))
    (dolist (super-name superclass-names)
      (when (sys:get-flavor super-name environment)
	(push super-name super-flavors)))
    (unless (null super-flavors)
      (let* ((flavor-name class-name)
	     (compiler:*compile-file-environment* environment)
	     (sys::*just-compiling* (and 
					 (compiler:environment-remote-p environment))))
	(sys::defflavor1 flavor-name
			 (collect-body
			    (dolist (sd (class-direct-slots class-object))
			       (declare (type slot-description sd))
			       (when (eq (slot-allocation sd) :instance)
				 (collect
				    (if (non-empty-p (slot-initform sd))
					(list (slot-name sd) (slot-definition-initform sd))
				      (slot-name sd))))))
	  (nreverse super-flavors)
	  (if (typep class-prototype 'hybrid-class-with-flavor-instances)
	      '()
	    '((:instantiation-flavor-function dont-make-hybrid-flavor)) ))
	(let ((fl (sys:get-flavor flavor-name environment)))
	   (setf (slot-value class-object 'flavor) fl)
	   (setf (class-description-class-object fl) class-object)
	    ;; shouldn't need this now that defflavor1 knows about this
	   ;;(setf (class-named class-name t environment) class-object)
	   )))
    class-object))

;; This function is referenced above for use by sys:instantiate-flavor.
;; 12/14/88 DNG - Original.
;;  5/07/89 DNG - Go ahead and invoke MAKE-INSTANCE instead of signalling an 
;;		error.  [Note that this requires a corresponding change in SYS:INSTANTIATE-FLAVOR.]
(defun dont-make-hybrid-flavor (flavor-description init-plist)
  ;;(declare (ignore init-plist))
  (let* ((name (sys:flavor-name flavor-description))
	 (class (find-class name nil)))
    (when (and (eq class (class-description-class-object flavor-description))
	       (typep class 'hybrid-class))
      ;;(error "Flavor ~S is not supposed to be instantiated; use ~S of class ~S instead."
      ;;     name 'ticlos:make-instance class)
      (return-from dont-make-hybrid-flavor (apply #'make-instance class (contents init-plist)))
      )
    name))

