;;; -*- Mode:Common-Lisp; Package:TICLOS; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND
;;;
;;; Use, duplication, or disclosure by the Government is subject to
;;; restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;; Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.


;;; A mapping table is a substitution vector mapping a super's slot in the class.
;;; It is also used to find the mapping table of a super before invoking
;;; the daemon method of a superclass.
;;;
;;; Mapping tables handle both class slots and instance slots. It also
;;; provides a way of calling methods instead of going directly to the
;;; slot. The structure of the mapping table is this:
;;; (aref mt 0..n) => slot access
;;; (array-leader mt 0) => 
;;; (array-leader mt 1) => Method-class 
;;; (array-leader mt 2) => Instance-class
;;; (locf (array-leader mt 3...)) =>Mapping table for daemon methods.
;;;
;;;Slot access:
;;; (aref mt 0..n) => fixnum: Instance slot offset within Class.
;;;                => locative: locative to the class-variable-location.
;;;                => NIL: Non existing slot in Class.
;;;                => (<read method> <write method>) Methods to call to
;;;                read or write the slot. LOCF ??

;;; Accessors and constructor.
(defconstant mapping-table-leader-start 3)

;;  2/27/89 DNG - Use background area.
(defun make-mt-internal (length leader-length)
  (make-array length :leader-length (+ mapping-table-leader-start  leader-length)
	      :area sys:background-cons-area))

(defsubst mapping-table-instance-class (mt)
  (array-leader mt 2))

(defsubst mapping-table-method-class (mt)
  (array-leader mt 1))

;;  2/08/89 DNG - Commented out unused MSLOTS and MSUPERS bindings.
;;  2/09/89 DNG - Permit mapping table for flavor instance.
(defun get-mapping-table (method-class class &optional recursion)
  (let (;;(mslots (and (typep method-class 'standard-class) (class-mapped-slot-names method-class)))
	;;(msupers (class-mapped-supers method-class))
	)
    (unless (or (typep class 'standard-class)
		(typep class 'flavor-class))
      (return-from get-mapping-table nil))
;    (unless (or mslots msupers)
;      (return-from get-mapping-table method-class))
    (or (cdr (assoc method-class (class-mapping-tables class ) :test #'eq))
	(and (member  method-class recursion :test #'eq) method-class)
	(create-mapping-table method-class class recursion)
	)))

;;  2/08/89 DNG - Added  :TEST #'EQ  for efficiency.  Add special case for T 
;;		class since it is a superclass of STANDARD-OBJECT.
;;  2/09/89 DNG - Permit mapping table for flavor instance.
(defun update-mapping-table (method-class  class)
  ;;remake the mapping table, because something has changed.
  ;;don't create them if it does not exist.
  (let ((acons (AND (or (typep class 'standard-class)
			(eq class *t-class*)
			(typep class 'flavor-class))
		    (assoc method-class (class-mapping-tables class) :test #'eq))))
    (when acons 
      (let* ((old (cdr acons))
	     (nmt (create-mapping-table method-class class nil)))
	(STRUCTURE-FORWARD old nmt)))))

;; 12/02/88 DNG - Skip flavor maps.
;; 12/06/88 DNG - Pass instance-slot-names to slot-offset-or-locative.
;;  2/08/89 DNG - Replace an old entry in the a-list instead of always pushing 
;;		new entries on the front, so that obsolete maps can be GC'd.  Removed 
;;		unnecessary call to CLASS-NAMED.  Added call to SUBCLASSP to avoid 
;;		generating spurious tables.   Add special case for T class since it is a 
;;		superclass of STANDARD-OBJECT.
;;  2/09/89 DNG - Don't call CLASS-ALL-SLOTS on a flavor class.  Change test 
;;		for calling CLASS-ALL-SLOTS to check CLASS instead of METHOD-CLASS.
;;  2/27/89 DNG - Enable building CLOS maps for flavor classes.  Use CONS-IN-AREA.
;;  3/01/89 DNG - Added error check for instance class not composed.
(defun create-mapping-table (method-class class recursion)
  ;; METHOD-CLASS is the class the method is specialized on.
  ;; CLASS is the actual class of the instance received as the method's argument.
  (let ((mslots (and (or (typep method-class 'standard-class)
			 (typep method-class 'flavor-class))
		     (class-mapped-slot-names method-class)))
	(msupers (and (or (typep method-class 'standard-class)
			  (typep method-class 'flavor-class)
			  (eq method-class *t-class*))
		      (class-mapped-supers method-class))))
    (let ((mt (make-mt-internal (length mslots) (length msupers)))
	  (all-slots (and (typep class 'standard-class) (class-all-slots class)))
	  (instance-slot-names (class-description-instance-slot-names (class-description class))))
      (when (and mslots (null instance-slot-names) (null (class-instance-size class)))
	(error "CLOS bug: ~S is not composed." class))
      (setf (mapping-table-instance-class mt) class)
      (setf (mapping-table-method-class mt) method-class)
      ;;Set up mapping-table-class-slots-structure 
      (do ((i 0 (1+ i))
	   (sl mslots (cdr sl)))
	  ((null sl ))
	(setf (aref mt i )
	      (slot-offset-or-locative (car sl) all-slots instance-slot-names)))
      (with-stack-list* (recursion method-class recursion)
	(do ((i mapping-table-leader-start  (1+ i))
	     (sl msupers (cdr sl)))
	    ((null sl ))
	  (when (subclassp class (car sl))
	    (setf (array-leader mt i)
		  (get-mapping-table (car sl) class recursion)))))
      ;; Store the new mapping table in the class description.
      (let* ((maps (class-mapping-tables class))
	     (acons (assoc method-class maps :test #'eq)))
	(if acons
	    (setf (cdr acons) mt)
	  (setf (class-mapping-tables class )
		(cons-in-area (cons-in-area method-class mt sys:background-cons-area)
			      maps sys:background-cons-area))))
      ;;Update other mapping table that might use this one.
      (dolist (acons (class-mapping-tables class))
	(when (arrayp (cdr acons)) ; skip any flavor maps
	  (do*((mmt (cdr acons))
	       (i mapping-table-leader-start (1+ i)))
	      ((= i (array-leader-length (cdr acons))))
	    (when (eq method-class (if (arrayp (array-leader mmt i))
				       (mapping-table-method-class  (array-leader mmt i))
				     (array-leader mmt i)))
	      (setf (array-leader mmt i) mt)))))
      mt
      ;;Update the method-hash-tables
      ;;Probably the method-hash-tables will have to use a locative instead of a direct pointer.
      ;;The alternative is the use of destructive operation on mapping tables.
      ;;I think we can get away with destructive operations on mapping tables. We can snap
      ;;out forwarding pointers when we dispatch.
      )))

;; 12/06/88 DNG - Do quick check of instance-slot-names before looking in 
;;		all-slots.  This has the following advantages:
;;		(1) Is faster (except for allocation :class).
;;		(2) Ensures consistency with LOCATE-IN-INSTANCE.
;;		(3) Enables hybrid-class to handle ordered instance variables consistently with flavors.
(defun slot-offset-or-locative (slot-name all-slots instance-slots)
  (or (position slot-name (the list instance-slots) :test #'eq)
      (do ((s all-slots (cdr s))
	   (i 0 ))
	  ((null s) nil)
	(when (eq slot-name (slot-name (car s)))
	  (return (if (eq :class (car-safe (slot-allocation (car s))))
		      (class-slot-location (cadr (slot-allocation (car s))) slot-name)
		    (progn (compiler:debug-assert nil nil "Internal CLOS error, instance slot inconsistency")
			   i))))
	(WHEN (eq :instance (slot-allocation (car s)))
	  (INCF i)))))


;;;We need to be able to support mapping-table-update when self-ref-index is called and the 
;;;slot was never asked before. in this mode, if the mapping table does not exist already, don't
;;;create it. it will be created when the class is composed or method-hash-table is created.
;;;Be carefull about it though,
;;;If self-ref-index was called for a sub-mapping table offset and we don't find it. we need to create
;;;this sub mapping table right away.

(p-defgeneric-options slot-reference-index (class slot-name))

;;  2/25/89 DNG - Add (THE LIST ...) for efficiency.
(defmethod (standard-class :slot-reference-index) (slot-name)
  ;;We need to update the mapping tables if they are already created.
  ;;And update all our subclasses that have mapping tables.
  (or (position slot-name (the list (class-mapped-slot-names self)) :test #'eq)
      (prog1 
	(progn
	  (setf (class-mapped-slot-names self)
		(nconc (class-mapped-slot-names self) (cons slot-name nil)))
	  (1- (length (class-mapped-slot-names self))))
	(update-mapping-tables self))))

;; 12/14/88 DNG - Added to get a more meaningful error message than unclaimed message.
(defmethod (class :default :slot-reference-index) (slot-name)
  (declare (ignore slot-name))
  (error "Loading a method for class ~S which was
   compiled as a STANDARD-CLASS, but is now defined as a ~S."
	 (class-proper-name self) (class-proper-name (class-of self))))

;;  2/25/89 DNG - New version to simplify and make applicable to flavor 
;;		classes.  Also use WITHOUT-INTERRUPTS to prevent two processes from 
;;		stepping on each other.
;;  3/31/89 DNG - Add use of BACKGROUND-CONS-AREA.
(defun mapping-table-reference-index (current-class next-method-class)
  (or (position current-class (the list (class-mapped-supers next-method-class)) :test #'eq)
      ;;We need to update the mapping tables if they are already created.
      ;;And update all our subclasses that have mapping tables.
      (progn (without-interrupts
	       (let ((msupers (class-mapped-supers next-method-class)))
		 (unless (member current-class msupers :test #'eq)
		   (setf (class-mapped-supers next-method-class)
			 (nconc msupers (cons-in-area current-class nil
						      sys:background-cons-area))))))
	     (update-mapping-tables next-method-class)
	     (mapping-table-reference-index current-class next-method-class))))

(comment ; old way 
(p-defgeneric-options  mapping-table-reference-index (class method-class))

(defmethod (standard-class :mapping-table-reference-index ) (method-class)
  ;;We need to update the mapping tables if they are already created.
  ;;And update all our subclasses that have mapping tables.
  (let ((pos (position self (class-mapped-supers method-class) :test #'eq)))
    (or pos
	(prog1 
	  (progn
	    (setf (class-mapped-supers method-class)
		  (nconc (class-mapped-supers method-class) (cons self nil)))
	    (1- (length (class-mapped-supers method-class))))
	  (update-mapping-tables method-class)))))

(defmethod (built-in-class :mapping-table-reference-index ) (method-class)
  ;;We need to update the mapping tables if they are already created.
  ;;And update all our subclasses that have mapping tables.
  (let ((pos (position self (class-mapped-supers method-class) :test #'eq)))
    (or pos
	(prog1 
	  (progn
	    (setf (class-mapped-supers method-class)
		  (nconc (class-mapped-supers method-class) (cons self nil)))
	    (1- (length (class-mapped-supers method-class))))
	  (update-mapping-tables method-class)))))
) ; end comment

;;  2/09/89 DNG - Add special cases for T class and for NIL number.
;;  2/27/89 DNG - Support maps for flavor instances.
;;  4/17/89 DNG - Add environment argument to CLASS-NAMED calls.
(defun clos-var-pointer (number class-name  name)
  (and number
       (let* ((class (class-named class-name nil compiler:*local-environment*))
	      (value (if (ldb-test sys:%%clos-self-ref-map-leader-flag number)
			 (let* ((to-class (class-named name nil compiler:*local-environment*)))
			   (if (and (not (typep class 'standard-class))
				    (not (typep to-class 'standard-class))
				    (not (and (eq class *t-class*)
					      (eq to-class *t-class*)))
				    (not (typep class 'flavor-class))
				    (not (typep to-class 'flavor-class)))
			       nil
			     (mapping-table-reference-index to-class
							    class)))
		       (slot-reference-index class name))))
	 (and value
	      (dpb value 
		   sys:%%clos-self-ref-slot-offset
		   number)))))

(defun %clos-map-list (&rest specs)
  (let ((list (make-list (length specs))))
    (do ((result-tail list (cdr result-tail))
	 (specs-tail specs (cdr specs-tail)))
	((null specs-tail))
      (setf (car result-tail)
	(apply #'clos-var-pointer (car specs-tail))))
    list))

;;  2/22/89 DNG - When updating map for class T, just update its mapped 
;;		supers, not all of its subclasses.
(defun update-mapping-tables (method-class &optional (instance-class method-class))
  ;;update the class mapping tables, starting at instance-class.
  ;;going to all of it dependents.
  (update-mapping-table method-class instance-class)
  (dolist (subc (if (typep instance-class 'built-in-class)
		    (class-mapped-supers instance-class)
		  (class-direct-subclasses instance-class)))
    (unless (eq subc instance-class)
      (update-mapping-tables method-class subc))))

;;  2/08/89 DNG - Added the following method.
;;  1/10/89 DNG - Changed to always return NIL, and deleted the set method.
;; This is really only used for class T.
(defmethod (built-in-class :class-mapping-tables) ignore) ; always returns empty list

;;Support for obsolete instances. 
(defun obsolete-mapping-table (class)
  (dolist (ale (class-mapping-tables class))
    (let ((mt (cdr ale)))
      (when (arrayp mt) ; skip flavor maps -- DNG 12/2/88
	(dotimes (i (array-total-size mt))
	  (setf (aref mt i) 'obsolete-instance))))))

;;  2/08/89 DNG - Fix to not error when ENTRY is NIL.
(defun describe-mapping-table (mt)
  (let* ((method-class (mapping-table-method-class mt))
	 (instance-class (mapping-table-instance-class mt))
	 (map-s (class-mapped-slot-names method-class))
	 (all-in-s (REMOVE :instance (class-all-slots instance-class) :key #'slot-allocation  :test-not #'eq)))
    (format t "~%Mapping Table mapping method class ~s to instance class ~s" method-class instance-class)
    (dotimes (i (length mt))
      (LET ((entry (aref mt i)))
	(IF (LOCATIVEP entry)
	    (format t "~% entry: ~d, Mapped slot name: ~s, Class slot locative: ~d"
	      i (nth i map-s) entry )
	    (format t "~% entry: ~d, Mapped slot name: ~s, Instance offset: ~d, Instance slot: ~s"
		    i (nth i map-s) entry (and entry (nth entry all-in-s))))))))

;;  2/10/89 DNG - Original, for use by the inspector.
;;  3/01/89 DNG - Use VECTORP instead of ARRAYP and ARRAY-RANK.			       
(defun mapping-table-p (obj)
  (and (vectorp obj)
       (array-has-leader-p obj)
       (>= (array-leader-length obj) mapping-table-leader-start)
       (eq (array-type obj) 'art-q)
       (sys:classp (mapping-table-method-class obj))
       (sys:classp (mapping-table-instance-class obj))))

;;Microcode support for slot-access failure:

;; 11/28/88 DNG - Fixed to pass correct slot name to SLOT-MISSING.
(defun slot-access-failure (map-value mode instance self-ref-pointer fef frame-loc new-value &optional mt)
  (declare (ignore frame-loc fef))
  (let ((slot-name (nth (ldb sys:%%CLOS-SELF-REF-SLOT-OFFSET self-ref-pointer )
			(class-mapped-slot-names (mapping-table-method-class mt)))))
    (cond ((null map-value)
	   (slot-missing (class-of instance) instance slot-name
			 (case  mode
			   (0 'slot-value)
			   (1 '(setf slot-value)))
			 (and (eql mode 1) new-value)))
	  ((eq map-value 'obsolete-instance)
	   (when (obsolete-instance-p instance)
	     (restructure-instance instance (class-of instance)))
	   (case mode
	     (0 (slot-value instance slot-name))
	     (1 (setf (slot-value instance slot-name) new-value))))
	  (t (ferror nil "slot access failure")))))
  
(si:%add-to-support-vector 'slot-access-failure #o160 t)


