;;; -*- cold-load:t; Mode:Common-Lisp; Package:EH; Fonts:(CPTFONT HL12B HL12BI); Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
1;;
;; Copyright (c) 1981 Massachusetts Institute of Technology 
;; Copyright (c) 1984-1989 Texas Instruments Incorporated.  All Rights Reserved.
;;

;; This file contains the  macro definitions and condition flavor definitions of lisp and ucode errors.*

;;;
;;; Edit History
;;;
;;;                   Patch
;;;   Date    Author  Number   Description
;;;--------------------------------------------------------------------
;;;   07-86     pf             Original, from parts of Rel2 EH files.
;;; 03-10-87    ab    *L.19    Added DEF-UCODE-FORMAT-ERROR for SELF-MAPPING-TABLE-NOT-ARRAY
;;; 05-28-87  ab/clm  *P.20    Fixed DECODE-NONEXISTENT-INSTANCE-VARIABLE for case
;;;                            when coming out of macrocode (instruction field is in M-1).
;;; 06-25-87    clm            Deleted some old debug code out of too-few/too-many args handler
;;; 12-04-87    RJF            Fixed :grow-pdl to only put the message out if the variable
;;;                            *allow-pdl-grow-message* is true.
;;; 12-11-87    KTN            modified bug-report-method of condition to print self and special
;;;                            bindings in the first 3 frames in the bug report, and changed
;;;                            *default-bug-report-frames* from 5 to 10.
;;;  1-14-88    RJF            Added small-bignum-or-fixnum to argtype error.
;;;  6/29/88    CLM            Renamed the "turd-alert" error flavor and methods to 
;;;                            NO-CURRENTLY-PREPARED-SHEET-ERROR.  (SPR 8338)
;;;  7/20/88    clm            added in Mary Fontana's (mjf) changes for CLEH.
;;;  8-5-88     ab             Added some more entries to *data-type-names*
;;;  9/06/88    clm            Added handler for new microcode trap KEY-NOT-FOUND (used for 
;;;                            invalid keyword args.
;;; 12/20/88    RJF            Added a bunch of clos errors
;;; 03/20/89    clm            Integrated CLOS changes into EH.
;;; 04/12/89    jlm	       Changed (PUTPROP ... usage to (SETF GET ...


1;; DEFSIGNAL and DEFSIGNAL-EXPLICIT*

(defmacro defsignal-explicit (signal-name flavor &optional args &body init-options)
  2"Define a signal name, which can be used in SIGNAL and FERROR.
SIGNAL-NAME is the signal name to be defined.
FLAVOR is the flavor of condition object to make,
 or a list (FLAVOR . ADDITIONAL-CONDITION-NAMES), where ADDITIONAL-CONDITION-NAMES
 is a list of condition-names to signal in addition to the flavor components.
 If you specify just a flavor name, the SIGNAL-NAME itself is used
 as the sole additional condition name.
 If you specify a one-element list (FLAVOR), there are no additional
 condition names, and you can optionally use :CONDITION-NAMES as one of
 the INIT-OPTIONS to control them dynamically.
DOCUMENTATION is a documentation string describing what this signal-name is for.
When SIGNAL-NAME is used as the first arg to MAKE-CONDITION, FERROR, SIGNAL, etc.
then you can refer to the remaining args with the arglist ARGS.
The argument names in that list may be used in the INIT-OPTIONS,
which are arguments to pass to MAKE-INSTANCE in addition to the flavor name."*
  (declare (arglist signal-name flavor &optional args documentation &body init-options))
  (let ((flavor (if (consp flavor) (car flavor) flavor))
	(documentation (if (stringp (car init-options)) (pop init-options)))
	(condition-names
	  (if (consp flavor) (cdr flavor) (list signal-name))))
    `(progn (si:record-source-file-name ',signal-name 'defsignal)
	    (defun (:property ,signal-name make-condition-function) (ignore . ,args)
	      (declare (sys:function-parent ,signal-name))
	      ,documentation
	      ,(if condition-names
		   `(make-instance ',flavor ,@init-options
				   :condition-names ',condition-names)
		 `(make-instance ',flavor ,@init-options))))))

(deff defsignal-format 'defsignal)

(defmacro defsignal (signal-name flavor args &body init-options)
  2"Define a signal name, which can be used in SIGNAL and FERROR.
SIGNAL-NAME is the signal name to be defined.
FLAVOR is the flavor of condition object to make,
 or a list (FLAVOR . ADDITIONAL-CONDITION-NAMES), where ADDITIONAL-CONDITION-NAMES
 is a list of condition-names to signal in addition to the flavor components.
 If you specify just a flavor name, the SIGNAL-NAME itself is used
 as the sole additional condition name.
DOCUMENTATION is a documentation string describing what this signal-name is for.
When SIGNAL-NAME is used as the first arg to MAKE-CONDITION, FERROR, SIGNAL, etc.
then the remaining arguments are treated as a format string and format arguments.
In addition, the first few (not counting the string) can be given names,
which you specify as the list ARGS.
These names (moved into the keyword package) become messages
that can be sent to the condition object to get the corresponding values.
In addition, the values of the ARGS may be used in the INIT-OPTIONS,
which are additional arguments to pass to MAKE-INSTANCE."*
  (declare (arglist signal-name flavor args &optional documentation &body init-options))
  (let ((flavor (if (consp flavor) (car flavor) flavor))
	(documentation (if (stringp (car init-options)) (pop init-options)))
	(condition-names
	  (if (consp flavor) (cdr flavor) (list signal-name)))
	(properties (mapcar '(lambda (sym) `',(intern (symbol-name sym) si:pkg-keyword-package))
			    args)))
    `(progn (si:record-source-file-name ',signal-name 'defsignal)
	    (defun (:property ,signal-name make-condition-function)
		   (ignore format-string &optional ,@args &rest format-args)
	      (declare (sys:function-parent ,signal-name))
	      ,documentation
	      (make-instance ',flavor
			     ,@init-options
			     :property-list (list . ,(mapcan 'list properties args))
			     :format-string format-string
			     :format-args (list* ,@args (copy-list format-args))
			     :condition-names ',condition-names)))))


1;; CONDITION*

(defflavor condition 
	   ((condition-identify nil)
	    (si:property-list nil)
	    (condition-names nil)
	    (format-string nil)
	    (format-args nil)
	    (initial-error-message-printed-p nil))
	   (si:property-list-mixin si:print-readably-mixin)
  (:initable-instance-variables condition-names format-string format-args)
  (:settable-instance-variables initial-error-message-printed-p) 
  (:gettable-instance-variables condition-names format-string format-args)
  :ordered-instance-variables
  (:init-keywords :proceed-types)
  (:default-handler condition-data-plist-lookup)
  (:method-combination (:or :base-flavor-last
			    :default-handler
			    :debugger-command-loop)
		       (:progn :base-flavor-last
			       :initialize-special-commands)
		       (:pass-on (:base-flavor-last proceed-types)
				 :user-proceed-types)
		       (:case :base-flavor-last
			      :proceed
			      :proceed-asking-user
			      :proceed-ucode-with-args
			      :document-proceed-type))
  (:outside-accessible-instance-variables condition-identify))

(defparameter *not-condition-flavors*
	      '(si:vanilla-flavor si:property-list-mixin si:print-readably-mixin
				  no-action-mixin proceed-with-value-mixin)
  2"These condition flavors are excluded from the condition names of a condition."*) 

(defmethod (condition :init) (&rest ignore)
1  ;; initialize condition-names to all flavor names which self depends on,
  ;; except for those flavor names which are in variable, *not-condition-flavors**
  (declare (special *not-condition-flavors*))
  (let ((flavors
	  (remove-if #'(lambda (symbol) (member symbol (the list *not-condition-flavors*) :test #'eq))
		    (the list (dont-optimize (si:flavor-depends-on-all (si:instance-flavor self)))))))
    (setq condition-names (nunion flavors condition-names :test #'eq))
;1    ;; This is needed since we now have global:error and cli:error;*
;1    ;; If one of these is a member of condition-names, need to include the other too*
;    (if (or (member 'global:error condition-names :test #'eq)
;	    (member 'error condition-names :test #'eq))
;	(setq condition-names (nunion condition-names '(error global:error) :test #'eq)))
    ))

(defmethod (condition :reconstruction-init-plist) ()
  (let (accum)
    (with-self-variables-bound
      (dolist (var (or (get (type-of self) 'keyword-var-alist)
		       (setup-keyword-var-alist (type-of self))))
	1;; When printing this flavor instance,
  *	1;; Mention only certain instance variables;*
	1;; Don't mention some variables at all (like initial-error-message-printed-p)*
	1;; and only mention others if they have a non-nil value (like format-string and format-args)*
	1;; or if they are bound.*
	(if (and (neq (cdr var) 'initial-error-message-printed-p)
		 (if (member (cdr var) '(si:property-list format-string format-args) :test #'eq) 
		     (symbol-value (cdr var)) 
		   (boundp (cdr var))))
	    (setq accum (list* (car var) (symbol-value (cdr var))  accum)))))
    accum))

(defprop condition (condition-identify) :dont-print-instance-variables)

(defun setup-keyword-var-alist (flavor-name)
1  ;; Return a list of all instance variables and values;
  ;; exclude those which in property list :dont-print-instance-variables*
  (let ((flavor (get flavor-name 'si:flavor))
	accum
	dont-mention)
    (dolist (component (dont-optimize (si:flavor-depends-on-all flavor)))
      (setq dont-mention (union (get component :dont-print-instance-variables) dont-mention
				:test #'eq)))
    (dolist (var (dont-optimize (si:flavor-all-instance-variables flavor)))
      (or (member var dont-mention :test #'eq) 
	  (push (cons (intern (string var) si:pkg-keyword-package) var)
		accum)))
1    ;; This alist only needs to be computed once for a given flavor-name;
    ;; (:method condition :reconstruction-init-plist) checks to
    ;; see if the flavor name has the property 'keyword-var-alist*
    ;;(putprop flavor-name accum 'keyword-var-alist)	; jlm 4/12/89
    (setf (get flavor-name 'keyword-var-list) accum)
    accum))


1;; The following four vanilla-flavor operations must be redefined
;; because conditions handle operations that vanilla-flavor's methods
;; do not realize that they handle.*

(defmethod (condition :which-operations) ()
  1;; include property-list items as operations*
  (nreconc (do ((l si:property-list (cddr l))
		accum)
	       ((null l) accum)
	     (push (car l) accum))
	   (funcall #'(:method si:vanilla-flavor :which-operations)
		    :which-operations)))

(defmethod (condition :operation-handled-p) (operation)
1  ;; check if operation is a member of property-list*
  (or (si:memq-alternated operation si:property-list)
      (funcall #'(:method si:vanilla-flavor :operation-handled-p)
	       :operation-handled-p operation)))

(defprop condition-data-plist-lookup t :error-reporter)
(defun condition-data-plist-lookup (operation &rest args)
 2"Any operation which is the same as the name of a property on the plist
of a condition instance just returns the value of the property.
Anything else is an error, as usual.
This is the default handler for format-string conditions."*
  (declare (:self-flavor condition))
  (if (si:memq-alternated operation si:property-list)
      (get (locf si:property-list) operation)
    (apply 'si:flavor-unclaimed-message operation args)))


(defmethod (condition :get-handler-for) (operation)
  (or (funcall #'(:method si:vanilla-flavor :get-handler-for)
	       :get-handler-for operation)
1      ;; also allow operation to be in property list*
      (if (si:memq-alternated operation si:property-list)
	  'condition-data-plist-lookup)))

(defmethod (condition :send-if-handles) (operation &rest args)
  (let ((handler (or (funcall #'(:method si:vanilla-flavor :get-handler-for)
			      :get-handler-for operation)
		1     ;; also allows operation to be in proberty list*
		     (if (si:memq-alternated operation si:property-list)
			 'condition-data-plist-lookup))))
    (if handler (apply handler operation args))))


1;; Reporting conditions*

(defmethod (condition :maybe-clear-input) (stream)
  2"Used after printing error message."*
  (send stream :clear-input))

(defun condition-report-abort-printing (condition)
  2"Handler to abort when error in printing."*
  (when (member :abort-printing (send condition :proceed-types) :test #'eq) 
    :abort-printing))

(defmethod (condition :around :print-self) 
	   (continuation map args stream ignore slashify-p &rest ignore)
  (if slashify-p
      1;; give up if an error while printing self*
      (condition-bind ((error 'condition-report-abort-printing))
        (lexpr-funcall-with-mapping-table continuation map args))
    (send self :report stream)))

;1;; 7/20/88 mjf/clm - changed for CLEH.*
(defmethod (condition :report) (stream)
  "2Using format-string and format-args,  print condition message on stream.*"
  (apply 'format stream format-string 
	 (or (send self :format-args)
	     (send self :send-if-handles :format-arguments))))

(defwrapper (condition :report) (ignore . body)
  `(let ((*print-level* *error-print-level*)
	 (*print-length* *error-print-length*))
1     ;; If there is an error while printing some object,
     ;; Just give up on that object.*
     (condition-bind ((error 'condition-report-abort-printing))
       . ,body)))

(defmethod (condition :print-initial-error-message) (sg)
  2"Print the initial error message only once.  This one usually done when just
entering the debugger."*
  (unless initial-error-message-printed-p
    (terpri) (terpri)
    (print-initial-error-message sg self)
    (setq initial-error-message-printed-p t)))

(defun print-initial-error-message (sg error-object)
  2"Print error message and backtrace for error-object when in sg."*
  (print-carefully "error message"
    (send *standard-output* :fresh-line)
    (send error-object :print-error-message sg nil *standard-output*))
  (print-brief-error-backtrace sg error-object)
  (send error-object :maybe-clear-input *standard-input*))

(defmethod (condition :string-for-printing) ()
1  ;; use the report message for condition object*
   (with-output-to-string (s) (send self :report s)))

(defmethod (condition :report-string) ()
1  ;; use the report message for condition object*
   (with-output-to-string (s) (send self :report s)))

(defmethod (condition :print-error-message-prefix) (sg brief stream)
   (declare (ignore sg brief))
   (princ ">>Condition: " stream))

(defmethod (condition :print-error-message) (sg brief stream)
  (send self :print-error-message-prefix sg brief stream)
  (send self :report stream)
  (terpri stream))

(defmethod (condition :bug-report-recipient-system) ()
  "LISPM")

(defvar *default-bug-report-frames* 10.
  "Default number of frames to include in a bug report backtrace.")
(forward-value-cell 'default-bug-report-frames '*default-bug-report-frames*)
(proclaim '(special default-bug-report-frames))

(defmethod (condition :bug-report-description) (stream &optional n-frames)
1  ;; bug report describption contains the error message,
  ;; and a backtrace of frames*
  (declare (special *error-sg* *error-locus-frame* *default-bug-report-frames*))
  (send self :print-error-message *error-sg* nil stream)
  (format stream "Backtrace from the debugger:")
  (let ((*standard-output* stream)
	total-frames)
1    ;; get total number of frames possible for this sg*
    (do ((frame *error-locus-frame* (sg-next-frame *error-sg* frame))
	 (i 0 (1+ i)))
	((null frame) (setq total-frames i)))
1    ;; show frame details for up to so many frames;
    ;; after so many frames, show brief detail until get to 10 frames;
    ;; after 10 frames, show a ...*
    (do ((frame *error-locus-frame* (sg-next-frame *error-sg* frame))
	 (i 0 (1+ i)))
	((null frame))
      (cond ((< i (or n-frames *default-bug-report-frames*))
	     (show-frame-for-bug-message *error-sg* frame)
	     (when (< i 3)
	       (let ((*current-frame* frame))
		 (com-print-frame-bindings *error-sg* self))))
	    ((= i (or n-frames *default-bug-report-frames*))
	     (format t "~%~%Remainder of stack:~%~%")
	     (show-frame-briefly-for-bug-message *error-sg* frame))
	    ((< i (+ 10. (or n-frames *default-bug-report-frames*)))
	     (show-frame-briefly-for-bug-message *error-sg* frame))
	    ((> i (- total-frames 10.))
	     (show-frame-briefly-for-bug-message *error-sg* frame))
	    ((= i (+ 10. (or n-frames *default-bug-report-frames*)))
	     (format t "~&..."))))))


1;; Operations used when signalling a condition and the Error Handler*

(defmethod (condition :find-current-frame) (sg)
1  ;; Returns four values, which are used to set *ERROR-LOCUS-FRAME*, *CURRENT-FRAME*,
  ;; *INNERMOST-VISIBLE-FRAME* and *INNERMOST-FRAME-IS-INTERESTING*.*
  (declare (values error-locus-frame current-frame innermost-visible-frame innermost-frame-is-interesting))
  (let* ((frame (sg-top-frame sg)) 
	 (*innermost-visible-frame* frame))
    (declare (special *innermost-visible-frame*)) 
1    ;; skip frame if a :ERROR-REPORTER function*
    (do ((rp (sg-regular-pdl sg)))
	((not (let ((f (function-name (rp-function-word rp frame))))
		(if (symbolp f) (get f :error-reporter)))))
1      ;; skip frames foothold -> fh-applier -> signal-condition
      ;; this is on the stack for microcode errors*
      (and (eq (function-name (rp-function-word rp frame))
	       'signal-condition)
	   (eq (function-name (rp-function-word rp (sg-next-frame sg frame)))
	       'fh-applier)
	   (eq (function-name (rp-function-word rp (sg-previous-nth-frame sg frame -2)))
	       'foothold)
	   (return (setq frame (sg-previous-nth-frame sg frame -3))))
      (setq frame (sg-next-frame sg frame)))
    (values frame (sg-out-to-interesting-frame sg frame) *innermost-visible-frame* nil)))

(defmethod (condition :debugger-command-loop) (sg &rest ignore)
1  ;; There are 2 debugger command loops now,  one is a ucl version*
  (declare (special *use-old-debugger*)) 
  (if (and (not *use-old-debugger*) (functionp 'ucl-dbg-command-loop))
      (funcall 'ucl-dbg-command-loop sg self)
    (command-loop sg self)))

(defmethod (condition :debugging-condition-p) ()
  nil)

(defmethod (condition :dangerous-condition-p) ()
  nil)

(defun condition-typep (condition-instance condition-name)
  2"T if CONDITION-NAME is one of the condition names possessed by CONDITION-INSTANCE."*
  (condition-typep-1 (send condition-instance :condition-names) condition-name))

(defun condition-typep-1 (names frob)
1  ;; Allow item AND, OR and MEMBER in frob list*
  (if (atom frob)
      (member frob (the list names) :test #'eq) 
      (case (car frob)
	(and 
	 (dolist (c (cdr frob) t)
	   (when (not (condition-typep-1 names c))
	     (return nil))))
	((or member)
	 (dolist (c (cdr frob) nil)
	   (when (condition-typep-1 names c)
	     (return t))))
	(not 
	 (not (condition-typep-1 names (cadr frob)))))))


1;; Proceeding, or trying to.*

(defmethod (condition :proceed) (&rest args)
   1;; First arg should be the proceed type*
   1;; If args exist, return all of them.*
   1;; If there are no args, call the appropriate :PROCEED-ASKING-USER method if any,*
   1;; or just return the proceed type (first arg)*
   (if (cdr args)
       (values-list args)
       1;; no args*
       (let ((proceed-type (car args)))
	 (if (send self :proceed-asking-user :operation-handled-p proceed-type)
	     (send self :proceed-asking-user proceed-type 'values 'read-object)
	     proceed-type))))

;1;; 7/20/88 mjf/clm - added for CLEH.*
(defun call-restart-interactively  (restart &aux interactive-function)
  2"Calls the interactive-function of restart in the erring stack group,
unless in window debugger.  Invokes the restart in the erring stack group."*
  (declare (special *window-debugger-old-window* *error-sg* *current-frame*))
    (cond ((not *window-debugger-old-window*)
	   1;; regular debugger*
	   1;; call interactive function and invoke restart in erring stack group*
	   (sg-apply *error-sg* 
		     'cleh:invoke-restart-interactively `(,restart) nil nil))
	  (:else ;; in window debugger
	   1;; The restart interactive function uses stream *query-io*.*
	   1;; When we are in the window-debugger mode in the EH-stack group,  *query-io* is the window debugger.*
	   1;; If run interactive-function in ERRING-stack-group, *query-io* is  *terminal-io*-syn-stream -*
	   1;; and we would attempt to typeout onto a stream that is not the window debugger *
	   1;; and is not currently selected/exposed.*
	   1;; This normally is not a problem when in the regular debugger*
	   1;; since the regular debusets *query-io* to the terminal-io stream of the erring stack-group*
	   1;;*
	   1;; Since we have a problem with typeout, we need to run interactive-function in eh-stack group.*
	   1;; and then invoke the restart in erring stack group.*
	   (if (setq interactive-function (cleh:restart-interactive-function restart))
	       (sg-apply *error-sg* 
			 'cleh:invoke-restart `(,restart ,@(funcall interactive-function)) nil nil)
	       (sg-apply *error-sg* 
			 'cleh:invoke-restart `(,restart nil) nil nil)))))

(defmethod (condition :proceed-asking-user) (proceed-type continuation read-object-function)
  (declare (ignore read-object-function))
  ;1; If the user has defined a case of :PROCEED instead of :PROCEED-ASKING-USER, call it.*
  (cond ((send self :proceed :operation-handled-p proceed-type)
	 (let ((values (multiple-value-list
			(send self :proceed proceed-type))))
	   (when (car values)
	     ;1; If :PROCEED handler returns NIL, don't really proceed.*
	     (apply continuation values))))
	((and (find-package 'cleh)
	      (type-specifier-p 'cleh:restart)
	      (typep proceed-type 'cleh:restart))
	1 ;; Common LISP restart - mjf/clm 7/20/88*
	1 ;; Call the Common LISP restart's interactive function*
	 (call-restart-interactively proceed-type))
	(:else
	 (funcall continuation proceed-type))))

;;; 7/20/88 mjf/clm - changed for CLEH
(defmethod (condition :document-proceed-type) 
	   (proceed-type stream &optional (resume-handlers *condition-resume-handlers*))
1  ;; If this method is called, the  :case :document-proceed-type :proceed-type for the condition flavor
  ;; either returned nil or didn't exist.
  ;; The documentation of this proceed type is either:
  ;; 1. the documentation of the :proceed-asking-user or :proceed method for this proceed-type
  ;; 2. the report-function of a Common LISP restart - mjf/clm 7/20/88
  ;; 3. the documentation string of the the resume-handler for this proceed type
  ;; 4. the string "<proceed-type> undocumented."  *
  (declare (special *condition-resume-handlers*)) 
  (let ((string (or (send self :proceed-asking-user :case-documentation proceed-type)
		    (send self :proceed :case-documentation proceed-type)
		    ;1; Resume handler documentation*
		    (do ((handler-list resume-handlers (cdr handler-list))
			 h r)
			((eq handler-list (cdr handler-list)))
		      ;1; h is a resume handler*
		      (setq h (car handler-list))
		      (cond 
		       ((and (consp h)
			     (find-package 'cleh)
			     (type-specifier-p 'cleh:restart)
			     (typep (car h) 'cleh:restart))
			1;; h is a list of Common LISP restarts - mjf/clm 7/20/88*
			(setq r (dolist (restart h)
				  (when (or (eq restart proceed-type)
					    (eq (cleh:restart-name restart) proceed-type)
					    (and (eq self *abort-object*)
						 (eq (intern (cleh:restart-name restart) :keyword) :abort)))
				    (return (list (format nil "~A" restart))))))
			(when r (return r)))
		       ((and 
			   (consp h)
			   ;1; 1st item is the resume handler`s condition name(s)*
			   ;1; if nil, this handler is for all conditions*
			   ;1; if a condition name or a list of condition names,*
			   ;1; one of these names must be a name this condition object knows about*
			   (cond ((null (car h)) t)
				 ((atom (car h)) 
				  (member (car h) condition-names :test #'eq))
				 (t
				  (dolist (c (car h))
				    (if (member c condition-names :test #'eq) (return t)))))
			   ;1; 2nd item in resume handler is the proceed type*
			   (eq (cadr h) proceed-type) 
			   ;1; 3rd item is a predicate to see if this handler is active*
			   (or (eq (caddr h) t) (funcall (caddr h) self)))
			;1; 4th item is the format string and format args in a list*
			(return (fourth h)))))
		1    ;; At least say something about being an undocumented proceed*
		    (list "~:[~;~@(~A ~)~](undocumented)." proceed-type proceed-type))))
    ;1; Return the documentation string*
    (if (consp string)
	(apply 'format stream string)
      (format stream string))))

(defmethod (condition :proceed-type-p) (proceed-type)
  2"Does proceed type exist for this condition?"
1  ;; proceed type should either be a member of *condition-proceed-types* (set when entering signal-condition)
  ;; be a resume handler**
  (declare (special *condition-proceed-types* *condition-resume-handlers*))
  (or (member proceed-type *condition-proceed-types* :test #'eq) 
      (do ((handler-list *condition-resume-handlers* (cdr handler-list))
	   (h))
	  ((eq handler-list (cdr handler-list)))
	1;; h is a resume handler*
	(setq h (car handler-list))
	(and (consp h)
	     1;; 1st item is the resume handler`s condition name(s)*
	     1;; if nil, this handler is for all conditions*
	     1;; if a condition name or a list of condition names,*
	     1;; one of these names must be a name this condition object knows about*
	     (cond ((null (car h)) t)
		   ((atom (car h)) 
		    (member (car h) condition-names :test #'eq))
		   (t (dolist (c (car h))
			(if (member c condition-names :test #'eq) (return t)))))
	     1;; 3rd item is a predicate to see if this handler is active*
	     (or (eq (caddr h) t) (funcall (caddr h) self))
	     (return t)))))

(defmethod (condition :user-proceed-types) (proceed-types) proceed-types)

(defmethod (condition :ucode-proceed-types) () nil)

(defmethod (condition :proceed-types) ()
1  ;; proceed types include those in *condition-proceed-types* (initialized by signal-condition)
  ;; and also those resume handler keywords which are active for self*
  (declare (special *condition-proceed-types*)) 
  (union *condition-proceed-types* (condition-resume-types self) :test #'eq))



1;; Setup special commands
;; Make sure we have a combined method for this,
;; so that the :WHICH-OPERATIONS suboperation works.*

(defmethod (condition :special-command) (&rest ignore) nil)

(defmethod (condition :initialize-special-commands) () nil)


1;; PROCEED-WITH-VALUE-MIXIN*

(defflavor proceed-with-value-mixin () ()
  (:required-flavors condition))

(defmethod (proceed-with-value-mixin :case :proceed-asking-user :new-value)
	   (continuation read-object-function)
  2"Return a value specified by the user."
1  ;; CONTINUATION is a function which resumes stack group and returns the proceed-type and the new value**
  (funcall continuation :new-value
	   (funcall read-object-function :eval-read
		    "Form to evaluate and return: ")))

1;; NO-ACTION-MIXIN*

(defflavor no-action-mixin () () 
  (:required-flavors condition))

(defmethod (no-action-mixin :case :proceed-asking-user :no-action)
	   (continuation ignore)
  2"Simply proceed."
1  ;; CONTINUATION is a function which resumes stack group and returns the proceed type**
  (funcall continuation :no-action))

1;; WARNING*

(defflavor warning () (no-action-mixin condition))

(defmethod (warning :print-error-message-prefix) (ignore ignore stream)
  (princ ">>Warning: " stream))

1;; BREAK is used for the M-Break and C-M-Break keys*

(defflavor break ((format-string "break.")) (condition))

(defmethod (break :debugging-condition-p) ()
  t)

(defmethod (break :print-error-message-prefix) (sg brief stream)
  sg brief
  (princ ">>Keyboard " stream))

(defmethod (break :case :proceed-asking-user :no-action) (continuation ignore)
  2"Proceed from keyboard break."*
  (format t " Continue from break.~%")
   ;1; Resume stack group and return the proceed type*
  (funcall continuation :no-action))

1;; TRACE-BREAKPOINT is used by (TRACE (FOO :ERROR))*

(defflavor trace-break () (break))

(defmethod (trace-break :maybe-clear-input) (stream)
  stream nil)

(defmethod (trace-break :print-error-message-prefix) (sg brief stream)
  sg brief
  (princ ">>Trace break: " stream))

(defsignal trace-breakpoint trace-break ()
  2"Used by (trace (foo :error))."*)

1;; ABORT is used for Abort, M-Abort and C-M-Abort keys*

(defsignal sys:abort condition ()
  2"If signalled, aborts back to the innermost command loop."*)

(defparameter *abort-object* (make-condition 'sys:abort "Abort.")
  2"A condition-object for condition SYS:ABORT, used every time we want to abort."*)
(forward-value-cell 'abort-object '*abort-object*)
(proclaim '(special abort-object))




1;; ERROR*

;;; 7/20/88 mjf/clm - added for cleh
;1;; SERIOUS-CONDITION*
(defflavor serious-condition () (condition))

(defvar *error-identify-value* (list nil)
  2"This object is the first instance variable of every error object.
It is a fast way of identifying error objects."*)

;1;; 7/20/88 mjf/clm - changed for  CLEH.*
(defflavor error ((condition-identify *error-identify-value*)) (serious-condition))

;(defflavor global:error () (error))

;1;; 7/20/88 mjf/clm - added for CLEH.*
(defmethod (error :after :init) (&rest ignore)
  (when (find-package 'cleh)  ;;clm - added this
    (if (not (member 'cleh:error condition-names :test #'eq))
	(push 'cleh:error condition-names))))

(defmethod (error :print-error-message-prefix) (sg brief stream)
  2"Prints error message prefix >>Error: or >>Trap: "
1  ;; For errors detected by microcode, the function SIGNAL-CONDITION is always 
  ;; called (in the erring stack group) with *UCODE-ERROR-STATUS* set to a NON-NIL value.**
  (declare (special *calls-sub-list*))
  (cond ((not (symeval-in-stack-group '*ucode-error-status* sg))
	 (princ ">>Error: " stream))
	(brief
	 (princ ">>Trap: " stream))
	(t 
	1 ;; **1UCODE-ERROR-STATUS**1 is a list:*
	1 ;; (trap-micro-pc error-table-entry frame-pointer frame-pointer saved-micro-pcs)*
	 (let (trap-micro-pc ete saved-micro-pcs sym flag)
	   (setf (list* trap-micro-pc ete 'a 'a saved-micro-pcs)
		 (symeval-in-stack-group '*ucode-error-status* sg))
	1   ;; Print:  >>Trap PC ETE -> micro-symbol-names..*
	   (format stream ">>Trap #o~O ~A" trap-micro-pc ete)
	1   ;; The symbol name for each pc in saved-micro-pcs*
	1   ;; are in alist *calls-sub-list**
	   (dolist (pc saved-micro-pcs)
	     (setq sym (assoc (1- (%pointer pc)) (the list *calls-sub-list*) :test #'eq))
	     (when sym
	       (unless flag (format stream " ->"))
	       (setq flag t)
	       (format stream "  ~A " (cdr sym))))
	   (terpri stream)))))

(defun errorp (object)
  2"T if OBJECT is a condition-object representing an error condition.
This is equivalent to (TYPEP OBJECT 'ERROR) but faster."
1  ;; CONDITION-IDENTIFY is an instance variable for CONDITION flavor
  ;; and is initialized to *ERROR-IDENTIFY-VALUE* for ERROR flavors**
  (declare (special *error-identify-value*))
  (and (typep object 'instance)
       (location-boundp (locf (condition-condition-identify object)))
       (eq (condition-condition-identify object) *error-identify-value*)))

1;; FERROR*

(defflavor ferror () (proceed-with-value-mixin no-action-mixin error))

(defsignal nil (ferror) ()
   2"This is signaled for (FERROR NIL ...), etc."*)

1;; MACLISP-ERROR
;; It makes the error message out of MESSAGE and OBJECT,
;; and the condition-names out of INTERRUPT*

(defun (:property maclisp-error make-condition-function) (ignore message object interrupt)
  2"Makes a Maclisp error object."*
  (make-instance 'ferror
		 :format-string (if object "~S ~A" "~*~A")
		 :format-args (list object message)
		 :condition-names (list (get interrupt 'condition-name))))

1;; COMMON-LISP-CERROR*

(defflavor common-lisp-cerror (continue-format-string)
	   (ferror)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (common-lisp-cerror :report) (stream)
  (apply 'format stream format-string format-args))

(defmethod (common-lisp-cerror :case :document-proceed-type :continue)
	   (stream &optional ignore)
1  ;; use continue-format-string for documentation*
  (apply 'format stream continue-format-string format-args))

(defmethod (common-lisp-cerror :case :proceed-asking-user :continue)
	   (continuation ignore)
1  ;; CONTINUATION  is a function which resumes the stack group and returns the proceed type*
  (funcall continuation :continue))

;1;; 7/20/88 mjf/clm - NEW flavors for CLEH*
;1;; SIMPLE-CONDITION*
(defflavor simple-condition (format-arguments) (condition))

;1;; SIMPLE-WARNING*
(defflavor simple-warning (format-arguments) (warning))

;1;; SIMPLE-ERROR*
(defflavor simple-error (format-arguments) (error))

;1;; PROGRAM-ERROR*
(defflavor program-error () (error))

;1;; CONTROL-ERROR*
(defflavor control-error () (error))

;1;; TYPE-ERROR*
(defflavor type-error (datum expected-type) (error))

;1;; SIMPLE-TYPE-ERROR*
(defflavor simple-type-error (format-arguments) (type-error))

;1;; STREAM-ERROR*
(defflavor stream-error (stream) (error))

1;; DANGEROUS-ERROR
;; used by ucode errors VIRTUAL-MEMORY-OVERFLOW and REGION-TABLE-OVERFLOW*

(defflavor dangerous-error () (error))

(defmethod (dangerous-error :dangerous-condition-p) ()
  t)


1;; Various Non-Microcode Conditions and Errors

;; :FQUERY *

(defsignal-explicit :fquery condition (options format-string &rest format-args)
  2"By default, calls to FQUERY signal this."*
  :property-list (list :options options)		    
  :format-string format-string
  :format-args (copy-list format-args))

(defsignal zwei:barf condition ()
  2"All calls to ZWEI:BARF signal this condition, normally ignored."*)

1;; UNKNOWN LOCF/SETF REFERENCE*

(defsignal sys:unknown-locf-reference error (form)
  2"Means that LOCF was used on a form which it did not know how to handle."*)

(defsignal sys:unknown-setf-reference error (form)
  2"Means that SETF was used on a form which it did not know how to handle."*)

1;; ARITHMETIC-ERROR
;; NEGATIVE-SQRT, NON-POSITIVE-LOG, SINGULAR-MATRIX*
;1;; 7/20/88 mjf/clm - added two instance variables to flavor for CLEH.*
(defflavor arithmetic-error (operation operands) (error))

(defsignal sys:negative-sqrt arithmetic-error (number)
  2"NUMBER was a negative argument to SQRT."*)

(defsignal sys:non-positive-log arithmetic-error (number)
  2"NUMBER was used as the argument to a logarithm function."*)

(defsignal math:singular-matrix arithmetic-error (matrix)
  2"Signaled when any matrix handling function finds a singular matrix."*)

1;; WRONG-TYPE-ARGUMENT*

(defsignal-explicit wrong-type-argument wrong-type-argument-error
		    (format-string &rest format-args)
  2"Wrong type argument from a CHECK-ARG macro."*
  :format-string format-string
  :format-args (copy-list format-args) 
  :property-list `(:description ,(first format-args)
		    :old-value ,(second format-args)
		    :arg-name ,(third format-args)
		    :function
  		      1;skip 3 frames,  back to function that got the error*
		      1;stack should look like:*
		      1;erring-function -> cerror -> make-condition -> this defsignal's function*
		      ,(let* ((rp (sg-regular-pdl *current-stack-group*))
			      (frame (%pointer-difference (%stack-frame-pointer) 
							  (locf (aref rp 0)))))
			 (rp-function-word rp 
					   (sg-previous-nth-frame
					     *current-stack-group* frame -3)))))

;1;; 7/20/88 mjf/clm - changed for CLEH to use type-error*
(defflavor wrong-type-argument-error () (type-error))

(defmethod (wrong-type-argument-error :case :proceed-asking-user :argument-value)
	   (continuation read-object-function)
  2"Ask for a replacement argument and proceed."
1  ;; Resume stack group and return the proceed type and new argument valur**
  (funcall continuation :argument-value
	   (funcall read-object-function :eval-read
	     (format nil "Form to evaluate and use as a replacement value for ~A: "
		     (send self :arg-name)))))

1;; UNCLAIMED-MESSAGE*

(defflavor unclaimed-message-error () (error))

(defmethod (unclaimed-message-error :case :proceed-asking-user :new-operation)
	   (continuation read-argument-function)
  2"Read operation to perform instead."
1  ;; Resume stack group and returns the proceed type and new operation**
  (funcall continuation :new-operation
	   (funcall read-argument-function :eval-read
		    "Form to evaluate and use as operation to perform instead: ")))

(defsignal sys:unclaimed-message unclaimed-message-error
	   (object message arguments)
  2"OBJECT, an instance or funcallable hash*-2table*,2didn't recognize operation MESSAGE."*)

1;; INVALID-FORM*

(defsignal sys:invalid-form error (form)
  2"EVAL was given FORM and couldn't make sense of it."*)

1;; INVALID-FUNCTION-SPEC*

(defsignal sys:invalid-function-spec error (function-spec)
  2"Invalid function spec passed to FDEFINITION, etc."*)

1;; INVALID-FUNCTION and INVALID-LAMBDA-LIST 
;; INVALID-FUNCTION flavor is defined in ucode area*

(defsignal sys:invalid-function invalid-function (function)
  2"Function was supposed to be applied, but it is malformatted, etc."*)

(defsignal sys:invalid-lambda-list invalid-function (function)
  2"The interpreted function FUNCTION's lambda list is malformatted."*)

1;; UNDEFINED-KEYWORD-ARGUMENT*

(defflavor sys:undefined-keyword-argument () (error))

(defmethod (sys:undefined-keyword-argument :case :proceed-asking-user :new-keyword)
	   (continuation read-object-function)
  2"Read a form to use as keyword."
1  ;; Resume stack group and return the proceed type and the new keyword**
  (funcall continuation :new-keyword
	   (funcall read-object-function :eval-read
		    "Form to evaluate to use as the keyword instead of ~S: "
		    (send self :keyword))))

(defsignal sys:undefined-keyword-argument sys:undefined-keyword-argument (keyword value)
  2"A function wanting keyword args to KEYWORD with VALUE, which it wasn't expecting."*)

1;; STREAM-CLOSED, STREAM-INVALID*

(defsignal stream-closed (error stream-closed stream-invalid) (stream)
  2"I/O to STREAM, which has been closed and no longer knows how to do I/O."*)

(defsignal stream-invalid error (stream)
  2"I/O to STREAM, which is in a state not valid for I/O."*)

1;; END-OF-FILE*

;1;; 7/20/88 mjf/clm - changed for CLEH (to use stream-error).*
(defflavor end-of-file () (stream-error))

(defsignal sys:end-of-file-1 end-of-file (stream)
  2"End of file on STREAM, not within READ."*)


1;; PARSE READ ERRORS*

(defflavor parse-error (stream) (ferror) 
  :settable-instance-variables)

(defmethod (parse-error :after :init) (ignore)
  (declare (special si:read-stream))
1  ;; STREAM is used when printing error message*
  (setq stream (if (variable-boundp si:read-stream) si:read-stream)))

(defmethod (parse-error :after :print-error-message) (ignore ignore -stream-)
  (when stream
    (format -stream- "Error occurred when reading from ~S.~%" stream)))

(defmethod (parse-error :case :proceed-asking-user :no-action) (continuation ignore)
  2"Continue reading, trying to ignore the problem."
1  ;; Resume stack group and simple return the proceed type**
  (funcall continuation :no-action))

1;; PARSE-ERROR*
(defsignal parse-error-1 parse-error ()
  2"Error in parsing input; rubout handler should handle it."*)

1;; PARSE-FERROR*
(defflavor parse-ferror () (parse-error) 
  :alias-flavor)

1;; READ-ERROR*
(defsignal read-error-1 (parse-error read-error) ()
  2"Error other than end of file, within READ."*)

1;; MISSING-CLOSEPAREN*
(defsignal missing-closeparen (parse-error read-error missing-closeparen) ()
  2"Error of open paren found in column 0 in middle of defun."*)

1;; PRINT-NOT-READABLE*
(defsignal print-not-readable ferror (object)
  2"Printing OBJECT, which cannot be printed so it can be read back."*)


1;; End Of File Errors*

(defflavor read-end-of-file () (end-of-file parse-error))

(defmethod (read-end-of-file :case :proceed-asking-user :no-action) (continuation ignore)
  2"Close off unfinished lists and proceed."*
  (funcall continuation :no-action))

1;; READ-END-OF-FILE*
(defsignal read-end-of-file (read-end-of-file read-error) (stream)
  2"End of file within READ on STREAM.  SYS:READ-LIST-END-OF-FILE or
SYS:READ-STRING-END-OF-FILE should be used if they apply."*)

1;; READ-LIST-END-OF-FILE *
(defsignal read-list-end-of-file (read-end-of-file read-error read-list-end-of-file)
	   (stream list)
  2"End of file within READ constructing a list, on STREAM.
LIST is the list constructed so far."*)

1;; READ-STRING-END-OF-FILE *
(defsignal read-string-end-of-file (read-end-of-file read-error read-string-end-of-file)
	   (stream string)
  2"End of file within READ constructing a string, on STREAM.
STRING is the string read so far."*)

1;; READ-SYMBOL-END-OF-FILE *
(defsignal read-symbol-end-of-file (read-end-of-file read-error read-symbol-end-of-file)
	   (stream string)
  2"End of file within READ constructing a symbol, on STREAM.
Occurs only within a vertical-bar construct.  STRING is the string read so far."*)



1;; PACKAGE-NOT-FOUND/PACKAGE-ERROR*
;1;; 7/20/88 mjf/clm - changed for CLEH.*
(defflavor package-error () (ferror))

(defflavor package-not-found () (package-error))

(defmethod (package-not-found :case :proceed-asking-user :retry)
	   (continuation ignore)
  2"Look for the package again.  Use this if you create it by hand."
1  ;; Resume stack group and simple return the :RETRY proceed type**
  (funcall continuation :retry))

(defmethod (package-not-found :case :proceed-asking-user :create-package)
	   (continuation ignore)
  2"Create the package and proceed."*
  (format t "Creating package ~A." (send self :name))
1  ;; Resume stack group and simply return the :CREATE-PACKAGE proceed type, *
  (funcall continuation :create-package))

(defmethod (package-not-found :case :proceed-asking-user :new-name)
	   (continuation prompt-and-read-function)
  2"Ask for a name of a package to use instead."
1  ;; Resume stack group and return the proceed type and new **package 1name value*
  (funcall continuation :new-name
	   (funcall prompt-and-read-function :string
		    "Name (not local nickname) of package to use instead: ")))

(defsignal package-not-found-1 (package-not-found package-error)
	   (name relative-to)
  2"Package NAME doesn't exist; relative-to is package whose local nicknames we are searching."*)

1;; READ-PACKAGE-NOT-FOUND *
(defsignal read-package-not-found (package-not-found package-error read-error parse-error)
	   (name relative-to)
  2"Package prefix not recognized as a package name."*)

1;; NAME-CONFLICT
;;;  Routines to handle name-conflict resolution for functions that alter the package structure.*
;1;; 7/20/88 mjf/clm - changed for CLEH to use package-error*
(defflavor name-conflict-error () (package-error))   	1;So appropriate proceed types may be defined.*

(defsignal name-conflict name-conflict-error
	   (symbol pack operation conflicts-list)
	   "Used to signal a name conflict error when more than one symbol of the same name is accessible")

1;;;* 1This signal is used by PACKAGES, which has a SYS attribute.*
(export 'name-conflict)

1;;;  Methods for printing documentation strings for the various proceed types.*

(defmethod (name-conflict-error :case :document-proceed-type :skip) (stream ignore)
  (let ((op (send self :operation)))
    (if (eq op :use-package)
	(format stream "Return without using the ~a package."
		(send self :pack))
	(format stream "Return without ~(~a~)ing ~s ~(~a~) the ~a package."
		op
		(send self :symbol)
		(if (eq op :import) 'into 'from)
		(send self :pack)))))

(defmethod (name-conflict-error :case :document-proceed-type :skip-all) (stream ignore)
  (let ((op (send self :operation)))
    (if (eq op :use-package)
	(format stream "Return without using any of the packages.")
	(format stream "Return without doing any of the ~(~a~) operations."
		op))))

1;;;  EXPORT can cause name conflicts in two ways, so there is a separate proceed type
;;;  for each type of conflict (since different actions are required).  There is also a 
;;;  proceed type to handle cases where both types of conflicts occur when a symbol 
;;;  is exported.  First, the symbol being exported might match a symbol directly PRESENT 
;;;  in an inheriting package.  Second, the symbol being exported might match a symbol 
;;;  already ACCESSIBLE BY INHERITANCE in an inheriting package. *
(defmethod (name-conflict-error :case :document-proceed-type :export-present) (stream ignore)
  (format stream "Export ~s from the ~a package and resolve name conflicts." 
	  (send self :symbol)
	  (send self :pack)))

(defmethod (name-conflict-error :case :document-proceed-type :export-accessible-by-inheritance) (stream ignore)
  (format stream "Export ~s from the ~a package and resolve name conflicts." 
	  (send self :symbol)
	  (send self :pack)))

(defmethod (name-conflict-error :case :document-proceed-type :export-both-conflict-types) (stream ignore)
  (format stream "Export ~s from the ~a package and resolve name conflicts." 
	  (send self :symbol)
	  (send self :pack)))

1;;;  USE-PACKAGE, like EXPORT, can also cause name conflicts in two ways. *
(defmethod (name-conflict-error :case :document-proceed-type :use-package-present) (stream ignore)
  (format stream "Use the ~a package and resolve name conflicts." 
	  (send self :pack)))

(defmethod (name-conflict-error :case :document-proceed-type :use-package-accessible-by-inheritance) (stream ignore)
  (format stream "Use the ~a package and resolve name conflicts." 
	  (send self :pack)))

(defmethod (name-conflict-error :case :document-proceed-type :use-package-both-conflict-types) (stream ignore)
  (format stream "Use the ~a package and resolve name conflicts." 
	  (send self :pack)))

1;;;  Used by both EXPORT and USE-PACKAGE.*
(defmethod (name-conflict-error :case :document-proceed-type :unintern-all) (stream ignore)
  (if (eq (send self :operation) :export)
      (format stream "Unintern all the conflicting symbols from the inheriting packages and proceed with the exporting.")
      (format stream "Unintern all the conflicting symbols from the inheriting package.")))

1;;;  Used by both EXPORT and USE-PACKAGE.*
(defmethod (name-conflict-error :case :document-proceed-type :shadow-all) (stream ignore)
  (if (eq (send self :operation) :export)
      (format stream "Shadow all the conflicting symbols in the inheriting packages and proceed with the exporting.")
      (format stream "Shadow all the conflicting symbols in the inheriting package.")))

1;;;  IMPORT, like EXPORT and USE-PACKAGE, can also cause name conflicts in two ways.  
;;;  However, unlike EXPORT and USE-PACKAGE, IMPORT can never cause many simultaneous conflicts.*
(defmethod (name-conflict-error :case :document-proceed-type :import-present) (stream ignore)
  (format stream "Import ~s into the ~a package, uninterning the conflicting symbol already present."
	  (send self :symbol)
	  (send self :pack)))

(defmethod (name-conflict-error :case :document-proceed-type :import-accessible-by-inheritance) (stream ignore)
  (format stream "Import ~s into the ~a package, shadowing the conflicting symbol accessible by inheritance."
	  (send self :symbol)
	  (send self :pack)))

1;;;  UNINTERN can only cause a name conflict when a shadowing symbol is uninterned.*
(defmethod (name-conflict-error :case :document-proceed-type :unintern) (stream ignore)
  (format stream "Unintern ~s from the ~a package and resolve name conflicts."
	  (send self :symbol)
	  (send self :pack)))

1;;; Redefine the :REPORT method to print a suitable error message.*
(defmethod (name-conflict-error :report) (stream)
  "Print condition message on STREAM."
  (format stream "Name Conflicts"))

1;;;  Methods for collecting arguments for the various proceed types.*

(defmethod (name-conflict-error :case :proceed-asking-user :skip)
	   (continuation ignore)
  "Return without doing the requested operation."
      (funcall continuation :skip))


1;;; CONFLICTS-LIST looks like: ((<sym> <pkg>) ... ) where PKG is a package that will
;;; inherit SYM, the symbol being exported.*
(defmethod (name-conflict-error :case :proceed-asking-user :export-present)
	   (continuation read-object-function)
  (funcall continuation :export-present
	   (dolist (conflict (send self :conflicts-list))
	     (let ((sym (car conflict))
		   (pkg (package-name (cdr conflict))))
	       (if (funcall read-object-function
			    '(:fquery . (:type :tyi :choices (((t "Unintern") #\U)
							      ((nil "Shadow") #\S))))
			    (format nil
				    "~%Unintern ~s from the ~a package (U) ~
~%or ~%Make ~s a shadowing symbol in the ~a package (S)?  "
				    sym pkg sym pkg))
		   (unintern sym pkg)
		   (shadow sym pkg))))))

(defmethod (name-conflict-error :case :proceed-asking-user :export-accessible-by-inheritance)
	   (continuation read-object-function)
  (funcall continuation :export-accessible-by-inheritance
	   (dolist (conflict (send self :conflicts-list))
		   (let* ((sym (car conflict))
			  (inheriting-pkg (package-name (cdr conflict)))
			  (pkg-exporting-from (send self :pack))
			  (pkg-accessible-in (multiple-value-bind (ignore ignore pack)
						 (find-symbol (symbol-name sym) inheriting-pkg)
					       (package-name pack))))
		     (if (funcall read-object-function
				  `(:fquery . (:type :readline :choices ((t ,pkg-exporting-from)
								    (nil ,pkg-accessible-in))))
				  (format nil
					  "~%Make ~? accessible as a shadowing symbol in the ~a package (~a) ~
~%or ~%Make ~? accessible as a shadowing symbol in the ~a package (~a)?  "
					  "~a:~a" `(,pkg-exporting-from ,sym) inheriting-pkg pkg-exporting-from 
					  "~a:~a" `(,pkg-accessible-in ,sym) inheriting-pkg pkg-accessible-in))
			 (shadowing-import (find-symbol (symbol-name sym) pkg-exporting-from) inheriting-pkg)
			 (shadowing-import (find-symbol (symbol-name sym) pkg-accessible-in) inheriting-pkg))))))

1;;; In this case, CONFLICTS-LIST looks like: 
;;; ((<set-of-directly-conflicting-symbols>) . (<set-of-inherited-conflicting-symbols>)). 
;;; Each set consists of items of the form (SYM PKG) where PKG is a package that will
;;; inherit SYM, the symbol being exported.*
(defmethod (name-conflict-error :case :proceed-asking-user :export-both-conflict-types)
	   (continuation read-object-function)
  (funcall continuation :export-both-conflict-types
	   (let ((conflicts-list (send self :conflicts-list)))
	1     ;; Process conflicts with symbols directly present in the inheriting package.*
	   (dolist (conflict (car conflicts-list))
		   (let ((sym (car conflict))
			 (pkg (package-name (cdr conflict))))
		     (if (funcall read-object-function
				  '(:fquery . (:type :tyi :choices (((t "Unintern") #\U)
								    ((nil "Shadow") #\S))))
				  (format nil
					  "~%Unintern ~s from the ~a package (U) ~
~%or ~%Make ~s a shadowing symbol in the ~a package (S)?  "
					  sym pkg sym pkg))
			 (unintern sym pkg)
			 (shadow sym pkg))))
	   (dolist (conflict (cdr conflicts-list))
		   1;; Process conflicts with symbols accessible by inheritance in the inheriting package.*
		   (let* ((sym (car conflict))
			  (inheriting-pkg (package-name (cdr conflict)))
			  (pkg-exporting-from (send self :pack))
			  (pkg-accessible-in (multiple-value-bind (ignore ignore pack)
						 (find-symbol (symbol-name sym) inheriting-pkg)
					       (package-name pack))))
		     (if (funcall read-object-function
				  `(:fquery . (:type :readline :choices ((t ,pkg-exporting-from)
									 (nil ,pkg-accessible-in))))
				  (format nil
					  "~%Make ~? accessible as a shadowing symbol in the ~a package (~a) ~
~%or ~%Make ~? accessible as a shadowing symbol in the ~a package (~a)?  "
					  "~a:~a" `(,pkg-exporting-from ,sym) inheriting-pkg pkg-exporting-from 
					  "~a:~a" `(,pkg-accessible-in ,sym) inheriting-pkg pkg-accessible-in))
			 (shadowing-import (find-symbol (symbol-name sym) pkg-exporting-from) inheriting-pkg)
			 (shadowing-import (find-symbol (symbol-name sym) pkg-accessible-in) inheriting-pkg)))))))

1;;; CONFLICTS-LIST looks like: ((<sym> <pkg>) ... ) where PKG is a package that will
;;; inherit SYM, the symbol being exported.*
(defmethod (name-conflict-error :case :proceed-asking-user :use-package-present)
	   (continuation read-object-function)
  (funcall continuation :use-package-present
	   (dolist (conflict (send self :conflicts-list))
	     (let ((sym (car conflict))
		   (pkg (package-name (cdr conflict))))
	       (if (funcall read-object-function
			    '(:fquery . (:type :tyi :choices (((t "Unintern") #\U)
							      ((nil "Shadow") #\S))))
			    (format nil
				    "~%Unintern ~s from the ~a package (U) ~
~%or ~%Make ~s a shadowing symbol in the ~a package (S)?  "
				    sym pkg sym pkg))
		   (unintern sym pkg)
		   (shadow sym pkg))))))

(defmethod (name-conflict-error :case :proceed-asking-user :use-package-accessible-by-inheritance)
	   (continuation read-object-function)
  (funcall continuation :use-package-accessible-by-inheritance
	   (dolist (conflict (send self :conflicts-list))
		   (let* ((sym (car conflict))
			  (inheriting-pkg (package-name (cdr conflict)))
			  (pkg-to-use (send self :pack))
			  (pkg-accessible-in (multiple-value-bind (ignore ignore pack)
						 (find-symbol (symbol-name sym) inheriting-pkg)
					       (package-name pack))))
		     (if (funcall read-object-function
				  `(:fquery . (:type :readline :choices ((t ,pkg-to-use)
								    (nil ,pkg-accessible-in))))
				  (format nil
					  "~%Make ~? accessible as a shadowing symbol in the ~a package (~a) ~
~%or ~%Make ~? accessible as a shadowing symbol in the ~a package (~a)?  "
					  "~a:~a" `(,pkg-to-use ,sym) inheriting-pkg pkg-to-use 
					  "~a:~a" `(,pkg-accessible-in ,sym) inheriting-pkg pkg-accessible-in))
			 (shadowing-import (find-symbol (symbol-name sym) pkg-to-use) inheriting-pkg)
			 (shadowing-import (find-symbol (symbol-name sym) pkg-accessible-in) inheriting-pkg))))))

1;;; In this case, CONFLICTS-LIST looks like: 
;;; ((<set-of-directly-conflicting-symbols>) . (<set-of-inherited-conflicting-symbols>)). 
;;; Each set consists of items of the form (SYM PKG) where PKG is a package that will
;;; inherit SYM, the symbol made accessible by USE-PACKAGE.*
(defmethod (name-conflict-error :case :proceed-asking-user :use-package-both-conflict-types)
	   (continuation read-object-function)
  (funcall continuation :use-package-both-conflict-types
	   (let ((conflicts-list (send self :conflicts-list)))
	1     ;; Process conflicts with symbols directly present in the inheriting package.*
	   (dolist (conflict (car conflicts-list))
		   (let ((sym (car conflict))
			 (pkg (package-name (cdr conflict))))
		     (if (funcall read-object-function
				  '(:fquery . (:type :tyi :choices (((t "Unintern") #\U)
								    ((nil "Shadow") #\S))))
				  (format nil
					  "~%Unintern ~s from the ~a package (U) ~
~%or ~%Make ~s a shadowing symbol in the ~a package (S)?  "
					  sym pkg sym pkg))
			 (unintern sym pkg)
			 (shadow sym pkg))))
	   (dolist (conflict (cdr conflicts-list))
		   1;; Process conflicts with symbols accessible by inheritance in the inheriting package.*
		   (let* ((sym (car conflict))
			  (inheriting-pkg (package-name (cdr conflict)))
			  (pkg-being-used (send self :pack))
			  (pkg-accessible-in (multiple-value-bind (ignore ignore pack)
						 (find-symbol (symbol-name sym) inheriting-pkg)
					       (package-name pack))))
		     (if (funcall read-object-function
				  `(:fquery . (:type :readline :choices ((t ,pkg-being-used)
									 (nil ,pkg-accessible-in))))
				  (format nil
					  "~%Make ~? accessible as a shadowing symbol in the ~a package (~a) ~
~%or ~%Make ~? accessible as a shadowing symbol in the ~a package (~a)?  "
					  "~a:~a" `(,pkg-being-used ,sym) inheriting-pkg pkg-being-used
					  "~a:~a" `(,pkg-accessible-in ,sym) inheriting-pkg pkg-accessible-in))
			 (shadowing-import (find-symbol (symbol-name sym) pkg-being-used) inheriting-pkg)
			 (shadowing-import (find-symbol (symbol-name sym) pkg-accessible-in) inheriting-pkg)))))))


1;;;  Used by both EXPORT and USE-PACKAGE.*
(defmethod (name-conflict-error :case :proceed-asking-user :unintern-all)
	   (continuation ignore)
  (funcall continuation :unintern-all
	   (dolist (conflict (send self :conflicts-list))
		   (let ((sym (car conflict))
			 (pkg (package-name (cdr conflict))))
		     (unintern sym pkg)))))

1;;;  Used by both EXPORT and USE-PACKAGE.*
(defmethod (name-conflict-error :case :proceed-asking-user :shadow-all)
	   (continuation ignore)
  (funcall continuation :shadow-all
	   (dolist (conflict (send self :conflicts-list))
		   (let ((sym (car conflict))
			 (pkg (package-name (cdr conflict))))
		     (shadow sym pkg)))))

(defmethod (name-conflict-error :case :proceed-asking-user :import-present)
	   (continuation ignore)
  (funcall continuation :import-present
	   (unintern (send self :symbol) (send self :pack))))

(defmethod (name-conflict-error :case :proceed-asking-user :import-accessible-by-inheritance)
	   (continuation ignore)
  (funcall continuation :import-accessible-by-inheritance
	   (shadowing-import (send self :symbol) (send self :pack))))

(defmethod (name-conflict-error :case :proceed-asking-user :unintern)
	   (continuation read-object-function)
  (funcall continuation :unintern
	   (let* ((conflicts (send self :conflicts-list))
		  (inheriting-pkg (package-name (send self :pack)))
		  (choices (mapcar #'(lambda (confl)
				      (list (cdr confl) (format nil "~a" (package-name (cdr confl)))))
				  conflicts))
		  (pkg-chosen (funcall read-object-function
				       `(:fquery . (:type :readline :choices ,choices))
				       (let ((control-string
					       (if (cddr conflicts)
						   ;;  Ensure the choices are always printed with prefixes 
						   ;;  since their p-names are equivalent.
						   "Make (~{~a:~a~#[~;~;, or ~:;, ~]~}) present as a shadowing ~
symbol in the ~a package?  "
						   "Make (~{~a:~a~#[~;~; or ~]~}) present as a shadowing ~
symbol in the ~a package?  ")))
					 
					 (format nil
						 control-string
						 (mapcan #'(lambda (pair)
							     `(,(package-name (cdr pair)) ,(car pair)))
							 conflicts)
						 inheriting-pkg))))
		  (winner (rassoc (find-package pkg-chosen) conflicts :test #'eq)))
	     (shadowing-import (find-symbol (car winner) pkg-chosen) inheriting-pkg))))
	   

1;; FAILED-ASSERTION*

;1;; 7/20/88 mjf/clm - changed for CLEH (uses simple-error instead of error).*
(defflavor failed-assertion (places) (simple-error)
  :inittable-instance-variables
  :gettable-instance-variables)

(defmethod (failed-assertion :or :document-proceed-type)
	   (proceed-type stream ignore)
  (and (member proceed-type places :test #'eq) 
       (or (format stream "Ask for a replacement value for ~S and proceed." proceed-type)
	   t)))

(defmethod (failed-assertion :or :proceed-asking-user)
	   (proceed-type continuation read-object-function)
  (when (member proceed-type places :test #'eq) 
1    ;; Resume stack group and return the proceed-type and the new value*
    (funcall continuation proceed-type
	     (funcall read-object-function :eval-read
		      "Form to evaluate and use as a replacement value for ~S: "
		      proceed-type))
    t))

;1;; 7/20/88 mjf/clm - added for CLEH.*
(defmethod (failed-assertion :case :proceed-asking-user :no-action) (continuation ignore)
  2"Retry assertion."*
   ;1; Resume stack group and return the proceed type*
  (funcall continuation :no-action))

1;; REDEFINITION Warning*

(defflavor redefinition () (warning))

(defmethod (redefinition :case :proceed-asking-user :proceed) (continuation ignore)
  2"Proceed, and do not ask again for redefinitions of that file by this one."*
  (funcall continuation :proceed))

(defmethod (redefinition :case :proceed-asking-user :inhibit-definition) (continuation ignore)
  2"Proceed without performing the redefinition."*
  (funcall continuation :inhibit-definition))

(defsignal sys:redefinition redefinition (definition-type name new-pathname old-pathname)
  2"Name's definition of type DEFINITION-TYPE was redefined in a different file.
The old definition was in OLD-PATHNAME and the new one is in NEW-PATHNAME.
Both of the last two are generic pathnames or NIL."*)



1;; DISK-ERROR*

(defflavor disk-error () (error))

(defmethod (disk-error :dangerous-condition-p) ()
1  ;; It's best to encourage handlers to let the user see dangerous errors!*
  t)

(defmethod (disk-error :case :proceed-asking-user :retry-disk-operation)
	   (continuation read-object-function)
  2"Try the disk operation again."*
  (if (funcall read-object-function '(:fquery) "Retry the disk operation? ")
      1;; Resume stack group and return proceed type*
      (funcall continuation :retry-disk-operation)))

(defsignal sys:disk-error disk-error ()
  2"A fatal disk error happened."*)

1;; TAPE-ERROR*

(defflavor tape-error () (error))

(defsignal si:tape-error tape-error ()
  2"A fatal tape error happened."*)

(defsignal si:tape-media-error tape-error ()
2  "A media error has occurred on tape."*)

(defmethod (tape-error :case :proceed-asking-user :retry-tape-operation)
	   (continuation read-object-function)
  2"Try the tape operation again."*
  (if (funcall read-object-function '(:fquery) "Retry the tape operation? ")
      1;; Resume stack group and return proceed type*
      (funcall continuation :retry-tape-operation)))


1;;; Explorer specific UNIT-ERROR*

(defflavor unit-error (unit) (error)
  :abstract-flavor
  :gettable-instance-variables
  :inittable-instance-variables)

(defmethod (unit-error :after :init) (ignore)
  (setq unit (get (locf si:property-list) :unit)))

(defmethod (unit-error :case :proceed-asking-user :new-unit)
	   (continuation read-object-function)
  2"Read a new unit and use that instead."*
  1;; Resume stack group and return proceed type* and a new unit number
  (funcall continuation :new-unit
	   (setq unit (funcall read-object-function
			       `(:number :defaults ,unit)
			       "Unit to use instead: (default ~A) "
			       unit))))

1;; UNIT-BAD-NUMBER-ERROR*

(defflavor unit-bad-number-error () (unit-error no-action-mixin))

(defsignal bad-number-error unit-bad-number-error (unit)
  2"Unit number is out of range."*)

1;; UNIT-OFFLINE-ERROR*

(defflavor unit-offline-error () (unit-error))

(defmethod (unit-offline-error :case :proceed-asking-user :configure-disk-system-and-retry)
	   (continuation ignore)
  2"Re-configure the disk-type-table, looking for online units."*
  (si:initialize-disk-system)
1  ;; Resume stack group and return proceed type to retry*
  (send continuation :configure-disk-system-and-retry))

(defsignal offline-error unit-offline-error (unit)
  2"Unit number is offline."*)



1;; Conventions for the microcode error handler routines:  how to add new ones.

;; Each place in the microcode where TRAP can be called is followed by
;; an ERROR-TABLE pseudo-op.  This should appear at the PC which is
;; going to be TRAP's return address.  An example is
;;    (ERROR-TABLE ARGTYP FIXNUM M-T 0)
;; (for example).  The CDR of this list is the ETE.  So, the FIRST element
;; is the name of the error, and the SECOND is the first "argument" to that
;; error's associated routines.
;;
;; All ETEs should be a list whose car is a ucode symbol.
;; Each ucode symbol should define a (:property ucode-symbol MAKE-UCODE-ERROR-FUNCTION) function
;; which makes an error condition object.
;; This function can be defined with DEF-UCODE-ERROR
;; Within the DEF-UCODE-ERROR, the variable ETE can be used to refer to the entire ETE list.
;;
;; There are many ways for the error-handler and the microcode to communicate about errors.
;; Some error-table entries indicate the locations of the important data (like the ARGTYP example
;; above);  some handlers assume the locations, and some errors just leave the data atop the
;; stack.  Some traps may be restarted by simply continuing, others must be started over at
;; restarts (defined by ERROR-TABLE RESTART in the microcode);  some handlers pick up the
;; restart from the ETE, others assume it (like TRANS-TRAP).  Some traps are distinguished by
;; different error-table entries, while others use the same ETE type and are distinguished by the
;; "arguments."
;;
;; The only rule is that the microcode and the error-handler must agree on method of communication.
;; Changes in the microcode may necessitate changes in the handlers, and it is the responsibility of
;; both parties to keep the two entities in sync.*

(defmacro def-ucode-error (error-name error-flavor &body init-options)
  2"Defines a microcode error.
ERROR-NAME is the ucode error name to be defined.
ERROR-FLAVOR is the flavor of the condition object to make,
 or a list (FLAVOR . ADDITIONAL-CONDITION-NAMES), where ADDITIONAL-CONDITION-NAMES
 is a list of condition-names to signal in addition to the flavor components.
 If you specify just a flavor name, the ERROR-NAME itself is used
 as the sole additional condition name.
INIT-OPTIONS are passed to MAKE-INSTANCE of ERROR-FLAVOR"*
  (let ((error-flavor (if (consp error-flavor) (car error-flavor) error-flavor))
	(condition-names (if (consp error-flavor) (cdr error-flavor) (list error-name))))
    `(progn (si:record-source-file-name ',error-name 'defsignal)
	    (defun (:property ,error-name make-ucode-error-function) (ignore sg ete)
	      (declare (sys:function-parent ,error-name))
	      sg ete
	      (make-instance ',error-flavor
			     :condition-names ',condition-names
			     ,@init-options)))))

(defmacro def-ucode-format-error (error-name error-flavor format-string &body format-args)
2"Defines a microcode error.
ERROR-NAME is the ucode error name to be defined.
ERROR-FLAVOR is the flavor of the condition object to make,
 or a list (FLAVOR . ADDITIONAL-CONDITION-NAMES), where ADDITIONAL-CONDITION-NAMES
 is a list of condition-names to signal in addition to the flavor components.
 If you specify just a flavor name, the ERROR-NAME itself is used
 as the sole additional condition name.
:FORMAT-STRING FORMAT-STRING and :FORMAT-ARGS FORMAT-ARGS are
init options for ERROR-FLAVOR"*
  `(def-ucode-error ,error-name ,error-flavor
     :format-string ,format-string
     :format-args (list . ,format-args)))

(def-ucode-format-error unknown-microcode-trap error
  "Unknown microcode trap.  No error table entry for PC ~S2.*"
  (second ete))

(defun sg-proceed-micro-pc (sg tag)
  2"Restart SG from an error in the microcode, at micro-pc determined by TAG.
TAG may be the name of an (ERROR-TABLE RESTART tag) in the microcode source.
If TAG is NIL, restart at the pc where the error happened.
If a PROCEED routine doesn't call SG-PROCEED-MICRO-PC, then
control will be returned from the micro-routine that got the error."*

  1;; Operates by pushing the specified PC onto the saved microstack,*
  1;; since resuming the stack group will do a POPJ.*
  (declare (special *restart-list*))
  (let ((pc (if tag
		(cdr (assoc tag *restart-list* :test #'eq))
		(1+ (sg-trap-micro-pc sg)))))
    (when (null pc) 1;; this should never happen*
      (bad-hacker tag "- no such restart tag")
      (throw 'quit nil))
    (let ((rp (sg-regular-pdl sg))
	  (sp (sg-special-pdl sg))
	  (spp (sg-special-pdl-pointer sg))
	  (frame (sg-top-frame sg))
	  function)
      1;; microstack state of top frame should be in sg struct*
      (unless (zerop (sg-restore-microstack sg))
	1;; the micro stack is saved backwards;*
	1;; the top of the stack is buried where it is hard to get at;*
	1;; need to shuffle stack up one to make room for PC*
	(do ((flag 0)) ((not (zerop flag)))
	  (setf (aref sp (1+ spp)) (aref sp spp))
	  (%p-dpb 0 %%specpdl-block-start-flag (locf (aref sp (1+ spp))))
	  (setq flag (%p-ldb %%specpdl-block-start-flag (locf (aref sp spp))))
	  (setq spp (1- spp))))
      1;; push PC onto Special Pdl & update microstack state*
      (setf (aref sp (setq spp (1+ spp))) pc) 
      (%p-dpb 1 %%specpdl-block-start-flag (locf (aref sp spp)))
      (setf (sg-special-pdl-pointer sg) (1+ (sg-special-pdl-pointer sg)))
      (setf (sg-restore-microstack sg) 1)
1       ;; This is needed for Explorer;
       ;; in some cases, we need to zero out the exit pc*
      (setq function (rp-function-word rp frame))
      (when (and (typep function 'compiled-function) 
		 (> (rp-location-counter-offset rp frame) (* 2 (fef-length function))))
	(setf (rp-location-counter-offset rp frame) 0)))))


1;; ARG-TYPE-ERROR
;; used by ARGTYP and FLONUM-NO-GOOD *
;1;; 7/20/88 mjf/clm -changed for CLEH to use type error.*
(defflavor arg-type-error
	   (arg-number function arg-pointer arg-data-type description
	    arg-location-in-sg restart-tag)
	   (type-error)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (arg-type-error :arg-name) ()
1  ;; This works for microcoded functions.*
  (if arg-number (nth arg-number (arglist function))))

(defmethod (arg-type-error :old-value) ()
  (%make-pointer arg-data-type arg-pointer))

(defmethod (arg-type-error :report) (stream)
  (format:output stream
    (format t "~:[Some~*~*~;The~:[ ~:R~;~*~]~] argument to ~S"
	    arg-number (eq arg-number t) (if (numberp arg-number) (1+ arg-number))
	    function)
    1;; print the arg*
    (cond ((or (null arg-data-type)
	       (null arg-pointer)))1  ;Do nothing if FLONUM-NO-GOOD error*
	  (t
	   (format t ", ")
	   (if (consp (errset (%make-pointer arg-data-type arg-pointer) nil))
	       1;;no error with this data type*
	       (prin1 (%make-pointer arg-data-type arg-pointer))
	     1;;else*
	     (format t "#<~S ~O>" (nth arg-data-type q-data-types) arg-pointer))
	   (format t ",")))
1     ;; the error*
    (format t (if (arrayp function)
		  " was an invalid array subscript.~%Use "      
		" was of the wrong type.~%The function expected "))
1     ;; print description describing what the type should be*
    (cond ((symbolp description)
	   (princ-type-name description))
	  ((listp description)
	   (if (< (length description) 2)
	       (princ-type-name (car description))
	     (do ((tl description (cdr tl)))
		 ((null tl))
	       (princ-type-name (car tl))
	       (princ (cond ((null (cdr tl)) "")
			    ((null (cddr tl)) " or ")
			    (t ", "))))))
	  (t (bad-hacker description "- not a valid type in ARGTYP error")))
    "."))

(defun princ-type-name (type)
  2"Print a string describing the meaning of TYPE, a type name found in the error table."*
  (declare (special *data-type-names*))
  (princ (or (cadr (assoc type *data-type-names* :test #'eq)) type)))

(defparameter *data-type-names*
	      '((fixnum "a fixnum")
		(number "a number")
		(bignum "a bignum")
		(integer "a fixnum or a bignum")
		(float "a floating-point number")
		(single-float "a single-precision floating-point number")
		(double-float "a double-precision floating-point number")
		(positive-fixnum "a fixnum  zero")
		(nonnegative-fixnum "a nonnegative fixnum")
		(nonnegative-integer "a nonnegative integer")	;ab 8-5-88
		(positive-integer "a positive integer")	;ab
		(negative-integer "a negative integer")	;ab
		(32-bit-unsigned-integer "a 32-bit unsigned integer value")	;ab
		(fixnum-greater-than-0 "a fixnum > zero")
		(fixnum-greater-than-1 "a fixnum > 1")
		(plusp "a number > zero")
		(zero "the number zero")	;ab
		(symbol "a symbol")
		(cons "a cons")
		(list "a list")
		(nil "the symbol NIL")
		(non-nil "something other than NIL")
		(locative "a locative")
		(symbol-or-locative "a symbol or locative")
		(closure "a closure")
		(instance "an instance")
		(stack-group "a stack group")
		(array "an array")
		(art-q-list-array "an ART-Q-LIST array")
		(q-array "an array of Lisp objects")
		(byte-array "an array of numbers")
		(8-bit-or-16-bit-array "an 8 or 16-bit byte array")
		(art-4b-array "an array of 4-bit bytes")
		(art-16b "an array of 16-bit bytes")
		(non-displaced-array "a non-displaced array")
		(non-physical-array "a non-physical array")	;ab
		(array-without-displaced-index-offset "array with no displaced-index-offset")	;ab
		(art-q-array "an array of type ART-Q")
		(numeric-array "an array of numeric type")
		(reasonable-size-array "an array of reasonable size")
		(area "an area number, NIL (default), or a symbol whose value is one")
		(fixnum-field "a byte pointer to a field that fits in a fixnum")
		(character "a character object")
                (small-bignum-or-fixnum "a integer between 2,147,483,647 and -2,147,483,648 inclusive"))
  2"Data type descriptions."*)

(defmethod (arg-type-error :ucode-proceed-types) ()
  1;; ARGTYP traps supply restart tag (FLONUM-NO-GOOD trap never supplies one)*
  (if restart-tag '(:argument-value)))

(defmethod (arg-type-error :case :proceed-asking-user :argument-value)
	   (continuation read-object-function)
  2"Ask for a replacement argument and proceed."*
  1;; Resume stack group and return the proceed type and new arg value*
  (funcall continuation :argument-value
	   (funcall read-object-function :eval-read
		    "Form to evaluate and use as a replacement argument: ")))

(defmethod (arg-type-error :case :proceed-ucode-with-args :argument-value)
	   (sg value &rest ignore)
  (sg-store value sg arg-location-in-sg)
  (if (consp restart-tag)
1      ;; ARGTYP traps can have a list of restart tags, like SUBSCRIPT-OOB errors.
      ;; This allows the restart handler to be a subroutine that returns to the point of the error.*
      (dolist (tag restart-tag)
	(sg-proceed-micro-pc sg tag))
1    ;; restart sg at pc for restart tag or at pc where error happened*
    (sg-proceed-micro-pc sg (if (neq restart-tag 'fall-through) restart-tag))))

1;; ARGTYP
;; First arg is the arg type description expected (like FIXNUM, LIST, SYMBOL, ...).  See variable *data-type-names*.
;; Second arg is the location in stack group of bad arg.
;; Third arg is the arg number (zero based) or T (for "the one arg") or NIL (for "one of the args").
;; Fourth arg is the restart tag.  If it is NIL or missing, then there is no way to recover.
;; It can also be a list of restart tags.  Then their addresses are pushed sequentially.
;; This is used to get the effect of making the microcode restart by calling
;; a subroutine which will return to the point of the error.
;; Fifth arg (optional) is the name of function with the bad arg type.*

(def-ucode-error argtyp (arg-type-error wrong-type-argument)
  :description (second ete)
  :arg-location-in-sg (third ete)
  :arg-pointer (sg-pointer-careful sg (third ete))
  :arg-data-type (sg-data-type sg (third ete))
  :arg-number (fourth ete)
  :restart-tag (fifth ete)
  :function 
1     ;; use function supplied in ete or
     ;; sg-erring-function or function in last frame*
     (or (sixth ete)
	 (and (eq (fifth ete) 'array-decode-n-error-restart)
	      (not (member (sg-erring-function sg) '(aref zlc:aset aloc) :test #'eq))
	      (rp-function-word (sg-regular-pdl sg) (sg-top-frame sg)))
	 (sg-erring-function sg)))

(defun sg-data-type (sg register)
  (%p-data-type (sg-locate sg register)))

(defun sg-pointer-careful (sg register)
  (let* ((location (sg-locate sg register)))
    (cond ((%p-contents-safe-p location)
	   (contents location))
	  ((%p-pointerp location)
	   (%p-contents-as-locative location))
	  (t
	   (%p-pointer location)))))

1;; FLONUM-NO-GOOD*

(def-ucode-error flonum-no-good (arg-type-error wrong-type-argument)
  :description 'fixnum
  :arg-location-in-sg nil
  :arg-pointer nil
  :arg-data-type nil
  :arg-number nil
  :restart-tag nil
  :function (sg-erring-function sg))



1;; FLOATING-EXPONENT-UNDERFLOW
;; FLOATING-EXPONENT-UNDERFLOW SFL
;; FLOATING-EXPONENT-UNDERFLOW FLO
;; FLOATING-EXPONENT-UNDERFLOW DBL
;; First Arg is SFL, FLO, or DBL*

(def-ucode-error floating-exponent-underflow floating-exponent-underflow-error 
  :flonum-type (second ete)
  :function (sg-erring-function sg))

;1;; 7/20/88 mjf/clm - changed for CLEH.  Changed floating-exponent-underflow-error to use the*
;1;; new floating-point-underflow flavor*
(defflavor floating-point-underflow () (arithmetic-error))
(defflavor floating-exponent-underflow-error (function flonum-type) (floating-point-underflow)
  :gettable-instance-variables
  :initable-instance-variables)

(defun flonum-name (flonum-type)
   (ccase flonum-type
     (sfl "short-float")
     (flo "single-float")
     (dbl "double-float")))

(defmethod (floating-exponent-underflow-error :report) (stream)
  (format stream "~S produced a result too small in magnitude to be a ~A."
	  function
	  (flonum-name flonum-type)))

(defmethod (floating-exponent-underflow-error :ucode-proceed-types) ()
  '(:use-zero))

(defmethod (floating-exponent-underflow-error :case :document-proceed-type :use-zero)
	   (stream ignore)
  ;; 03/25/87 CLM - Add this method to fix proceed type documentation problem (SPR4133).
  (format stream 2"Use 0.0*~A2 as the result instead."*
	  (case flonum-type
		   (sfl "s0")
		   (dbl "d0")
		   (t ""))))

(defmethod (floating-exponent-underflow-error :case :proceed-asking-user :use-zero)
	   (continuation read-object-function)
  2"Use 0.0 as the result instead."*
  ;; 03/25/87 CLM -  Added T clause to CASE to fix print-out problem (SPR4133).
  (when (funcall read-object-function
		 '(:fquery :list-choices nil :fresh-line nil)
		 "Proceed using 0.0~A as the value instead? "
		 (case flonum-type
		   (sfl "s0")
		   (dbl "d0")
		   (t ""))) ;added the last clause 'flo to fix spr4133
1    ;; Resume stack group return proceed type*
    (funcall continuation :use-zero)))

(defmethod (floating-exponent-underflow-error :case :proceed-ucode-with-args :use-zero)
	   (sg &rest ignore)
  1;; Restart sg at pc where error happened*
  (sg-proceed-micro-pc sg nil))

1;; FLOATING-EXPONENT-OVERFLOW
;; FLOATING-EXPONENT-OVERFLOW SFL
;; FLOATING-EXPONENT-OVERFLOW FLO
;; FLOATING-EXPONENT-OVERFLOW DBL
;; First arg is SFL, FLO, or DBL
;; Result is to be placed in M-T*

(def-ucode-error floating-exponent-overflow floating-exponent-overflow-error
  :flonum-type (second ete)
  :function (sg-erring-function sg))

;1;; 7/20/88 mjf/clm - changed for CLEH.  Changed floating-exponent-overflow-error to use *
;1;; the new floating-point-overflow.*
(defflavor floating-point-overflow () (arithmetic-error))
(defflavor floating-exponent-overflow-error (function flonum-type) (floating-point-overflow)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (floating-exponent-overflow-error :report) (stream)
  (format stream "~S produced a result too large in magnitude to be a ~A."
	  function
	  (flonum-name flonum-type)))

(defmethod (floating-exponent-overflow-error :ucode-proceed-types) ()
  '(:new-value))

(defmethod (floating-exponent-overflow-error :case :proceed-asking-user :new-value)
	   (continuation read-object-function)
  2"Ask for a flonum to use as the result and proceed."*
  1;; Resume stack group and return proceed type and new value*
  (funcall continuation :new-value
	   1;; loop until get a flonum*
	   (do ((num (funcall read-object-function :eval-read
			      "Form to evaluate and return instead: "
			      )))
	       (())
	     (when (or (and (small-floatp num) (eq flonum-type 'sfl))
		       (and (typep num 'single-float) (eq flonum-type 'flo))
		       (and (typep num 'double-float) (eq flonum-type 'dbl)))
	       (return num))
	     (format t "Please use a ~A.~%" (flonum-name flonum-type)))))

(defmethod (floating-exponent-overflow-error :case :proceed-ucode-with-args :new-value)
	   (sg value &rest ignore)
  1;; new value is expected in M-T*
  (sg-store value sg 'm-t))


1;; DIVIDE-BY-ZERO
;; The 1st arg of this ETE can be the location of the dividend.
;; You cannot recover from this error*
;1;; 7/20/88 mjf/clm - changed for CLEH.  Changed divide-by-zero to use the new*
;1;; division-by-zero flavor.*
(defflavor division-by-zero () (arithmetic-error))
(defun (:property divide-by-zero make-ucode-error-function) (ignore sg ete)
  (make-instance 'division-by-zero
		 :condition-names (if (eq (sg-erring-function sg) 'expt)
				      '(sys:zero-to-negative-power)
				      '(sys:divide-by-zero))
		 :property-list (if (second ete)
				    `(:function ,(sg-erring-function sg)
				      :dividend ,(sg-contents sg (second ete)))
				    `(:function ,(sg-erring-function sg)))
		 :format-string (if (eq (sg-erring-function sg) 'expt)
				    "There was an attempt to raise zero to a negative power in ~S."
				    "There was an attempt to divide a number by zero in ~S.")
		 :format-args (list (sg-erring-function sg))))


1;; Array ucode errors

;; BAD-ARRAY-MIXIN*
(defflavor bad-array-mixin (array array-location-in-sg restart-tag) ()
  (:required-flavors error)
  :gettable-instance-variables
  :inittable-instance-variables)

(defmethod (bad-array-mixin :ucode-proceed-types) ()
  (if restart-tag '(:new-array)))

(defmethod (bad-array-mixin :case :proceed-asking-user :new-array)
	   (continuation read-object-function)
  2"Read an array to use instead."
1  ;; Resume stack group and return proceed type and new value **
  (funcall continuation :new-array
	   (funcall read-object-function :eval-read
		    "Form to evaluate and use as the array to use instead: ")))

(defmethod (bad-array-mixin :case :proceed-ucode-with-args :new-array)
	   (sg -array-)
1  ;; store new array and restart sg at pc for restart tag*
  (sg-store -array- sg array-location-in-sg)
  (sg-proceed-micro-pc sg restart-tag))

1;; BAD-ARRAY-ERROR 
;; Used by ucode errors: ARRAY-HAS-NO-LEADER,
;; NUMBER-ARRAY-NOT-ALLOWED, and FILL-POINTER-NOT-FIXNUM *
(defflavor bad-array-error () (bad-array-mixin error))

1;; ARRAY-NUMBER-DIMENSIONS-ERROR 
;; used by ucode error ARRAY-NUMBER-DIMENSIONS*
(defflavor array-number-dimensions-error (subscripts-used) (bad-array-error)
  :initable-instance-variables
  :gettable-instance-variables)

(defmethod (array-number-dimensions-error :dimensions-expected) () (array-rank array))

(defmethod (array-number-dimensions-error :dimensions-given) ()
  (length subscripts-used))

(defmethod (array-number-dimensions-error :report) (stream)
  (format stream	 
	  1;; Was this array applied or aref'ed?*
	  (if (eq restart-tag 'qaryr-restart)
	      "The ~D-dimensional array ~S was erroneously applied to ~D argument~:P ~S."
	    "The ~D-dimensional array ~S was given ~D subscript~:P: ~S.")
	  (array-rank array) array (length subscripts-used) subscripts-used))

1;; ARRAY-NUMBER-DIMENSIONS
;; ARRAY-NUMBER-DIMENSIONS M-D 2    M-A
;; ARRAY-NUMBER-DIMENSIONS M-D 1    M-A
;; ARRAY-NUMBER-DIMENSIONS M-D 1    PP
;; ARRAY-NUMBER-DIMENSIONS M-D M-R M-A QARYR-RESTART
;; ARRAY-NUMBER-DIMENSIONS M-D M-R M-A ARRAY-DECODE-N-NEW-ARRAY-RESTART
;; ARRAY-NUMBER-DIMENSIONS M-D 3    M-A ARRAY-DECODE-3-NEW-ARRAY-RESTART
;; ARRAY-NUMBER-DIMENSIONS M-D 2    M-A ARRAY-DECODE-2-NEW-ARRAY-RESTART
;; ARRAY-NUMBER-DIMENSIONS M-D 1    M-A ARRAY-DECODE-1-A
;; ARRAY-NUMBER-DIMENSIONS NIL   NIL  M-A ARRAY-IN-BOUNDS-P
;; First arg is no longer used.
;; Second arg is how many dimensions we want (as a constant), or NIL if variable number.
;; Third arg is the array.
;; Fourth arg is the restart tag
;;  or QARYR if this is array called as function.*

(def-ucode-error array-number-dimensions
		 (array-number-dimensions-error array-wrong-number-of-dimensions)
  :array (sg-contents sg (fourth ete))
  :array-location-in-sg (fourth ete)
  :subscripts-used (array-number-dimensions-subscript-list sg ete)
  :restart-tag (fifth ete))

(defun array-number-dimensions-subscript-list (sg ete)
  (let* ((frame (sg-top-frame sg))
	 (rp (sg-regular-pdl sg))
	 (fn (rp-function-word rp frame)))
    (if (not (integerp (third ete)))
	1;; If number of dimensions is not supplied,*
	1;; this is AREF, ALOC, ASET, SET-AREF, or array called as function.*
	1;; get dimensions from args in stack*
	(do ((p (+ (rp-argument-offset sg rp frame)
		   (- (rp-number-of-arguments rp frame)
		      (if (eq fn #'set-aref) 2 1)))
		(1- p))
	     subscripts
	     (limit (+ (rp-argument-offset sg rp frame)
		       (cond ((eq fn #'zlc:aset) 2)
			     ((eq fn #'set-aref) 1)
			     ((typep fn 'array) 0)
			     (t 1)))))
	    ((< p limit) subscripts)
	  (push (aref rp p) subscripts))
1      ;; ELSE
      ;; (third ete) tells us whether the error was from ARRAY-DECODE-1, -2 or -3.
      ;; get dimensions from registers*
      (let ((dim (case (third ete)
		       (1 (list (sg-fixnum-contents sg 'm-q)))
		       (2 (list (sg-contents sg 'm-j)
				(sg-contents sg 'm-q)))
		       (3 (list (sg-contents sg 'm-i) 
				(sg-contents sg 'm-j) 
				(sg-contents sg 'm-q))))))
1;;  Array-index-order is always T in Release 3, or would be if we supported it.*
;;	(if array-index-order
;;	    dim
;;	    (nreverse dim))
	dim))))

1;; NUMBER-ARRAY-NOT-ALLOWED
;; NUMBER-ARRAY-NOT-ALLOWED M-A
;; First arg is where to find the array.
;; Second arg is the restart tag for new array.*

(def-ucode-error number-array-not-allowed bad-array-error
  :array (sg-contents sg (second ete))
  :array-location-in-sg (second ete)
  :restart-tag (third ete)
  :format-string "The array ~S, which was given to ~S, is not allowed to be a number array."
  :format-args (list (sg-contents sg (second ete)) (sg-erring-function sg)))

1;; ARRAY-HAS-NO-LEADER
;; ARRAY-HAS-NO-LEADER M-C XFARY-RESTART
;; ARRAY-HAS-NO-LEADER M-A XFLAD1-RESTART
;; First arg is where array pointer is.
;; Second arg is restart tag for new array.*

(def-ucode-error array-has-no-leader bad-array-error
  :array (sg-contents sg (second ete))
  :array-location-in-sg (second ete)
  :restart-tag (third ete)
  :format-string "The array given to ~S, ~S, has no leader."
  :format-args (list (sg-erring-function sg) (sg-contents sg (second ete))))

1;; FILL-POINTER-NOT-FIXNUM
;; FILL-POINTER-NOT-FIXNUM M-C XFARY-RESTART
;; First arg is where array pointer is.
;; Second arg is restart tag for new array.*

(def-ucode-error fill-pointer-not-fixnum bad-array-error
  :array (sg-contents sg (second ete))
  :array-location-in-sg (second ete)
  :restart-tag (third ete)
  :format-string "The fill-pointer of the array given to ~S, ~S, is not a fixnum."
  :format-args (list (sg-erring-function sg) (sg-contents sg (second ete))))

1;;;  BAD-ACTIVE-LENGTH
;;;  BAD-ACTIVE-LENGTH M-D M-T M-A GAHDR
;;;  Trap when the array's fill-pointer is greater than its total length
;;;  Arguments:
;;;     1: fill pointer for array on trap
;;;     2: total length for array on trap
;;;     3: array pointer
;;;     4: restart tag*

(def-ucode-error bad-active-length bad-array-error
  :array (sg-contents sg (fourth ete))
  :array-location-in-sg (fourth ete)
  :restart-tag (fifth ete)
  :format-string "The fill pointer, ~S, of the array, ~S, given to ~S~%  is greater than the total length ~S."
  :format-args (list (sg-contents sg (second ete))
		     (sg-contents sg (fourth ete))
		     (sg-erring-function sg)
		     (sg-contents sg (third ete))))

1;; BAD-ARRAY-TYPE
;; BAD-ARRAY-TYPE M-A
;; BAD-ARRAY-TYPE M-K
;; First arg is where array header is.
;; Note that it may well have a data type of DTP-TRAP. 
;; You cannot recover from this error*

(def-ucode-format-error bad-array-type error
  "The array type, ~S, was invalid in ~S."
  (ldb %%array-type-field (%p-pointer (sg-locate sg (second ete))))
  (sg-erring-function sg))

1;; BAD-ARRAY-DIMENSION-NUMBER
;; BAD-ARRAY-DIMENSION-NUMBER M-A M-C XARRAY-DIMENSION-1
;; First arg is where to find array, 
;;* 1Second arg is where to find dimension number.*

(def-ucode-format-error bad-array-dimension-number error
  "The dimension number ~S is out of range for ~S."
  (sg-fixnum-contents sg (third ete))
  (sg-contents sg (second ete)))



1;; SUBSCRIPT-ERROR
;; Used by ucode errors:  INDIVIDUAL-SUBSCRIPT-OOB and SUBSCRIPT-OOB*

(defflavor subscript-error
	(function subscripts-used subscript-used index-location-in-sg
	 subscript-limit restart-tag object (dimension-number nil))
	(error)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (subscript-error :report) (stream)
  (cond (dimension-number
	 (format stream "The subscript for dimension ~D was ~S, which is out of range for ~S."
		 dimension-number (nth dimension-number subscripts-used) object))
	(object
	 (if (= (length subscripts-used) 1)
	     (format stream "The subscript ~S for ~S was out of range in ~S."
		     subscript-used object function)
	   (format stream "The subscripts ~S for ~S were out of range in ~S."
		   subscripts-used object function)))
	((< subscript-used 0)
	 (format stream "The index, ~S, was negative in ~S."
		 subscript-used function))
	(t 
	 (format stream "The index, ~S, was beyond the length, ~S, in ~S."
		 subscript-used subscript-limit function))))

(defmethod (subscript-error :ucode-proceed-types) ()
  '(:new-subscript))

(defmethod (subscript-error :case :proceed-asking-user :new-subscript)
	   (continuation read-object-function)
  2"Read new subscripts to use instead."*
  (if (or (not (arrayp object))
	  (neq (LENGTH SUBSCRIPTS-USED) (ARRAY-RANK OBJECT)) )
1      ;; get one subscript and resume stack group*
      (let (num)
	(do () (())
	  (setq num (funcall read-object-function :eval-read
			     "Form to evaluate and use as subscript: "))
	  (if (and (integerp num) (< -1 num subscript-limit))
	      (return))
	  (format t "Please use a positive fixnum less than ~D.~%" subscript-limit))
	1;; Resume stack group and return proceed type and new subscript*
	(funcall continuation :new-subscript num))
1    ;; else
    ;; get list of subscript and resume stack group 
 *   (do ((i 0 (1+ i))
	 subscripts)
	((= i (length subscripts-used))
	 (apply continuation :new-subscript (nreverse subscripts)))
      (push (funcall read-object-function :eval-read
		     " Subscript ~D: " i)
	    subscripts))))

(defmethod (subscript-error :case :proceed-ucode-with-args :new-subscript)
	   (sg &rest subscripts) 
  (cond (dimension-number
	 1;; For error in ARRAY-ROW-MAJOR-INDEX,*
	 1;; store back all the subscripts into the frame.*
	 1;; We will restart the function to examine them.*
	 (do ((rp (sg-regular-pdl sg))
	      (frame (sg-top-frame sg))
	      (tail subscripts (cdr tail))
	      (i 0 (1+ i)))
	     ((= i (length subscripts-used)))
	   (setf (aref rp
		       (+ (rp-argument-offset sg rp frame) i 1))	;1 *index to arg i+2
		 (car tail))))
	(:else
	 1;; Error is based on the cumulative index,*
	 1;; so store back the updated cumulative index.*
	 (if (or (not (arrayp object))
		 (neq (LENGTH SUBSCRIPTS-USED) (ARRAY-RANK OBJECT)))
	     (sg-fixnum-store (first subscripts) sg index-location-in-sg)
	     (sg-fixnum-store
	       (do ((i 0 (1+ i))
		    (index 0)
		    (rest subscripts (cdr rest)))
		   ((null rest) index)
		 (setq index (* index (array-dimension object i)))
		 (incf index (car rest)))
	       sg
	       index-location-in-sg))
	 1;; Restart sg at pc for restart tag*
	 (if (consp restart-tag)
	     (dolist (tag restart-tag)
	       (sg-proceed-micro-pc sg tag))
	     (sg-proceed-micro-pc sg restart-tag)))))

1;; INDIVIDUAL-SUBSCRIPT-OOB
;; INDIVIDUAL-SUBSCRIPT-OOB M-A M-C XARRAY-ROW-MAJOR-INDEX
;; First arg is the location of array.
;; Second arg is the dimension number.
;; Third arg is the restart tag.
;; We assume that the current frame's args are the array and the subscripts,
;; and find the actual losing subscript that way.*

(def-ucode-error individual-subscript-oob (subscript-error subscript-out-of-bounds)
  :function (sg-erring-function sg)
  :object (sg-contents sg (second ete))
  :subscripts-used (individual-subscript-oob-subscript-list sg ete)
  :dimension-number (sg-contents sg (third ete))
  :restart-tag (fourth ete))

(defun individual-subscript-oob-subscript-list (sg ete)
  (let* ((object (sg-contents sg (second ete)))
	 (rank (array-rank object))
	 (frame (sg-top-frame sg)))
1    ;; get list of subscripts from arg list for top frame*
    (case rank
      (1 (list (sg-frame-arg-value sg frame 1)))
      (2 (list (sg-frame-arg-value sg frame 1)
	       (sg-frame-arg-value sg frame 2)))
      (3 (list (sg-frame-arg-value sg frame 1)
	       (sg-frame-arg-value sg frame 2)
	       (sg-frame-arg-value sg frame 3))))))

1;; SUBSCRIPT-OOB
;; SUBSCRIPT-OOB M-J M-S
;; SUBSCRIPT-OOB M-T M-S
;; SUBSCRIPT-OOB M-R M-S
;; SUBSCRIPT-OOB M-1 M-2 XINSTANCE-LOC-RESTART   M-T
;; SUBSCRIPT-OOB M-Q RMD XFLAD1-A
;; SUBSCRIPT-OOB M-Q M-S
;; SUBSCRIPT-OOB M-Q M-S (NIL XAREF-RECHECK-INDEX) M-A T
;; SUBSCRIPT-OOB M-Q M-S (NIL XAREF-RECHECK-INDEX) M-A 1
;; SUBSCRIPT-OOB M-Q M-S (NIL XAREF-RECHECK-INDEX) M-A
;; First arg is how many we gave.
;; Second is the legal limit.
;; Third optional arg is a restart tag.
;;  It can also be a list of restart tags.  Then their addresses are pushed sequentially.
;;  This is used to get the effect of making the microcode restart by calling
;;  a subroutine which will return to the point of the error.
;; Fourth optional arg is where the array is.
;; Fifth arg is either T if indices are on the stack,
;;  or 1 if this is AR-1-FORCE or such like and there is only one index (first arg says where)
;;  or missing if this is AR-1, AR-2, AR-3 and the array's rank should be
;;  used to decide where the args are.*

(def-ucode-error subscript-oob (subscript-error subscript-out-of-bounds)
  :function (sg-erring-function sg)
  :index-location-in-sg (second ete)
  :subscript-used (sg-fixnum-contents sg (second ete))
  :subscripts-used (subscript-oob-subscript-list sg ete)
  :subscript-limit (sg-fixnum-contents sg (third ete))
  :restart-tag (fourth ete)
  :object (if (fifth ete) (sg-contents sg (fifth ete))))

(defun subscript-oob-subscript-list (sg ete)
   (let* ((object (if (fifth ete) (sg-contents sg (fifth ete))))
	  (frame (sg-top-frame sg))
	  (rp (sg-regular-pdl sg))
	  (fn (rp-function-word rp frame)))
     (if (and (arrayp object) (neq (sixth ete) 1))
	 (if (sixth ete)
	     1;; This is AREF, ALOC, ASET, SET-AREF, or array called as function.*
	     1;; get subscripts from arg list on stack*
	     (do ((p (+ (rp-argument-offset sg rp frame)
			(- (rp-number-of-arguments rp frame)
			   (if (eq fn #'set-aref) 2 1)))
		     (1- p))
		  subscripts
		  (limit (+ (rp-argument-offset sg rp frame)
			    (cond ((eq fn #'zlc:aset) 2)
				  ((eq fn #'set-aref) 1)
				  ((typep fn 'array) 0)
				  (t 1)))))
		 ((< p limit) subscripts)
	       (push (aref rp p) subscripts))
	     1;; ELSE ((sixth ete) is NIL)*
	     1;; This is AR-1, AR-2 or AR-3 called as a function.*
	     1;; We can tell which array function by the rank of the array.*
	     1;; Get subscripts from first arg of ETE (number of subscripts) and/or SG registers
;;  Array-index-order is always T in Rel 3, if we support it.*
;;	     (if array-index-order
		 (let ((rank (array-rank object)))
		   (case rank
		     (1 (list (sg-fixnum-contents sg (second ete))))
		     (2 (list (sg-contents sg 'm-j)
			      (- (sg-fixnum-contents sg (second ete))
				 (* (array-dimension object 1)
				    (sg-contents sg 'm-j)))))
		     (3 (list (sg-contents sg 'm-i)
			      (sg-contents sg 'm-j)
			      (- (sg-fixnum-contents sg (second ete))
				 (* (array-dimension object 2)
				    (+ (sg-contents sg 'm-j)
				       (* (array-dimension object 1)
					  (sg-contents sg 'm-i)))))))))
;;		 (let ((rank (array-rank object)))
;;		   (case rank
;;		     (1 (list (sg-fixnum-contents sg (second ete))))
;;		     (2 (list (- (sg-fixnum-contents sg (second ete))
;;				 (* (array-dimension object 0)
;;				    (sg-contents sg 'm-j)))
;;			      (sg-contents sg 'm-j)))
;;		     (3 (list (- (sg-fixnum-contents sg (second ete))
;;				 (* (array-dimension object 0)
;;				    (+ (sg-contents sg 'm-j)
;;				       (* (array-dimension object 1)
;;					  (sg-contents sg 'm-i)))))
;;			      (sg-contents sg 'm-j)
;;			      (sg-contents sg 'm-i)))))
		 ) ;1 *)
	 1;; ELSE ((and (arrayp object) (neq (sixth ete) 1)) is NIL)*
	 1;; If object is not known, or not an array, or if AR-1-FORCE.*
	 1;; Get subscripts from first arg of ETE*
	 (list (sg-fixnum-contents sg (second ete))))))

(defflavor subscript-past-fill-pointer-error () (subscript-error))

(defmethod (subscript-past-fill-pointer-error :report) (stream)
  (format stream "The subscript ~S was past the fill-pointer of ~S while in ~S."
	  subscript-used object function))

(def-ucode-error subscript-past-fill-pointer subscript-past-fill-pointer-error
  :function (sg-erring-function sg)
  :index-location-in-sg (second ete)
  :subscript-used (sg-fixnum-contents sg (second ete))
  :subscripts-used (list (sg-fixnum-contents sg (second ete)))
  :subscript-limit (sg-fixnum-contents sg (third ete))
  :restart-tag (fourth ete)
  :object (if (fifth ete) (sg-contents sg (fifth ete))))



1;; More random losses.

;; CONS-ZERO-SIZE 
;; CONS-ZERO-SIZE M-B*
(def-ucode-format-error cons-zero-size error
  "There was an attempt to allocate zero storage by ~S."
  (sg-erring-function sg))


1;; NUMBER-CALLED-AS-FUNCTION 
;; NUMBER-CALLED-AS-FUNCTION M-T
;; First arg is number*

(def-ucode-error number-called-as-function (invalid-function)
  :property-list `(:function ,(sg-contents sg (second ete)))
  :format-string "The number, ~S, was called as a function."
  :format-args (list (sg-contents sg (second ete))))

(defflavor invalid-function () (error))

(defmethod (invalid-function :ucode-proceed-types) ()
  '(:new-function))

(defmethod (invalid-function :case :proceed-asking-user :new-function)
	   (continuation read-object-function)
  2"Read a function to use instead."
1  ;; Resume stack group and return proceed type and new function**
  (funcall continuation :new-function
	   (funcall read-object-function
		    :eval-read "Form to evaluate and use as function: ")))

(defmethod (invalid-function :case :proceed-ucode-with-args :new-function)
	   (sg new-function)
1  ;; Store new function in register and
  ;; restart sg at pc where error happened*
  (sg-store new-function sg 'm-t)		1; Was M-A in VM1.*
  (sg-proceed-micro-pc sg nil))

1;; WRONG-SG-STATE
;; WRONG-SG-STATE M-A
;; WRONG-SG-STATE A-SG-PREVIOUS-STACK-GROUP
;; Arg is stack group.
;; You cannot recover from this error*

(def-ucode-error wrong-sg-state (error wrong-stack-group-state)
  :format-string "The state of the stack group ~S, given to ~S, was invalid."
  :format-args (list (sg-contents sg (second ete))
		     (sg-erring-function sg)))

1;; SG-RETURN-UNSAFE
;; No args, since the frob is in the previous-stack-group of the current one.
;; You cannot recover.*

(def-ucode-format-error sg-return-unsafe error
  "An unsafe stack group attempted a STACK-GROUP-RETURN.")

1;; MICRO-CODE-ENTRY-OUT-OF-RANGE 
;; MICRO-CODE-ENTRY-OUT-OF-RANGE M-D
;; First arg is invalid MISC instruction*

(def-ucode-format-error micro-code-entry-out-of-range error
  "The microcode-entry function ~S is not an implemented function."
  (sg-fixnum-contents sg (second ete)))

1;; BAD-INTERNAL-MEMORY-SELECTOR-ARG 
;; BAD-INTERNAL-MEMORY-SELECTOR-ARG M-B
;; FIrst arg is the first arg of %WRITE-INTERNAL-PROCESSOR-MEMORIES. *

(def-ucode-format-error bad-internal-memory-selector-arg error
  "~S is not valid as the first argument to %WRITE-INTERNAL-PROCESSOR-MEMORIES."
  (sg-contents sg (second ete)))

1;; WRITE-IN-READ-ONLY
;; WRITE-IN-READ-ONLY VMA
;; Arg is read only address*

(def-ucode-error write-in-read-only error
  :property-list `(:address ,(sg-contents sg (second ete)))
  :format-string "There was an attempt to write into ~S, which is a read-only address~@[;~@
		  the code was probably trying to modify a program constant~]."
  :format-args (list (sg-contents sg (second ete))
		     (eq (area-name (%area-number (sg-contents sg (second ete))))
			 'macro-compiled-program)))

1;; BIGNUM-NOT-BIG-ENOUGH-DPB *

(def-ucode-format-error bignum-not-big-enough-dpb error
  "There is an internal error in bignums;  please report this bug.")


1;; More general machine lossages.

;; BAD-CDR-CODE
;; BAD-CDR-CODE VMA
;; ETE Arg is address where loser is.*

(def-ucode-format-error bad-cdr-code error
  "A bad cdr-code was found in memory (at address ~S)."
  (sg-fixnum-contents sg (second ete)))1  ;Can't use Lisp print since will err again*

(def-ucode-format-error LENGTH-OF-CIRCULAR-LIST error
  "There was an attempt to determine the length of the circular list ~s."
  (sg-contents sg (second ete)))

1;; CDR-CODE-SCREWUP
;; CDR-CODE-SCREWUP SELECTQ-TABLE*

(def-ucode-format-error cdr-code-screwup error
  "A bad cdr-code was found in a selectq-table within the function ~S."
  (sg-erring-function sg))

1;; ILLEGAL-INSTRUCTION
;; No args.*

(def-ucode-format-error illegal-instruction error
  "There was an attempt to execute an invalid instruction: 2 *~S."
  (let ((rp (sg-regular-pdl sg)) 
	(frame (sg-next-frame sg (sg-top-frame sg))))
    (fef-instruction (rp-function-word rp frame)
		     (rp-location-counter-offset rp frame))))

1;; DATA-TYPE-SCREWUP
;; DATA-TYPE-SCREWUP ARRAY
;; DATA-TYPE-SCREWUP DTP-INSTANCE-HEADER
;; DATA-TYPE-SCREWUP CDR-CODE-IN-INSTANCE-BINDINGS
;; DATA-TYPE-SCREWUP %INSTANCE-DESCRIPTOR-BINDINGS
;; DATA-TYPE-SCREWUP SELECTQ-TABLE
;; DATA-TYPE-SCREWUP DISPATCH-TABLE
;; DATA-TYPE-SCREWUP SUBR-RETURN-PC
;; ETE arg is the name.
;; This happens when some internal data structure contains the wrong data type. 
;; Not a continuable error.*

(def-ucode-format-error data-type-screwup error
  "A bad data-type was found in the internals of a~A."
  (case (second ete)
    (array
     (format nil "n array:~%DTP-ARRAY #o~o is pointing at something not DTP-ARRAY-HEADER,~@
                            or its displaced-index-offset is not a fixnum"
	     (sg-fixnum-contents sg 'm-a)))
    (dtp-instance-header
     (format nil "n instance:~%DTP-INSTANCE #o~o is pointing at something not DTP-INSTANCE-HEADER"
	     (sg-fixnum-contents sg 'vma)))
    (cdr-code-in-instance-bindings
     (format nil "n instance:~%Instance bindings list is not properly cdr-coded in instance at #o~o"
	     (sg-fixnum-contents sg 'm-a)))
    (%instance-descriptor-bindings
     (format nil "n instance:~%Instance bindings are not a list in instance at #o~o"
	     (sg-fixnum-contents sg 'm-a)))
    (selectq-table
     (format nil " function:~%Selectq-table offset not a fixnum in the function ~S"
	     (sg-erring-function sg)))
    (dispatch-table
     (format nil " function:~%Dispatch-table size or entry not a fixnum in the function ~S"
	     (sg-erring-function sg)))
    (subr-return-pc
     (format nil " function:~%Return-PC for macro-POPJ not a positive fixnum in the function ~S"
	     (sg-erring-function sg)))
    (otherwise
     (format nil " ~S" (second ete)))))

1;; STACK-FRAME-TOO-LARGE*

(def-ucode-format-error stack-frame-too-large error
  "There was an attempt to make a stack frame larger than 256. words.")

1;;  STACK-FRAME-OVERPOPPED*

(def-ucode-format-error stack-frame-overpopped error
  "There was an attempt to pop more words than possible in this stack frame.")

1;;  ILLEGAL-AREA*

(def-ucode-format-error illegal-area error
  "Tried to cons in free, fixed, or unused-code region.  Please report this error.")

1;;  ILLEGAL-DATA-TYPE*

(def-ucode-format-error illegal-data-type error
  "Illegal data type found in internal storage.")

1;; RCONS-FIXED*

(def-ucode-error rcons-fixed (error cons-in-fixed-area)
  :property-list `(:area ,(area-name (sg-contents sg 'm-s)))
  :format-string "There was an attempt to allocate storage in the fixed area ~S."
  :format-args (list (area-name (sg-contents sg 'm-s))))

1;; RPLACD-WRONG-REPRESENTATION-TYPE
;; RPLACD-WRONG-REPRESENTATION-TYPE M-S
;; Arg entry for this ETE is first argument of RPLACD*

(def-ucode-format-error rplacd-wrong-representation-type error
  "There was an attempt to RPLACD a list which is embedded in a structure and therefore cannot be RPLACD'ed.
The list is ~S."
  (sg-contents sg (second ete)))

1;; Overflow problems

;; AREA-OVERFLOW
;; AREA-OVERFLOW M-S
;; Arg for this ETE is register containing area number.*
;1;; 7/20/88 mjf/clm - added for CLEH.*
(defflavor storage-condition () (error))
(defflavor stack-overflow () (storage-condition))
(defflavor storage-exhausted () (storage-condition dangerous-error))

;1;; 7/20/88 mjf/clm - changed for CLEH.  Changed to use the new storage-exhausted.*
(def-ucode-format-error area-overflow storage-exhausted
  "2Allocation in the area ~A exceeded the maximum of ~D. words.*"
  (area-name (sg-fixnum-contents sg (second ete)))
  (area-maximum-size (sg-fixnum-contents sg (second ete))))

1;; VIRTUAL-MEMORY-OVERFLOW*
;1;; 7/20/88 mjf/clm - changed for CLEH to use the new storage-exhausted.*
(def-ucode-format-error virtual-memory-overflow storage-exhausted
  2"No more virtual memory"*)

1;; REGION-TABLE-OVERFLOW*
;1;; 7/20/88 mjf/clm - changed for CLEH to use the new storage-exhausted.*
(def-ucode-format-error region-table-overflow storage-exhausted
  "2Unable to create a new region because the region tables are full.*")

1;; PDL-OVERFLOW
;; PDL-OVERFLOW REGULAR
;; PDL-OVERFLOW SPECIAL
;; Arg is specific pdl*

(def-ucode-error pdl-overflow pdl-overflow-error
  :pdl-name (second ete))

;1;; 7/20/88 mjf/clm - changed for CLEH to use the new stack-overflow.*
(defflavor pdl-overflow-error (pdl-name) (stack-overflow)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (pdl-overflow-error :report) (stream)
  (case pdl-name
    ((regular :regular)				1; sg-maybe-grow-pdls (and maybe others) uses keywords.*
     (format stream "The regular push-down list has overflowed."))
    ((special :special)
     (format stream "The special push-down list has overflowed."))
    (otherwise
     (bad-hacker "~S is not a known pdl-name." pdl-name))))

(defmethod (pdl-overflow-error :ucode-proceed-types) () '(:grow-pdl))

(defmethod (pdl-overflow-error :case :proceed-asking-user :grow-pdl) (continuation ignore)
  2"Grow the pdl and proceed."*
  1;; Simply resume stack group and return proceed type*
  (funcall continuation :grow-pdl))

(defmethod (pdl-overflow-error :case :proceed-ucode-with-args :grow-pdl) (sg &rest ignore)
  1;; Grow sg and then restart it at pc where error happened*
  (declare (special *allow-pdl-grow-message*))
  (multiple-value-bind (value bound-p)
       (symeval-in-stack-group '*allow-pdl-grow-message* sg nil t)
    (if (and bound-p value) 
     (format t "Continuing with more pdl.~%")))
  (sg-maybe-grow-pdls sg t nil nil t)		1;Make very sure that there is enough room*
  (sg-proceed-micro-pc sg nil))			1;Then continue after ucode check for room*





1;; Select method, mapping table, and instance problems

;; ZERO-ARGS-TO-SELECT-METHOD 
;; ZERO-ARGS-TO-SELECT-METHOD M-1
;; First arg is select method  *

(def-ucode-format-error zero-args-to-select-method error
  "No method was specified;  ~S was applied to no arguments."
  (sg-contents sg (second ete)))

1;; NO-MAPPING-TABLE *

(def-ucode-format-error no-mapping-table error
  "Flavor ~S is not a component of SELF's flavor, ~S,
on a call to a function which assumes SELF is a ~S."
  (si:fef-flavor-name (rp-function-word (sg-regular-pdl sg) (sg-top-frame sg)))
  (type-of (symeval-in-stack-group 'self sg))
  (si:fef-flavor-name (rp-function-word (sg-regular-pdl sg) (sg-top-frame sg))))

1;; NO-MAPPING-TABLE-1 *

(def-ucode-format-error no-mapping-table-1 error
  "SYS:SELF-MAPPING-TABLE is NIL in a combined method.")

1;; SELF-MAPPING-TABLE-NOT-ARRAY*

(def-ucode-format-error self-mapping-table-not-array error
  "The SELF-MAPPING-TABLE for ~s is ~s (it should be an array)."
  (SYMEVAL-IN-STACK-GROUP 'self sg)
  (SYMEVAL-IN-STACK-GROUP 'self-mapping-table sg))

1;; SELF-NOT-INSTANCE *

(def-ucode-format-error self-not-instance error
  "A method is referring to an instance variable,2 *but SELF, ~S, is not an instance."
  (symeval-in-stack-group 'self sg))

1;; INSTANCE-LACKS-INSTANCE-VARIABLE 
;; INSTANCE-LACKS-INSTANCE-VARIABLE M-C M-A
;; First arg is instance variable
;; Second arg is flavor instance
;; Signaled by LOCATE-IN-INSTANCE*

(def-ucode-format-error instance-lacks-instance-variable error
  "There is no instance variable ~S in instance ~S."
  (sg-contents sg (second ete))
  (sg-contents sg (third ete)))

1;; NONEXISTENT-INSTANCE-VARIABLE *

(defun decode-nonexistent-instance-variable (sg loc)
  ;1;* 1loc will either be vma or m-1 depending on how the microcode signalled the error.*
  (let* ((rp (sg-regular-pdl sg)) 
	 (frame (sg-top-frame sg))
	 (flavor (si:fef-flavor-name (rp-function-word rp frame)))
	 (loc-contents (sg-contents sg loc)))
    (if (eq loc 'vma)
	;; From transporter:  VMA contains SRP.
	(list (si:flavor-decode-self-ref-pointer flavor (%p-pointer loc-contents))
	      flavor)
	;; From macrocode:  M-1 has instruction offset field.
	(list (nth (ldb si:%%QMI-IVAR-INDEX loc-contents)
		   (si:flavor-mapped-instance-variables (get flavor 'si:flavor)))
	      flavor))))

(def-ucode-error nonexistent-instance-variable (error instance-lacks-instance-variable)
  :format-string "Compiled code referred to instance variable ~S, which is no longer present in flavor ~S."
  :format-args  (decode-nonexistent-instance-variable sg (second ete)) )

;;03/20/89 clm - Integrated the following CLOS changes into the EH.
(eval-when (compile)
(defun build-decode-table ()
  (let ((arr (make-array 128 :element-type 'bit))
	(ar (compiler:INSTRUCTION-DECODE-TABLE)))
    (dotimes (i 128)
      (if (or (null (AREF ar i)) (get (AREF ar i) 'compiler:no-reg))
	  (setf (aref arr i ) 0)
	  (setf (aref arr i ) 1)))
    arr)))

;;;02/21/89 clm - problem occurred when trying to access an unbound class (i.e., shared) 
;;;slot.  In this case the local instance was being set to NIL; we then called SLOT-UNBOUND
;;;with the class of NIL as the class argument and NIL as the instance argument.  This would 
;;;result in a no-applicable method error for SLOT-UNBOUND on built-in-class NIL.  The fix: if
;;;we can't obtain the instance, pass the class obtained from the containing structure
;;;and NIL for the instance.   
(defun clos-unbound ( containing-structure symbol sg)
  (declare (special ce-type))
  (let* ((instance
	  (if (neq :class-slot ce-type)
	      containing-structure 
	      (let* ((pdl (sg-regular-pdl sg))
		     (frame (sg-top-frame sg))
		     (fef (rp-function-word  pdl frame))
		     (lc (1- (rp-location-counter-offset pdl frame)))
		     (wd (FEF-INSTRUCTION fef lc)))
		(if (aref #.(build-decode-table) (ASH WD -9))
		    (if ( < (ldb (byte 3 6) wd) 3)
			;;fef address mode
			(let ((self-ref (SYSTEM:%P-POINTER-OFFSET fef (ldb (byte 6 0) wd))))
			  (if (zerop (ldb sys:%%CLOS-SELF-REF-INSTANCE-REF-ADDRESSING-MODE self-ref ))
			      ;;arg number
			      (sg-frame-arg-value sg frame
						  (ldb sys:%%CLOS-SELF-REF-INSTANCE-REF-INDEX self-ref ))
			      ;;local
			      (sg-frame-local-value sg frame
						    (ldb sys:%%CLOS-SELF-REF-INSTANCE-REF-INDEX self-ref ))))


			)))))
	 (class (if instance (ticlos:class-of instance)
		    (when (eq (named-structure-p containing-structure)
				  'ticlos:class-slot-structure )
			  (let ((class-description (array-leader  containing-structure 0)))
			    (ticlos:class-description-class-object class-description )))))
	 )
      
    (let* ((*sg* sg)
	   (new-value (ticlos:slot-unbound class instance symbol)))
      (declare (special *sg*))
      (sg-regpdl-insert sg new-value)  ;; ** used to be (SG-REGPDL-PUSH VALUE SG)
      (sg-proceed-micro-pc sg 'trans-trap-restart))
    (setf (sg-current-state sg) sg-state-resumable)
    (proceed-sg sg)))

(defun eh-slot-missing (class instance slot-name operation new-value sg)
  (let* ((*sg* sg)
	 (new-value (ticlos:slot-missing class instance slot-name operation new-value)))
    (declare (special *sg*))
    (sg-regpdl-insert sg new-value)  ;; ** used to be (SG-REGPDL-PUSH VALUE SG)
    (sg-proceed-micro-pc sg 'trans-trap-restart))
  (setf (sg-current-state sg) sg-state-resumable)
  (proceed-sg sg))

(defun default-slot-missing (ignore instance slot-name operation)
  (declare (special eh:*sg*)
	   (ignore operation))
  (if (boundp '*sg*)
      (let ((sg *sg*)
	    (co (MAKE-INSTANCE (QUOTE ERROR)
			       :CONDITION-NAMES (QUOTE (INSTANCE-LACKS-INSTANCE-VARIABLE))
			       :FORMAT-STRING "There is no slot  ~S in instance ~S."
			       :FORMAT-ARGS (LIST slot-name instance))))
	(let ((signal-result
		(eh:sg-funcall-no-restart
		  sg 'signal-condition
		  co (send co :ucode-proceed-types)
		  t )))
	  ;; Try to proceed with a proceed type
	  (when (car signal-result) 
	    (lexpr-send co :proceed-ucode-with-args
			(car signal-result) sg (cdr signal-result))
	    (setf (sg-current-state sg) sg-state-resumable)
	    (proceed-sg sg))
	  (ferror nil "Proceed-type ~S~%not handled by microcode error ~S."
		  (car signal-result) co)))
      (multiple-value-bind (ignore value)
	  (signal 'error
		  :CONDITION-NAMES (QUOTE (INSTANCE-LACKS-INSTANCE-VARIABLE))
		  :FORMAT-STRING "There is no slot  ~S in instance ~S."
		  :FORMAT-ARGS (LIST slot-name instance))
	value)))

(defun default-slot-unbound (ignore instance name)
  (declare (special eh:*sg*  ce-type cont-structure or-address cu-address))
  (declare (ignore instance))
  (if (boundp 'eh:*sg*)
      (let ((eh:sg eh:*sg*))
	(let ((co (make-instance 'unbound-variable
				 :address or-address
				 :current-address cu-address
				 :cell-type ce-type
				 :symbol name
				 :data-type 
				 ;; look at top of stack
				 (aref (sg-regular-pdl sg) (- (sg-regular-pdl-pointer sg) %call-state-length)) 
				 :pointer 
				 ;; look at 2nd from top of stack
				 (aref (sg-regular-pdl sg) (- (sg-regular-pdl-pointer sg) %call-state-length 1))
				 :containing-structure cont-structure
				 :condition-names '(unbound-instance-variable))))
	  (let ((signal-result
		  (eh:sg-funcall-no-restart
		    sg 'signal-condition
		    co (send co :ucode-proceed-types)
		    t )))
	    ;; Try to proceed with a proceed type
	    (when (car signal-result) 
	      (lexpr-send co :proceed-ucode-with-args
			  (car signal-result) sg (cdr signal-result))
	      (setf (sg-current-state sg) sg-state-resumable)
	      (proceed-sg sg))
	    (ferror nil "Proceed-type ~S~%not handled by microcode error ~S."
		    (car signal-result) co))))
      (multiple-value-bind (ignore value)
	  (signal 'unbound-variable
		  :address or-address
		  :current-address cu-address
		  :cell-type ce-type
		  :data-type (sys:%p-data-type or-address)
		  :pointer (%p-pointer or-address)
		  :symbol name
		  :containing-structure cont-structure
		  :condition-names '(unbound-instance-variable))
	value)))

;;03/20/89 clm - Integrated CLOS changes into EH.
(defun (:property instance-lacks-instance-variable  make-ucode-error-function) (ignore sg ete)
  (let ((instance (sg-contents sg (third ete)))
	(slot-name (sg-contents sg (second ete))))
    (if (ticlos:clos-instance-p instance)	
	(let* ((pdl (sg-regular-pdl sg))
	       (frame (sg-top-frame sg))
	       (fef (rp-function-word  pdl frame)))
	  (if (eq fef #'ticlos:locate-slot)
	      (eh-slot-missing (ticlos:class-of instance) instance  slot-name (sg-frame-arg-value sg frame 2)
			       (sg-frame-arg-value sg frame 3) sg)
	      (MAKE-INSTANCE (QUOTE ERROR)
		       :CONDITION-NAMES (QUOTE (INSTANCE-LACKS-INSTANCE-VARIABLE))
		       :FORMAT-STRING "There is no slot ~S in instance ~S."
		       :FORMAT-ARGS (LIST slot-name instance))))
	(MAKE-INSTANCE (QUOTE ERROR)
		       :CONDITION-NAMES (QUOTE (INSTANCE-LACKS-INSTANCE-VARIABLE))
		       :FORMAT-STRING "There is no instance variable ~S in instance ~S."
		       :FORMAT-ARGS (LIST slot-name instance)))))


1;; BITBLT and TV problems

;; BITBLT-DESTINATION-TOO-SMALL *

(def-ucode-format-error bitblt-destination-too-small error
  "The destination of a BITBLT was too small to hold the rectangle being BITBLT'd.")

1;; BITBLT-ARRAY-FRACTIONAL-WORD-WIDTH *

(def-ucode-format-error bitblt-array-fractional-word-width error
  "An array passed to BITBLT has an invalid second dimension.
The second dimension, times the number of bits per element, must be a multiple of 32.")

1;; RASTER-WIDTH-TOO-WIDE*

(def-ucode-format-error raster-width-too-wide error
  "The raster width of a font passed to %DRAW-CHAR must be less than or equal to 32 pixels.")

1;; TV-ERASE-OFF-SCREEN*
(def-ucode-format-error tv-erase-off-screen (error draw-off-end-of-screen)
  "An attempt was made to do graphics past the end of the screen.")

1;; NO-CURRENTLY-PREPARED-SHEET
;; NO-CURRENTLY-PREPARED-SHEET M-C
;; Arg is the sheet*

(def-ucode-error no-currently-prepared-sheet (no-currently-prepared-sheet-error draw-on-unprepared-sheet)
  :sheet (sg-contents sg (second ete))
  :format-string "There was an attempt to draw on the sheet ~S without preparing it first."
  :format-args (list (sg-contents sg (second ete))))

1;; NO-CURRENTLY-PREPARED-SHEET-ERROR (same as NO-CURRENTLY-PREPARED-SHEET)*
;1;(renamed from TURD-ALERT and removed separate def-ucode-error - clm 6/29/88)  *

(defflavor no-currently-prepared-sheet-error (sheet) (error)
  :gettable-instance-variables
  :inittable-instance-variables)

(defmethod (no-currently-prepared-sheet-error :ucode-proceed-types) ()
  '(:no-action))

(defmethod (no-currently-prepared-sheet-error :case :proceed-asking-user :no-action) (continuation ignore)
  2"Proceed, perhaps writing garbage on the screen."
1  ;; Simply resume stack group and return proceed type**
  (funcall continuation :no-action))

(defmethod (no-currently-prepared-sheet-error :case :proceed-ucode-with-args :no-action) (sg &rest ignore)
1  ;; Restart sg at pc where error happened*
  (sg-proceed-micro-pc sg nil))


1;; TRANS-TRAP*

(defvar *use-fsh-safe* t "2Temporary variable to control the use of find-structure-header-safe.
When T, use the safer Lisp eh:find-structure-header-safe;
when NIL, use the less-safe microcoded sys:%find-structure-header.*")

;;03/20/89 clm - Integrated CLOS changes into EH.
(defun trans-trap-decode (sg)
  "Given the address on which a trans-trap occurred, determine where the contents of
that address is stored now, and if it is a null pointer, what cell of which symbol was unbound. 
The symbol can actually be a function-spec."
  (declare (values original-address current-address cell-type symbol))
  (let* ((original-address (sg-saved-vma sg))
	 (current-address (cell-location-in-stack-group original-address sg)))
    ;; If NULL on top of stack (because trans-trap pushes the pointer and then the data-type).
    (if (= (aref (sg-regular-pdl sg) (- (sg-regular-pdl-pointer sg) %call-state-length)) 
	   dtp-null)
	(let ((cell-type :unknown)		; Means we don't know what's unbound.
	      symbol)
	  (multiple-value-bind (containing-structure explanation)
	      (if *use-fsh-safe*
		  (find-structure-header-safe original-address)
		  (%find-structure-header original-address))
	    (if (null explanation)		; No problems.
		;; Is it a symbol, list, function, ...
		(cond ((symbolp containing-structure)
		       (case (%pointer-difference original-address containing-structure)
			     (1 (setq cell-type :value))
			     (2 (setq cell-type :function)))
		       (setq symbol containing-structure))
;;  This appears to be an unbound cell that points at a list whose car is a :method function-spec.
;;  I don't understand how such a case could come about, and don't think that it does anymore
;;  so I'm commenting it out.  - pf, Aug 3, 1986
;		       ((and (listp symbol)		; I don't understand this case.  - pf, Aug 1, 1986
;		             (= (%p-data-type symbol) dtp-list)
;		             (si:%p-contents-eq (car symbol) :method))
;		        (setq cell-type :function
;		              symbol    (si:meth-function-spec symbol)))
		      ((ticlos:clos-instance-p  containing-structure) 
		       (setq cell-type :clos-instance)
		      (let ((flavor (instance-flavor containing-structure))
			     (index (1- (%pointer-difference original-address containing-structure))))
			 (if flavor
			     (setq symbol (nth index (flavor-all-instance-variables flavor)))
			     (setq symbol (intern (format nil "INSTANCE-VAR-~A" index))))))
		      ((eq (named-structure-p containing-structure)
			   'ticlos:class-slot-structure )
		       (setf cell-type :class-slot)
		       (let ((class-description (array-leader  containing-structure 0)))
			 (setf symbol
			       (car (rassoc (%make-pointer sys:dtp-locative original-address )
				       (ticlos:class-description-class-slot-alist class-description ))))))
		      ((instancep containing-structure)
		       (setq cell-type :instance)
		       (let ((flavor (instance-flavor containing-structure))
			     (index (1- (%pointer-difference original-address containing-structure))))
			 (if flavor
			     (setq symbol (nth index (flavor-all-instance-variables flavor)))
			     (setq symbol (intern (format nil "INSTANCE-VAR-~A" index))))))
		      ((listp containing-structure)	; Because closures are built in list space.
		       (let ((contents (%make-pointer dtp-locative
						      ;; 2nd from top of stack ..
						      (aref (sg-regular-pdl sg)
							    (- (sg-regular-pdl-pointer sg) %call-state-length 1))))
			     explanation)
			 (multiple-value-setq (symbol explanation)
			     (if *use-fsh-safe*
				 (find-structure-header-safe contents)
				 (%find-structure-header contents)))
			 (if (null explanation)	; No problems with the symbol.
			     (setq cell-type :closure)
			     (setq cell-type :unknown
				   symbol explanation)))))
		(setq symbol explanation))	; Couldn't find the structure, cell-type is still :unknown.
	    (values original-address current-address cell-type symbol containing-structure)))
	(values original-address current-address))))

;;03/20/89 clm - Integrated CLOS changes into EH.
(defun (:property trans-trap make-ucode-error-function) (ignore sg ignore)
  (without-interrupts
    (multiple-value-bind (original-address current-address cell-type symbol containing-structure)
	(trans-trap-decode sg)
      (if (or (eq cell-type :clos-instance) (eq cell-type :class-slot))
	  (let ((ce-type cell-type)
		(or-address original-address )
		(cu-address current-address )
		(cont-structure containing-structure))
	    (declare (special cont-structure  ce-type or-address cu-address ))
	    (clos-unbound  containing-structure symbol sg))
	  (make-instance (case cell-type
			   ((:value :closure :clos-instance :instance :class-slot) 'unbound-variable)
			   (:function 'undefined-function)
			   (otherwise 'cell-contents-error))
			 :address original-address
			 :current-address current-address
			 :cell-type cell-type
			 :symbol symbol
			 :data-type 
			 ;; look at top of stack
			 (aref (sg-regular-pdl sg) (- (sg-regular-pdl-pointer sg) %call-state-length)) 
			 :pointer 
			 ;; look at 2nd from top of stack
			 (aref (sg-regular-pdl sg) (- (sg-regular-pdl-pointer sg) %call-state-length 1))
			 :containing-structure containing-structure
			 :condition-names (case cell-type
					    (:value    '(unbound-symbol))
					    (:closure  '(unbound-closure-variable))
					    (:instance '(unbound-instance-variable))
					    (:function '(undefined-function))
					    (otherwise '(bad-data-type-in-memory))))))))

;;;
;;;  A Lisp version of %find-structure-header that won't die on bad data types.

(defun find-structure-header-safe (ptr)
   (let ((region-number (sys:%region-number ptr)))
     (if (numberp region-number)
	 (let* ((origin (sys:region-origin region-number))
		(limit (+ origin (sys:region-free-pointer region-number)))
		(ptr (sys:%pointer ptr)))
	   (if (< ptr limit)
	       (let* ((bits (sys:region-bits region-number))
		      (representation-type (ldb sys:%%region-representation-type bits)))
		 (select representation-type
		   (sys:%region-representation-type-list
		    (fsh-safe-list ptr origin))
		   (sys:%region-representation-type-structure
		    (fsh-safe-structure ptr origin))
		   (otherwise
		    (values nil "Illegal region representation type"))))
	       (values nil "Pointer beyond end of region")))
	 (values nil "Pointer not in any region"))))

;;;
;;;  The idea here is to back up until the cdr-code is cdr-nil, meaning the end of
;;;  the preceding list, or cdr-error, meaning a full node from a preceding list.
;;;  Also, if we reach the beginning of the region, we're finished, because cdr-coded
;;;  lists are not consed across regions.  Once we reach the end, we return a list
;;;  pointer to the apparent first word of our list.

(defun fsh-safe-list (ptr origin)
   (loop until (= ptr origin)
	 as previous-data-type = (sys:%p-data-type (1- ptr))
	 when (= previous-data-type dtp-header-forward)	1       ; Forwarding pointer from a rplacd.*
	 do (loop-finish)
	 as previous-cdr-code = (sys:%p-cdr-code (1- ptr))
	 doing (select previous-cdr-code
		 ((cdr-normal cdr-next)
		  (decf ptr))
		 ((cdr-error cdr-nil)
		  (loop-finish)))
	 finally (return (sys:%make-pointer sys:dtp-list ptr))))

(defun fsh-safe-structure (ptr origin)
   (loop until (< ptr origin)
	 finally (return nil "Invalid structure:  header not found in region")
	 as data-type = (sys:%p-data-type ptr)
	 doing (select data-type
		 ((sys:dtp-gc-forward sys:dtp-free sys:dtp-gc-young-pointer)	; +++ Should be sys, someday.
		  (return nil
			  (format nil "Unexpected data type in find-structure-header:  ~A" data-type)))
		 ((sys:dtp-header-forward sys:dtp-body-forward)
		  (return nil "Can't handle header-forwards or body-forwards yet")
;			      (multiple-value-bind (header reason)
;				  (transport-header (if (= data-type sys:dtp-header-forward)
;							ptr
;							(sys:%p-pointer ptr)))
;				(if (null reason)
;				    (setq data-type (sys:%p-data-type header)
;					  ptr       header)
;				    (return nil reason)))
		  )
		 (sys:dtp-symbol-header
		  (return (sys:%make-pointer sys:dtp-symbol ptr)))
		 (sys:dtp-header
		  (let ((header-type (sys:%p-ldb sys:%%header-type-field ptr)))
		    (select header-type
		      ((sys:%header-type-error si:%header-unused-1 si:%header-type-unused-3)	; +++ Sys, someday.
		       (return nil (format nil "Illegal header type:  ~A" header-type)))
		      ((sys:%header-type-complex sys:%header-type-bignum
		        sys:%header-type-rational sys:%header-type-double-float)
		       (return (sys:%make-pointer sys:dtp-extended-number ptr)))
		      (sys:%header-type-array-leader
		       (let ((offset (sys:%p-ldb sys:%%array-leader-length ptr)))
			 (return (sys:%make-pointer-offset sys:dtp-array-pointer ptr offset))))
		      (si:%header-type-single-float	; +++ Sys, someday.
		       (return (sys:%make-pointer sys:dtp-single-float ptr))))))
		 (sys:dtp-array-header
		  (let ((array-type (sys:%p-ldb sys:%%array-type-field ptr)))
		    (if (= array-type (lsh art-stack-group-head sys:array-type-shift))
			(return (sys:%make-pointer sys:dtp-stack-group ptr))
			(return (sys:%make-pointer si:dtp-array ptr)))))	; +++ Sys, someday.
		 (sys:dtp-instance-header
		  (return (sys:%make-pointer sys:dtp-instance ptr)))
		 (sys:dtp-fef-header
		  (return (sys:%make-pointer si:dtp-function ptr)))	; +++ Sys, someday.
		 (otherwise
		  nil))
	 doing (decf ptr)))		; Will only reach here if needs to loop around.

;;;
;;;  Simulate the microcode's transport-header.  This really should do something about
;;;  oldspace pointers, but I don't know how to do that without possibly chasing pointers
;;;  into illegal data types or invalid places.

(defun transport-header (ptr)
   (loop as next-ptr = (sys:%p-pointer ptr)
	 as next-type = (sys:%p-data-type ptr)
	 doing (select next-type
		 ((sys:dtp-trap sys:dtp-gc-forward sys:dtp-external-value-cell-pointer sys:dtp-one-q-forward
		   sys:dtp-self-ref-pointer sys:dtp-free sys:dtp-null)	; sys:dtp-ones-trap sys:dtp-unused-28
		  (return nil (format nil "Illegal data type:  ~A" next-type)))
		 (sys:dtp-body-forward
		  (return nil "Don't know how to handle body-forwards yet in transport-header"))
		 (sys:dtp-header-forward
		  nil)				; Keep looping.
		 (otherwise
		  (return next-ptr)))))



1;; CELL-CONTENTS-ERROR, UNBOUND-VARIABLE, and UNDEFINED-FUNCTION
;; are used by TRANS-TRAP UCODE ERROR
;; provides proceed types :NO-ACTION, :PACKAGE-DWIM, :NEW-VALUE and :STORE-NEW-VALUE*

;1;; 7/20/88 mjf/clm -changed for CLEH, changed cell-contents-error flavor to use the*
;1;; new cell-error flavor*
(defflavor cell-error (name) (error))

(defflavor cell-contents-error
	(address current-address cell-type symbol 
	 containing-structure data-type pointer)
	(cell-error)
  :gettable-instance-variables
  :initable-instance-variables)

(defflavor unbound-variable () (cell-contents-error))

(defmethod (unbound-variable :variable-name) () symbol)

(defmethod (unbound-variable :instance) ()
  (if (typep containing-structure 'instance) containing-structure))

(defflavor undefined-function () (cell-contents-error))

(defmethod (undefined-function :function-name) () symbol)

;;03/20/89 clm - Integrated CLOS changes into EH.
(defmethod (cell-contents-error :report) (stream)
  (let* ((contents-changed
	   (or (neq (%p-data-type current-address) data-type)
	       (neq (%p-pointer current-address) pointer)))
	 (verb (if contents-changed "was" "is")))
    (case cell-type
      (:value    (format stream "The variable ~S ~A unbound." symbol verb))
      (:function (format stream "The function ~S ~A undefined." symbol verb))
      (:closure  (format stream "The variable ~S ~A unbound (in a closure value-cell)." symbol verb))
      (:instance (format stream "The instance variable ~S ~A unbound in ~S." symbol verb containing-structure))
      (:clos-instance (format stream "The instance slot ~S ~A unbound in ~S." symbol verb containing-structure))
      (:class-slot (format stream "The class slot ~S ~A unbound in an instance of class ~S."
			   symbol verb (ticlos:class-description-class-object (array-leader containing-structure 0))))
      (otherwise
       (format stream "The word #<~S ~O> was read from location #o~O ~@[(in ~S)~].~:[~;~@
		       Problem in find-structure-header -- ~A.~]"
	       (q-data-types data-type)
	       pointer
	       (%pointer address)
	       (let ((area (%area-number address)))
		 (if area (area-name area)))
	       (and (eq cell-type :unknown)
		    (not (null symbol)))
	       symbol)))))
			1; Here, it's the explanation of the find-structure-header problem.*

;;03/20/89 clm - Integrated CLOS changes into EH.
(defmethod (cell-contents-error :after :print-error-message) (sg brief stream)
  (declare (special *error-locus-frame*))
  (unless brief
    (let ((contents-changed
	    (or (neq (%p-data-type current-address) data-type)
		(neq (%p-pointer current-address) pointer))))
      (case cell-type
	((:value :closure :instance :clos-instance :class-slot)
	 (when contents-changed
	   (print-new-cell-contents "It now has the value" current-address stream)))
	(:function
	 (when contents-changed
	   (print-new-cell-contents "It now has the definition" current-address stream))
	 (when (and (symbolp symbol)
		    (get symbol 'compiler:opcode))
	   (let ((fn (rp-function-word (sg-regular-pdl sg)
				       (sg-out-to-interesting-frame sg *error-locus-frame*))))
	     (format stream
		     "Note: ~S is supported as a function when used in compiled code.~%"
		     symbol)
	     (if (consp fn)
		 (format stream
			 "You may have evaluated the definition of ~S, which is now not compiled.~%"
			 (function-name fn)))))
	 )
	(otherwise
	 (when contents-changed
	   (print-new-cell-contents "The cell now contains" current-address stream)))))))

(defun print-new-cell-contents (cell-description address stream)
  (if (%p-contents-safe-p address)
      (format stream "~A ~S.~%" cell-description (%p-contents-offset address 0))
    (format stream "~A #<~S ~O>.~%" cell-description
	    (q-data-types (%p-data-type address))
	    (%p-pointer address))))

1;; Some people would rather not spend the time for this feature, so let them turn it off*
(defparameter *enable-trans-trap-dwim* t
 2"Non-NIL means look spontaneously in other packages on undefined function or
unbound variable error"*)

1;; If problem is symbol in wrong package, offer some dwimoid assistance.*
(defmethod (cell-contents-error :debugger-command-loop) (*error-sg* &optional (error-object self))
  (declare (special *error-sg* *enable-trans-trap-dwim*))
  (declare (special error-object))
  (catch 'quit
    (catch-error-restart ((sys:abort error) "Return to debugger command loop.")
      (if *enable-trans-trap-dwim*
	  (send error-object :proceed-asking-user :package-dwim
		'proceed-error-sg 'read-object))))
  nil)

(defmethod (cell-contents-error :user-proceed-types) (proceed-types)
  (remove :new-value (the list proceed-types) :test #'eq))

(defmethod (cell-contents-error :ucode-proceed-types) ()
1  ;; Don't include :package-dwim if not a variable or function *
  (if (member cell-type '(:value :function) :test #'eq) 
      '(:no-action :new-value :store-new-value :package-dwim)
    '(:no-action :new-value :store-new-value)))

1;; :NO-ACTION*

(defmethod (cell-contents-error :case :proceed-asking-user :no-action) (continuation read-object-function)
  2"Use current contents if legal, or read replacement value and proceed."*
  (if (not (%p-contents-safe-p current-address))
      1;; Location still contains garbage, get a replacement value.*
      (funcall self :proceed-asking-user :new-value
	       continuation read-object-function)
      (funcall continuation :no-action)))

(defmethod (cell-contents-error :case :proceed-ucode-with-args :no-action) (sg)
1  ;; Push 0 on top of pdl*
  (sg-regpdl-insert sg 0)1  ;; ** used to be (SG-REGPDL-PUSH 0 SG)
  ;; Transfer the current contents to what MD will be got from.
  ;; Restart sg at pc for restart tag, TRANS-TRAP*
  (%blt-typed current-address (sg-locate sg 'pp) 1 0)
  (sg-proceed-micro-pc sg 'trans-trap-restart))

1;; :PACKAGE-DWIM*

(defmethod (cell-contents-error :case :proceed-asking-user :package-dwim) (continuation ignore)
  2"Look for symbols with the same name but in different packages."*
  ;; 03/09/87 CLM - Fixed to use SYMBOL-VALUE instead of SYMEVAL.
  (declare (special *error-sg*))
  (let (cell new-val)
    (and (symbolp symbol)
	 (setq cell (assoc cell-type '((:value boundp symbol-value) (:function fdefinedp fdefinition))
			   :test #'eq))
	 1;; Attempt to get new value from erring stack group by looking in other packages*
	 (car (setq new-val
		    (sg-funcall *error-sg* 'cell-contents-error-dwimify symbol cell *terminal-io*)))
	 1;; Resume stack group and return proceed type and new value (2nd item in new-val)*
	 (funcall continuation :new-value (cadr new-val)))))

1;; CELL is a list (symbolic-name dwimify-definition-type value-extractor)*
(defun cell-contents-error-dwimify (sym cell *query-io*)
  (declare (values success-p new-value new-symbol))
1  ;; cell has format (:value boundp sym*bol-value1) or (:function fdefinedp fdefinition)*
  (let ((dwim-value (dwimify-package-0 sym (second cell)))) 
    (send *query-io* :fresh-line)
    (if dwim-value (values t (funcall (third cell) dwim-value) dwim-value))))

1;; This has to exist, but it's not intended to ever be used,
;; so do the same thing as :NO-ACTION if it ever does get used.*
(defmethod (cell-contents-error :case :proceed-ucode-with-args :package-dwim) (sg)
  1;; Push 0 on top of pdl*
  (sg-regpdl-insert sg 0)1  ;; ** used to be (SG-REGPDL-PUSH 0 SG)*
  1;; Transfer the current contents to what MD will be got from.*
  1;; Restart sg at pc for restart tag, TRANS-TRAP*
  (%blt-typed current-address (sg-locate sg 'pp) 1 0)
  (sg-proceed-micro-pc sg 'trans-trap-restart))

1;; :NEW-VALUE*

(defmethod (cell-contents-error :case :proceed-asking-user :new-value) (continuation read-object-function)
  2"Read a value to use instead."*
  (let ((prompt (case cell-type
		  (:value
		   (format nil "Form to evaluate and use instead of ~S's value: " symbol))
		  (:function
		   (format nil "Form to evaluate and use instead of ~S's function definition: " symbol))
		  (otherwise
		    "Form to evaluate and use instead of cell's contents: "))))
    1;; Resume stack group and return proceed type and new value*
    (funcall continuation :new-value (funcall read-object-function :eval-read prompt))))

(defmethod (cell-contents-error :case :proceed-ucode-with-args :new-value) (sg value)
1  ;; Push new value on top of pdl
  ;; and restart sg at pc for restart tag, TRANS-TRAP*
  (sg-regpdl-insert sg value)1  ;; ** used to be (SG-REGPDL-PUSH VALUE SG)*
  (sg-proceed-micro-pc sg 'trans-trap-restart))

1;; :STORE-NEW-VALUE*

(defmethod (cell-contents-error :case :proceed-asking-user :store-new-value)
	   (continuation read-object-function)
  2"Read a value to store and use instead."*
  (let ((value (funcall read-object-function :eval-read
			(case cell-type
			  (:value
			   (format nil "Form to evaluate and SETQ ~S to: " symbol))
			  (:function
			   (format nil "Form to evaluate and FSET ~S to: " symbol))
			  (otherwise
			   "Form to evaluate and store back: ")))))
    1;; Resume stack group and return proceed type and new value*
    (send continuation :store-new-value value)))

(defmethod (cell-contents-error :case :proceed-ucode-with-args :store-new-value) (sg value)
1  ;; Store new value at current address;
  ;; also push new value on top of pdl
  ;; and restart sg at pc for restart tag, TRANS-TRAP*
  (%p-store-contents current-address value)
  (sg-regpdl-insert sg value)1  ;; ** used to be (SG-REGPDL-PUSH VALUE SG)*
  (sg-proceed-micro-pc sg 'trans-trap-restart))


1;; TOO-MANY-ARGS and TOO-FEW-ARGS  (new for VM2)

;;  Of the six conditions handled by the FUNCTION-ENTRY error in VM1, only two are supported
;;  in VM2:  too-many-args and too-few-args.  We will contrive to signal the same flavor
;;  with the same condition-names, but use two make-ucode-error-functions to do it.
;;  Also, these traps are signalled before the function we're calling has a chance to set up its
;;  state words, so we have to pick things up out of registers (there may be a better way).*
(defun (:property too-many-args make-ucode-error-function) (ignore sg ignore)
  ;; 03/10/87 CLM - Changed argument collection to reflect new stack setup
  ;;                (a new frame has been pushed).
  ;; 06/25/87 CLM - Deleted some unused debugging code.
  (let ((nargs (sg-fixnum-contents sg 'm-h)))	; A pure number, so be careful.
    (make-instance 'function-entry-error
		   :function (sg-ac-t sg)
		   :argument-list (collect-args-from-stack sg nargs)
		   :nargs nargs
		   :condition-names '(too-many-arguments))))

(defun (:property too-few-args make-ucode-error-function) (ignore sg ignore)
  ;; 03/10/87 CLM - Changed argument collection to reflect new stack setup
  ;;                (a new frame has been pushed).
  ;; 06/25/87 CLM - Deleted some unused debugging code.
  (let ((nargs (sg-fixnum-contents sg 'm-h)))	; A pure number, so be careful.
    (make-instance 'function-entry-error
		   :function (sg-ac-t sg)
		   :argument-list (collect-args-from-stack sg nargs)
		   :nargs nargs
		   :condition-names '(too-few-arguments))))

(defun collect-args-from-stack (sg n)
  ;; 03/10/87 CLM - Added new function to gather args from stack in cases of
  ;;                too-few/too-many args.
   (loop with rp = (sg-regular-pdl sg)
	 with frame = (sg-top-frame sg)
	 with list = nil
	 for i from (1- (+ (rp-argument-offset sg rp frame) n))
	      downto (rp-argument-offset sg rp frame)
	 doing (push (aref rp i) list)
	 finally (return list)))

(defun collect-top-n-from-stack (sg n)
   (loop with rp = (sg-regular-pdl sg)
	 with list = nil
	 for i from (- (sg-regular-pdl-pointer sg) %call-state-length)
	      above (- (sg-regular-pdl-pointer sg) %call-state-length n)
	 doing (push (aref rp i) list)
	 finally (return list)))


1;; SYS:TOO-FEW-ARGUMENTS and SYS:TOO-MANY-ARGMENTS are errors signaled in lisp*

(defsignal-explicit sys:too-few-arguments (function-entry-error sys:too-few-arguments)
		    (ignore function nargs argument-list)
  2"FUNCTION is called with only NARGS args, which are in ARGUMENT-LIST."*
  :function function :nargs nargs :argument-list argument-list)

(defsignal-explicit sys:too-many-arguments (function-entry-error sys:too-many-arguments)
		    (ignore function nargs argument-list)
  2"FUNCTION is called with only NARGS args, which are in ARGUMENT-LIST."*
  :function function :nargs nargs :argument-list argument-list)

1;; FUNCTION-ENTRY-ERROR*

(defflavor function-entry-error	(function argument-list nargs) (error)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (function-entry-error :report) (stream)
  (cond ((member 'too-few-arguments (the list condition-names) :test #'eq)
	 (format stream "Function ~S called with only ~D argument~1@*~P."
		 (function-name function) nargs))
	((member 'too-many-arguments (the list condition-names) :test #'eq)
	 (format stream "Function ~S called with too many arguments (~D)."
		 (function-name function) nargs))
	((member 'bad-data-type (the list condition-names) :test #'eq)
	 (format stream "Function ~S called with an argument of bad data type."
		 (function-name function)))
	(t
	 (format stream "Function ~S called with bad arguments." 
		 (function-name function)))))

(defmethod (function-entry-error :user-proceed-types) (proceed-types)
  (let ((ptypes (copy-list proceed-types))
	first)
    (dolist (p ptypes)
      (if (member p '(:additional-arguments :fewer-arguments :new-argument-list) :test #'eq) 
	  (return (setq first p))))
    1;; Get rid of all occurrences of proceed types :additional-arguments, :fewer-arguments, *
    1;; and :new-argument-list, except the first one.*
    (when first
      (setf (cdr (member first ptypes :test #'eq))
	    (delete :additional-arguments
		    (the list (delete :fewer-arguments
				      (the list (delete :new-argument-list
							(the list (cdr (member first ptypes :test #'eq)))
							:test #'eq))
				      :test #'eq))
		    :test #'eq) ))
    ptypes))

(defmethod (function-entry-error :ucode-proceed-types) ()
  (if (member 'too-few-arguments (the list condition-names) :test #'eq) 
      '(:additional-arguments :fewer-arguments :new-argument-list)
      '(:fewer-arguments :additional-arguments :new-argument-list)))

(defmethod (function-entry-error :case :proceed-asking-user :additional-arguments) (continuation read-object-function)
  2"Try the call again, asking for additional arguments to use."*
  (send self :proceed-asking-user :new-argument-list 
	continuation read-object-function :additional-arguments))

(defmethod (function-entry-error :case :proceed-ucode-with-args :additional-arguments) (sg args)
  (send self :proceed-ucode-with-args :new-argument-list sg args))

(defmethod (function-entry-error :case :proceed-asking-user :fewer-arguments) (continuation read-object-function)
2  "Try the call again, dropping some of the arguments."*
  (send self :proceed-asking-user :new-argument-list 
	continuation read-object-function :fewer-arguments))

(defmethod (function-entry-error :case :proceed-ucode-with-args :fewer-arguments) (sg args)
  (send self :proceed-ucode-with-args :new-argument-list sg args))

(defmethod (function-entry-error :case :proceed-asking-user :new-argument-list)
	   (continuation read-object-function &optional (ptype :new-argument-list))
  2"Try the call again, with new argument list."*
  (let* ((-function- (send self :function))
	 (-argument-list- (send self :argument-list))
	 (-nargs- (send self :nargs))
	 (form (cons -function- -argument-list-))
	 new-args)
    (multiple-value-bind (args-wanted max-args rest-flag)
	(args-desc -function-)
      1;; Function may have been redefined to take the supplied number of arguments*
      1;; so don't look at the original error, but check everything again.*
      (cond ((< -nargs- args-wanted)
	     1;; too few args, ask for values for missing args*
	     (do ((i -nargs- (1+ i)))
		 ((unless rest-flag (eq i max-args)))
	       (multiple-value-bind (value flag)
		   (funcall read-object-function
			    (if (>= i args-wanted)
				:eval-read-or-end
				:eval-read)
			    (if (>= i args-wanted)
				"Arg ~D~A, or ~C: "
				"Arg ~D~A: ")
			    i
			    (with-output-to-string (*standard-output*)
			      (display-arg-name " (~A)" -function- i))
			    #\end)
		 (if flag (return))
		 (setq new-args (nconc new-args (cons value nil)))))
	     1;; Resume stack group and return proceed type and all args*
	     (funcall continuation ptype (append (firstn -nargs- (cdr form)) new-args)))
	    1;; Offer to proceed with current args*
	    ((or (<= -nargs- max-args) rest-flag)
	     (if (funcall read-object-function '(:fquery) "Try ~S again? " form)
		 (funcall continuation ptype (cdr form))))
	    1;; too many args, offer to drop args*
	    ((funcall read-object-function '(:fquery)
		      "Call again with the last ~[~1;~:;~:*~D ~]argument~:P dropped? "
		      (- -nargs- max-args))
	     1;; resume stack group and return proceed type and args*
	     (funcall continuation ptype (firstn max-args (cdr form))))))))

(defmethod (function-entry-error :case :proceed-ucode-with-args :new-argument-list) (sg arguments)
  (declare (special *current-stack-group* *error-handler-running*))
  (let* ((form (cons function arguments))
	 (frame (sg-top-frame sg))
	 (*error-locus-frame* frame)
	 (*current-frame* frame)
	 (*innermost-visible-frame* frame))
    (declare (special *error-locus-frame*  *current-frame* *innermost-visible-frame*))
    1;; If we haven't quit before getting here, *
    1;; resume SG at FRAME, replacing fef and args of FRAME with FORM*
    (sg-unwind-to-frame-and-reinvoke sg frame form)
    (leaving-error-handler)
    (without-interrupts
      (when *error-handler-running*
	(free-second-level-error-handler-sg *current-stack-group*))
      (stack-group-resume sg nil))))

(defmethod (function-entry-error :case :proceed-asking-user :return-value) (continuation read-object-function)
  2"Read a form to evaluate and return instead."*
  1;; Resume stack group and return proceed type and new value to return*
  (funcall continuation :return-value
	   (funcall read-object-function :eval-read
		    "Form to evaluate and return from ~S: "
		    (function-name (send self :function)))))


1;; MAR-BREAK

;; This code won't work if write-data is a DTP-NULL because of trap out of MAKUNBOUND*

(def-ucode-error mar-break mar-break
  :direction (cdr (assoc (second ete) '((write . :write) (read . :read) (global:read . :read))	1; Just in case.*
			 :test #'eq))
  :value (sg-contents sg 'pp)
  :object (let ((%mar-high -2) (%mar-low -1))
	     (%find-structure-header (sg-saved-vma sg)))
  :offset (let ((%mar-high -2) (%mar-low -1))
	     (%pointer-difference (sg-saved-vma sg)
				  (%find-structure-header (sg-saved-vma sg)))))

(defflavor mar-break (direction value object offset) (condition)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (mar-break :debugging-condition-p) ()
  t)

(defmethod (mar-break :report) (stream)
  (if (eq direction :write)
      (format stream "The MAR has gone off because of an attempt to write ~S into offset ~A in ~S."
	      value offset object)
      (format stream "The MAR has gone off because of an attempt to read from offset ~A in ~S."
	      offset object)))

(defmethod (mar-break :ucode-proceed-types) ()
  (if (eq direction :write)
      '(:no-action :proceed-no-write)
      '(:no-action)))

(defmethod (mar-break :case :proceed-asking-user :no-action) (continuation ignore)
  "Proceed."
  (funcall continuation :no-action))

(defmethod (mar-break :case :proceed-ucode-with-args :no-action) (sg &rest ignore)
  1;; By simply returning without calling SG-PROCEED-MICRO-PC, the PGF-R will return*
  (if (eq direction :write)
      (sg-funcall sg #'(lambda (vma md)				1;Simulate the write*
			 (let ((%mar-high -2) (%mar-low -1))	1;Disable MAR*
			   (rplaca vma md)))
		  (sg-saved-vma sg) 
		  (sg-regpdl-delete sg))))1  ;; ** used to be (SG-REGPDL-POP SG)*

(defmethod (mar-break :case :proceed-asking-user :proceed-no-write) (continuation ignore)
  2"Proceed, not changing the cell contents."*
  1;; Resume stack group and return proceed type*
  (funcall continuation :proceed-no-write))

(defmethod (mar-break :case :proceed-ucode-with-args :proceed-no-write) (&rest ignore)
  1;; By simply returning without calling SG-PROCEED-MICRO-PC, the PGF-R will return*
  nil)


1;; DONT-CLEAR-INPUT-UCODE-BREAKPOINT
;; Used by STEP-BREAK-ERROR, CALL-TRAP-ERROR, EXIT-TRAP-ERROR, 
;; and THROW-EXIT-TRAP-ERROR *

(defflavor dont-clear-input-ucode-breakpoint () (condition))

(defmethod (dont-clear-input-ucode-breakpoint :maybe-clear-input) (stream)
   (declare (ignore stream))
   nil)

(defmethod (dont-clear-input-ucode-breakpoint :print-error-message-prefix) (sg brief stream)
  (declare (ignore sg brief))
  (princ ">> " stream))

(defmethod (dont-clear-input-ucode-breakpoint :debugging-condition-p) ()
  t)

(defmethod (dont-clear-input-ucode-breakpoint :ucode-proceed-types) ()
  '(:no-action))

(defmethod (dont-clear-input-ucode-breakpoint :case :proceed-asking-user :no-action) (continuation ignore)
  2"Proceed."*
  (format t " Continue from break.~%")
  (funcall continuation :no-action))

(defmethod (dont-clear-input-ucode-breakpoint :case :proceed-ucode-with-args :no-action) (sg &rest ignore)
  1;; Restart sg at pc where error happened*
  (sg-proceed-micro-pc sg nil))

1;; BREAKPOINT*

(def-ucode-error breakpoint step-break-error
  :format-string "Breakpoint.")

1;; STEP-BREAK ERROR*

(def-ucode-error step-break step-break-error
  :format-string "Step break.")

(defflavor step-break-error () (dont-clear-input-ucode-breakpoint))

(defmethod (step-break-error :ucode-proceed-types) ()
  ())

1;; CALL-TRAP ERROR
;; For "Break on entry to function"*

(def-ucode-error call-trap call-trap-error
  :function (rp-function-word (sg-regular-pdl sg) (sg-top-frame sg)))

(defun (:property call-trap enter-error-handler) (sg ignore)
1  ;Make frame being entered visible.*
  (let ((in-frame (sg-top-frame sg))
	(rp (sg-regular-pdl sg))) 
1     ;;  The microcode has advanced the location-counter to the second instruction, even
     ;;  though it hasn't executed the first instruction yet (we caught it in the middle), so
     ;;  we back it up for display purposes and move it forward again in the wrapper on
     ;;  :proceed-ucode-with-args :no-action below (if only we could put daemons on :case
     ;;  methods).
     ;;    - pf, Sept 23, 1986*
    (decf (rp-location-counter-offset rp in-frame))
    (setf (rp-trap-on-exit rp in-frame) 1)))

(defflavor call-trap-error (function) (dont-clear-input-ucode-breakpoint)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (call-trap-error :around :find-current-frame) (cont mt args sg)
  1;; Uses frame in *UCODE-ERROR-STATUS* for first 2 (of 4) values returned, *
  1;; ie, *ERROR-LOCUS-FRAME* and *CURRENT-FRAME**
  1;; (*ucode-error-status* is set when enter signal-condition)*
  (let ((frame (+ (fourth (symeval-in-stack-group '*ucode-error-status* sg))
		  (* 2 (array-leader-length sg)))))1     ; Have to compensate for the foothold "args."*
    (multiple-value-bind (ignore ignore innermost-frame innermost-visible-p)
	(lexpr-funcall-with-mapping-table cont mt args)
      (declare (ignore ignore))
      (values frame frame innermost-frame innermost-visible-p))))

(defmethod (call-trap-error :report) (stream)
   (format stream "Break on entry to function ~S." (function-name function)))

1;;  
;;  What I really want is (:method call-trap-error :before :case :proceed-ucode-with-args :no-action),
;;  but we can't do that.  This method is a :before daemon on the :no-action proceed-type that moves
;;  the location-counter up one to undo the moving back one of (:property call-trap enter-error-handler).
;;    - pf, Sept 23, 1986*
(defwrapper (call-trap-error :proceed-ucode-with-args) ((subop sg) . body)
  `(progn (when (eq subop :no-action)
	    (let ((in-frame (sg-top-frame sg))
		  (rp (sg-regular-pdl sg)))
	      (incf (rp-location-counter-offset rp in-frame))))
	  ,@body))


1;; EXIT-TRAP ERROR
;; EXIT-TRAP ERROR %THROW
;; EXIT-TRAP ERROR %THROW-N
;; For "Break on exit from function"*

(def-ucode-error exit-trap exit-trap-error
  :function (rp-function-word (sg-regular-pdl sg) (sg-top-frame sg))
  :values (collect-value-list sg (sg-top-frame sg) (second ete))
  :throw-p (second ete))

(defun collect-results-from-stack (sg)
  ;; 09/03/87 CLM - Original, in these cases we know the number of results
  ;;                is the first value on top of the stack.
   (loop with rp = (sg-regular-pdl sg)
	 with n = (aref rp (- (sg-regular-pdl-pointer sg) %call-state-length))
	 with list = nil
	 for i from (- (sg-regular-pdl-pointer sg) %call-state-length 1)  ;;one subtracted to get past count
	      above (- (sg-regular-pdl-pointer sg) %call-state-length 1 n)
	 doing (push (aref rp i) list)
	 finally (return list)))

(defun collect-value-list (sg frame throw-p)
  ;; 03/19/87 CLM - Fixed for %return-all-values-with-count-on-stack to get count
  ;;                from m-a; at this point, the count is no longer on top of the stack.
  ;; 09/03/87 CLM - Change handling of %normal-return and %return-all-values-with-count-on-stack,
  ;;                the number of values to be returned in both cases is on top of the
  ;;                stack now.  Call COLLECT-RESULTS-FROM-STACK to collect values. 
   (cond ((null throw-p)
	  (let ((rp (sg-regular-pdl sg)))
	    (if (= (rp-saved-destination rp frame) %qmi-calldest-inds)
		nil				; D-IGNORE, no values.
		(select (rp-return-type rp frame)	; D-PDL or D-RETURN (D-TAIL-REC wouldn't get here).
		  (%only-one-result-needed
		   (list (sg-contents sg 'm-t)))
		  (%multiple-value-list-return
		   (sg-contents sg 'm-t))	; Already a list.
		  (%normal-return		; Multiple values.
		   (collect-results-from-stack sg))
		  (%return-all-values-with-count-on-stack
		   (collect-results-from-stack sg)  ) ) ) ))
	 ((eq throw-p '%throw)
	  (list (sg-contents sg 'm-t)))
	 ((eq throw-p '%throw-n)
	  (let ((n-values (sg-fixnum-contents sg 'm-g)))
	    (if (zerop n-values)
		nil
		(collect-top-n-from-stack sg n-values) )))
	 (:else
	  (bad-hacker "~S is not a valid second argument in EXIT-TRAP." throw-p))))

(defun (:property exit-trap enter-error-handler) (sg ignore)
1  ;; Don't catch this trap again if user tries to reinvoke this frame.*
  (setf (rp-trap-on-exit (sg-regular-pdl sg) (sg-top-frame sg)) 0))

(defflavor exit-trap-error (function values throw-p) (dont-clear-input-ucode-breakpoint)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (exit-trap-error :around :find-current-frame) (cont mt args sg)
1  ;; Uses frame in *UCODE-ERROR-STATUS* for for first 2 (of 4) values returned, 
  ;; ie, *ERROR-LOCUS-FRAME* and *CURRENT-FRAME*
  ;; (*ucode-error-status* is set when enter signal-condition)*
  (let ((frame (+ (third (symeval-in-stack-group '*ucode-error-status* sg))
		  (* 2 (array-leader-length sg)))))1     ; Have to compensate for the foothold "args."*
    (multiple-value-bind (ignore ignore innermost-frame innermost-visible-p)
	(lexpr-funcall-with-mapping-table cont mt args)
      (declare (ignore ignore))
      (values frame frame innermost-frame innermost-visible-p))))

(defmethod (exit-trap-error :report) (stream)
  (format stream "Break on ~:[exit from~;throw through~] ~S."
	  throw-p (function-name function)))

;;;03/06/89 clm - added optional arg stream to terpri call so that newline
;;would be inserted in bug report backtraces.
(defmethod (exit-trap-error :after :print-error-message) (sg brief stream)
  (declare (special *error-print-length* *error-print-level*)
	   (ignore sg brief))
  1;; Report values being returned*
  (cond ((null values)
	 (format stream "No values are being ~:[returned~;thrown~].~%" throw-p))
	(:else
	 (let ((n-values (length values)))
	   (format stream "The ~D value~:P being ~:[returned~;thrown~] ~:[are~;is~]:"
		   n-values throw-p (= n-values 1)))
	 (let ((*print-length* *error-print-length*)
	       (*print-level* *error-print-level*))
	   (dolist (val values)
	     (format stream "~%   ~S" val)))
	 (terpri stream))))


1;;
;;  New throw-trap handlers.*
;1;; 7/20/88 mjf/clm - changed for CLEH to use the new control-error.*
(defflavor throw-trap-error (tag restart values) (control-error)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (throw-trap-error :after :print-error-message) (sg brief stream)
   (declare (ignore sg))
   1;;  Report value(s) being thrown.*
   (unless brief
     (let ((n-values (length values)))
       (cond ((zerop n-values)
	      (format stream "No values are being thrown.~%"))
;	     ((= n-values 1)
;	      (format stream "The value being thrown is ~S." (first values)))
	     (:else
	      (format stream "The ~D value~:P being thrown ~:[are~;is~]:"
		      n-values (= n-values 1))
	      (let ((*print-length* *error-print-length*)
		    (*print-level* *error-print-level*))
		(loop for val in values
		      doing (format stream "~%   ~S" val)
		      finally (terpri))))))))

(defmethod (throw-trap-error :ucode-proceed-types) ()
  '(:new-tag))

(defmethod (throw-trap-error :case :proceed-asking-user :new-tag) (continuation read-object-function)
  2"Ask for a new tag to throw to instead."*
  1;; Resume stack group and return proceed type and new tag.*
  (funcall continuation :new-tag
	   (funcall read-object-function :eval-read
		    "Form to evaluate and use as tag: ")))

(defmethod (throw-trap-error :case :proceed-ucode-with-args :new-tag) (sg new-tag)
  1;; Store new tag in register and restart sg at restart.*
  (sg-store new-tag sg 'm-a)
  (sg-proceed-micro-pc sg restart))

1;;  Throw-Tag-Not-Found
;;  Throw-Tag-Not-Found Throw-Restart
;;  Throw-Tag-Not-Found Throw-N-Restart
;;
;;  First arg is the restart tag, which is also used to distinguish %throw from %throw-n.*

(def-ucode-error throw-tag-not-found throw-trap-error
  :tag (sg-contents sg 'm-a)
  :restart (second ete)
  :values (collect-thrown-values sg (second ete))
  :format-string "There was no pending CATCH for the tag ~S."
  :format-args (list (sg-contents sg 'm-a)))

1;;  Reserved-Internal-Throw-Tag
;;  Reserved-Internal-Throw-Tag Throw-Restart
;;  Reserved-Internal-Throw-Tag Throw-N-Restart
;;
;;  First arg is the restart tag, which is also used to distinguish %throw from %throw-n.*

(def-ucode-error reserved-internal-throw-tag throw-trap-error
  :tag (sg-contents sg 'm-a)
  :restart (second ete)
  :values (collect-thrown-values sg (second ete))
  :format-string "~S is a throw tag reserved for internal purposes;  it may not be used in THROW."
  :format-args (list (sg-contents sg 'm-a)))

(defun collect-thrown-values (sg restart)
   (if (eq restart 'throw-restart)
       (list (sg-contents sg 'm-t))		1; %THROW*
       (let ((n-values (sg-fixnum-contents sg 'm-g)))	1; %THROW-N*
	 (if (zerop n-values)
	     nil
	     (collect-top-n-from-stack sg n-values)))))

1;;  *Unwind-Stack-Illegal-Args
;;  *Unwind-Stack-Illegal-Args Tag
;;  *Unwind-Stack-Illegal-Args Count
;;  *Unwind-Stack-Illegal-Args Action
;;
;;  First arg indicates which argument(s) were illegal.*

(def-ucode-format-error *unwind-stack-illegal-args error
   (case (second ete)
     (tag "The tag argument to *UNWIND-STACK must be T.")
     (count "The count argument to *UNWIND-STACK must be either NIL or a positive fixnum.")
     (action "The action argument to *UNWIND-STACK must be non-NIL if the count is NIL.")))


1;; New microcode error entries for CHAPARRAL:

;; Internal-Memory-Location-OOB
;; INTERNAL-MEMORY-LOCATION-OOB M-B M-A
;; first arg is the memory selector
;; second arg is the index*

(def-ucode-format-error internal-memory-location-oob error
  "Internal memory location is out of range:  memory selector ~A, index ~A."
  (sg-contents sg (second ete))
  (sg-contents sg (third ete)))

1;; Cant-Initiate-On-This-Device-Type
;; CANT-INITIATE-ON-THIS-DEVICE-TYPE M-T
;; first arg is the device type*

(def-ucode-format-error cant-initiate-on-this-device-type error
  "Can't initiate on device type ~S."
  (sg-contents sg (second ete)))

1;; Unimplemented-Hardware
;; UNIMPLEMENTED-HARDWARE UNIBUS 
;; UNIMPLEMENTED-HARDWARE XBUS
;; first arg is the hardware type*

(def-ucode-format-error unimplemented-hardware error
  "~A is an unimplemented hardware type."
  (second ete))

1;; Too-Many-Page-Devices*

(def-ucode-format-error too-many-page-devices error
  "There is no room for another logical page device.")

1;; User-Nubus-Gacbl-Limit*

(def-ucode-format-error user-nubus-gacbl-limit error
  "Number of GACBLs exceeded limit in user NuBus operation.")

1;; User-Nubus-Parity-Limit-Exceeded*

(def-ucode-format-error user-nubus-parity-limit-exceeded error
  "Number of parity-error retries exceeded limit in user NuBus operation.")

1;; Page-Device-Is-Read-Only*

(def-ucode-format-error page-device-is-read-only error
  "Cannot return cluster to a read-only page device.")

1;; Invalid-Page-Cluster*

(def-ucode-format-error invalid-page-cluster error
  "Page cluster number was outside of page device bitmap.")

1;; Invalid-Page-Device-Number*

(def-ucode-format-error invalid-page-device-number error
  "The page device number was not valid.")

1;; No-Space-Left-For-Bitmap*

(def-ucode-format-error no-space-left-for-bitmap error
  "Not enough memory left in Device-Descriptor-Area for allocating the page device bitmap.")

1;; User-NuBus-Error
;; USER-NUBUS-ERROR M-B M-A M-D
;; first and second args are the hi and low addresses
;; third arg is the NuBus-TMs type*

(def-ucode-error user-nubus-error user-nubus-error
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :nubus-tms (ldb si:%%mem-error-nubus-tms (sg-contents sg (fourth ete)))
  :error-bits (sg-contents sg (fourth ete)))

(defflavor user-nubus-error (hi-addr lo-addr nubus-tms error-bits) (error)
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (user-nubus-error :report) (stream)
   (let ((error-type (select nubus-tms
		       (si:%nubus-tms-no-error "No Error")
		       (si:%nubus-tms-error    "Error")
		       (si:%nubus-tms-timeout  "Timeout")
		       (si:%nubus-tms-gacbl    "GACBL"))))
     (format stream "User NuBus Error of type ~S, at address #x~16R~16,6,'0R.~%Error Bits: #x~16R."
	     error-type hi-addr lo-addr error-bits)))

1;; User-NuBus-Error-Explicit M-A M-B M-D
;; M-A = VMA HI
;; M-B = VMA LO
;; M-D = NUBUS STATUS REGISTER*
(defflavor user-nubus-error-explicit (NSR-register fmt-string) (error user-nubus-error)
  :gettable-instance-variables
  :initable-instance-variables)
#|
(defflavor user-nubus-error-explicit (hi-addr lo-addr NSR-register fmt-string) (error)
  :gettable-instance-variables
  :initable-instance-variables)
|#
(defmethod (user-nubus-error-explicit :report) (stream)
  (format stream "User NuBus Error at address #x~16R~16,6,'0R with Nubus Status Register = #x~16R."
	  hi-addr lo-addr nsr-register)
  (format stream fmt-string))

(def-ucode-error USER-NUBUS-ERROR-NOT-MASTER user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII not enabled as NuBus Master in the configuration register.")

(def-ucode-error USER-NUBUS-ERROR-XCRFIFORESET user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII XCRFIFORESET bit active during NuBus cycle.")

(def-ucode-error USER-NUBUS-ERROR-WRITE-READ user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%Possible EXPII hardware failure!!! MCR indicates write, NSR indicates read.")

(def-ucode-error USER-NUBUS-ERROR-READ-WRITE user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%Possible EXPII hardware failure!!! MCR indicates read, NSR indicates write.")

(def-ucode-error USER-NUBUS-ERROR-ILLEGAL-BLOCKWRITE user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%Possible EXPII hardware failure!!! NSR indicates illegal mapped blockwrite.")

(def-ucode-error USER-NUBUS-ERROR-READ-FIFO user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII blockwrite address from READ FIFO does not match address in VMA.")

(def-ucode-error USER-NUBUS-ERROR-BUFFERED-BLOCKWRITE-BLOCKSIZE user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII pending BUFFERED BLOCKWRITE encoded blocksize <> number of words in WRITE FIFO.")

(def-ucode-error USER-NUBUS-ERROR-BUFFERED-WRITE user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII received NuBus Error or Timeout on a BUFFERED WRITE cycle with a cycle pending.")

(def-ucode-error USER-NUBUS-ERROR-BUFFERED-WRITE-GACBL user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII GACBL retry limit exhausted on a BUFFERED WRITE cycle with a cycle pending.")

(def-ucode-error USER-NUBUS-ERROR-UNSUPPORTED-BLOCKREAD user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII Blockread attempted to a NuBus slave that does not support blocks.")

(def-ucode-error USER-NUBUS-ERROR-READ user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII received NuBus Error or Timeout on a READ CYCLE.")

(def-ucode-error USER-NUBUS-ERROR-GACBL-LIMIT-ON-READ user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII GACBL retry limit exhausted on a READ CYCLE.")

(def-ucode-error USER-NUBUS-ERROR-WRITE user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII received NuBus Error or Timeout on a WRITE CYCLE.")

(def-ucode-error USER-NUBUS-ERROR-GACBL-LIMIT-ON-WRITE user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII GACBL retry limit exhausted on a WRITE CYCLE.")

(def-ucode-error USER-NUBUS-ERROR-BLOCKWRITE user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII Blockwrite number of words recovered for retry <> encoded blocksize in address.")

(def-ucode-error USER-NUBUS-ERROR-GREEDY-SLAVE user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII blockwrite addressed slave took more words than the encoded blocksize.")

(def-ucode-error USER-NUBUS-ERROR-OVERLOADED user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII blockwrite number of words loaded > encoded blocksize.")

(def-ucode-error USER-NUBUS-ERROR-BLOCK-NOT-HANDLED user-nubus-error-explicit
  :hi-addr (sg-contents sg (second ete))
  :lo-addr (sg-contents sg (third ete))
  :NSR-register (sg-contents sg (fourth ete))
  :fmt-string "~%EXPII Blockwrite attempted to a NuBus slave that does not support blocks.")


1;;  %Enable-NUPI-Locking-Not-Installed*

(def-ucode-format-error %enable-nupi-locking-not-installed error
   "The instruction %ENABLE-NUPI-LOCKING is not installed in the current microcode.")

1;;  %Disable-NUPI-Locking-Not-Installed*

(def-ucode-format-error %disable-nupi-locking-not-installed error
   "The instruction %DISABLE-NUPI-LOCKING is not installed in the current microcode.")

1;;  INVALID-SUBR-RETURN-PC
;;  INVALID-SUBR-RETURN-PC M-A*

(def-ucode-format-error invalid-subr-return-pc error
   "The return-PC for macro-POPJ is ~A, which is outside the range of the function ~S."
   (sg-contents sg (second ete))
   (sg-erring-function sg))

1;;
;;  Lexical-closure stuff.

;;  NO-LEXICAL-SUPERIOR*

(def-ucode-format-error no-lexical-superior error
  "There is no lexically-superior environment.")

1;;  OUT-OF-LEX-LEVELS*

(def-ucode-format-error out-of-lex-levels error
  "There are not ~D more lexically-superior environments."
  (sg-contents sg 'm-1))


1;;
;; CLOS ERRORS
;;*

;; Clos-Ran-out-of-args-before-method-found 

(def-ucode-format-error Clos-Ran-out-of-args-before-method-found error 
  "TICLOS - Ran out of arguments to dispatch on before method found
         during a Generic Function Dispatch.")

;; Generic-function-hash-table-value-error

(def-ucode-format-error Generic-function-hash-table-value-error error  
  "TICLOS - During Generic Function dispatching, a value was found in a Generic 
         function hash-table which was neither an array or locative.")

;; Generic-function-Individual-hash-table-value-error

(def-ucode-format-error Generic-function-Individual-hash-table-value-error error  
  "TICLOS - During Generic Function dispatching, a value was found in a Generic 
         function individual hash-table which was neither an array or locative.")

;; Support-vector-failure-for-clos

(def-ucode-format-error Support-vector-failure-for-clos error  
  "TICLOS - During Class description look up in the in the Class-Description table
         or the Array Class-Description table in the Support Vector area, neither
         a class-description nor a function to call was found.") 

;; Class-description-not-found-after-call-out

(def-ucode-format-error Class-description-not-found-after-call-out error  
  "TICLOS - The Class description was not found after the function in the 
         Class-Description table or the Array Class-Description table in the
         Support Vector area was called.")

;; CLOS-looping-forever-looking-for-key-in-hash-table

(def-ucode-format-error CLOS-looping-forever-looking-for-key-in-hash-table error
  "TICLOS - The microcode has has gone through the whole hash table without
         finding the key or finding a empty slot.  Hash table may be full.")

;; CLOS-Dispatch-failure-bad-hash-table

(def-ucode-format-error CLOS-Dispatch-failure-bad-hash-table error  
  "TICLOS - During Generic Function dispatching, the array being used as a hash
         table is not a valid TICLOS has table.")

;; CLOS-Permuation-list-does-not-contain-fixnums

(def-ucode-format-error CLOS-Permuation-list-does-not-contain-fixnums error  
  "TICLOS - During Generic Function dispatching, the permutation list
         passed in contains objects other than fixnums.  The 
         permutation list must be cdr-coded list of fixnums.")

;; CLOS-Permuation-list-not-a-cdr-coded-list

(def-ucode-format-error CLOS-Permuation-list-not-a-cdr-coded-list error  
  "TICLOS - During Generic Function dispatching, the permutation list
         passed in is not a cdr-coded list.  The permutation list
         must be cdr-coded list of fixnums.")

;; CLOS-Permutation-list-ended-before-method-found

(def-ucode-format-error CLOS-Permutation-list-ended-before-method-found error  
  "TICLOS - During Generic Function dispatching, the end of the permutation
         list was reached before the method to use was found.  Either the
         permutation list is incorrect or the dispatch tables are wrong.") 

;; CLOS-Permuation-list-not-fixnum-or-list

(def-ucode-format-error CLOS-Permuation-list-not-fixnum-or-list error  
  "TICLOS - The second argument to %dispatch-method, the permutation list
         was not a fixnum or a cdr-coded list of fixnums.")

;; CLOS--no-hash-table-after-calling-out

(def-ucode-format-error CLOS--no-hash-table-after-calling-out error  
  "TICLOS - The %Dispatch-method miscop was called with no hash table so
         the microcode called out. The function called out to did not
         return a hash table.")

;; CLOS-Method-was-found-but-is-not-a-function

(def-ucode-format-error CLOS-Method-was-found-but-is-not-a-function error  
  "TICLOS - During generic dispatching, the method was found but was not
         a compiled function.")

;; Cant-handle-self-mapping-table-bit-in-fef-header-in-generic-functions 

(def-ucode-format-error Cant-handle-self-mapping-table-bit-in-fef-header-in-generic-functions error  
  "TICLOS - During generic dispatching, the method was found but the self-mapping
         table bit is set which is illegal for TICLOS methods.")

;; CLOS-cant-handle-optionals-in-method

(def-ucode-format-error CLOS-cant-handle-optionals-in-method error  
  "TICLOS - During generic dispatching, the method was found but it contained 
         optional arguments which can't be handled by generic dispatching.")

;; CLOS-Non-cdr-coded-list-for-list-of-clos-mapping-tables 

(def-ucode-format-error CLOS-Non-cdr-coded-list-for-list-of-clos-mapping-tables error
  "TICLOS - during function calling, the Clos-info-provided bit was set in the call info
         word, but the mapping table list provided was not a cdr-coded list.")

;; Clos-mapping-table-not-array 

(def-ucode-format-error Clos-mapping-table-not-array error  
  "TICLOS - During transport of a SELF-REF-POINTER, a mapped reference was 
         indicated, but the value obtained from the local block was not a
         mapping table (array).")

;; CLOS-LOCAL-OR-ARG-NOT-INSTANCE 

(def-ucode-format-error CLOS-LOCAL-OR-ARG-NOT-INSTANCE error  
  "TICLOS - During transport of a SELF-REF-POINTER, the value obtained from
         the local block or argument block was not an instance.")

;; Clos-mapping-table-not-array-during-map-leader-ref 

(def-ucode-format-error Clos-mapping-table-not-array-during-map-leader-ref error  
  "TICLOS - During transport of a SELF-REF-POINTER, a map leader reference was
         indicated, but the value read from the array leader was not a mapping
         table (array).")

1;; UNDEFINED-KEYWORD-ARGUMENT*
;;; 09/06/88 clm - added for new keyword microcode-trap
;;; 03/15/89 clm - positioned after definition of macro def-ucode-error;
;;;                problem arising because of make-instance in macro-expansion
(def-ucode-error key-not-found sys:undefined-keyword-argument
  :format-string "~S is not one of the valid keyword arguments ~S."
  :format-args (list (sg-contents sg (second ete))
		     (sg-contents sg (third ete))))



;;;
;;; Misc Errors 
;;;

1;;
;;  Trap-on-Illop*

(def-ucode-format-error trap-on-illop dangerous-error
   "Something extremely nasty has happened while the trap-on-illop flag is set.")


1;;  Finale.

;; the big one*
(compile-flavor-methods condition
			simple-condition
			serious-condition
			warning
			simple-warning
			break
			trace-break
			error
			simple-error
			program-error
			control-error
			type-error
			simple-type-error
			stream-error
;			global:error
			ferror
			common-lisp-cerror
			dangerous-error
			arithmetic-error
			division-by-zero
			wrong-type-argument-error
			unclaimed-message-error
			sys:undefined-keyword-argument
			end-of-file
			parse-error
			read-end-of-file
			package-error
			package-not-found
			name-conflict-error
			failed-assertion
			redefinition
			disk-error
			tape-error
			unit-error
			unit-bad-number-error
			unit-offline-error
			arg-type-error
			floating-point-overflow
			floating-point-underflow
			floating-exponent-underflow-error
			floating-exponent-overflow-error
			bad-array-error
			array-number-dimensions-error
			subscript-error
			subscript-past-fill-pointer-error
			invalid-function
			storage-condition
			stack-overflow
			storage-exhausted 
			pdl-overflow-error
			no-currently-prepared-sheet-error
			cell-error
			cell-contents-error
			unbound-variable
			undefined-function
			function-entry-error
			mar-break
			step-break-error
			call-trap-error
			exit-trap-error
			throw-trap-error
			user-nubus-error
			)


