;;; -*- Mode: COMMON-LISP; Base: 10.; cold-load:t; Package: File-System -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985-1989 Texas Instruments Incorporated.  All rights reserved.

;;;  1/26/86  HW Modified get-bytes to not evaluate the return value.

;;;; Basic Structures

;;; -ab TGC
;;; 02-16-87 DAB :set-pointer changes. Changed Allocate-disk-block.
;;; 02-23-87 DAB Write-property-list. Error handling BIGNUM.
;;; 02.27.87 DAB Write-property-list. Errors off when value is a keyword.
;;; 03.26.87 DAB Changed write-property-list to special case :backup-date.
;;; 04.01.87 DAB Added Locking-with-timeout and locking-recursively-with-timeout

(DEFVAR LOCAL-FILE-SYSTEM-AREA si:working-storage-area
   "All of the internal structures made by the file system are consed in this area.
It may be reset only when the file system is dismounted.")

(DEFSTRUCT (DISK-CONFIGURATION (:TYPE :NAMED-ARRAY)
			       (:CONC-NAME DC-)
			       (:ALTERANT NIL)
			       (:MAKE-ARRAY (:AREA LOCAL-FILE-SYSTEM-AREA))
			       (:DEFAULT-POINTER DISK-CONFIGURATION)
			       (:CALLABLE-CONSTRUCTORS NIL)
			       (:PREDICATE NIL)
			       (:COPIER NIL))
  VERSION					;The version of the format on disk
  PARTITION-SIZE
  PUT-BASE					;Relative address of the page usage table
  PUT-SIZE
  ROOT-DIRECTORY				;The root directory.
  VBAT						;The virtual block table
  )  



(DEFSTRUCT (DIRECTORY (:TYPE :NAMED-ARRAY)
		      (:ALTERANT NIL)
		      (:MAKE-ARRAY (:AREA LOCAL-FILE-SYSTEM-AREA))
		      (:CONC-NAME "DIRECTORY-")
		      (:CALLABLE-CONSTRUCTORS NIL)
		      (:PREDICATE NIL)
		      (:COPIER NIL))
  NAME						;The name of the directory
  LOCK						;The lock associated with the directory
  MAP						;The map of the directory.
  FILES						;The files specified in the map.
  )  



(DEFSTRUCT (BASIC-FILE (:TYPE :NAMED-ARRAY)
		       (:CONC-NAME FILE-)
		       (:CONSTRUCTOR NIL)
		       (:ALTERANT NIL)
		       (:CALLABLE-CONSTRUCTORS NIL)
		       (:PREDICATE NIL)
		       (:COPIER NIL))
  NAME						;A string.
  LOCK						;A contention lock.
  MAP						;The map of the data.
  FILES						;Used only if the file is a directory.
  TYPE						;A string.
  VERSION					;A number (16b).
  displacement                                  ;a number (16b).
  entry-length                                  ;a number (16b).
  )  



(DEFSTRUCT (FILE (:TYPE :NAMED-ARRAY)
		 (:INCLUDE BASIC-FILE)
		 (:MAKE-ARRAY (:AREA LOCAL-FILE-SYSTEM-AREA))
                 (:CONC-NAME "FILE-")
		 (:CALLABLE-CONSTRUCTORS NIL)
		 (:ALTERANT ALTER-FILE)
		 (:PREDICATE NIL)
                 (:COPIER NIL))
  DEFAULT-BYTE-SIZE
  OVERWRITE-FILE				;Links overwriting/overwritten files.
  AUTHOR-INTERNAL				;A string
  CREATION-DATE-INTERNAL			;A number (32 bits)
  DIRECTORY					;Mommy.
  OPEN-COUNT					;The number of times this file is open
						;-1 if open for overwrite.
  ATTRIBUTES					;A 16-bit word of common properties
  PLIST						;Arbitrary properties.
  )  



(DEFSTRUCT (LINK (:TYPE :NAMED-ARRAY)
		 (:INCLUDE BASIC-FILE)
		 (:MAKE-ARRAY (:AREA LOCAL-FILE-SYSTEM-AREA))
		 (:CONC-NAME "LINK-")
		 (:CALLABLE-CONSTRUCTORS NIL)
		 (:ALTERANT ALTER-LINK)
		 (:PREDICATE NIL)
		 (:COPIER NIL))
  TO-STRING					;The string linked to, ie "AI: DLA; FOO >"
  )  


;;; Abort Protection Macro



(DEFMACRO ABORT-PROTECT (FORM IGNORE CLEANUP);9/30/83 RAF
  "Execute FORM. If execution of FORM is aborted, Execute the CLEANUP form."
  `(LET ((ABORT? T))
     (UNWIND-PROTECT (PROG1
		      ,FORM
		      (SETQ ABORT? ()))
       (IF ABORT?
	 ,CLEANUP))))  



;;;; Property Bit Definitions



(DEFVAR ATTRIBUTES (MAKE-ARRAY 16.))  



(DEFMACRO DEFINE-ATTRIBUTES (&BODY NUMBERS-AND-NAMES)
  `(EVAL-WHEN (COMPILE EVAL LOAD)
      ,@(LOOP FOR (NUMBER NAME) ON NUMBERS-AND-NAMES BY #'CDDR NCONCING
	   `((DEFPROP ,NAME ,(1+ (* NUMBER 64.)) ATTRIBUTE)
	     (SETF (AREF ATTRIBUTES ,NUMBER) ',NAME))))) 



(DEFINE-ATTRIBUTES 0. :DONT-DELETE;File cannot be deleted if on.
   1. :CLOSED;File is still being written if off.
   2. :DELETED;File is deleted but still open somewhere.
   3. :DUMPED;File has been dumped
   4. :DONT-REAP;Dont reap this file (not used internally)
   5. :CHARACTERS;The file is a CHARACTERS file.
   6. :DIRECTORY;The file is a subdirectory.
   7.
   :DONT-SUPERSEDE;Dont write another version past this file.
   ; 13 :MAY-BE-REAPED	  ;This bit is used by reaping programs.
   ;If it is on, the file is liable to be
   ;reaped.  Also, a reaping program may
   ;turn it on as a warning that it is liable
   ;to be reaped the next time around.
   )  



(DEFSUBST TEST-ATTRIBUTE (ATTRIBUTES INDICATOR)
  "Returns T if the INDICATOR is set in the ATTRIBUTES field."
  (LDB-TEST (GET INDICATOR 'ATTRIBUTE) ATTRIBUTES))  



(DEFSUBST FILE-ATTRIBUTE (FILE INDICATOR)
  (LDB-TEST (GET INDICATOR 'ATTRIBUTE) (FILE-ATTRIBUTES FILE)))  


(DEFUN SET-FILE-ATTRIBUTE (FILE INDICATOR VAL)
  (SETF (FILE-ATTRIBUTES FILE)
	(DPB (IF VAL
	       1.
	       0.)
	     (GET INDICATOR 'ATTRIBUTE) (FILE-ATTRIBUTES FILE)))
  (NOT (NULL VAL)))  

(DEFSETF FILE-ATTRIBUTE SET-FILE-ATTRIBUTE) 
;;; Syntax for attributes.



(DEFSUBST FILE-CLOSED? (FILE)
  (FILE-ATTRIBUTE FILE :CLOSED))  



(DEFSUBST FILE-DELETED? (FILE)
  (FILE-ATTRIBUTE FILE :DELETED))  



(DEFSUBST DIRECTORY? (FILE)
  (FILE-ATTRIBUTE FILE :DIRECTORY))  



(DEFSUBST FILE-DONT-SUPERSEDE? (FILE)
  (FILE-ATTRIBUTE FILE :DONT-SUPERSEDE))  



(DEFSUBST ROOT-DIRECTORY? (FILE)
  (NULL (FILE-DIRECTORY FILE)))  



(DEFSUBST FILE-NPAGES (FILE)
  (MAP-NPAGES (FILE-MAP FILE)))  

;;;; Maps



(DEFSUBST MAP-NBLOCKS (MAP)
  (ARRAY-LEADER MAP 0.))  



(DEFSUBST MAP-BLOCK-LOCATION (MAP INDEX)
  (AREF MAP INDEX 0.))  



(DEFSUBST MAP-BLOCK-SIZE (MAP INDEX)
  (AREF MAP INDEX 1.))  

;;;; Printers



(DEFun FILE-NSI (operation &optional file stream &rest arg)
  (declare (ignore arg))
  (case operation
  (:PRINT-SELF 
   (PRINTING-RANDOM-OBJECT (FILE STREAM :NO-POINTER)
      (COND
	((DIRECTORY? FILE) (FORMAT STREAM "LMFS-Directory~{ ~A~}" (DIRECTORY-FULL-NAME FILE)))
	(T
	 (FORMAT STREAM "LMFS-File ~S"
		 (LM-NAMESTRING () () (DIRECTORY-FULL-NAME (FILE-DIRECTORY FILE))
				(FILE-NAME FILE) (FILE-TYPE FILE) (FILE-VERSION FILE)))))))
  (:which-operations
   '(:print-self))))



(DEFun DIRECTORY-NSI (operation directory stream &rest arg)
  (declare (ignore arg))
  (case operation
    (:PRINT-SELF 
     (PRINTING-RANDOM-OBJECT (DIRECTORY STREAM :TYPEP :NO-POINTER)
       (PRIN1 (DIRECTORY-NAME DIRECTORY) STREAM)))
    (:which-operations
     '(:print-self))))  



(DEFUN DIRECTORY-FULL-NAME (DIRECTORY)
  (LET ((RESULT))
    (LOOP
     (IF (NOT (ROOT-DIRECTORY? DIRECTORY))
       (PROGN
	 (PUSH (DIRECTORY-NAME DIRECTORY) RESULT)
	 (SETQ DIRECTORY (FILE-DIRECTORY DIRECTORY)))
       (RETURN RESULT)))))  



(DEFun MAP-NSI (operation map &optional stream &rest arg)
  (declare (ignore arg))
  (case operation
    (:PRINT-SELF 
     (PRINTING-RANDOM-OBJECT (MAP STREAM :TYPEP) (FORMAT STREAM "~D block~:P" (MAP-NBLOCKS MAP))))
    (:DESCRIBE (DESCRIBE-MAP MAP))
    (:which-operations
     '(:print-self :describe))))  



(DEFPROP FILE FILE-NSI NAMED-STRUCTURE-INVOKE)  


(DEFPROP LINK FILE-NSI NAMED-STRUCTURE-INVOKE)  


(DEFPROP DIRECTORY DIRECTORY-NSI NAMED-STRUCTURE-INVOKE)  


(DEFPROP MAP MAP-NSI NAMED-STRUCTURE-INVOKE)  



;;; Randomness.






(DEFUN LMFS-KEYWORD-OPTION (OPTIONS KEYWORD DEFAULT)
  (COND
    ((NULL OPTIONS) DEFAULT)
    ((EQ KEYWORD (CAR OPTIONS)) (CADR OPTIONS))
    (T (LMFS-KEYWORD-OPTION (CDDR OPTIONS) KEYWORD DEFAULT))))  

;;;; Locking Macros



(DEFMACRO LOCKING (LOCK &BODY BODY)
  "Execute BODY with LOCK locked."
  (LET ((LOCK-CELL (GENSYM)))
    `(LET ((,LOCK-CELL (LOCF ,LOCK)))
       (UNWIND-PROTECT (PROGN
			(PROCESS-LOCK ,LOCK-CELL)
			,@BODY)
	 (%STORE-CONDITIONAL ,LOCK-CELL CURRENT-PROCESS ())))))  

(DEFMACRO LOCKING-with-timeout (LOCK timeout &BODY BODY)   ;04.01.87 DAB
  "Execute BODY with LOCK locked. If timeout is non-nil, returns error signal SYS:lock-timeout if a lock can not be 
   returned in the timeout interval. Timeout is in 60th of a second."
  (LET ((LOCK-CELL (GENSYM)))
    `(LET ((,LOCK-CELL (LOCF ,LOCK)))
       (UNWIND-PROTECT (PROGN
			(PROCESS-LOCK ,LOCK-CELL () "lock w/timeout" ,timeout)
			,@BODY)
	 (%STORE-CONDITIONAL ,LOCK-CELL CURRENT-PROCESS ())))))


(DEFMACRO LOCKING-RECURSIVELY (LOCK &BODY BODY)
  "Execute BODY with LOCK locked; don't die if already locked."
  (LET ((LOCK-CELL (GENSYM))
	(LOCK-OWNED (GENSYM)))
    `(LET* ((,LOCK-CELL (LOCF ,LOCK))
	    (,LOCK-OWNED (EQ (CAR ,LOCK-CELL) CURRENT-PROCESS)))
       (UNWIND-PROTECT (PROGN
			(IF (NOT ,LOCK-OWNED)
			  (PROCESS-LOCK ,LOCK-CELL))
			,@BODY)
	 (IF (NOT ,LOCK-OWNED)
	   (%STORE-CONDITIONAL ,LOCK-CELL CURRENT-PROCESS ()))))))  

(DEFMACRO LOCKING-RECURSIVELY-with-timeout (LOCK TIMEOUT &BODY BODY) ;04.01.87 DAB
  "Execute BODY with LOCK locked; don't die if already locked."
  (LET ((LOCK-CELL (GENSYM))
	(LOCK-OWNED (GENSYM)))
    `(LET* ((,LOCK-CELL (LOCF ,LOCK))
	    (,LOCK-OWNED (EQ (CAR ,LOCK-CELL) CURRENT-PROCESS)))
       (UNWIND-PROTECT (PROGN
			(IF (NOT ,LOCK-OWNED)
			  (PROCESS-LOCK ,LOCK-CELL () "Lock w/timeout" ,timeout))
			,@BODY)
	 (IF (NOT ,LOCK-OWNED)
	   (%STORE-CONDITIONAL ,LOCK-CELL CURRENT-PROCESS ()))))))


(DEFMACRO REQUIRE-LOCK (LOCK)
  (LET ((ERSTR
	 (FORMAT () "~S should be locked here but is not." (IF (ATOM LOCK)
							     LOCK
							     (CAR LOCK)))))
    `(IF (NOT (EQ ,LOCK CURRENT-PROCESS))
       (FERROR () ,ERSTR))))  



(DEFMACRO MODIFYING-DIRECTORY (DIRECTORY &BODY BODY)
  "Lock DIRECTORY, execute BODY, and then mark DIRECTORY as modified."
  (LET ((LOCK-CELL (GENSYM)))
    `(LET ((,LOCK-CELL (LOCF (DIRECTORY-LOCK ,DIRECTORY))))
       (UNWIND-PROTECT (PROGN
			(PROCESS-LOCK ,LOCK-CELL)
			,@BODY)
	 (PROGN
	   (SETF (FILE-CLOSED? DIRECTORY) ())
	   (%STORE-CONDITIONAL ,LOCK-CELL CURRENT-PROCESS ()))))))  



(DEFMACRO OPENING-INPUT-FILE ((FILE DIRECTORY NAME TYPE VERSION) &BODY BODY)
  `(LET ((,FILE NIL))
     (UNWIND-PROTECT (PROGN
		      (SETQ ,FILE (LOOKUP-FILE ,DIRECTORY ,NAME ,TYPE ,VERSION :ERROR () T))
		      ,@BODY)
       (IF (NOT (NULL ,FILE))
	 (LMFS-CLOSE-FILE ,FILE)))))  



(DEFPARAMETER LM-VERSION 5.
   "Version of the file system which will support vbats in the configuration area.")  



(DEFPARAMETER LM-MINIMUM-VERSION 4. "Lowest version of file system we can handle.")  



(DEFVAR LM-UNIT 0.)  				;See SET-LM-BAND


(DEFVAR LM-PARTITION "FILE")  


(DEFVAR LM-PARTITION-POSSIBILITIES '((0. "FILE") (1. "FILE")))  


(DEFVAR LM-PARTITION-BASE)  			;The address of the partition.

;; ***** This should be replaced by a device scheme.  *****


(DEFVAR DISK-CONFIGURATION)  


(DEFVAR DISK-CONFIGURATION-RQB)  


(DEFVAR DISK-CONFIGURATION-BUFFER)  


(DEFVAR DISK-CONFIGURATION-BUFFER-POINTER)  


(DEFVAR DISK-CONFIGURATION-LOCK ())  



(DEFVAR PUT-LOCK ())  


(DEFVAR PUT-SCANNING-INDEX 0.)  


(DEFVAR PUT-RQB)  		;The page usage table is indirected into this.


(DEFVAR PAGE-USAGE-TABLE)  	;The actual table.

;; The first entry in the page usage table tells whether the table is consistent
;; with the rest of the system.  If it is not, this only reflects the fact that
;; some pages are lost...


(DEFPARAMETER PUT-CONSISTENT 1.)  


(DEFPARAMETER PUT-INCONSISTENT 2.)  

;; The remaining entries describe all the remaining pages in the file (not counting
;; the configuration page, which is always used).


(DEFPARAMETER PUT-FREE 0.)  				;Page is up for grabs.


(DEFPARAMETER PUT-RESERVED 1.)  			;Page is allocated but (likely) deletable.


(DEFPARAMETER PUT-USED 2.)  				;Page is allocated in core and on disk.


;(DEFPARAMETER PUT-UNUSABLE 3.)  			;Page should not be referenced ever.
(DEFPARAMETER puT-RSVD-NOT-USED 3.)                     ;Pages allocated to a directory but not written too.


(DEFPARAMETER *DIRECTORY-DIVIDER* 200.
   "Fraction of the file-band for directories.  A value of N will
   cause the allocator to try to allocate directories in the 
   first 1Nth of the file partition and files in the rest.")  



(DEFVAR PAGE-SIZE-IN-BITS (* si:disk-block-word-size 32.))  
(DEFVAR PAGE-SIZE-IN-bytes si:disk-block-byte-size)
;; This is also in MTDEFS
;;;(DEFVAR PAGE-SIZE-IN-BITS (* PAGE-SIZE 32.))
;;;(DEFVAR PAGE-SIZE-IN-bytes (* PAGE-SIZE 4.))



(REMPROP 'QUOTIENT-CEILING 'SOURCE-FILE-NAME)  
#+MIT


(DEFSUBST QUOTIENT-CEILING (Y X)
  (CEILING Y X))  
#-MIT


(DEFUN QUOTIENT-CEILING (Y X)
  "The smallest integer greater than or equal to Y/X"
   (QUOTIENT (+ Y (1- X)) X))  

;; Get a buffer of the correct byte size


(DEFUN GET-RQB-ARRAY (RQB BYTE-SIZE &AUX TYPE)
  (COND
    ((= BYTE-SIZE 16.) (RQB-BUFFER RQB))
    ((= BYTE-SIZE 8.) (RQB-8-BIT-BUFFER RQB))
    ((SETQ TYPE (CDR (ASSOC BYTE-SIZE '((4. . ART-4B) (2. . ART-2B) (1. . ART-1B)) :TEST #'EQ)))
     (MAKE-ARRAY (FLOOR (* (RQB-NPAGES RQB) PAGE-SIZE-IN-BITS) BYTE-SIZE)
		 :AREA LOCAL-FILE-SYSTEM-AREA
		 :TYPE TYPE
		 :DISPLACED-TO RQB
		 :DISPLACED-INDEX-OFFSET		     ;; This is a system bug.  Don't try to figure it out.
		     (* (%P-CONTENTS-OFFSET (RQB-BUFFER RQB) 3.) (FLOOR 16. BYTE-SIZE))))
    (T (FERROR () "~D is an invalid byte size." BYTE-SIZE))))  


;;; Use this to write out new maps.



(DEFMACRO WITH-MAP-STREAM-OUT ((STREAM . ARGS) &BODY BODY)
  "A variant of WITH-OPEN-STREAM, it returns a new map."
  `(LET ((.FILE-ABORTED-FLAG. ':ABORT)
	 (,STREAM NIL))
     (UNWIND-PROTECT (PROGN
		      (SETQ ,STREAM (MAKE-MAP-STREAM-OUT . ,ARGS))
		      ,@BODY
		      (SETQ .FILE-ABORTED-FLAG. ()))
       (IF (AND (NOT (NULL ,STREAM)) (NOT (ERRORP ,STREAM)))
	 (SEND ,STREAM ':CLOSE .FILE-ABORTED-FLAG.)))
     (AND (NULL .FILE-ABORTED-FLAG.) (SEND ,STREAM ':MAP))))  

(DEFMACRO WITH-MAP-STREAM-UPDATE ((STREAM . ARGS) &BODY BODY) 
  "A variant of WITH-OPEN-STREAM."
  `(LET ((.FILE-ABORTED-FLAG. ':ABORT)
	 (,STREAM NIL))
     (UNWIND-PROTECT
       (PROGN (SETQ ,STREAM (MAKE-MAP-STREAM-UPDATE . ,ARGS))
	      ,@BODY
	      (SETQ .FILE-ABORTED-FLAG. ()))
       (IF (AND (NOT (NULL ,STREAM))
		(NOT (ERRORP ,STREAM)))
	   (SEND ,STREAM ':CLOSE .FILE-ABORTED-FLAG.)))))

(DEFMACRO WITH-MAP-STREAM-IN ((STREAM . ARGS) &BODY BODY)
  "A variant of WITH-OPEN-STREAM."
  `(LET ((.FILE-ABORTED-FLAG. ':ABORT)
	 (,STREAM NIL))
     (UNWIND-PROTECT (PROG1
		      (PROGN
			(SETQ ,STREAM (MAKE-MAP-STREAM-IN . ,ARGS))
			,@BODY)
		      (SETQ .FILE-ABORTED-FLAG. ()))
       (IF (AND (NOT (NULL ,STREAM)) (NOT (ERRORP ,STREAM)))
	 (SEND ,STREAM ':CLOSE .FILE-ABORTED-FLAG.)))))  


;; Directory information on the disk is stored in packed format.  The
;; following functions allow the file system to read and write fixnums
;; and bignums in such a packed format.


;; These macros do not use ONCE-ONLY because it loses in the simple case of NBYTES = 1.

;(DEFMACRO GET-BYTES (STREAM NBYTES)
;  `(LOOP FOR I FROM (1- ,NBYTES) DOWNTO 0
;	 AS X = (FUNCALL ,STREAM ':TYI)
;	 WHEN (PLUSP I) DO (SETQ X `(* ,X ,(expt 256. I)))
;	 COLLECT X INTO FORMS
;	 FINALLY (IF (= ,NBYTES 1)
;		     (SETQ FORMS (CAR FORMS))
;		     (SETQ FORMS (CONS '+ FORMS)))
;	 (RETURN  (eval FORMS))))

(PROCLAIM '(inline get-bytes))
(DEFUN GET-BYTES (STREAM NBYTES)
     (LET ((VALUE 0))
       (DOTIMES (I NBYTES VALUE)
         (SETF VALUE (+ (* VALUE 256.)
   		     (SEND STREAM :TYI))))))
  

(DEFMACRO PUT-BYTES (STREAM NBYTES VALUE)
  (if (fixnump nbytes)
      `(PROGN .  ,(LOOP FOR PPSS FROM (+ (* (1- nbytes) 512.) 8.) DOWNTO 8. BY 512.
		   COLLECTING `(FUNCALL ,STREAM ':TYO (LDB ,PPSS ,VALUE))))
      `(LOOP FOR PPSS FROM (+ (* (1- ,nbytes) 512.) 8.) DOWNTO 8. BY 512.
		   do (FUNCALL ,STREAM ':TYO (LDB PPSS ,VALUE)))))


(DEFUN PUT-STRING (STRING STREAM)
  (FUNCALL STREAM :TYO (ARRAY-ACTIVE-LENGTH STRING))
  (FUNCALL STREAM :STRING-OUT STRING))  



(DEFUN GET-STRING (STREAM)
  (LET* ((LEN (FUNCALL STREAM :TYI))
	 (ARR (MAKE-ARRAY LEN :ELEMENT-TYPE 'STRING-CHAR)))
    (FUNCALL STREAM :STRING-IN () ARR 0. LEN)
    ARR))  


(DEFPARAMETER GLOBAL-OR-USER-OR-NIL (LIST (FIND-PACKAGE "GLOBAL") (FIND-PACKAGE "USER") ()))      ;Uninterned (interned later)


(DEFUN WRITE-PROPERTY-LIST (STREAM PLIST &AUX (X (FLOOR (LENGTH PLIST) 2.)))
  (PUT-BYTES STREAM 1. X)
  (DO ((I X (1- I))
       (L PLIST (CDDR L))
       (PROP))
      ((ZEROP I))
    (PUT-STRING (FORMAT () "~s" (CAR L)) STREAM)
    (COND
      ((NULL (SETQ PROP (CADR L))) (FUNCALL STREAM :TYO 0.))
      ((EQ PROP T) (FUNCALL STREAM :TYO 1.))
      ((SYMBOLP PROP)
       (COND
	 ((MEMBER (SYMBOL-PACKAGE PROP) GLOBAL-OR-USER-OR-NIL :TEST #'EQ)
	  (FUNCALL STREAM :TYO 2.) (PUT-STRING (SYMBOL-NAME PROP) STREAM))
	 (T (FUNCALL STREAM :TYO 3.)
	    ;; Use this odd order for ease in interning later.
	    (PUT-STRING (SYMBOL-NAME PROP) STREAM)
	    (PUT-STRING  (si:package-name (or (SYMBOL-PACKAGE PROP) ;may return nil, if so use keywaor package.
					      (si:find-package "KEYWORD"))) STREAM)))) ;02.27.87
      ((AND (NUMBERP PROP)			;modified to handle neagtive numbers and bignums in property list
	    (integerp PROP))
       (cond ((and (typep prop 'BIGNUM)
		   (not (eq (car L) :backup-date)))  ;03.26.87 DAB Write :BACKUP-DATE as a fixnum,even it is a bignum.
	      (let* ((prop-value (abs prop))
		     (length-prop (length (format nil "~a" prop-value))))   
		(funcall stream :tyo 9.)
		(funcall stream :tyo length-prop)
		(if (plusp prop)
		    (funcall stream :tyo 0)	;positive bignum
		    (funcall stream :tyo 1)	;negative bignum
		    )
		(put-bytes stream length-prop  prop-value)
		))
	     (t (if (or (plusp prop)
			(zerop prop))
		    (progn
		      (FUNCALL STREAM :TYO 4.)	;positive integers
		      (PUT-BYTES STREAM 3 PROP))	;negative integers
		    (FUNCALL STREAM :TYO 8.)
		    (PUT-BYTES STREAM 3 PROP)))))
      ((STRINGP PROP) (FUNCALL STREAM :TYO 5.) (PUT-STRING PROP STREAM))
      ((CONSP PROP) (FUNCALL STREAM :TYO 6.)
		    (LET ((*PACKAGE* SI:PKG-USER-PACKAGE)
			  (*READ-BASE* 10.)
			  (*NOPOINT T)
			  (SI:PRINT-READABLY T))
		      (PRIN1 PROP STREAM)))
      ((TYPEP PROP 'PATHNAME) (FUNCALL STREAM :TYO 7.)
			      (LET ((*PACKAGE* SI:PKG-USER-PACKAGE) (*PRINT-BASE* 10.) (*NOPOINT T)
				    (SI:PRINT-READABLY T))
				;; These are the arguments to MAKE-FASLOAD-PATHNAME.
				(FUNCALL STREAM :TYO #\()
				(PRIN1-THEN-SPACE (FUNCALL (PATHNAME-HOST PROP) :NAME-AS-FILE-COMPUTER) STREAM)
				(PRIN1-THEN-SPACE (PATHNAME-DEVICE PROP) STREAM)
				(PRIN1-THEN-SPACE (PATHNAME-DIRECTORY PROP) STREAM)
				(PRIN1-THEN-SPACE (PATHNAME-NAME PROP) STREAM)
				(PRIN1-THEN-SPACE (PATHNAME-TYPE PROP) STREAM)
				(PRIN1 (PATHNAME-VERSION PROP) STREAM)
				(FUNCALL STREAM :TYO #\))))
      (T (FERROR () "I don't know how to write ~S as a property." PROP)))))  



(DEFUN READ-PROPERTY-LIST (STREAM &AUX LIST (PAK PKG-KEYWORD-PACKAGE))
  (SETQ LIST (MAKE-LIST (* (FUNCALL STREAM :TYI) 2.)))
  (DO ((L LIST (CDDR L)))
      ((NULL L))
    (RPLACA L (LET ((*PACKAGE* PAK))
		(IGNORE-ERRORS (READ-FROM-STRING (GET-STRING STREAM) T))))
    (CASE (FUNCALL STREAM :TYI)
      (0.)
      (1. (SETF (CADR L) T))
      (2. (SETF (CADR L) (INTERN (GET-STRING STREAM) PAK)))
      (3.
       (SETF (CADR L)
	     (INTERN (READ-FROM-STRING (GET-STRING STREAM) T)
		     (PKG-FIND-PACKAGE (GET-STRING STREAM) :FIND))))
      (4. (SETF (CADR L) (GET-BYTES STREAM 3.)))
      (5. (SETF (CADR L) (GET-STRING STREAM)))
      (6.
       (SETF (CADR L) (PKG-BIND SI:PKG-USER-PACKAGE (LET ((*READ-BASE* 10.))
							  ;; This can lose pretty badly with #<'s, etc.  -- DLA
						      (READ STREAM T)))))
      (7.
       (PKG-BIND ""
	  (LET* ((*READ-BASE* 10.) (FORM (READ STREAM T))
		 (DEFAULT-CONS-AREA WORKING-STORAGE-AREA));<-*
	    (SETF (CADR L)
		  (IF (= (LENGTH FORM) 6.)
		    (APPLY #'MAKE-FASLOAD-PATHNAME FORM)
		    (FERROR () "Invalid pathname in property list."))))))
      (8. (SETF (CADR L) (%logdpb #x01 #o3001 (GET-BYTES STREAM 3.))))    ;negative numbers 
      (9. (let* ((length-prop (get-bytes stream 1))                       ;BIGNUMs
		 (sign-prop (get-bytes stream 1))
		 (prop-value (get-bytes stream length-prop)))
	         (setf (cadr l) (if (eq sign-prop 0) prop-value (* -1 prop-value)))))
      (OTHERWISE (FERROR () "Invalid Plist property designator."))))
  LIST)  


;; Functions for creating and manipulating MAPs.



(DEFUN MAP-CREATE (&OPTIONAL (NBLOCKS 32.) INPUTP &AUX MAP)
  (SETQ MAP
	(MAKE-ARRAY (LIST-IN-AREA LOCAL-FILE-SYSTEM-AREA NBLOCKS 2.) :LEADER-LENGTH 2.
		    :NAMED-STRUCTURE-SYMBOL 'MAP))
  (SETF (MAP-NBLOCKS MAP) (IF INPUTP
			    NBLOCKS
			    0.))
  MAP)  

;; Returns the number of pages used (as opposed to the number of blocks) in the map.


(DEFUN MAP-NPAGES (MAP)
  (LOOP WITH NBLOCKS = (MAP-NBLOCKS MAP) FOR I FROM 0. BELOW NBLOCKS SUMMING
     (QUOTIENT-CEILING (MAP-BLOCK-SIZE MAP I) PAGE-SIZE-IN-BITS)))  

;; This is used during output, when a block is added onto the end of a file.


(DEFUN MAP-APPEND-BLOCK (MAP LOCATION SIZE &AUX (NBLOCKS (MAP-NBLOCKS MAP)))
  (AND (= NBLOCKS (ARRAY-DIMENSION MAP 0.)) (ADJUST-ARRAY MAP `(,(+ NBLOCKS 32.) 2.)))
  (WITHOUT-INTERRUPTS (SETF (MAP-NBLOCKS MAP) (1+ NBLOCKS))
     (SETF (MAP-BLOCK-LOCATION MAP NBLOCKS) LOCATION) (SETF (MAP-BLOCK-SIZE MAP NBLOCKS) SIZE)))  

;; Returns the length of the area mapped in bits.


(DEFUN MAP-LENGTH (MAP)
  (LOOP WITH NBLOCKS = (MAP-NBLOCKS MAP) FOR I FROM 0. BELOW NBLOCKS SUMMING
     (MAP-BLOCK-SIZE MAP I)))  



(DEFUN MAP-READ (STREAM)
  (LOOP WITH NBLOCKS = (GET-BYTES STREAM 2.)
	WITH MAP = (MAP-CREATE NBLOCKS T)
	FOR I FROM 0.
        BELOW NBLOCKS
	DOING (SETF (MAP-BLOCK-LOCATION MAP I) (GET-BYTES STREAM 3.))
              (SETF (MAP-BLOCK-SIZE MAP I) (GET-BYTES STREAM 3.))
	      FINALLY (RETURN MAP)))  



(DEFUN MAP-WRITE (STREAM MAP)
  (LET ((NBLOCKS (MAP-NBLOCKS MAP)))
    (PUT-BYTES STREAM 2. NBLOCKS)
    (DOTIMES (I NBLOCKS)
      (MAP-WRITE-1-BLOCK STREAM (MAP-BLOCK-LOCATION MAP I) (MAP-BLOCK-SIZE MAP I))))
  MAP)  



(DEFUN MAP-WRITE-1-BLOCK (STREAM LOCATION SIZE)
  (PUT-BYTES STREAM 3. LOCATION)
  (PUT-BYTES STREAM 3. SIZE))  



(DEFUN DESCRIBE-MAP (MAP &OPTIONAL (STREAM *STANDARD-OUTPUT*))
  (FORMAT STREAM "~&~S maps out the following blocks:~@
 ~2t Index   Location      Size   Status~%"
	  MAP)
  (LET ((NBITS 0.))
    (DOTIMES (I (MAP-NBLOCKS MAP))
      (FORMAT STREAM "~6O: ~10O ~10O   ~[Free~;Reserved~;Used~;Bad~]~%" I
	      (MAP-BLOCK-LOCATION MAP I) (MAP-BLOCK-SIZE MAP I)
	      (AREF PAGE-USAGE-TABLE (MAP-BLOCK-LOCATION MAP I)))
      (INCF NBITS (MAP-BLOCK-SIZE MAP I)))
    (FORMAT STREAM "~%Total of ~D block~:P, ~D bit~:P.~%" (MAP-NBLOCKS MAP) NBITS)))  



(DEFUN FILE-DATA-LENGTH (FILE)
  (MAP-LENGTH (FILE-MAP FILE)))  



(DEFUN FILE-TRUENAME (FILE)
  (MAKE-PATHNAME-INTERNAL () SI:LOCAL-HOST ()
			  (OR (DIRECTORY-FULL-NAME (FILE-DIRECTORY FILE)) :ROOT)
			  (FILE-NAME FILE)
			  (IF (EQUAL (FILE-TYPE FILE) "")
			    :UNSPECIFIC
			    (FILE-TYPE FILE))
			  (FILE-VERSION FILE)))  


;; Error handler interface.



(DEFMACRO IDENTIFY-FILE-OPERATION (OPERATION &BODY BODY)
  `(LET ((*CURRENT-FILE-OPERATION* (OR *CURRENT-FILE-OPERATION* ,OPERATION)))
     ,@BODY))  

;; Wrap this around things which want to do things the simple way.
;; If ERROR-P is non-NIL and an error occurs, the error object is returned from this form

(DEFMACRO HANDLING-ERRORS (ERROR-P &BODY BODY)
  `(CATCH 'LM-TRAP-ERRORS
		    (LET ((LM-TRAP-ERRORS (NOT ,ERROR-P)))
		      (FILE-OPERATION-RETRY . ,BODY))))  

;; This is bound to T if one is inside LM-TRAP-ERROR, in which case control
;; is thrown to LM-TRAP-ERROR


(DEFVAR LM-TRAP-ERRORS ())  



(DEFVAR *CURRENT-FILE-OPERATION* ())  



(DEFPROP LM-SIGNAL-ERROR T :ERROR-REPORTER)

(DEFUN LM-SIGNAL-ERROR (SYMBOL &OPTIONAL SOURCE PROCEEDABLE &REST MAKE-CONDITION-ARGS)
  (declare (special lm-signal-pathname-object))	; object of type pathname. enclosed in a LET in high level file functions.
  (OR SOURCE
      (AND (boundp 'lm-signal-pathname-object)
	   (OR (ignore-errors (TYPEP lm-signal-pathname-object 'PATHNAME))
	       (ignore-errors (TYPEP  lm-signal-pathname-object 'SI:FILE-STREAM-MIXIN)))
	   (SETQ SOURCE  lm-signal-pathname-object)))
  (AND SOURCE (NOT (TYPEP SOURCE 'PATHNAME)) (SETQ SOURCE (SEND SOURCE :PATHNAME)))
  (LET ((ERROR
	  (APPLY 'MAKE-CONDITION (OR (GET SYMBOL 'ERROR-SIGNALER) SYMBOL)
		 (LET ((STRING (OR (GET SYMBOL 'ERROR-STRING) (SYMBOL-NAME SYMBOL))))
		   (IF SOURCE
		       (STRING-APPEND STRING " for "
				      (STRING-SUBST-CHAR #\* #\~ (SEND SOURCE :STRING-FOR-PRINTING)))
		       STRING))
		 SOURCE MAKE-CONDITION-ARGS)))
    (IF LM-TRAP-ERRORS
	(THROW 'LM-TRAP-ERRORS
	       ERROR))
    (SIGNAL ERROR :PROCEED-TYPES
	    (COND
	      ((CONSP PROCEEDABLE) PROCEEDABLE)
	      (PROCEEDABLE '(:RETRY-FILE-OPERATION))))))  

;; Use this when an error occurs in file lookup.


(DEFPROP LM-LOOKUP-ERROR T :ERROR-REPORTER)  


(DEFUN LM-LOOKUP-ERROR (SYMBOL DIR NAM TYP VER &OPTIONAL (OPERATION *CURRENT-FILE-OPERATION*))
  (LM-SIGNAL-ERROR SYMBOL
		   (MAKE-PATHNAME :HOST "LM" :DEVICE "DSK" :DIRECTORY
				  (COND
				    ((STRINGP DIR) DIR)
				    ((CONSP DIR) DIR)
				    (T (DIRECTORY-FULL-NAME DIR)))
				  :NAME NAM :TYPE TYP :VERSION VER)
		   () OPERATION))  



;; Error Definitions

;; ALLOCATE-DISK-BLOCK returns the base and size of a block which has pages
;; pre-set to PUT-RESERVED.
;; {VERIFY, CHANGE, SET}-{BLOCK, MAP}-DISK-SPACE.  They should all be called
;; inside the special form USING-PUT.  SET-... should be used minimally.
;; WRITE-PUT (&optional force-p) may be called any time inside a USING-PUT
;; form to force the put to be written onto the disk.  If FORCE-P is nil, and
;; the PUT hasn't been modified in core, this is a no-op.  (WRITE-PUT) is implicit
;; when exitting a USING-PUT.  The USING-PUT special form may be enterred recursively,
;; and in fact usually is.

;; USE-{MAP,BLOCK}-DISK-SPACE??

;; This variable is T if the put has been modified in core and not written to the disk.


(DEFVAR PUT-MODIFIED ())  

;; This array contains four numbers:  The number of pages in the PUT which
;; have the index as their status.


(DEFVAR PUT-USAGE-ARRAY (MAKE-ARRAY 4 ':TYPE 'ART-32B))  

;; This is the minimum number of free pages that must be available in order
;; to write files.  Note that internal structures may still be written.


(DEFVAR PUT-MINIMUM-FREE-PAGES 50.)  

;; This is bound by LM-WRITE-DIRECTORY to T.  This tells ALLOCATE-DISK-BLOCK
;; that it is OK to go below PUT-MINIMUM-FREE-PAGES.


(DEFVAR WRITING-INTERNAL-STRUCTURES ())  

(DEFVAR STANDARD-BLOCK-SIZE 20.)  

(DEFVAR LAST-ALLOCATED-BLOCK 0.)

(DEFVAR DIRECTORY-AREA 0)

(DEFVAR LAST-ALLOCATED-directory-BLOCK 0.)


(DEFUN ALLOCATE-DISK-BLOCK (&OPTIONAL (PAGES STANDARD-BLOCK-SIZE) (new-status put-reserved))
  "Returns the displacement and length of a free block of disk in
   the file partition.  The block is marked as reserved before returning."
  (BLOCK ALLOCATE-DISK-BLOCK
    (DO ()
	(NIL)
      (LET ((DISPLACEMENT NIL)
	    (LENGTH ())
	    )
	(LOCKING-RECURSIVELY PUT-LOCK
	  (IF WRITING-INTERNAL-STRUCTURES
	      (PROGN				;allocate space for a directory
		(OR (MULTIPLE-VALUE-setq (DISPLACEMENT LENGTH)	;look first in the directory area
		      (FIND-FREE-SPACE last-allocated-directory-block DIRECTORY-AREA))
		    (MULTIPLE-VALUE-setq (DISPLACEMENT LENGTH)	;look first in the directory area
		      (FIND-FREE-SPACE 0 LAST-ALLOCATED-directory-BLOCK))
		    (MULTIPLE-VALUE-setq (DISPLACEMENT LENGTH)	;then in the rest of the partition
		      (FIND-FREE-SPACE LAST-ALLOCATED-BLOCK  (DC-PARTITION-SIZE)))
		    (MULTIPLE-VALUE-setq (DISPLACEMENT LENGTH)	;then in the rest of the partition
		      (FIND-FREE-SPACE DIRECTORY-AREA LAST-ALLOCATED-BLOCK)))
		(WHEN DISPLACEMENT		;if we found some space, mark it reserved
		  (setq LAST-ALLOCATED-directory-BLOCK (if (< LAST-ALLOCATED-directory-BLOCK
							      DIRECTORY-AREA )
							   displacement 0))
		  (SETQ LENGTH (MIN LENGTH PAGES))
		  (CHANGE-BLOCK-DISK-SPACE DISPLACEMENT
					   LENGTH
					   PUT-FREE
					   NEW-STATUS
					   ())))
	      
	      (LET ((AVAILABLE (- (AREF PUT-USAGE-ARRAY PUT-FREE)	;allocate space for a file
				  PUT-MINIMUM-FREE-PAGES)))
		(WHEN (PLUSP AVAILABLE)		;if we still have file space available
		  (OR
		    (MULTIPLE-VALUE-setq (DISPLACEMENT LENGTH)	;look first in the main partition
		      (FIND-FREE-SPACE (OR LAST-ALLOCATED-BLOCK DIRECTORY-AREA)
				       (DC-PARTITION-SIZE)))
		    (AND LAST-ALLOCATED-BLOCK
			 (MULTIPLE-VALUE-setq (DISPLACEMENT LENGTH)	;CHECK UPPER PORTION OF MAIN PARTITION
			   (FIND-FREE-SPACE DIRECTORY-AREA LAST-ALLOCATED-BLOCK))))
		  (WHEN DISPLACEMENT		;if we found some space, mark it reserved
		    (SETQ LAST-ALLOCATED-BLOCK DISPLACEMENT)
		    (SETQ LENGTH (MIN LENGTH PAGES AVAILABLE))
		    (CHANGE-BLOCK-DISK-SPACE DISPLACEMENT LENGTH PUT-FREE new-status ()))))))
	(IF DISPLACEMENT
	    (RETURN-FROM ALLOCATE-DISK-BLOCK DISPLACEMENT LENGTH)	;return the space we found
	    (LM-SIGNAL-ERROR 'NO-MORE-ROOM () T *CURRENT-FILE-OPERATION*))))))




(DEFSUBST SCAN-PUT (START END TYPE)
  "Scans the PUT looking for the first occurance of space of 
   TYPE. It starts looking at START and finishes at END.  
   Returns the first occurance or END if no occurances found."
  (LET ((FIRST (%STRING-SEARCH-CHAR TYPE PAGE-USAGE-TABLE START END)))
    (IF FIRST
      FIRST
      END)))  



(DEFUN FIND-FREE-SPACE (START END)
  "Locates free space in the PUT. Begins looking at
   START and does not look past END.  Returns the beginning 
   block number and number of blocks free.  Returns NIL if
   there is free space between START and END."
  (LET ((FREE (%STRING-SEARCH-CHAR PUT-FREE PAGE-USAGE-TABLE START END)))
    (WHEN FREE
      (VALUES FREE
	      (- (MIN (SCAN-PUT FREE END PUT-USED)
		      (SCAN-PUT FREE END PUT-RESERVED))
		 FREE)))))  



(DEFUN WRITE-PUT (&OPTIONAL FORCE-P)
  (REQUIRE-LOCK PUT-LOCK)
  (when (OR PUT-MODIFIED FORCE-P)
        (SETQ PUT-MODIFIED ())))  



(DEFMACRO USING-PUT (&BODY BODY)
  (LET ((LOCK-OWNED (GENSYM)))
    `(LET ((,LOCK-OWNED (EQ PUT-LOCK CURRENT-PROCESS))
	   OLD-STATE)
       (UNWIND-PROTECT (PROGN
			(COND
			  ((NOT ,LOCK-OWNED) (PROCESS-LOCK (LOCF PUT-LOCK))
			   (IF (NOT (NULL PUT-MODIFIED))
			     (FERROR () "PUT evidently modified while unlocked"))
			   (SETQ OLD-STATE (AREF PAGE-USAGE-TABLE 0.))
			   (SETF (AREF PAGE-USAGE-TABLE 0.) PUT-INCONSISTENT)))
			,@BODY)
	 (COND
	   ((NOT ,LOCK-OWNED) (SETF (AREF PAGE-USAGE-TABLE 0.) OLD-STATE) (WRITE-PUT)
	    (PROCESS-UNLOCK (LOCF PUT-LOCK))))))))  



(DEFUN CHANGE-BLOCK-DISK-SPACE (LOC NPAGES OLD-STATUS NEW-STATUS &OPTIONAL (MARK-MODIFIED T) (ignore-errors-p nil) &AUX LIM)
  "Change the PUT status of a disk block starting at LOC for NPAGES.  Status is
   changed from OLD-STATUS to NEW-STATUS.  An error is signaled if the pages are
   not in OLD-STATUS.  MARK-MODIFIED indicates whether to mark the PUT as modified."
  (REQUIRE-LOCK PUT-LOCK)
  (SETQ LIM (+ LOC NPAGES))
  (AND MARK-MODIFIED (PLUSP NPAGES)
     (NOT (AND (= OLD-STATUS PUT-RESERVED) (= NEW-STATUS PUT-FREE))) (SETQ PUT-MODIFIED T))
  (DO ((I LOC (1+ I)))
      ((>= I LIM)
       (INCF (AREF PUT-USAGE-ARRAY NEW-STATUS) NPAGES)
       T)
    (if (or (= (AREF PAGE-USAGE-TABLE I) OLD-STATUS)
	    ignore-errors-p)
	(progn
	  (DECF (AREF PUT-USAGE-ARRAY (AREF PAGE-USAGE-TABLE I)))
	  (setf (aref page-usage-table I) new-status))
	(FERROR NIL "Unexpected disk space status in CHANGE-BLOCK-DISK-SPACE. ~d expected, ~d found."
		new-status (AREF PAGE-USAGE-TABLE I)))
    ))  



(DEFUN CHANGE-MAP-DISK-SPACE (MAP OLD-STATUS NEW-STATUS)
  "Change the status of the space in MAP from OLD-STATUS to
   NEW-STATUS.  An error is generated if the space is not
   already OLD-STATUS or NEW-STATUS.  No meaningful return."
  (REQUIRE-LOCK PUT-LOCK)
  (LET ((NBLOCKS (MAP-NBLOCKS MAP)))
    (AND (PLUSP NBLOCKS) (NOT (AND (= OLD-STATUS PUT-RESERVED) (= NEW-STATUS PUT-FREE)))
       (SETQ PUT-MODIFIED T))
    (DOTIMES (MAP-INDEX NBLOCKS T);loop over all map entries
      (LET* ((LOC (MAP-BLOCK-LOCATION MAP MAP-INDEX))
	     (COUNT (QUOTIENT-CEILING (MAP-BLOCK-SIZE MAP MAP-INDEX) PAGE-SIZE-IN-BITS))
	     (LIM (+ LOC COUNT)))
	(DO ((I LOC (1+ I)));loop over all blocks in entry
	    ((>= I LIM))
	  (COND
	    ((= (AREF PAGE-USAGE-TABLE I) OLD-STATUS);if status needs to change,
	     (DECF (AREF PUT-USAGE-ARRAY OLD-STATUS));change it.
	     (INCF (AREF PUT-USAGE-ARRAY NEW-STATUS))
	     (SETF (AREF PAGE-USAGE-TABLE I) NEW-STATUS))
	    ((= (AREF PAGE-USAGE-TABLE I) NEW-STATUS) T);if status already set
	    (T;if block has wrong status, error
	     (FERROR () "Unexpected disk space status in CHANGE-MAP-DISK-SPACE"))))))))  



(DEFUN SET-MAP-DISK-SPACE (MAP NEW-STATUS)
  (REQUIRE-LOCK PUT-LOCK)
  (WHEN (> (MAP-NBLOCKS MAP) 0.)
    (SETQ PUT-MODIFIED T)
    (DOTIMES (MAP-INDEX (MAP-NBLOCKS MAP))
      (LET ((COUNT (QUOTIENT-CEILING (MAP-BLOCK-SIZE MAP MAP-INDEX) PAGE-SIZE-IN-BITS))
	    (LOC (MAP-BLOCK-LOCATION MAP MAP-INDEX)))
	(DOTIMES (I COUNT)
	  (DECF (AREF PUT-USAGE-ARRAY (AREF PAGE-USAGE-TABLE (+ LOC I))))
	  (INCF (AREF PUT-USAGE-ARRAY NEW-STATUS))
	  (SETF (AREF PAGE-USAGE-TABLE (+ LOC I)) NEW-STATUS))))))  



(DEFUN DETERMINE-MAP-DISK-SPACE-STATUS (MAP)
  (REQUIRE-LOCK PUT-LOCK)
  (LET ((STATUS ()))
    (DOTIMES (MAP-INDEX (MAP-NBLOCKS MAP))
      (LET ((COUNT (QUOTIENT-CEILING (MAP-BLOCK-SIZE MAP MAP-INDEX) PAGE-SIZE-IN-BITS))
	    (LOC (MAP-BLOCK-LOCATION MAP MAP-INDEX)))
	(DOTIMES (I COUNT)
	  (COND
	    ((NULL STATUS) (SETQ STATUS (AREF PAGE-USAGE-TABLE (+ LOC I))))
	    ((NOT (= STATUS (AREF PAGE-USAGE-TABLE (+ LOC I))))
	     (FERROR () "Inconsistent status in DETERMINE-MAP-DISK-SPACE-STATUS"))))))
    STATUS))  

;; Binding this variable to non-NIL causes auto creation of directories.
;; This is used internally to create directories, but may be bound by things
;; such as magtape restoration programs.


(DEFVAR LM-AUTOMATICALLY-CREATE-DIRECTORIES ())  

;; This is a variable with all file streams open since boot time.
;; The :OPEN-STREAMS message to a LOCAL-LISPM-HOST returns COPYLIST of this.


(DEFVAR LM-FILE-STREAMS-LIST ())  
