;;; -*- Mode:Common-Lisp; Package:addin; Base:10.; Cold-Load:t; -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1987- 1989 Texas Instruments Incorporated. All rights reserved.
;;;
;;; Edit History
;;;
;;;                   Patch
;;;   Date    Author  Number   Description
;;;--------------------------------------------------------------------
;;; 09-87     bj/ab     --     Original
;;; 12-16-87    ab             o Clean up compiler warnings.
;;; 02-08-88    ab     4.2     o Define GET-COMMAND, somehow deleted accidentally.
;;; 3-16-87     bj             o Add without-interrupts around all of with-acb.
;;; 8-3-88      ab             o Restore my fixes of 4-26-88 to CHANNEL-TIMER-LOOP
;;;                              and PROCESS-WAIT-LOOP.


;;;
;;; Channel Flavor
;;;


(DEFVAR *CHANNEL-LIST* nil "An ordered list of the addin communications channels.")
(DEFVAR *channel-boot-init* nil "A flag inidicating communications channels may be used..")

(defflavor CHANNEL
  (number
   name
   mx-queue
   host-queue
   status-register-array
   (event-handler nil)
   (reset-init-list nil)
   ;; debug info
   (input-complete-timeout)
   (command-complete-timeout)
   (requestor-complete-timeout)
   (send-command-timeout)
   (debug-handler)
   (last-command)
   (error-list)
   )
  (si:property-list-mixin)
  :outside-accessible-instance-variables
  :ordered-instance-variables
  :gettable-instance-variables
  (:settable-instance-variables number name)
  :inittable-instance-variables)


;;;
;;; Timing/Waiting support
;;;


(DEFMACRO timer-loop (seconds &body function-and-args)
  "Time delay for SECONDS."
  `(LOOP with future = (time:time-increment (time:time) (* ,seconds 60))
	 when (FUNCALL ,@function-and-args)
	 do (RETURN t)
	 when (time:time-lessp future (time:time))
	 do (RETURN nil)))

(DEFMACRO channel-timer-loop (channel timeout-name &body function-and-args
			      &aux timeout-fn)
  (SETQ timeout-fn
	(INTERN (STRING-APPEND (STRING-APPEND "CHANNEL-" (STRING-UPCASE timeout-name))) 'add))
  (UNLESS (FUNCTIONP timeout-fn)
    (FERROR nil "~a is not a channel timeout function" timeout-fn))
  (LET ((seconds (GENSYM)))
    `(LET ((,seconds (,timeout-fn ,channel)))
       (COND (,seconds (timer-loop ,seconds ,@function-and-args) t)
	     (t (loop until (FUNCALL ,@function-and-args)
		      finally (RETURN t))))))
  )

;;ab 4/26/88.  Don't use channel-timer-loop if channel timeout is NIL.  
(DEFMACRO process-wait-loop (channel timeout-name whostate
			     &body function-and-args &aux timeout-fn)
  (SETQ timeout-fn
	(INTERN (STRING-APPEND (STRING-APPEND "CHANNEL-" (STRING-UPCASE timeout-name))) 'add))
  (UNLESS (FUNCTIONP timeout-fn)
    (FERROR nil "~a is not a channel timeout function" timeout-fn))
  (LET ((seconds (GENSYM))
	(sixtieths (GENSYM)))
    `(LET* ((,seconds (,timeout-fn ,channel))
	    (,sixtieths (IF ,seconds (* ,seconds 60.))))
       (COND ((EQ si:current-stack-group si:%scheduler-stack-group)
	      (IF ,seconds
		  (PROGN (timer-loop ,seconds ,@function-and-args) t)
		  (loop until (FUNCALL ,@function-and-args)
			finally (RETURN t))))
	     ((PROCESS-WAIT-WITH-TIMEOUT
		,whostate
		,sixtieths
		,@function-and-args) t)
	     (t nil))))
  )


;;;
;;; Queueing
;;;


;(DEFUN queue-acb (acb channel)
;  (q-acb (channel-queue (find-channel channel)) acb));
;
;(DEFUN dequeue-acb (channel)
;  (deq-acb (channel-queue (find-channel channel))))



;;;
;;; Send/Get/Reserved ACB Support
;;;

(DEFMACRO with-acb ((acb-var acb-form) (channel-var channel-form) &body body)
  "This macro acts as a wrapper for general addin communications, performing
the necessary blocking protocol for commands.
   ACB-VAR and CHANNEL-VAR should be names that will be bound to an ACB
/(Addin Communication Block) and a channel respectively within the body of
the code.
   ACB-FORM should either be the name of a variable which already holds an ACB, 
or a form to allocate an ACB.  If the form contains a call to ADD:GET-ACB, the
ACB will be deallocated automatically (via a call to ADD:RETURN-ACB) upon
exit from the body.
   CHANNEL-FORM should be an expression which computes the channel to use
in the communication."
  (UNLESS (AND (SYMBOLP acb-var) (SYMBOLP channel-var))
    (FERROR nil "ACB-VAR and CHANNEL-VAR must be symbols (~s and ~s supplied)"
	    acb-var channel-var))
  (LET ((return-acb (AND (CONSP acb-form)
			 (EQ (FIRST acb-form) 'get-acb))))
    (ONCE-ONLY (acb-form channel-form)
      `(without-interrupts
	 (LET ((,acb-var ,acb-form)
	       (,channel-var ,channel-form))
	   (UNWIND-PROTECT
	       (progn
		 (wait-requestor-complete ,acb-var ,channel-var)
		 (SETF (requestor-complete ,acb-var) nil)
		 (wait-input-complete ,acb-var ,channel-var)
		 ,@body)
	     (SETF (requestor-complete ,acb-var) t)
	     ,(WHEN return-acb
		`(return-acb ,acb-var)))))))
  )

(DEFMACRO initiate-reserved-acb (acb channel parameters-list)
  (LET ((a (GENSYM))
	(c (GENSYM)))
    `(with-acb (,a ,acb) (,c ,channel)
       (load-parms-16b ,a ,@parameters-list)
       (send-command ,a ,c)))
  )

(DEFMACRO send-reserved-acb-wait-values (acb channel parameters-list &rest values-forms)
  (LET ((a (GENSYM))
	(c (GENSYM)))
    `(with-acb (,a ,acb) (,c ,channel)
       (load-parms-16b ,a ,@parameters-list)
       (send-command-and-wait-complete ,a ,c)
       ,@values-forms))
  )

