;;; -*- Mode:Common-Lisp; Mode:COMMON-LISP; Package:NET; Base:10 -*-

;;;                        Restricted Rights Legend

;;;  Use, duplication, or disclosure by the Government is subject to
;;;  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;  Technical Data and Computer Software clause at 52.227-7013.

;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;
;;;  Copyright (C) 1986- 1989 Texas Instruments Incorporated.  All rights reserved.


;; For each network define two functions. 
;;
;;   (:property NETWORK-NAME :network-address-parser) (address-spec &optional error-p) 
;;
;;   (:property NETWORK-NAME :network-address-printer) (address-spec &optional stream error-p)

;;;  12-12-88 DAB Changed :network-address-parser in (:PROPERTY :Dna :NETWORK-ADDRESS-PARSER) to handle float address
;;;               diffently. Always returns address as an integer. [6445]
;;;  *************  REl 5.0 OS 12-12-88 DAB


(defun PARSE-NETWORK-ADDRESS (address-spec network &optional (error-p t))
  "Return the parsed version of address-spec for network. 
   Returns two values:  parsed-address and error-occurred-p.   
If ERROR-P = NIL and an error occurred, then the first value returned is address-spec, and the 
                 second value returned is T.
                 If an error did not occur, then the first value returned is the parsed address,
                 and the second value returned is NIL.
                 If no address parser exists, then the first value returned is address-spec, and
                 the second value returned is :NO-PARSER-EXISTS."

  (unless (keywordp network)
    (setf network (intern-as-keyword network)))
  (let ((address-parser (get network :network-address-parser))
	(parsed-address)
	(error-occurred-p))
    (cond ((null address-parser)
	   (if error-p
	       (ferror nil "No network address parser defined for network ~s." network)
	       ;;else
	       (values address-spec :NO-PARSER-EXISTS)))
	  (error-p
	   (values (funcall address-parser address-spec error-p) nil))
	  (t
	   ;; catch errors
	   (setf (values parsed-address error-occurred-p)
		 (ignore-errors
		   (funcall address-parser address-spec t)))
	   (if error-occurred-p
	       (values address-spec t)
	       ;;else
	       (values parsed-address nil))))))

;(defmacro PARSE-NETWORK-ADDRESS (address-spec network &optional (error-p t))
;  "Return the parsed version of address-spec for network."
;    (cond ((and (keywordp network)
;		;; If the parser is not defined now defer to eval time.
;		(get network :network-address-parser))
;	   `(funcall (get ,network :network-address-parser) ,address-spec ,error-p))
;	  (t
;	   `(let ((address-parser (get (intern-as-keyword ,network) :network-address-parser)))
;	      (cond (address-parser
;		     (funcall address-parser ,address-spec ,error-p))
;		    ((not ,error-p) ,address-spec)
;		    (t
;		     (ferror nil "No network address parser defined for network ~s." ,network)))))))

(defmacro PRINT-NETWORK-ADDRESS (address-spec network &optional stream (error-p t))
  "Print the pretty version of address-spec for network on stream."
  (cond ((and (keywordp network)
	      ;; If the printer is not defined now defer to eval time.
	      (get network :network-address-printer))
	 `(funcall (get ,network :network-address-printer)
		   (parse-network-address ,address-spec ,network ,error-p) ,stream ,error-p))
	(t	   
	 `(let ((address-printer (get (intern-as-keyword ,network) :network-address-printer))
		(address (parse-network-address ,address-spec ,network ,error-p)))
	    (cond (address-printer
		   (funcall address-printer address ,stream ,error-p))
		  ((not ,error-p) (format ,stream "~a" address))
		  (t
		   (ferror nil "No network address printer defined for network ~s." ,network)))))))

;;;
;;; Chaos
;;;

(defun (:PROPERTY :CHAOS :NETWORK-ADDRESS-PARSER) (address &optional (errorp t))
  "Coerce the argument to a chaosnet address. The argument can be a string (read in octal) or an integer"
  (cond
    ((integerp address) address)
    ((and (stringp address) (parse-number address 0 nil 8 t)))
    (errorp (ferror nil "Unknown format ~s for Chaosnet address." address))))

(defun (:PROPERTY :CHAOS :NETWORK-ADDRESS-PRINTER) (address &optional stream (errorp t))
  (condition-call-if (not errorp) ()
      (format stream "~o" address)
    (t (format stream "~s" address))))

;;;
;;; IP
;;;

(defsubst IP-ADDRESS-PARSER (address-spec &optional (error-p t))
  (parse-network-address address-spec :ip error-p))

(defun (:PROPERTY :IP :NETWORK-ADDRESS-PARSER) (address-spec &optional (error-p t) &aux (start 0) (address 0) delimiter)
  "The address corresponding to address-spec is returned if address-spec is either a string or a number 
and represents a valid IP address.  Errors are signaled when error-p is t. When error-p is nil,
two values are returned, the first value is the address.  The second value is T if an error occured otherwise nil."
  (flet
    ((parse-ip-byte-spec (address-spec *print-base* delim-char start)
			 ;; Go thru a list of n/n/n/n where n is in specified base
			 ;; and / is specified delimiter character, accumulating
			 ;; byte values and checking for syntax errors.  Return accumulated value or nil.
			 (let ((slen (length address-spec))
			       (delim-cnt 0)
			       (address 0))
			   (loop do
				 (multiple-value-bind (n stop-ix)
				     (parse-integer address-spec :start start :radix *print-base* :junk-allowed t)
				   (when (null n)
				     (return ()))
				   (setq address (+ (* address 256) n))
				   (when (= stop-ix slen)
				     (return address))
				   (when (neq (aref address-spec stop-ix) delim-char)
				     (return ()))
				   (when (> (incf delim-cnt) 3)
				     (return ()))
				   (setq start (1+ stop-ix)))))))
    (setq address (ignore-errors
		      (cond ((stringp address-spec)	      
			 (cond
			   ;; form = #
			   ((eql (aref address-spec start) #\#)
			    (case (aref address-spec (1+ start))
			      ((#\X #\x) (parse-integer address-spec :start (+ start 2) :radix 16))
			      ((#\O #\o) (parse-integer address-spec :start (+ start 2) :radix 8))
			      ((#\B #\b) (parse-integer address-spec :start (+ start 2) :radix 2))))
			   ;; form = o,o,o,o
			   ((position #\, address-spec) (parse-ip-byte-spec address-spec 8 #\, start))
			   ;; form = d.d.d.d
			   ((position #\. address-spec) (parse-ip-byte-spec address-spec 10 #\. start))
			   ;; form = d:d 
			   ((setq delimiter (position #\: address-spec))
			    (let ((part1
				    (parse-integer address-spec :start start :end delimiter :radix 10
						   :junk-allowed t))
				  (part2 (parse-integer address-spec :start (1+ delimiter) :radix 10)))
			      ;; Note: LSH isn't used below because of sign extension 
			      (cond
				((or (null part1) (null part2)) nil)
				;; Leading bits are 111 => illegal
				((> part1 14680063) nil)
				;; Leading bits are 110, 3 byte : 1 byte
				((> part1 49151) (dpb part1 (byte 24 8) part2))
				;; Leading bits are 10, 2 byte : 2 byte
				((> part1 127) (dpb part1 (byte 16 16) part2))
				;; Leading bit is 0, 1 byte : 3 byte
				(t (dpb part1 (byte 8 24) part2)))))
			   ;; form = number (decimal)
			   (t (parse-integer address-spec :start start :radix 10))))
			((numberp address-spec) address-spec))))
    (if error-p
	(if (or (null address) (< address 16777216) (>= address 3758096384))
	    (ferror nil "Illegal IP address ~A" address-spec)
	    address)
	(values address (or (null address) (< address 16777216) (>= address 3758096384))))))

(defun (:PROPERTY :IP :NETWORK-ADDRESS-PRINTER) (address &optional stream (error-p t))
  (let (b1 b2 b3 b4)

    (cond ((ignore-errors
	     (setf b1 (ldb (byte 8 24) address)
		   b2 (ldb (byte 8 16) address)
		   b3 (ldb (byte 8 8) address)
		   b4 (ldb (byte 8 0) address)))    
	   (format stream "~D.~D.~D.~D" b1 b2 b3 b4))
	  (error-p
	   (ferror nil "Illegal IP address ~A" address))
	  (t
	   (format stream "~a" address)))))
  
;;;
;;; Decnet
;;;
(Defun (:PROPERTY :Dna :NETWORK-ADDRESS-PARSER) (address-spec &optional error-p &aux address)
  "Returns the DNA address corresponding to address-spec.
   An error is signalled if the string is formatted incorrectly."
  
  (setf address (ignore-errors
		  
		  (COND 
		    ((or (floatp address-spec)     
			 (STRINGP address-spec))
		     (when (floatp address-spec)
		       (setf address-spec (format nil "~d" address-spec)))

		     (COND 
		       ;; form = #
		       ((EQL (AREF address-spec 0) #\#)
			(CASE (aref address-spec 1)
			  ((#\X #\x) (PARSE-INTEGER address-spec :start 2 :radix 16))
			  ((#\O #\o) (PARSE-INTEGER address-spec :start 2 :radix 8))
			  ((#\B #\b) (PARSE-INTEGER address-spec :start 2 :radix 2))))
		       
		       ;; form = d.d
		       ((find #\. address-spec)
			(LET ((slen (LENGTH address-spec))
			      (delim-cnt 0) 
			      (address 0)
			      (start 0))
			  (LOOP
			    (MULTIPLE-VALUE-BIND (n stop-ix)
				(PARSE-INTEGER address-spec :start start :radix 10. :Junk-allowed t)
			      (WHEN (NULL n)                              (RETURN nil))
			      (SETF address (+ (* address 1024.) n))
			      (WHEN (= stop-ix slen)                      (RETURN address))
			      (WHEN (NOT (EQL (AREF address-spec stop-ix) #\.)) (RETURN nil))
			      (WHEN (> (INCF delim-cnt) 3)                (RETURN nil))
			      (SETF start (1+ stop-ix))))))
		       
		       ;; form = number (decimal)
		       ((PARSE-INTEGER address-spec :radix 10 :junk-allowed t))))
		    ((NUMBERP address-spec) address-spec)
		    
		    (error-p (FERROR nil "Illegal DECnet address ~A" address-spec)))))
  (if (and error-p (null address))
      (ferror nil "Illegal DNA address ~A" address-spec))
  
  (or address address-spec))


(defun (:PROPERTY :Dna :NETWORK-ADDRESS-PRINTER) (address &optional stream (errorp t))
  (condition-call-if (null errorp) ()
      (multiple-value-bind (area node) (truncate address 1024)
	(format stream "~d.~d" area node))
    (t (format stream "~s" address))))
