;;; -*- Mode:Common-Lisp; Package:USER; Base:10 -*-

;;  DEF-AUTO-LOAD -- for causing functions to be loaded automatically when they are called.
;;
;;	This would typically be used in a LOGIN-INIT file to declare functions that 
;;	you occasionally use even though they aren't included in the load band.
;;	
;;	Some examples:
;;
;;	  (def-auto-load Set-Alarm "SYS:PUBLIC;ALARM-CLOCK")
;;	  (def-auto-load Hex-Dump  "TOOLS:TOOLS;HEXDUMP")
;;	  (def-auto-load Explore (make-system 'explore :noconfirm))
;;
;;  11/13/87 D.N.G.


(defmacro def-auto-load ( fspec &rest forms-and-pathnames )
  "Provide a temporary definition for a function that will cause the real
definition to be loaded the first time it is called.
The first argument is the name of the function, and the remaining arguments 
are one or more pathnames to be loaded or forms to be executed to define the 
function.  Note that this only works for functions, not macros, special forms, 
or variables."
  (let ((decl (and (fdefinedp fspec)
		   `(declare (arglist . ,(arglist fspec)))))
	(doc (and (fdefinedp fspec)
		  (documentation fspec)))
	(body-forms (loop for x in forms-and-pathnames
			  collect (if (consp x)
				      x
				    `(load ,x))))
	(quoted-name `(quote ,fspec)))
    (if (symbolp fspec)
	`(unless (fboundp ,QUOTED-NAME)
	   (fset ,quoted-name
		 #'(named-lambda ,FSPEC (&rest args)
				 ,decl
				 ,doc
				 (let ((sys:default-cons-area sys:background-cons-area))
				   . ,body-forms)
				 (apply #',FSPEC args) ))
	   (when (fboundp 'compiler:compilation-define) ; unless in cold band
	     (compiler:compilation-define ,quoted-name))
	   ,quoted-name)
      `(unless (fdefinedp ,QUOTED-NAME)
	 (fdefine ,quoted-name
		  #'(named-lambda ,FSPEC (&rest args)
				  ,decl
				  ,doc
				  (let ((sys:default-cons-area sys:background-cons-area))
				    (si:function-spec-putprop ,quoted-name nil :source-file-name)
				    . ,body-forms)
				  (apply #',FSPEC args) )
		  T) ))))
