;;;  -*- Mode:Common-Lisp; Package:USER; Base:10 -*-

(DEFUN BBOARD ( &OPTIONAL BULLETIN-BOARD-NAME SINCE (HOST "Austin"))
  "Display messages from a bulletin-board on a Unix system.
The first argument is a string that is the name of the bulletin board (also 
known as a \"newsgroup\"); if nil, all the valid names are displayed.  The 
second argument is a string specifying a date; messages posted prior to that 
date are skipped.  The date defaults to the time of the last message you read 
from the selected bulletin board, so that it begins with the last message or 
two you saw in the previous session."

  ;;  3/31/87 DNG - Fixed to sort files by creation date rather than name.
  ;;  3/30/88 DNG - Change host from Tilde to Austin.
  ;;  3/31/88 DNG - Permit access to newsgroups in directories other than 
  ;;		"mail".  Remember the time of the last message read.  
  ;;		Adjust line count for small windows.
  (IF (NULL BULLETIN-BOARD-NAME)
      (PROGN (FORMAT T "~%The valid bulletin-board names are ...")
	     (map-sub-directories
	       #'(lambda (path)
		   (let ((nodes (rest (member "news" (send path :directory)
					      :test #'string-equal))))
		     (when (> (length nodes) 1) ; don't show "junk" and "control".
		       (format t "~&  \"")
		       (do ((tail nodes (rest tail)))
			   ((null tail))
			 (write-string (string-downcase (first tail)))
			 (unless (null (rest tail))
			   (write-char #\.)))
		       (write-char #\"))))
	       (STRING-APPEND HOST ":/usr/spool/news/*")
	       2)
	     (values))
    (LET* ((NAME (string-downcase (string BULLETIN-BOARD-NAME)))
	   (DIR-NAME (COND ((FIND #\. NAME) (NSUBSTITUTE #\/ #\. NAME))
			   ((FIND #\/ NAME) NAME)
			   ;; Else default to "mail" directory for compatibility
			   ;; with older versions of this program.
			   (T (STRING-APPEND "mail/" NAME))))
	   (DIR-LIST (FS:DIRECTORY-LIST
			;; Ought to use MAKE-PATHNAME instead of STRING-APPEND,
			;; but can't because of SPR 7682.
			(STRING-APPEND HOST ":/usr/spool/news/" DIR-NAME "/*")
			:SORTED NIL))
	   (LOG-FILE (SEND (FS:USER-HOMEDIR) :NEW-PATHNAME
			   :NAME "BBOARD-HISTORY" :CANONICAL-TYPE :LISP :VERSION :NEWEST))
	   (LAST-READ (WITH-OPEN-FILE (STREAM LOG-FILE :IF-DOES-NOT-EXIST NIL)
			(AND STREAM
			     (LET ((*PACKAGE* *USER-PACKAGE*)
				   (*READ-BASE* 10))
			       (SYS:WITH-COMMON-LISP-ON
				 (READ STREAM ))))))
	   (NEWEST-READ (OR (SECOND (ASSOC DIR-NAME LAST-READ :TEST #'EQUAL))
			    0))
	   (SINCE (IF SINCE
		      (TIME:PARSE-UNIVERSAL-TIME SINCE 0 NIL NIL)
		    NEWEST-READ))
	   (NUMLINES (MIN (- (IF (send *STANDARD-OUTPUT* :operation-handled-p :line-height)
				 ;; compute number of lines in window
				 (truncate (send *STANDARD-OUTPUT* :inside-height)
					   (send *STANDARD-OUTPUT* :line-height))
			       25) 2)
			  40))
	   )
      (DECLARE (STRING NAME DIR-NAME) (INTEGER SINCE NEWEST-READ NUMLINES)
	       (LIST DIR-LIST))
   (UNWIND-PROTECT
     (BLOCK READ-BBOARD 
      (SETQ DIR-LIST (STABLE-SORT DIR-LIST #'<
				  :KEY #'(LAMBDA (X) (GETF (CDR X) :CREATION-DATE 0))))
      (FLET ((DISPLAY-FILE (NAME)
	      (CATCH-ERROR-RESTART (ERROR "Give up reading file \"~A\"" NAME)
		(WITH-OPEN-FILE (FILE-STREAM NAME)
		  (FUNCALL *STANDARD-OUTPUT* :FRESH-LINE)
		  (BLOCK VIEW
		    (LOOP
		      (DOTIMES (I NUMLINES)
			(LET ((LINE (READ-LINE FILE-STREAM NIL :EOF)))
			  (IF (EQ LINE :EOF)
			      (RETURN-FROM VIEW NIL)
			    (WRITE-LINE LINE *STANDARD-OUTPUT*) )))
		      (LET ((CHOICE (FQUERY `(:TYPE :TYI :CHOICES 
					  (((:MORE "show more of the current message") 
					           #\space #\return #\Y #\y)
					   ((:EDIT "Edit the current message") #\E #\e)
					   ((:NEXT "skip to the Next message") #\N #\n)
					   ((:BACK "go Back to the previous message") #\B #\b #\P #\p)
					   ((:QUIT "Quit") #\Q #\q)
					   ))
				  "-- more --")))
			(IF (EQ CHOICE ':MORE)
			    (TERPRI)
			  (RETURN-FROM VIEW CHOICE))
			 ))))) ))
	(COND ((NULL DIR-LIST)
	       (FORMAT T "~%Directory not found for ~A." NAME))
	      ((NULL (CDR DIR-LIST))
	       (FORMAT T "~%No messages for ~A." NAME))
	      (T (LET ((FORWARD-LIST DIR-LIST)
		       (BACKWARD-LIST NIL)
		       (ACTION :NEXT)
		       (INFO NIL))
		   (LOOP
		     (BLOCK DO-ONE
		       (CASE (OR ACTION
				 (FQUERY
				   `(:TYPE :TYI :CHOICES 
					   #!Z
					   (((:EDIT "Edit the current message") #\E #\e)
					    ((:NEXT "Show the next messsage")
					     #\N #\n #\space #\return #\Y #\y)
					    ((:BACK "Go back to the previous message")
					     #\B #\b #\P #\p)
					    ((:QUIT "Quit") #\Q #\q)
					    ))
				   "bboard>"))
			 (:NEXT (UNLESS (NULL (CAR INFO))
				  (PUSH INFO BACKWARD-LIST))
				(SETQ INFO (POP FORWARD-LIST))
				(SETQ ACTION :NEXT))
			 (:EDIT (ED (CAR INFO))
				(SETQ ACTION NIL)
				(RETURN-FROM DO-ONE))
			 (:QUIT (RETURN-FROM READ-BBOARD (VALUES)))
			 (:BACK (PUSH INFO FORWARD-LIST)
				(SETQ INFO (POP BACKWARD-LIST))
				(SETQ ACTION :BACK))
			 )
		       (IF (NULL INFO)
			   (PROGN (FORMAT T "~&No more messages.")
				  (SETQ ACTION NIL))
			 (LET ((NAME (CAR INFO))
			       (PLIST (CDR INFO)))
			   (DECLARE (LIST PLIST))
			   (WHEN (AND (NOT (NULL NAME))
				      (GETF PLIST :CHARACTERS T)
				      (NOT (GETF PLIST :DIRECTORY NIL))
				      (>= (GETF PLIST :CREATION-DATE SINCE)
					  SINCE))
			     (SETQ ACTION (DISPLAY-FILE NAME))
			     (LET ((TIME (GETF PLIST :CREATION-DATE 0)))
			       (WHEN (> TIME NEWEST-READ)
				 (SETQ NEWEST-READ TIME)))
			     )))
		       )		   ; end of block do-one
		     )			   ; end of loop
		   ))))) ; end block READ-BBOARD
	(WHEN (> NEWEST-READ 0)
	  (BLOCK UPDATE
	    (LET ((X (ASSOC DIR-NAME LAST-READ :TEST #'EQUAL)))
	      (IF X
		  (IF (> NEWEST-READ (SECOND X))
		      (SETF (SECOND X) NEWEST-READ)
		    (RETURN-FROM UPDATE))
		(PUSH (LIST DIR-NAME NEWEST-READ) LAST-READ)))
	    (catch-error-restart ((error break) "Give up trying to write history file.")
	      (WITH-OPEN-FILE (STREAM LOG-FILE :DIRECTION :OUTPUT)
		(LET ((*PACKAGE* *USER-PACKAGE*)
		      (*PRINT-BASE* 10))
		  (SYS:WITH-COMMON-LISP-ON
		    (WRITE-STRING ";; A-list of newsgroups and universal time of last message read by BBOARD"
				  STREAM)
		    (TERPRI STREAM)
		    (PPRINT LAST-READ STREAM)))) ))	; end block UPDATE
	  )) ; end UNWIND-PROTECT
  (VALUES))))

(defun map-sub-directories (function path max-depth)
  (let ((dir-path (pathname path)))
    (setq dir-path (send dir-path :new-pathname
			 :name (or (send dir-path :name) :wild)
			 :type :wild
			 :version :newest))
    (labels ((collect-directories (path depth)
	       (catch-error-restart ((error break) "Give up looking at directory \"~A\"." path)
		 (let ((files (fs:directory-list path))
		       (messages nil))
		   (dolist (info files)
		     (let ((file (car info))
			   (plist (cdr info)))
		       (cond ((null file))
			     ((getf plist :directory nil)
			      (let ((sub-dir (send file :new-pathname
						   :directory (append (send file :directory)
								      (list (send file :name)))
						   :name :wild
						   :type :wild
						   :version :newest)))
				(if (>= depth max-depth)
				    (funcall function sub-dir)	; assume terminal instead of reading
				  (collect-directories sub-dir (+ depth 1)))))
			     (t (setq messages t)))))
		   (when messages
		     (funcall function path)) ))
	       (values)))
      (collect-directories dir-path 0))
    (values)))
