;;; -*- Mode:Common-Lisp; Package:ZLC; Base:10 -*-

;; These two forms are temporary and will not go into the released
;; system.  They are only needed to process the update arglist data.


si:(defun set-arglist (function-spec arglist &optional (values-list nil supplied-p))
  "This function changes the :Descriptive-Arglist property and the
   :Values property of FUNCTION-SPEC to ARGLIST and VALUES-LIST."
  (let ((info-struct (get-debug-info-struct function-spec)))
    ;;set the arglist property
    (setf (get-debug-info-field info-struct :descriptive-arglist) arglist)
    ;;setf the return-list property is supplied	
    (if supplied-p
	(setf (get-debug-info-field  info-struct :values) values-list)))
  ;;make sure the right number of values is returned incase this was
  ;;called from SETF
  (if supplied-p
      (values arglist values-list)
      arglist)
  )


si:(defsetf arglist (function-spec) (new-value values-list)
  "This SETF form modifies the arglist and return-values declarations
   for FUNCTION-SPEC.  If NEW-VALUE is a single value, the only the
   arglist is modified.  If NEW-VALUE generates multiples then the
   first value is the new arglist and the second value modifies the
   return value."
  `(multiple-value-call 'set-arglist ,function-spec ,new-value))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Here are the documentation changes for ZLC

(setf (arglist 'DEFCONST) '(VARIABLE INITIAL-VALUE . ARGS))
(setf (documentation 'DEFCONST 'function) 
  "Declares a special variable named VARIABLE whose value is INITIAL-VALUE.  ARGS may take on the
   value of the documentation string.  The use of DEFCONST does not license the compiler to build
   assumptions about the value of VARIABLE into the program being compiled.") 

(setf (arglist 'DELQ) (VALUES '(ITEM LIST &OPTIONAL (TIMES -1)) '(ALTERED-LIST)))
(setf (documentation 'DELQ 'function)
  "Returns a list created by deleting all elements in LIST that are EQ to ITEM.  LIST is destructively
   modified by DELQ if any items are deleted.  If TIMES is specified, it is an integer stating how
   many occurrences of ITEM should be deleted from LIST.")

;(setf (arglist ERR) (&OPTIONAL VALUE-FORM FLAG))
;(setf (documentation 'ERR 'function) 
;  NIL)

(setf (arglist 'PIXEL-ARRAY-WIDTH) (VALUES '(ARRAY) '(LENGTH)))
(setf (documentation 'PIXEL-ARRAY-WIDTH 'function)
  "Returns the LENGTH of the dimension that varies the fastest in memory of the two dimension array
   ARRAY.  For a screen array, this dimension is the width.") 

(setf (arglist 'DECLARE-FLAVOR-INSTANCE-VARIABLES) '((FLAVOR-NAME MAP-SET-BY-CALLER) &BODY BODY))
(setf (documentation 'DECLARE-FLAVOR-INSTANCE-VARIABLES 'function)
  "Macro used to allow a function which is itself not a flavor method to use the flavor object SELF. 
   This assumes that SELF's flavor will include FLAVOR-NAME as a component flavor; an error is 
   signaled if it does not.  BODY may be either one or more expressions in a function or the entire
   function definition.")

(setf (arglist 'PLIST) (values '(SYMBOL) '(PROPERTY-LIST-CONTENTS)))
(setf (documentation 'PLIST 'function)
  "Returns the contents of the property list of SYMBOL.  SYMBOL may be a symbol, an instance
   supporting the :PROPERTY-LIST operation, or a locative or cons cell whose CDR is the
   property list.")

(setf (arglist 'REST4) (VALUES '(LIST) '(MODIFIED-LIST)))
(setf (documentation 'REST4 'function)
  "Returns a list created by applying the function CDR 4 times to LIST, analogous to removing the
   first 4 elements of LIST.") 

(setf (arglist 'STRING-LENGTH) (VALUES '(STRING) '(LENGTH)))
(setf (documentation 'STRING-LENGTH 'function) 
  "Returns as a non-negative integer the length of STRING in characters if STRING is a string, the
   active length of STRING if it is an array, or 1 if STRING is a number or character object.")

(setf (arglist 'STRING-REVERSE) (VALUES '(STRING) '(MODIFIED-STRING)))
(setf (documentation 'STRING-REVERSE 'function) 
  "Returns a string whose elements are those of STRING but in reverse order.  STRING-REVERSE will
   work on any one-dimensional array.")

(setf (arglist 'AS-2-REVERSE) (VALUES '(VALUE ARRAY INDEX2 INDEX1) '(MODIFIED-ARRAY)))
(setf (documentation 'AS-2-REVERSE 'function) 
  "Replaces the element in array ARRAY with VALUE at the given index positions.  ARRAY is modified
   by AS-2-REVERSE.  INDEX2 is the value for the fastest varying index in memory; therefore, even
   though it is the first index argument given, the index arguments may be reversed upon use
   depending upon how the array is stored in memory.")

(setf (arglist 'ADJUST-ARRAY-SIZE) (VALUES '(ARRAY NEW-INDEX-LENGTH) '(REDIMENSIONED-ARRAY)))
(setf (documentation 'ADJUST-ARRAY-SIZE 'function)
  "Alters the dimensions of array ARRAY.  If ARRAY is one-dimensional, its size is changed to
   NEW-INDEX-LENGTH; if multi-dimensional, its total size is changed to NEW-INDEX-LENGTH by
   altering only the last subscript.  If ARRAY is increased in size, it may become displaced;
   ADJUST-ARRAY-SIZE returns this new array.  In all other cases ADJUST-ARRAY-SIZE returns ARRAY.")    

(setf (arglist 'ENTITYP) (VALUES '(OBJ) '()))
(setf (documentation 'ENTITYP 'function) 
  "Returns NIL, for entities no longer exist.  Entities have been made obsolete by flavors.")

;(setf (arglist 'SEND-IF-HANDLES) '(TO-OBJECT MESSAGE &REST OTHER-ARGUMENTS))
;(setf (documentation 'SEND-IF-HANDLES 'function) 
;  "Send a SEND-IF-HANDLES message to to-object")

(setf (arglist 'LOAD-FILE-LIST) '(File-LIST &OPTIONAL KEYLIST))
(setf (documentation 'LOAD-FILE-LIST 'function) 
  "Loads all the files in FILE-LIST that are either not loaded or have changed.  Each element of
   FILE-LIST is a list containing a filename string.  KEYLIST is a list of keywords that are
   acceptable to the MAKE-SYSTEM function.")

(setf (arglist ') (VALUES '(NUMBER &REST NUMBERS) '(T-OR-NIL)))
(setf (documentation ' 'function)
  "Returns true if the arguments are in nonincreasing numerical order, otherwise NIL.")

(setf (arglist 'CHAR) (VALUES '(CHAR &REST CHARS) '(T-OR-NIL)))
(setf (documentation 'CHAR 'function)
  "Returns true if the arguments are nonincreasing, considering bits, fonts, and case; otherwise,
   NIL is returned.")

(setf (arglist 'REST1) (VALUES '(LIST) '(MODIFIED-LIST)))
(setf (documentation 'REST1 'function)
  "Returns a list created by applying the function CDR once to LIST, analogous to removing the
   first element of LIST.") 

(setf (arglist 'AR-2-REVERSE) (VALUES '(ARRAY INDEX2 INDEX1) '(ARRAY-ELEMENT)))
(setf (documentation 'AR-2-REVERSE 'function) 
  "Returns the element of ARRAY at the given index positions.  INDEX2 is the value for the fastest
   varying index in memory; therefore, even though it is the first index argument given, the index
   arguments may be reversed upon use depending upon how the array is stored in memory.")

(setf (arglist 'SELECTQ) '(TEST-OBJECT &BODY CLAUSES))
(setf (documentation 'SELECTQ 'function)
  "Control structure designed to execute only one case from a set of cases.  Syntax of 
   (SELECTQ TEST-OBJECT (test1 body1) ... (testN bodyN)) is defined as :  the (test# body#)
   construct is called a clause.  TEST-OBJECT is evaluated.  <test1> is evaluated and compared
   against TEST-OBJECT; if they are EQL, all forms in <body1> are evaluated sequentially and the
   value of the last expression in <body1> is returned by SELECTQ.  If <test1> does not match,
   <test2> is evaluated and compared, and so on, until either one <test#> matches TEST-OBJECT or
   all clauses are exhausted.  SELECTQ returns NIL if no clause is selected.  An OTHERWISE or T as
   the <testN> matches any TEST-OBJECT.")

(setf (arglist 'COPYLIST) (VALUES '(LIST &OPTIONAL FORCE-DOTTED) '(NEW-LIST)))
(setf (documentation 'COPYLIST 'function)
  "Returns a list that is EQUAL to but not EQ to LIST.  Dotted pair termination of the list will
   also be copied.") 

(setf (arglist ') (VALUES '(&REST NUMBERS) '(T-OR-NIL)))
(setf (documentation ' 'function)
  "Returns true if all arguments are not equal in magnitude, otherwise NIL.")

(setf (arglist 'LEXPR-FUNCALL-SELF) '(&REST ARGS))
(setf (documentation 'LEXPR-FUNCALL-SELF 'function)
  "Obsolete version of LEXPR-FUNCALL of SELF; replaced by LEXPR-SEND using SELF as first argument.")

(setf (arglist 'RETURN-LIST) (VALUES '(VALUES) '(LIST-OF-VALUES)))
(setf (documentation 'RETURN-LIST 'function)
  "Exits a block named NIL and returns each element in VALUES as a separate value.  Forms such as
   DO, PROG, and LOOP create blocks named NIL.")

(setf (arglist 'NCONS-IN-AREA) '(CAR AREA))
(setf (documentation 'NCONS-IN-AREA 'function) 
  "Returns a cons cell whose CAR is CAR and CDR is NIL.  This cons cell is constructed in area AREA.")

(setf (arglist '\\) (VALUES '(&REST NUMBERS) '(GREATEST-COMMON-DIVISOR)))
(setf (documentation '\\ 'function)
  "Returns the greatest common divisor of NUMBERS, or either the absolute value of NUMBERS if only one 
   argument is given or 0 if no arguments are given.")

(setf (arglist 'LOAD-BYTE) (VALUES '(FROM-VALUE POSITION WIDTH) '(INTEGER)))
(setf (documentation 'LOAD-BYTE 'function)
  "Returns as a non negative integer the byte of size WIDTH at position POSITION from FROM-VALUE.
   POSITION is zero-origin starting at the least significant bit.")

(setf (arglist 'STRING) (VALUES '(STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2) '(INDEX)))
(setf (documentation 'STRING 'function)
  "Returns the INDEX into STRING1 of the first point of difference between it and STRING2 that
   satisfies the test less than or equal to.  Both case and font are considered.  If the test
   is never satisfied, NIL is returned.  If the strings match, the length of the shorter string
   is returned.  Although the default string length of both STRING1 and STRING2 is the entire
   length, a substring of STRING1 starting at :START1 and ending at :END1 and/or and substring
   of STRING2 starting at :START2 and ending at :END2 may be specified.")

(setf (arglist 'STRING-REVERSE-SEARCH-CHAR) (VALUES '(CHAR STRING &OPTIONAL FROM (TO 0) CONSIDER-CASE) '(INDEX)))
(setf (documentation 'STRING-REVERSE-SEARCH-CHAR 'function) 
  "Returns the INDEX into STRING of the last occurrence of CHAR.  Even though STRING is searched in
   reverse order, the index is relative to the beginning of STRING.  Since the search is in reverse
   order, the last occurrence is found first.  If CHAR is not found, STRING-REVERSE-SEARCH-CHAR
   returns NIL.  Although the entire length of STRING is the default, a substring of STRING starting
   at one less than FROM and ending at TO may be specified.  TO should normally be less than FROM. 
   Both case and font are considered if CONSIDER-CASE is true.")

(setf (arglist 'FIND-POSITION-IN-LIST-EQUAL) (VALUES '(ITEM IN-LIST) '(INDEX)))
(setf (documentation 'FIND-POSITION-IN-LIST-EQUAL 'function) 
  "Returns the INDEX into list IN-LIST of the first element that is EQUAL to ITEM.  Zero-origin 
   indexing is used.  If not match is found, FIND-POSITION-IN-LIST-EQUAL returns NIL.")

(setf (arglist 'SETPLIST) '(SYMBOL L))
(setf (documentation 'SETPLIST 'function) 
  "Sets the property list of SYMBOL to L (a list of alternating properties and values).  SYMBOL may
   be an instance that handles the :SETPLIST operation, instead of a symbol.  This function should
   be used with caution.")

(setf (documentation 'PRINLEVEL 'variable)
  "Variable whose value, when non-NIL, is an integer specifying how many levels deep a nested
   object will be printed.  Any structure at a level lower that PRINLEVEL is replaced by #. 
   A NIL value for PRINLEVEL specifies no control.")

(setf (arglist 'DEL-IF-NOT) (VALUES '(PRED LIST) '(MODIFIED-LIST)))
(setf (documentation 'DEL-IF-NOT 'function)
  "Returns a list where those elements in list LIST that do not satisfy test PRED have been deleted.
   LIST is destructively modified by this operation.")

(setf (arglist 'DO-NAMED) '(NAME . BODY))
(setf (documentation 'DO-NAMED 'function)
  "Macro used to provide a generalized iteration facility.  Syntax of   
   (DO-NAMED NAME ((var1 init1 step1) ... (varN initN stepN)) (end-test return-result) [(declare...)] body)
   is defined as :  in addition to the implicit DO block named NIL, DO-NAMED defines a  block named
   NAME.  NAME is not evaluated, and allows the DO-NAMED to be exited with a RETURN-FROM statement.
   Before the first iteration, all init forms <init1> thru <initN> are evaluated and these results
   saved.  The variables <var1> thru <varN> are then bound to the value of <init1> thru <initN>,
   respectively.  After all variables are bound, <end-test> is evaluated.  If <end-test> returns
   true, the forms in <return-result> are evaluated as a PROGN and DO-NAMED returns the result of
   the last form evaluated.  If <end-test> returns NIL, the forms in <body> are evaluated as in a
   PROGN.  After the last form in <body> is evaluated, the step forms <step1> thru <stepN> are
   evaluated and their return values saved.  These values are assigned in parallel to <var1> thru
   <varN>, respectively.  <end-test> is again evaluated.  This iteration continues until <end-test>
   returns true.  NIL is the default for any form not specified.")

(setf (arglist 'GET-PNAME) (VALUES '(SYMBOL) '(PRINT-NAME)))
(setf (documentation 'GET-PNAME 'function)
  "Returns the print name of symbol SYMBOL.")

(setf (arglist 'NCONS) (VALUES '(CAR) '(CONS-CELL)))
(setf (documentation 'NCONS 'function) 
  "Returns a cons cell whose CAR is CAR and CDR is NIL.")

(setf (arglist 'QC-FILE-LOAD) '(INPUT-FILE &OPTIONAL OUTPUT-FILE))
(setf (documentation 'QC-FILE-LOAD 'function) 
  "Compiles the Lisp file INPUT-FILE and loads the resulting binary file.  Although the name of the
   binary file will default to INPUT-FILE but with an type component of XLD, the binary file can be
   written elsewhere by specifying its pathname with the OUTPUT-FILE argument.")

(setf (documentation 'ARRAY-INDEX-ORDER 'variable)
  "Constant whose value is true if arrays are stored in row-major order or NIL if arrays are stored
   in column-major order.")

(setf (arglist 'ASSQ) (VALUES '(X ALIST) '(MODIFIED-LIST)))
(setf (documentation 'ASSQ 'function)
  "Returns the first cons in association list ALIST whose CAR is EQ to X.  If no match is found,
   ASSQ returns NIL.") 

(setf (arglist 'ATAN2) (VALUES '(Y &OPTIONAL (X (FLOAT 1 Y))) '(RADIANS)))
(setf (documentation 'ATAN2 'function)
  "Returns the angle in radians whose tangent is Y.  If X is specified, the tangent used is Y/X.
   The sign of Y and X determine which angle, differing by pi and with the same tangent, is
   returned.  Y may be complex if X is not specified; X may be 0 if Y is not 0.") 

(setf (arglist 'LEXPR-FUNCALL) '(FCT &REST ARGS))
(setf (documentation 'LEXPR-FUNCALL 'function)
  "Applies the function FCT to the argument list ARGS which has been appended into one list.  The  
   last element of ARGS must be a list.  The return value of LEXPR-FUNCALL is the result obtained
   by applying the function to the argument list.  When ARGS is not specified, the CAR of FCT is
   applied to the CDR of FCT.")

(setf (arglist 'DEPOSIT-BYTE) (VALUES '(INTO-VALUE POSITION WIDTH BYTE-VALUE) '(INTEGER)))
(setf (documentation 'DEPOSIT-BYTE 'function)
  "Returns the integer composed by substituting the WIDTH number of bits of BYTE-VALUE into INTO-VALUE
   at position POSITION.  POSITION is zero-origin starting from the least significant bit.")

(setf (arglist 'DIFFERENCE) (VALUES '(NUMBER &REST NUMBERS) '(NEW-NUMBER)))
(setf (documentation 'DIFFERENCE 'function)
  "Returns the value of the first argument minus the remaining arguments, or NUMBER if only one argument
   is given.")

(setf (arglist 'LOAD-FILE-ALIST) '(ALIST &OPTIONAL (DONT-ASK-INDIVIDUALLY 0) (DONT-CARE-IF-LOADED-P 0)
                                 DONT-ASK-FOR-CONFIRMATION PKG))
(setf (documentation 'LOAD-FILE-ALIST 'function) 
  "Loads all the files in ALIST that are either not loaded or have changed. Each element of ALIST
   is a list whose CAR is a filename string.  They are loaded into package PKG.  The remaining
   arguments can be defined as :
      DONT-ASK-INDIVIDUALLY - Controls whether the user is queried about each file.
      DONT-CARE-IF-LOADED-P - Controls whether unchanged files are reloaded.
      DONT-ASK-FOR-CONFIRMATION - Controls whether the user is queried for one final confirmation.")

(setf (arglist 'MULTIPLE-VALUE) (VALUES '(&QUOTE VAR-LIST EXP) '(MODIFIED-VAR-LIST)))
(setf (documentation 'MULTIPLE-VALUE 'function)
  "Evaluates EXP, and assigns the variables in VAR-LIST to them.  If EXP returns more values than
   there are variables, the extra values are ignored; if there are more variables than values, they
   are assigned the value NIL.  If NIL is used as a variable name, its corresponding value is ignored.
   MULTIPLE-VALUE returns the first value returned by EXP.")

(setf (arglist 'COPYTREE) (VALUES '(TREE &OPTIONAL (DEFAULT-CONS-AREA DEFAULT-CONS-AREA)) '(NEW-TREE)))
(setf (documentation 'COPYTREE 'function)
  "Returns a tree which is a copy of TREE.  All cons cells in the tree are copied; objects that
   are not cons cells are EQ to those in TREE.")

;(setf (arglist '*EXPR) (&QUOTE &REST L))
;(setf (documentation '*EXPR 'function)   
;  NIL)

(setf (arglist 'CHAR) (VALUES '(CHAR &REST CHARS) '(T-OR-NIL)))
(setf (documentation 'CHAR 'function)
  "Returns true if the arguments are nondecreasing, considering bits, fonts, and case;
   otherwise, NIL is returned.")

(setf (arglist 'REST2) (VALUES '(LIST) '(MODIFIED-LIST)))
(setf (documentation 'REST2 'function)
  "Returns a list created by applying the function CDR 2 times to LIST, analogous to removing the
   first 2 elements of LIST.")

(setf (arglist 'COPYSYMBOL) (VALUES '(SYMBOL &OPTIONAL COPYPROPS) '(NEW-SYMBOL)))
(setf (documentation 'COPYSYMBOL 'function)
  "Returns a new uninterned symbol with the same print name as SYMBOL. If COPYPROPS is true,
   the value, function definition, and properties of SYMBOL are also copied into the new symbol.  
   If COPYPROPS is NIL, SYMBOL is unbound and undefined.")

(setf (arglist 'STRING-REVERSE-SEARCH) (VALUES
					 '(KEY STRING &OPTIONAL FROM (TO 0) (KEY-FROM 0) KEY-TO CONSIDER-CASE)
					 '(INDEX)))
(setf (documentation 'STRING-REVERSE-SEARCH 'function)
  "Returns the index into STRING of the last occurrence of string KEY.  Even though STRING is
   searched in reverse order, the index is relative to the beginning of STRING.  Since the search is
   in reverse order, the last occurrence is found first.  If KEY is not found, STRING-REVERSE-SEARCH
   returns NIL.  Although the entire length of both STRING and KEY is the default, a substring of
   STRING starting at one less than FROM and ending at TO and/or a substring of KEY starting at
   KEY-FROM and ending at KEY-TO may be specified.  TO should normally be less than FROM.   Both
   case and font are considered if CONSIDER-CASE is true.")

(setf (arglist '^) (VALUES '(BASE EXPONENT) '(VALUE)))
(setf (documentation '^ 'function)
  "Returns the VALUE of BASE raised to the EXPONENT power.") 

(setf (arglist 'FILLARRAY) (VALUES '(ARRAY SOURCE) '(MODIFIED-ARRAY)))
(setf (documentation 'FILLARRAY 'function)
  "Returns the array created by destructively modifying ARRAY by replacing each element with
   SOURCE.  If SOURCE is a list, its last element is repeated to fill any part of ARRAY not 
   specifically replaced.  If SOURCE is an array, elements of ARRAY not replaced by SOURCE are left
   unmodified.  If SOURCE is NIL, the array is filled with the default type for the array.  If ARRAY
   is NIL, a new array the same length as SOURCE is returned.")

(setf (arglist 'ASET) (VALUES '(VALUE ARRAY &REST SUBSCRIPTS) '(VALUE MODIFIED-ARRAY)))
(setf (documentation 'ASET 'function) 
  "Replaces the element in array ARRAY specified by subscripts SUBSCRIPTS with VALUE.  ARRAY is
   destructively modified by ASET.  ASET returns VALUE.")

(setf (arglist 'DEL-IF) (VALUES '(PRED LIST) '(MODIFIED-LIST)))
(setf (documentation 'DEL-IF 'function)
  "Returns a list where those elements in list LIST that satisfy test PRED have been deleted.  LIST
   is destructively modified by this operation.")

(setf (documentation 'ALLOW-VARIABLES-IN-FUNCTION-POSITION-SWITCH 'variable) 
  "Variable whose value, when true, allows the use of a variable name in a function position, with
   FUNCALL to be applied to the value of this variable.  This is for compatibility with old MacLisp
   programs.")

(setf (arglist 'STRING-SEARCH-NOT-CHAR) (VALUES '(CHAR STRING &OPTIONAL (FROM 0) TO CONSIDER-CASE) '(INDEX)))
(setf (documentation 'STRING-SEARCH-NOT-CHAR 'function) 
  "Returns the INDEX into STRING of the first character that is not equal to CHAR.  If all characters
   in STRING are equal to CHAR, STRING-SEARCH-NOT-CHAR returns NIL.  Although the entire length of
   STRING is the default, a substring starting at FROM and ending at TO may be specified.  Both case
   and font are considered if CONSIDER-CASE is true.")

(setf (arglist 'CHECK-ARG-TYPE) (VALUES '(ARG-NAME TYPE &OPTIONAL TYPE-STRING) '(ERROR-OR-NIL)))
(setf (documentation 'CHECK-ARG-TYPE 'function)
  "Generates an error if ARG-NAME is not of type TYPE.  ARG-NAME, the contents of ARG-NAME, and 
   TYPE are mentioned in the error message.  In addition, if TYPE-STRING is specified, it is a 
   string that is used in generating the error message.  This error is correctable.  Continuing
   from the error causes CHECK-ARG-TYPE to accept a new value for ARG-NAME and redo the test.
   CHECK-ARG-TYPE returns NIL.")

(setf (arglist 'FIXP) (VALUES '(X) '(T-OR-NIL)))
(setf (documentation 'FIXP 'function)
  "Returns true if X is an integer, otherwise NIL.")

(setf (documentation 'TERMINAL-IO 'variable)
  "Global variable whose value is the default stream for I/O to the console, which is normally
   the selected window for output and the keyboard for input when the operations are interactive.") 

(setf (documentation 'STANDARD-INPUT 'variable)
  "Global variable whose value is the default stream for input functions such as READ.")

(setf (documentation 'ERROR-OUTPUT 'variable)
  "Global variable whose value is the default stream for unanticipated noninteractive output,
   such as warnings.")

(setf (arglist 'STRING-REVERSE-SEARCH-NOT-CHAR) (VALUES
						  '(CHAR STRING &OPTIONAL FROM (TO 0) CONSIDER-CASE)
						  '(INDEX)))
(setf (documentation 'STRING-REVERSE-SEARCH-NOT-CHAR 'function)
  "Returns the INDEX into STRING of the last character that is not equal to CHAR.  Even though 
   the search is in reverse order, the index is relative to the beginning of STRING.  Since the 
   search is in reverse order, the last occurrence is found first.  If all characters in STRING
   are equal to CHAR, STRING-REVERSE-SEARCH-NOT-CHAR returns NIL.  Although the entire length of
   STRING is the default, a substring starting at 1 less than FROM and ending at TO may be specified.
   TO should normally be less than FROM.  Both case and font are considered if CONSIDER-CASE is true.")

(setf (arglist 'PROCESS-CREATE) (VALUES
				  '(NAME &KEY SIMPLE-P FLAVOR STACK-GROUP WARM-BOOT-ACTION QUANTUM PRIORITY
                                    SG-AREA REGULAR-PDL-AREA SPECIAL-PDL-AREA REGULAR-PDL-SIZE
                                    SPECIAL-PDL-SIZE &ALLOW-OTHER-KEYS)
				  '(NEW-PROCESS)))
(setf (documentation 'PROCESS-CREATE 'function) 
  "Creates a process with name NAME.  The keyword arguments are defined as :
     :SIMPLE-P  - If true, flavor SI:SIMPLE-PROCESS is used. 
     :FLAVOR  - Specifies the flavor of process to make.  If both :FLAVOR and :SIMPLE-P
                are NIL, the flavor SI:PROCESS is used.
     :STACK-GROUP - Specifies the stack group for this process to run in. If it is omitted,
                    the keyword arguments :SG-AREA, :REGULAR-PDL-AREA, :SPECIAL-PDL-AREA,
                    :REGULAR-PDL-SIZE, and :SPECIAL-PDL-SIZE are passed to function MAKE-STACK-GROUP.
     :WARM-BOOT-ACTION - Specifies either a function to call on warm booting or the value :FLUSH.
                         :FLUSH means flush the process.  The default is to restart it. 
                         SI:PROCESS-WARM-BOOT-RESET will kill the process. 
                         SI:PROCESS-WARM-BOOT-RESTART will restart it at an earlier stage of booting.
     :QUANTUM - Specifies in 60'ths of a second the time quantum; defaults to one second.
     :PRIORITY - Specifies execution priority; the larger the number the more ofter it runs.
                 The default is 0."
     )

(setf (documentation 'STANDARD-OUTPUT 'variable)
  "Global variable whose value is the default stream for output functions such as PRINT.")

(setf (arglist 'STRING) (VALUES '(STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2) '(INDEX)))
(setf (documentation 'STRING 'function)
  "Returns the INDEX into STRING1 of the first point of difference between it and STRING2 that
   satisfies the test greater than or equal to.  Both case and font are considered.  If the test
   is never satisfied, NIL is returned.  If the strings match, the length of the shorter string
   is returned.  Although the default string length of both STRING1 and STRING2 is the entire
   length, a substring of STRING1 starting at :START1 and ending at :END1 and/or and substring
   of STRING2 starting at :START2 and ending at :END2 may be specified.")

(setf (arglist 'MAKE-SYN-STREAM) (VALUES '(STREAM-SYMBOL) '(NEW-STREAM)))
(setf (documentation 'MAKE-SYN-STREAM 'function)
  "Returns a stream which will perform all operations by redirecting those operations to the
   stream that is then the value of the variable STREAM-SYMBOL.")

(setf (arglist 'STRING-NREVERSE) (VALUES '(STRING) '(MODIFIED-STRING)))
(setf (documentation 'STRING-NREVERSE 'function)
  "Returns a string whose elements are those of STRING but in reverse order.  STRING-REVERSE will
   work on any one-dimensional array.  STRING is destructively modified by STRING-NREVERSE.")

(setf (arglist 'FUNCALL-SELF) '(&REST ARGS))
(setf (documentation 'FUNCALL-SELF 'function) 
  "Obsolete function equivalent to FUNCALL with SELF as the first argument.  It should be replaced
   with FUNCALL or SEND.")

(setf (arglist 'BIT-TEST) (VALUES '(BITS WORD) '(T-OR-NIL)))
(setf (documentation 'BIT-TEST 'function)
  "Returns true if any of the 1 bits in BITS are also 1's in WORD, otherwise NIL.") 

(setf (arglist 'QC-FILE) '(INFILE &OPTIONAL OUTFILE LOAD-FLAG IN-CORE-FLAG PACKAGE-SPEC
                           FILE-LOCAL-DECLARATIONS DONT-SET-DEFAULT-P IGNORE IGNORE))
(setf (documentation 'QC-FILE 'function)
  "Compiles source file INFILE into object file OUTFILE.  The arguments are defined as :
     LOAD-FLAG - Always set this to NIL.
     IN-CORE-FLAG - Always set this to NIL.
     PACKAGE-SPEC - The name of the package to read the source file into.
     FILE-LOCAL-DECLARATIONS - Used to compile multiple files as if they were one file. 
     DONT-SET-DEFAULT-P - If NIL, the compile default pathname components will not be updated.") 

(setf (arglist 'SYMEVAL) (VALUES '(SYMBOL) '(VALUE)))
(setf (documentation 'SYMEVAL 'function)
  "Returns the current VALUE of the dynamic symbol SYMBOL.  An error is signaled if SYMBOL
   is unbound.")

(setf (arglist 'STRING) (VALUES '(STRING1 STRING2 &KEY (START1 0) END1 (START2 0) END2) '(INDEX)))
(setf (documentation 'STRING 'function)
  "Returns the index, relative to STRING1, of the first difference found between STRING1 and
   STRING2.  Although the entire length of both STRING1 and STRING2 is the default, a substring
   of STRING1 starting at :START1 and ending at :END1 and/or a substring of STRING2 starting at
   :START2 and ending at :END2 may be specified.  Case and font are significant during the
   comparison.  STRING returns NIL if the strings are equal.")

(setf (documentation 'ALL-SPECIAL-SWITCH 'variable) 
  "Variable whose value, when true, causes all variables to be compiled as special variables no
   matter how they were declared.")

(setf (documentation 'BASE 'variable) 
  "Variable whose value is the radix used to print integers and ratios.") 

(setf (arglist 'COMPILE-FILE-ALIST) '(FILE-ALIST &OPTIONAL
						 (DONT-ASK-P 0)
						 (DONT-CARE-IF-UNCHANGED-P 0)
						  DONT-ASK-FOR-CONFIRMATION PACKAGE-SPEC))
(setf (documentation 'COMPILE-FILE-ALIST 'function)
  "Compiles all the files in FILE-ALIST that are either not compiled or have changed. Each element of
   FILE-ALIST is a list whose CAR is a filename string.  They are read into package PACKAGE-SPEC.  The
   remaining arguments can be defined as :
      DONT-ASK-P - Controls whether the user is queried about each file.
      DONT-CARE-IF-UNCHANGED-P - Controls whether unchanged files are recompiled.
      DONT-ASK-FOR-CONFIRMATION - Controls whether the user is queried for one final confirmation.")

(setf (arglist 'SUBRP) (VALUES '(OBJECT) '(T-OR-NIL)))
(setf (documentation 'SUBRP 'function) 
  "Returns true if OBJECT is either a compiled code object or microcoded function, otherwise NIL.")

(setf (arglist 'ARRAY-PUSH) (VALUES '(ARRAY VALUE) '(MODIFIED-ARRAY)))
(setf (documentation 'ARRAY-PUSH 'function)
  "Stores the new element VALUE in array ARRAY at the array position specified by the fill pointer.
   This position should be the end of the array.  The fill pointer is incremented by one and the
   former value of the fill pointer is returned by ARRAY-PUSH.  If the element cannot be added,
   ARRAY-PUSH returns NIL.")

(setf (arglist 'MINUS) (VALUES '(NUMBER) '(VALUE)))
(setf (documentation 'MINUS 'function) 
  "Returns the VALUE of the first argument minus the remaining arguments, or the negation of 
   NUMBER if only one argument is given.")

(setf (arglist 'READLINE) (values '(&OPTIONAL STREAM EOF-OPTION OPTIONS) '(string)))
(setf (documentation 'READLINE 'function) 
  "Read a line from STREAM and return it as a string.  This string does not include a Return
   character and is empty when a blank line is read.  If EOF-OPTION is non-NIL, EOF-OPTION is
   returned when end of file is encountered;  if NIL, end of file is signaled as an error.  OPTIONS
   is a list of rubout handler options passed to WITH-INPUT-EDITING if STREAM supports the
   :RUBOUT-HANDLER operation.  READLINE returns two addition values.  Value2 is EOF-OPTION if
   termination is due to end of file.  VALUE3 is the delimiter which ended the input, or NIL if
   it ended due to EOF.")

(setf (arglist 'PIXEL-ARRAY-HEIGHT) (VALUES '(ARRAY) '(LENGTH)))
(setf (documentation 'PIXEL-ARRAY-HEIGHT 'function)
  "Returns the LENGTH of the dimension that varies the slowest in memory of the two dimension array
   ARRAY.  For a screen array, this dimension is the height.") 

(setf (arglist 'REMQ) (VALUES '(ITEM LIST &OPTIONAL (TIMES -1)) '(NEW-LIST)))
(setf (documentation 'REMQ 'function)
  "Returns a list created by removing all element in LIST that are EQ to ITEM.  LIST is not altered
   by REMQ.  If TIMES is included, it is an integer specifying how many occurrences of ITEM should
   be removed from LIST.")

;(setf (arglist 'STORE) (ARRAY-REFERENCE VALUE))
;(setf (documentation 'STORE 'function) 
;  NIL)

(setf (arglist 'RASSQ) (VALUES '(ITEM ALIST) '(FIRST-MATCHING-CONS)))
(setf (documentation 'RASSQ 'function)
  "Returns the first cons in association list ALIST whose CDR is EQ to ITEM.  If no match is found,
   RASSQ returns NIL.") 

(setf (arglist 'REST3) (VALUES '(LIST) '(MODIFIED-LIST)))
(setf (documentation 'REST3 'function)
  "Returns a list created by applying the function CDR 3 times to LIST, analogous to removing the
   first 3 elements of LIST.") 

(setf (arglist 'FIND-POSITION-IN-LIST) (VALUES '(ELEMENT LIST) '(INDEX)))
(setf (documentation 'FIND-POSITION-IN-LIST 'function)
  "Returns the INDEX of the element in list LIST that is EQ to ELEMENT.  If ELEMENT is not a member
   of LIST, FIND-POSITION-IN-LIST returns NIL.  Zero-origin numbering is used.")  

(setf (documentation 'PRINLENGTH 'variable)
  "Variable whose value, when non-NIL, is an integer specifying how many elements of a list
   are printed before the remaining list is representated by ellipses (...).")

(setf (arglist ') (VALUES '(NUMBER &REST NUMBERS) '(T-OR-NIL)))
(setf (documentation ' 'function)
  "Returns true if the arguments are in nondecreasing numerical order, otherwise NIL.")

(setf (arglist 'CHAR) (VALUES '(&REST CHARS) '(T-OR-NIL)))
(setf (documentation 'CHAR 'function)
  "Returns true if all the characters are not equal, considering bits, font, and case; otherwise,
   NIL is returned.")

(setf (arglist '\\ ) (VALUES '(X Y) '(REMAINDER)))
(setf (documentation '\ 'function)
  "Returns the REMAINDER after dividing DIVIDEND by MODULUS.  The sign of a non-zero return value
   is the same as the sign of X.") 

(setf (arglist 'MEMQ) (VALUES '(X LIST) '(MODIFIED-LIST)))
(setf (documentation 'MEMQ 'function)
  "Returns a list consisting of the tail of list LIST whose CAR is the first element of LIST that
   is EQ to X.  If no match is found, MEMQ returns NIL.")

(setf (arglist 'STRING-SEARCH) (VALUES
				 '(KEY STRING &OPTIONAL (FROM 0) TO (KEY-FROM 0) KEY-TO CONSIDER-CASE)
				 '(INDEX)))
(setf (documentation 'STRING-SEARCH 'function)
  "Returns the INDEX in STRING of the first substring that matches KEY.  If no match is found,
   NIL is returned.  Although the entire length of both STRING and KEY is the default, a substring
   of STRING starting at FROM and ending at TO and/or a substring of KEY starting at KEY-FROM and
   ending at KEY-TO may be specified.  Both case and font are considered if CONSIDER-CASE is true.")

(setf (arglist 'STRING-SEARCH-CHAR) (VALUES
				      '(CHAR STRING &OPTIONAL (FROM 0) TO CONSIDER-CASE)
				      '(INDEX)))
(setf (documentation 'STRING-SEARCH-CHAR 'function)
  "Returns the INDEX in STRING of the first occurrence of character CHAR.  If no match is found,
   NIL is returned.  Although the entire length of STRING is the default, a substring of STRING
   starting at FROM and ending at TO may be specified.  Both case and font are considered if
   CONSIDER-CASE is true.")

(setf (arglist 'DEL) (VALUES '(PRED ITEM LIST &OPTIONAL (TIMES -1)) '(MODIFIED-LIST)))
(setf (documentation 'DEL 'function)
  "Returns a list where those elements in LIST that satisfy test PRED have been deleted.  PRED is
   a function that takes two arguments, ITEM and an element of LIST.  When PRED returns true, that
   element is deleted.  If TIMES is specified, it is an integer stating how many elements that
   match ITEM should be deleted from LIST.")

(setf (arglist 'FSYMEVAL) (VALUES '(SYMBOL) '(FUNCTION-DEFINITION)))
(setf (documentation 'FSYMEVAL 'function)
  "Returns the FUNCTION DEFINITION of SYMBOL, otherwise signal an error if one does not exit.")

(setf (arglist 'MAKE-PIXEL-ARRAY) (VALUES '(WIDTH HEIGHT &REST OPTIONS) '(ARRAY-OBJECT)))
(setf (documentation 'MAKE-PIXEL-ARRAY 'function)
  "Creates a two dimensional array a returns the ARRAY OBJECT.  The dimensions are specified by
   WIDTH, the subscript that varies the fastest in memory, and HEIGHT, the subscript that varies
   the slowest in memory.  :OPTIONS are identical to those of MAKE-ARRAY.  The array returned by
   MAKE-PIXEL-ARRAY should be accessed by the AR-2-REVERSE and AS-2-REVERSE functions to guarantee
   that accessing is independent of array dimension order.")

(setf (arglist 'FIX) (VALUES '(NUMBER) '(INTEGER)))
(setf (documentation 'FIX 'function)
  "Returns the largest integer that is not larger than NUMBER.")

(setf (arglist '*THROW) (VALUES '(TAG &QUOTE VALUE-EXPRESSION) '(MODIFIED-VALUES)))
(setf (documentation '*THROW 'function)
  "Transfers control out of a matching CATCH.  After TAG is evaluated to produce a throw tag, 
   VALUE-EXPRESSION is evaluated and all of its values saved.  A search of the stack is made to
   find the innermost CATCH that is EQ to the throw tag, that CATCH construct is exited, and
   the saved values are returned as the value(s) of the CATCH.")

(setf (arglist 'SUBSTRING) (VALUES '(STRING FROM &OPTIONAL TO (AREA NIL)) '(NEW-STRING)))
(setf (documentation 'SUBSTRING 'function) 
  "Returns a substring of STRING starting at FROM and ending at TO.  If TO is not specified, the
   end of the string is used.  If AREA is specified, the new string is created in that area.")

(setf (arglist 'SWAPF) (VALUES '(&REST PLACES) '(MODIFIED-PLACES)))
(setf (documentation 'SWAPF 'function)
  "Rotates the place values left one position by performing a left circular shift one position
   on all the specified PLACES.  That is, place-value-1 = place-value-2, place-value-2 =
   place-value-3, ..., place-value-last = place-value-1.")

(setf (arglist '*CATCH) (VALUES '(TAG &QUOTE &REST BODY) '(RESULTS)))
(setf (documentation '*CATCH 'function)
  "Construct used for non-local exits.  TAG is evaluated to produce an object that names the
   CATCH.  BODY is then evaluated as an implicit PROGN with the results of the last form
   evaluated returned as the values of CATCH.  However, if during this evaluation of BODY a
   *THROW is executed with the same tag as CATCH, evaluation of the forms in BODY immediately
   halts and CATCH returns the values specified by the THROW.")

(setf (arglist 'COPYALIST) (VALUES '(ALIST &OPTIONAL (DEFAULT-CONS-AREA DEFAULT-CONS-AREA))
				   '(NEW-ALIST)))
(setf (documentation 'COPYALIST 'function)
  "Returns an association list that is a copy of ALIST.  Dotted pair termination of list will also
   be copied.")

(setf (arglist 'LOCAL-DECLARE) '(DECLARATIONS &BODY BODY))
(setf (documentation 'LOCAL-DECLARE 'function)
  "Construct used for embedding declarations within executable code.  These declarations are in 
   effect throughout the construct at the head of whose body the LOCAL-DECLARE appears.  BODY is 
   evaluated with DECLARATIONS in effect.  DECLARATIONS is a list of declarations, each of which
   is a list.  Declarations include (SPECIAL variables...), (ARGLIST 'argument-names...),
   (RETURN-LIST value-names...), and (:SELF-FLAVOR flavorname).")

;(setf (arglist '*FEXPR) (&QUOTE &REST L))
;(setf (documentation '*FEXPR 'function) 
;  NIL)

(setf (arglist 'REMAINDER) (VALUES '(DIVIDEND MODULUS) '(REMAINDER)))
(setf (documentation 'REMAINDER 'function)
  "Returns the REMAINDER after dividing DIVIDEND by MODULUS.  The sign of a non-zero return value
   is the same as the sign of DIVIDEND.") 

;(setf (arglist '*LEXPR) (&QUOTE &REST L))
;(setf (documentation '*LEXPR 'function) 
;  NIL)

(setf (arglist 'ARRAY-POP) (VALUES '(ARRAY) '(ARRAY-ELEMENT)))
(setf (documentation 'ARRAY-POP 'function)
  "Decrements the fill pointer of ARRAY by 1 and returns the ELEMENT pointed to by this new
   value.  An error is signaled if ARRAY is empty.")

(setf (arglist 'TYO) '(CHAR &OPTIONAL STREAM))
(setf (documentation 'TYO 'function) 
  "Outputs CHAR to stream STREAM.")

(setf (arglist 'CASEQ) (VALUES '(TEST-OBJECT &BODY CLAUSES) '(RESULT)))
(setf (documentation 'CASEQ 'function)
  "Control structure designed to execute only one case from a set of cases.  Syntax of 
   (CASEQ TEST-OBJECT (test1 body1) ... (testN bodyN)) is defined as :  the (test# body#)
   construct is called a clause.  TEST-OBJECT is evaluated.  <test1> is evaluated and compared
   against TEST-OBJECT; if they are EQL, all forms in <body1> are evaluated sequentially and the
   value of the last expression in <body1> is returned by CASE.  If <test1> does not match,
   <test2> is evaluated and compared, and so on, until either one <test#> matches TEST-OBJECT or
   all clauses are exhausted.  CASEQ returns NIL if no clause is selected.  A T as the <testN>
   matches any TEST-OBJECT.")

(setf (arglist 'ARRAY-PUSH-EXTEND) (VALUES '(ARRAY DATA &OPTIONAL EXTENSION) '(INDEX MODIFIED-ARRAY)))
(setf (documentation 'ARRAY-PUSH-EXTEND 'function)
  "Stores the new element DATA at the end of ARRAY, extending ARRAY by EXTENSION number of cells
   if required.  ARRAY must have a fill pointer.  ARRAY-PUSH-EXTEND returns the INDEX of DATA.")  

(setf (documentation 'IBASE 'variable)
  "Variable whose value is the default radix for reading integers and ratios.")

;;;Number of  functions  : 99
;;;Number of  variables  : 11
;;;Number of  structures : 0
;;;Number of  types      : 0
;;;Number of  setf       : 0

;;;Total number combined is 110












