;;; -*- Mode:Common-Lisp; Package:TV; Base:10; Fonts:(COURIER HL12B HL12I); PATCH-FILE: T -*-

;;************************************************************************

;; This was written by Garr S. Lystad, 3-26-88

;; This file adds a search capability to the inspector in the first inspect pane.
;; This is done through use of the command string-search to #\C-s.

;; You are prompted for the search string.  Type in what you want to find.
;; The search normally begins at the top line of the display (not counting
;; the title line.  An exception is made for a second search of the same
;; object for the same string as the last search.  In that case the search
;; is made from the line following the line where the last find was made.
;; A #\C-s entered for the search string will bring up the last thing
;; typed to the inspector, whether for a search or not.

;; Numeric args are given to the search as #\C-<digit>, #\C-<digit>, ....
;; Giving an arg to the search which is > 10 will cause the search to
;; start at the top of the displayed item no matter what.
;; Giving an arg to the search which has a remainder of 1 mod 10 will
;; cause a search through an instance variable to search only the slot names.
;; Giving an arg to the search which has a remainder of 2 mod 10 will
;; cause a search through an instance variable to search only the slot values.

;; For further documentation see the string-search command.

;; Items found are marked by moving the mouse to their beginning.
;; Horizontal placement which should place it off the right of the screen
;;  leaves the cursor at the right side of the screen.  Things off the
;;  right are searched too.

;;************************************************************************


(defvar *inspectors-search-states* nil "Keeps track of what has been found where for various inspectors,
   arranged as an alist by inspector pane.")

(defstruct (iss :type :list)
  window object line index string)

(DEFCOMMAND string-search (numarg)	
  '(:DESCRIPTION "Search for a typed in string, Arg > 9 starts at very top, see documentation."
    :DOCUMENTATION "Search for a typed in string, With arg > 9 search from top of object. 
      With smaller arg search from top of screen unless this string is the same as the
      last and the object is the same, in which case we search from the line after the
      last find.  With various (mod args 10) values certain types of items can be eliminated.
      1 eliminates instance-value
      2 eliminates instance-slot
      C-s brings back the last thing typed."
    :NAMES ("Search")
    :KEYS (#\C-S )
    :ARGUMENTS (UCL:NUMERIC-ARGUMENT))
  (DECLARE (SPECIAL inspectors))
  (multiple-value-bind (scroll-line line-index string)
      (send (CAR inspectors) :prompt-and-search
	    (if (and numarg (> numarg 9.)) :very-top :top-item-shown)
	    (case (if numarg (mod numarg 10) nil)
	      (nil nil)
	      (1  (list 'tv:instance-value))
	      (2  (list 'tv:instance-slot))
	      ))
    (let ((input-window (send (send (CAR inspectors) :superior) :selection-substitute)))
      (if scroll-line
	  (progn
	    (center-on-line (CAR inspectors) scroll-line line-index)
	    (save-inspector-search-state
	      (CAR inspectors)
	      (send (CAR inspectors) :current-object)
	      scroll-line
	      line-index
	      string)
	    (send input-window :clear-screen)
	    (format input-window "Found it.~%Inspect:") )
	(tv:beep)
	(send input-window :clear-screen)
	(format input-window "Search failed.~%Inspect:") ))))

(defmethod (inspect-window :prompt-and-search) (start-position &optional excluded-item-types)
  2"This routine prompts for the search string and returns the line, line-index, and the string.
   Nil is returned if the string is not found.
   The args are a start position: one of (:VERY-TOP :TOP-ITEM-SHOWN :CURRENT-POSITION*
   2  or NIL which is like :top-item-shown unless a second control s is used for string input.)
    and a list of item types not to search: chosen from
       (  *tv:instance-slot tv:instance-value ...2)"*
  (let ((old-data (assoc SELF *inspectors-search-states*))
	  string)
      ;1;Get the sting*
      (send *standard-input* :clear-screen)
      (setq string (send *standard-input* :rubout-handler
			 '((:prompt "Search string:")) *standard-input* :read-string))
      ;;(format tv:initial-lisp-listener "~&insp-search string=~s~%" string)
      ;1;Take care of special returned values*
      (if (or (eq string :old-string)
	      (and string old-data
		   (string-equal string (iss-string old-data))
		   (eq current-object (iss-object old-data)) ))
	  (progn
	    (setq string (if old-data (iss-string old-data) nil))
	    (unless (eq start-position :very-top) (setq start-position 2:CURRENT-POSITION*)) ))
      (send *standard-input* :clear-screen)
      (WHEN string
	;1;Let user know we're looking*
	(send *standard-input* :string-out "Looking for: ")
	(send *standard-input* :string-out string)
	;1;Search for the string*
	(send self :search
	      string
	      (case start-position
		(:very-top 0)
		((:top-item-shown nil) top-item)
		(:current-position (iss-line old-data))
		(:otherwise (ferror nil "Unknown start position for search.")) )
	      excluded-item-types))))

(defmethod (inspect-window :search) (string first-line excluded-item-types)
  "This loops from the first-line-th line of the window looking for string in
   a string representation of each window line, until string is found.
   excluded-item-types is used only when the line is normally printed by
   4INSPECT-PRINTER*, in which case printed items of the types in excluded-item-types
   are replaced by an equivalent number of blanks in the string representation of
   the line, and thus excluded from the search."
  (let (index)
    (loop for ITEM-NO from first-line below (SEND self :number-of-items)
	  do (let ((printed-line
		     (with-output-to-string (*standard-output*) 
		       (APPLY PRINT-FUNCTION
			      (SEND self :ITEM-OF-NUMBER ITEM-NO)
			      PRINT-FUNCTION-ARG
			      *standard-output*
			      (APPEND (LIST ITEM-NO)
				      (IF (EQ PRINT-FUNCTION '4INSPECT-PRINTER*)
					  (LIST excluded-item-types) ))))))
	       ;;(format tv:initial-lisp-listener "~&~a~%" printed-line)
	       (setq index
		     (search (the string string)
			     (the string printed-line)
			     :test #'string-equal)))
	  when index
	    do (return (1+ ITEM-NO) index string) )))

;; have to add :item1 method to WITH-OUTPUT-TO-STRING
(let ((compile-encapsulations-flag t))
  (advise format:format-string-stream :around string-search nil
    (case (first arglist)
      (:item1
       (let ((st
	       (with-output-to-string (*standard-output*)
		 (let* ((item (second arglist))
			;;(type (third arglist))
			(function (or (fourth arglist) #'PRIN1))
			(print-args (nthcdr 4 arglist)) )
		   (APPLY FUNCTION ITEM *standard-output* PRINT-ARGS) ))))
	 ;;(format tv:initial-lisp-listener " i1 string=~s~%" st)
	 (or format::format-string
	     (setq format::format-string (make-array (length st) :element-type 'string-char
						     :fill-pointer 0.)) )
	 (adjust-array format::format-string (+ (length format::format-string) (length st)))
	 (dotimes (i (length st))
	   (vector-push (aref st i) format::format-string) )
	 st))
      (:otherwise :do-it) )))

(defmethod (INSPECTOR-INTERACTION-PANE :read-string) ()
  "This is called by the rubout handler method for the interactive pane."
  (let (result)
    (loop as char = (send self :read-any)
	  when (char= char (code-char 254.));1;254 = #\Symbol-s = *
	  do (return :old-string)
	  when (char= char #\return)
	  do (let* ((length (length result))
		    (string (make-array length :fill-pointer length :type 'art-string)) )
	       (return (loop for ch in result
			     for i from 0 below length
			     do (setf (aref string i) ch)
			     finally (progn (setf (fill-pointer string) length)
					    (return string)) )))
	  do (setq result (nconc result (list char))) )))

(defun save-inspector-search-state (window object line index string)
  "This records what each inspector has searched for last.  It is
   more general than needed for the sake of future enhancement."
  (let* ((entry (assoc window *inspectors-search-states*)))
    (if line
	(if entry
	    (setf (iss-object entry) object
		  (iss-line entry) line
		  (iss-index entry) index
		  (iss-string entry) string)
	  (push (make-iss :window window :object object :line line :index index :string string)
		*inspectors-search-states*) )
      (tv:beep))))

(defmethod (inspect-window :lines) ()
  (floor height line-height) )
      
(defun center-on-line (window line line-index)
  "This centers the indicated line of window and moves mouse to the centered line and line-index col."
  (let* ((char-height (W:SHEET-LINE-HEIGHT WINDOW))
	 (char-width (W:SHEET-CHAR-WIDTH WINDOW))
	 (lines-in-window (send window :lines));;Might use :top-item msg & :relative for no unnecess scroll.
	 (lines-scrolled (if (> line (floor lines-in-window 2));lines-in-window)
			     (- line (floor lines-in-window 2))
			   0)) )
    (SEND window :SCROLL-TO lines-scrolled :absolute)
    (SEND WINDOW :SET-MOUSE-CURSORPOS
	  (min (+ (* line-index char-width) -1
		  (FLOOR (W:SHEET-CHAR-WIDTH WINDOW) 2))
	       (- (send window :width) (* 3 (W:SHEET-CHAR-WIDTH WINDOW))) )
	  (+ (* char-height
		(if (zerop lines-scrolled)
		    line
		  (floor lines-in-window 2)) )
	     (FLOOR (* -3 (W:SHEET-LINE-HEIGHT WINDOW)) 4) ))))

;;********
;;Put the command in place;

(DEFPARAMETER inspector-non-menu-cmds	
	  '(page-up-cmd
	     page-down-cmd
	     page-to-top
	     page-to-bottom
	     break-cmd
	     string-search))

(BUILD-COMMAND-TABLE 'inspector-other-cmd-table 'inspect-frame
  inspector-non-menu-cmds
  :INIT-OPTIONS
  '(:NAME "Other Inspector commands"))

;;********************************************************************************
;; Fixes to system code
;;********************************************************************************

(DEFUN UPDATE-PANES (&optional ignore)                                ;!
  "Update inspection and history panes to current objects.  Should be called
   by each command that effects the inspected objects list."
  (DECLARE (SPECIAL TOP-ITEM HISTORY FRAME INSPECTORS))
  (LET ((ITEMS (SEND HISTORY :ITEMS))
        (PANE)
        (IDX)
        (NUM-PANES (LENGTH INSPECTORS))
        (locked-objects nil)
        (history-object nil))   ;!
    (SETQ IDX (ARRAY-ACTIVE-LENGTH ITEMS))
    (SETQ locked-objects nil)
    ;; Make sure the inspection windows reflect the state of the history buffer
    (DOLIST (I INSPECTORS)
      ;; Update datastructure to reflect current TOP-ITEMs
      (LET ((DISP (SEND I :CURRENT-DISPLAY)))
	(AND DISP (SETF (FOURTH DISP) (SEND I :TOP-ITEM))))
      (when (send i :locked-p) (setq locked-objects (cons (send i :current-object) locked-objects))))   ;!
    (DOTIMES (I NUM-PANES)
      (SETQ PANE (NTH I INSPECTORS))
           (COND ((SEND pane :locked-p)
		  (w:prepare-sheet (pane)
		    (w:draw-char fonts:icons 98. (- (SEND pane :width) 50.) 3. w:alu-andca pane))
		  (SEND history :inspect-object (SEND pane :current-object) pane top-item nil t)
		  (w:prepare-sheet (pane)
		    (w:draw-char fonts:icons 98. (- (SEND pane :width) 50.) 3. w:alu-xor pane)))
		 (t ;;really display something
		  (SETQ IDX (1- IDX)) ;1;number of items in history*
		  (COND
		    ((< IDX 0)
		     (SEND PANE :SET-CURRENT-DISPLAY
			   (SEND pane :SETUP
				 `(INSPECT-PRINTER NIL NIL NIL
						   (NIL NIL NIL NIL
							,(LABEL-FONT (SEND pane :LABEL))
							"2Empty*"))))
		     (SEND PANE :SET-CURRENT-OBJECT (LIST NIL)))
		    (T ;1;THERE IS SOMETHING TO PRING*
		     (SETQ history-object
					   ;1;advance past items displayed in locked panes*
			   (do* ((history-object (aref items idx) (IF (>= idx 0) (aref items idx))))	
				((or (< idx 0)
				     (not (member history-object locked-objects :test #'EQ)))
				 history-object)
			     (setq idx (1- idx))))
		     (COND (history-object1 *;1;got something to inspect*
			    (SEND HISTORY :INSPECT-OBJECT history-object PANE TOP-ITEM NIL T))
			   
			   (t	 ;1;nothing to inspect, so need to clear the pane*
			    (SEND PANE :SET-CURRENT-DISPLAY
				  (SEND pane :SETUP
					`(INSPECT-PRINTER NIL NIL NIL
							  (NIL NIL NIL NIL
							       ,(LABEL-FONT (SEND pane :LABEL))
							       "2NIL*")))) ;;gsl. 3-26-88
			     ;;**The last T should be NIL, but it acts haywire if it is!!!
			    (SEND PANE :SET-CURRENT-OBJECT (LIST NIL))))))))
      );!
    (SETQ TOP-ITEM NIL))                                                         
  ;; Insure last item in history is on the screen
  (SEND HISTORY :PUT-LAST-ITEM-IN-WINDOW)
  ;; Give *, ** and *** the right values.
  (send frame :update-*)  )

(DEFUN 4INSPECT-PRINTER* (LINE ARG STREAM ITEM-NO &optional excluded-item-types)      ;1fi*
  ;1;;Make sure base is consistent since sometimes this is called from the mouse process.*
  (DOLIST (ELT LINE)
    (COND
      ((NUMBERP ELT) (FORMAT STREAM "3~VT*" ELT))
      ((STRINGP ELT) (PRINC ELT STREAM))
      ((NOT (LISTP ELT)) (FERROR NIL "3Unknown element type: ~S*" ELT))
      ((STRINGP (CAR ELT)) (APPLY #'FORMAT STREAM ELT))
      (T
       (CASE (FIRST ELT)
	 (:FUNCTION (APPLY (SECOND ELT) ARG STREAM ITEM-NO (CDDR ELT)))
	 (:COLON (FORMAT STREAM "3:~VT *" (SECOND ELT)))
	 ;1;Provides a mechanism for inserting bold or italicized text (for column headers).*
	 (:FONT
	  (UNWIND-PROTECT (PROGN
			    (SEND STREAM :SET-CURRENT-FONT (SECOND ELT))
			    (PRINC (THIRD ELT) STREAM))
	    (SEND STREAM :SET-CURRENT-FONT 0)))
	 (:ITEM1
	  (if (member (second elt) excluded-item-types);1;print out appropriate num of blanks.*
	      (format stream "~V@T"
		      (length (with-output-to-string (stream2)
				(apply (OR (FOURTH ELT)
					   #'INSPECTION-DATA-PRINT-ITEM-CONCISELY)
				       (THIRD ELT)
				       stream2
				       (NTHCDR 5 elt) ))))	       
	    (LEXPR-SEND STREAM :ITEM1
			ELT
			(SECOND ELT)
			#'(LAMBDA (ELT &REST ARGS)
			    (APPLY (OR (FOURTH ELT)
				       #'INSPECTION-DATA-PRINT-ITEM-CONCISELY)
				   (THIRD ELT)
				   ARGS))
			(NTHCDR 4 elt))))
	 (OTHERWISE (FERROR () "3Unknown item type ~A*" (FIRST ELT))))))))


(DEFUN 4inspection-data-print-item-concisely* (thing stream &OPTIONAL (level 0))        ;1fi*
  ;1;;Send changed to Send-if-handles.  Frames, which are not inspect frames may still have this*
  ;1;;sort of window (debugger frames).  Such frames may not have the extra IVs to cope with this message.*
  ;;(break)
  (IF (AND (TYPEP thing 'inspection-data)
	   (Si:send-if-handles stream :SUPERIOR) ;;for the search we have no superior. gsl.
           (Si:send-if-handles (SEND stream :SUPERIOR) :inspection-data-active?))
      (SEND thing :format-concisely stream)
    (print-item-concisely thing stream level)))

;;------------------------------------------------------------------------------END
