;;; -*- Mode:Common-Lisp; Package:MT; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1987, Texas Instruments Incorporated. All rights reserved.

;;; Copyright (C) 1987 Unisys Corporation
;;; All Rights Reserved

;;; ************************************************************************

;;; This software was developed by the Knowledge Systems Project of Unisys
;;; Corporation and is provided as a courtesy to all Explorer users.  It
;;; enables the Explorer to access more than one tape drive simultaneously.
;;; It was developed to facilitate software manufacturing on Explorers.
;;; Surprisingly, multiple tape drives can be accessed in only marginally
;;; greater time than a single tape drive.  For instance, a large directory
;;; might take 10 minutes to copy to a single tape drive, but only take
;;; 15 minutes to copy to 4 tape drives simultaneously.

;;; ************************************************************************

;;; Using the following code, an Explorer can support multiple tape drives,
;;; each supporting independent operation.  The key to this activity is that
;;; the operations to each tape drive must run in separate Explorer processes.
;;; The following two interfaces are provided:
;;;
;;; LISP function interface -
;;;
;;;   Allows a user to call the standard tape manipulation functions (e.g.,
;;;   mt:backup, mt:verify, etc.) for each tape drive.  An interactive
;;;   option is supported so that the operations can be selected from a
;;;   standard Explorer choose-variable-values menu.
;;;
;;; Backup System Window interface -
;;;
;;;   Allows a user to instantiate any number of Backup Windows (using SYSTEM-B
;;;   and SYSTEM-CONTROL-B sequences), then associate each Backup Window with
;;;   a particular tape unit.
;;;
;;; 
;;; Development of the multi-tape facilities was accomplished by writing three
;;; new functions, changing two existing functions, one existing flavor, and
;;; one existing method.
;;;
;;; LISP Function interface: Allow the use of multiple tapes by starting
;;;                          separate processes for each tape operation.
;;;
;;;   MULTI-TAPE:            New function to perform error checking and start
;;;                          separate processes.
;;;
;;;   MULTI-TAPE-INTERNAL:   New function.  Note that the binding of the six
;;;                          variables at the beginning is a kludge to overcome
;;;                          the design limitations where unit # is not passed
;;;                          downward as a parameter, and where global variables
;;;                          are relied upon for buffering.
;;;
;;; Backup Window interface: The approach here is to instantiate a separate Backup frame
;;;                          for each tape so that each will run in its own process.
;;;
;;;   BACKUP-FRAME:          Redefined flavor to add a TAPE-UNIT variable.
;;;
;;;   COMMAND-LOOP:          Redefined function to perform the same kludge as
;;;                          MULTI-TAPE-INTERNAL - binds global variables locally
;;;                          so there will be no interference.
;;;
;;;   PREPARE-TAPE:          Redefined method.  Once user picks the drive, insure
;;;                          that no other Backup frame is already associated
;;;                          with the selected drive.  Stores the drive in the
;;;                          TAPE-UNIT variable for the frame.
;;;
;;;   CHECK-TAPE-ASSIGNED:   New function to verify that no other Backup frame has selected
;;;                          unit in its TAPE-UNIT variable.
;;;
;;;   MAKE-BOOT-TAPE:        Redefined function.  Problem is that this function was
;;;                          advising MT:WRITE-EOF to do nothing.  If another process
;;;                          was backing-up files, we'd want MT:WRITE-EOF to do the write.
;;;                          A screwy design requires a screwy solution:  Locally bind
;;;                          a variable *NO-EOF-NEEDED* in MAKE-BOOT-TAPE, and declare
;;;                          it special; advise MT:WRITE-EOF, but with some smarts that
;;;                          if the *NO-EOF-NEEDED* variable is bound, don't write the EOF.
;;;                          This way, other processes that didn't call MAKE-BOOT-TAPE
;;;                          will get an EOF.
;;;
;;;  These seem to work with Unisys Release 3.0.1 and TI Release 3.0 - BAC

;;;  Known deficiencies:
;;;
;;;    When using using multiple Backup Frames, the exposed one will prevent the unexposed
;;;    ones from running.  However, if all are unexposed, all will run.
;;;
;;;    The function MULTI-TAPE doesn't support the partition operations.

(defun multi-tape (&key (interactive t)
		   (units nil)
		   (pre-operation 'REWIND)
		   (operation 'NONE)
		   (pathname nil)
		   (post-operation 'REWIND))
  
  "Perform tape operations using multiple tape units.  The operations
are run in separate processes, one for each selected tape unit.
The following KEYWORD arguments are supported:
INTERACTIVE:      T means pop up a CVV to make tape choices.
UNITS:            If non-nil, must be a list of valid, on-line tape units.
PRE-OPERATION:    One of 'REWIND or 'REWIND-ERASE.
OPERATION:        One of 'LIST-CONTENTS 'BACKUP 'RESTORE 'VERIFY 'REWIND-ERASE
PATHNAME:         Pathname to perform the operation on.
POST-OPERATION:   one of 'REWIND 'REWIND-VERIFY-REWIND"

  (declare (special units pre-operation operation pathname post-operation))
  (let*
    ((valid-pre-operations '(REWIND REWIND-ERASE NONE ))
     (valid-operations '(NONE LIST-CONTENTS BACKUP RESTORE VERIFY REWIND-ERASE))
     (valid-post-operations '(REWIND REWIND-VERIFY-REWIND NONE ))
     (all-process-names (mapcar 'process-name system:all-processes))
     (all-available-units (loop for u in (si:all-tape-units)
				when (not (member (format nil "MT~D" u) all-process-names))
				collect u)))
    (si:initialize-disk-system)                      ; Try to make all units available
    (if (not (si:all-tape-units))                    ; Any tape units on-line?
	(format t "~%No tape units online.")         ; - No,  say so!
	(if (not all-available-units)                ; -- Yes, any tape units free?
	    (format t "~%All tape units in use.")    ; --- No, all units in use.
	    (progn                                   ; - Yes, proceed...
	      (when interactive
		(when (and (= (length all-available-units) 1) (null units))
		  (setq units (list (car  all-available-units))))
		(tv:choose-variable-values
		  `((units "Unit"
			   :documentation "Select tape drive(s)."
			   :set ,all-available-units)
		    (pre-operation "Before action"
				   :documentation "Action to take before the main operation."
				   :choose ,valid-pre-operations)
		    (operation "Operation"
			       :documentation "Main action to perform."
			       :choose ,valid-operations)
		    (pathname "Pathname"
			      :documentation "Name of files to perform operations on."
			      :pathname)
		    (post-operation "After action"
				    :documentation "Action to take after the main operation."
				    :choose ,valid-post-operations))
		  :label "Select a Multi-tape Operation"
		  :margin-choices '("Do It" ("Abort" (signal-condition eh:abort-object)))))
	      
	      (dolist (drive units)                             ; Start a process for each selected drive...
		(if (not (member drive all-available-units))	; Already a process for this tape unit?
		    (format t "~%A process <MT~D> already exists for unit ~D." drive drive)
		    (process-run-function
		      (format nil "MT~D" drive)                 ; Name the process MTn where n = drive number
		      'mt:multi-tape-internal
		      drive pre-operation operation pathname post-operation)))
	      T)))))

;;; The following function binds global variables locally so that the local ones will
;;; be used instead of globals.  This is pretty bad, but it works and saves a structural
;;; redesign of the tape system.  A better solution would be to pass these values as
;;; keywords.  This might work if everyone allowed other keys...

(defun multi-tape-internal (unit pre-operation operation pathname post-operation)
  "Internal function to perform multi-tape operations.  Should be run in a 
separate process.  Called by MT:MULTI-TAPE."
  (let
    ((mt:*nupi-default-unit* unit)                 ; Bind global variables locally so local ones will
     (mt:*current-unit* unit)                      ; be used instead of globals.
     (mt:file-buf nil)                             ; From MT:SOURCE-COMPARE-STREAMS and MT:INITIALIZE-SOURCE-COMPARE-BUFFERS
     (mt:tape-buf nil)                             ; From MT:SOURCE-COMPARE-STREAMS and MT:INITIALIZE-SOURCE-COMPARE-BUFFERS
     (mt:file-buf-16 nil)                          ; From MT:SOURCE-COMPARE-STREAMS and MT:INITIALIZE-SOURCE-COMPARE-BUFFERS
     (mt:tape-buf-16 nil))                         ; From MT:SOURCE-COMPARE-STREAMS and MT:INITIALIZE-SOURCE-COMPARE-BUFFERS
    (mt:initialize-tape-drive :unit unit)
    (mt:setup-quarter-inch-tape :unit unit)
    (when pre-operation
      (tv:notify nil "MT~D: Performing ~A pre-operation..." unit pre-operation))
    (case pre-operation
	  ('rewind (mt:rewind unit))
	  ('rewind-erase (mt:rewind unit)
			 (mt:erase unit)))
    (when operation
      (tv:notify nil "MT~D: Performing ~A operation..." unit operation))
    (case operation
	  ('list-contents (mt:list-contents))
	  ('backup (mt:backup pathname :VERBOSE NIL))
	  ('restore (mt:restore pathname :VERBOSE NIL))
	  ('verify (if (mt:verify pathname nil :VERBOSE NIL)
				     (tv:notify nil "MT~D: Verification successfull." unit)
				     (tv:notify nil "MT~D: **** Verfication error." unit)))
	  ('rewind-erase (mt:rewind unit)
			 (mt:erase unit)))
    (when post-operation
      (tv:notify nil "MT~D: Performing ~A post-operation..." unit post-operation))
    (case post-operation
	  ('rewind (mt:rewind unit))
	  ('rewind-verify-rewind (mt:rewind unit)
				 (if (mt:verify pathname nil :VERBOSE NIL)
				     (tv:notify nil "MT~D: Verification successfull." unit)
				     (tv:notify nil "MT~D: **** Verfication error." unit))
				 (mt:rewind unit)))
    (tv:notify nil "MT~D: Completed." unit))
  )

;;; Thought this was needed but turns out it's already a patch!
;;; (defun init-unit-var (unit)
;;;   (setf mt:*current-unit* unit)
;;;   (setf mt:*nupi-default-unit* unit))


;;; Redefined from STREAMER-TAPE;DISPLAY.LISP

(DEFFLAVOR BACKUP-FRAME
	   (OUT-STREAM
	    (tape-unit NIL))                              ; BAC - add instance variable
	   (TV:INFERIORS-NOT-IN-SELECT-MENU-MIXIN
	    ;;tv:full-screen-hack-mixin
	    TV:PROCESS-MIXIN;Process-Mixin should come before Select-Mixin (See doc for P-M)
	    TV:SELECT-MIXIN
	    TV:CENTERED-LABEL-MIXIN
	    TV:TOP-LABEL-MIXIN
	    TV:LABEL-MIXIN
	    TV::ESSENTIAL-MOUSE
	    TV:WINDOW-WITH-TYPEOUT-MIXIN
	    TV:STREAM-MIXIN
	    TAPE-MIXIN
	    TV:BORDERED-CONSTRAINT-FRAME-WITH-SHARED-IO-BUFFER)
  (:SETTABLE-INSTANCE-VARIABLES OUT-STREAM tape-unit)
  (:GETTABLE-INSTANCE-VARIABLES OUT-STREAM tape-unit)
  (:DEFAULT-INIT-PLIST :PROCESS '(TOP-LEVEL;Function to be run by the process
				  :REGULAR-PDL-SIZE 3000);Ensure the PDL stack isn't overflown
   :SAVE-BITS :DELAYED;Don't create a bit-save array until we need it
   :LABEL (LIST ':STRING "Backup System"))
  (:DOCUMENTATION "Window interface to the magtape Backup System."))

(DEFUN COMMAND-LOOP (FRAME)
  (LET* ((TEXT-PANE (SEND FRAME :GET-PANE 'TEXT-PANE))
	 (BLINKY (CAR (SEND TEXT-PANE :BLINKER-LIST)))
	 (*TERMINAL-IO* TEXT-PANE)
	 (LOCAL-OUT-STREAM TEXT-PANE)
    	                                                          ; BAC - Define variables locally so that
	 (mt:*first-time-through* nil)                            ; BAC - these will be used instead of the
	 (tape-unit (send frame :tape-unit))                      ; BAC - global bindings.
	 (mt:*nupi-default-unit* tape-unit)                       ; BAC
	 (mt:*current-unit* tape-unit)                            ; BAC
	 (mt:file-buf nil)                                        ; BAC
	 (mt:tape-buf nil)                                        ; BAC
	 (mt:file-buf-16 nil)                                     ; BAC
	 (mt:tape-buf-16 nil))                                    ; BAC
    (SEND FRAME :SET-OUT-STREAM LOCAL-OUT-STREAM)
    (SEND BLINKY :SET-VISIBILITY :BLINK)
    (FORMAT LOCAL-OUT-STREAM "~&");this gives a new line after ABORTing out of error
    (CATCH :EXIT
      (LOOP FOR INPUT = (SEND FRAME :ANY-TYI) DO
	 (COND
	   ((AND (CONSP INPUT) (NOT (EQUALP (CAR INPUT) :MOUSE-BUTTON)))
	    (LET ((VAL (AND (CONSP INPUT) (SECOND INPUT)))
		  (*print-BASE* 10)
		  (*read-BASE* 10))
	      (SETQ VAL (GET VAL :VALUE))
	      (SEND BLINKY :SET-VISIBILITY :BLINK)
	      (CASE (SEND FRAME :SEND-IF-HANDLES VAL)
		(:ABORT (FORMAT LOCAL-OUT-STREAM "~&~A aborted.~%" VAL))
		(:COMPLETE (FORMAT LOCAL-OUT-STREAM " ~A complete.~%" VAL))
		(:MUTE NIL)
		(NIL (FORMAT LOCAL-OUT-STREAM "~&Command ~A not yet handled.~%" VAL)))))
	   ((EQl INPUT (char-code #\END)) (THROW :EXIT
				    :EXIT-FROM-KEYBOARD))
	   ((EQl INPUT (char-code #\HELP));need any others ???
	    (SEND FRAME :HELP))
	   ((EQl INPUT (char-code #\clear-screen))  ;need any others ???
	    (SEND TEXT-PANE :CLEAR-SCREEN))
	   (T T))
	 (SETQ INPUT ()) UNTIL ()))))

(DEFMETHOD (tape-mixin :prepare-tape) ()
  "Prepare tape."
  (let ((fini :ABORTED)
	(unit (pick-drive))
	(block-mode :fixed)
	(tape-type :quarter-inch)
	(block-size 512.)
	(density 0))
    (if unit
	(if (and (not (EQ unit (send self :tape-unit))) (check-tape-assigned unit))	; BAC - check for other tapes
	    (format (send self ':out-stream) "~&Tape Unit ~D already assigned to another Backup System window." unit)	; BAC
	    (let ((fun (w:menu-choose initialize-format-alist :label "Choose format"
				      :default-item (first initialize-format-alist))))
	      (send self :set-tape-unit unit)	; BAC - assign this frame
	      (send self :set-label (format nil "Backup System for Tape Unit ~D" unit))	; BAC - and change the label
	      (selector fun equal
		(('setup-half-inch-tape )
		 (setq tape-type :half-inch)
		 (catch 'abort-block	   
		   (tv:choose-variable-values '((*temp-block-mode* "Block Mode" :choose (:Variable :Fixed))
						(*temp-density* "Density" :choose (:single :double)))
					      :MARGIN-CHOICES '("Do it"
								("Abort" (THROW 'abort-block :ABORTED)))
					      :LABEL "Select block mode and density" :EXTRA-WIDTH 25)
		   (setq block-mode *temp-block-mode*))
		 (if (eq  *temp-density* :single)
		     (setq density 2.)
		     (setq density 6.))
		 (if (eq block-mode :variable)
		     (setq block-size 0.)
		     (progn
		       (catch 'block
			 (tv:choose-variable-values '((*temp-block-size* "Block size" :choose (512. 1024.))) 
						    :margin-choices '(("Do it" (and (setf outok t)
										    (throw 'block nil)))
								      "Abort")
						    :label "Enter a value for the block size"))
		       (setq block-size *temp-block-size*)))
		 (prepare-tape-drive unit :block-size block-size :block-mode block-mode	;hw
				     :tape-type tape-type :density density)
		 
		 (setq fini :COMPLETE)
		 (send (send self :get-pane 'command-pane)
		       :set-label `(:FONT ,*LABEL-FONT* :STRING "Half Inch Tape Menu"))
		 (send (send self :get-pane 'command-pane) :set-item-list  half-inch-command-menu-list)
		 )
		(('setup-quarter-inch-tape)
		 (prepare-tape-drive unit)	;hw
		 (setq fini :COMPLETE)
		 (send (send self :get-pane 'command-pane)
		       :set-label `(:FONT ,*LABEL-FONT* :STRING "Quarter Inch Tape Menu"))
		 (send (send self :get-pane 'command-pane) :set-item-list
		       quarter-inch-command-menu-list)
		 )
		(('carry-tape)
		 (prepare-tape-drive unit)	;hw
		 (setq fini :COMPLETE)
		 (send (send self :get-pane 'command-pane)
		       :set-label `(:FONT ,*LABEL-FONT* :STRING "Carry Tape Menu"))
		 (send (send self :get-pane 'command-pane) :set-item-list carry-tape-command-menu-list)
		 )		    
		))))	      
    fini))

;;; New function to check it there is already a backup frame assigned to this tape unit.
;;; The sort of kludgee way to do this is to append the tape unit to the backup window label
;;; when it is assigned.

(defun check-tape-assigned (unit)
  "Returns T if tape drive is already assigned to a backup window, or NIL if drive is unassigned."
  (let
    ((backup-units nil))
    (dolist (x (SEND tv:main-screen :INFERIORS))                            ; For all Backup frames...
      (WHEN (EQUAL 'mt:backup-frame
		   (SI:FLAVOR-NAME (SYSTEM-INTERNALS:INSTANCE-FLAVOR x)))
	(LET
	  ((drive (send x :tape-unit)))
	  (when drive
	    (setq backup-units (cons drive backup-units))))))               ; make a list of the units they apply to
    (if (member unit backup-units)                                          ; Is this unit in the list?
	T                                                                   ; - Yes, drive is already assigned
	NIL)))                                                              ; - No, drive is available.

;;; Redefined from STREAMER-TAPE;BOOTABLE.LISP
;;; As if the global variables problem wasn't enough,
;;; this is the worst: advising MT:WRITE-EOF on the fly.

(DEFUN MAKE-BOOT-TAPE (PARTITION-LIST &OPTIONAL (UNIT SI::*DEFAULT-DISK-UNIT*)
                                (OUT-STREAM *TERMINAL-IO*))
  "Makes a bootable tape. Accepts a list of lists and an optional default disk unit.
 Each of the lists specifies a partition name and an optional disk unit.
 If the disk unit is not specified with the partition name, the default unit is used.
 Creates a label and partition table on the tape and copies each partition in the
 list from the specified disk to the tape." ; Ex: (si:make-boot-tape '(("mcr1" 1) ("mcr2") ("mcr3")) 0)
  (LET
    ((COMPILE-ENCAPSULATIONS-FLAG T))
    (DECLARE (SPECIAL COMPILE-ENCAPSULATIONS-FLAG))  
       (UNWIND-PROTECT
	   (LET                                                       ; BAC - from streamer-tape;bootable.lisp
	     ((*no-eof-needed* T))                                    ; BAC - Bind a variable so that WRITE-EOF
	     (declare (special *no-eof-needed*))                      ; BAC - can check it in this stack-group
	     (advise mt:write-eof :around nil 0                       ; BAC - There's got to be a better way...
	       (local-declare ((special *no-eof-needed*))             ; BAC - Check the *no-eof-needed* variable.
		 (if (VARIABLE-BOUNDP *no-eof-needed*)                ; BAC - Is it bound?
		     (print "Partition copied to bootable tape")      ; BAC - - Yes, we're making a bootable tape
		     :do-it)))                                        ; BAC - - No,  a files tape so call WRITE-EOF
	     (SETQ *FIRST-TIME-THROUGH* (CONS UNIT PARTITION-LIST))
	     ;I know, this is absolutely AWFUL!!!
	     (DOLIST (PART PARTITION-LIST)	;let part be next element in list
	       (MT:BACKUP-PARTITION
		 (IF (> (LENGTH PART) 1)
		     (CADR PART) UNIT)
		 (CAR PART) 0 NIL :STYLE :BOOT)	
	       (SETQ *FIRST-TIME-THROUGH* NIL)))
	 (UNADVISE MT:WRITE-EOF :AROUND 0)
	 )))
