;;;-*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;;  7/06/87    TWE	Fixed the distance-from-point method to work properly for lines
;;;			with arbitrary slope.  Also moved some common code from LINE.
;;;  5/19/87    TWE	Initial creation.  Moved from LINE file because of methods which
;;;			need to be shared between lines and vectors.

(DEFFLAVOR BASIC-LINE-OBJECT
  ((DELTA-X   NIL :TYPE INTEGER-OR-NIL)
   (DELTA-Y   NIL :TYPE INTEGER-OR-NIL)
   (LENGTH    NIL :TYPE FLOAT-OR-NIL)
   (THICKNESS NIL :TYPE INTEGER-OR-NIL))
  (BASIC-OBJECT)
  :SETTABLE-INSTANCE-VARIABLES
  :INITTABLE-INSTANCE-VARIABLES
  :GETTABLE-INSTANCE-VARIABLES)

;;; The following two functions were separated out from the main place they were
;;; used so that we can calculate the location of the other endpoint when we
;;; need to determine the maximum dimensions of this line.
(DEFUN CALCULATE-X1 (X0 DELTA-X LENGTH)
  (IF (ZEROP DELTA-X)
      X0
      ;;1ELSE*
      (+ X0
         (IF (MINUSP DELTA-X) (- LENGTH) LENGTH))))

(DEFUN CALCULATE-Y1 (Y0 DELTA-X DELTA-Y LENGTH)
  (- Y0 (IF (ZEROP DELTA-X)
            (IF (PLUSP DELTA-Y) LENGTH (- LENGTH))
            ;;1ELSE*
            (TRUNCATE (* LENGTH
                         (/ (FLOAT DELTA-Y) (ABS DELTA-X)))))))

;;; The primary idea of the following function is to compute the endpoint of a
;;; line specified by the user such that its length looks reasonable.  The
;;; secondary idea is to compute the distance from a point to a line.  The
;;; first (simplest) approach was to simply use the value of the X argument to
;;; determine the length.  This approach, when one moves towards the Y axis,
;;; makes the length become too short to be reasonable.  This approach is to
;;; find a line length that is independent of the movement towards any axis.
;;;
;;; The algorithm was derived as follows:
;;; let SLOPE=DELTA-Y / DELTA-X
;;; let f(x) = y = SLOPE * x + b be the line which contains the ray defined by
;;;                              (X0,Y0) and DELTA-X DELTA-Y
;;; then f1(x) = -(1 / SLOPE) * x + b1 defines a line which is perpendicular
;;;                                    to f.  We choose a value for b1 such
;;;                                    that f1 contains the point (X,Y).
;;;                                    Since we already know an instance of
;;;                                    the x, f1(x), and SLOPE components of
;;;                                    f1, solving for b1 is easy.
;;; Y = -(1 / SLOPE) * X + b1
;;; b1 = Y + (1 / SLOPE) * X
;;; We can now define f1 in terms of this value for b1:
;;; f1(x)= -(1 / SLOPE) * x + Y + (1 / SLOPE) * X
;;;
;;; The point on f which is closest to (X,Y) is where f and f1 intersect.
;;; Equating f and f1 will yield this point.
;;; SLOPE * X2 + b = -(1 / SLOPE) * X2 + Y + (1 / SLOPE) * X
;;; 
;;;      Y + (1 / SLOPE) * X - b   Y * SLOPE + X - b * SLOPE
;;; X2 = ----------------------- = -------------------------
;;;        SLOPE + (1 / SLOPE)           SLOPE**2 + 1
;;; 

;;; For the BEST-LINE-LENGTH function we don't need to calculate the Y2
;;; component.  For the :DISTANCE-FROM-POINT method we need to have the Y2
;;; component as well.  All we need to do is to substitute the value of X2 into
;;; either f1 or f to get:

;;;
;;;      Y * SLOPE + X - b * SLOPE
;;; Y2 = ------------------------- * SLOPE + b
;;;            SLOPE**2 + 1
;;; 
;;; All that remains is to calculate the value of b, the Y intercept for f.
;;; Since the point (X0,Y0) is on the line defined by f, we can solve for
;;; b by using the definition of f.
;;; b = Y0 - SLOPE * X0
;;;
;;; For the :DISTANCE-FROM-POINT method we now have the two points (X,Y) and
;;; (X2,Y2) so we can calculate the smallest distance between (X,Y) and the
;;; line easily.


(DEFUN BEST-LINE-LENGTH (X0 Y0 DELTA-X DELTA-Y X Y)
  "2Compute the length of a line segment starting at (X0,Y0) and ending at a point (X1,Y1) such that
the slope of the line segment is DELTA-Y / DELTA-X and (X1,Y1) is the point nearest to the point (X,Y).*"
  (LET* ((SLOPE (/ DELTA-Y DELTA-X))
         (INTERCEPT (- Y0 (* SLOPE X0)))
         (X1 (/ (+ (* Y SLOPE) X (- (* INTERCEPT SLOPE)))
                (1+ (* SLOPE SLOPE)))))
    ;; For LaTeX, the line length is the length of the X component.
    (ABS (- X1 X0))))


(DEFMETHOD (BASIC-LINE-OBJECT :TRANSLATE-UNITS) (NEW-OVER-OLD X-ORIGIN Y-ORIGIN)
  (SETQ LENGTH     (* NEW-OVER-OLD LENGTH)
        X-POSITION (+ (* NEW-OVER-OLD (- X-POSITION X-ORIGIN)) X-ORIGIN)
        Y-POSITION (+ (* NEW-OVER-OLD (- Y-POSITION Y-ORIGIN)) Y-ORIGIN)))

(DEFMETHOD (BASIC-LINE-OBJECT :DISTANCE-FROM-POINT) (WINDOW X Y)
  "Returns the smallest distance that this object is from the point (X,Y).
X and Y are measured in terms of the current unit."
  (DECLARE (IGNORE WINDOW))
  ;; Compute the length of a line segment starting at (X-POSITION,Y-POSITION)
  ;; and ending at a point (X1,Y1) such that the slope of the line segment is
  ;; DELTA-Y / DELTA-X and (X1,Y1) is the point nearest to the point (X,Y).
  (LET (SLOPE-DIRECTION
        END-X-POSITION
        END-Y-POSITION)
    (COND ((ZEROP DELTA-X)
           (SETQ SLOPE-DIRECTION (SIGN DELTA-Y)
                 END-X-POSITION X-POSITION 
                 END-Y-POSITION (+ Y-POSITION (* SLOPE-DIRECTION LENGTH)))
           ;; This is a vertical line.  The distance between any point and this line is
           ;; simply the differences between the X components.  However, this isn't really
           ;; a line but is just a line segment.  For those points which are to the
           ;; immediate left or right of this line segment, the distance is easy.  For the
           ;; other points, the distance formula must be applied between (X,Y) and the
           ;; appropriate endpoint.
           (IF (OR (<= Y-POSITION Y END-Y-POSITION)
                   (>= Y-POSITION Y END-Y-POSITION))
               (ABS (- X-POSITION X))
               ;;ELSE
               (IF (OR (AND (PLUSP  SLOPE-DIRECTION) (< Y Y-POSITION))
                       (AND (MINUSP SLOPE-DIRECTION) (> Y Y-POSITION)))
                   ;; (X,Y) is below or above the line segment, respectively.
                   (W:DIST X Y X-POSITION Y-POSITION)
                   ;;ELSE
                   (W:DIST X Y X-POSITION END-Y-POSITION))))
          ((ZEROP DELTA-Y)
           ;;Horizontal line.  Very similar to the vertical case.
           (SETQ SLOPE-DIRECTION (SIGN DELTA-X)
                 END-X-POSITION (+ X-POSITION (* SLOPE-DIRECTION LENGTH))
                 END-Y-POSITION Y-POSITION)
           ;; This is a horizontal line.  This is very similar to the vertical line case.
           (IF (OR (<= X-POSITION X END-X-POSITION)
                   (>= X-POSITION X END-X-POSITION))
               (ABS (- Y-POSITION Y))
               ;;ELSE
               (IF (OR (AND (PLUSP  SLOPE-DIRECTION) (< X X-POSITION))
                       (AND (MINUSP SLOPE-DIRECTION) (> X X-POSITION)))
                   ;; (X,Y) is to the left or right of the line segment, respectively.
                   (W:DIST X Y X-POSITION Y-POSITION)
                   ;;ELSE
                   (W:DIST X Y END-X-POSITION Y-POSITION))))
          (T
           (SETQ END-X-POSITION (+ X-POSITION (* (SIGN DELTA-X) LENGTH))
                 END-Y-POSITION (+ Y-POSITION (* (SIGN DELTA-Y) (ABS (/ DELTA-Y DELTA-X)) LENGTH)))
           ;; This is the complex case.  We have an arbitrarily (well, almost)
           ;; sloped line segment.  For an explanation of the following
           ;; calculation, see the comments before the definition of the
           ;; BEST-LINE-LENGTH function.  The idea is to first determine if the point (X,Y)
           ;; is closer to some point on the line segment or if it is closer to one of the
           ;; endpoints.  If this were a line instead of a line segment, then we would just
           ;; need to compute the distance from the point (X2,Y2) and (X,Y).  But this is
           ;; not a line, it is a line segment,  We need to determine if (X2,Y2) is within
           ;; the line segment before we can do that kind of distance calculation.
           (LET* ((SLOPE (/ DELTA-Y DELTA-X))
                  (INTERCEPT (- Y-POSITION (* SLOPE X-POSITION)))
                  (X2 (/ (+ (* Y SLOPE) X (- (* INTERCEPT SLOPE)))
                         (1+ (* SLOPE SLOPE))))
                  (Y2 (+ (* SLOPE X2) INTERCEPT)))
             (IF (AND (FUNCALL (IF (PLUSP DELTA-X) #'<= #'>=) X-POSITION X2 END-X-POSITION)
                      (FUNCALL (IF (PLUSP DELTA-Y) #'<= #'>=) Y-POSITION Y2 END-Y-POSITION))
                 (W:DIST X Y X2 Y2)
                 ;;1ELSE*
                 ;; At this point we know that the point (X,Y) is nearest to one
                 ;; of the endpoints.  We could try to figure out which endpoint
                 ;; it is nearest, but we don't really need to do that.  At the
                 ;; expense of some extra calculations, all we need to do is to
                 ;; compute the distances between (X,Y) and each of the endpoints.
                 ;; The smallest distance is the correct one.
                 (MIN (W:DIST X Y X-POSITION Y-POSITION)
                      (W:DIST X Y END-X-POSITION END-Y-POSITION))))))))

(DEFMETHOD (BASIC-LINE-OBJECT :OBJECT-LIMITS) (WINDOW)
  "Returns the minimum and maximum limits of this object.
The values returned are the left top right and bottom unit values."
  (DECLARE (IGNORE WINDOW))
  ;; Note that the (X,Y) position of the box is the upper left corner.
  ;; This makes the maximum Y value of the box the Y-POSITION.
  (LET ((X1 (CALCULATE-X1 X-POSITION DELTA-X LENGTH))
        ;; Note that we need to calculate the Y1 value differently since the
        ;; calculate-y1 function calculates screen coordinates, not LaTeX coordinates.
        (Y1 (+ Y-POSITION (IF (ZEROP DELTA-X)
                              (IF (PLUSP DELTA-Y) LENGTH (- LENGTH))
                              ;;1ELSE*
                              (TRUNCATE (* LENGTH (/ (FLOAT DELTA-Y) (ABS DELTA-X))))))))
    (VALUES (MIN X-POSITION X1) (MAX Y-POSITION Y1)
            (MAX X-POSITION X1) (MIN Y-POSITION Y1))))
