;;; -*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;;  1/03/89    TWE	Changed edit-object to pop up a menu when the object being picked
;;;			is not unique.  Also changed this function to look like
;;;			move-object.
;;; 10/14/88    TWE	Put support in place for read-only mode.
;;;  6/08/88    TWE	Changed the `no objects' message to correct the grammar.
;;;  8/31/87    TWE	Fixed up distance-from-box to handle interior points differently.
;;;			Previously, it returned 0 for such points.  Now it returns the
;;;			distance from the point to the edges of the box.  This makes
;;;			picking much more acceptable, particularly for nested objects.
;;;  7/17/87    TWE	Fixed up so that the user can see the coordinate being specified
;;;			when a single point is specified.
;;;  5/21/87	TWE	Initial creation.  Moved from BASE.

#|

This file defines the command that allows one to edit an object.

|#

(DEFUN DISTANCE-FROM-BOX (X Y LEFT TOP RIGHT BOTTOM)
  "Compute the minimum distance between the point (X,Y) and the box defined
by LEFT TOP RIGHT BOTTOM."
  ;; Note that SECTOR-CODE assumes that we are using the fourth quadrant, but
  ;; we are using the first quadrant.  The following weird translation swaps
  ;; the top row of sector codes with the bottom row.
  (LET ((SECTOR-CODE (W:SECTOR-CODE X Y LEFT BOTTOM RIGHT TOP)))
    (WHEN (NOT (ZEROP (LDB (BYTE 2 2) SECTOR-CODE)))
      (SETF (LDB (BYTE 2 2) SECTOR-CODE) (LOGXOR #o3 (LDB (BYTE 2 2) SECTOR-CODE))))
    (CASE SECTOR-CODE
      (10                              ; Upper right
       (W:DIST X Y RIGHT TOP))
      (9                               ; Upper left
       (W:DIST X Y LEFT TOP))
      (8                               ; Upper middle
       (- Y TOP))
      (6                               ; Lower right
       (W:DIST X Y RIGHT BOTTOM))
      (5                               ; Lower left
       (W:DIST X Y LEFT BOTTOM))
      (4                               ; Lower middle
       (- BOTTOM Y))
      (2                               ; Middle right
       (- X RIGHT))
      (1                               ; Middle left
       (- LEFT X))
      (0                               ; Center
       (MIN (- X LEFT)   (- RIGHT X)
            (- Y BOTTOM) (- TOP Y))))))

(DEFCONSTANT EDIT-OBJECT-MOUSE-DOCUMENTATION
             '(:MOUSE-L-1 "Pick an object to edit"
               :MOUSE-M-1 "Abort out of this operation"))


(DEFUN EDIT-OBJECT ()
  "Handles all interaction for editing objects."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (LOOP WITH X = NIL
            WITH Y = NIL
            DO (PROGN
                 (WITH-GRID-DISABLED
                   (MULTIPLE-VALUE-SETQ (X Y)
                     (MOUSE-SPECIFY-POINT NIL W:MOUSE-GLYPH-CIRCLE-X
                                          EDIT-OBJECT-MOUSE-DOCUMENTATION NIL)))
                 (WHEN (NULL X)
                   (RETURN NIL))
                 (SETQ X (TRANSLATE-PIXEL-TO-UNIT X)
                       Y (TRANSLATE-PIXEL-TO-UNIT Y))
                 (LET ((OBJECT-TO-EDIT (LOOP FOR OBJECT IN DISPLAY-LIST
                                             WITH OBJECT-DISTANCE = NIL
                                             WITH NEAREST-OBJECTS = NIL
                                             WITH NEAREST-DISTANCE = MOST-POSITIVE-SHORT-FLOAT
                                             WHEN (NOT (EQ OBJECT MARGIN-OBJECT))
                                             DO (PROGN
                                                  (SETQ OBJECT-DISTANCE (SEND OBJECT
                                                                              :DISTANCE-FROM-POINT
                                                                              SELF X Y))
                                                  (COND ((< OBJECT-DISTANCE NEAREST-DISTANCE)
                                                         (SETQ NEAREST-DISTANCE OBJECT-DISTANCE
                                                               NEAREST-OBJECTS `(,OBJECT)))
                                                        ((= OBJECT-DISTANCE NEAREST-DISTANCE)
                                                         (PUSH OBJECT NEAREST-OBJECTS))))
                                             FINALLY (RETURN NEAREST-OBJECTS)))
                       (INTERACTION-PANE (GET-INTERACTION-PANE SELF)))
                   ;; Identify the object being edited.
                   (IF (NULL OBJECT-TO-EDIT)
                       ;; There aren't any object.  Nice try.
                       (SEND INTERACTION-PANE :LINE-OUT "There are no objects to edit.")
                       ;;1ELSE*
                       (PROGN
                         (SETQ OBJECT-TO-EDIT (IF (> (LENGTH OBJECT-TO-EDIT) 1)
                                                  (W:MENU-CHOOSE
                                                    (LOOP FOR OBJECT IN OBJECT-TO-EDIT
                                                          COLLECT `(,(REMOVE-RETURNS
                                                                       (SEND OBJECT :NAME)
                                                                       #\SPACE)
                                                                    :VALUE ,OBJECT)))
                                                  (CAR OBJECT-TO-EDIT)))
                         (WHEN OBJECT-TO-EDIT
                           (SEND INTERACTION-PANE :STRING-OUT "Editing ")
                           (SEND INTERACTION-PANE :LINE-OUT (SEND OBJECT-TO-EDIT :NAME))
                           (SETQ CURRENT-MOUSE-BLINKER W:MOUSE-GLYPH-CIRCLE-X
                                 CURRENT-MOUSE-DOCUMENTATION EDIT-OBJECT-MOUSE-DOCUMENTATION)
                           (FUNCALL (SYMBOL-FUNCTION (SEND OBJECT-TO-EDIT :EDIT-FUNCTION))
                                    OBJECT-TO-EDIT)
                           (RECALCULATE-BOUNDS)))))
                 (SEND SELF :MOUSE-STANDARD-BLINKER))))
  (SETQ CURRENT-MOUSE-BLINKER       NIL
        CURRENT-MOUSE-DOCUMENTATION NIL))
