;;; -*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 10/14/88    TWE	Put support in place for read-only mode.
;;;  8/17/88    TWE	Fixed up grid drawing to take the margins into account,
;;;			especially when they are not the size of the entire picture
;;;			window.
;;;  9/11/87    TWE	Put the unit in the CVV window for new grid values.
;;;  6/29/87    TWE	Fixed up the argument VARIABLE to be CVV-VARIABLE to get around
;;;			a compiler warning message.
;;;  6/18/87    TWE	Added the GRID-STATE-TOGGLE function for GED compatibility.
;;;  6/02/87    TWE	Implemented the macro WITH-GRID-DISABLED to make object selection
;;;			work better.
;;;  5/22/87    TWE	Put in error checking for the CVV code.
;;;  5/20/87    TWE	Added a way to specify new grid values.
;;;  5/19/86	TWE	Initial creation.

#|

This file implements the grid mechanism which allows one to specify points of an objects
in such a way that they will join properly.

|#

(DEFMACRO WITH-GRID-DISABLED (&body body)
  `(LET ((.OLD-GRID-STATUS. GRID-STATUS-P))
     (SETQ GRID-STATUS-P NIL)
     ,@body
     (SETQ GRID-STATUS-P .OLD-GRID-STATUS.)))

(DEFUN GRID-SPECIFY-VALUES ()
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (LET ((NEW-GRID-X-SIZE GRID-X-SIZE)
            (NEW-GRID-Y-SIZE GRID-Y-SIZE)
            (PICTURE-BASE-UNIT BASE-UNIT))
        (FLET ((VALIDATE-WIDTH (WINDOW CVV-VARIABLE OLD-VALUE NEW-VALUE)
                               (DECLARE (IGNORE WINDOW CVV-VARIABLE OLD-VALUE))
                               (VALIDATE-PAGE-WIDTH NEW-VALUE PICTURE-BASE-UNIT))
               (VALIDATE-HEIGHT (WINDOW CVV-VARIABLE OLD-VALUE NEW-VALUE)
                                (DECLARE (IGNORE WINDOW CVV-VARIABLE OLD-VALUE))
                                (VALIDATE-PAGE-HEIGHT NEW-VALUE PICTURE-BASE-UNIT)))
          (CONDITION-CASE ()
              (W:CHOOSE-VARIABLE-VALUES
               `((,(LOCF NEW-GRID-X-SIZE)  "Grid X size"  :CONSTRAINT ,#'VALIDATE-WIDTH
                  :DOCUMENTATION "Width of a grid box.")
                 (,(LOCF NEW-GRID-Y-SIZE) "Grid Y size" :CONSTRAINT ,#'VALIDATE-HEIGHT
                  :DOCUMENTATION "Height of a grid box."))
               :MARGIN-CHOICES '("Do It"
                                 ("Abort" (SIGNAL-CONDITION EH:ABORT-OBJECT)))
               :LABEL (FORMAT NIL "Enter values for the grid box size in ~a."
                              (UNIT-NAME-STRING BASE-UNIT))
               ;; We need the extra space to display error messages in case the user tries
               ;; bad values.
               :EXTRA-WIDTH 30.)
            (EH:ABORT (SETQ NEW-GRID-X-SIZE NIL))))
        (WHEN NEW-GRID-X-SIZE
          (WHEN GRID-STATUS-P
            ;; We are changing the grid size and the grid is ON.  We need
            ;; to erase the old grid and then draw the new one.
            (SETQ GRID-STATUS-P NIL)
            (SEND SELF :DRAW-GRID)
            (SETQ GRID-STATUS-P T))
          (SETQ GRID-X-SIZE (FLOAT NEW-GRID-X-SIZE)
                GRID-Y-SIZE (FLOAT NEW-GRID-Y-SIZE))
          (WHEN GRID-STATUS-P
            (SEND SELF :DRAW-GRID))))))

(DEFUN GRID-STATE-TOGGLE ()
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (SETQ GRID-STATUS-P (NULL GRID-STATUS-P))
  (SEND SELF :DRAW-GRID))

(DEFUN GRID-STATE-ON ()
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (PROGN
        (SETQ GRID-STATUS-P T)
        (SEND SELF :DRAW-GRID))))

(DEFUN GRID-STATE-OFF ()
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (PROGN
        (SETQ GRID-STATUS-P NIL)
        (SEND SELF :DRAW-GRID))))


(DEFUN NEAREST-MULTIPLE (X Y)
  "2Finds the nearest multiple of Y to X.*"
  (* Y (W:G-ROUND X Y))) 


(DEFMETHOD (PICTURE-WINDOW :DRAW-GRID) ()
  "2This method draws a grid of dots corresponding to the specified grid spacing.
The grid covers the entire window area on the screen.*"
  (LET* ((INSIDE-RIGHT (W:SHEET-INSIDE-RIGHT))
         (INSIDE-BOTTOM (1- (W:SHEET-INSIDE-BOTTOM)))
         (WINDOW-X-SPACE GRID-X-SIZE)
	 (WINDOW-Y-SPACE GRID-Y-SIZE)
	 (WINDOW-INITIAL-GRID-X (* (CEILING (* (UNITS-PER-INCH BASE-UNIT) LEFT-MARGIN)
                                            GRID-X-SIZE) GRID-X-SIZE))
	 (WINDOW-INITIAL-GRID-Y (* (FLOOR (* (UNITS-PER-INCH BASE-UNIT) TOP-MARGIN)
                                            GRID-Y-SIZE) GRID-Y-SIZE))
         (ALU (IF GRID-STATUS-P W:CHAR-ALUF W:ERASE-ALUF)))
    ;; Implementation note: the prepare-sheet needs to be done out to this level instead
    ;; of just being around the last LOOP located in this function because we need to
    ;; make sure that we have a screen array on which to call array-dimension.  If we
    ;; do not then we may have a situation where we don't have a screen array
    ;; temporarily and will error out.  This actually happens when we toggle
    ;; configurations.
    (W:PREPARE-SHEET (SELF)
      (LET ((GRID (W:MAKE-SHEET-BIT-ARRAY SELF (ARRAY-DIMENSION W:SCREEN-ARRAY 1)1 *1))
            (BITBLT-ALU (IF (SEND SELF :REVERSE-VIDEO-P)
                            (W:REVERSE-ALU ALU)
                            ALU)))
        (WHEN (AND (> (TRANSLATE-UNIT-TO-PIXEL WINDOW-X-SPACE) 3)
                   (> (TRANSLATE-UNIT-TO-PIXEL WINDOW-Y-SPACE) 3))
          (LOOP FOR I FROM WINDOW-INITIAL-GRID-X BY WINDOW-X-SPACE
                BELOW (* (UNITS-PER-INCH BASE-UNIT) RIGHT-MARGIN)
                DO (SETF (AREF GRID 0 (W:G-ROUND (TRANSLATE-UNIT-TO-PIXEL I))) 1))
          (LOOP FOR J FROM (* (MAX-PAGE-HEIGHT) (UNITS-PER-INCH BASE-UNIT)) BY WINDOW-Y-SPACE
                ABOVE WINDOW-INITIAL-GRID-Y
                DO (WHEN (< J (* (UNITS-PER-INCH BASE-UNIT) BOTTOM-MARGIN))
                       (BITBLT BITBLT-ALU INSIDE-RIGHT 1
                               GRID 0 0
                               (W:SHEET-SCREEN-ARRAY SELF) 0
                               (MIN INSIDE-BOTTOM (W:G-ROUND (TRANSLATE-UNIT-TO-PIXEL J)))))))
        (RETURN-ARRAY (PROG1
                        GRID
                        (SETQ GRID NIL)))))))

(DEFMETHOD (PICTURE-WINDOW :GRIDIFY-POINT) (X Y)
  "2Round a point to the nearest grid location.*"
  (IF GRID-STATUS-P
      (VALUES (TRANSLATE-UNIT-TO-PIXEL (NEAREST-MULTIPLE (TRANSLATE-PIXEL-TO-UNIT X)
                                                           GRID-X-SIZE))
              (TRANSLATE-UNIT-TO-PIXEL (NEAREST-MULTIPLE (TRANSLATE-PIXEL-TO-UNIT Y)
                                                           GRID-Y-SIZE)))
      ;;1ELSE*
      (VALUES X Y)))

(DEFMETHOD (PICTURE-WINDOW :GRIDIFY-MEASURE) (X Y)
  "2Round a linear unit of measure to the nearest grid location.*"
  (IF GRID-STATUS-P
      (VALUES (NEAREST-MULTIPLE X GRID-X-SIZE)
              (NEAREST-MULTIPLE Y GRID-Y-SIZE))
      ;;1ELSE*
      (VALUES X Y)))



