;;; -*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 10/14/88    TWE	Put support in place for read-only mode.
;;;  9/20/88    TWE	Fixed bug where changing the line thickness also changed the
;;;			thickness of the margin.
;;;  8/17/88    TWE	Fixed up the margin code clean out more of the margin information
;;;			when the margin is being turned off.
;;;  6/15/88    TWE	Write new code to allow one to turn off the margins.
;;; 10/01/87    TWE	Updated to reflect the adjust arguments to :TRANSLATE-UNITS.
;;;  6/29/87    TWE	Fixed up the argument VARIABLE to be CVV-VARIABLE to get around
;;;			a compiler warning message.
;;;  5/22/87    TWE	Fixed up the drawing of a margin when one is already present.
;;;			The fix was to erase the old margin.  Also put in error checking
;;;			for the CVV code.
;;;  5/05/86	TWE	Initial creation.  Moved code from BASE

#|

This file handles the margin setting for the user.  The margins show the user
what part of the page LaTeX will be using.

|#



;;; Define a new variable type for CVV which will print out
;;; a unit using PRINT-UNIT and get a new unit from a menu. 
(SETF (GET :UNIT 'TV:CHOOSE-VARIABLE-VALUES-KEYWORD)
      '(PRINT-UNIT NIL NIL
                   NIL GET-UNIT-MENU "Click left to input a new unit of measure for the margins."))

(DEFUN PRINT-UNIT (UNIT &OPTIONAL (STREAM T))
  (FORMAT STREAM "~A" (UNIT-NAME-STRING UNIT)))

(DEFUN SPECIFY-MARGINS ()
  "Handles all interaction for specifying the margins."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (LET ((MARGIN-BASE-UNIT MARGIN-UNIT))
        ;; The following local function is called whenever the margin-unit is changed.
        ;; When a unit changes, the values of the margins need to change accordingly,
        ;; and the window needs to be refreshed to reflect the new values.  We need to
        ;; make this function local so we can update the margin values.
        (FLET ((UNIT-CHANGE (WINDOW CVV-VARIABLE OLD-UNIT NEW-UNIT)
                            (DECLARE (IGNORE CVV-VARIABLE))
                            (WHEN (NOT (EQ OLD-UNIT NEW-UNIT))
                              (LET ((NEW-OVER-OLD (/ (UNITS-PER-INCH NEW-UNIT)
                                                     (UNITS-PER-INCH OLD-UNIT))))
                                (SETQ LEFT-MARGIN   (* LEFT-MARGIN   NEW-OVER-OLD)
                                      TOP-MARGIN    (* TOP-MARGIN    NEW-OVER-OLD)
                                      RIGHT-MARGIN  (* RIGHT-MARGIN  NEW-OVER-OLD)
                                      BOTTOM-MARGIN (* BOTTOM-MARGIN NEW-OVER-OLD)))
                              (SEND WINDOW :REFRESH)))
               (VALIDATE-WIDTH (WINDOW CVV-VARIABLE OLD-UNIT NEW-UNIT)
                               (DECLARE (IGNORE WINDOW CVV-VARIABLE OLD-UNIT))
                               (VALIDATE-PAGE-WIDTH NEW-UNIT MARGIN-BASE-UNIT))
               (VALIDATE-HEIGHT (WINDOW CVV-VARIABLE OLD-UNIT NEW-UNIT)
                                (DECLARE (IGNORE WINDOW CVV-VARIABLE OLD-UNIT))
                                (VALIDATE-PAGE-HEIGHT NEW-UNIT MARGIN-BASE-UNIT)))
          (WHEN GRID-STATUS-P
            ;; The grid is turned on.  We need to erase it before we deal with the margins.
            (SETQ GRID-STATUS-P NIL)
            (SEND SELF :DRAW-GRID)
            (SETQ GRID-STATUS-P T))
          (W:CHOOSE-VARIABLE-VALUES
            `((,(LOCF MARGIN-UNIT  ) "Unit for margin"
               :SIDE-EFFECT ,#'UNIT-CHANGE :UNIT)
              (,(LOCF LEFT-MARGIN  ) "Left margin"   :CONSTRAINT ,#'VALIDATE-WIDTH)
              (,(LOCF TOP-MARGIN   ) "Top margin"    :CONSTRAINT ,#'VALIDATE-HEIGHT)
              (,(LOCF RIGHT-MARGIN ) "Right margin"  :CONSTRAINT ,#'VALIDATE-WIDTH)
              (,(LOCF BOTTOM-MARGIN) "Bottom margin" :CONSTRAINT ,#'VALIDATE-HEIGHT))
            ;; We need the extra width in case we have to display an error message.
            :EXTRA-WIDTH 25.))

        (IF MARGIN-OBJECT
            ;; Erase the old margin before we draw the new one.
            (SEND MARGIN-OBJECT :DRAW SELF W:ALU-XOR)
            ;;1ELSE*
            (PROGN
              (SETQ MARGIN-OBJECT (MAKE-INSTANCE 'HOLLOW-BOX-OBJECT))
              (PUSH MARGIN-OBJECT DISPLAY-LIST)))
        (SEND MARGIN-OBJECT :SET-X-POSITION (FLOAT LEFT-MARGIN))
        (SEND MARGIN-OBJECT :SET-Y-POSITION (* (- 11.0 (/ TOP-MARGIN (UNITS-PER-INCH MARGIN-UNIT)))
                                               (UNITS-PER-INCH MARGIN-UNIT)))
        (SEND MARGIN-OBJECT :SET-WIDTH  (- (FLOAT RIGHT-MARGIN)  LEFT-MARGIN))
        (SEND MARGIN-OBJECT :SET-HEIGHT (- (FLOAT BOTTOM-MARGIN) TOP-MARGIN))
        (SEND MARGIN-OBJECT :SET-THICKNESS *THIN-LINES*)
        (SEND MARGIN-OBJECT :TRANSLATE-UNITS
              (/ (UNITS-PER-INCH BASE-UNIT) (UNITS-PER-INCH MARGIN-UNIT)) 0 0)
        (SEND MARGIN-OBJECT :DRAW SELF)
        (WHEN GRID-STATUS-P
          ;; The grid is turned on.  We need to draw it again so it appears
          ;; in the right place relative to the new margins.
          (SEND SELF :DRAW-GRID)))))

(DEFUN TURN-OFF-MARGINS ()
  "Gets rid of the margins when the user doesn't want them anymore."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (PROGN
        ;; Update the instance variables that indicate the margin size.
        (SETQ LEFT-MARGIN   0.0
              TOP-MARGIN    0.0
              RIGHT-MARGIN  (MAX-PAGE-WIDTH)
              BOTTOM-MARGIN (MAX-PAGE-HEIGHT))
        ;; Get rid of the margin-object too.
        (WHEN MARGIN-OBJECT
          (SETQ DISPLAY-LIST (REMOVE MARGIN-OBJECT DISPLAY-LIST))
          ;; Erase the old margin.
          (SEND MARGIN-OBJECT :DRAW SELF W:ALU-XOR)
          (SETQ MARGIN-OBJECT NIL))
        (WHEN GRID-STATUS-P
          ;; The grid is turned on.  We need to draw it again so it appears
          ;; over the entire picture window.
          (SEND SELF :DRAW-GRID)))))