;;; -*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 10/11/88    TWE	Added code to :WRITE-LATEX to not texify the text at the user's
;;;			request.
;;;  6/15/88    TWE	Fixed drawing to copy the object to the picture window when the
;;;			user has specified enough parameters (instead of XORing the
;;;			object and interfering with other objects).  This keeps the user
;;;			from having to refresh the screen so often.
;;;  6/08/88    TWE	Fixed DRAW-TEXT-STACK-OBJECT and :DISTANCE-FROM-POINT to handle
;;;			the case where the user removes the text from the object.  Also
;;;			changed DRAW-TEXT-STACK to ask the user if removing the text from
;;;			an old object was an attempt to delete the object.
;;;  1/05/88    TWE	Fixed a bug where a fonted text object was moved and the font
;;;			for the new object was changed back to the default font.
;;;  8/31/87    TWE	Created simple-distance-from-box which returns 0 when the point
;;;			is inside the box.  This differs from the distance-from-box which
;;;			assumes that distances are measured from the edges of the box.
;;;  8/24/87    TWE	Added centered and right justification for the text.
;;;  7/20/87    TWE	Update the :NAME method to reflect the shorter interaction pane.
;;;			Also fixed up coordinate calculation by computing the gridified
;;;			values like the box code does.  Moved remove-returns to
;;;			BASIC-BOX.
;;;  7/17/87    TWE	Fixed up so that the user can see the coordinate being specified
;;;			when a single point is specified.
;;;  6/29/87    TWE	Added the support for MINIMUM-X/Y to make the picture only as big
;;;			as the user makes it.
;;;  6/23/87    TWE	Added support for fonts.
;;;  6/02/87    TWE	Fixed aborting out while editing an object.
;;;  6/01/87    TWE	Partial Fix for a bug in move mode.  For in move mode, if one
;;;			aborts out of the move, the object's image gets erased from the
;;;			screen.  To fix this, all of the functions that can be invoked by
;;;			the :PICK-CORNER-FUNCTION method now return NIL for the abort
;;;			case and non-NIL for all other cases.
;;;  5/29/87    TWE	fixed up aborting out of the creation of a text stack object
;;;			so that the aborting would erase the image too.  Implemented menu
;;;			item sequencing.
;;;  5/26/87    TWE	Added more error checking when entering text.  It now checks to
;;;			make sure that the user has specified a corner first.  Fixed up
;;;			all of this code to have the Y-POSITION refer to the lower corner
;;;			of the text stack instead of the upper corner.  Removed all code
;;;			that did anything to the MAXIMUM-HEIGHT instance variable.
;;;  5/19/87    TWE	Implemented editing for text.
;;;  5/07/87	TWE	Initial creation.


#|

This file defines the code to handle text objects.

|#

(DEFFLAVOR TEXT-STACK-OBJECT ()
  (POSITION-MIXIN TEXT-MIXIN BASIC-OBJECT)
  :GETTABLE-INSTANCE-VARIABLES
  :SETTABLE-INSTANCE-VARIABLES
  :INITTABLE-INSTANCE-VARIABLES
  (:DEFAULT-INIT-PLIST :POSITION "l"))

(DEFMETHOD (TEXT-STACK-OBJECT :WRITE-LATEX) (LATEX-STREAM TEXIFY)
  "2Write the description of this object to a LaTeX file.*"
  (FORMAT LATEX-STREAM "\\put(~F,~F){\\shortstack~:[~*~;[~A]~]{~A}}~%"
          X-POSITION Y-POSITION
          (NOT (ZEROP (LENGTH POSITION))) POSITION
              (TEXIFY TEXT TEXT-FONT TEXIFY)))


(DEFMETHOD (TEXT-STACK-OBJECT :WRITE-PICTURE) (PICTURE-STREAM)
  "2Write the description of this object to an auxilary file.*"
  ;; We write a string which, when read using the READ function, will create this object.
  (FORMAT PICTURE-STREAM "(MAKE-INSTANCE 'PICTURE:TEXT-STACK-OBJECT ~
                                         :X-POSITION ~D :Y-POSITION ~D ~
                                         :TEXT ~S :TEXT-FONT ~S)~%"
          X-POSITION Y-POSITION TEXT TEXT-FONT))

(DEFMETHOD (TEXT-STACK-OBJECT :TRANSLATE-UNITS) (NEW-OVER-OLD X-ORIGIN Y-ORIGIN)
  (SETQ X-POSITION (+ (* NEW-OVER-OLD (- X-POSITION X-ORIGIN)) X-ORIGIN)
        Y-POSITION (+ (* NEW-OVER-OLD (- Y-POSITION Y-ORIGIN)) Y-ORIGIN)))

(DEFUN SIMPLE-DISTANCE-FROM-BOX (X Y LEFT TOP RIGHT BOTTOM)
  "Compute the minimum distance between the point (X,Y) and the box defined
by LEFT TOP RIGHT BOTTOM."
  ;; Note that SECTOR-CODE assumes that we are using the fourth quadrant, but
  ;; we are using the first quadrant.  The following weird translation swaps
  ;; the top row of sector codes with the bottom row.
  (LET ((SECTOR-CODE (W:SECTOR-CODE X Y LEFT BOTTOM RIGHT TOP)))
    (WHEN (NOT (ZEROP (LDB #o0202 SECTOR-CODE)))
      (SETF (LDB #o0202 SECTOR-CODE) (LOGXOR #o3 (LDB #o0202 SECTOR-CODE))))
    (CASE SECTOR-CODE
      (10                              ; Upper right
       (W:DIST X Y RIGHT TOP))
      (9                               ; Upper left
       (W:DIST X Y LEFT TOP))
      (8                               ; Upper middle
       (- Y TOP))
      (6                               ; Lower right
       (W:DIST X Y RIGHT BOTTOM))
      (5                               ; Lower left
       (W:DIST X Y LEFT BOTTOM))
      (4                               ; Lower middle
       (- BOTTOM Y))
      (2                               ; Middle right
       (- X RIGHT))
      (1                               ; Middle left
       (- LEFT X))
      (0                               ; Center
       0))))

(DEFMETHOD (TEXT-STACK-OBJECT :DISTANCE-FROM-POINT) (WINDOW X Y)
  "Returns the smallest distance that this object is from the point (X,Y).
X and Y are measured in terms of the current unit.
If the point (X,Y) is inside of the object then 0.0 is returned."
  (IF (ZEROP (LENGTH TEXT))
      ;; If the user didn't enter text, then the distance computation is easy.
      (W:DIST X Y X-POSITION Y-POSITION)
      ;;ELSE
      (LOOP WITH VSP    = (SEND WINDOW :VSP)
            WITH FONT   = (LOOKUP-EXPLORER-FONT (SEND SELF :TEXT-FONT))
            WITH LEFT   = NIL
            WITH TOP    = NIL
            WITH RIGHT  = NIL
            WITH BOTTOM = NIL
            WITH CURRENT-X = X-POSITION
            WITH CURRENT-Y = (+ Y-POSITION (MULTIPLE-VALUE-BIND (IGNORE TOP IGNORE BOTTOM)
                                               (TEXT-LINE-BOUNDING-BOX TEXT WINDOW FONT VSP)
                                             (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT
                                                   `(TRANSLATE-PIXEL-TO-UNIT ,(- BOTTOM TOP)))))
            WITH SMALLEST-DISTANCE = MOST-POSITIVE-SINGLE-FLOAT
            FOR LAST-RETURN = NIL THEN RETURN-LOC
            FOR TEXT-START-INDEX = 0 THEN (1+ (OR LAST-RETURN (1- (LENGTH TEXT))))
            FOR RETURN-LOC = (POSITION #\RETURN TEXT :TEST #'CHAR= :START TEXT-START-INDEX)
            WHILE (OR RETURN-LOC LAST-RETURN (= TEXT-START-INDEX 0))
            FOR STRING = (SUBSEQ TEXT TEXT-START-INDEX 
                                 (IF (NULL RETURN-LOC)
                                     ;; Hit the last part of TEXT and it doesn't
                                     ;; have a RETURN in it.
                                     (LENGTH TEXT)
                                     ;;ELSE
                                     RETURN-LOC))
            FOR CURRENT-DISTANCE = 
            (MULTIPLE-VALUE-BIND (LEFT-PIXEL TOP-PIXEL RIGHT-PIXEL BOTTOM-PIXEL)
                (TEXT-LINE-BOUNDING-BOX STRING WINDOW FONT VSP)
              (SETQ LEFT   (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT
                                 `(TRANSLATE-PIXEL-TO-UNIT ,LEFT-PIXEL))
                    TOP    (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT
                                 `(TRANSLATE-PIXEL-TO-UNIT ,TOP-PIXEL))
                    RIGHT  (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT
                                 `(TRANSLATE-PIXEL-TO-UNIT ,RIGHT-PIXEL))
                    BOTTOM (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT
                                 `(TRANSLATE-PIXEL-TO-UNIT ,BOTTOM-PIXEL)))
              (PROG1
                (SIMPLE-DISTANCE-FROM-BOX X Y
                                          (+ CURRENT-X LEFT)  (- CURRENT-Y TOP)
                                          (+ CURRENT-X RIGHT) (- CURRENT-Y BOTTOM))
                (SETQ CURRENT-Y (- CURRENT-Y BOTTOM))))
            DO (WHEN (< CURRENT-DISTANCE SMALLEST-DISTANCE)
                 (SETQ SMALLEST-DISTANCE CURRENT-DISTANCE))
            FINALLY (RETURN SMALLEST-DISTANCE))))

(DEFUN TEXT-LINE-BOUNDING-BOX (STRING WINDOW FONT VSP)
  "Calculates the bounding box for STRING drawn on WINDOW using FONT.
Returns the left top right bottom pixel positions of the bounding box
relative to the origin (0,0)."
  (LET ((DESCENDER-HEIGHT    (+ VSP (W:FONT-CHAR-HEIGHT FONT)))
        (NO-DESCENDER-HEIGHT (+ VSP (W:FONT-BASELINE    FONT))))
    (VALUES 0 0
            (SEND WINDOW :STRING-LENGTH STRING 0 NIL (W:SHEET-INSIDE-WIDTH WINDOW) FONT)
            (IF (ZEROP (LENGTH STRING))
                *EMPTY-LINE-HEIGHT*
                ;;1ELSE*
                (IF (STRING-DESCENDERS-P STRING)
                    DESCENDER-HEIGHT
                    ;;1ELSE*
                    NO-DESCENDER-HEIGHT)))))

(DEFMETHOD (TEXT-STACK-OBJECT :OBJECT-LIMITS) (WINDOW)
  "Returns the minimum and maximum limits of this object.
The values returned are the left top right and bottom unit values."
  (MULTIPLE-VALUE-BIND (LEFT TOP RIGHT BOTTOM)
      (TEXT-LINE-BOUNDING-BOX TEXT WINDOW
                              (LOOKUP-EXPLORER-FONT TEXT-FONT)
        (SEND WINDOW :VSP))
    (VALUES X-POSITION
            (+ (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT
                     `(TRANSLATE-PIXEL-TO-UNIT ,(- BOTTOM TOP)))
               Y-POSITION)
            (+ (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT
                     `(TRANSLATE-PIXEL-TO-UNIT ,(- RIGHT LEFT)))
               X-POSITION)
            Y-POSITION)))

(DEFMETHOD (TEXT-STACK-OBJECT :NAME) ()
  "Return a string which identifies the object."
  (FORMAT NIL "Text \"~A\" at (~D,~D)" (REMOVE-RETURNS TEXT) X-POSITION Y-POSITION))

(DEFMETHOD (TEXT-STACK-OBJECT :SHORT-NAME) ()
  "Return a narrow string which identifies the object."
  (FORMAT NIL "Text ~
               ~:[~;at~%(~,2F,~,2F)~]"
          X-POSITION X-POSITION Y-POSITION))

(DEFUN MAXIMUM-TEXT-WIDTH (WINDOW TEXT FONT)
  (MULTIPLE-VALUE-BIND (IGNORE IGNORE IGNORE MAXIMUM-X)
    (SEND WINDOW :COMPUTE-MOTION TEXT 0 NIL 0 0 NIL 0 NIL NIL NIL FONT)
    MAXIMUM-X))

(DEFUN DISPLAY-TEXT-STACK-CENTERED (TEXT WINDOW X Y FONT ALU)
  "Display TEXT centered inside the box defined by X Y and the maximum width of the text.
The (X,Y) points to the lower left corner of the text."
  (LOOP WITH VSP = (SEND WINDOW :VSP)
        WITH DESCENDER-HEIGHT    = (+ VSP (W:FONT-CHAR-HEIGHT FONT))
        WITH NO-DESCENDER-HEIGHT = (+ VSP (W:FONT-BASELINE    FONT))
        WITH WIDTH = (MAXIMUM-TEXT-WIDTH WINDOW TEXT FONT)
        FOR LAST-RETURN = NIL THEN RETURN-LOC
        FOR TEXT-END-INDEX = (LENGTH TEXT) THEN (OR LAST-RETURN 0)
        FOR RETURN-LOC = (POSITION #\RETURN TEXT :TEST #'CHAR= :END TEXT-END-INDEX :FROM-END T)
        WHILE (OR RETURN-LOC LAST-RETURN (= TEXT-END-INDEX (LENGTH TEXT)))
        FOR STRING = (SUBSEQ TEXT (IF (NULL RETURN-LOC)
                                      ;; Hit the last part of TEXT and it doesn't have a RETURN.
                                      0
                                      ;;ELSE
                                      (1+ RETURN-LOC))
                             TEXT-END-INDEX)
        DO (PROGN
             (DECF Y (IF (ZEROP (LENGTH STRING))
                         *EMPTY-LINE-HEIGHT*
                         ;;1ELSE*
                         (IF (STRING-DESCENDERS-P STRING)
                             DESCENDER-HEIGHT
                             ;;1ELSE*
                             NO-DESCENDER-HEIGHT)))
             (SEND WINDOW :STRING-OUT-CENTERED-EXPLICIT STRING
                   X Y (+ X WIDTH)
                   (W:SHEET-INSIDE-BOTTOM WINDOW)
                   FONT ALU))))

(DEFUN DISPLAY-TEXT-STACK-LEFT (TEXT WINDOW X Y FONT ALU)
  "Display TEXT left justified in the horizontal direction using the values X Y
and the maximum width of the text.
The (X,Y) points to the lower left corner of the text."
  (LOOP WITH VSP = (SEND WINDOW :VSP)
        WITH DESCENDER-HEIGHT    = (+ VSP (W:FONT-CHAR-HEIGHT FONT))
        WITH NO-DESCENDER-HEIGHT = (+ VSP (W:FONT-BASELINE    FONT))
        FOR LAST-RETURN = NIL THEN RETURN-LOC
        FOR TEXT-END-INDEX = (LENGTH TEXT) THEN (OR LAST-RETURN 0)
        FOR RETURN-LOC = (POSITION #\RETURN TEXT :TEST #'CHAR= :END TEXT-END-INDEX :FROM-END T)
        WHILE (OR RETURN-LOC LAST-RETURN (= TEXT-END-INDEX (LENGTH TEXT)))
        FOR STRING = (SUBSEQ TEXT (IF (NULL RETURN-LOC)
                                      ;; Hit the last part of TEXT and it doesn't have a RETURN.
                                      0
                                      ;;ELSE
                                      (1+ RETURN-LOC))
                             TEXT-END-INDEX)
        DO (PROGN
             (DECF Y (IF (ZEROP (LENGTH STRING))
                         *EMPTY-LINE-HEIGHT*
                         ;;1ELSE*
                         (IF (STRING-DESCENDERS-P STRING)
                             DESCENDER-HEIGHT
                             ;;1ELSE*
                             NO-DESCENDER-HEIGHT)))
             (SEND WINDOW :STRING-OUT-EXPLICIT STRING
                   X Y
                   (W:SHEET-INSIDE-RIGHT WINDOW)
                   (W:SHEET-INSIDE-BOTTOM WINDOW)
                   FONT ALU))))

(DEFUN DISPLAY-TEXT-STACK-RIGHT (TEXT WINDOW X Y FONT ALU)
  "Display TEXT right justified in the horizontal direction using the values X Y
and the maximum width of the text.
The (X,Y) points to the lower left corner of the text."
  (LOOP WITH VSP = (SEND WINDOW :VSP)
        WITH DESCENDER-HEIGHT    = (+ VSP (W:FONT-CHAR-HEIGHT FONT))
        WITH NO-DESCENDER-HEIGHT = (+ VSP (W:FONT-BASELINE    FONT))
        WITH WIDTH = (MAXIMUM-TEXT-WIDTH WINDOW TEXT FONT)
        FOR LAST-RETURN = NIL THEN RETURN-LOC
        FOR TEXT-END-INDEX = (LENGTH TEXT) THEN (OR LAST-RETURN 0)
        FOR RETURN-LOC = (POSITION #\RETURN TEXT :TEST #'CHAR= :END TEXT-END-INDEX :FROM-END T)
        WHILE (OR RETURN-LOC LAST-RETURN (= TEXT-END-INDEX (LENGTH TEXT)))
        FOR STRING = (SUBSEQ TEXT (IF (NULL RETURN-LOC)
                                      ;; Hit the last part of TEXT and it doesn't have a RETURN.
                                      0
                                      ;;ELSE
                                      (1+ RETURN-LOC))
                             TEXT-END-INDEX)
        FOR STRING-WIDTH = (SEND WINDOW :STRING-LENGTH STRING 0 NIL
                                 (W:SHEET-INSIDE-WIDTH WINDOW) FONT)
        DO (PROGN
             (DECF Y (IF (ZEROP (LENGTH STRING))
                         *EMPTY-LINE-HEIGHT*
                         ;;1ELSE*
                         (IF (STRING-DESCENDERS-P STRING)
                             DESCENDER-HEIGHT
                             ;;1ELSE*
                             NO-DESCENDER-HEIGHT)))
             (SEND WINDOW :STRING-OUT-EXPLICIT STRING
                   (+ X (- WIDTH STRING-WIDTH))
                   Y
                   (W:SHEET-INSIDE-RIGHT WINDOW)
                   (W:SHEET-INSIDE-BOTTOM WINDOW)
                   FONT ALU))))

(DEFUN DRAW-TEXT-STACK-OBJECT (OBJECT &OPTIONAL (ALU W:CHAR-ALUF))
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((PIXEL-X (TRANSLATE-X-UNIT-TO-PIXEL (SEND OBJECT :X-POSITION)))
        (PIXEL-Y (TRANSLATE-Y-UNIT-TO-PIXEL (SEND OBJECT :Y-POSITION)))
        (POSITION (SEND OBJECT :POSITION))
        (TEXT     (SEND OBJECT :TEXT)))
    ;; If this object doesn't have any text, then don't display anything.
    (WHEN (AND TEXT (PLUSP (LENGTH TEXT)))
      (FUNCALL (COND ((OR (NULL POSITION) (STRING-EQUAL POSITION "")) 'DISPLAY-TEXT-STACK-CENTERED)
                     ((STRING-EQUAL POSITION "L")                     'DISPLAY-TEXT-STACK-LEFT)
                     ((STRING-EQUAL POSITION "R")                     'DISPLAY-TEXT-STACK-RIGHT))
               (SEND OBJECT :TEXT) SELF PIXEL-X PIXEL-Y
               (LOOKUP-EXPLORER-FONT (SEND OBJECT :TEXT-FONT)) ALU))))


(DEFUN MOVE-TEXT-STACK (OBJECT)
  "Use the mouse to move text to a new location."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (COND ((NULL (SEND OBJECT :X-POSITION))
         (PICTURE-ERROR "A corner position is not specified yet.")
         T)
        ((ZEROP (LENGTH (SEND OBJECT :TEXT)))
         (PICTURE-ERROR "Text is not specified yet.")
         T)
        (T
         (LET ((OLD-X-POSITION (SEND OBJECT :X-POSITION))
               (OLD-Y-POSITION (SEND OBJECT :Y-POSITION))
               USER-INPUT)
           (SEND SELF :TURN-ON-RUBBERBAND
                 #'(LAMBDA (X Y)
                     (MULTIPLE-VALUE-BIND (X-UNIT Y-UNIT)
                         (SEND SELF :GRIDIFY-MEASURE
                               (TRANSLATE-PIXEL-TO-UNIT X)
                               (TRANSLATE-PIXEL-TO-UNIT Y))
                       (SETF (SEND OBJECT :X-POSITION) X-UNIT)
                       (SETF (SEND OBJECT :Y-POSITION) Y-UNIT))
                     (SEND SELF :WRITE-OBJECT-STATUS OBJECT)
                     (DRAW-TEXT-STACK-OBJECT OBJECT W:ALU-XOR))
                 (TRANSLATE-X-UNIT-TO-PIXEL OLD-X-POSITION)
                 (TRANSLATE-Y-UNIT-TO-PIXEL OLD-Y-POSITION))
           (SETQ USER-INPUT (READ-MOUSE-XY-INPUT))
           ;; Turn off rubberbanding.
           (SEND SELF :TURN-OFF-RUBBERBAND)
           (IF USER-INPUT
               ;; Draw the object with a IOR alu after we have finished moving it.
               (PROGN
                 (SEND OBJECT :DRAW SELF)
                 T)
               ;;ELSE
               (PROGN
                 ;; Set the text back the way it was.
                 (SEND OBJECT :DRAW SELF W:ALU-XOR)
                 (SETF (SEND OBJECT :X-POSITION) OLD-X-POSITION)
                 (SETF (SEND OBJECT :Y-POSITION) OLD-Y-POSITION)
                 (SEND OBJECT :DRAW SELF W:ALU-XOR)
                 NIL))))))

(DEFCONSTANT TEXT-CORNER-MOUSE-DOCUMENTATION
             '(:MOUSE-L-1 "Pick the upper left corner for the text"
               :MOUSE-M-1 "Abort out of this operation"))

(DEFUN PICK-TEXT-STACK-POINT (OBJECT)
  "Pick an (X,Y) point for text via the mouse."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF (SEND OBJECT :X-POSITION)
      (PROGN
        ;; The user has already specified the position.  All
        ;; that is being done now is to change the position.
        (MOVE-TEXT-STACK OBJECT))
      ;;ELSE
      (MULTIPLE-VALUE-BIND (NEW-X NEW-Y)
          (MOUSE-SPECIFY-POINT OBJECT W:MOUSE-GLYPH-BLOCK-LETTER-T TEXT-CORNER-MOUSE-DOCUMENTATION)
        (IF NEW-X
            (PROGN
              (MULTIPLE-VALUE-BIND (X-UNIT Y-UNIT)
                  (SEND SELF :GRIDIFY-MEASURE
                        (TRANSLATE-PIXEL-TO-UNIT NEW-X)
                        (TRANSLATE-PIXEL-TO-UNIT NEW-Y))
                (SETF (SEND OBJECT :X-POSITION) X-UNIT)
                (SETF (SEND OBJECT :Y-POSITION) Y-UNIT))
              ;; Only draw the text some has been specified.
              (WHEN (PLUSP (LENGTH (SEND OBJECT :TEXT)))
                (DRAW-TEXT-STACK-OBJECT OBJECT))
              T)
            ;;1ELSE*
            NIL))))

(DEFUN ENTER-TEXT-STACK-TEXT-LINES (OBJECT)
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF (NULL (SEND OBJECT :X-POSITION))
      (PICTURE-ERROR "A corner position is not specified yet.")
      ;;ELSE
      (LET ((OLD-TEXT (SEND OBJECT :TEXT))
            (TEXT-LINES (ENTER-TEXT-LINES OBJECT "Text Lines")))
        (WHEN TEXT-LINES
          (WHEN (NOT (ZEROP (LENGTH OLD-TEXT)))
            ;; Erase the old text before we display the new text.
            (SETF (SEND OBJECT :TEXT) OLD-TEXT)
            (DRAW-TEXT-STACK-OBJECT OBJECT W:ALU-ANDCA))
          (SETF (SEND OBJECT :TEXT) TEXT-LINES)
          (WHEN (NOT (ZEROP (LENGTH TEXT-LINES)))
            (DRAW-TEXT-STACK-OBJECT OBJECT))))))

(DEFMETHOD (TEXT-STACK-OBJECT :DRAW) (WINDOW &OPTIONAL (ALU NIL))
  (IF ALU
      (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT (LIST 'DRAW-TEXT-STACK-OBJECT SELF ALU))
      ;;1ELSE*
      (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT (LIST 'DRAW-TEXT-STACK-OBJECT SELF))))

(DEFMETHOD (TEXT-STACK-OBJECT :DRAW-TEXT) (WINDOW &OPTIONAL (ALU NIL))
  (IF ALU
      (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT (LIST 'DRAW-TEXT-STACK-OBJECT SELF ALU))
      ;;1ELSE*
      (SEND WINDOW :SETUP-INSTANCE-ENVIRONMENT (LIST 'DRAW-TEXT-STACK-OBJECT SELF))))

(DEFMETHOD (TEXT-STACK-OBJECT :PICK-CORNER-FUNCTION) ()
  'PICK-TEXT-STACK-POINT)

(DEFMETHOD (TEXT-STACK-OBJECT :REPLICATE) ()
  "2Make a copy of this object.*"
  (MAKE-INSTANCE 'TEXT-STACK-OBJECT
                 :X-POSITION X-POSITION
                 :Y-POSITION Y-POSITION
                 :TEXT TEXT
                 :TEXT-FONT TEXT-FONT))
                 

(DEFCONSTANT VALID-TEXT-STACK-POSITION-ITEM-LIST
             '(("Centered" "")
               ("Left"     "l")
               ("Right"    "r"))
  "2Item list for positions which are valid for text stack objects.*")

(DEFUN GET-TEXT-STACK-POSITION-DATA (OBJECT)
  "2Obtain text position data from the user and redisplay the text, if any exists.*"
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((OLD-POSITION (SEND OBJECT :POSITION))
        (NEW-POSITION (W:MENU-CHOOSE VALID-TEXT-STACK-POSITION-ITEM-LIST
                                     :LABEL "Specify Position of Text"))
        (TEXT (SEND OBJECT :TEXT)))
    ;; Only update the display when we have a position and there is text.
    ;; If the user aborts out of GET-POSITION-DATA then the position will be nil.
    (WHEN NEW-POSITION
      ;; Erase the text before we display it at the new position.
      (SETF (SEND OBJECT :POSITION) OLD-POSITION)
      (WHEN (NOT (ZEROP (LENGTH TEXT)))
        (DRAW-TEXT-STACK-OBJECT OBJECT W:ALU-ANDCA))
      (SETF (SEND OBJECT :POSITION) NEW-POSITION)
      (WHEN (NOT (ZEROP (LENGTH TEXT)))
        (DRAW-TEXT-STACK-OBJECT OBJECT)))))

(DEFCONSTANT TEXT-STACK-ACTION-ITEM-LIST
             '(("Pick a corner" :VALUE PICK-TEXT-STACK-POINT
                :DOCUMENTATION "Pick the lower left corner for the text.")
               ("Lines of text" :VALUE ENTER-TEXT-STACK-TEXT-LINES
                :DOCUMENTATION "Enter the lines of text.")
               ("Text Position" :VALUE GET-TEXT-STACK-POSITION-DATA
                :DOCUMENTATION "Specify how the text is to be positioned in the box.")
               ("Text font" :VALUE GET-FONT-DATA
                :DOCUMENTATION "Specify a font for the text")
               ("Abort" :VALUE :ABORT
                :DOCUMENTATION "Quit without specifying the text.")
               ("Done" :VALUE :DONE
                :DOCUMENTATION "Click here when you are done specifying the text."))
  "Menu item-list through which a user specifies text.")

(DEFCONSTANT TEXT-STACK-ITEM-SEQUENCING
             (GENERATE-ACTION-ITEM-SEQUENCE TEXT-STACK-ACTION-ITEM-LIST
                                            '((NIL 1) (1 2) (2 3) (3 4) (4 6))))

(DEFUN DRAW-TEXT-STACK (&OPTIONAL (OLD-OBJECT NIL))
  "Handles all interaction for drawing text objects.
Note that this is different from the text within objects."
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (LET ((OBJECT NIL))
    (IF OLD-OBJECT
        (SETQ OBJECT OLD-OBJECT)
        ;;1ELSE*
        (SETQ OBJECT (MAKE-INSTANCE 'TEXT-STACK-OBJECT)))
    (IF (GET-OBJECT-DATA OBJECT TEXT-STACK-ACTION-ITEM-LIST "Specify text"
                         TEXT-STACK-ITEM-SEQUENCING)
        (IF (OR (ZEROP (LENGTH (SEND OBJECT :TEXT)))
                (NULL  (SEND OBJECT :X-POSITION))
                (NULL  (SEND OBJECT :Y-POSITION)))
            ;; The text object needs to have some text and a position.
            (PROGN
              (PICTURE-ERROR "The ~a component of the text object must be specified."
                             (COND ((ZEROP (LENGTH (SEND OBJECT :TEXT))) "text")
                                   ((NULL  (SEND OBJECT :X-POSITION)) "initial corner")))
              (WHEN (AND OLD-OBJECT (MOUSE-Y-OR-N-P "Do you really want to
delete this object?"))
                ;; The user had an old object and got rid of the text.  Let
                ;; the user decide what to do about this.  The user may want
                ;; to keep the object around since it still has a position
                ;; and font.
                (SETQ DISPLAY-LIST (REMOVE OBJECT DISPLAY-LIST))))
            ;;ELSE
            (PROGN
              (MULTIPLE-VALUE-BIND (LEFT TOP RIGHT BOTTOM)
                  (SEND OBJECT :OBJECT-LIMITS SELF)
                (SETQ MAXIMUM-X-UNIT (MAX MAXIMUM-X-UNIT RIGHT )
                      MINIMUM-X-UNIT (MIN MINIMUM-X-UNIT LEFT  )
                      MAXIMUM-Y-UNIT (MAX MAXIMUM-Y-UNIT TOP   )
                      MINIMUM-Y-UNIT (MIN MINIMUM-Y-UNIT BOTTOM)))
              ;; If we didn't start out with an object, then we created a new one.
              ;; This means that we need to put this new object onto the display list.
              (WHEN (NULL OLD-OBJECT)
                (PUSH OBJECT DISPLAY-LIST))))
        ;;1ELSE*
        ;; See if enough of the object's parameters were entered for it to be displayed.
        (WHEN (AND (PLUSP (LENGTH (SEND OBJECT :TEXT)))
                   (SEND OBJECT :X-POSITION)
                   (SEND OBJECT :Y-POSITION))
          ;; Don't XOR the object if the user aborted out of the menu for an existing object.
          (WHEN (NULL OLD-OBJECT)
            (SEND OBJECT :DRAW SELF W:ALU-XOR))))))


(DEFMETHOD (TEXT-STACK-OBJECT :EDIT-FUNCTION) ()
  "Returns the symbol which is the function to do editing for this object."
  'DRAW-TEXT-STACK)
