;;;-*- Mode:Common-Lisp; Package:PICTURE; Base:10; Fonts:(MEDFNB HL12B HL12BI) -*-



;;; This software was developed at the Texas Instruments Corporation
;;; and is provided "as-is".  Texas Instruments Corporation and the
;;; author disclaim all warranties on the software, including without
;;; limitation, all implied warranties of merchantability and fitness.
;;;
;;; This software does not contain any technical data or information
;;; that is proprietary in nature.  It may be copied, modified, and
;;; distributed on a non-profit basis and with the inclusion of this
;;; notice.
;;;
;;; Copyright (C) 1989, Texas Instruments Incorporated.




;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 10/14/88    TWE	Put support in place for read-only mode.
;;; 10/13/88    TWE	Updated to be able to handle both the Explorer and microExplorer
;;;			screen sizes.
;;;  9/20/88    TWE	Fixed up scaling to be more general and to not touch the margins.
;;;			Made scaling be relative to the margins if they are present, and
;;;			allowed the user to select the origin for the scaling operations.
;;;  9/10/87    TWE	Added support for landscape mode.
;;;  6/29/87    TWE	Added the support for MINIMUM-X/Y to make the picture only as big
;;;			as the user makes it.
;;;  6/18/87    TWE	Changed the window scale factor to be in tune with the larger
;;;			picture window size.
;;;  5/26/87    TWE	Updated get-new-unit to scale the maximum-x and maximum-y units.
;;;  4/16/87    TWE	Changed TV package prefixes to W.
;;;  3/30/87	TWE	Changed the translation functions to call another function which
;;;			does not have a flavor dependency.  This allows us to perform a
;;;			unit translation during a flavor instantiation.  Also changed
;;;			the name of points and picas to the singular form.  Changed the
;;;			unit->pixel translation so that a different function is called
;;;			depending upon which coordinate (X or Y) is being translated.
;;;  1/06/87	TWE	Changed truncate to round in the unit -> pixel translation.  In
;;;			addition to working properly (so that unit -> pixel -> unit yields
;;;			at least a close approximation of the starting number) it will be
;;;			more accurate.
;;; 12/30/86	TWE	Made units more general by defining unit properties on a symbol.
;;; 12/29/86	TWE	Initial creation.


#|

This file defines translations for UNITS into PIXELS, and
other related functions.

|#

(DEFPARAMETER LIST-OF-ALL-UNITS NIL
  "2List of all units of measure.*")

(DEFUN DEFINE-UNIT (KEYWORD UNITS-PER-INCH NAME-STRING)
  "2Define a unit of measure given a keyword, the number of those units in an inch and a pluralized name string.*"
  (SETF (GET 'UNITS-PER-INCH   KEYWORD) UNITS-PER-INCH)
  (SETF (GET 'UNIT-NAME-STRING KEYWORD) NAME-STRING)
  (PUSH KEYWORD LIST-OF-ALL-UNITS))


;;; The following define the default units.  For each unit there must be
;;; an inch equivalent expressed in units/inch and a plural unit name string
;;; which is used for printing.  To define a new unit, furlongs for example,
;;; all one needs to do is to invoke the function PICTURE:DEFINE-UNIT with the
;;; arguments :FURLONG, 1/7920, and "furlongs".

(EVAL-WHEN (LOAD EVAL)
  (DEFINE-UNIT :MILLIMETER  25.4        "millimeters")
  (DEFINE-UNIT :CENTIMETER  2.54        "centimeters")
  (DEFINE-UNIT :INCH        1.0         "inches")
  (DEFINE-UNIT :POINT      72.27        "points")
  (DEFINE-UNIT :PICA       (/ 72.27 12) "picas"))

(DEFCONSTANT PIXELS-PER-INCH (/ (+ 86.87 90.14) 2)
  "2Number of pixels in an inch.  This is estimated to be average of the number of
pixels/inch horizontally and the number of pixels/inch vertically.*")

(DEFPARAMETER MAGIC-SCREEN-CONSTANT 88.5
  "This constant makes the scale-factor work for various screen sizes.
It was determined empirically and has work for both Explorer and
microExplorer screens.  I don't know why it works or what it really
means, but I do know that it does work.")

;;; The following function returns a number which is used to
;;; scale all dimensions.  This number represents the size of
;;; the screen in relation to the real size of a piece of paper
;;; in the indicated orientation.
(DEFUN SCALE-FACTOR ()
  (IF (EQ *CURRENT-ORIENTATION* :PORTRAIT)
      (/ (/ PORTRAIT-HEIGHT 11.0) MAGIC-SCREEN-CONSTANT)
      ;;ELSE
      (/ (/ LANDSCAPE-HEIGHT 8.5) MAGIC-SCREEN-CONSTANT)))

(DEFUN UNITS-PER-INCH (UNIT)
  (GET 'UNITS-PER-INCH UNIT))

(DEFUN TRANSLATE-UNIT-TO-PIXEL-INTERNAL (UNIT-MAGNITUDE BASE-UNIT)
  (LET ((UNIT-CONVERSION (/ 1 (GET 'UNITS-PER-INCH BASE-UNIT))))
    ;; In the following form, there are about PIXELS-PER-INCH pixels per inch, and
    ;; we need to scale the pixels down by SCALE-FACTOR to be able to fit a
    ;; sheet of paper into a window.
    (VALUES (ROUND (* PIXELS-PER-INCH (SCALE-FACTOR) UNIT-CONVERSION UNIT-MAGNITUDE)))))

(DEFUN TRANSLATE-X-UNIT-TO-PIXEL (UNIT-MAGNITUDE)
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (WHEN UNIT-MAGNITUDE
    (TRANSLATE-UNIT-TO-PIXEL-INTERNAL UNIT-MAGNITUDE BASE-UNIT)))

(DEFUN TRANSLATE-UNIT-TO-PIXEL (UNIT-MAGNITUDE)
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (WHEN UNIT-MAGNITUDE
    (TRANSLATE-UNIT-TO-PIXEL-INTERNAL UNIT-MAGNITUDE BASE-UNIT)))

(DEFUN TRANSLATE-Y-UNIT-TO-PIXEL (UNIT-MAGNITUDE)
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (WHEN UNIT-MAGNITUDE
    (- Y-PIXEL-ORIGIN (TRANSLATE-UNIT-TO-PIXEL-INTERNAL UNIT-MAGNITUDE BASE-UNIT))))

(DEFUN TRANSLATE-PIXEL-TO-UNIT-INTERNAL (PIXEL-MAGNITUDE BASE-UNIT)
  (LET ((UNIT-CONVERSION (GET 'UNITS-PER-INCH BASE-UNIT)))
    ;; In the following form, there are about PIXELS-PER-INCH pixels per inch, and
    ;; we need to scale the pixels up by 1/SCALE-FACTOR to be able to fit a
    ;; sheet of paper into a window.
    (* (/ PIXEL-MAGNITUDE PIXELS-PER-INCH) (/ 1 (SCALE-FACTOR)) UNIT-CONVERSION)))

;;; The (X,Y) translation is localized for pixel-->unit translations so we only
;;; need one function here.
(DEFUN TRANSLATE-PIXEL-TO-UNIT (PIXEL-MAGNITUDE)
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (WHEN PIXEL-MAGNITUDE
    (TRANSLATE-PIXEL-TO-UNIT-INTERNAL PIXEL-MAGNITUDE BASE-UNIT)))

(DEFUN UNIT-NAME-STRING (UNIT)
  (GET 'UNIT-NAME-STRING UNIT))

(DEFUN GET-UNIT-MENU (&OPTIONAL (STREAM NIL))
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW)
           (IGNORE STREAM))
  (W:MENU-CHOOSE (LOOP FOR UNIT IN LIST-OF-ALL-UNITS
                       COLLECTING (LIST (UNIT-NAME-STRING UNIT) :VALUE UNIT))
                 :LABEL "Specify new unit of measure"))

(DEFUN GET-NEW-UNIT ()
  (DECLARE (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (LET ((NEW-UNIT (GET-UNIT-MENU)))
        (WHEN NEW-UNIT
          (LET ((UNIT-SCALE-FACTOR (/ (UNITS-PER-INCH NEW-UNIT) (UNITS-PER-INCH BASE-UNIT))))
            ;; Go through all objects translating the old unit to the new one.
            (LOOP FOR OBJECT IN DISPLAY-LIST
                  DO (SEND OBJECT :TRANSLATE-UNITS (/ (UNITS-PER-INCH NEW-UNIT)
                                                      (UNITS-PER-INCH BASE-UNIT)) 0 0))
            (SETQ GRID-X-SIZE    (* GRID-X-SIZE UNIT-SCALE-FACTOR)
                  GRID-Y-SIZE    (* GRID-Y-SIZE UNIT-SCALE-FACTOR)
                  MAXIMUM-X-UNIT (* MAXIMUM-X-UNIT UNIT-SCALE-FACTOR)
                  MAXIMUM-Y-UNIT (* MAXIMUM-Y-UNIT UNIT-SCALE-FACTOR)
                  MINIMUM-X-UNIT (* MINIMUM-X-UNIT UNIT-SCALE-FACTOR)
                  MINIMUM-Y-UNIT (* MINIMUM-Y-UNIT UNIT-SCALE-FACTOR)
                  BASE-UNIT      NEW-UNIT)
            (SEND SELF :REFRESH))))))

(DEFPARAMETER PREVIOUS-SCALE-FACTOR 1.0)

(DEFCONSTANT SPECIFY-SCALE-POINT
             '(:MOUSE-L-1 "Move to a point which will be the scaling origin"
               :MOUSE-M-1 "Abort out of this operation"))


;;; Previous point specified in a point-relative scaling operation.
(DEFPARAMETER PREVIOUS-POINT-SCALE-X NIL)
(DEFPARAMETER PREVIOUS-POINT-SCALE-Y NIL)

;;; Scaling a picture is really a hack since it doesn't work for all
;;; cases, for example, text and circles.  It may be useful in those
;;; cases where the user doesn't care about those objects that don't
;;; scale well.  An attempt was made to make the scaling command easy
;;; to repeat.  The old scale factor entered previously by the user
;;; is saved and the mouse positions (over the menu item) are saved
;;; so the user can repeat the command by simply clicking the mouse
;;; and pressing the END key.
(DEFUN SCALE-PICTURE (POSITION)
  "Scale all objects according to a particular place on the page.
POSITION can be :UPPER-LEFT for the upper left part of the page,
	:MIDDLE for the center of the page, or
	:POINT for a user-specified place on the page."
  (DECLARE (TYPE ATOM POSITION)
           (SELF-FLAVOR PICTURE-WINDOW))
  (IF READ-ONLY-PICTURE
      (PICTURE-ERROR "Picture is read-only")
      ;;ELSE
      (LET ((SCALE-FACTOR PREVIOUS-SCALE-FACTOR)
            (OLD-X W:MOUSE-X)
            (OLD-Y W:MOUSE-Y)
            (MARGIN-X NIL)
            (MARGIN-Y NIL))
        (FLET ((VALIDATE-SCALE-FACTOR (WINDOW CVV-VARIABLE OLD-VALUE NEW-VALUE)
                                      (DECLARE (IGNORE WINDOW CVV-VARIABLE OLD-VALUE))
                                      (IF (> NEW-VALUE 0)
                                          NIL
                                          ;;1ELSE*
                                          "Scale factor must be positive")))
          (CONDITION-CASE ()
              (W:CHOOSE-VARIABLE-VALUES
               `((,(locf scale-factor)  "Scale factor"  :CONSTRAINT ,#'VALIDATE-SCALE-FACTOR
                  :DOCUMENTATION "Positive scale factor."))
               :MARGIN-CHOICES '("Do It"
                                 ("Abort" (SIGNAL-CONDITION EH:ABORT-OBJECT)))
               :LABEL "Scale entire picture. >1 means expand, <1 means shrink."
               ;; We need the extra space to display error messages in case the user tries
               ;; bad values.
               :EXTRA-WIDTH 30.)
            (EH:ABORT (SETQ SCALE-FACTOR NIL))))
        (WHEN SCALE-FACTOR
          ;; Determine the origin point for scaling.
          (COND ((EQ POSITION :UPPER-LEFT)
                 (SETQ MARGIN-X (IF MARGIN-OBJECT
                                    (SEND MARGIN-OBJECT :X-POSITION)
                                    ;;ELSE
                                    0)
                       MARGIN-Y (IF MARGIN-OBJECT
                                    (SEND MARGIN-OBJECT :Y-POSITION)
                                    ;;ELSE
                                    (* (UNITS-PER-INCH BASE-UNIT)
                                       (MAX-PAGE-HEIGHT)))))
                ((EQ POSITION :MIDDLE)
                 (SETQ MARGIN-X (IF MARGIN-OBJECT
                                    (+ (SEND MARGIN-OBJECT :X-POSITION)
                                       (TRUNCATE (SEND MARGIN-OBJECT :WIDTH) 2))
                                    ;;ELSE
                                    (TRUNCATE (* (UNITS-PER-INCH BASE-UNIT)
                                                 (MAX-PAGE-WIDTH)) 2))
                       MARGIN-Y (IF MARGIN-OBJECT
                                    (- (SEND MARGIN-OBJECT :Y-POSITION)
                                       (TRUNCATE (SEND MARGIN-OBJECT :HEIGHT) 2))
                                    ;;ELSE
                                    (TRUNCATE (* (UNITS-PER-INCH BASE-UNIT)
                                                 (MAX-PAGE-HEIGHT)) 2))))
                ((EQ POSITION :POINT)
                 ;; Position the mouse to where the user had placed it earlier.
                 (WHEN PREVIOUS-POINT-SCALE-X
                   (SETQ W:MOUSE-X PREVIOUS-POINT-SCALE-X
                         W:MOUSE-Y PREVIOUS-POINT-SCALE-Y))
                 (MULTIPLE-VALUE-BIND (X Y)
                     (MOUSE-SPECIFY-POINT NIL
                                          W:MOUSE-GLYPH-CIRCLE-X SPECIFY-SCALE-POINT NIL)
                   (WHEN X 
                     (SETQ PREVIOUS-POINT-SCALE-X W:MOUSE-X
                           PREVIOUS-POINT-SCALE-Y W:MOUSE-Y
                           MARGIN-X (TRANSLATE-PIXEL-TO-UNIT X)
                           MARGIN-Y (TRANSLATE-PIXEL-TO-UNIT Y))))))
          (SETQ PREVIOUS-SCALE-FACTOR SCALE-FACTOR)
          (WHEN MARGIN-X
            (LOOP FOR OBJECT IN DISPLAY-LIST
                  WHEN (NOT (EQ OBJECT MARGIN-OBJECT))
                  DO (SEND OBJECT :TRANSLATE-UNITS SCALE-FACTOR MARGIN-X MARGIN-Y))))
        (SETQ W:MOUSE-X OLD-X
              W:MOUSE-Y OLD-Y)
        (RECALCULATE-BOUNDS)
        (SEND SELF :REFRESH))))
