;;;   -*- Mode: Zetalisp; Package: Compiler; Base: 10. -*-
;;;
;;;           COMPLETENESS PROBE UTILITY    
;;;
;;; Copyright (C) 1984, 1987 Texas Instruments Incorporated. All rights reserved.
;;;
;;;        See file TOOLS;PROBES.TEXT for user documentation.
;;;
;;;
;;;  (PROBE-FUNCTION function-spec ...)  
;;;            Inserts probes into the function(s) named and recompiles.
;;;            The function definition must have been evaluated or 
;;;            compiled in an editor buffer so that the interpreted 
;;;            version of the definition is available.
;;;
;;;  (PROBE-FUNCTION "pathname" ...)
;;;            Compiles from a Lisp file into memory, with probes 
;;;            inserted into all functions, macros, methods, and 
;;;            non-trivial DEFSUBSTs defined in the file.  
;;;            [Warning: this one may still have bugs in it.  Using 
;;;            (PROBING-ON) and m-X Compile Buffer is probably safer.]
;;;
;;;  (COMPILE-FILE-WITH-PROBES input-file [ options ] )
;;;            Compiles a Lisp file into an object file, with probes 
;;;            inserted into all functions, macros, methods, and 
;;;            non-trivial DEFSUBSTs defined in the file.  
;;;            Accepts all of the same optional keyword arguments as
;;;            COMPILE-FILE.
;;;
;;;  (PROBING-ON)
;;;            Causes probes to be inserted into everything compiled
;;;            afterwards until (PROBING-OFF) or (LOGOUT).
;;;
;;;  (PROBING-OFF)
;;;            Undoes (PROBING-ON), returning the compiler to its
;;;            normal state.
;;;
;;;  (PROBE-REPORT)         
;;;            For all of the probed functions, lists the function name,
;;;            number of times called, number of basic blocks of code,
;;;            and how many of them were executed.  The functions are
;;;            listed in alphabetical order.
;;;
;;;  (PROBE-REPORT-FREQUENCY)
;;;            Like (PROBE-REPORT) except the functions are listed 
;;;            in order of the number of times called.
;;;
;;;  (PROBE-REPORT T)
;;;            In addition to the above information, for each function
;;;            for which some but not all of the basic blocks were 
;;;            executed, displays the function definition with executed
;;;            portions in lower case and unused portions in upper case.
;;;
;;;  (PROBE-REPORT function-spec ...)
;;;            Give information for indicated function(s) only.
;;;
;;;  (PROBE-REPORT 0)
;;;            List the names of functions which have an 
;;;            execution count of zero; ie, were never called.
;;;            
;;;  (CLEAR-PROBE-DATA)
;;;            Re-initializes the probe data.  The probes remain in 
;;;            the functions but all other information is forgotten.
;;;            Any probed functions will be recorded with a call count
;;;            of 1 the next time they are called.
;;;
;;;  (FASD-PROBE-DATA "pathname")
;;;            Writes the probe data to an XLD file.  The file can be
;;;            loaded later to restore the data.
;;;
;;;
;;; Written by D.N.G. October, 1984.
;;; 10/19/84 - Fixed to handle variable name in a DO not in a list.
;;; 10/23/84 - Improve handling of macro arguments.
;;; 11/07/84 - Additional report options.
;;; 11/17/84 - Modified for use with or without window interface.
;;; 11/19/84 - Correct font for display of never-executed function.
;;; 12/04/84 - Fix to handle "(DO NIL ((..." [used in CHECK-ARG macro].
;;; 12/07/84 - Bind SAVE-INTERP-DEF flag for QCOMPILE0.
;;;  1/03/85 - Fix to handle destructuring in LOOP macro.
;;;  2/14/85 - Alleviate problems with fasdumping interpreted defs in probes data.
;;;  3/21/85 - Fix so function spec lists in the compiler temporary 
;;;            area are copied before being entered into the hash table.
;;;  2/21/86 - Fix to recognize *THROW and CLI:THROW; shorten the report heading
;;;	       to leave more room for the function name.
;;;  3/26/86 - Fix for when compiler invoked from interpreting flavor definitions.
;;; 11/08/86 - Add handling for ECASE and CCASE.
;;; 01/27/87 - Modified for use with Rel.3 (CLM).
;;;  4/25/87 DNG - Add handling for ETYPECASE and CTYPECASE.
;;;  5/19/87 DNG - Treat PROG1 and PROG2 as functions instead of using the 
;;;		alternate macro definition; add handling for MULTIPLE-VALUE-CALL; 
;;;		fixed handling of ZLC:EVERY.
;;;  5/20/87 DNG - Treat INCF, DECF, and REMF as ordinary functions; permit probing 
;;;		sub-forms of generalized variables in SETF, PUSH, etc.
;;;  5/21/87 DNG - Fix saving of interpreted definition in COMPILE-FILE.
;;;  6/05/87 DNG - Add use of CONDITIONAL-PROBE-ENTRY for probing kernel code.
;;;  8/15/88 DNG - Load "PPRINT" from current directory instead of "TOOLS" host.
;;;  1/17/89 DNG - Don't probe local macros defined by MACROLET since they get 
;;;		evaluated at compile time.
;;;

;;; load in a special version of the Pretty Printer that allows printing
;;; with fonts in the Window version of PROBE.
(sys:load-if (send SI:FDEFINE-FILE-PATHNAME :new-pathname
		   :name "PPRINT" :type nil :version :newest)
	     :verbose t)

(EXPORT '( PROBE-FUNCTION PROBE-REPORT PROBE-FUNCTIONS-IN-FILE
	  PROBE-FUNCTION-1 PROBE-REPORT-1 CLEAR-PROBE-DATA FASD-PROBE-DATA
	  PROBING-ON PROBING-OFF PROBE-REPORT-FREQUENCY
	  *PROBE-NOW* WITH-PROBES ))


#+Symbolics
(GLOBALIZE 'PROBE-FUNCTION)
#+Symbolics
(GLOBALIZE 'PROBE-REPORT)
#+Symbolics
(GLOBALIZE 'COMPILE-FILE-WITH-PROBES)

;;for Rel3
(IMPORT 'PROBE-FUNCTION 'USER)
(IMPORT 'PROBE-REPORT 'USER)
(IMPORT 'COMPILE-FILE-WITH-PROBES 'USER)

;;;
;;;   =====   Data declarations and accessor functions   =====
;;;
(DEFSTRUCT PROBE-DATA
  ( PROBE-DATA-NAME NIL ) ; the name (function spec) of the function
  ( PROBE-DATA-COUNT 0 ) ; the number of times executed
  ( PROBE-DATA-ARRAY NIL ) ; the bit array for the probe points, or
			   ;  the number of probes before first execution.
  ( PROBE-DATA-INTERP-DEF NIL ) ; interpreted definition of the function.
 )
(DEFVAR *LAST-PROBE-NUMBER* 0)
(DEFVAR *GRIND-PROBE-ARRAY* NIL)
(DEFVAR *PROBE-NEXT* NIL)
(DEFVAR *PROBING-MACRO-ARGS* NIL)
(DEFVAR *MULTI-PATH-MACRO* )
(DEFPROP .PROBE-ARRAY. T COMPILER:IGNORABLE-VARIABLE)
(DEFSUBST GET-PROBE-NUMBER ()
  (SETQ *LAST-PROBE-NUMBER* (+ 1 *LAST-PROBE-NUMBER*)) )

#| old data base using association list
(DEFVAR *PROBES-DATA-BASE* NIL)

(DEFSUBST ADD-PROBE-DATA ( X )
  (DECLARE (TYPE PROBE-DATA X))
  (PUSH X *PROBES-DATA-BASE*) )

(DEFUN FIND-PROBE-DATA ( NAME )
  (DOLIST ( X  *PROBES-DATA-BASE* NIL )
    (WHEN (EQUAL NAME (PROBE-DATA-NAME X))
      (RETURN X) ) ) )

(DEFUN CLEAR-PROBE-DATA () (SETQ *PROBES-DATA-BASE* NIL))

(DEFSUBST PROBES-DATA-LIST () (COPYLIST *PROBES-DATA-BASE*))
|#


(DEFVAR *PROBES-HASH-TABLE*
	(MAKE-HASH-TABLE :SIZE 200
                               #-SYMBOLICS :REHASH-SIZE #-SYMBOLICS 200 :TEST #'EQUAL))

(DEFSUBST ADD-PROBE-DATA ( X )
  #+Common-Lisp (DECLARE (TYPE PROBE-DATA X))
  (LET ((XX X))					; ensure single evaluation
    (PUTHASH (PROBE-DATA-NAME XX) XX *PROBES-HASH-TABLE*) ) )

(DEFSUBST FIND-PROBE-DATA ( NAME )
  (LET (( DATA (GETHASH NAME *PROBES-HASH-TABLE*) ))
    (COMMENT
    (ASSERT (OR (NULL DATA)
		(EQUAL NAME (PROBE-DATA-NAME DATA))) )
    )
    DATA ) )

(DEFUN CLEAR-PROBE-DATA () (CLRHASH *PROBES-HASH-TABLE*) T)

(DEFUN FASD-PROBE-DATA ( FILE )
  (COMPILER:FASD-SYMBOL-VALUE FILE '*PROBES-HASH-TABLE*) )

(DEFSUBST PROBES-DATA-LIST ()
  (LET (( PROBES-DATA-LIST NIL ))
    #+Symbolics (DECLARE (SPECIAL PROBES-DATA-LIST))
    ;; Note: not using MAPHASH-RETURN because it is broken in DELUXE 1.79.
    (MAPHASH (FUNCTION (LAMBDA ( KEY VALUE )
			 (IGNORE KEY)
			 (PUSH VALUE PROBES-DATA-LIST)))
	     *PROBES-HASH-TABLE* )
  PROBES-DATA-LIST ) )
;;;
;;;   =====   Code to insert probes into lisp code   =====
;;;
(DEFUN PROBE-FUNCTION-BLOCK ( EXP &OPTIONAL (NAME (SECOND EXP)) )
   (LET* (( *LAST-PROBE-NUMBER* 0 )
	  ( *PROBE-NEXT* NIL )
          ( PROBED-EXP (PROBE-FORM EXP) ))
   `(LET (( .PROBE-ARRAY. (PROBE-ENTRY ',NAME ,*LAST-PROBE-NUMBER* ) ))
           ,PROBED-EXP ) ) )

(DEFUN PROBE-FUNCTION (&QUOTE &REST FUNCTIONS)
  "Place completeness testing probes in the functions given as arguments."
  (DOLIST ( F FUNCTIONS )
    (IF (STRINGP F)
	(PROBE-FUNCTIONS-IN-FILE F)
      (PROBE-FUNCTION-1 
	(IF (AND (CONSP F)
		 (EQ (FIRST F) 'QUOTE))
	    (SECOND F)
	  F ) ) ) )
  (VALUES) )

(DEFUN PROBE-FUNCTION-1 ( FNAME )
  "Insert probes into a function.  The argument is a function spec."
  ;; 01/23/87 CLM - Use COMPILEDP to get the function's interpreted definition.
  (LET ( FDEF INTERP-DEF FSPEC (MACRO-FLAG NIL) )
    (SETQ FSPEC (SI:UNENCAPSULATE-FUNCTION-SPEC FNAME))
    (COND ((AND (SYMBOLP FSPEC)
		(FBOUNDP FSPEC)
		(SETQ FDEF (FSYMEVAL FSPEC)) ))
	  ((AND (CONSP FSPEC)
		(VALIDATE-FUNCTION-SPEC FSPEC)
		(FDEFINEDP FSPEC)
		(SETQ FDEF (FDEFINITION FSPEC)) ))
	  (T (FERROR NIL "~S is not a defined function." FNAME)
	     (RETURN-FROM PROBE-FUNCTION-1 NIL) ) )
    (WHEN (AND (CONSP FDEF) (EQ (CAR FDEF) 'MACRO))
      (SETQ FDEF (CDR FDEF))
      (SETQ MACRO-FLAG T) )
    (IF (CONSP FDEF)
	(SETQ INTERP-DEF FDEF)
      (SETQ INTERP-DEF
	    #-Symbolics (COMPILEDP FDEF)
	    #+Symbolics (SECOND (ASSQ ':INTERPRETED-DEFINITION
				      (DEBUGGING-INFO FDEF) ) ) ) )
    (IF (CONSP
	  INTERP-DEF)
	(LET (( *LAST-PROBE-NUMBER* 0 ) PROBED-DEF
	      ( DEFAULT-CONS-AREA BACKGROUND-CONS-AREA ))
	  (SETQ PROBED-DEF (PROBED-NAMED-LAMBDA INTERP-DEF FNAME))
	  (WHEN MACRO-FLAG (PUSH 'MACRO PROBED-DEF))
	  (COMPILE FSPEC PROBED-DEF )
	  (LET (( TEM (FIND-PROBE-DATA FNAME) ))
	    (IF (NULL TEM)
		(ADD-PROBE-DATA
		  (SETQ TEM (MAKE-PROBE-DATA PROBE-DATA-NAME FNAME
					     PROBE-DATA-ARRAY *LAST-PROBE-NUMBER*)))
		(PROGN
		  (SETF (PROBE-DATA-ARRAY TEM) *LAST-PROBE-NUMBER*)
		  (SETF (PROBE-DATA-COUNT TEM) 0) ) )
	    (SETF (PROBE-DATA-INTERP-DEF TEM) PROBED-DEF)
	    ) )
	(FERROR NIL "~S does not have an interpreted definition." FNAME))
    (VALUES) ) )

(DEFUN PROBED-NAMED-LAMBDA ( INTERP-DEF &OPTIONAL FNAME)
  ;; 01/27/87 CLM - use PARSE-BODY instead of obsolete EXTRACT-DECLARATIONS.
  ;;  5/20/87 DNG - Fix to not lose doc string or second DECLARE.
  ;;  6/05/87 DNG - Use CONDITIONAL-PROBE-ENTRY for probing kernel code.
  (IF (EQ (FIRST INTERP-DEF) 'MACRO)
      (CONS 'MACRO (PROBED-NAMED-LAMBDA (CDR INTERP-DEF) FNAME))
    (*CATCH 'ALREADY-PROBED
      (MULTIPLE-VALUE-BIND (BODY DECLS DOC)
	  (PARSE-BODY (REST3 INTERP-DEF) NIL T)
	(LET ((NEW-BODY
		(LET* (( *PROBE-NEXT* NIL )
		       ( *PROBING-MACRO-ARGS* NIL )
		       ( PROBED-EXP (PROBE-FORMS BODY) )
		       ( NAME (COND ((NOT (NULL FNAME)) FNAME)
				    ((ATOM (SECOND INTERP-DEF))
				     (SECOND INTERP-DEF))
				    ((CONSP (FIRST (SECOND INTERP-DEF)))
				     (FIRST (SECOND INTERP-DEF)))
				    ((KEYWORDP (FIRST (SECOND INTERP-DEF)))
				     (SECOND INTERP-DEF))
				    (T (FIRST (SECOND INTERP-DEF))) )) )
		  `(LET (( .PROBE-ARRAY.
			  (,(IF (OR (EQ PACKAGE COMPILER:KERNEL-PACKAGE)
				    SI:FILE-IN-COLD-LOAD) ; be careful with kernel code
				'CONDITIONAL-PROBE-ENTRY
			      'PROBE-ENTRY)
			   ',NAME
			   ,*LAST-PROBE-NUMBER* ) ))
		     ,@PROBED-EXP ) ) ))
	  (SETQ NEW-BODY (APPEND DECLS (LIST NEW-BODY)))
	  (WHEN DOC (PUSH DOC NEW-BODY))
	  (RETURN-FROM PROBED-NAMED-LAMBDA
	    (LIST* (FIRST INTERP-DEF)		; NAMED-LAMBDA
		   (SECOND INTERP-DEF)
		   (THIRD INTERP-DEF)		; argument list
		   NEW-BODY 	      
		   ) ) ) ) )
    INTERP-DEF ) )

(DEFUN (:PROPERTY PROBE-ENTRY PROBES) ( FORM )
  (SETQ *LAST-PROBE-NUMBER* (THIRD FORM))
  (*THROW 'ALREADY-PROBED NIL) )

(PUTPROP 'CONDITIONAL-PROBE-ENTRY (GET 'PROBE-ENTRY 'PROBES) 'PROBES)

(DEFUN QCOMPILE1-WRAPPER ( )
  ;; 3/21/85 DNG - Copy non-symbol function spec if it is in a temporary area.
  ;; 3/26/86 DNG - Check QC-FILE-IN-PROGRESS before QC-FILE-LOAD-FLAG because
  ;;		the latter may be undefined when making flavor methods.
  ;; 1/27/87 CLM - This is now wrapped around QCOMPILE1.
  ;; 2/05/87 CLM - Don't try to probe unnamed functions, e.g. top level lambdas.
  (DECLARE (SPECIAL ARGLIST))
  (LET ((FNAME (COMPILAND-FUNCTION-NAME (FIRST ARGLIST)))
	(DEFAULT-CONS-AREA BACKGROUND-CONS-AREA))
    (UNLESS (OR (NULL FNAME)
		(EQ (CAR-SAFE FNAME) ':INTERNAL)
		(EQ FNAME 'PROBE-ENTRY)
		(EQ FNAME 'CONDITIONAL-PROBE-ENTRY)
		(AND (CONSP FNAME)
		     (EQ (FIRST FNAME) ':METHOD)
		     (MEMQ (THIRD FNAME) '(:COMBINED SI:FASLOAD-COMBINED)) ) )
      (LET* (( *LAST-PROBE-NUMBER* 0 )
	     ( NEW-EXP (PROBED-NAMED-LAMBDA (COMPILAND-DEFINITION (FIRST ARGLIST)))) )
	(UNLESS (OR (AND (MEMQ (FIRST NEW-EXP)
			       '(#!z NAMED-SUBST #!c NAMED-SUBST))
			 (< *LAST-PROBE-NUMBER* 2))
		    (EQ NEW-EXP (COMPILAND-DEFINITION (FIRST ARGLIST))) )

	  (SETF (COMPILAND-DEFINITION (FIRST ARGLIST)) NEW-EXP)

	  (IF COMPILER:COMPILER-VERBOSE
	      (FORMAT T " ~40,2T with probes.")
	    (FORMAT T "~&Inserting probes in ~S" FNAME) )
	  (LET (( TEM (FIND-PROBE-DATA FNAME) ))
	    (IF (NULL TEM)
		(ADD-PROBE-DATA
		  (SETQ TEM (MAKE-PROBE-DATA PROBE-DATA-NAME
					        (SI:COPY-OBJECT-TREE FNAME T)
					     PROBE-DATA-ARRAY *LAST-PROBE-NUMBER*)))
	      (PROGN
		(SETF (PROBE-DATA-ARRAY TEM) *LAST-PROBE-NUMBER*)
		(SETF (PROBE-DATA-COUNT TEM) 0) ) )
	    (WHEN (OR (NOT COMPILER:QC-FILE-IN-PROGRESS)
		      COMPILER:QC-FILE-LOAD-FLAG)
	      ;; Don't save interpreted definition in QC-FILE because it is
	      ;; probably in a temporary memory area.
	      (SETF (PROBE-DATA-INTERP-DEF TEM) NEW-EXP) )
	    )) ) ) ) )

(DEFUN PROBING-ON ()
  "Cause probes to be inserted during compilation."
  ;; 01/27/87 CLM - Advise QCOMPILE1.
  ;;  5/21/87 DNG - Set SAVE-INTERP-DEF instead of binding it since it is now
  ;;		tested in QCOMPILE2 instead of QCOMPILE1.
  (DECLARE (SPECIAL COMPILER:SAVE-INTERP-DEF))
  (ADVISE COMPILER:QCOMPILE1 :BEFORE INSERT-PROBES NIL
    (QCOMPILE1-WRAPPER))
  (UNLESS (MEMBER '(PROBING-OFF) LOGOUT-LIST)
    (PUSH '(PROBING-OFF) LOGOUT-LIST) )
  (UNLESS COMPILER:SAVE-INTERP-DEF 
    (SETQ COMPILER:SAVE-INTERP-DEF 'PROBING-ON))
  T
  )

(DEFUN PROBING-OFF ()
  "Turns off insertion of probes during compilation."
  (DECLARE (SPECIAL COMPILER:SAVE-INTERP-DEF))
  (UNLESS (EQ 'COMPILER:QCOMPILE1
	      (SI:UNENCAPSULATE-FUNCTION-SPEC 'COMPILER:QCOMPILE1))
    (UNADVISE COMPILER:QCOMPILE1 :BEFORE INSERT-PROBES) )
  (WHEN (EQ COMPILER:SAVE-INTERP-DEF 'PROBING-ON)
    (SETQ COMPILER:SAVE-INTERP-DEF NIL))
  NIL
 )

#-Symbolics
(DEFMACRO WITH-PROBING (&BODY BODY)
  `(UNWIND-PROTECT
      (PROGN
	(PROBING-ON)
	,@BODY )
     (PROBING-OFF) ) )

#-Symbolics
(DEFUN COMPILE-FILE-WITH-PROBES (&REST ARGS)
  "Compile a file, with probes inserted into all the functions."
  (DECLARE (ARGLIST . #,(ARGLIST 'COMPILE-FILE)))
  (LET (( COMPILER:SAVE-INTERP-DEF T ))
    (DECLARE (SPECIAL COMPILER:SAVE-INTERP-DEF))
    (WITH-PROBING
      (APPLY #'COMPILE-FILE ARGS) ) ))

#-Symbolics
(DEFUN PROBE-FUNCTIONS-IN-FILE ( INFILE )
  "Compile from a file to memory with probe insertion."
  (WITH-PROBING
    (QC-FILE-MEM INFILE) ) )

#-Symbolics
(DEFUN QC-FILE-MEM (INFILE &OPTIONAL  PACKAGE-SPEC
                                 FILE-LOCAL-DECLARATIONS
                                 DONT-SET-DEFAULT-P )
  ;; 01/19/87 CLM - the latest version from DEBUG.
  (LET ( GENERIC-PATHNAME
	 QC-FILE-PACKAGE
	 QC-FILE-MACROS-EXPANDED
	 (QC-FILE-RECORD-MACROS-EXPANDED T)
	 ( DECLARATIONS-IGNORED (AND (BOUNDP 'DECLARATIONS-IGNORED)
				     DECLARATIONS-IGNORED) )
	 ( INLINE-DECLARATIONS (AND (BOUNDP 'INLINE-DECLARATIONS)
				    INLINE-DECLARATIONS) )
	 ( *RETURN-STATUS* 0 )
	 #+MIT
	 (QC-FILE-REL-FORMAT NIL)
	 ( FASD-STREAM NIL ) )
  (DECLARE (SPECIAL INLINE-DECLARATIONS *RETURN-STATUS* DECLARATIONS-IGNORED))
  "Compile from file to memory."
  ;; Default the specified input and output file names.  Open files.
  (SETQ INFILE (FS:MERGE-PATHNAME-DEFAULTS INFILE FS:LOAD-PATHNAME-DEFAULTS NIL))
  (WITH-OPEN-STREAM (INPUT-STREAM
                      (FILE-RETRY-NEW-PATHNAME (INFILE FS:FILE-ERROR)
                        (SEND INFILE :OPEN-CANONICAL-DEFAULT-TYPE ':LISP)))
    ;; The input pathname might have been changed by the user in response to an error.
    ;; Also, find out what type field was actually found.
    (SETQ INFILE (SEND INPUT-STREAM :PATHNAME))
    (OR DONT-SET-DEFAULT-P (FS:SET-DEFAULT-PATHNAME INFILE FS:LOAD-PATHNAME-DEFAULTS))
    (SETQ GENERIC-PATHNAME (SEND INFILE :GENERIC-PATHNAME))

    ;; Get the file property list again, in case we don't have it already or it changed
    (FS:READ-ATTRIBUTE-LIST GENERIC-PATHNAME INPUT-STREAM)

    ;; Bind all the variables required by the file property list.
    (MULTIPLE-VALUE-BIND (VARS VALS) (FS:FILE-ATTRIBUTE-BINDINGS GENERIC-PATHNAME)
      (PROGV VARS VALS
	(LOCKING-RESOURCES-NO-QFASL
	  (COMPILE-STREAM INPUT-STREAM GENERIC-PATHNAME
			  NIL			; fasd flag
			  #'QC-MEM-WORK-COMPILE
			  T			; load flag
			  NIL			; in core flag
			  PACKAGE-SPEC
			  FILE-LOCAL-DECLARATIONS
			  NIL			; READ-THEN-PROCESS-FLAG
			  T)			; whole-file-p
	  ))))
  (VALUES NIL *RETURN-STATUS*) ))

(DEFUN QC-MEM-WORK-COMPILE (FORM)
  (COMPILE-DRIVER FORM #'QC-MEM-COMMON NIL))                                                                              
(DEFUN QC-MEM-COMMON (FORM TYPE)
  ;;  8/21/85 Delete (RPLACA (FUNCTION-CELL-LOCATION TEM) FV).
  ;; 10/28/85 Fix to evaluate SPECIAL declarations when cross-compiling for :JUST-COUNT.
  ;;  8/18/86 Rewritten for new COMPILE-DRIVER.
  ;; 01/19/87 Added lates version from DEBUG.
  (DECLARE (LIST FORM) (SYMBOL TYPE))
  (UNLESS (ATOM FORM)
    (IF (EQ TARGET-PROCESSOR HOST-PROCESSOR)
	(IF (EQ TYPE 'SPECIAL)
	    (SI:EVAL1 FORM)
	  (COMPILE-TOP-LEVEL-FORM FORM 'COMPILE-TO-CORE #'SI:EVAL1))
      (LET ((UNDO-DECLARATIONS-FLAG T))
	(IF (MEMBER TYPE '(SPECIAL DECLARE))
	    (COMPILE-TIME-EVAL FORM TYPE)
	  (COMPILE-TOP-LEVEL-FORM FORM ':JUST-COUNT #'IGNORE))))))

(DEFSUBST MACRO-ARG-VALUE ( PAIR ) (CAR PAIR))
(DEFSUBST MACRO-ARG-USE   ( PAIR ) (CDR PAIR))

(DEFUN PROBE ( FORM )
  (SETQ *PROBE-NEXT* T)
  (PROBE-FORM FORM) )

(DEFUN PROBE-FORM ( FORM &OPTIONAL PROBE-ARGS-ONLY )
  (DECLARE COMPILER:(OPTIMIZE SPEED))
  (UNLESS (NULL FORM)
    (LET (( MACRO-ARG (ASSQ FORM *PROBING-MACRO-ARGS*) ))
      (UNLESS (NULL MACRO-ARG)
	;; found one of the macro arguments
	(UNLESS PROBE-ARGS-ONLY
	  (WHEN (AND *PROBE-NEXT*
		     (NOT (ATOM FORM)))
	    (SETQ *MULTI-PATH-MACRO* T) )
	  (WHEN (NULL (MACRO-ARG-USE MACRO-ARG))
	    (SETF (MACRO-ARG-USE MACRO-ARG) 'EVAL) ))
	(RETURN-FROM PROBE-FORM FORM) )  ) )
  (LET ( HANDLER MACRO-ARG )
    (COND ((AND *PROBE-NEXT* (NULL *PROBING-MACRO-ARGS*) (NOT PROBE-ARGS-ONLY))
           (IF (AND (CONSP FORM) (EQ (FIRST FORM) 'DECLARE))
               FORM
             (PROGN
	       (SETQ *PROBE-NEXT* NIL)
	       (LIST 'PROBE-POINT (GET-PROBE-NUMBER) (PROBE-FORM FORM)) ) ) )
	  ((ATOM FORM)
	   FORM )
	  ((AND (SETQ MACRO-ARG (ASSQ (FIRST FORM) *PROBING-MACRO-ARGS*) )
		(NEQ (MACRO-ARG-USE MACRO-ARG) 'QUOTE)
		(SETF (MACRO-ARG-USE MACRO-ARG) 'FUNCTION)
		NIL ))
	  ((NOT (NULL (SETQ HANDLER
			    (AND (SYMBOLP (FIRST FORM))
				 (GET (FIRST FORM) 'PROBES)))))
	   (FUNCALL HANDLER FORM))
	  ((AND (CONSP (FIRST FORM))
		(MEMQ (FIRST (FIRST FORM)) '(LAMBDA #-Symbolics CLI:LAMBDA)) )
	   (IF (NULL MACRO-ARG)
	       (LET ( LEXP )
		 (SETQ *PROBE-NEXT* T)
		 (SETQ LEXP (LET-PROBE (FIRST FORM)))
		 (SETQ *PROBE-NEXT* T)
		 (CONS LEXP (PROBE-FORMS (REST1 FORM))) )
	     (CONS (FIRST FORM) (PROBE-FORMS (REST1 FORM))) ) )
	  ((NULL (REST1 FORM)) FORM)
	  ((MACRO-FUNCTION (FIRST FORM))
	   (IF (NULL *PROBING-MACRO-ARGS*)
	       (PROBE-MACRO-CALL FORM)
	     (PROBE-FORM (FUNCALL (MACRO-FUNCTION (FIRST FORM)) FORM))) )
	  ((EVERY (REST1 FORM) #'ATOM) FORM)
	  ((SPECIAL-FORM-P (FIRST FORM))
	   (IF *PROBING-MACRO-ARGS*
	       (*THROW 'TOP-LEVEL-MACRO NIL)
	     `(PROBE-UNKNOWN ,FORM)) )
	  (T (CONS (FIRST FORM) (PROBE-FORMS (REST1 FORM))) ) ) )  )

(DEFUN PROBE-MACRO-CALL ( FORM )
      (LET* (( ARGS (LOOP FOR ARG IN (REST1 FORM)
			  COLLECT (IF (ATOM ARG)
				      ARG
				    (COPY-LIST ARG))) )
	     ( *PROBING-MACRO-ARGS*
	       (LOOP FOR ARG IN ARGS
		     COLLECT (CONS ARG NIL)) )
	     (*MULTI-PATH-MACRO* NIL)
	     ( EXPANDED-FORM
	      (FUNCALL (MACRO-FUNCTION (FIRST FORM)) (CONS (FIRST FORM) ARGS)) ) )
	(IF (*CATCH 'TOP-LEVEL-MACRO 
	      (PROBE-FORM EXPANDED-FORM) )
	    (LET ((ARGS-ALIST *PROBING-MACRO-ARGS*)
		  (*PROBING-MACRO-ARGS* NIL)
		  (PROBE-ARGS *MULTI-PATH-MACRO*))
	      (DOLIST ( MACRO-ARG ARGS-ALIST )
		(SETQ *PROBE-NEXT* NIL)
		(SELECTQ (MACRO-ARG-USE MACRO-ARG)
		  ( EVAL
		   (SETF (MACRO-ARG-VALUE MACRO-ARG)
			 (PROBE-FORM (MACRO-ARG-VALUE MACRO-ARG)) ) )
		  ( FUNCTION
		   (UNLESS (ATOM (MACRO-ARG-VALUE MACRO-ARG))
		     (SETF (MACRO-ARG-VALUE MACRO-ARG)
			   (LET-PROBE (MACRO-ARG-VALUE MACRO-ARG)))) )
		  ( NIL (MACRO-ARG-VALUE MACRO-ARG) )
		  ( QUOTE (MACRO-ARG-VALUE MACRO-ARG) ) )
		(WHEN *PROBE-NEXT*
		  (SETQ PROBE-ARGS T)) )
	      (WHEN PROBE-ARGS
		(SETQ *PROBE-NEXT* T))
	      (CONS (FIRST FORM)
		    (LOOP FOR MACRO-ARG IN ARGS-ALIST
			  COLLECT
			  (IF (AND PROBE-ARGS
				   (EQ (MACRO-ARG-USE MACRO-ARG) 'EVAL))
			      (LIST 'PROBE-POINT
				    (GET-PROBE-NUMBER)
				    (MACRO-ARG-VALUE MACRO-ARG))
			    (MACRO-ARG-VALUE MACRO-ARG))
			  ) ) )
	  `(PROBE-UNKNOWN ,FORM)) ) )

(DEFUN PROBE-QUOTED-ARG ( FORM )
  (LET (( MACRO-ARG (ASSQ FORM *PROBING-MACRO-ARGS*) ))
    (UNLESS (NULL MACRO-ARG)
      ;; This is a macro argument; mark it to not be probed.
      (SETF (MACRO-ARG-USE MACRO-ARG) 'QUOTE) ) )
  FORM )

(DEFUN PROBE-PLACE ( FORM )
  ;; a generalized variable; can't place a probe here, but do want to be able to probe any subforms.
  (PROBE-QUOTED-ARG FORM)
  (LET* ((NEXT *PROBE-NEXT*)
	 (RESULT (PROBE-FORM FORM T)))
    (WHEN NEXT (SETQ *PROBE-NEXT* NEXT))
    RESULT))

(DEFUN PROBE-FORMS ( FORMS )
  (DECLARE COMPILER:(OPTIMIZE SPEED))
  (LOOP FOR FORM IN FORMS
        COLLECT (PROBE-FORM FORM) ) )

(DEFUN PROBE-FORMS-BEFORE-OR-AFTER ( FORMS PROBE-FIRST PROBE-AFTER )
  (DECLARE COMPILER:(OPTIMIZE SPEED))
  (LET ( NEW-FORMS )
    (WHEN PROBE-FIRST (SETQ *PROBE-NEXT* T))
    (SETQ NEW-FORMS 
	  (LOOP FOR FORM IN FORMS
		COLLECT (PROBE-FORM FORM) ) )
    (WHEN PROBE-AFTER (SETQ *PROBE-NEXT* T))
    NEW-FORMS ) )

(DEFMACRO PROBE-POINT ( NUMBER &OPTIONAL (FORM NIL FORM-SUPPLIED) )
  ;;  6/05/87 DNG - Use SETF instead of ASET to avoid obsolete warnings when probing kernel code.
  (IF FORM-SUPPLIED
      `(PROGN (SETF (AREF .PROBE-ARRAY. ,NUMBER) 1)
              ,FORM)
    `(SETF (AREF .PROBE-ARRAY. ,NUMBER) 1) ) )

(DEFSUBST PROBE-UNKNOWN ( FORM )
  ;; This marker is needed only in the interpreted definition,
  ;; it can be removed from the compiled object code.
  FORM )

#+Symbolics
(DEFUN INTERPRETED-DEFINITION ( FNAME )
  ;; Given a function spec, return its interpreted definition.
  ;; 01/27/87 CLM - Obsolete. Now using COMPILEDP.
  (LET* ((FSPEC (SI:UNENCAPSULATE-FUNCTION-SPEC FNAME))
	 ( FDEF (FDEFINITION FSPEC) ))
    (IF (AND (CONSP FDEF)
	     (OR (NEQ (CAR FDEF) 'MACRO)
		 (CONSP (CDR FDEF)) ))
	FDEF
      ;;clm - debug-info now a struct
      #-Symbolics
      (si:get-debug-info-field  (si:get-debug-info-struct fdef)
				:INTERPRETED-DEFINITION)
      #+Symbolics
      (SECOND (ASSQ #+Symbolics ':INTERPRETED-DEFINITION
		    (DEBUGGING-INFO FDEF) ) ) ) ) )

(DEFUN PROBE-ENTRY ( NAME NUMBER )
  ;; 2/14/85 - Use interpreted def from function in preference to the
  ;;           one recorded when the function was probed.
  ;; 1/27/87 CLM - Use COMPILEDP to obtain interpreted definition.
  (DECLARE COMPILER:(OPTIMIZE (SPEED 2) (SPACE 1)))
  (LET ( TEM ARRAY )
    (SETQ TEM (FIND-PROBE-DATA NAME))
    (WHEN (OR (NULL TEM)
	      (NOT (ARRAYP (SETQ ARRAY (PROBE-DATA-ARRAY TEM))))
	      (NOT (= (1+ NUMBER) (ARRAY-LENGTH ARRAY))) )
      (LET (( DEFAULT-CONS-AREA BACKGROUND-CONS-AREA ))
	(SETQ ARRAY
	      (MAKE-ARRAY (1+ NUMBER) ':TYPE 'ART-1B :INITIAL-VALUE 0) )
	(IF (NULL TEM)
	    (ADD-PROBE-DATA (SETQ TEM (MAKE-PROBE-DATA PROBE-DATA-NAME NAME
				   PROBE-DATA-ARRAY ARRAY)))
	  (SETF (PROBE-DATA-ARRAY TEM) ARRAY)
	  (SETF (PROBE-DATA-COUNT TEM) 0) )
	(LET (( IDEF (COMPILEDP NAME) ))
	  (WHEN (CONSP IDEF)
	    (SETF (PROBE-DATA-INTERP-DEF TEM)
		  IDEF) ) )
      ) )
    (INCF (PROBE-DATA-COUNT TEM))
    ARRAY
    ) )

(defparameter *probe-now* t)

(defparameter *dummy-point-array* (MAKE-ARRAY 1000 ':TYPE 'ART-1B :INITIAL-VALUE 0))

(defmacro with-probes (&body body)
  "Execute body forms with collection of probe data enabled."
  `(let ((*probe-now* t))
     . ,body))

(DEFUN CONDITIONAL-PROBE-ENTRY ( NAME NUMBER )
  ;;  Used instead of PROBE-ENTRY to make collection of probe data conditional
  ;;  on *PROBE-NOW*.  This is needed for probing low-level system code that 
  ;;  could be called unexpectedly from places other than the test program.
  ;;
  ;; 6/05/87 DNG - Original.
  (DECLARE (OPTIMIZE (SPEED 2) (SPACE 1) (SAFETY 0)))
  (if *probe-now*
      (LET ((*probe-now* nil)) ; avoid recursive entry
	(PROBE-ENTRY NAME NUMBER))
    *dummy-point-array*))

;; The following is a kludge to prevent COMPILER:CHECK-COLD from writing out 
;; warnings that PROBE-ENTRY is not in the cold band when inserting probes in
;; a cold-load file.
#+TI
(WHEN SYS:FDEFINE-FILE-PATHNAME
  (PUSHNEW SYS:FDEFINE-FILE-PATHNAME COMPILER:COLD-LOAD-FILES))

;;;
;;;   =====   handlers for inserting probes in special forms   =====
;;;
(DEFUN (:PROPERTY SI:DISPLACED PROBES) ( FORM )
  ;; Probe the original form and discard the macro expansion.
  (PROBE-FORM (SECOND FORM)) )

(DEFPROP PROGN PROBE-ORDINARY PROBES)
(DEFPROP DONT-OPTIMIZE PROBE-ORDINARY PROBES)
(DEFPROP FUNCALL-WITH-MAPPING-TABLE PROBE-ORDINARY PROBES)
(DEFPROP LEXPR-FUNCALL-WITH-MAPPING-TABLE PROBE-ORDINARY PROBES)
(DEFPROP FUNCALL-SELF PROBE-ORDINARY PROBES)
(DEFPROP LEXPR-FUNCALL-SELF PROBE-ORDINARY PROBES)
(DEFPROP INCF PROBE-ORDINARY PROBES)
(DEFPROP DECF PROBE-ORDINARY PROBES)
(DEFPROP REMF PROBE-ORDINARY PROBES)
(DEFPROP SIGNP PROBE-ORDINARY PROBES)
(DEFPROP MULTIPLE-VALUE-CALL PROBE-ORDINARY PROBES)
(DEFPROP MULTIPLE-VALUE-LIST PROBE-ORDINARY PROBES)
(DEFPROP MULTIPLE-VALUE-PROG1 PROBE-ORDINARY PROBES)
(DEFPROP PROG1 PROBE-ORDINARY PROBES)
(DEFPROP PROG2 PROBE-ORDINARY PROBES)
(DEFPROP PROGV PROBE-ORDINARY PROBES)
(DEFPROP PROGW PROBE-ORDINARY PROBES)
(DEFPROP NTH-VALUE PROBE-ORDINARY PROBES)
(DEFPROP UNWIND-PROTECT PROBE-ORDINARY PROBES)
(DEFPROP VARIABLE-BOUNDP PROBE-ORDINARY PROBES)
(DEFPROP VARIABLE-LOCATION PROBE-ORDINARY PROBES)
(DEFPROP global:EVERY PROBE-ORDINARY PROBES)
(DEFPROP global:SOME PROBE-ORDINARY PROBES)
(DEFPROP SEND PROBE-ORDINARY PROBES)
(DEFPROP LEXPR-SEND PROBE-ORDINARY PROBES)
(DEFPROP POP PROBE-ORDINARY PROBES)
(DEFPROP LOCF PROBE-ORDINARY PROBES)
(DEFPROP LOCALLY PROBE-ORDINARY PROBES)
(DEFPROP *THROW PROBE-ORDINARY PROBES)
(DEFPROP WITHOUT-INTERRUPTS PROBE-ORDINARY PROBES)
#-Symbolics
(DEFPROP CLI:THROW PROBE-ORDINARY PROBES)
(DEFUN PROBE-ORDINARY ( FORM )
  ;; Handler for special forms or macros that can be treated as ordinary functions
  ;; because always evaluate all of their arguments in normal left-to-right order.
  (CONS (FIRST FORM) (PROBE-FORMS (REST1 FORM))) )

(DEFPROP SETQ PROBE-SETQ PROBES)
(DEFPROP SETF PROBE-SETQ PROBES)
(DEFPROP PSETQ PROBE-SETQ PROBES)
(DEFPROP PSETF PROBE-SETQ PROBES)
(DEFUN PROBE-SETQ ( FORM )
  (CONS (FIRST FORM)
	(LOOP FOR ARGS FIRST (REST1 FORM) THEN (REST2 ARGS)
	      UNTIL (NULL ARGS)
	      COLLECT (PROBE-PLACE (FIRST ARGS))
	      COLLECT (PROBE-FORM (SECOND ARGS)) ) ) )

(DEFPROP PUSH PROBE-PUSH PROBES)
(DEFPROP PUSHNEW PROBE-PUSH PROBES)
(DEFPROP PUSH-END PROBE-PUSH PROBES)
(DEFUN PROBE-PUSH ( FORM )
  (LIST* (FIRST FORM)				; PUSH
	 (PROBE-FORM (SECOND FORM))		; item
	 (PROBE-PLACE (THIRD FORM))		; destination
	 (PROBE-FORMS (REST3 FORM))))		; options

(DEFPROP LET LET-PROBE PROBES)
(DEFPROP LET* LET-PROBE PROBES)
(DEFPROP COMPILER-LET LET-PROBE PROBES)
(DEFUN LET-PROBE ( FORM )
  (LIST* (FIRST FORM)
	 (PROBE-BIND-LIST (SECOND FORM))
         (PROBE-FORMS (REST2 FORM))) )

(DEFUN (:PROPERTY LET-IF PROBES) ( FORM )
  (LET (( TEST (PROBE-FORM (SECOND FORM)) )
	RESULT )
    (SETQ *PROBE-NEXT* T)
    (SETQ RESULT (LIST* (FIRST FORM)
			TEST
			(PROBE-BIND-LIST (THIRD FORM))
			(PROBE-FORMS (REST3 FORM))) )
    (SETQ *PROBE-NEXT* T)
    RESULT ) )

(DEFUN (:PROPERTY FUNCTION PROBES) ( FORM )
  ;; 2/14/85 - Use COPY-TREE to make sure form is in a safe area.
  (LET (( F (SECOND FORM) ))
    (IF (AND (CONSP F)
	     #-Symbolics (MEMQ (FIRST F) '(LAMBDA CLI:LAMBDA))
             #+Symbolics (EQ (FIRST F) 'LAMBDA) )
	(PROG2 (SETQ *PROBE-NEXT* T)
	       (LIST 'FUNCTION
		     (LET-PROBE F))
	       (SETQ *PROBE-NEXT* T) )
      (COPY-TREE FORM) ) ) )

(DEFPROP FLET PROBE-FLET PROBES)
(DEFPROP LABELS PROBE-FLET PROBES)
(DEFUN PROBE-FLET ( FORM )
  (LIST* (FIRST FORM)
	 (LOOP FOR XF IN (SECOND FORM)
	       COLLECT (LIST* (FIRST XF)
			      (PROBE-BIND-LIST (SECOND XF))
			      (PROBE-FORMS-BEFORE-OR-AFTER (REST2 XF) T T) ) )
         (PROBE-FORMS (REST2 FORM)) )
  )
(DEFPROP MACROLET PROBE-MACROLET PROBES)
(DEFUN PROBE-MACROLET ( FORM )
  (LIST* (FIRST FORM)
	 (SECOND FORM)
         (PROBE-FORMS (REST2 FORM)) )
  )

(DEFPROP MULTIPLE-VALUE-BIND PROBE-QUOTE-SECOND PROBES)
(DEFPROP MULTIPLE-VALUE-SETQ PROBE-QUOTE-SECOND PROBES)
(DEFPROP MULTIPLE-VALUE      PROBE-QUOTE-SECOND PROBES)
(DEFPROP EVAL-WHEN           PROBE-QUOTE-SECOND PROBES)
(DEFPROP THE                 PROBE-QUOTE-SECOND PROBES)
(DEFUN PROBE-QUOTE-SECOND ( FORM )
  ;; Insert probes in special form with quoted first argument.
  (LIST* (FIRST FORM)
	 (SECOND FORM)
	 (PROBE-FORMS (REST2 FORM)) ) )

(DEFUN (:PROPERTY IF PROBES) ( FORM )
  (LIST* (FIRST FORM)				; IF
	 (PROBE-FORM (SECOND FORM))		; test clause
	 (PROBE (THIRD FORM))			; then clause
	 (PROBE-FORMS-BEFORE-OR-AFTER (REST3 FORM) T T)	; else clauses(s)
 ) )

(DEFPROP COND COND-PROBE PROBES)
(DEFPROP COND-EVERY COND-PROBE PROBES)
(DEFUN COND-PROBE ( FORM )
  (CONS (FIRST FORM)
    (LOOP FOR CLAUSE IN (REST1 FORM)
          COLLECT (LIST* (PROBE-FORM (FIRST CLAUSE))
			 (PROBE-FORMS-BEFORE-OR-AFTER (REST1 CLAUSE) T T) )
 ) ) )

(DEFPROP BLOCK PROBE-BLOCK PROBES)
(DEFPROP *CATCH PROBE-BLOCK PROBES)
#-Symbolics (DEFPROP CLI:CATCH PROBE-BLOCK PROBES)
(DEFUN PROBE-BLOCK ( FORM )
  (LIST* (FIRST FORM)				; BLOCK
	 (PROBE-QUOTED-ARG (SECOND FORM))	; name of block
	 (PROBE-FORMS-BEFORE-OR-AFTER (REST2 FORM) NIL T)	; body forms
	 ) )
(DEFPROP AND AND-OR-PROBE PROBES)
(DEFPROP OR  AND-OR-PROBE PROBES)
(DEFUN AND-OR-PROBE ( FORM )
  (LIST* (FIRST FORM)
         (LOOP FOR ARG IN (REST1 FORM)
               COLLECT (PROBE-FORM ARG)
	       DO (SETQ *PROBE-NEXT* T) ) ) )

(DEFPROP WHEN   WHEN-UNLESS-PROBE PROBES)
(DEFPROP UNLESS WHEN-UNLESS-PROBE PROBES)
(DEFUN WHEN-UNLESS-PROBE ( FORM )
  (LIST* (FIRST FORM) 
	 (PROBE-FORM (SECOND FORM))
	 (PROBE-FORMS-BEFORE-OR-AFTER (REST2 FORM) T T)) )

(DEFPROP CASE CASE-PROBE PROBES)
(DEFPROP ECASE CASE-PROBE PROBES)
(DEFPROP CCASE CASE-PROBE PROBES)
(DEFPROP TYPECASE CASE-PROBE PROBES)
(DEFPROP ETYPECASE CASE-PROBE PROBES)
(DEFPROP CTYPECASE CASE-PROBE PROBES)
(DEFPROP SELECT CASE-PROBE PROBES)
(DEFPROP SELECTQ CASE-PROBE PROBES)
(DEFPROP SELECTQ-EVERY CASE-PROBE PROBES)
(DEFPROP CASEQ CASE-PROBE PROBES)
(DEFUN CASE-PROBE ( FORM )
  (SETQ *PROBE-NEXT* T)
  (LIST* (FIRST FORM)
	 (PROBE-FORM (SECOND FORM))
         (LOOP FOR CLAUSE IN (REST2 FORM)
               COLLECT (CONS (PROBE-QUOTED-ARG (FIRST CLAUSE))
                             (PROBE-FORMS-BEFORE-OR-AFTER (REST1 CLAUSE) T T)) ) ) )

(DEFPROP SELECTOR SELECTOR-PROBE PROBES)
(DEFPROP DISPATCH SELECTOR-PROBE PROBES)
(DEFUN SELECTOR-PROBE ( FORM )
  (LIST* (FIRST FORM)
	 (PROBE-FORM (SECOND FORM))
	 (THIRD FORM)
         (LOOP FOR CLAUSE IN (REST3 FORM)
               COLLECT (CONS (FIRST CLAUSE)
                             (PROBE-FORMS-BEFORE-OR-AFTER (REST1 CLAUSE) T T))
	       ) ) )
(DEFUN (:PROPERTY SELECT-MATCH PROBES) ( FORM )
  (LIST* (FIRST FORM)
	 (PROBE-FORM (SECOND FORM))
         (LOOP FOR CLAUSE IN (REST2 FORM)
               COLLECT (LIST* (FIRST CLAUSE)
			      (SECOND CLAUSE)
                              (PROBE-FORMS-BEFORE-OR-AFTER (REST2 CLAUSE) T T))
	       ) ) )

;;added these to test problem
(DEFPROP TAGBODY PROBE-TAG-BODY PROBES)
(DEFPROP DO-FOREVER PROBE-TAG-BODY PROBES)
;;this is the current version
;;(DEFPROP TAGBODY PROBE-TAGBODY PROBES)
;;(DEFPROP DO-FOREVER PROBE-TAGBODY PROBES)
(DEFUN (:PROPERTY TAGBODY PROBES) ( FORM )
  (CONS (FIRST FORM)
	(PROBE-TAG-BODY (REST1 FORM)) ) )

(DEFUN PROBE-TAG-BODY ( FORMS )
  (PROG1
    (LOOP FOR FORM IN FORMS
	  IF (ATOM FORM)
	  COLLECT (PROG1 FORM (SETQ *PROBE-NEXT* T) )
	  ELSE COLLECT (PROBE-FORM FORM) )
    (SETQ *PROBE-NEXT* T) ) )

(DEFUN PROBE-BIND-LIST ( VLIST )
  (LOOP FOR VAR IN VLIST
	COLLECT (IF (ATOM VAR)
		    (PROBE-QUOTED-ARG VAR)
		  (LIST* (PROBE-QUOTED-ARG (FIRST VAR))
			 (PROBE-FORM (SECOND VAR))
			 (REST2 VAR)) ) ) )

(DEFPROP PROG PROG-PROBE PROBES)
(DEFPROP PROG* PROG-PROBE PROBES)
(DEFUN PROG-PROBE ( FORM )
  (LET ( NAME VLIST BODY )
    (IF (AND (ATOM (SECOND FORM))
	     (NOT (NULL (SECOND FORM))))
	(SETQ NAME (SECOND FORM)
	      VLIST (THIRD FORM)
	      BODY (REST3 FORM) )
      (SETQ VLIST (SECOND FORM)
	    BODY (REST2 FORM)) )
    (SETQ VLIST (PROBE-BIND-LIST VLIST))
    (SETQ BODY (PROBE-TAG-BODY BODY))
    (IF NAME
	(LIST* (FIRST FORM) NAME VLIST BODY)
      (LIST* (FIRST FORM) VLIST BODY) )
    ) )

(DEFPROP DO DO-PROBE PROBES)
(DEFPROP DO* DO-PROBE PROBES)
(DEFUN DO-PROBE ( FORM )
  (IF (AND (ATOM (SECOND FORM))			; "old style"
	   (NOT (NULL (SECOND FORM))))
      (LIST* (FIRST FORM)			; DO
	     (PROBE-QUOTED-ARG (SECOND FORM))	; var
	     (PROBE-FORM (THIRD FORM))		; init
	     (PROBE (FOURTH FORM))		; repeat
	     (PROBE (FIFTH FORM))		; end test
	     (PROGN (SETQ *PROBE-NEXT* T)
		    (PROBE-TAG-BODY (NTHCDR 5 FORM) ) ) )	; body
    ;; else "new style"
    (LIST* (FIRST FORM)
	   (LOOP FOR VAR IN (SECOND FORM)
		 COLLECT (IF (ATOM VAR)
			     (PROBE-QUOTED-ARG VAR)
			   (CONS (PROBE-QUOTED-ARG (FIRST VAR))
				 (PROBE-FORMS (REST1 VAR)) )) )
	   (CONS (PROBE (FIRST (THIRD FORM)))
		 (PROBE-FORMS-BEFORE-OR-AFTER (REST1 (THIRD FORM)) T T))
	   (PROBE-TAG-BODY (REST3 FORM))) ) )

(DEFPROP DO-NAMED DO-NAMED-PROBE PROBES)
(DEFPROP DO*-NAMED DO-NAMED-PROBE PROBES)
(DEFUN DO-NAMED-PROBE ( FORM )
  (LIST* (FIRST FORM)
	 (SECOND FORM)
	 (REST1 (DO-PROBE (CONS 'DO (REST2 FORM)))) ) )
(DEFPROP DOLIST PROBE-DOLIST PROBES)
(DEFPROP DOTIMES PROBE-DOLIST PROBES)
(DEFPROP DO-SYMBOLS PROBE-DOLIST PROBES)
(DEFPROP DO-EXTERNAL-SYMBOLS PROBE-DOLIST PROBES)
(DEFPROP DO-ALL-SYMBOLS PROBE-DOLIST PROBES)
(DEFUN PROBE-DOLIST ( FORM )
  (LIST* (FIRST FORM)
	 (LIST* (PROBE-QUOTED-ARG (FIRST (SECOND FORM)))
		(PROBE-FORM (SECOND (SECOND FORM)))
		(PROBE-FORMS-BEFORE-OR-AFTER (REST2 (SECOND FORM)) T T))
	 (PROBE-TAG-BODY (REST2 FORM))) )


(DEFUN (:PROPERTY LOOP PROBES) ( FORM )
  ;;  1/03/85 - Fixed to handle destructuring.
  (PROG1 (LOOP WITH DONT-EVAL = NIL
	       FOR X IN FORM
	       IF (OR (ATOM X) DONT-EVAL)
	       COLLECT X
	       ELSE COLLECT (PROBE X)
	       WHEN (SYMBOLP X)
	       DO (SETQ DONT-EVAL
			(MEMBER (GET-PNAME X)
				'( "FOR" "WITH" "AS" "AND" ))) )
	 (SETQ *PROBE-NEXT* T) ) )

(DEFUN (:PROPERTY GO PROBES) (FORM)
  (SETQ *PROBE-NEXT* T)
  FORM )

(DEFPROP RETURN PROBE-RETURN PROBES)
(DEFPROP RETURN-FROM PROBE-RETURN PROBES)
(DEFUN PROBE-RETURN ( FORM )
  (CONS (FIRST FORM)
	(PROBE-FORMS-BEFORE-OR-AFTER (REST1 FORM) NIL T) ) )

(DEFPROP WITH-STACK-LIST  PROBE-WITH-STACK-LIST PROBES)
(DEFPROP WITH-STACK-LIST* PROBE-WITH-STACK-LIST PROBES)
(DEFUN PROBE-WITH-STACK-LIST ( FORM )
  (LIST* (FIRST FORM)
	 (CONS (FIRST (SECOND FORM))
	       (PROBE-FORMS (REST1 (SECOND FORM))) )
	 (PROBE-FORMS (REST2 FORM)) ) )

(DEFUN (:PROPERTY QUOTE PROBES) ( FORM )
  ;; 2/14/85 - Use COPY-TREE to put in safe area.
  (PROBE-QUOTED-ARG (SECOND FORM))
  (COPY-TREE FORM) )

(DEFPROP DECLARE IDENTITY PROBES)
(DEFPROP COMMENT IDENTITY PROBES)

;;;  ====   I/O details   ====

(DECLARE (SPECIAL *HEADLINE-FONT* *LIGHT-FONT* *HEAVY-FONT* SI:*PP-CURRENT-FONT*))
(DEFUN PROBE-REPORT-STREAM ( &REST ARGS )
  (DECLARE (SPECIAL *PROBE-OUTPUT*))
  (IF (AND (EQ (FIRST ARGS) ':SET-CURRENT-FONT)
	   (KEYWORDP (SECOND ARGS)) )
      (SETQ *PRINT-CASE* (SECOND ARGS))
    (APPLY *PROBE-OUTPUT* ARGS) ) )

;;;
;;;   =====   Report probe data   =====
;;;
(DEFUN PROBE-REPORT (&QUOTE &REST FUNCTIONS)
  "Report the probe information that has been collected.
The arguments are unevaluated function specs whose probe data is desired.
When no arguments are given or a single argument of T is given, report
data for all probed functions.  The function definition is displayed 
with exercised paths in lower case and unused paths in upper case when
not all paths have been used and at least one argument has been supplied
to this function.   The information is written to *STANDARD-OUTPUT*."
  ;; This function is for direct calling by the user from a Lisp Listener.
  ;; It uses letter case selection instead of font changes.
  (LET (( *PROBE-OUTPUT* *STANDARD-OUTPUT* ) ; for PROBE-REPORT-STREAM
	( *PRINT-CASE* *PRINT-CASE* )
	( *HEADLINE-FONT* ':UPCASE )
	( *LIGHT-FONT* ':DOWNCASE )
	( *HEAVY-FONT* ':UPCASE )
	( *STANDARD-OUTPUT* #'PROBE-REPORT-STREAM ))
    (DECLARE (SPECIAL *PROBE-OUTPUT*))
    (PROBE-REPORT-INTERNAL FUNCTIONS) ) )

(DEFUN PROBE-REPORT-INTERNAL ( FUNCTIONS )
  ;; This is called by the user-callable function PROBE-REPORT above and
  ;; from the UCL-window interface program.
  (IF (OR (NULL FUNCTIONS)
	  (EQUAL FUNCTIONS '(T))
	  (EQUAL FUNCTIONS '(0)))
      (LET (( DATA-LIST (PROBES-DATA-LIST) ))
	(WHEN (EQUAL FUNCTIONS '(0))
	  ;; show only functions that were never called
	  (SETQ DATA-LIST
		(DELETE-IF-NOT #'(LAMBDA (D) (ZEROP (PROBE-DATA-COUNT D)))
			       DATA-LIST) ) )
	(COND ((NULL DATA-LIST) (RETURN-FROM PROBE-REPORT-INTERNAL NIL))
	      ((NULL (REST1 DATA-LIST))	; if only one function
	       (PROBE-REPORT-2 (FIRST DATA-LIST) NIL) )
	      (T (LET (( SORTED-LIST (SORT DATA-LIST
					   #'SI:FUNCTION-SPEC-LESSP
					   :KEY #'PROBE-DATA-NAME) ))
		   (DOLIST ( TEM SORTED-LIST )
		     (PROBE-REPORT-2 TEM (IF (NULL FUNCTIONS)
					     'FALSE
					   NIL) ) ) ) ) ) )
    (DOLIST ( F FUNCTIONS )
      (PROBE-REPORT-1 (IF (AND (CONSP F)
			       (EQ (FIRST F) 'QUOTE))
			  (SECOND F)
			F ) ) ) )
  (VALUES) )

(DEFUN PROBE-REPORT-FREQUENCY ()
  "Report the probe information that has been collected
with the functions listed in order of number of times called."
  (LET (( DATA-LIST (PROBES-DATA-LIST) ))
    (UNLESS (NULL DATA-LIST)
      (LET (( SORTED-LIST (SORT DATA-LIST
				#'>
				:KEY #'PROBE-DATA-COUNT) ))
	(DOLIST ( TEM SORTED-LIST )
	  (PROBE-REPORT-2 TEM 'FALSE) ) )
      (VALUES) ) ) )

(DEFUN PROBE-REPORT-1 ( FSPEC &OPTIONAL ALWAYS-PRINT-DEF )
  "Report probe data for a function"
  (LET (( TEM (FIND-PROBE-DATA FSPEC) ))
    (IF (NULL TEM)
	(FORMAT T "~%There is no probe information recorded for ~S" FSPEC)
      (PROBE-REPORT-2 TEM (AND ALWAYS-PRINT-DEF 'TRUE)) ) )
  (VALUES) )

(DEFUN PROBE-REPORT-2 ( TEM &OPTIONAL PRINT-DEF )
  ;;  2/21/86 DNG - Shorten heading to allow more room for function name.
  ;; 01/27/87 CLM - Changed for Rel3.
  ;;  5/21/87 DNG - Fix to not error when COMPILEDP returns T or when a doc
  ;;		string is present.  Shorten heading a little more to enable
  ;;		printing on one line on 2015 printer.  Fix conversion of
  ;;		CLI:NAMED-LAMBDA to DEFUN form.
  (LET ( PA COUNT NUMBER-OF-PROBES NUMBER-OF-PATHS PATHS-TAKEN )
    (FORMAT T "~%~S~35,5T" (PROBE-DATA-NAME TEM))
    (SETQ PA (PROBE-DATA-ARRAY TEM))
    (IF (NUMBERP PA)
	(SETQ NUMBER-OF-PROBES PA)
      (SETQ NUMBER-OF-PROBES (- (ARRAY-LENGTH PA) 1)) )
    (SETQ NUMBER-OF-PATHS (+ NUMBER-OF-PROBES 1))
    (SETQ COUNT 0)
    (IF (OR (NUMBERP PA) (ZEROP (PROBE-DATA-COUNT TEM)))
	(SETQ PATHS-TAKEN 0)
      (PROGN
	(LOOP FOR N FROM 1 TO NUMBER-OF-PROBES
	      WHEN (NOT (ZEROP (AREF PA N)))
	      DO (INCF COUNT) )
	(SETQ PATHS-TAKEN (+ COUNT 1)) ) )
    (FORMAT T "called~6D times, using ~3D% (~3D//~3D) of branches."
	    (PROBE-DATA-COUNT TEM)
	    (ROUND (%DIV (* PATHS-TAKEN 100.) NUMBER-OF-PATHS))
	    PATHS-TAKEN
	    NUMBER-OF-PATHS )
    (UNLESS (OR (EQ PRINT-DEF 'FALSE)
		(AND (NEQ PRINT-DEF 'TRUE)
		     (OR (>= COUNT NUMBER-OF-PROBES)
			 (NOT (ARRAYP PA))) ) )
      (LET* ( ( *PRINT-CASE* :DOWNCASE)
	      ( SI:*PP-CURRENT-FONT* *LIGHT-FONT*)   ;;fonts for PROBER
	      ( *GRIND-PROBE-ARRAY* (AND (ARRAYP PA) PA) )
	      ( INTERP-DEF (OR (PROBE-DATA-INTERP-DEF TEM)
			       (COMPILEDP (PROBE-DATA-NAME TEM))
			       ) ))
	(SEND *STANDARD-OUTPUT*
	      ':SET-CURRENT-FONT (IF (ZEROP (PROBE-DATA-COUNT TEM))
				     *HEAVY-FONT*
				   *LIGHT-FONT*))
	(MULTIPLE-VALUE-BIND (BODY-FORMS DECLS DOC)
	    (SI:PARSE-BODY (AND (CONSP INTERP-DEF) (REST3 INTERP-DEF)) NIL T)
	  (DECLARE (IGNORE DECLS DOC))
	  (LET ((BODY (FIRST BODY-FORMS)))
	    (IF (AND (CONSP BODY)
		     (EQ (FIRST BODY) 'LET)
		     (EQ (FIRST (FIRST (SECOND BODY))) '.PROBE-ARRAY.)
		     (NULL (REST BODY-FORMS)))
		;;clm - use PPRINT instead of grinding
		(PPRINT
		  (LIST* (SELECTQ (FIRST INTERP-DEF)
				  ( ( #!z NAMED-LAMBDA #!c NAMED-LAMBDA ) 'DEFUN )
				  ( ( #!z NAMED-SUBST  #!c NAMED-SUBST ) 'DEFSUBST )
				  ( OTHERWISE (FIRST INTERP-DEF) ) )
			 (SECOND INTERP-DEF)	; function name
			 (THIRD INTERP-DEF)	; argument list
			 (LET* (( LET-BODY #-Symbolics
				 (SI:PARSE-BODY (REST2 BODY) NIL T)
				 #+Symbolics (REST2 BODY) )
				( FIRST-FORM (FIRST LET-BODY)) )
			   (IF (AND (CONSP FIRST-FORM)
				    (EQ (FIRST FIRST-FORM) 'BLOCK)
				    (EQ (SECOND FIRST-FORM) (SECOND INTERP-DEF)) )
			       (REST2 FIRST-FORM)
			     LET-BODY ) ) ) 
		  )
	      (PPRINT INTERP-DEF) )))
	(TERPRI)
	(SEND *STANDARD-OUTPUT* ':SET-CURRENT-FONT *HEADLINE-FONT*) ) ) )
  (VALUES) )

;;;  ============================================================
;;;==============================================================================================
;;;changes needed in PPRINT
;;;==============================================================================================
;;clm - A special pprint-handler for probe-points in probe reports
si:
(defun pp-probe-point (object location currlevel)
  ;; The following is needed to prevent the 
  ;; PROBE-POINTs from appearing on the display in certain contexts.
  (declare (ignore location))
  (when compiler:*GRIND-PROBE-ARRAY*      ;;within PROBE
    (if (typep *pp-current-font* 'font)
	(setf *pp-current-font* (IF (ZEROP (AREF compiler:*GRIND-PROBE-ARRAY* (SECOND object)))
				    compiler:*HEAVY-FONT*  ; for un-executed code
				    compiler:*LIGHT-FONT*  ; for executed code
				    ))
	(setf *print-case* (IF (ZEROP (AREF compiler:*GRIND-PROBE-ARRAY* (SECOND object)))
			       :upcase	   ;compiler:*HEAVY-FONT* ; for un-executed code
			       :downcase   ;compiler:*LIGHT-FONT* ; for executed code
			       )) ) )

  (pp-objify (third object) (locf (third object)) currlevel) ) 


si:
(defprint compiler:probe-point (pprint-handler pp-probe-point))




