;;;  -*- Mode:Common-Lisp; Package:IP; Base:10 -*-
;;;
;;;  Functions to do rsh from the Explorer, and some useful commands using it.

(export '(rsh cscwho lpq vmsq mailq ruptime rwho mailbox)
	'ticl)

(defvar *remote-user-id* nil "Either NIL, if the unix user-id is the same as the lispm one,
or a string containing the unix user-id.")

(DEFUN rsh (host shell-command &key (lispm-user-id user-id)
				    (unix-user-id (or *remote-user-id* user-id))
				    (output-stream *standard-output*))
  "On the HOST, execute SHELL-COMMAND using the Berkeley \"rsh\" protocol.
Optionally specify your LISPM-USER-ID on the host or forge a different UNIX-USER-ID.
The command output goes to OUTPUT-STREAM, which defaults to *standard-output*."
  (LET ((*print-base* 10)
	(*read-base* 10)
	(local-port nil))
    (unwind-protect
	(progn
	  (setq local-port (generate-port-number-in-range 256 1024))
	  (WITH-OPEN-STREAM (stream (open-stream host
						 :remote-port 514
						 :local-port local-port
						 :timeout 60
						 :error t
						 :characters :ascii))
	    ;; send error port number.
	    (write-char 0 stream)		   ; Error port would be here, but let's blow it off.
	    
	    ;; send lispm user-id
	    (write-string lispm-user-id stream)
	    (write-char 0 stream)
	    
	    ;; send unix user-id
	    (write-string unix-user-id stream)
	    (write-char 0 stream)
	    
	    ;; send command
	    (write-string shell-command stream)
	    (write-char 0 stream)
	    (force-output stream)
	    
	    ;; first char is 0 if successful, 1 if not. 
	    (LET ((CHAR (read-char stream)))
	      (unless (zerop (char-code char))
		(format output-stream "~&Error in rsh:")
		(unless (= (code-char char) 1)	; Paranoia.
		  (unread-char char stream))))
	    
	    ;; translate the output
	    (let ((line nil)
		  (eof  nil))
	      (loop (multiple-value-setq (line eof)
		      (read-delimited-string #\linefeed stream nil))
		    (when eof
		      (terpri output-stream)
		      (return))
		    (when line
		      (fresh-line output-stream)
		      (write-string line output-stream))))
	    (values)))
      (when local-port
	(unreserve-tcp-port local-port)))))

(defun generate-port-number-in-range (min max)
   "Generates and reserves a TCP port between MIN and MAX, inclusive.
Caller must unreserve the port when finished."
   (assert (<= min max) (min max)
	   "Minimum port number, ~D, must be less than or equal to maximum, ~D." min max)
   (loop repeat 1000
	 as port = (generate-tcp-port)
	 when (<= min port max)
	 return port
	 else do (unreserve-tcp-port port)
	 finally (error "Cannot get TCP port between ~D and ~D." min max)))

(defvar *default-rsh-host* "aim")

(defun rsh-with-switches (host command switches)
   (rsh host (format nil "~A ~A" command (or switches ""))))

(defun cscwho (name &optional (remote-host *default-rsh-host*))
   (rsh-with-switches remote-host "/usr/local/bin/cscwho" name))

(defun lpq (&optional (switches "-Pip") (remote-host *default-rsh-host*))
   (rsh-with-switches remote-host "/usr/ucb/lpq" switches))

(defun mailbox (address &optional (remote-host *default-rsh-host*))
   (rsh-with-switches remote-host "/usr/local/bin/mailbox" address))

(defun mailq (&optional (remote-host *default-rsh-host*))
   (rsh remote-host "/usr/local/bin/mailq"))

(defun ruptime (&optional (switches "") (remote-host *default-rsh-host*))
   (rsh-with-switches remote-host "/usr/ucb/ruptime" switches))

(defun rwho (&optional (switches "") (remote-host *default-rsh-host*))
   (rsh-with-switches remote-host "/usr/ucb/rwho" switches))

(defun vmsq (&optional (remote-host *default-rsh-host*))
   (rsh remote-host "/usr/local/bin/vmsq"))

