;;; -*- Mode:Common-Lisp; Package:USER; Fonts:(MEDFNB TR12BI); Base:10 -*-

;;;===============================================================================
;;;                                    RESTRICTED RIGHTS LEGEND 
;;; Use,  duplication, or  disclosure  by  the  Government is subject to restrictions
;;; as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data and
;;; Computer Software clause at 52.227-7013. 
;;;
;;; TEXAS INSTRUMENTS INCORPORATED, P.O. BOX 2909 AUSTIN, TEXAS 78769  
;;; Copyright (C) 1984, Texas Instruments Incorporated. All rights reserved.
;;;===============================================================================

;; Blank the screen after SCREEN-SAVER-TIME-DELAY (default 20) minutes.

;;;
;;; Change history:
;;;
;;;  Date	Author	Description
;;; -------------------------------------------------------------------------------------
;;;  7/07/88    TWE	Added comments after viewing the July 4 fireworks.
;;;  6/22/88    TWE	Added a gravity pattern and grays to simulate an ember dying out.
;;;  5/17/88    TWE	Created.


;;; Simulates fireworks as a screen saver hack.


;;; If you have trouble running this as a screen saver, then you probably need to
;;; 1RELOAD THE SCREEN-SAVER* file.  The old screen saver file uses an old window
;;; system flavor which causes problems here.


;;; Parameters that are handled here are:
;;;   Size of final explosion
;;;   Height reached by rocket
;;;   Direction of launch
;;;   Speed of launch
;;;   Pattern generated
;;;   Gravity
;;;   Ember of explosion fading out


;;; The following are the author's comments written after seeing the July 4th
;;; fireworks.  Although there is a great variety in fireworks, these comments
;;; pertain to those that are initially spherical in shape.
;;;
;;; The fireworks implemented here are three dimensional (X,Y,Time), but
;;; fireworks in real life have four dimensions (X,Y,Z,Time).  What needs to be
;;; done is to change the simple explosion and gravity explosion to include the
;;; Z component when calculating points, and then throw the Z component away to
;;; display the result.  This makes the viewer's eye's location be at an
;;; infinite distance away along positive the Z axis.
;;;
;;; Another problem relates to the embers.  This implementation uses embers
;;; that are really rectangles of a recognizable size.  In real fireworks, the
;;; embers are normally quite small.  Changing the embers to be much smaller
;;; and traversing a larger distance should correct this problem.  Another
;;; related problem is the time at which the embers are visible.  In real
;;; fireworks they are usually visible after they have traveled about 1/3 of
;;; their path, with the ember particles being visible only in the latter 2/3
;;; of their path.  Another kind of burning was observed where embers were
;;; left behind while the main ember projectiles were moving toward their
;;; destination.  If the enbers burned fast enough, this gave the appearance of
;;; lines emanating from a central point; slow burning embers generated arcs as
;;; gravity influenced their trajectory.  All of these problems: smaller
;;; embers, linear embers and arcing embers are doable.
;;;
;;; There are various kinds of rocket launchings.  The implementation here
;;; assumes that all launchings are the same.  In real fireworks the rocket
;;; launchings vary from being invisible, slightly visible, very visible, and
;;; very visible spiral path.  For the purposes of this program, an invisible
;;; launching is probably not appropriate.  A slightly visible or very visible
;;; rocket burning is doable, while a very visible spiral path is probably not
;;; work the effort of implementing.  A very visible burning is where embers
;;; are emitted as the rocket is moving, forming an image of the rockets path
;;; for a few seconds before the embers decay.  This involves saving the points
;;; of the path, much like the qix program does for the endpoints of its lines.
;;; The implementation done by the qix program is probably appropriate as it
;;; does not generate lots of garbage.  An array implementation may be
;;; appropriate too.
;;;
;;; In this implementation of fireworks there is no sound.  This probably
;;; should be an option, as the sound of fireworks may be disturbing in a
;;; terminal room.  There should be sound for three things: the initial
;;; launching (a kind of swoosh), the launching itself (very faint), and the
;;; final explosion.  The initial launchings are pretty much identical, so that
;;; should be simple.  The sound for the launching itself could be
;;; non-existent, or faint.  There was much variation in the final explosion,
;;; from just a pop to a loud boom.


(DEFPARAMETER EXPLOSION-ARRAY-LIST NIL
  "List of explosion arrays used to draw an explosion pattern.")

(DEFPARAMETER *DEFAULT-LAUNCHER-COUNT* 3
  "Default number of rocket launchers in fireworks display.")

(DEFPARAMETER GRAYS-ARRAY (LET ((ARRAY (MAKE-ARRAY 8)))
                            (LOOP FOR GRAY IN (LIST W:100%-BLACK W:88%-GRAY W:75%-GRAY W:66%-GRAY
                                                    W:50%-GRAY W:33%-GRAY W:25%-GRAY W:12%-GRAY)
                                  FOR INDEX FROM 0 BY 1
                                  DO (SETF (AREF ARRAY INDEX) GRAY))
                            ARRAY)
  "Array of grays used to simulate an ember dimming.")

(DEFUN LAUNCH-ROCKET (WINDOW X Y ANGLE SPEED MAX-DISTANCE)
  "Simulate a rocket launch starting from (X,Y) in direction ANGLE at SPEED out to MAX-DISTANCE.
ANGLE is in degrees, and SPEED is the number of 60ths of a second to wait between
drawing of each piece of the rocket trail."
  (DECLARE (TYPE W:MINIMUM-WINDOW WINDOW)
           (TYPE NUMBER X Y ANGLE SPEED MAX-DISTANCE)
           (VALUES LAST-X LAST-Y))
  ;; Divide the distance into reasonable chunks, and show a track for each one.
  (LOOP WITH PIXEL-INCREMENT = 10
        WITH DELTA-X = (* PIXEL-INCREMENT (COSD ANGLE))
        WITH DELTA-Y = (* PIXEL-INCREMENT (SIND ANGLE))
        FOR DISTANCE FROM 0 BELOW MAX-DISTANCE BY PIXEL-INCREMENT
        FOR CURRENT-X FIRST X THEN NEXT-X
        FOR CURRENT-Y FIRST Y THEN NEXT-Y
        FOR NEXT-X = (+ CURRENT-X DELTA-X)
        FOR NEXT-Y = (- CURRENT-Y DELTA-Y)
        DO (PROGN
	     ;; Draw the trace of the rocket's engine, wait a little while and then
	     ;; erase it.  The monitor's phosphors will take time to dim, making this
	     ;; a reasonable pattern.
	     ;; 1POSSIBLE ENHANCEMENTS:*  (1) Draw a curved line to simulate gravity.
	     ;; (2) Change the size of the increment according to the speed.
             (SEND WINDOW :DRAW-LINE (TRUNCATE CURRENT-X) (TRUNCATE CURRENT-Y)
                   (TRUNCATE NEXT-X) (TRUNCATE NEXT-Y) 2 W:BLACK W:ALU-XOR)
             (PROCESS-SLEEP SPEED "Launching")
             (SEND WINDOW :DRAW-LINE (TRUNCATE CURRENT-X) (TRUNCATE CURRENT-Y)
                   (TRUNCATE NEXT-X) (TRUNCATE NEXT-Y) 2 W:BLACK W:ALU-XOR))
        FINALLY (RETURN (VALUES NEXT-X NEXT-Y))))


;;; The main idea behind the following function is to make the drawing of the explosion
;;; pattern not take so long to compute.  The previous implementation used the :draw-line
;;; method.  That implementation caused the rocket launching to slow down noticably when
;;; an explosion pattern was drawn.  This function does more calculation ahead of time so
;;; that we can drop down into microcode when we want to draw the explosion pattern.
;;;
;;; This function takes the endpoints for a line segment and returns the four corners of
;;; a rectangle which is the equivalent graphics object.  By doing this we can convert a
;;; call to :draw-line into 2 calls to sys:%draw-shaded-triangle, which is much faster.
(ZWEI:DEFINE-INDENTATION CALCULATE-LINE-COORDINATES (1 1))
(DEFUN CONVERT-LINES-TO-RECTANGLES (LINE-SEGMENT-ARRAY)
  "Translate the array of line segment (x1,y1) (x2,y2) points to an array of rectangle 
points (x1,y1) (x2,y2) (x3,y3) (x4,y4) so we can use microcode to do the drawing."
  (LET ((RECTANGLE-POINTS-ARRAY (MAKE-ARRAY `(8 ,(ARRAY-DIMENSION LINE-SEGMENT-ARRAY 1)
                                                ,(1- (ARRAY-DIMENSION LINE-SEGMENT-ARRAY 2)))
                                            :ELEMENT-TYPE 'INTEGER)))
    (FLET ((CALCULATE-LINE-COORDINATES (X1 Y1 X2 Y2)
             "This calculates a set of 4 points which define a box which represents a line to draw"
             (LET ((THICKNESS 3)
                   ;; Use a window that has w:graphics-mixin as a component flavor.
                   (WINDOW W:INITIAL-LISP-LISTENER)
                   DX DY)
               (MULTIPLE-VALUE-SETQ (DX DY)
                 (SEND WINDOW :TRANSFORM-DELTAS THICKNESS THICKNESS))
               (IF (AND (>= 1.3 DX)
                        (>= 1.3 DY))
                   (PROGN
                     (MULTIPLE-VALUE-SETQ (X1 Y1)
                       (SEND WINDOW :TRANSFORM-POINT X1 Y1))
                     (MULTIPLE-VALUE-SETQ (X2 Y2)
                       (SEND WINDOW :TRANSFORM-POINT X2 Y2))
                     (LIST X1 Y1 X2 Y2 NIL NIL NIL NIL))
                   ;;ELSE
                   (PROGN
                     (SETQ THICKNESS (TRUNCATE THICKNESS 2))
                     (LET (X3 Y3 DX DY PX1 PY1 PX2 PY2 PX3 PY3 PX4 PY4)
                       (setq x3 x1
                             y3 y1)
                       (multiple-value-setq (dx dy)
                         (W:line-deltas x3 y3 x1 y1 x2 y2 thickness))
                       (setq px1 (- x1 dx)
                             py1 (- y1 dy)
                             px2 (+ x1 dx)
                             py2 (+ y1 dy))
                       (setq x3 x2
                             y3 y2)
                       (multiple-value-setq (dx dy)
                         (W:line-deltas x1 y1 x2 y2 x3 y3 thickness))
                       (setq px3 (- x2 dx)
                             py3 (- y2 dy)
                             px4 (+ x2 dx)
                             py4 (+ y2 dy))
                       (LIST PX1 PY1 PX3 PY3 PX4 PY4 PX2 PY2)))))))
      (LOOP
        FOR ITERATION-INDEX FROM 1 BELOW (ARRAY-DIMENSION LINE-SEGMENT-ARRAY 2)
        DO (LOOP
             FOR ANGLE-INDEX FROM 0 BELOW (ARRAY-DIMENSION LINE-SEGMENT-ARRAY 1)
             FOR X1 = (AREF LINE-SEGMENT-ARRAY 0 ANGLE-INDEX (1- ITERATION-INDEX))
             FOR Y1 = (AREF LINE-SEGMENT-ARRAY 1 ANGLE-INDEX (1- ITERATION-INDEX))
             FOR X2 = (AREF LINE-SEGMENT-ARRAY 0 ANGLE-INDEX ITERATION-INDEX)
             FOR Y2 = (AREF LINE-SEGMENT-ARRAY 1 ANGLE-INDEX ITERATION-INDEX)
             FOR (RX1 RY1 RX2 RY2 RX3 RY3 RX4 RY4) = (CALCULATE-LINE-COORDINATES
                                                       X1 Y1 X2 Y2)
             DO (PROGN
                  (SETQ RX1 (ROUND RX1)
                        RY1 (ROUND RY1)
                        RX2 (ROUND RX2)
                        RY2 (ROUND RY2)
                        RX3 (ROUND RX3)
                        RY3 (ROUND RY3)
                        RX4 (ROUND RX4)
                        RY4 (ROUND RY4))
                  (SETF (AREF RECTANGLE-POINTS-ARRAY 0 ANGLE-INDEX (1- ITERATION-INDEX)) RX1)
                  (SETF (AREF RECTANGLE-POINTS-ARRAY 1 ANGLE-INDEX (1- ITERATION-INDEX)) RY1)
                  (SETF (AREF RECTANGLE-POINTS-ARRAY 2 ANGLE-INDEX (1- ITERATION-INDEX)) RX2)
                  (SETF (AREF RECTANGLE-POINTS-ARRAY 3 ANGLE-INDEX (1- ITERATION-INDEX)) RY2)
                  (SETF (AREF RECTANGLE-POINTS-ARRAY 4 ANGLE-INDEX (1- ITERATION-INDEX)) RX3)
                  (SETF (AREF RECTANGLE-POINTS-ARRAY 5 ANGLE-INDEX (1- ITERATION-INDEX)) RY3)
                  (SETF (AREF RECTANGLE-POINTS-ARRAY 6 ANGLE-INDEX (1- ITERATION-INDEX)) RX4)
                  (SETF (AREF RECTANGLE-POINTS-ARRAY 7 ANGLE-INDEX (1- ITERATION-INDEX)) RY4)))))
    RECTANGLE-POINTS-ARRAY))

;;; The idea of this and the following variable is to precalculate the
;;; pattern to be drawn.  This calculation isn't particularly efficient,
;;; but who cares - it is only done once.  To handle other patterns, just
;;; define another array which calculates the points, and be sure to push
;;; it onto the EXPLOSION-ARRAY-LIST; the rest is handled automatically.
;;; The easy way to calculate the points is to calculate endpoints for
;;; line segments and use FINAL-EXPLOSION-SEGMENTS to debug it.  When you
;;; have finished debugging it call the CONVERT-LINES-TO-RECTANGLES
;;; function.  This function takes an array of point pairs which
;;; represent line segments and translates them into a quads of points
;;; which represent the equivalent rectangles.  The reason for this is
;;; that the :draw-line method used to draw a line segment takes a
;;; noticable amount of time to compute the rectangle.  The
;;; CONVERT-LINES-TO-RECTANGLES function optimizes this loop invarient by
;;; computing the rectangle ahead of time.  We can then call microcode to
;;; do the drawing, making it very fast.
;;;
;;; An assumption made for patterns is that all patterns are connected
;;; points and that line segments are drawn, and erased, for each line
;;; segment.  This works for most patterns, but will not work well for
;;; others: for example, a rocket with a primary explosion and a cluster of
;;; secondary explosions would work better as several independent explosions
;;; for each secondary explosion.
(DEFPARAMETER
  SIMPLE-EXPLOSION
  (LET* ((ANGLE-INCREMENT 30)
         (PIXEL-INCREMENT 10)
         (MAX-SIZE 200)
         (MAX-ITERATIONS (TRUNCATE MAX-SIZE PIXEL-INCREMENT))
         (ARRAY (MAKE-ARRAY `(2 ,(/ 360 ANGLE-INCREMENT)
                                ,MAX-ITERATIONS)
                            :ELEMENT-TYPE 'INTEGER))
         COORDINATE-ARRAY)
    ;; Everything starts out at the center.
    (LOOP FOR ANGLE FROM 0 BELOW 360 BY ANGLE-INCREMENT
          FOR ANGLE-INDEX FROM 0 BY 1
          DO (PROGN
               (SETF (AREF ARRAY 0 ANGLE-INDEX 0) 0)
               (SETF (AREF ARRAY 1 ANGLE-INDEX 0) 0)))
    
    (LOOP FOR ANGLE FROM 0 BELOW 360 BY ANGLE-INCREMENT
          FOR DELTA-X = (* PIXEL-INCREMENT (COSD ANGLE))
          FOR DELTA-Y = (* PIXEL-INCREMENT (SIND ANGLE))
          FOR ANGLE-INDEX FROM 0 BY 1
          DO (LOOP FOR ITERATION FROM 1 BELOW MAX-ITERATIONS
                   DO (PROGN
                        (SETF (AREF ARRAY 0 ANGLE-INDEX ITERATION) (+ DELTA-X
                                                                      (AREF ARRAY 0 
                                                                            ANGLE-INDEX
                                                                            (1- ITERATION))))
                        (SETF (AREF ARRAY 1 ANGLE-INDEX ITERATION) (+ DELTA-Y
                                                                      (AREF ARRAY 1
                                                                            ANGLE-INDEX
                                                                            (1- ITERATION)))))))
    (SETQ COORDINATE-ARRAY (CONVERT-LINES-TO-RECTANGLES ARRAY))
    (PUSH COORDINATE-ARRAY EXPLOSION-ARRAY-LIST)
    COORDINATE-ARRAY)
  "3-D Array of points.  The outer dimension controls the iteration (moving outward with
increasing indices), the middle dimension controls the angle (starting at 0 and moving out
just below 360 degrees) and the inner dimension is for X (index 0) and Y (index 1).")


(ZWEI:DEFINE-INDENTATION CALCULATE-SINGLE-EMBER (1 1))
(ZWEI:DEFINE-INDENTATION COMPUTE-EMBERS (1 1))
(DEFUN INITIALIZE-GRAVITY-ARRAY (GRAVITY)
  (LET* ((ANGLE-INCREMENT 20)
         (MAX-ITERATIONS 50)
         ANGLE-INDEX
         (ARRAY (MAKE-ARRAY `(2 ,(/ 360 ANGLE-INCREMENT)
                                ,MAX-ITERATIONS)
                            :ELEMENT-TYPE 'INTEGER
                            :INITIAL-ELEMENT MOST-NEGATIVE-FIXNUM))
         
         COORDINATE-ARRAY)
    (LABELS ((CALCULATE-SINGLE-EMBER (INITIAL-X INITIAL-Y X-VELOCITY INITIAL-Y-VELOCITY)
               (LET* (PREVIOUS-X
                      PREVIOUS-Y
                      (Y-VELOCITY INITIAL-Y-VELOCITY)
                      (X INITIAL-X)
                      (Y INITIAL-Y)
                      (ITERATION-INDEX 0))
                 (SETF (AREF ARRAY 0 ANGLE-INDEX ITERATION-INDEX) X)
                 (SETF (AREF ARRAY 1 ANGLE-INDEX ITERATION-INDEX) Y)
                 (LOOP
                   (INCF ITERATION-INDEX)
                   (WHEN (>= ITERATION-INDEX MAX-ITERATIONS)
                     (RETURN NIL))
                   ;; Calculate the next point.
                   (SETQ PREVIOUS-X X
                         PREVIOUS-Y Y)
                   (SETQ X (+ PREVIOUS-X X-VELOCITY)
                         Y (+ PREVIOUS-Y Y-VELOCITY)
                         Y-VELOCITY (+ Y-VELOCITY GRAVITY))
                   (SETF (AREF ARRAY 0 ANGLE-INDEX ITERATION-INDEX) X)
                   (SETF (AREF ARRAY 1 ANGLE-INDEX ITERATION-INDEX) Y))))
             (COMPUTE-EMBERS ()
               (SETQ ANGLE-INDEX 0)
               (LOOP WITH INITIAL-X = 0
                     WITH INITIAL-Y = 0
                     WITH PIXEL-INCREMENT = 10
                     FOR ANGLE FROM 0 BY ANGLE-INCREMENT BELOW 360
                     FOR DELTA-X = (* PIXEL-INCREMENT (COSD ANGLE))
                     FOR DELTA-Y = (* PIXEL-INCREMENT (SIND ANGLE))
                     DO (PROGN
                          (CALCULATE-SINGLE-EMBER INITIAL-X INITIAL-Y DELTA-X DELTA-Y)
                          (INCF ANGLE-INDEX)))))
      (COMPUTE-EMBERS))
    (SETQ COORDINATE-ARRAY (CONVERT-LINES-TO-RECTANGLES ARRAY))
    (PUSH COORDINATE-ARRAY EXPLOSION-ARRAY-LIST)
    COORDINATE-ARRAY))

;;; Try two different kinds of gravity for variety.
(DEFPARAMETER LIGHT-GRAVITY-EXPLOSION (INITIALIZE-GRAVITY-ARRAY .5))
(DEFPARAMETER HEAVY-GRAVITY-EXPLOSION (INITIALIZE-GRAVITY-ARRAY 1))


;;; Debug function to show what the explosion array looks like.
(DEFUN DISPLAY-EXPLOSION-ARRAY (&OPTIONAL (ARRAY SIMPLE-EXPLOSION))
  "Display data for an explosion array."
  (DECLARE (TYPE ARRAY ARRAY))
  (LOOP FOR ITERATION FROM 0 BELOW (ARRAY-DIMENSION ARRAY 2)
        DO (LOOP FOR ANGLE-INDEX FROM 0 BELOW (ARRAY-DIMENSION ARRAY 1)
                 INITIALLY (FORMAT T "~2%Iteration number ~D " ITERATION)
                 DO (FORMAT T "~%~2D (~3D,~3D) (~3D,~3D) (~3D,~3D) (~3D,~3D)  "
                            ANGLE-INDEX
                            (TRUNCATE (AREF ARRAY 0 ANGLE-INDEX ITERATION))
                            (TRUNCATE (AREF ARRAY 1 ANGLE-INDEX ITERATION))
                            (TRUNCATE (AREF ARRAY 2 ANGLE-INDEX ITERATION))
                            (TRUNCATE (AREF ARRAY 3 ANGLE-INDEX ITERATION))
                            (TRUNCATE (AREF ARRAY 4 ANGLE-INDEX ITERATION))
                            (TRUNCATE (AREF ARRAY 5 ANGLE-INDEX ITERATION))
                            (TRUNCATE (AREF ARRAY 6 ANGLE-INDEX ITERATION))
                            (TRUNCATE (AREF ARRAY 7 ANGLE-INDEX ITERATION))))))


;;; Old version of FINAL-EXPLOSION which uses :draw-line.
(DEFUN FINAL-EXPLOSION-SEGMENTS (WINDOW X Y EXPLOSION-ARRAY SPEED SIZE)
  "Generate the final explosion of the rocket on WINDOW starting at (X,Y) of category
TYPE with swiftness SPEED and of SIZE."
  ;; Currently, only one type of explosion exists.
  (DECLARE (TYPE NUMBER X Y SPEED SIZE))
  (LOOP WITH PIXEL-INCREMENT = 10
        WITH INITIAL-X = X
        WITH INITIAL-Y = Y
        FOR ITERATION FROM 0 BELOW (1- (ARRAY-DIMENSION EXPLOSION-ARRAY 2))
        FOR DISTANCE FROM 0 BELOW SIZE BY PIXEL-INCREMENT
        FOR DREW-A-POINT = NIL
        DO (PROGN
             ;; Draw the segments representing ember traces of the explosion, wait
             ;; a little while, and then erase them.  The monitor's phosphors will
             ;; help here by making the traces persist a little.
             (LOOP FOR ANGLE-INDEX FROM 0 BELOW (ARRAY-DIMENSION EXPLOSION-ARRAY 1)
                   FOR X =  (+ INITIAL-X (AREF EXPLOSION-ARRAY 0 ANGLE-INDEX ITERATION))
                   FOR Y =  (+ INITIAL-Y (AREF EXPLOSION-ARRAY 1 ANGLE-INDEX ITERATION))
                   FOR NEXT-X = (+ INITIAL-X (AREF EXPLOSION-ARRAY 0 ANGLE-INDEX (1+ ITERATION)))
                   FOR NEXT-Y = (+ INITIAL-Y (AREF EXPLOSION-ARRAY 1 ANGLE-INDEX (1+ ITERATION)))
                   WHEN (AND (PLUSP X) (PLUSP Y) (PLUSP NEXT-X) (PLUSP NEXT-Y))
                   DO (PROGN
                        (SETQ DREW-A-POINT T)
                        (SEND WINDOW :DRAW-LINE X Y NEXT-X NEXT-Y 2 W:BLACK W:ALU-XOR)))
             (WHEN DREW-A-POINT
               (PROCESS-SLEEP SPEED "Exploding")
               (LOOP FOR ANGLE-INDEX FROM 0 BELOW (ARRAY-DIMENSION EXPLOSION-ARRAY 1)
                     FOR X =  (+ INITIAL-X (AREF EXPLOSION-ARRAY 0 ANGLE-INDEX ITERATION))
                     FOR Y =  (+ INITIAL-Y (AREF EXPLOSION-ARRAY 1 ANGLE-INDEX ITERATION))
                     FOR NEXT-X = (+ INITIAL-X (AREF EXPLOSION-ARRAY 0 ANGLE-INDEX (1+ ITERATION)))
                     FOR NEXT-Y = (+ INITIAL-Y (AREF EXPLOSION-ARRAY 1 ANGLE-INDEX (1+ ITERATION)))
                     WHEN (AND (PLUSP X) (PLUSP Y) (PLUSP NEXT-X) (PLUSP NEXT-Y))
                     DO (SEND WINDOW :DRAW-LINE X Y NEXT-X NEXT-Y 2 W:BLACK W:ALU-XOR))))))

(DEFUN FINAL-EXPLOSION (WINDOW X Y EXPLOSION-ARRAY SPEED SIZE)
  "Generate the final explosion of the rocket on WINDOW starting at (X,Y) using EXPLOSION-ARRAY
with swiftness SPEED and of SIZE."
  ;; Currently, only one type of explosion exists.
  (DECLARE (TYPE NUMBER X Y SPEED SIZE))
  (LOOP
    WITH PIXEL-INCREMENT = 10
    WITH INITIAL-X = (TRUNCATE X)
    WITH INITIAL-Y = (TRUNCATE Y)
    WITH GRAY-DIVISOR = (MAX 1 (CEILING (/ SIZE (1- (LENGTH GRAYS-ARRAY))) 2))
    WITH HALFWAY = (CEILING SIZE 2)
    FOR ITERATION FROM 0 BELOW (ARRAY-DIMENSION EXPLOSION-ARRAY 2)
    FOR DISTANCE FROM 0 BELOW SIZE BY PIXEL-INCREMENT
    ;; The idea here is that the ember will die out with time.  We simulate this decay by
    ;; having it burn brightly for the first half of its trip, and then evenly decay
    ;; through the grays-array for the second half of its trip.
    FOR GRAY = (IF (< DISTANCE HALFWAY)
                   (AREF GRAYS-ARRAY 0)
                   ;;ELSE
                   (AREF GRAYS-ARRAY (TRUNCATE (- DISTANCE HALFWAY) GRAY-DIVISOR)))
    FOR DREW-A-POINT = NIL
    DO (PROGN
         ;; Draw the segments representing ember traces of the explosion, wait
         ;; a little while, and then erase them.  The monitor's phosphors will
         ;; help here by making the traces persist a little.
         (LOOP
           FOR ANGLE-INDEX FROM 0 BELOW (ARRAY-DIMENSION EXPLOSION-ARRAY 1)
           FOR X1 = (+ INITIAL-X (AREF EXPLOSION-ARRAY 0 ANGLE-INDEX ITERATION))
           FOR Y1 = (+ INITIAL-Y (AREF EXPLOSION-ARRAY 1 ANGLE-INDEX ITERATION))
           FOR X2 = (+ INITIAL-X (AREF EXPLOSION-ARRAY 2 ANGLE-INDEX ITERATION))
           FOR Y2 = (+ INITIAL-Y (AREF EXPLOSION-ARRAY 3 ANGLE-INDEX ITERATION))
           FOR X3 = (+ INITIAL-X (AREF EXPLOSION-ARRAY 4 ANGLE-INDEX ITERATION))
           FOR Y3 = (+ INITIAL-Y (AREF EXPLOSION-ARRAY 5 ANGLE-INDEX ITERATION))
           FOR X4 = (+ INITIAL-X (AREF EXPLOSION-ARRAY 6 ANGLE-INDEX ITERATION))
           FOR Y4 = (+ INITIAL-Y (AREF EXPLOSION-ARRAY 7 ANGLE-INDEX ITERATION))
           WHEN (AND (PLUSP X1) (PLUSP Y1))
           DO (PROGN
                (W:PREPARE-SHEET (WINDOW)
                  (SYS:%DRAW-SHADED-TRIANGLE X1 Y1 X2 Y2 X3 Y3 W:ALU-XOR T T NIL GRAY WINDOW)
                  (SYS:%DRAW-SHADED-TRIANGLE X1 Y1 X4 Y4 X3 Y3 W:ALU-XOR T T T GRAY WINDOW))
                (SETQ DREW-A-POINT T)))
         (WHEN DREW-A-POINT
           (PROCESS-SLEEP SPEED "Exploding")
           (LOOP
           FOR ANGLE-INDEX FROM 0 BELOW (ARRAY-DIMENSION EXPLOSION-ARRAY 1)
           FOR X1 = (+ INITIAL-X (AREF EXPLOSION-ARRAY 0 ANGLE-INDEX ITERATION))
           FOR Y1 = (+ INITIAL-Y (AREF EXPLOSION-ARRAY 1 ANGLE-INDEX ITERATION))
           FOR X2 = (+ INITIAL-X (AREF EXPLOSION-ARRAY 2 ANGLE-INDEX ITERATION))
           FOR Y2 = (+ INITIAL-Y (AREF EXPLOSION-ARRAY 3 ANGLE-INDEX ITERATION))
           FOR X3 = (+ INITIAL-X (AREF EXPLOSION-ARRAY 4 ANGLE-INDEX ITERATION))
           FOR Y3 = (+ INITIAL-Y (AREF EXPLOSION-ARRAY 5 ANGLE-INDEX ITERATION))
           FOR X4 = (+ INITIAL-X (AREF EXPLOSION-ARRAY 6 ANGLE-INDEX ITERATION))
           FOR Y4 = (+ INITIAL-Y (AREF EXPLOSION-ARRAY 7 ANGLE-INDEX ITERATION))
           WHEN (AND (PLUSP X1) (PLUSP Y1))
           DO (PROGN
                (W:PREPARE-SHEET (WINDOW)
                  (SYS:%DRAW-SHADED-TRIANGLE X1 Y1 X2 Y2 X3 Y3 W:ALU-XOR T T NIL GRAY WINDOW)
                  (SYS:%DRAW-SHADED-TRIANGLE X1 Y1 X4 Y4 X3 Y3 W:ALU-XOR T T T GRAY WINDOW))
                (SETQ DREW-A-POINT T)))))))

(DEFUN LAUNCH-AND-EXPLODE (WINDOW X Y ANGLE MAX-DISTANCE SIZE)
  "Launch and explode a single rocket."
  (DECLARE (TYPE W:MINIMUM-WINDOW WINDOW)
           (TYPE NUMBER X Y ANGLE MAX-DISTANCE SIZE))
  (MULTIPLE-VALUE-BIND (MAX-ROCKET-X MAX-ROCKET-Y)
      (LAUNCH-ROCKET WINDOW X Y ANGLE 1 MAX-DISTANCE)
    ;; Start the explosion where the rocket stopped.
    (FINAL-EXPLOSION WINDOW MAX-ROCKET-X MAX-ROCKET-Y
                     (NTH (RANDOM (LENGTH EXPLOSION-ARRAY-LIST))
                          EXPLOSION-ARRAY-LIST) (* 2 (1+ (RANDOM 4))) SIZE)))

(DEFUN FIREWORK-DISPLAY (WINDOW X Y)
  "Display rockets serially, one after another, launching on WINDOW starting at (X,Y)."
  (DECLARE (TYPE W:MINIMUM-WINDOW WINDOW)
           (TYPE NUMBER X Y))
  (LOOP WITH WINDOW-WIDTH = (SEND WINDOW :WIDTH)
        ;; Divide the window up into three parts: left, center and right.
        ;; The idea here is that if we launch a rocket near the center of
        ;; the window then we should be able to launch it to the left or
        ;; to the right.  The left part can have an angle from 10 to 90
        ;; degrees, the center part from 10 to 170 degrees, and the right
        ;; part from 90 to 170 degrees.
        WITH WINDOW-PART = (COND ((< X (* 1 (TRUNCATE WINDOW-WIDTH 3))) :LEFT)
                                 ((< X (* 2 (TRUNCATE WINDOW-WIDTH 3))) :CENTER)
                                 (T :RIGHT))
        WITH MIN-ANGLE = (COND ((EQ WINDOW-PART :LEFT  ) 10)
                               ((EQ WINDOW-PART :CENTER) 10)
                               ((EQ WINDOW-PART :RIGHT ) 90))
        WITH MAX-ANGLE = (COND ((EQ WINDOW-PART :LEFT  ) 90)
                               ((EQ WINDOW-PART :CENTER) 170)
                               ((EQ WINDOW-PART :RIGHT ) 170))
        ;; Make sure that the rocket travels at least 50 pixels.  If
        ;; travels a shorter distance than that it might hurt the person
        ;; launching it :-)
        WITH MIN-LAUNCH-DISTANCE = 50
        FOR ANGLE = (+ MIN-ANGLE (RANDOM (- MAX-ANGLE MIN-ANGLE)))
        ;; Size of the explosion in pixels.
        FOR SIZE = (RANDOM 200)
        ;; This is a feeble attempt to calculate the distance the rocket will
        ;; travel.  It works most of the time, but doesn't work all of the time.
        ;; If I had more time to work on this I would figure out why, and make
        ;; this calculation correct.
        FOR MAX-LAUNCH-DISTANCE = (- (COND ((< ANGLE 90)
                                         (MIN (/ (- WINDOW-WIDTH X)
                                                 (COSD ANGLE))
                                              (/ Y (COSD (- 90 ANGLE)))))
                                        ((= ANGLE 90)
                                         Y)
                                        (T
                                         (MIN (/ X (- (COSD ANGLE)))
                                              (/ Y (COSD (- 90 ANGLE))))))
                                     SIZE)
        DO (LAUNCH-AND-EXPLODE WINDOW X Y ANGLE (+ MIN-LAUNCH-DISTANCE
                                                   (RANDOM (- MAX-LAUNCH-DISTANCE
                                                              MIN-LAUNCH-DISTANCE)))
                               SIZE)))

(DEFUN FIREWORKS-DISPLAY (WINDOW &OPTIONAL (NUMBER-LAUNCHERS *DEFAULT-LAUNCHER-COUNT*))
  "Top-level function to display the fireworks."
  (DECLARE (TYPE W:MINIMUM-WINDOW WINDOW)
           (TYPE INTEGER NUMBER-LAUNCHERS))
  ;; Get rid of the clouds in the sky!
  (SEND WINDOW :SET-CURSORPOS 0 0)
  (SEND WINDOW :CLEAR-EOF)
  (LET* ((WINDOW-WIDTH (SEND WINDOW :WIDTH))
         ;; Determine where to put the launchers based upon random places between
         ;; the min/max (X,Y)'s.
         (MIN-X 50)
         (MAX-X (- WINDOW-WIDTH MIN-X))
         (X-TRAVERSAL (- MAX-X MIN-X))
         (MIN-Y 500)
         (MAX-Y 700)
         (Y-TRAVERSAL (- MAX-Y MIN-Y)))
  (IF (> NUMBER-LAUNCHERS 1)
      ;; We have more than one launcher, so we need to create a process for each one.
      (LET ((PROCESSES (LOOP FOR PROCESS-INDEX FROM 0 BELOW NUMBER-LAUNCHERS
                             COLLECTING (PROCESS-RUN-FUNCTION "Firework display"
                                              'FIREWORK-DISPLAY WINDOW
                                              (+ MIN-X (RANDOM X-TRAVERSAL))
                                              (+ MIN-Y (RANDOM Y-TRAVERSAL))))))
        (UNWIND-PROTECT
            ;; Wait until we stop.
            (SLEEP MOST-POSITIVE-FIXNUM)
          ;; Kill off each launcher process before getting out.
          (LOOP FOR PROCESS IN PROCESSES
                DO (SEND PROCESS :KILL))))
      ;;ELSE - Only one display
      (FIREWORK-DISPLAY WINDOW
                        (+ MIN-X (RANDOM X-TRAVERSAL))
                        (+ MIN-Y (RANDOM Y-TRAVERSAL))))))

;;; If the user has the MIT screen saver stuff loaded, put this on that list,
;;; otherwise, always call this as a screen saver.
(IF (BOUNDP 'TV:*SCREEN-SAVER-HACKS-LIST*)
  (PUSHNEW 'FIREWORKS-DISPLAY TV:*SCREEN-SAVER-HACKS-LIST*)
  ;;ELSE
  (SETQ TV:*SCREEN-SAVER-HACKS* 'FIREWORKS-DISPLAY))
