;;; -*- Mode:Common-Lisp; Package:ZWEI; Fonts:CPTFONT,CPTFONTB; Base:10 -*-

;;;
;;; Copyright (c) 1987 Douglas Johnson.  All rights reserved. 
;;; 
; Copyright (C) 1980, Massachusetts Institute of Technology
; Copyright (C) 1984, Texas Instruments Incorporated. All rights reserved.


;***********************************************************************************
;
;        This file contains the ZWEI interface into the spelling 
;        facilities.
;
;***********************************************************************************



;  This table defines the syntax of a word for spelling correction.
;  A word is a-z, A-Z, and  single qoute, delimited by anything else."


(DEFPARAMETER *SPELLING-WORD-SYNTAX-TABLE*
   (MAKE-SYNTAX-TABLE
    '((32 WORD-DELIMITER) WORD-DELIMITER;040 space
      WORD-DELIMITER;041 !
      WORD-DELIMITER;042 "
      WORD-DELIMITER;043 #
      WORD-DELIMITER;044 $
      WORD-DELIMITER;045 %
      WORD-DELIMITER;046 &
      WORD-ALPHABETIC;047 '
      WORD-DELIMITER;050 (
      WORD-DELIMITER;051 )
      WORD-DELIMITER;052 *
      WORD-DELIMITER;053 +
      WORD-DELIMITER;054 ,
      WORD-DELIMITER;055 -
      WORD-DELIMITER;056 .
      WORD-DELIMITER;057 /
      (10 WORD-DELIMITER);Digits
      WORD-DELIMITER;072 :
      WORD-DELIMITER;073 ;
      WORD-DELIMITER;074 <
      WORD-DELIMITER;075 =
      WORD-DELIMITER;076 >
      WORD-DELIMITER;077 ?
      WORD-DELIMITER;100 @
      (26 WORD-ALPHABETIC);Upper case letters
      WORD-DELIMITER;133 [
      WORD-DELIMITER;134 \
      WORD-DELIMITER;135 ]
      WORD-DELIMITER;136 ^
      WORD-DELIMITER;137 _
      WORD-DELIMITER;140 `
      (26 WORD-ALPHABETIC);Lower case letters
      (133 WORD-DELIMITER)))
   "This defines the syntax of a word for purposes of the spelling correction.") 


(DEFSUBST CURRENT-WORD (BP)
  "Return the BP that begins the current word."
  (FORWARD-WORD (FORWARD-WORD (FORWARD-CHAR BP -1 T) 1 T) -1 T)) 


(DEFMACRO WITH-SPELLING-WORD-SYNTAX (&BODY BODY)
  "Execute BODY in spelling word mode.
The spelling word syntax table is temporarily installed as the word syntax table."
  `(LET ((*MODE-WORD-SYNTAX-TABLE* *SPELLING-WORD-SYNTAX-TABLE*))
     ,@BODY)) 



(DEFCOM COM-CORRECT-SPELLING
	"Correct the spelling for the entire buffer or just the region if one exists."
	()
  (SPELLER:EDITOR-INIT-SPELLER)
  (LET ((CORRECTION-FUNCTION (IF (MEMBER (SEND *INTERVAL* :MAJOR-MODE) '(COMMON-LISP-MODE ZETALISP-MODE))
				 'CORRECT-SPELLING-LISP
				 'CORRECT-SPELLING-INTERVAL))
	(REDISPLAY DIS-ALL))
  (WITH-SPELLING-WORD-SYNTAX
    (PRESERVE-POINT 
      (COND ((WINDOW-MARK-P *WINDOW*)
	     (TYPEIN-LINE "Correcting region spelling...")
	     (SETF (WINDOW-MARK-P *WINDOW*) nil)	;1Get rid of region display*
	       (REDISPLAY *WINDOW* :RELATIVE 0)
	       (CATCH-ALL
		 (UNLESS (FUNCALL CORRECTION-FUNCTION (POINT) (MARK))
		   (SETF REDISPLAY DIS-NONE))))

	    (T   (TYPEIN-LINE "Correcting Spelling...")
		 (CATCH-ALL
		 (UNLESS (FUNCALL CORRECTION-FUNCTION *interval*)
		   (SETF REDISPLAY DIS-NONE)))))))
  (TYPEIN-LINE-MORE " Done.")
  REDISPLAY))



(DEFCOM COM-CORRECT-WORD-SPELLING
	"Correct the spelling of the word at point." ()
  (SPELLER:EDITOR-INIT-SPELLER)	
  (LET ((DISPLAY-CODE DIS-NONE))
    (WITH-SPELLING-WORD-SYNTAX
	    (LET* ((BP1 (CURRENT-WORD (POINT)))
		      (BP2 (FORWARD-WORD BP1 1 T))
		      (WORD (STRING-INTERVAL BP1 BP2)))
		 (TYPEIN-LINE (STRING-APPEND "Correcting the spelling of " #\" WORD #\" "...")) 
		 (CATCH 'SPELLER:ABORT
		 (IF (SPELLER:EDITOR-CHECK-WORD-P WORD)
		     (TYPEIN-LINE-MORE "2 *OK")
		     (LET ((NEW-WORD (SPELLER:EDITOR-CORRECT-WORD WORD)))
		       (UNLESS (STRING-EQUAL WORD NEW-WORD)	
			 (CASE-REPLACE BP1 BP2  NEW-WORD)
			 (SETQ DISPLAY-CODE DIS-TEXT)))))))
  DISPLAY-CODE))



(DEFUN CORRECT-SPELLING-INTERVAL (START-BP &OPTIONAL END-BP)
  "Corrects the spelling of an group of words specified by 
   START-BP and END-BP as the start and end of the interval
   or by START-BP as an interval.2  *Returns t if any words were
   changed, nil otherwise."
1  *(LET ((CHANGED NIL))
  (GET-INTERVAL START-BP END-BP NIL)
  (SETQ START-BP (COPY-BP START-BP)
	END-BP (FORWARD-WORD (FORWARD-WORD END-BP -1 T) 1 T))
  (WITH-UNDO-SAVE  ("Spelling Correction" START-BP END-BP T)
    (LET ((*REGION-MARKING-MODE* :REVERSE-VIDEO))	1;reverse video misspelled words*
      (DO* ((BP1 (CURRENT-WORD START-BP))	
	    (BP2 (FORWARD-WORD BP1 1 T))
	    (WORD))
	   (NIL)				
	(SETQ WORD (STRING-INTERVAL BP1 BP2))
	(UNLESS (SPELLER:EDITOR-CHECK-WORD-P WORD)
	  (SETF CHANGED T)
	  (MOVE-BP (POINT) BP1)
	  (MOVE-BP (MARK) BP2)
	  (SETF (WINDOW-MARK-P *WINDOW*) T)
	  (REDISPLAY *WINDOW* :ABSOLUTE)   ;display word in center of screen
	  (LET ((NEW-WORD (SPELLER:EDITOR-CORRECT-WORD WORD)))
	    (UNLESS (STRING-EQUAL WORD NEW-WORD)
	      (CASE-REPLACE BP1 BP2  NEW-WORD))))
	(SETQ BP2 (FORWARD-WORD BP2 1))
	(WHEN (OR (NULL BP2) (BP-< END-BP BP2)) (RETURN))
	(SETQ BP1 (FORWARD-WORD BP2 -1)))
      (SETF (WINDOW-MARK-P *WINDOW*) nil)  ;1Get rid of region display*
      ))
  CHANGED))

(DEFUN CORRECT-SPELLING-LISP (START-BP &OPTIONAL END-BP)
  "Corrects the spe2l*ling of LISP strings and comments in the 
   interval specified by START-BP and END-BP as the start and
   end of the interval or by START-BP as an interval.  
   This function assumes LISP code in the buffer.2 Returns t if anything 
   was changed.  NIL otherwise.*"
  (LET ((CHANGED NIL))
  (GET-INTERVAL START-BP END-BP NIL)
  (SETQ START-BP (COPY-BP START-BP)
	END-BP (FORWARD-WORD (FORWARD-WORD END-BP -1 T) 1 T))
  (WITH-UNDO-SAVE  ("Spelling Correction" START-BP END-BP T)
    (LET ((*REGION-MARKING-MODE* :REVERSE-VIDEO))	;reverse video misspelled words
		     (DO* ((BP1 (CURRENT-WORD START-BP))	
			   (BP2 (FORWARD-WORD BP1 1 T))
			   (WORD))
			  (NIL)
		       (MULTIPLE-VALUE-BIND (STRING NIL COMMENT)
			   (LISP-BP-SYNTACTIC-CONTEXT BP1)
			 (WHEN (OR STRING COMMENT) 
			   (SETQ WORD (STRING-INTERVAL BP1 BP2))
			   (UNLESS (SPELLER:EDITOR-CHECK-WORD-P WORD)
			     (SETF CHANGED T)
			     (MOVE-BP (POINT) BP1)
			     (MOVE-BP (MARK) BP2)
			     (SETF (WINDOW-MARK-P *WINDOW*) T)
			     (REDISPLAY *WINDOW* :ABSOLUTE)	;display word in center of screen
			     (LET ((NEW-WORD (SPELLER:EDITOR-CORRECT-WORD WORD)))
			       (UNLESS (STRING-EQUAL WORD NEW-WORD)
				 (CASE-REPLACE BP1 BP2  NEW-WORD)))))
			 (SETQ BP2 (FORWARD-WORD BP2 1))
			 (WHEN (OR (NULL BP2) (BP-< END-BP BP2)) (RETURN))
			 (SETQ BP1 (FORWARD-WORD BP2 -1)))
		       (SETF (WINDOW-MARK-P *WINDOW*) nil))))	;Get rid of region display
    CHANGED)) 





;
;  Add commands to correct spelling to command table.
;
#-elroy #!z (SET-COMTAB *STANDARD-COMTAB* 
	    '(#/META-$       COM-CORRECT-WORD-SPELLING
	      #/M-C-$  COM-CORRECT-SPELLING)
		   (MAKE-COMMAND-ALIST '(COM-CORRECT-SPELLING
					  COM-CORRECT-WORD-SPELLING)))

#+elroy (SET-COMTAB *STANDARD-COMTAB* 
	    '(#\META-$       COM-CORRECT-WORD-SPELLING
	      #\M-C-$  COM-CORRECT-SPELLING)
		   (MAKE-COMMAND-ALIST '(COM-CORRECT-SPELLING
					  COM-CORRECT-WORD-SPELLING)))




