;;; -*- Mode:Common-Lisp; Package:user; Base:10; Fonts:(CPTFONT HL12B HL12I MEDFNB) -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;          PURPOSE:  To find the :HOST or :USER information for STRING in a given NAMESPACE.
;                       
;  FUNCTION NAMES:  WHERE             PARAMETERS:  STRING
;                                                         SEARCH-RULE    (:USER)
;                                                         NAMESPACE      ("TI-CSC")
;                                                         REFRESH-CACHE  (NIL)
;
;                      WHERE-IS-HOST   PARAMETERS:  STRING
;                                                         NAMESPACE
;                                                         REFRESH-CACHE
;
;                      WHERE-IS-USER    PARAMETERS:  STRING
;                                                         NAMESPACE
;                                                         REFRESH-CACHE
;
;            AUTHOR:  DIANA M. SPARACIN  09/16/87
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(DEFUN WHERE-IS-HOST (STRING NAMESPACE REFRESH-CACHE)
  "Find hosts in the namespace with the string STRING in their Explorer host name or their host location."
  
; Find an exact match (if there is one) and print its HOST-NAME, LOCATION, and ADDRESSES.
  
  (LET ((ADDRESSES)
	(LIST-OF-HOST-NAMES)
	(LOCATION)
	(OBJECT (NAME:LIST-OBJECT STRING :HOST :NAMESPACE NAMESPACE)))
    (COND
      ((NOT (NULL OBJECT))
       (FORMAT T "~%~%~@26,1,1,A:  ~A"  "HOST-NAME"  (CAR OBJECT))

       ; Get ATTR-LIST for OBJECT and search for LOCATION AND ADDRESSES
       ; since they can be anywhere in the ATTR-LIST.

       (LOOP WITH ATTR-LIST = (CADDR OBJECT) AND ATTR AND VALUE
	     WHILE ATTR-LIST DO
             (SETF ATTR  (CAR ATTR-LIST))
	     (SETF VALUE (CADR ATTR-LIST))
	     (COND
	       ((EQUALP ':LOCATION ATTR)                    
		(SETF LOCATION VALUE))
	       ((EQUALP '(:ADDRESSES :GROUP) ATTR)	   
		(SETF ADDRESSES VALUE))
	       ); END COND
	     (SETF ATTR-LIST (CDDR ATTR-LIST))
	     ); END LOOP
       (FORMAT T "~%~@26,1,1,A:  ~A"    "LOCATION"   LOCATION)
       (FORMAT T "~%~@26,1,1,A:  ~A"    "ADDRESSES"  ADDRESSES)
       (FORMAT T "~%~%Continuing search for host ~A..." STRING)
       ); END NOT
      (T (FORMAT T "~%~%~A ~A ~A" "Host name" STRING "not found.  Continuing search..."))
      ); END COND

    ; FIND OTHER OCCURRENCES OF STRING

    ; Set up the internal symbol "namespace-HOST" and bind it to the variable LIST-OF-HOST-NAMES
    ; to save the list of hosts for a given NAMESPACE.  (This makes subsequent searches very fast.)
  
    (SETF LIST-OF-HOST-NAMES (INTERN (FORMAT NIL "*~A-HOST*" NAMESPACE)))

    ; Get the list of hosts from the namespace server if this is the first time a given
    ; namespace has been searched or REFRESH-CACHE has been set to T (default is NIL).
    ; Each namespace will have its own "namespace-HOST" symbol.

    (UNLESS (AND (BOUNDP LIST-OF-HOST-NAMES)
		 (CONSP (FORMAT NIL "*~A-HOST*" NAMESPACE))
		 (NULL REFRESH-CACHE))
      (SET LIST-OF-HOST-NAMES (NAME:LOOKUP-OBJECTS-FROM-PROPERTIES
				:NAMESPACE NAMESPACE
				:CLASS :HOST
				:READ-ONLY T
				:ATTRIBUTE-LIST '(:MACHINE-TYPE :EXPLORER))))

    ; Find all occurences of STRING, but do not print
    ; an exact match since it has already been printed.
  
    (LOOP FOR HST IN (EVAL LIST-OF-HOST-NAMES)
	  WHEN (AND (NOT (EQUALP STRING (CAR (NAME:LIST-OBJECT HST :HOST))))
		    (OR (SEARCH STRING (NAME:OBJECT-NAME HST) :TEST 'STRING-EQUAL)
			(SEARCH STRING (NAME:GET-ATTRIBUTE-VALUE HST :LOCATION)
				:TEST 'STRING-EQUAL))) DO

	  ; Print host's HOST-NAME, LOCATION, and ADDRESSES.

	  (FORMAT T "~%~%~@26,1,1,A:  ~A" "HOST-NAME"     
		  (NAME:OBJECT-NAME HST))

	  (FORMAT T "~%~@26,1,1,A:  ~A"   "LOCATION"
		  (NAME:GET-ATTRIBUTE-VALUE HST :LOCATION))

	  (FORMAT T "~%~@26,1,1,A:  ~A"   "ADDRESSES"
		  (NAME:GET-ATTRIBUTE-VALUE HST :ADDRESSES))
	  ); END LOOP
    );END LET
  T
  ); END WHERE-IS-HOST



(DEFUN WHERE-IS-USER (STRING NAMESPACE REFRESH-CACHE)
  "Find users in the namespace with the string STRING in their Explorer login-name or their personal name."
  
; Find an exact match (if there is one) and print its attribute list.
  
  (LET ((LIST-OF-LOGIN-NAMES)
	(OBJECT (NAME:LIST-OBJECT STRING :USER :NAMESPACE NAMESPACE)))
    (COND
      ((NOT (NULL OBJECT))
       (FORMAT T "~%~%~@26,1,1,A:  ~A" "LOGIN-NAME" (CAR OBJECT)) 
       (LOOP WITH ATTR-LIST = (CADDR OBJECT) AND ATTR AND VALUE
	     WHILE ATTR-LIST DO
	     (SETF ATTR  (CAR  ATTR-LIST))
	     (SETF VALUE (CADR ATTR-LIST))
	     (FORMAT T "~%~@26,1,1,A:  ~A" ATTR VALUE)   
	     (SETF ATTR-LIST (CDDR ATTR-LIST))
	     ); END LOOP
       (FORMAT T "~%~%Continuing search for user ~A..." STRING)
       ); END NOT
      (T (FORMAT T "~%~%~A ~A ~A" "Login name" STRING "not found.  Continuing search..."))
      ); END COND

    ; FIND OTHER OCCURRENCES OF STRING.

    ; Set up the internal symbol "namespace-USER" and bind it to the variable LIST-OF-LOGIN-NAMES
    ; to save the list of users for a given NAMESPACE.  (This makes subsequent searches very fast.)
  
    (SETF LIST-OF-LOGIN-NAMES (INTERN (FORMAT NIL "*~A-USER*" NAMESPACE)))

    ; Get the list of users from the namespace server if this is the first time a given
    ; namespace has been searched or REFRESH-CACHE has been set to T (default is NIL).
    ; Each namespace will have its own "namespace-USER" symbol.

    (UNLESS (AND (BOUNDP LIST-OF-LOGIN-NAMES)
		 (NOT (NULL (FORMAT NIL "*~A-USER*" NAMESPACE)))
		 (NULL REFRESH-CACHE))
      (SET LIST-OF-LOGIN-NAMES (NAME:LOOKUP-OBJECTS-FROM-PROPERTIES
				 :NAMESPACE NAMESPACE
				 :CLASS :USER
				 :READ-ONLY T)))

    ; Find all occurences of STRING, but do not print
    ; an exact match since it has already been printed.  
  
    (LOOP FOR USR IN (EVAL LIST-OF-LOGIN-NAMES)
	  WHEN (AND (NOT (EQUALP STRING (CAR (NAME:LIST-OBJECT USR :USER))))
		    (OR (SEARCH STRING (NAME:OBJECT-NAME USR) :TEST 'STRING-EQUAL)
			(SEARCH STRING (NAME:GET-ATTRIBUTE-VALUE USR :PERSONAL-NAME)
				:TEST 'STRING-EQUAL))) DO

	  ; Print the login-name and its attribute list.

	  (FORMAT T "~%~%~@26,1,1,A~A  ~A" "LOGIN-NAME" ":" (NAME:OBJECT-NAME USR))
	  (LOOP WITH ATTR-LIST = (CADDR (NAME:LIST-OBJECT USR :USER)) AND ATTR AND VALUE
		WHILE ATTR-LIST
		(SETF ATTR  (CAR  ATTR-LIST))  
		(SETF VALUE (CADR ATTR-LIST))
		(FORMAT T "~%~@26,1,1,A:  ~A" ATTR VALUE)  
		(SETF ATTR-LIST (CDDR ATTR-LIST))
		); END LOOP
	  (FORMAT T "~%")
	  ); END LOOP
    ); END LET
  T
  ); END WHERE-IS-USER



(DEFUN WHERE (STRING &KEY (SEARCH-RULE :HOST) (NAMESPACE "TI-CSC") (REFRESH-CACHE NIL))
  "Find  STRING in namespace NAMESPACE using SEARCH-RULE :HOST, :USER, or :BOTH (:HOST and :USER)" 
  (COND
    ((EQUALP :BOTH SEARCH-RULE) (CATCH-ERROR (WHERE-IS-USER STRING NAMESPACE REFRESH-CACHE))
			        (CATCH-ERROR (WHERE-IS-HOST STRING NAMESPACE REFRESH-CACHE)))
    ((EQUALP :HOST SEARCH-RULE) (CATCH-ERROR (WHERE-IS-HOST STRING NAMESPACE REFRESH-CACHE)))
    ((EQUALP :USER SEARCH-RULE) (CATCH-ERROR (WHERE-IS-USER STRING NAMESPACE REFRESH-CACHE)))
    (T (FORMAT T "~%INVALID SEARCH-RULE OPTION - USE :USER, :HOST, OR :BOTH (:USER AND :HOST)"))
    ); END COND
  T
  ); END WHERE
