;;; -*- Mode:Common-Lisp; Fonts:(HL12 HL12B HL12BI); Base:10 -*-

;;; THIS IS A FILE TRANSFER PROGRAM FOR TRANSFERRING FILES BETWEEN THE TI EXPLORER AND
;;;      ANY OTHER MACHINE WITH XMODEM SOFTWARE, VIA THE EXPLORER'S SERIAL PORT.

;1 *************************************************************************************************
;1;; This software is not supported by Texas Instruments and no warranty is intended or implied.*
;1 *************************************************************************************************

;;; IT USES THE XMODEM PROTOCOL (Due to W. Christensen; my apologies to him if I have
;;; mis-spelled his name), simple checksum (not crc) version.

;;;  **** written by J. Morgan, Dept. of Linguistics, U. of Illinois as Urbana-Champaign, March 1986****
;;;            and may be freely disseminated as long as it is not sold for profit.

;;; MODIFICATIONS:
;;;   4/08/87 Converted to Explorer release 3.0
;;;   7/10/86 
;;;    This code defaults to work at 19200 baud, 8 data bits, one stop bit, and no parity.
;;;       To work with a dial-up modem, the baud rate may have to be changed to a lower speed such
;;;       as 300, 1200, or 2400, and this file recompiled.
;;;       See functions xmodem-in and xmodem-out for baud parity stop and data parameters being used.
;;;     Added close of outstream in xmodem-out so that additional file transfers wouldn't get error.
;;;     Added close of instream in xmodem-in so that additional file transfers wouldn't get an error.
;;;     Added passing of com, win, and file arguments.
;;;     Added using win argument in format statements.
;;;     Modified to not ask for file if one is passed.     
;;;     Added stripping off of eof's on rt type files.
;;;     Added code to close streams in event of abnormal termination.
;;;     Added code to terminate xmodem if return entered for file name.          
;;;     Added code to time out if character not able to be read on send or receive.
;;;     Added code to check for valid file name on send and receive.
;;;     

;;; INSTRUCTIONS FOR USING THIS PROGRAM:

;;;      A. SYNTAX:  (xmodem arg &optional file win baud parity stop data)
;;;                   where arg is one of {st, sb, rt, rb}; arg NOT quoted

;;;                   st = send an explorer file to ibm, translating from explorer character set to ascii
;;;                   sb = send an explorer file to ibm, without any character translation

;;;                   rt = receive a file from the ibm and store it as an explorer file, translating from
;;;                                  ascii to the explorer character set
;;;                   rb = receive a file from the ibm and store it as an explorer file, without translation
;;;                   optional   default
;;;                   arguments
;;;
;;;                   file       nil(will ask for file if nil)
;;;                   win        t  
;;;                   baud      19200
;;;                   parity     none
;;;                   stop       1
;;;                   data       8
;;;                   time       60 seconds
;;;
;;;
;;;      B. USING XMODEM:

;;;         1. SENDING A FILE TO IBM:

;;;            a. First type (xmodem st) or (xmodem sb).  You will be prompted for a file name;
;;;            enter the name of the explorer file you wish to send.
;;;            Give the full path name of the file unless it is in the current default directory

;;;            b. Prepare the ibm by calling up some modem program (e.g. procomm, qmodem, etc.)
;;;            that uses the xmodem protocol (also known as modem7 or Christensen protocol).
;;;            Do whatever is necessary to prepare it to receive a file.  When it is ready it will
;;;            send a NAK to the explorer and transmission will begin.  You need do nothing more.

;;;         2. RECEIVING A FILE FROM IBM:

;;;            a. First type (xmodem rt) or (xmodem rb).  You will be prompted for a file name;
;;;            enter the explorer name you wish to assign to the file received from the ibm.
;;;            Give the full path name of the file unless you wish it to be created in the current
;;;            default directory.

;;;            b. The explorer will now say "Press any key when the ibm is ready to send."
;;;            DO NOT do this (i.e. press a key) until step c is completed.

;;;            c. Prepare the ibm by calling up some modem program (e.g. procomm, qmodem, etc.)
;;;            that uses the xmodem protocol. Do whatever is necessary to prepare it to send a
;;;            file.  When the ibm is ready, press any key on the explorer.  The explorer will then
;;;            send a NAK to the ibm as a signal to start the transfer.  You need do nothing more.


(DEFVAR NEEDLF NIL)				; USED IN ascii version to record whether last char was a return

(DEFVAR soh 1)         ; first character in xmodem header
(DEFVAR eot 4)          ; end-of-transmission signal
(DEFVAR ack 6)          ; acknowledge receipt signal
(DEFVAR nak 21)         ; signals checksum error; also used for ready-to-receive
(DEFVAR eof 26)


(DEFUN xmodem  (mode &optional (file nil) (win t) (xbaud 19200) (xparity :none) (xstop 1) (xdata 8)(xtime 60))
"                  mode
                               st = send an explorer file to pc, translating from explorer character set to ascii
                               sb = send an explorer file to pc, without any character translation

                               rt = receive a file from the pc and store it as an explorer file, translating from
                                    ascii to the explorer character set
                               rb = receive a file from the pc and store it as an explorer file, without translation

                   argument possible values

                   file        user supplied file name(nil will ask for file)   
                   win        user supplied window(t for *termianl-io*)            
                   xbaud     from 50 to 19200 
                   xparity    :odd,:even,:none
                   xstop     1,1.5,2
                   xdata     5,6,7,8
                   xtime     an integer representing seconds till timeout"
; main function
  (setq xparity
	 (cond
	   ((equal xparity "odd") :odd)
	   ((equal xparity "even") :even)
	   ((equal xparity "none") :none)
	   (t xparity)))
  (Case mode
	  (st (xmodem-out t file win xbaud xparity xstop xdata xtime))	; send a file from explorer to pc, translated to ascii first
	  (sb (xmodem-out nil file win xbaud xparity xstop xdata xtime))	; send a file from explorer to pc, raw
	  (rt (xmodem-in t file win xbaud xparity xstop xdata xtime))	; receive a file from pc, translated from ascii to ti character set
	  (rb (xmodem-in nil file win xbaud xparity xstop xdata xtime))))	; receive a file from pc, raw (no translation)

    ; xmodem-out  manages the transmission process

(DEFUN xmodem-out (ascii-flag     ; if ascii-flag = t, translate to ascii
			  file win xbaud xparity xstop xdata xtime
			  &aux
			  (block-nr 0)
			  outstream
			  (out-file nil))
  "Send a file via XMODEM to PC. If ascii-flag = t, Explorer character set is translated to ascii."
  (if file (FORMAT win "~%Place remote PC in xmodem receive mode for file ~A~%" file))
  (setq out-file (if file file (get-file "" win)))		;first get explorer file name  7/26/86
  (if out-file
      (WITH-OPEN-FILE (instream out-file :if-does-not-exist nil)	;7/27/86
        (if instream
	    (with-open-stream (outstream (si:make-serial-stream ))	;initialize the serial port  7/26/86
	      (configure-serial-stream outstream xbaud xparity xstop xdata)	;7/10/86 10:29:49
;;; If wait times out, exit
	      (if (wait-for nak outstream xtime)			;wait for signal to start from pc
		  (progn
		    (FORMAT win "~%Received NAK~%")
		    (SEND outstream :clear-output)		;this may be useless, but who knows?
		    (LOOP while (xmod-send-block ascii-flag	; keep sending 128-char blocks until file is empty
							 instream
							 outstream
							 (setq block-nr (1+ block-nr)) win xtime))
		    (SEND outstream :tyo eot)		; send end-of-transmission signal
		    (wait-for ack outstream xtime)			; wait for acknowledgement
		    (FORMAT win "~%ACK received. Transmission completed.~%")
		    t						; return t for success  7/26/86
		    )
		  (progn
		    (FORMAT win "~%NAK Not received, Transmission not completed.~%")
		    nil						; return false for failure  7/26
		    )
		  )
	      )
	    (FORMAT win "~%File ~A not found." out-file)
	    nil; return false for failure  7/28
	    )
	)
      )
  
  )

(DEFUN get-file(file win &aux (not-abort nil))                                                ;get explorer file name as a string
  (FORMAT win "~%     Explorer filename: ")
  (setq not-abort (READ-LINE win))			;7/26/86
  (if (not (equal not-abort ""))
      (concatenate :string file not-abort)
      nil))

(DEFUN wait-for (char STREAM xtime &aux (beg-second 0)(beg-minute 0)(end-second 0)(end-minute 0) return-val)
  (multiple-value-setq (beg-second beg-minute) (get-decoded-time))
  (setq beg-second (+ beg-second (* beg-minute 60)))
; loop till we get the character we want or 20 seconds elapses
  (do ()(return-val (if (equal return-val 'found) t nil))
    (multiple-value-setq (end-second end-minute) (get-decoded-time))
    (setq end-second (+ end-second (* end-minute 60)))
    (if (< end-second beg-second)
        (setq end-second (+ end-second 60)))
    (setq return-val (IF (EQUAL (SEND stream :tyi-no-hang)
				     char)
			   'found
			   (if (> end-second (+ beg-second xtime))
			       t nil)))))

(DEFUN bit-complement (nr)                                    ; 8-bit compelement of its argument
  (LOGANDC1 nr 255))

(DEFUN xmod-send-block (ascii-flag                          ; send a block of 128 characters
			        instream
				outstream
				block-nr win xtime
				&aux
				(more t)			; value of 'more' is what the function returns.  
				block				; t = file not yet empty
				outbuffer			; nil = end-of-file was encountered, file is empty
				char
				(checksum 0)
				(eof-not-written t))
  (prog ();
	  (setq outbuffer (MAKE-STRING-OUTPUT-STREAM))
	  (setq block (WITH-OUTPUT-TO-STRING (outbuffer)	;block is a copy of the output, to use if
			  (DOTIMES (n 128)			;re-transmission is necessary
			    (IF needlf				; if last char was a return and ascii-flag is on,
			        (PROGN (SETQ char 10)	;then make the next char a LF
					   (SETQ needlf nil))
				(SETQ char (SEND instream :tyi)))	;otherwise get the next char from the file
			    (IF (AND char			; if there is a char and ascii-flag is on,
				       ascii-flag)		;then translate to ascii
			        (SETQ char (translate-exp-pc char instream)))	;NB if file is empty, char will be NIL
								;(IF eof-not-written
			    (WRITE-CHAR (IF char		; Finally, if there is a char, write it to the buffer
						 char
						 (PROGN
						   (SETQ more nil)	;else write pc end-of-file instead		   
						   (SETQ eof-not-written NIL)
						   (SETQ char 26)))	;7/15/86 14:39:58
					      outbuffer)
								;);END IF
			    (SETQ checksum (make-8-bits (+ char checksum)))	; update the checksum, which is just
			    (IF (AND ascii-flag		;the 8-bit sum of the block
				       (= char 13))
			        (SETQ needlf t))		; if this char was a return and ascii-flag
			    )))					; is on, next char should be LF
	  (FORMAT win "~%Sending block #~D" block-nr)
       loop       (SEND outstream :tyo soh)		;xmodem protocol header = SOH, BLOCK NUMBER,
	  (SEND outstream :tyo (make-8-bits block-nr))	; 8-BIT COMPLEMENT OF BLOCK NUMBER  
	  (SEND outstream :tyo (bit-complement block-nr))
	  (WITH-INPUT-FROM-STRING (instring block)	; now send the data: 128 characters
	    (DOTIMES (n 128)
	      (SEND outstream :tyo (TYI  instring))))
	  (SEND outstream :tyo (make-8-bits checksum))	; now send the checksum and wait for
	  (IF (= (SEND outstream :tyi)			; acknowledgement
		 ack)
	      (RETURN more)
	      (FORMAT win "Checksum error on block ~D~%" block-nr)	;if header or checksum error, 
	      (GO loop))))					; send it again til you get it right
          
;;; these two functions are self-evident
          
(DEFUN make-8-bits (n)
  (LOGAND n 255))
          
(DEFUN configure-serial-stream (STREAM xbaud xparity xstop xdata)
  (SEND stream :put :baud xbaud)
  (SEND stream :put :number-of-data-bits xdata)
  (SEND stream :put :number-of-stop-bits xstop)
  (SEND stream :put :parity xparity))

;;; the following translates from the explorer character set to ascii
;;;       it doesn't give full coverage; for example, the symbol-shift characters are ignored.
;;;       here is all it does:

         ;;; explorer line-end is just CR (no LF), moreover with the high bit set (i.e. 8Dh, not 0Dh)
         ;;;  so the high bit is cleared and (by another function) a LF is added
         ;;;  likewise, TAB is 89h, not 09h, so it has to be translated too.

         ;;; the second thing this function does is look for 06h, which appears to be the explorer
         ;;; marker of a font change.  It is always followed by a second non-character byte which
         ;;;  is probably a font number.  This function just skips past such pairs of characters.

(DEFUN translate-exp-pc (CHAR stream)
  (LOGAND 127
	      (IF (EQL char 6)
		  (PROGN (SEND stream :tyi)
			     (translate-exp-pc (SEND stream :tyi)
						  stream))
		  char)))

(DEFUN wait-for-char (STREAM xtime &aux (beg-second 0)(beg-minute 0)(end-second 0)(end-minute 0) return-val)
  (multiple-value-setq (beg-second beg-minute) (get-decoded-time))
  (setq beg-second (+ beg-second (* beg-minute 60)))
; loop till we get the character we want or 20 seconds elapses
  (do ()(return-val return-val )
    (multiple-value-setq (end-second end-minute) (get-decoded-time))
    (setq end-second (+ end-second (* end-minute 60)))
    (if (< end-second beg-second)
        (setq end-second (+ end-second 60)))
    (setq return-val  (SEND stream :tyi-no-hang))
    (setq return-val
	   (if return-val
	       return-val
	       (if (> end-second (+ beg-second xtime))
		   'error
		   nil)
	       )
	   )
    )
  )

;;; xmodem-in manages the receiving process very analogously to xmodem-out

(DEFUN xmodem-in (ascii-flag                                           ; if the ascii-flag is set,
		        file win xbaud xparity xstop xdata xtime
			&aux					;  ascii is translated to explorer 
			(block-nr 0)				; character set
			instream
			(file-in t)
			(last-char '()) fst (lst t))
"Receive a file via XMODEM from a PC. If ascii-flag = t, ascii is translated to Explorer character set."      
  
  (if (do () ((or (not file-in) (not lst)) (if file-in t nil)); wait for valid pathname or return
        (setq file-in (get-file file win))
;If probe-file returns T or NIL, ok, go on and receive it (T = exists, NIL = doesn't), and ignore-errors returns
;     NIL for fst and NIL for lst
;If probe returns error, ignore-errors will capture it and return NIL for fst and T for lst
;     repeat file name question.
	(multiple-value-setq (fst lst) (ignore-errors (probe-file file-in)))	;9/17/86 09:21:02
	(if lst
	    (FORMAT win "~%     File name incorrect as entered, check and re-enter.")));9/18/86 08:52:00
;if valid pathname
      (WITH-OPEN-FILE (outstream file-in :direction :output)	; first get the explorer file name
        (with-open-stream (instream (si:make-serial-stream ))	; 7/26/86
	  (configure-serial-stream instream xbaud xparity xstop xdata)	; 7/10/86 10:29:21 ; then wait until the pc is set up on the 
	  (FORMAT win "Press any key when the pc is ready to send")	; other end and looking for a nak
	  (READ-CHAR win)
	  (FORMAT win "~%Sending  NAK~%")		; then send the nak to signal 
	  (SEND instream :tyo nak)				; ready-to-receive
	  (SEND instream :clear-input)			; does this do anything useful?
	  (do () ((or (equal last-char eot)(equal last-char 'error))
		   (if (equal last-char eot)
		       (progn
			 (FORMAT win "~%EOT received. Transmission completed.~%")
			 t)
		       (if (equal last-char 'error)
			   (progn
			     (FORMAT win "~%Data not received. Transmission not completed.~%")
			     nil))))
	    (setq last-char (xmod-receive-block ascii-flag	; then read in data 128 characters at a
						       instream	; time until the pc sends eot to signal end
						       outstream	; of transmission      
						       (SETQ block-nr (1+ block-nr)) win xtime)))))))

;;; this function does the work of reading in the data a character at a time, checking for errors, and storing
;;; in the destination file.  If EOT was received, nil is returned, signalling done; else t is returned

(DEFUN xmod-receive-block (ascii-flag          ;if ascii-flag =t do the translation from ascii to explorer
				  instream
				  outstream
				  block-nr win xtime
				  &aux
				  head				; these variables are for the three header bytes that precede
				  in-block-nr			; each received block 
				  in-block-compl
				  checksum-in
				  inbuffer
				  block
				  char
				  (checksum 0)
				  (eof-not-read t))		; 7/15/86 11:43:10
  (block (); 4/7/87
	  (setq inbuffer (MAKE-STRING-OUTPUT-STREAM))	;inbuffer is where the data is 
	  (if (equal 'error (setq head (wait-for-char instream xtime)))	; stored 1 char at a time
	      (return 'error)
	      (if (equal eot (prog (); 8/22/86 16:18:56	      
				  loop  (SETQ checksum 0)
				     (IF (=  head eot)		; if the first character is eot,  the transmission is done
					 (PROGN (SEND instream :tyo ack)	; so acknowledge the EOT and quit
						    (RETURN eot)))
				     (SETQ in-block-nr (SEND instream :tyi))
				     (SETQ in-block-compl (SEND instream :tyi))
				     
; inbuffer is really a stream pointer, not a string, so we need 'block' as the string to be written to the file.      
				     
				     (SETQ block                                                       
					      (WITH-OUTPUT-TO-STRING (inbuffer)
					        (DOTIMES (n 128)	; get 128 chars, 
						  (SETQ char (SEND instream :tyi))                      
						  
; keep a checksum by simply summing the characters (before they're translated, of course)
						  
						  (SETQ checksum (make-8-bits (+ char checksum)))  
						  
; if eof encountered, store its location in buffer
						  
						  (IF (and ascii-flag (= eof char))	; 7/15/86 09:10:03
						      (SETQ eof-not-read nil))
						  
; if the ascii-flag is on translate them, and store them in the input buffer
						  
						  (IF ascii-flag                                                
						      (SETQ char (translate-pc-exp char)))               
						  
;  if it's a LF and ascii-flag is on, throw it away
						  
						  (IF (NOT (AND ascii-flag
								    (= 10 char)))                  
						      (if eof-not-read	; 7/15/86 11:25:20
							  (WRITE-CHAR char inbuffer)
							  )
						      )
						  )
						)
					      )
; then read in what the pc thinks the checksum should be
				     
				     (SETQ checksum-in (SEND instream :tyi))                      
				     
; check the header and checksum for errors
				     
				     (IF (NOT (AND  (= head soh)                                  
						        (= (mod block-nr 256) in-block-nr)	; 7/11/86 16:37:32
							(= in-block-compl (bit-complement block-nr))))
					 (PROGN (FORMAT win "~%Header error on block ~D" block-nr) 
						    
; if there's an error, send nak as a request for re-transmission, and loop.
						    
						    (SEND instream :tyo nak)                                     
						    (GO loop)
						    )
					 )
				     (IF (NOT (= checksum checksum-in))
					 (PROGN (FORMAT win "~%Checksum error on block ~D" block-nr)
						    (SEND instream :tyo nak)
						    (GO loop)
						    )
					 )
				     
; otherwise, if header and checksum are ok, then ;chances are the data is ok, so send ack as a signal
; of successful receipt of the block,
				     
				     (SEND instream :tyo ack)                                          
				     (FORMAT win "~%Block ~D received" block-nr)                       
				     
;  write the block to the file and 
				     
				     (princ block outstream)
				     
; return t to signal there is  more to come  
				     
				     (RETURN t)
				     )
			 )
		  
		  (return eot)
		  )
	      (return nil)
	      )
	  )
  )
  

;;; this is a minimal translation function.  It changes asci CR to 8Dh, and TAB to 89h.  
;;; ascii LF is stripped off elsewhere in the program

(DEFUN translate-pc-exp (CHAR)
  (CASE char
	   (13 141)
	   (09 137)
	   (otherwise char)))


