;;; -*- Mode:Common-Lisp; Package:CLUE; Base:10; Fonts:(CPTFONT HL12B HL12BI) -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (b)(3)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1988, Texas Instruments Incorporated. All rights reserved.


1;;;
;;; Change history:
;;;
;;;  Date        Author    Description
;;; ----------------------------------------------------------------------
;;;  2/17/88              Original
;;;  3/28/88     CAT     Updated for CLUE 0.11--make-instance changed to make-contact*

;;;--------------------------------------------
;;; This is the explanation of how ICONs are managed:
;;;
;;; * The basic icon glyphs are maintained as LISPM font characters
;;;   since that is the best pixel editor available (now).  Look
;;;   for the glyphs in SYS:XWM;ICON-FONT.XLD.  This font file
;;;   need not be loaded in the runtime environment for XWM.
;;; * The function GENERATE-ICON-IMAGES translates the font glyphs
;;;   into IMAGES, as defined by CLX, and dumps them to a file; look in
;;;   SYS:XWM;ICON-IMAGES.XLD.  This image file should be loaded in the
;;;   environment that is to run XWM.
;;; * Once a channel has been opened for XWM to run in, the function
;;;   INITIALIZE-ICON-PIXMAPS must be run to generate the PIXMAPS.


;;;--------------------------------------------
;;; The following definitions support the translation of font glyphs
;;; into images.  It works in the following way: 
;;;  - The assoc list *XWM-SYMBOL-CHARACTER-IMAGE* itemizes all the
;;;    ICON images to be made.  
;;;  - GENERATE-ICON-IMAGES displays them on the screen;
;;;    if width and height are not specified you are prompted, via
;;;    W:MOUSE-SPECIFY-RECTANGLE to delimit the ICON pixels.
;;;    if width and height are specified then area of the ICON 
;;;    is computed from that.

;;; These variables control where the origin of the interactive display
;;; of the font will be.
(DEFVAR interactive-display-x 300)
(DEFVAR interactive-display-y 300)

;;; This is an association list of all the ICONs to be made.
;;; For each element in this list the structure is:
;;;  - Symbol name for the image; "IMAGE" must be the last word in the symbol name
;;;  - The LISPM font that contains the ICON glyph
;;;  - A string which presumably contains the glyph
;;;  - width of the image
;;;  - height of the image
(DEFVAR *xwm-symbol-character-image*
	'((move-image fonts:xwm-icons "M" 32 32)
	  (zoom-out-image fonts:xwm-icons "Z" 32 32)
	  (zoom-in-image fonts:xwm-icons "z" 32 32)
	  (resize-image fonts:xwm-icons "R" 32 32)
	  (iconify-image fonts:xwm-icons "I" 32 32)
	  (TI-LOGO-image fonts:TI-LOGO "T" 130 130)))

;;; This returns a bit array that contains the LISPM bit image
(defun mouse-array (&optional max-width max-height left-top-right-bottom)
 (MULTIPLE-VALUE-BIND (left top right bottom)
     (IF  left-top-right-bottom
	  (VALUES-LIST left-top-right-bottom)
	  (w:mouse-specify-rectangle))
   (when (and max-width (< max-width (- right left)))
     (setf right (+ left max-width)))
   (when (and max-height (< max-height (- bottom top)))
     (setf bottom (+ top max-height)))
  (let* ((array (tv:sheet-screen-array tv:mouse-sheet))
	 (width (- right left))
	 (height (- bottom top))
	 (x left)
	 (y top)
	 (bits-per-pixel (ticl:ARRAY-ELEMENT-SIZE array))
	 (real-width (TRUNCATE (DPB 0 5 (+ (* WIDTH bits-per-pixel) 31.)) bits-per-pixel))
	 (new-array (MAKE-ARRAY `(,HEIGHT ,real-width)
			:element-type (tv:sheet-array-type-cl tv:default-screen))))
    (ticl:BITBLT tv:alu-seta real-width height array x y new-array 0 0)
    new-array)))

;;; This function mearly puts the image up on the screen
(defun display-icon (&key (font fonts:ti-logo) (string "T")
		     ;; It looks like all text strings need to be moved left a bit
		     ;; the number 10 is an eyeball approximation.
		     &aux (left-kerning 10)) 
  (funcall *terminal-io* :clear-screen) 
  (funcall *terminal-io* :string-out-explicit string
	    (- interactive-display-x left-kerning)
	    interactive-display-y 
	    nil nil font w:alu-seta))

;;; This function accepts an association list in the form of *XWM-SYMBOL-CHARACTER-IMAGE*
;;; and sets the symbol names to the generated XLIB IMAGES.  If an output file is specified
;;; the appropriate forms will be dumped to file to recreate the image symbols ina new environment.
;;; This function should only need to be run whenever the set of ICONs change.
(DEFUN generate-icon-images (&optional output-file (symbol-character-images *xwm-symbol-character-image*))
  (LOOP for (symbol font string width height) in symbol-character-images
	do (SET symbol (create-image :width width :height height :depth 1 :format :image-xy :data
				     (PROGN (display-icon :font font :string string)
					    (mouse-array width height
							 (LIST interactive-display-x
							       interactive-display-y
							       (+ interactive-display-x width)
							       (+ interactive-display-y height)))))))
  (WHEN output-file
    (ticl:dump-forms-to-file output-file
			     (LOOP for (symbol . rest) in symbol-character-images
				   collect `(DEFPARAMETER ,symbol ,(SYMBOL-VALUE symbol))))))

;;;--------------------------------------------
;;; This function is used to initialize PIXMAPs with IMAGES once the XWM channel has been opened.
(DEFUN initialize-icon-pixmaps (&optional (drawable-screen *xwm-root-window* )
				          (symbol-character-images *xwm-symbol-character-image*))
  "Generate ICON PIXMAPs described in SYMBOL-CHARACTER-IMAGES for DRAWABLE-SCREEN ."
  (LOOP for (symbol . rest) in symbol-character-images
	do (SET (INTERN (ticl:STRING-APPEND (STRING-RIGHT-TRIM "IMAGE" (SYMBOL-NAME symbol)) "PIXMAP")
			'clue)
		(LET* ((image (symbol-value symbol))
		       (width (image-width image))
		       (height (image-height image))
		       (pixmap (create-pixmap :drawable drawable-screen 
					      :width width
					      :height height
					      :depth 1))
		       (gcontext (create-gcontext :drawable drawable-screen)))
		  (put-image pixmap gcontext image
			     :x 0 :y 0
			     :width width
			     :height height)
		  (free-gcontext gcontext)
		  pixmap))))

;;; --------------------------------------------
;;; Once the PIXMAPs have been initialized this routine can be used to display them on
;;; an X-SERVER just to see if they look OK.
(DEFUN display-icon-pixmaps (&optional (contact-display (open-contact-display 'XWM-icon-display :host *default-host*)
							supplied-p)
			               (symbol-character-images  *xwm-symbol-character-image*))
  "Look at PIXMAPs associated with SYMBOL-CHARACTER-IMAGES on CONTACT-DISPLAY"  
  (LET ((contacts
	  (LOOP for (symbol font string width height) in symbol-character-images
		for next-pixmap = (SYMBOL-VALUE (INTERN (ticl:STRING-APPEND (STRING-RIGHT-TRIM "IMAGE"
											       (SYMBOL-NAME symbol))
									    "PIXMAP")))
		for x = 10
		for y from 10 to 900 by gadget-width
		collect (MAKE-contact 'contact
				       :parent contact-display
				       :width width
				       :height height
				       :border 1  :y y :x x
				       :contact-background next-pixmap))))
    (MAPC 'present contacts)
    (display-finish-output contact-display)
    (Y-OR-N-P "OK?")
    (MAPC 'unmap contacts)
    (MAP 'destroy contacts)
    (display-finish-output contact-display)
    (unless supplied-p
      (close-display  contact-display))))
