;;; -*- Mode:common-Lisp; Package:System-Internals; Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; COPYRIGHT (C) 1984-1989 Texas Instruments INCORPORATED.  ALL RIGHTS RESERVED.

;;; Revised:
;;; With standard parallel printer cable, the 855 won't work, but Imagen will.  So... make
;;;   :RESET-HARDWARE properly toggle the init line, instead of holding it down, (which tells the
;;;   855 to init forever.)  Also wait for it to come back ONLINE afterwards. 10-14-85 MBC
;;;
;;; Stream support for the explorer parallel port.
;;; This code uses the following system variables:
;;;   *Parallel-Port*          - The device descriptor for the parallel port
;;;   *Parallel-Output-Buffer* - the output buffer where the microcode gets characters to go out



(DEFUN MAKE-PARALLEL-STREAM (&REST KEYWORDS)
  "Make a stream that talks over the explorer parallel port.
Keywords are:
:FLAVOR             -- flavor name to make this stream an instance of
:OUTPUT-BUFFER-SIZE -- 256. words default (1 page)
:FORCE-OUTPUT       -- NIL (the default) means you must :FORCE-OUTPUT to cause transmission,
                       T means an automatic :FORCE-OUTPUT occurs after :TYO

"
  (DECLARE (ARGLIST &KEY &OPTIONAL FLAVOR OUTPUT-BUFFER-SIZE FORCE-OUTPUT))
  (INSTANTIATE-FLAVOR (OR (GETF KEYWORDS :FLAVOR);user specified flavor or ...
			 'PARALLEL-STREAM) (LOCF KEYWORDS) T)) 



(DEFFLAVOR PARALLEL-STREAM-MIXIN
	   ((OUTPUT-BUFFER-SIZE 256)
	    (FORCE-OUTPUT ()))
	   (LINE-OUTPUT-STREAM-MIXIN)
  (:REQUIRED-FLAVORS OUTPUT-STREAM CHARACTER-STREAM)
  (:INITABLE-INSTANCE-VARIABLES OUTPUT-BUFFER-SIZE FORCE-OUTPUT)
  (:INIT-KEYWORDS :FLAVOR :OUTPUT-BUFFER-SIZE :FORCE-OUTPUT)
  (:DOCUMENTATION :MIXIN "The guts of parallel I/O Streams")) 


(DEFFLAVOR PARALLEL-STREAM
	   ()
	   (PARALLEL-STREAM-MIXIN
	    OUTPUT-STREAM
	    CHARACTER-STREAM)
  (:DOCUMENTATION :COMBINATION "Parallel I/O Stream")) 


;;; Housekeeping methods


(DEFMETHOD (PARALLEL-STREAM-MIXIN :INIT) (INIT-PLIST)
 ;; first make sure the *Parallel-Port* structure has been set up
 ;; this should have been part of the cold initialization list
 ;; NOTE: this must only be done once!
  (declare (special *parallel-port*))
  (IF (EQ *PARALLEL-PORT* ())
    (INITIALIZE-PARALLEL-INTERRUPT))
  ;; set up parameters from init list
  (LOOP FOR (KWD VAL) ON (CDR INIT-PLIST) BY 'CDDR UNLESS (MEMBER KWD '(:FLAVOR) :TEST #'EQ) DO
     (ERRSET (SEND SELF :PUT KWD VAL) ()))
  (SEND SELF :RESET)) 


(DEFMETHOD (PARALLEL-STREAM-MIXIN :RESET) (&OPTIONAL (ABORT-OUTPUT ':ABORT))
  (declare (special *parallel-port* *PARALLEL-OUTPUT-BUFFER*))
  (SEND SELF :CLOSE ABORT-OUTPUT);Forget about any remaining output characters
  (SEND SELF :RESET-HARDWARE)
  (SETQ *PARALLEL-OUTPUT-BUFFER* (GET-SERIAL-BUFFER OUTPUT-BUFFER-SIZE))
  (ARRAY-DPB-OFFSET (1+ (%POINTER *PARALLEL-OUTPUT-BUFFER*)) %%Q-POINTER *PARALLEL-PORT*
     %PARALLEL-TRANSMIT-BUFFER)
  (ENABLE-PARALLEL-EVENT)) 


(DEFVAR *PARALLEL-HAS-BEEN-RESET* ()) 


(DEFMETHOD (PARALLEL-STREAM-MIXIN :RESET-HARDWARE) (&OPTIONAL (OVERRIDE NIL))
  (declare (special  *PARALLEL-HAS-BEEN-RESET*))
  (WHEN (OR OVERRIDE (NOT *PARALLEL-HAS-BEEN-RESET*))
    (PARALLEL-COMMAND #o12);hold init line down
    (PARALLEL-COMMAND #o16);then bring it back up, "off"
    (SETF *PARALLEL-HAS-BEEN-RESET* T)
    (PROCESS-WAIT-WITH-TIMEOUT "Parallel Reset" 600                              ;wait a reasonable time for ONLINE
			       #'(LAMBDA ();Comes ONLINE or if not,
				   (= 1 (LDB (BYTE 1 2) (PARALLEL-STATUS))))))) 	;Let FORCE-OUTPUT handle error report.
  
  

(DEFMETHOD (PARALLEL-STREAM-MIXIN :CLOSE) (&OPTIONAL ABORT-P)
  (declare (special *parallel-port* *PARALLEL-OUTPUT-BUFFER*))
  (IF *PARALLEL-OUTPUT-BUFFER*
    (PROGN
      (IF (NOT ABORT-P)
	(SEND SELF :FINISH))
      (RETURN-SERIAL-BUFFER *PARALLEL-OUTPUT-BUFFER*)
      (SETQ *PARALLEL-OUTPUT-BUFFER* ())))
  (DISABLE-PARALLEL-EVENT)
  (ARRAY-DPB-OFFSET 0 %%Q-POINTER *PARALLEL-PORT* %PARALLEL-TRANSMIT-BUFFER)) 



(DEFMETHOD (PARALLEL-STREAM-MIXIN :FINISH) ()
 ;; get things started if they aren't already
  (declare (special *PARALLEL-OUTPUT-BUFFER*))
  (SEND SELF :FORCE-OUTPUT)
  (PROCESS-WAIT "Parallel Finish" #'(LAMBDA (BUFFER)
				      (NOT (SERIAL-BUFFER-NOT-EMPTY-P BUFFER)))
		*PARALLEL-OUTPUT-BUFFER*)) 



(DEFMETHOD (PARALLEL-STREAM-MIXIN :STATUS) ()
  "Print out the status of the parallel i/o interface."
  (LET ((STAT (PARALLEL-STATUS)))
    (FORMAT T "~%Parallel status ~16R : " (LDB (BYTE 8 0) STAT))
    (IF (= 1 (LDB (BYTE 1 0) STAT))
      (FORMAT T " BUSY"))
    (IF (= 1 (LDB (BYTE 1 1) STAT))
      (FORMAT T " PAPER OUT"))
    (IF (= 1 (LDB (BYTE 1 2) STAT))
      (FORMAT T " ONLINE"))
    (IF (= 0 (LDB (BYTE 1 3) STAT))
      (FORMAT T " FAULT")))) 


;;; Get the current parallel port parameters

(DEFMETHOD (PARALLEL-STREAM-MIXIN :GET) (PROP)
  (LET ((STAT (PARALLEL-STATUS)))
    (CASE PROP
      (:STATUS STAT)
      (:BUSY (IF (= 1 (LDB (BYTE 1 0) STAT))
	       T
	       ()))
      (:PAPER-OUT (IF (= 1 (LDB (BYTE 1 1) STAT))
		    T
		    ()))
      (:ONLINE (IF (= 1 (LDB (BYTE 1 2) STAT))
		 T
		 ()))
      (:FAULT (IF (= 0 (LDB (BYTE 1 3) STAT))
		T
		()))
      (:OUTPUT-BUFFER-SIZE OUTPUT-BUFFER-SIZE)
      (:FORCE-OUTPUT FORCE-OUTPUT)
      (:OTHERWISE (FERROR () "~S not a valid property name" PROP))))) 

;;; Modify parallel port parameters

(DEFMETHOD (PARALLEL-STREAM-MIXIN :PUT) (PROP VAL)
  (LET ()
    (CASE PROP
      (:OUTPUT-BUFFER-SIZE (SETQ OUTPUT-BUFFER-SIZE VAL))
      (:FORCE-OUTPUT (SETQ FORCE-OUTPUT (IF VAL
					  T
					  ())))
      (:OTHERWISE (FERROR () "~S not a valid property name" PROP))))) 


;;; Parallel output methods

(DEFMETHOD (PARALLEL-STREAM-MIXIN :TYO) (CH)
  (declare (special *PARALLEL-OUTPUT-BUFFER*))
  (IF (NOT (SERIAL-BUFFER-NOT-FULL-P *PARALLEL-OUTPUT-BUFFER*))
    (PROGN
      (SEND SELF :FORCE-OUTPUT)
      (PROCESS-WAIT "Parallel out" #'SERIAL-BUFFER-NOT-FULL-P *PARALLEL-OUTPUT-BUFFER*)))
  (WRITE-SERIAL-BUFFER *PARALLEL-OUTPUT-BUFFER* CH)
  (IF FORCE-OUTPUT
    (SEND SELF :FORCE-OUTPUT))) 

;;; The microcoded %io routine for the parallel port will return either NIL if the
;;; operation completed successfully, or the contents of the parallel status register
;;; if the operation did not complete (port busy or port errors).
;;; The 855 printer unfortunately turns on the error status bits while the busy
;;; bit is on so we have to do the status decoding a bit differently than we
;;; would normally.


(DEFMETHOD (PARALLEL-STREAM-MIXIN :FORCE-OUTPUT) ()
  (declare (special *parallel-port* *PARALLEL-OUTPUT-BUFFER*))
  (IF (SERIAL-BUFFER-NOT-EMPTY-P *PARALLEL-OUTPUT-BUFFER*)
    (PROGN
      (PROCESS-WAIT "Parallel Out"
		    #'(LAMBDA ()
			(LET ((STATUS (%IO 0 *PARALLEL-PORT*)))
			  (OR (NOT STATUS);ready or
			     (AND (NOT (= 1 (LDB (BYTE 1 0) STATUS)));not busy &
				(NOT (= 6 (LDB (BYTE 3 1) STATUS))))))));errors
      ;;now check for errors -- must read the register again since we can't pass status
      ;;out of the process-wait -- 855 might be busy here !
      (LET ((STATUS (PARALLEL-STATUS)))
	(IF (AND (NOT (= 1 (LDB (BYTE 1 0) STATUS))) (NOT (= 6 (LDB (BYTE 3 1) STATUS))))
	  (COND
	    ((= 1 (LDB (BYTE 1 1) STATUS)); ERROR: Paper-out.
	     (FERROR () "Printer is out of paper"))
	    ((= 0 (LDB (BYTE 1 2) STATUS)); ERROR: Off-line.
	     (FERROR () "Printer is off-line."))
	    ((= 0 (LDB (BYTE 1 3) STATUS)); ERROR: Fault.
	     (FERROR () "Printer fault.")))))))) 


;;; The following method is used to test parallel stream operation without microcode support
;;; It should normally be commented out.
;(defmethod (PARALLEL-STREAM-MIXIN :FORCE-OUTPUT) ()
;  (if (serial-buffer-not-empty-p *Parallel-Output-Buffer*)
;      (parallel-tyo (read-serial-buffer *Parallel-Output-Buffer*))))


(COMPILE-FLAVOR-METHODS PARALLEL-STREAM) 



