;;; -*- Mode:Common-Lisp; Package:sys; Base:8.; -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1986-1989 Texas Instruments Incorporated. All rights reserved.

;;;
;;; This file contains ECMB error correction event processing routines


;;; Edit History:
;;; 
;;; Date      Patcher Patch #  Description
;;;----------------------------------------------------------------
;;; 12-86       ab      --     Original. 



;;; Error checking support for ECMBs (Error Correcting Memory Boards)
;;;
;;; A warm-init function checks all slots in the chassis for ECMBs.
;;; The function to check for ECMB correctable error occurrences runs
;;; in the TM-UPDATE process, which executes every 5 minutes.  Detected
;;; ECMB correctable error events are posted to the system log.

(DEFVAR *all-ecm-boards* nil)

(DEFVAR *ecmb-part-number-list* '("00002540835"))

(DEFVAR *ecmb-syndrome-decode-table*
	(MAKE-ARRAY 128.
		    :element-type t
		    :initial-contents
		    '(multi-bit  2-bit      2-bit      multi-bit
		      2-bit      multi-bit  multi-bit  2-bit
		      2-bit      multi-bit  DB-31      2-bit
		      multi-bit  2-bit      2-bit      DB-30
		      2-bit      multi-bit  DB-29      2-bit
		      DB-28      2-bit      2-bit      DB-27
		      DB-26      2-bit      2-bit      DB-25
		      2-bit      DB-24      multi-bit  2-bit

		      2-bit      multi-bit  DB-7       2-bit
		      DB-6       2-bit      2-bit      DB-5
		      DB-4       2-bit      2-bit      DB-3
		      2-bit      DB-2       multi-bit  DB-2
		      DB-0       2-bit      2-bit      multi-bit
		      2-bit      DB-1       multi-bit  2-bit
		      2-bit      multi-bit  multi-bit  2-bit
		      multi-bit  2-bit      2-bit      CB-6

		      2-bit      multi-bit  multi-bit  2-bit
		      multi-bit  2-bit      2-bit      multi-bit
		      multi-bit  2-bit      2-bit      DB-15
		      2-bit      multi-bit  DB-14      2-bit
		      multi-bit  2-bit      2-bit      DB-13
		      2-bit      DB-12      DB-11      2-bit
		      2-bit      DB-10      DB-9       2-bit
		      DB-8       2-bit      2-bit      CB-5

		      multi-bit  2-bit      2-bit      DB-23
		      2-bit      DB-22      DB-21      2-bit
		      2-bit      DB-20      DB-19      2-bit
		      DB-18      2-bit      2-bit      CB-4
		      2-bit      DB-16      multi-bit  2-bit
		      DB-17      2-bit      2-bit      CB-3
		      multi-bit  2-bit      2-bit      CB-2
		      2-bit      CB-1       CB-0       none)))



(DEFUN Find-All-Ecm-Boards ()
  (SETQ *all-ecm-boards* nil)
  (LOOP WITH slots-i-own = (get-paging-parameter %Slots-I-Own)
	FOR  slot-num FROM 0 BY 1
	FOR  slot = (DPB slot-num (BYTE 4. 0.) #xF0)
	FOR  i-own-slot = (LDB-TEST (BYTE 1 slot-num) slots-i-own)
	UNTIL (>= slot-num 16.) DO
	
	;; Make sure we own the board, and don't include the main
	;; memory board, which has already been added to the configuration.
	(WHEN (AND i-own-slot
		   (memory-board-p slot)
		   (ecm-memory-p slot))
	  ;; It is an error-correcting memory board.
	  (PUSH slot *all-ecm-boards*)))
  )

(DEFUN ecm-memory-p (slot)
  (DOLIST (pn *ecmb-part-number-list*)
    (LOOP FOR i FROM 0 TO (1- (LENGTH (THE string pn)))
	  FOR char = (AREF pn i)
	  FOR rom-byte = (%nubus-read-8b-careful
				     slot (+ CROMO-Board-Part-Number (* i 4.)))
	  FOR rom-char = (WHEN (NUMBERP rom-byte) (INT-CHAR rom-byte))
	  DO
	  (WHEN (OR (NULL rom-char)
		    (CHAR-NOT-EQUAL char rom-char))
	    (RETURN nil))
	  FINALLY (RETURN-FROM ecm-memory-p t))))



(DEFUN fmt-ecmb-memory-error (record stream &aux tem)
  (FORMAT stream
	  "~2%ECM MEMORY BOARD MEMORY ERROR ~
            ~%   Time logged:         ~@18a  ~
            ~%   Memory board slot:   ~18,,d. ~
            ~%   Error type:          ~@18a  ~
            ~%   Error type bits:     ~18,,d. ~
            ~%   Row and bank:        ~18,,d. ~
            ~%   Syndrome code:       ~@18,,a ~
            ~%   Syndrome code bits:               #x+~2,'0,x  ~
            ~%   All error code bits:        #x+~8,'0,x)"
	  (time:print-universal-time
	    (SECOND (ASSOC :time record :test #'EQ)) nil)
	  (SECOND (ASSOC :board-slot record :test #'EQ))
	  (SELECT (SETQ tem (SECOND (ASSOC :error-type record :test #'EQ)))
	    (%ECMB-Failure-Type-No-Error      'no-error)
	    (%ECMB-Failure-Type-Correctable   'correctable)
	    (%ECMB-Failure-Type-Cant-Happen   'cant-happen)
	    (%ECMB-Failure-Type-Uncorrectable 'uncorrectable))
	  tem
	  (SECOND (ASSOC :row-and-bank record :test #'EQ))
	  (AREF *ecmb-syndrome-decode-table*
		(SETQ tem (SECOND (ASSOC :syndrome record :test #'EQ))))
	  tem
	  (SECOND (ASSOC :all-error-bits record :test #'EQ)))
  )

;; Formatting function must be available in the SYSLOG package.
(DEFF syslog:fmt-ecmb-memory-error 'sys:fmt-ecmb-memory-error)


(Add-Initialization "Find all ECM memory boards"
		    '(find-all-ecm-boards) '(:Warm :First))


(DEFUN check-ecm-boards-for-errors ()
  (DOLIST (slot *all-ecm-boards*)
    (LET* ((err-lo
	     (%nubus-read-8b-careful slot %ECMB-Memory-Failure-Register))
	   (err-hi
	     (%nubus-read-8b-careful slot (1+ %ECMB-Memory-Failure-Register)))
	   (error (WHEN (AND (NUMBERP err-lo) (NUMBERP err-hi))
		    (DPB err-hi (BYTE 8. 8.) err-lo)))
	   record)
      (WHEN (AND error
		 (LDB-TEST %%ECMB-Failure-Error-Bit error)
		 syslog:*log-enabled*)
	  (SETQ record
		(LIST
		  (LIST :FORMAT-F              'Fmt-ECMB-Memory-Error)
		  (LIST :ALL-ERROR-BITS        error)
		  (LIST :BOARD-SLOT            (LDB (BYTE 4 0) slot))
		  (LIST :ERROR-TYPE            (LDB %%ECMB-Failure-Error-Type error))
		  (LIST :ROW-AND-BANK          (LDB %%ECMB-Failure-Row-And-Bank error))
		  (LIST :SYNDROME              (LDB %%ECMB-Failure-Syndrome-Bits error))))
	  (syslog:add-record :ECM-MEMORY-BOARD-ERROR record))
      ;; Clear error status.
      (%nubus-write-16b slot %ECMB-Memory-Failure-Register 0))))
