;;; -*- Mode: Common-Lisp; Package: SYSLOG; Base: 8.; Patch-File: T -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1985-1989 Texas Instruments Incorporated. All rights reserved.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   Revisions:
;;;     12-19-88 DAB    Fixed View-log to not go into cold load when crash table is not found.
;;;     ************    Rel 5.0 OS 12-14-88 DAB
;;;     12-13-88 DAB    Added a new function called VIEW-LOG. THis function allows a user to
;;;                     view selected record in a constraint window.
;;;     11-16-87 DAB    Fixed Purge to initialize the log is necessary.
;;; 	10.13.87 MBC 	Eliminate munged output from SYSLOG:DUMP caused by bad header info by
;;;			making sure we don't use *HEADER-RQB* after a SYSLOG:PURGE has returned it.
;;;	03.25.87 MBC 	Be sure to set *CURRENT-BUFFER-BLOCK* so writes know where to.
;;;	03.24.87 MBC 	Drop use of Buffer-Address macro in GET-BUFFER-IN-MEMORY since 
;;;			 MSC board doesn't preserve the disk address field.
;;;	03.12.87 MBC	Check *log-enabled* before accessing *header-rqb* in Add-Record.
;;;			 [SPRs 3996 3921 4077 4088] 
;;;	3-24-86 MBC	Fix self lock out problem when adding a record that requires the
;;;			deletion (wrap around) of more than one record in Add-Record.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   ENTRY Points:
;;;   
;;;   BOOT-INIT      - This should be called only during boot.  It finds the System Log and
;;;                    initializes pointers to it.  If the log has never been initialized,
;;;                    it is initialized now.

;;;   PURGE          - Initializes the log and resets it's pointers to the start.

;;;   ADD-RECORD     - Writes an EVENT and RECORD to the log.  If there isn't enough room one or more of the
;;;                    oldest records will be deleted.  
;;;                    A record should be a list of ((keyword value) (keyword value) .... ).
;;;                    If Keyword :FORMAT-F is present then the value must be either a function
;;;                    or a list whose car is a function to be used to format the record.
;;;                    See Format-Record for details.  After conversion to a string and addition
;;;                    of a time stamp, record length is checked for an upper bound, about 1000 bytes.

;;;   READ-RECORD    - Read one record.  Starts at oldest and works up to newest record, returning
;;;                    one record at a time.  When NIL is returned there are no more records.
;;;                    One more call starts over at oldest record.

;;;   DUMP-LOG       - Just loops on Read-record & dumps entire log to specified stream.

;;;   TURN-OFF-LOG   - Sets a flag that disables logging.

;;;   TURN-ON-LOG    - Sets a flag to enable logging, but only if log has been initialized.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   DESIGN:
;;;      The system log is a parition that is managed as a circular buffer.
;;;   The first page is dedicated to pointers and overhead info.  There is a head pointer
;;;   which points to the next byte to put a record at.  The tail pointer points to the
;;;   last (oldest) record in the log. 
;;;      Wrap around of pointers is handled such that a record is never spit up between the end and
;;;   the start of the log.  If a record doesn't fit before the physical end of the partition,
;;;   some dead space is left (temporarily) and the head (insert) pointer is reset to start before
;;;   inserting the record.
;;; 
;;;   Two global flags can inhibit logging.
;;;     *LOG-ENABLED* (if nil) indicates the log is not initialized or has been turned-off.
;;;     *LOG-BUSY* (if T) indicates that the log is being manipulated and should not be modified.
;;;   Either of these two flags can cause a log record to fall in the proverbial bit bucket.
;;;
;;;
;;;
;;; USER HINTS:
;;;   NEWLINES not allowed in records --> they'll be converted to BLANKs.
;;;   RECORD must be in proper format
;;;         Max record length is 1000. bytes. 
;;;
;;;
;;;
;;;

(DEFUN LOG-SELF-TEST-FAILURES ()
  "If the last boot had any self-test failures, add a record about it to the system log."
  ()) 
;; Remove this print statement, since it shows up during boot otherwise.  -ab
;;  (print "Log-Self-Test-Failures : not implemented."))  

;;;
;;;
;;;

(DEFUN POINTER-INIT ()
  "Initialize physical disk pointers and internal pointers to the system log."
  (LET (START
	BLOCK-LENGTH)
    (MULTIPLE-VALUE-SETQ (START BLOCK-LENGTH)
			 (SYSTEM:FIND-DISK-PARTITION *LOG-NAME* () *LOG-UNIT*))
    (SETF *START-BLOCK* START
	  *END-BLOCK* (+ BLOCK-LENGTH START))
    (SETF *READ-RECORD* (MAKE-ARRAY 2000 :ELEMENT-TYPE 'STRING-CHAR :FILL-POINTER 0))
    (READ-HEADER))) 


(DEFUN ASSURE-INITIALIZED ()
  (READ-HEADER)
  (IF (AND (= (SI::AREF-32B (SYSTEM:RQB-BUFFER *HEADER-RQB*) INITED-KEY-INDEX) INITED-VALUE)
	   (= (SI::AREF-32B (SYSTEM:RQB-BUFFER *HEADER-RQB*) BLOCK-LENGTH-INDEX)
	      (- *END-BLOCK* *START-BLOCK*)))
      T
      (PURGE ()))) 


(DEFUN FIND-LOG ()
  "Look for a partition whose name is defined by SYSLOG:*LOG-NAME* on all
online disk units.  Returns the unit number or NIL if not found."
  (DOLIST (U (SI::ALL-DISK-UNITS) NIL)
    (IF (SYSTEM:FIND-DISK-PARTITION *LOG-NAME* () U () () :ATTRIBUTE SYSTEM:%BT-LOG-BAND :CPU-TYPE
				    SYSTEM:%CPU-CHAPARRAL)
	(RETURN U)))) 
;;;
;;;  Make sure all user modifiable things that affect the 
;;;  conversion of lisp objects to string output, are bound
;;;  for proper use.
;;;

(DEFUN PREP-RECORD (EVENT RCD)
  "Prepare a record by inserting time stamp and checking length."
  (LET ((*PRINT-BASE* 12)
	(*PRINT-LENGTH* NIL)
	(*PRINT-LEVEL* NIL)
	(*PRINT-RADIX* T)
	(*PRINT-PRETTY* NIL)
	(SI:PRINT-READABLY T))
    (SETF RCD
	  (FORMAT () "~s"
		  (CONS (LIST :EVENT EVENT) (CONS (LIST :TIME (GET-UNIVERSAL-TIME)) RCD)))))
  (IF (< 1750 (LENGTH RCD))
      (FERROR "Invalid SYSTEM LOG record :  Record size: ~d. too large." (LENGTH RCD)))
  RCD) 


(DEFUN ADD-1 (RCD)
  "Real workhorse for adding a record to the system log."
  (SETF *LOG-BUSY* T)
  (LET ((HEAD (GET-HEAD)) (TAIL (GET-TAIL)) (START 0) (END (GET-LENGTH))
	(SIZE (+ 2 (LENGTH RCD))))		;room for record seperators.
    (COND
      ((<= TAIL HEAD)				;first pass, tail follow head
       (IF (> SIZE (- END HEAD))
	   (PROGN
	     (WHEN (= TAIL START)
	       (DELETE-1 TAIL ()))
	     (PUT-HEAD START)			;wrap around head pointer
	     (ADD-1 RCD))
	   (COPY-RECORD RCD HEAD)))		;updates head pointer
      (T					;Normal, head follows tail
       (COND
	 ((> SIZE (- TAIL HEAD)) (DELETE-1 TAIL ()) (ADD-1 RCD))	;remove one message starting at tail
	 (T (COPY-RECORD RCD HEAD))))))) 



(DEFUN COPY-RECORD (RCD POINTER)
  "Copy record to buffer starting at ponter."
  (LET* ((BUFFER-OFFSET (GET-BUFFER-IN-MEMORY POINTER))	;force it into mem
	 (BUFFER (SYSTEM:RQB-8-BIT-BUFFER *BUFFER-RQB*))
	 (SIZE (LENGTH RCD)))
    (DOTIMES (I SIZE
		(PROGN
		  (SETF (AREF BUFFER (+ I BUFFER-OFFSET)) #\NEWLINE)	;two newlines [returns]
		  (SETF (AREF BUFFER (+ 1 I BUFFER-OFFSET)) #\NEWLINE)))
      (LET ((A-CHAR (AREF RCD I)))		;convert NEWLINES to SPACES!
	(SETF (AREF BUFFER (+ BUFFER-OFFSET I)) (IF (= #\NEWLINE A-CHAR)
						    #\SPACE
						    A-CHAR))))
    (WRITE-LOG)					;write out active buffer
    (PUT-HEAD (+ 1 POINTER SIZE))		;update header pointer [length plus newline]
    (WRITE-HEADER))
  T) 

;;;
;;;
;;;
(DEFUN GET-BUFFER-IN-MEMORY (LOG-OFFSET)
  "Read buffer into memory if offset is not already in buffer.
Return LOG-OFFSET's corresponding buffer offset into this buffer."
  (OR *BUFFER-RQB* (ALLOCATE-THE-BUFFER-RQB))
  (LET ((BLOCK-ADDRESS (+ (START-DATA-BLOCK) (FLOOR LOG-OFFSET 2000))))
    (IF (> (1+ BLOCK-ADDRESS)
	   *END-BLOCK*)
	(CERROR T () "Reading past end of log +1"))
    (setf *current-buffer-block* block-address)	;3.25.87
    (SYSTEM:DISK-READ *BUFFER-RQB* *LOG-UNIT* BLOCK-ADDRESS))	;Read every time.  3.24.87
  (- LOG-OFFSET (* (FLOOR LOG-OFFSET 2000) 2000)))


(DEFUN ALLOCATE-THE-BUFFER-RQB ()
  (SETF *BUFFER-RQB* (SYSTEM:GET-DISK-RQB 2))	;could be reading past the end of SYSTEM LOG partition!
  (SI::ASET-32B 37777777760 *BUFFER-RQB* 6)	;ILLEGAL it for it's address for check
  *BUFFER-RQB*) 

;;;
;;;
;;; 

(DEFUN DELETE-1 (POINTER &OPTIONAL (WRITE-HEADER? T))	;nulls ?  Off center?
  (LET* ((BUFFER-OFFSET (GET-BUFFER-IN-MEMORY POINTER))	;force it into mem
						;watch for wrap around
	 (SIZE (FIND-END-OF-NEXT-RECORD BUFFER-OFFSET)))	;figure out its length
    (IF (>= (+ 1 POINTER SIZE) (GET-LENGTH))	;if beyond end of log
	(PUT-TAIL 0)				;wrap around
	(PUT-TAIL (+ 1 POINTER SIZE)))		;change tail pointer in RQB
    (AND WRITE-HEADER? (WRITE-HEADER)))) 




(DEFUN FIND-END-OF-NEXT-RECORD (BUFFER-OFFSET)
  "Return the next end of the record starting at buffer-offset.  Buffer must be in memory."
  (LET* ((BUFFER (SYSTEM:RQB-8-BIT-BUFFER *BUFFER-RQB*))
	 (BUFFER-LENGTH (LENGTH BUFFER)))
    (DO ((I BUFFER-OFFSET (1+ I)))
	((>= I BUFFER-LENGTH) I)		;(cerror t nil "End of next record not present"))	;Maybe read next buffer? Check for end-block?
      (WHEN (= (AREF BUFFER I) #\NEWLINE)
	(RETURN (- I BUFFER-OFFSET)))))) 


;;;
;;;
;;;

(DEFUN GET-RECORD-FROM-BUFFER (BUFFER-OFFSET)
  "Return the next end of the record starting at buffer-offset and buffer-offset to the next record.
Buffer must be in memory."
  (LET* ((BUFFER (SYSTEM:RQB-8-BIT-BUFFER *BUFFER-RQB*))
	 (BUFFER-LENGTH (LENGTH BUFFER))
	 (RCD *READ-RECORD*))
    (SETF (FILL-POINTER RCD) 0)
    (DO ((I BUFFER-OFFSET (1+ I)))
	((>= I BUFFER-LENGTH)
	 I)
      (LET ((A-CHAR (AREF BUFFER I)))
	(WHEN (= (AREF BUFFER I) #\NEWLINE)
	  (SETF (FILL-POINTER RCD) (- I BUFFER-OFFSET))
	  (RETURN (VALUES RCD (1+ I))))
	(STRING-NCONC RCD (STRING A-CHAR)))))) 


;;;
;;;
;;;

(DEFUN DUMP-LOG-1 (OUT RAW)
  (LOOP
    (LET ((RCD (READ-RECORD)))
      (IF RCD
	  (IF RAW
	      (FORMAT OUT "~%~a" RCD)
	      (FORMAT-RECORD RCD OUT))
	  (RETURN ()))))) 

;;;
;;;  Format-record allows records to designate their own formating function.
;;;  Should be in SYSLOG package or specify its own package.  It must be a function
;;;  of at least two arguments, RECORD (the keyword/value list) and STREAM, the i/o
;;;  stream where the formatted information is written.  Other arguments are allowed,
;;;  and will follow RECORD and STREAM if present.
;;;  The general form in the keyword list is thus (:FORMAT-F SYSLOG:FN-NAME arg1 arg2).
;;;  This translates to the function call (SYSLOG:FN-NAME record stream arg1 arg2).
;;;    If the formatting function doen't exist, or is not of the proper form, the record
;;;  is output RAW.
;;;

(DEFUN FORMAT-RECORD (RECORD OUT)
  "Look within RECORD for the name of a function to format this record.  If found call it
to format the record, otherwise copy RECORD to OUT stream, looking for EVENT & TIME."
  (LET* ((*READ-BASE* 10.)
	 (*PRINT-BASE* 10.)
	 (*package* (find-package "SYSLOG"))	;force format fn into SYSLOG package. -ab
	 (RECORD-LIST (READ-FROM-STRING RECORD))
	 FORMAT-FUNCTION
	 BAD-ONE)
    (MULTIPLE-VALUE-SETQ (FORMAT-FUNCTION BAD-ONE)
			 (IGNORE-ERRORS (CDR (ASSOC :_+$$-GARBAGE RECORD-LIST))))
    (IF BAD-ONE
	(FORMAT OUT "~%~a " RECORD)
	(IF (SETF FORMAT-FUNCTION (CDR (ASSOC :FORMAT-F RECORD-LIST)))
	    (LET ((FORMAT-FUNCTION (CAR FORMAT-FUNCTION))
		  (REST-ARGS (CDR FORMAT-FUNCTION)))
	      (IF REST-ARGS
		  (IF (FUNCTIONP FORMAT-FUNCTION)
		      (APPLY FORMAT-FUNCTION (CONS RECORD-LIST (CONS OUT REST-ARGS)))	;This will work, -ab
		      (FORMAT OUT "~%~a" RECORD))
		  (IF (FUNCTIONP FORMAT-FUNCTION)
		      (FUNCALL FORMAT-FUNCTION RECORD-LIST OUT)	;Use list version, not string -ab
		      (FORMAT OUT "~%~a" RECORD))))
	    (PROGN
	      (FORMAT OUT "~%")
	      (LET ((ADD-TIME (ASSOC :TIME RECORD-LIST))
		    (THE-EVENT (ASSOC :EVENT RECORD-LIST)))
		(WHEN THE-EVENT
		  (FORMAT OUT "~a " (CADR THE-EVENT))
		  (SETF RECORD-LIST (DELETE THE-EVENT (THE LIST RECORD-LIST) :TEST #'EQ :COUNT 1))
		  (WHEN ADD-TIME
		    (TIME:PRINT-UNIVERSAL-TIME (CADR ADD-TIME) OUT)
		    (SETF RECORD-LIST (DELETE ADD-TIME (THE LIST RECORD-LIST) :TEST #'EQL :COUNT 1)))
		  (FORMAT OUT " ~a" RECORD-LIST)))))))) 


;;;;; External Functions:
(DEFUN BOOT-INIT ()
  "Initialize system log pointers, log abnormal shutdowns, and log self-test failures."
  (SETF *LOG-ENABLED* ())
  (WHEN (SETF *LOG-UNIT* (FIND-LOG))
    (UNLESS (SI::RESOURCE-OBJECT-IN-USE *HEADER-RQB* 'SI::RQB)
      (SETF *HEADER-RQB* (SYSTEM:GET-DISK-RQB)))
    (POINTER-INIT)
    (WHEN (ASSURE-INITIALIZED)
      (PROGN
	(LOG-SELF-TEST-FAILURES)
	(SETF *LOG-ENABLED* T)
	(WHEN (FUNCTIONP 'LOG-ABNORMAL-SHUTDOWN)
	  (LOG-ABNORMAL-SHUTDOWN)))))) 


(ADD-INITIALIZATION "Initialize the System Log" '(BOOT-INIT)) 


(DEFUN PURGE (&OPTIONAL (CLOBBER-QUERY T))
  "Initialize the system log partition, SYSLOG:*LOG-NAME*, on unit, SYSLOG:*LOG-UNIT*, purging all records.
Log partition must be at least five pages long.  Clobber partition with or without querying."
  (LET (START
	BLOCK-LENGTH)
    (unless *LOG-UNIT* (setq *LOG-UNIT* (find-log)))  ;11-2-87 DAB  BUG:6819
    (IF CLOBBER-QUERY
	(MULTIPLE-VALUE-SETQ (START BLOCK-LENGTH)
			     (SYSTEM:FIND-DISK-PARTITION-FOR-WRITE *LOG-NAME* () *LOG-UNIT*))
	(MULTIPLE-VALUE-SETQ (START BLOCK-LENGTH)
			     (SYSTEM:FIND-DISK-PARTITION *LOG-NAME* () *LOG-UNIT*)))
    (IF (AND START (> BLOCK-LENGTH 4))
	(PROGN
	  (SETF *START-BLOCK* START
		*END-BLOCK* (+ BLOCK-LENGTH START)
		*LOG-ENABLED* T)
	  (WHEN *HEADER-RQB*
	    (SYSTEM:RETURN-DISK-RQB *HEADER-RQB*)
	    (SETF *HEADER-RQB* NIL))		;10.13.87 MBC
	  (READ-HEADER)				;allocates new rqb if neccessary.
	  (LET ((LENGTH (* 2000 (1- BLOCK-LENGTH))))
	    (PUT-HEAD 0)
	    (PUT-TAIL 0)
	    (PUT-LENGTH LENGTH)
	    (PUT-INIT-TIME)
	    (SI::ASET-32B INITED-VALUE (SYSTEM:RQB-BUFFER *HEADER-RQB*) INITED-KEY-INDEX)
	    (SI::ASET-32B BLOCK-LENGTH (SYSTEM:RQB-BUFFER *HEADER-RQB*) BLOCK-LENGTH-INDEX)
	    (WRITE-HEADER)
	    (SYSTEM:UPDATE-PARTITION-COMMENT *LOG-NAME* (FORMAT () "System Log") *LOG-UNIT*)
	    (POINTER-INIT)
	    T))
	()))) 


;;;
;;;  This determines how soon to flag a warning about checking your
;;;  log for repetitous entries.  
;;;

(DEFVAR *WRAP-WARNING-TIME-DELTA* (* 250600 36)	;86400 ticks in a day.
  "If log wraps around before this number of ticks, issue notification.") 


(DEFUN ADD-RECORD (EVENT RECORD)
  "Add a RECORD of keyword type EVENT to the system log.  RECORD must be an A-LIST."
  (WHEN *LOG-ENABLED*				;3.12.87
    (LET ((OLDEST-TIME (OLDEST-RECORD-TIME))
	  (OLD-TAIL (GET-TAIL)))
      (UNLESS *LOG-BUSY*
	(UNWIND-PROTECT (ADD-1
			  (PREP-RECORD EVENT RECORD))
	  (SETF *LOG-BUSY* ())))
      (IF (AND OLDEST-TIME (/= OLD-TAIL (GET-TAIL)))
	  (IF (> *WRAP-WARNING-TIME-DELTA* (- (GET-UNIVERSAL-TIME) OLDEST-TIME))
	      (SETF (SYSTEM:SYSTEM-COMMUNICATION-AREA SI::%SYS-COM-SYSLOG-WRAP-EVENT) OLDEST-TIME)))))) 

;;;
;;;
;;;

(DEFUN READ-RECORD (&OPTIONAL (RESET NIL))
  "Reads one record from system log.  First call to gets oldest record.
Each successive call gets the next record, returning non-NIL value.  When last
record is read NIL is returned, and next call starts all over again.
If RESET then start over."
  (WHEN *LOG-ENABLED*
    (WHEN (OR RESET (NULL READ-LOG-STATE))
      (SETF READ-LOG-STATE (GET-TAIL)))
    (LET* ((POINTER READ-LOG-STATE)
	   (BUFFER-OFFSET (GET-BUFFER-IN-MEMORY POINTER)))
      (COND
	((= POINTER (GET-HEAD)) (SETF READ-LOG-STATE ()))
	(T
	 (LET* ((RCD (GET-RECORD-FROM-BUFFER BUFFER-OFFSET))
		(Size (LENGTH RCD)))
	   (IF (ZEROP SIZE)			;found double NEWLINE
	       (IF (< (+ POINTER 2000) (GET-LENGTH))	;extra debug check
		   (CERROR () () 'BAD-WRAP "~&Wraping pointer too early? Pointer=~d.  Length=~d."
			   POINTER (GET-LENGTH))
		   (PROGN
		     (SETF READ-LOG-STATE 0)	;Wrap around
		     (READ-RECORD)))
	       (PROGN
		 (SETF READ-LOG-STATE (+ 1 POINTER SIZE))
		 RCD)))))))) 

;;;
;;;
;;; New stuff to view log interactively.

(defvar *syslog-constraint-frame* ())
(defvar *SYSLOG-raw-list* ())

(defflavor syslog-window
	   ()
	   (w:line-truncating-mixin w:window)
  :settable-instance-variables
  :gettable-instance-variables
  :initable-instance-variables)



(defmethod (syslog-window :print-item) (item ignore ignore)
  (send self :string-out item))


(defflavor syslog-menu-frame
	   ()
	   (w:menu)
   :settable-instance-variables
  :gettable-instance-variables
  :initable-instance-variables)

(defmethod (syslog-menu-frame :before :init) (&rest ignore)
  (declare (special *SYSLOG-raw-list* ))
  (setq w:default-font fonts:MEDFNT)
  (setq w:margin-choices '(:abort))
  (setq w:item-list
	(or *SYSLOG-raw-list* 
	    (do ((record-string (read-record) (read-record))
		 (record)  (itemlist ()))
		((null record-string) (setq *SYSLOG-raw-list* itemlist))
	      (setq record (read-from-string record-string))
	      (push (list (format nil "~a : ~a"
				  (time:print-universal-time (second (second record)) ()) (second (car record)))
			  :value (string-append record-string)) itemlist))))
  
  (setq TV:label "SYSLOG Menu")
  )

(compile-flavor-methods  syslog-menu-frame  syslog-window)



(defun create-frame ()
  (setq *syslog-constraint-frame* 
	  (make-instance 'w:bordered-constraint-frame
			 :panes '((syslog-menu  syslog-menu-frame)
				  (syslog-detail syslog-window))
			 :constraints '((main . ((whole-thing)
						 ((whole-thing :horizontal (:even)
							       (syslog-menu syslog-detail)
							       ((syslog-menu .3)
								(syslog-detail .7))))
						 )))))
  
    (send *syslog-constraint-frame* :expose))

(DEFUN VIEW-LOG (&optional FORCE-reread)
  "Read log partition record by record and dump to OUT, which can be
a stream or pathname.  Format the output, headers, etc unless RAW is T."
  (declare (special *SYSLOG-raw-list* ))
  (SETF READ-LOG-STATE ())
  (when  FORCE-reread (setq *SYSLOG-raw-list* nil))
  (let (*syslog-constraint-frame* pane1 pane2 choice)
    (create-frame)
    (do-forever
      (setq pane1 (send *syslog-constraint-frame* :get-pane 'syslog-menu))
      (setf (send pane1 :chosen-item) nil)
      (process-wait "choose" #'(lambda (pane)
				 (setq choice (send pane :chosen-item))) pane1)
      (if choice
	  (progn 
	    (setq pane2 (send *syslog-constraint-frame* :get-pane 'syslog-detail))
	    (send pane2 :clear-screen)
	    (condition-case (condition)  ;12-19-88 DAB Otherwise you will get thrown int cold load on errors.
		(format-record (third choice) pane2)
	      (error (send pane2 :string-out (format nil "~%**ERROR:~%~s" (send condition :report-string))))))
	  (return)))
    ))

;;;;

(DEFUN DUMP-LOG (&OPTIONAL (OUT *STANDARD-OUTPUT*) (RAW NIL) (interactively? Nil))
  "Read log partition record by record and dump to OUT, which can be
a stream or pathname.  Format the output, headers, etc unless RAW is T.
If interactively, view log records in a window."
  (if interactively?
      (view-log )
      (SETF READ-LOG-STATE ())
      (IF (STRINGP OUT)
	  (WITH-OPEN-FILE (S OUT :DIRECTION :OUTPUT)
	    (DUMP-LOG-1 S RAW))
	  (DUMP-LOG-1 OUT RAW))))


(DEFUN OLDEST-RECORD-TIME ()
  (LET* ((*READ-BASE* 12)
	 (OLDEST-RECORD (READ-RECORD T))
	 RECORD-LIST)
    (IF OLDEST-RECORD
	(PROGN
	  ()
	  (SETF RECORD-LIST (READ-FROM-STRING OLDEST-RECORD))
	  (CADR (ASSOC :TIME RECORD-LIST)))
	()))) 


(DEFUN TURN-OFF-LOG ()
  (SETF *LOG-ENABLED* ())) 


(DEFUN TURN-ON-LOG ()
  (IF *START-BLOCK*
      (SETF *LOG-ENABLED* T))) 


;; Make available to syslog package.
(Deff Fmt-Crash-Record 'Sys:Fmt-Crash-Record) 

(Defvar *Syslog-Number-Crash-Records-To-Scan* #o3)  
;;; For use in logging crash information to the system log.
;;; Called during warm inits from the Syslog:Boot-Init function.

(Defun Log-Abnormal-Shutdown ()
  ;; Don't read it in unless NVRAM ok.
  (When (And (Sys:Nvram-Accessible-P) (Sys:Nvram-Initialized-P) (Sys:Crec-Format-Matches-P))
    (Sys:Clear-Crec-Vars)			;Just to be safe
    ;; Look at last couple of shutdown records.  Don't just look at last one,
    ;; because that may have been a crash from a warmboot attempt.  In that case,
    ;; we would never log the REAL crash record which is one before that.
    (Do* ((Crec (Sys:Crash-Rec-Find-Previous sys:Current-Crash-Rec-Offset)
		(Sys:Crash-Rec-Find-Previous Crec))
	  (Record (Sys:Get-Crash-Record Crec) (Sys:Get-Crash-Record Crec))
	  (I 0 (1+ I)))
	 ((>= I (Min *Syslog-Number-Crash-Records-To-Scan* (Sys:Number-Of-Crash-Records-In-Ring))))
      ;; Only log if record is reasonable, if shutdown abnormal, and if unlogged.
      (When (And (Not (Sys:Record-Unreasonable Record)) (Not (Sys:Normal-Shutdown-P Record))
		 (Not (Sys:Record-System-Logged Crec)))
	(Add-Record :Crash Record)
	;; So we won't log it again next time.
	(Sys:Mark-Record-System-Logged Crec))))) 





;;; This defines the format function for the system log records.  All hardware events will
;;;; be displayed with the same format when the user displays the system log via syslog:dump-log.
;;;; This function must be in the SYSLOG package.
(defun hardware-event-log-display (event-record log-stream)
  "This function will format the a-list recorded in the system log by above event handlers."
  (format log-stream
	  "~%~% Hardware Warning: ~a~% Time Recorded: "
	  (second (assoc :event event-record))	   ;extract hardware event type
	  )
  (time:print-universal-time
    (second (assoc :time event-record))	   ;extract the time
    log-stream)
  
  (format log-stream			   ;extract the event count
	  "~% Event count: ~d ~%"
	  (second (assoc :event-count event-record))
	  )
  (cond ( (equal (second (assoc :event event-record)) :mass-storage-overheat)
	 (format log-stream " Logical Units reporting overheat: ~a~%"
		 (second (assoc :logical-units event-record)))
	 )
	( (equal (second (assoc :event event-record)) :disk-retry-event)
	 (format log-stream " Unit Number:    #x~x ~% Status:         #x~x ~% Block Number:   #x~x~%"
		 (second (assoc :unit-number event-record))
		 (logand #xffffffff (second (assoc :status event-record)))
		 (logand #xffffffff (second (assoc :block event-record)))
		 )
	 )
	( (equal (second (assoc :event event-record)) :keyboard-error-event)
	 (format log-stream " Keyboard Error(s) reported: ~a~%"
		 (second (assoc :error-type  event-record)))
	 )
	( (equal (second (assoc :event event-record)) :NuBus-parity-error)
	 (format log-stream " NuBus address: #x~16r~%"
		 (second (assoc :NuBus-address event-record)))
	 )
	
	)				   ;cond
  )
