;;; -*- Mode:Common-Lisp; Package:UCL; Fonts:(CPTFONT HL10B); Base:10 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (c) 1984- 1989 Texas Instruments Incorporated.  All rights reserved.


1;;The following is the flavor definition and methods for UCL:COMMAND-TABLE.*

(DEFFLAVOR COMMAND-TABLE
	   ((name                "Command Table")
	    (symbol-name         NIL)
	    (documentation       "")
	    (commands            (MAKE-TABLE 20.))
	    (name-table          (MAKE-TABLE 20.))
	    (key-table           (MAKE-TABLE 20.))
	    (table-sorts         '(:COMMAND sort-command-table :NAME sort-name-table :KEY sort-key-table))
	    (command-lookup-fun  'lookup-command)
	    (name-lookup-fun     'lookup-name)
	    (key-lookup-fun      'lookup-key)
;;Remove these comments after the J-Build.  These vars aren't needed anymore.
;;	    (parsed-display-form NIL)
;;	    (parsed-edit-form    NIL)
	    (visible?            T)
	    (editable?           T))
	   (SI:PROPERTY-LIST-MIXIN)
  :GETTABLE-INSTANCE-VARIABLES
  :INITABLE-INSTANCE-VARIABLES 
  :SETTABLE-INSTANCE-VARIABLES
  (:SPECIAL-INSTANCE-VARIABLES documentation)
  (:DOCUMENTATION :COMBINATION
   1"This is the definition of a command table for use in the Universal Command
Loop System.  It is a grouping of UCL:COMMAND instances which have some
relation to each other, whether by functionality or by command context."

;;;The UCL:COMMAND-TABLE instance variables are:
;;;
;;;  :NAME       --Name for this command table.  This should describe what kind of commands are in this
;;;                  table to differentiate them from other command-tables when the user is in the UCL
;;;                  Command Display.  
;;;
;;;  :SYMBOL-NAME --This is initialized in BUILD-COMMAND-TABLE to the symbol you have set to this
;;;                  COMMAND-TABLE.  The symbol makes it easy to restore command environments from disk,
;;;                  since all we do to install saved COMMAND-TABLEs is to set the SYMBOL-NAMEs to new
;;;                  COMMAND-TABLE instances.  It is the SYMBOL-NAMEs, not the instances, which are stored
;;;                  in the instance variables of the command loop and used for command lookup.  
;;;
;;;  :DOCUMENTATION --General help for this command table.  May be a string or a form to evaluate to
;;;                  produce a string.  The string may be viewed by the user when in the Command Display. 
;;;                  Your strings should describe basically what commands in this table do (compared to those
;;;                  in other command tables).  
;;;
;;;  :COMMANDS  --All instances of UCL:COMMAND in this COMMAND-TABLE.  This is an array whose
;;;                  elements are UCL:COMMAND instances.  
;;;
;;;  :NAME-TABLE --Name-to-command table automatically generated for looking up typed command names. 
;;;                  It is an array of lists of command name followed by the UCL:COMMAND instance.  The
;;;                  names are derived from each command's :NAMES field, modified by substituting dashes for
;;;                  blanks.  
;;;
;;;  :KEY-TABLE  --Keystroke-to-command table automatically generated for looking up keystroke
;;;                  commands.  The table is generated from the :KEYS fields of each command in :COMMANDS.
;;;
;;;                  These :KEYS fields contain lists of keystroke sequences (lists of lists of fixnums).  Usually
;;;                  the keystroke sequences are one long.  For instance, the FORWARD command in Zmacs
;;;                  would have a :KEYS field of ((#\control-a)); but the Save-buffer command would have
;;;                  ((#\control-x #\control-s)).  The key-table is an array of lists of character followed
;;;                  either the UCL:COMMAND instance (in the case of single-key sequences) or another
;;;                  key-table (in the case of multi-key sequences).  These key-subtables are exactly the
;;;                  same as the as the KEY-TABLE,so recursion is used in table lookup and build.  In a Zmacs
;;;                  command table, there would be an entry (#\control-x sub-table) where sub-table is an
;;;                  array of lists such as (#\control-s <ucl:command>).  
;;;
;;;  :TABLE-SORTS --Keyword value pairs providing sorting functions for the three tables.  The allowed
;;;                  keywords are :COMMAND, :NAME, and :KEY which correspond to the instance variables
;;;                  COMMANDS, NAME-TABLE, and KEY-TABLE.  If a keyword isn't provided or its value is NIL,
;;;                  its table isn't sorted.  Otherwise the value is a function to call to sort the table.  This
;;;                  function should take one argument, the table (an array), and sort its contents
;;;                  destructively.  If you turn off or supply your own sorts, you must coordinate with the
;;;                  three lookup functions below.  There are three lookup functions available for use when
;;;                  turning off the sorts: UCL:UNSORTED-LOOKUP-COMMAND, UCL:UNSORTED-LOOKUP-KEY, and
;;;                  UCL:UNSORTED-LOOKUP-NAME.  
;;;
;;;  :COMMAND-LOOKUP-FUN -- Function which looks up a command instance in the COMMANDS table,
;;;                  returning NIL if not found or the instance and array index if found.  Takes two arg; the
;;;                  command instance to look up and the COMMANDS table to search.
;;;
;;;  :NAME-LOOKUP-FUN --Function which looks up a command name in the name table, returning a
;;;                  ucl:COMMAND or NIL.  Takes 2 args; the string to look up and the name table to search.  
;;;
;;;  :KEY-LOOKUP-FUN --Function which looks up a keystroke (or mouse button) in a key table or subtable. 
;;;                  See above on :KEY-TABLE for an explanation of key-subtables.  The function takes two
;;;                  arguments; a keystroke (character) and an array, and returns either the data for the
;;;                  keystroke (a UCL:COMMAND or a subtable) or NIL if not found.  See also the method
;;;                  :LOOKUP-KEYS which uses this function.
;;;
;;;  :VISIBLE?     --Indicates whether or not to display.  If not T nor NIL,should be a form to EVAL to get a
;;;                  list of scroll window items to be displayed one per line in the command display. Something
;;;                  like (list (tv:scroll-parse-item ...) (tv:scroll-parse-item ...))  would be apropriate. 
;;;
;;;  :EDITABLE?   --Indicates whether or not to allow this command-table to be edited.   
;;;
;;;The SI:PROPERTY-LIST-MIXIN has been included to facilitate storing of application- dependent data relevant
;;;to command tables.  Feel free to use it.*

))


(Defmethod (Command-Table :ADD-COMMAND) (command &Optional (update-tables? T))
  1"Adds a new UCL:COMMAND instance to this command table and returns it.  If
UPDATE-TABLES is non-NIL, updates the various command tables for SELF.  If a command
with the same set of names already exists then it is replaced instead of adding a new
command."*
  (DECLARE (values command))
  (CHECK-ARG command (TYPEP command 'COMMAND) "a UCL:COMMAND" command)
  (MULTIPLE-VALUE-BIND (found index) (SEND SELF :LOOKUP-COMMAND command)
    (IF found
	(SETF (AREF Commands index) command) 
      (VECTOR-PUSH-EXTEND command Commands)))
  (WHEN update-tables?1                              ;Update the tables.*
    (SEND SELF :BUILD-TABLES T))
  command) 


(Defmethod (Command-Table :BUILD-TABLES) (&Optional (sort-commands? NIL))
1  "Builds all of the automatically generated lookup tables of the COMMAND-TABLE, and
reparses the command table for the command display and editor.  This should be invoked
after a certain changes have been made to the UCL commands in a COMMAND-TABLE."*
  (LET ((command-sort (GET (MAKE-PLIST Table-Sorts) ':COMMAND)))   
    (WHEN (AND sort-commands? command-sort)1      ;Sort the commands if necessary.*
      (FUNCALL command-sort Commands)))
  (SEND SELF :BUILD-KEY-TABLE)
  (SEND SELF :BUILD-NAME-TABLE))

(Defun SORT-COMMAND-TABLE (commands)
  1"Default sort function for the COMMANDS instance variable of UCL:COMMAND-TABLE."*
  (SORT Commands #'Command-Less-Than?))

(Defmethod (Command-Table :BUILD-KEY-TABLE) ()
  1"Re-builds the table which maps characters to UCL commands.  This message should be sent to
the COMMAND-TABLE whenever the KEYS field of one of its active commands is changed, when an
inactive command with a non-null KEYS field is made active, or when a new active command with a
non-null KEYS field is added."
  ;; The KEY TABLE is a sorted array, which is overlayed with a list, whose elements are
  ;; lists of the character for the command followed by one of two things:
  ;; 1.  The command (an instance of ucl:COMMAND).  This is the case where a command 
  ;;     is invoked by one keystroke.  
  ;; 2.  An array which is another key table.  This is the case where commands are
  ;;     invoked by more than one keystroke.  For example, CONTROL-X CONTROL-F.  The
  ;;     array is structured exactly like the KEY-TABLE, so that commands may be defined
  ;;     having any number of keystrokes.  For instance, (#\control-x <array>) would be
  ;;     an entry in a Zmacs command-table, and (#\control-f <ucl:command for Find-FIle>)
  ;;     would be in an entry in control-x's key array/key table.*
  (DECLARE (values key-table))
  (SETF (FILL-POINTER Key-Table) 0)1             ; Empty the KEY TABLE by setting the fill pointer to zero.
  ;;Loop through all of the active commands in this COMMAND-TABLE, *
  (LOOP FOR command BEING THE ARRAY-ELEMENTS OF Commands
	WHEN (SEND command :ACTIVE?)
	DO (DOLIST (key-sequence (SEND command :KEYS))
             (Add-To-Key-Table KEY-TABLE key-sequence command)))
1  ;;Now sort the key table (including its sub-tables if any) to allow fast binary searches.*
  (LET ((key-table-sort (GET (MAKE-PLIST Table-Sorts) :KEY)))
    (WHEN key-table-sort
      (FUNCALL key-table-sort KEY-TABLE))))


(Defun ADD-TO-KEY-TABLE (table key-sequence command)
  1"Adds a new KEY-SEQUENCE for COMMAND to command table, TABLE.  DOES NOT SORT the
table!  KEY-SEQUENCE is a list of one or more keys which invokes COMMAND.  Each key in the
sequence indexes to either a command, or a subtable for the next key (as CNTL-X does).  This
function doesn't error check for ambiguous keys (that is left up to a future feature of the UCL
Command Editor) so ambigous keystrokes are added, and the search will find commands however
they fall."
  ;;   Does one of the following:
  ;; 1. Key-sequence is a single key, so add an array entry of (<key> <command>).
  ;; 2. Key-sequence starts with a key already in the table, so add the rest of the sequence 
  ;;    to its subtable (recursively).
  ;; 3. Key-sequence starts with a key not in the table, so create a subtable using a copy 
  ;;    of the same structure (list or some array type) of TABLE, add an entry of 
  ;;    (<first key> <subtable>), and add the rest of the key sequence to the new subtable (recursively).*
  (LET ((found-entry (Unsorted-Lookup-Key (CAR key-sequence) table))
	(subtable))
    (COND
      ((= 1 (LENGTH key-sequence))1              ;Add a (new) 1-char command to the table.*
       (VECTOR-PUSH-EXTEND (LIST (CAR key-sequence) command) table))
      ((AND found-entry (NOT (TYPEP found-entry 'COMMAND)))1     ;Recursively add rest of key seq. to the found*
       (ADD-TO-KEY-TABLE found-entry (CDR key-sequence) command))1       ;subtable.*
      (T
       (SETQ subtable (MAKE-TABLE 20.))
       (replace subtable table)1         ;Add a new multi-key subtable*  -GRH for BJ 11/13/86
       ;;(SETQ subtable (COPY-SEQ table)) 
       (SETF (FILL-POINTER subtable) 0)
       (ADD-TO-KEY-TABLE subtable (CDR key-sequence) command)
       1;;Pair key with its sub-table.*
       (VECTOR-PUSH-EXTEND (LIST (CAR KEY-SEQUENCE) SUBTABLE) TABLE)))))


(Defun SORT-KEY-TABLE (table)
  1"Sorts the vector, TABLE, into increasing order by the Character in the Car of each array
element.  Default sort function for the KEY-TABLE instance variable of ucl:COMMAND-TABLE."*
  (Declare (values sorted-table))
  (SORT table #'CHAR< :key #'car)1                       ;Sort TABLE in collating order.*
  (LOOP FOR (key command-or-subtable) BEING THE ARRAY-ELEMENTS OF table
	UNLESS (TYPEP command-or-subtable 'COMMAND)1     ;Recursively sort any subtables.*
	DO (SORT-KEY-TABLE command-or-subtable))
  table)


(DEFMETHOD (COMMAND-TABLE :BUILD-NAME-TABLE) ()
1  "Re-Builds the table which maps names to commands.  The table is a sorted array (overlayed with
a list) whose elements are lists containing a name of the command and the command.  Each name has
dashes substituted for spaces.  This message should be sent to the COMMAND-TABLE whenever the
NAMES field of one of its active commands is changed, when an inactive command with a non-null
NAMES field is made active, or when a new active command with a non-null NAMES field is
added."*
  (DECLARE (values Name-Table))
  (SETF (FILL-POINTER Name-Table) 0)1            ;Clear out the table.
   ;;Loop through all of the active commands in this COMMAND-TABLE.*
  (LOOP FOR command BEING THE ARRAY-ELEMENTS OF COMMANDS
        DO    1 ;;Add each typein name to the Name-Table.*
        (DOLIST (typein-name (SEND command :TYPEIN-NAMES))
	(VECTOR-PUSH-EXTEND (LIST typein-name command) Name-Table)))
  1;; Now sort the name table to allow fast binary searches.*
  (LET ((name-table-sort (GET (MAKE-PLIST table-sorts) :NAME)))
    (WHEN name-table-sort
      (FUNCALL name-table-sort Name-Table))))

(DEFUN sort-name-table (name-table)
  1"Sorts the array TABLE into alphabetical order by the CAR of each element.  Default sort
  function for the NAME-TABLE instance variable of UCL:COMMAND-TABLE."*
  (SORT name-table #'STRING-LESSP :key #'car))


(Defmethod (Command-Table :DISPLAY-LIST) ()
  1"Returns the display list for the UCL Command Display.  This display list is in a form suitable
for use by a scroll window with each active and visible command in the COMMAND-TABLE
displayed on one line in the scroll window, if VISIBLE?  is T.  If NIL, nothing is displayed."*
  (DECLARE (values display-list-or-NIL))
  (WHEN visible?
    (list* ""                                  1; Include white space to separate command tables.*
           SELF
           (MULTIPLE-VALUE-BIND (name-field key-field)  1;Column headers:*
               (Fields-For-Display T)
             (FORMAT NIL "~vA~vA~A" name-field "Command Name" key-field "Assigned Keystroke" "Description"))
	   1;; For each active and visible command in the COMMAND-TABLE, get its parsed*
	   1;; display form and add it to the display list.*
           (LOOP FOR command BEING THE ARRAY-ELEMENTS OF Commands
                 IF (AND (SEND command :ACTIVE?) (SEND command :VISIBLE?))
                 COLLECT command))))


(Defmethod (Command-Table :EDIT-DISPLAY-LIST) ()
  1"Returns a display list for the UCL Command Editor, including inactive and invisible commands.
This display list is in a form suitable for use by a scroll window with each command in the
COMMAND-TABLE displayed on one line in the scroll window.  If the command-table isn't
:EDITABLE?  then Nil is returned as an empty display list."*
  (DECLARE (values display-list-or-NIL))
  (WHEN Editable?
    (list* ""                                   1;Include white space to separate command tables.*
           SELF
           (MULTIPLE-VALUE-BIND (name-field key-field)  1;Column headers:*
               (Fields-For-Display T)
             (FORMAT NIL "~vA~vA~A" name-field "Command Name" key-field "Assigned Keystroke" "Description"))
           1;;G-L-P is a list of commands stored in array COMMANDS.*
           (G-L-P Commands))))

(defmethod (command-table :delete-command) (command)
  ;; remove a command from the table.
  (declare (type command command))
  (flet ((delete-from-vector (object vector key)
	     (let ((j 0)
		   (count (fill-pointer vector)))
	       (dotimes (i count)
		 (let ((elt (aref vector i)))
		   (unless (eq object (funcall key elt))
		     (unless (= i j)
		       (setf (aref vector j) elt))
		     (incf j))))
	       (setf (fill-pointer vector) j)
	       (loop while (< j count)
		     do (progn (setf (aref vector j) nil)
			       (incf j)))
	       vector)))
    (setf commands (delete-from-vector command commands #'identity))
    (setf key-table (delete-from-vector command key-table #'second))
    (setf name-table (delete-from-vector command name-table #'second))
    (send self :build-tables t)
    (values)))

(Defmethod (Command-Table :LOOKUP-KEYS) (key-sequence &Optional (table KEY-TABLE)
                                         &Aux command-or-subtable)
  1"Looks up a KEYSTROKE-SEQUENCE (or mouse buttons) in this COMMAND-TABLE's Key
Table.  KEY-SEQUENCE is a list of characters.  Returns: <UCL:COMMAND instance>, if a
command matches KEY-SEQUENCE.  Or NIL, if no command matches KEY-SEQUENCE.  Or
(:FETCH-MORE-KEYS subtable), when there are one or more commands which start with
KEY-SEQUENCE, but require more keys, SUBTABLE is the key-table where the search will start
with the next key."*
  (DECLARE (values command-or-flag-or-nil nil-or-subtable))
  (COND
    ((NULL key-sequence)                1;Ran out of keys, but there were command tables still claiming them.*
     (VALUES :FETCH-MORE-KEYS table))
    ((SETQ command-or-subtable (FUNCALL key-lookup-fun (FIRST key-sequence) table))
     (IF (TYPEP command-or-subtable 'COMMAND)
	 command-or-subtable            1;Return the Command found in this table.*
       ;1; Check the next key in the subtable.*
       (IF (NULL (CDR key-sequence))
           (VALUES :FETCH-MORE-KEYS command-or-subtable)        ;1Ran out of keys, get more.*
         ;1;Recursive call to this method.  Look up the rest of the keys in the key subtable we found.*
         (SEND SELF :lookup-keys (CDR key-sequence) command-or-subtable))))))


(Defun LOOKUP-KEY (key array &Optional (first 0.) (last  (1- (ARRAY-ACTIVE-LENGTH array))))
  1"Uses a binary search to look up KEY, a character, in ARRAY and if found, returns the KEY's
data item, otherwise Nil.  The array must have entries of the form (<key> <data>).  This is the
default function used by (UCL:Command-Table :LOOKUP-KEYS)."*
  1;; Note that this is compiled into a loop (not recursion), because of the Tail Recursion.*
  (DECLARE (OPTIMIZE (SPEED 3) (SAFETY 0))
	   (values command-or-subtable))
  (UNLESS (> first last)                1;If first is bigger than last, then the item obviously cannot be found.*
    (LET* ((middle      (HALF (+ first last)))  1; Calculate the index of the middle item in the range*
	   (middle-item (AREF array middle))    1; and get it.*
	   (middle-key  (FIRST middle-item)))
      (COND ((CHAR= key middle-key) (SECOND middle-item))       1;Found the key, so return the value.*
	    ((CHAR< key middle-key) (LOOKUP-KEY key array first (1- middle)))   1;Search lower half*
	    (T (LOOKUP-KEY key array (1+ middle) last)) 1;Search upper half*
            )))) 


(Defmethod (Command-Table :LOOKUP-COMMAND) (command)
  1"Looks up the UCL command, COMMAND, in this COMMAND-TABLE."*
  (DECLARE (values command-entry command-index))
  (CHECK-ARG command (TYPEP command 'COMMAND) "a UCL:COMMAND instance" command)
  (FUNCALL COMMAND-LOOKUP-FUN command COMMANDS))

(Defun LOOKUP-COMMAND (command array &Optional (first 0.) (last (1- (ARRAY-ACTIVE-LENGTH array))))
  1"Looks up COMMAND in ARRAY and returns the found command & its index in ARRAY.  Returns
NIL if not found.  ARRAY is an array of UCL:Command instances & COMMAND is a
UCL:Command instance."*
  1;; Note - the Tail Recursion compiles to interation.*
  (DECLARE (OPTIMIZE (SPEED 3) (SAFETY 0))
	   (values (command index))
	   (INLINE Command-Equal? Command-Less-Than?))
  (UNLESS (> first last)1                   ;If first is bigger than last, the item obviously cannot be found.*
    (LET* ((middle         (HALF (+ first last)))       1;Calculate the index of the middle item in the range *
	   (middle-command (AREF array middle)))        1;and get it.*
      (COND ((Command-Equal? command middle-command)
	     (VALUES middle-command middle))    1;Found it, so return command & its index.*
	    ((Command-Less-than? command middle-command)
	     (LOOKUP-COMMAND command array first (1- middle)))  1;Search lower half.*
	    (T (LOOKUP-COMMAND command array (1+ middle) last)) 1;Search upper half*
            ))))


(Defmethod (Command-Table :LOOKUP-NAME) (command-name)
1  "Searchs this Command Table and returns the UCL:Command named COMMAND-NAME.  Returns
NIL if no command has this name"*
  (DECLARE (values command-instance-or-nil))
  (FUNCALL NAME-LOOKUP-FUN command-name NAME-TABLE))



(Defun LOOKUP-NAME (name array &Optional (first 0.) (last  (1- (ARRAY-ACTIVE-LENGTH array))))
  1"Searches ARRAY for NAME and if found, returns NAME's data item.  The array must have
entries of the form (<name> <data>).  Returns NIL if nothing found."
  ;; This function is used by (UCL:COMMAND-TABLE :LOOKUP-NAME) where ARRAY's entires are
  ;; (<name> <ucl:command instance>).
  ;; Note - the Tail Recursion compiles to iteration*
  (DECLARE (OPTIMIZE (SPEED 3) (SAFETY 0))
	   (values cadr-of-found-entry))
  (UNLESS (> first last)                1;If first is bigger than last, the item obviously cannot be found.*
    (LET* ((middle      (HALF (+ first last)))  1;Calculate the index of the middle item in the range*
	   (middle-item (AREF array middle))    1; and get it.*
	   (middle-name (FIRST middle-item)))
      (COND ((STRING-EQUAL name middle-name)
	     (SECOND middle-item))      1;Found it, so return its DATA item.*
	    ((STRING-LESSP name middle-name)
	     (LOOKUP-NAME name array first (1- middle)))        1;Search lower half*
	    (T (LOOKUP-NAME name array (1+ middle) last))       1;Search upper half*
            ))))


(Defmethod (Command-Table :CLEAR) ()
  1"Clears all command in this Command-Table from the KEY-, NAME-, and COMMAND- tables."*
  (DOLIST (table (LIST key-table commands name-table))
    (DOTIMES (i (ARRAY-ACTIVE-LENGTH table))
      (SETF (AREF TABLE I) nil) )
    (SETF (FILL-POINTER table) 0)))


(Defmethod (Command-Table :MENU-LIST) ()
  1"Utility to produce a list of menus (menu symbols) which COMMANDs in this
COMMAND-TABLE are on."*
  (DECLARE (values menu-symbols))
  (LET (menus)
    (LOOP FOR command BEING THE ARRAY-ELEMENTS OF Commands
	  DO (LOOP FOR (menu . options) IN (SEND command :MENUS)
		   UNLESS (MEMBER menu menus :TEST #'EQUAL) 
		   DO (PUSH menu menus)))
    menus))


(Defmethod (Command-Table :BUILD-COMMAND-TABLE-FORM) (command-table-name)
  1"Returns a text string (a call to BUILD-COMMAND-TABLE) storing the current state of this
COMMAND-TABLE.  COMMAND-TABLE-NAME should be the symbol whose value is this
COMMAND-TABLE instance.  It is used as the first argument to BUILD-COMMAND-TABLE.

If you design your own flavor of COMMAND-TABLE with more data fields than
COMMAND-TABLE has, you need to also redefine this method for your own flavor.  This is a
good reason to make use of the property list mixin of COMMAND-TABLE, instead of adding your
own instance variables."*
  (declare (values build-command-table-source-string))
  (PKG-BIND 'global
    (FORMAT NIL "~%(BUILD-COMMAND-TABLE '~s NIL ~%  '(~{~S ~%    ~}))"
	  command-table-name
	  (LOOP FOR command BEING THE ARRAY-ELEMENTS OF Commands
		COLLECT (SEND command :DEFNAME)))))

(Compile-Flavor-Methods Command-Table)

1;; These three exist mainly for the UCL:TEMPORARY-COMMAND-TABLE flavor, but are put
;; here because UCL:COMMAND-TABLE makes use of UNSORTED-LOOKUP-KEY in building the key table.*

(Defun UNSORTED-LOOKUP-KEY (key array &Rest ignore)
  1"This function does the same thing as UCL:LOOKUP-KEY, but assumes that the key table
(ARRAY) is unsorted.  KEY is a charactor."*
  (DECLARE (values cadr-of-found-entry-or-nil))
  (LOOP FOR entry BEING THE ARRAY-ELEMENTS OF array
	WHEN (CHAR= key (FIRST entry))
	RETURN (SECOND entry)))

(Defun UNSORTED-LOOKUP-NAME (name array &Rest ignore)
  1"This function does the same thing as UCL:LOOKUP-NAME, but assumes that the name table
(ARRAY) is unsorted."*
  (DECLARE (values cadr-of-found-entry-or-nil))
  (LOOP FOR entry BEING THE ARRAY-ELEMENTS OF array
	WHEN (STRING-EQUAL name (FIRST entry))
	RETURN (SECOND entry)))

(Defun UNSORTED-LOOKUP-COMMAND (command array &Rest ignore)
  1"This function does the same thing as UCL:LOOKUP-COMMAND, but assumes that the command
table (ARRAY) is unsorted."*
  (DECLARE (values command-entry command-index))
  (LOOP FOR index FROM 0 BELOW (ARRAY-ACTIVE-LENGTH array)
	WHEN (Command-Equal? command (AREF array index))
	RETURN (VALUES command index)))


