;;; -*- Mode:Common-Lisp; Package:X11; Fonts:(MEDFNB HL12B HL12BI); Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

;;;
;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 12/14/88    LGO	Fixed BITBLT-UNDER-MASK to not barf on negative width or height
;;; 11/03/88    DAN     Fixed BITBLT-UNDER-MASK to not BITBLT if width or height is negative.
;;; 10/27/88	WJB	Made argument checking optional in BITBLT-UNDER-MASK -- let BITBLT do it.
;;;			Also note that the range checking is done before the occlusion stack is
;;;			taken into account, which may cause uneccesary errors.
;;;  8/25/88	WJB	Fixed bug in BITBLT-UNDER-MASK causing *only* the ones in the
;;;			source to be changed in the destination and leaving the zeros
;;;			in the source uncopied (regardless of the mask).  Also eliminates
;;;			2 bitblts and removes the need for 1 temp pixmap.
;;;  8/16/88	WJB	Changed BITBLT-UNDER-MASK to use a resource for its temp pixmaps.
;;;  2/19/88    TWE	Changed trace code to call the server-trace function instead.
;;;  1/18/88    TWE	Fixed bitblt-under-mask's to use Y component with HEIGHT instead
;;;			of X.
;;;  1/13/88    TWE	Changed references to tv:initial-lisp-listener to
;;;			*monochrome-server-trace*.  Commented out debug code.
;;;  1/06/88    TWE	Updated BITBLT-UNDER-MASK to associate the box offsets only with
;;;			the destination and mask coordinates
;;; 12/22/87    TWE	Updated BITBLT-UNDER-MASK to take an optional argument which is
;;;			the occlusion stack.
;;; 11/04/87    TWE	Initial creation.

;;; This file contains those functions which are too slow to execute in Lisp
;;; and which need to be implemented in microcode.

(DEFCONSTANT W:ALU-SET1 #B1111 "Turn on bits in the destination.")

;;; The differences between BITBLT and this function are described here:
;;;   The SOURCE and DEST arrays must be of the same type.  BITBLT allows
;;;   them to be of different integer types and does a different kind of
;;;   element combination.  This type of capatility will not be needed by X
;;;   so we don't bother with implementing it.
;;;
;;;   The widths of the SOURCE and DEST arrays times the array element sizes
;;;   need not be an integral multiple of 32.  For example, for 8 bit elements,
;;;   the width of such an array need not be an integral multiple of 4.  This
;;;   is a restriction of BITBLT and does not exist with BITBLT-UNDER-MASK.

;;; If the size of the mask rectangle is smaller than the copy rectangle then
;;; that part of the copy rectangle outside of the mask rectangle is not
;;; copied.  If the mask rectangle is larger than the mask array then the
;;; mask rectangle is altered to just fit within the mask array.  The (X,Y)
;;; point for the mask may specify a point outside of the mask (above or to
;;; the left) so that the mask would be inside of the copy rectangle.
(DEFUN BITBLT-UNDER-MASK (X-ALU WIDTH HEIGHT MASK MASK-X MASK-Y
                          SOURCE SOURCE-X SOURCE-Y DEST DEST-X DEST-Y
                          &OPTIONAL (OCCLUSION-STACK T) CHECK-ARGS)
  "Combines a SOURCE pixel with a DEST pixel when the corresponding MASK bit it 1."
  ;; Perform type checking first.
;;;  (server-trace "~%bitblt-under-mask, alu-~d, (w,h)=(~d,~d), ~
;;;mask=~a, (~d,~d), source=~a, (~d,~d), dest=~a, (~d,~d), occlusion-stack=~a"
;;;          X-ALU WIDTH HEIGHT MASK MASK-X MASK-Y
;;;                          SOURCE SOURCE-X SOURCE-Y DEST DEST-X DEST-Y
;;;                          occlusion-stack)
  (WHEN CHECK-ARGS
    (WHEN (OR (NOT (INTEGERP X-ALU))
	      (> X-ALU 16)
	      (MINUSP X-ALU))
      (ERROR "ALU, ~S,  must be an integer between 0 and 15." X-ALU))
    (WHEN (NOT (INTEGERP WIDTH))
      (ERROR "Width, ~S, must be an integer." WIDTH))
    (WHEN (NOT (INTEGERP HEIGHT))
      (ERROR "Height, ~S, must be an integer." HEIGHT))
    (WHEN (OR (NOT (ARRAYP MASK))
	      (NOT (= (ARRAY-RANK MASK) 2))
	      (NOT (EQ (ARRAY-ELEMENT-TYPE MASK) 'BIT)))
      (ERROR "Mask array, ~S, must be a 2-dimensional bit array." MASK))
    (WHEN (NOT (INTEGERP MASK-X))
      (ERROR "Mask X, ~S, must be an integer." MASK-X))
    (WHEN (NOT (INTEGERP MASK-Y))
      (ERROR "Mask Y, ~S, must be an integer." MASK-Y))
    (WHEN (OR (NOT (ARRAYP SOURCE))
	      (NOT (= (ARRAY-RANK SOURCE) 2)))
      (ERROR "Source array, ~S, must be a 2-dimensional array." SOURCE))
    (WHEN (OR (NOT (INTEGERP SOURCE-X))
	      (MINUSP SOURCE-X))
      (ERROR "Source X, ~S, must be a positive integer." SOURCE-X))
    (WHEN (OR (NOT (INTEGERP SOURCE-Y))
	      (MINUSP SOURCE-Y))
      (ERROR "Source Y, ~S, must be a positive integer." SOURCE-Y))
    (WHEN (OR (NOT (ARRAYP DEST))
	      (NOT (= (ARRAY-RANK DEST) 2)))
      (ERROR "Destination array, ~S, must be a 2-dimensional array." DEST))
    (WHEN (OR (NOT (EQ (ARRAY-ELEMENT-TYPE SOURCE) (ARRAY-ELEMENT-TYPE DEST)))
	      (NOT (SUBTYPEP (ARRAY-ELEMENT-TYPE SOURCE) 'INTEGER)))
      (ERROR "Elements of the source array, ~S, and the destination array, ~S,~
            ~%must be of the same integer type." SOURCE DEST))
    (WHEN (OR (NOT (INTEGERP DEST-X))
	      (MINUSP DEST-X))
      (ERROR "Destination X, ~S, must be a positive integer." DEST-X))
    (WHEN (OR (NOT (INTEGERP DEST-Y))
	      (MINUSP DEST-Y))
      (ERROR "Destination Y, ~S, must be a positive integer." DEST-Y)))

  ;; Now for some semantic checking and initial setup.
  (WHEN (AND (NOT (ZEROP WIDTH)) (NOT (ZEROP HEIGHT)))
    (LET ((REAL-WIDTH  (ABS WIDTH))
          (REAL-HEIGHT (ABS HEIGHT))
          (FORWARD-X-DIRECTION (PLUSP WIDTH))
          (FORWARD-Y-DIRECTION (PLUSP HEIGHT))
          (MASK-WIDTH  (ARRAY-DIMENSION MASK 1))
          (MASK-HEIGHT (ARRAY-DIMENSION MASK 0))
          (DEST-WIDTH    (ARRAY-DIMENSION DEST   1))
          (DEST-HEIGHT   (ARRAY-DIMENSION DEST   0))
	  DEST-SAVE-bitmap)
      ;; To make this close to the functionality of BITBLT, we don't check the size
      ;; of the source rectangle against the source dimensions.  If the source array
      ;; is too small when compared with the source rectangle then it wraps around.
      ;; This allows one to create stipple patterns.
      (WHEN CHECK-ARGS
	(WHEN (> (+ REAL-WIDTH DEST-X) DEST-WIDTH)
	  (ERROR "Destination array width of ~D is too small for destination rectangle of width ~D."
		 DEST-WIDTH (+ REAL-WIDTH DEST-X)))
	(WHEN (> (+ REAL-HEIGHT DEST-Y) DEST-HEIGHT)
	  (ERROR "Destination array height of ~D is too small for destination rectangle of height ~D."
		 DEST-HEIGHT (+ REAL-HEIGHT DEST-Y))))
      ;; Adjust the copy rectangle to fit within the mask array.
      (WHEN (> REAL-WIDTH (+ MASK-X MASK-WIDTH))
        ;; The width being copied is beyond the mask, which means that it won't
        ;; really get copied.
        (SETQ REAL-WIDTH (+ MASK-X MASK-WIDTH)))
      (WHEN (> REAL-HEIGHT (+ MASK-Y MASK-HEIGHT))
        ;; Part of the height being copied is below the mask, which means that
        ;; it won't really get copied.
        (SETQ REAL-HEIGHT (+ MASK-Y MASK-HEIGHT)))
      (WHEN (MINUSP MASK-X)
        ;; We are starting the clipping mask within the copy rectangle.  Adjust the
        ;; destination X and width appropriately.
        (INCF REAL-WIDTH MASK-X)
        (IF FORWARD-X-DIRECTION
            (INCF WIDTH MASK-X)
          (DECF WIDTH MASK-X))
        (SETQ MASK-X 0))
      (WHEN (MINUSP MASK-Y)
        ;; We are starting the clipping mask within the copy rectangle.  Adjust the
        ;; destination Y and height appropriately.
        (INCF REAL-HEIGHT MASK-Y)
        (IF FORWARD-Y-DIRECTION
            (INCF HEIGHT MASK-Y)
          (DECF HEIGHT MASK-Y))
        (SETQ MASK-Y 0))
      (USING-RESOURCE (DEST-SAVE-pixmap PIXMAP nil real-width real-height
					(ARRAY-ELEMENT-SIZE DEST) nil)
	(setq DEST-SAVE-bitmap (pixmap.array DEST-SAVE-pixmap))
	(IF (EQ OCCLUSION-STACK T)
	      (PROGN
		;; The idea is to only copy those pixels in the source which are on.  This means
		;; that we need to save away a copy of the destination so that we can prepare
		;; it for combination with the source ahead of time.

		(BITBLT W:ALU-SETA WIDTH HEIGHT DEST   DEST-X   DEST-Y   DEST-SAVE-BITMAP 0 0)
		;; dest-save = copy-of-dest

		(BITBLT W:ALU-ANDCA WIDTH HEIGHT MASK MASK-X MASK-Y DEST DEST-X DEST-Y)
		;; dest = zeros-where-ones-in-mask

		;; Combine the source pixels with the destination pixels and store in the temporary.
		(BITBLT (AREF X-ALU-To-Explorer-ALU X-ALU) WIDTH HEIGHT SOURCE SOURCE-X SOURCE-Y
			DEST-SAVE-BITMAP 0 0)
		;; dest-save = source-op-dest

		;; Only leave those pixels on which are part of the mask.
		(BITBLT W:ALU-AND WIDTH HEIGHT MASK MASK-X MASK-Y DEST-SAVE-BITMAP 0 0)
		;; dest-save = source-op-dest-in-mask-zeros-elswere

		;; Merge the new pixels into the destination.
		(BITBLT W:ALU-IOR WIDTH HEIGHT DEST-SAVE-BITMAP 0 0 DEST DEST-X DEST-Y)
		;; dest = source-op-dest-in-mask-zeros-elswere .or. zero-where-ones-in-mask
		)

	    ;;1ELSE*
	    ;; The occlusion stack indicates that only part of the window is to be written
	    ;; to the destination.  We will deal with each box individually.
	    (LET (X Y BOX-WIDTH BOX-HEIGHT BOX-X-OFFSET BOX-Y-OFFSET)
	      (DOLIST (BOX OCCLUSION-STACK)
		(SETQ X (BOX.LEFT BOX)
		      Y (BOX.TOP  BOX)
		      BOX-WIDTH  (BOX.WIDTH  BOX)
		      BOX-HEIGHT (BOX.HEIGHT BOX))
		;; Since the occlusion stack might be outside of the destination, make
		;; sure that it is inside.  This may happen since the occlusion stack
		;; relates to the entire window, including the borders, while graphics
		;; drawn on the destination only relate to the interior part of the
		;; window.
		;; Adjust when the occlusion stack includes the left border.
		(WHEN (< X DEST-X)
		  (DECF BOX-WIDTH (- DEST-X X))
		  (SETQ X DEST-X))
		;; Adjust when the occlusion stack includes the top border.
		(WHEN (< Y DEST-Y)
		  (DECF BOX-HEIGHT (- DEST-Y Y))
		  (SETQ Y DEST-Y))
		(WHEN (> (+ X BOX-WIDTH) (+ DEST-X WIDTH))
		  ;; This point is outside of the destination completely.
		  (SETQ BOX-WIDTH (- (+ DEST-X WIDTH) X)))
		(WHEN (> (+ Y BOX-HEIGHT) (+ DEST-Y HEIGHT))
		  ;; This point is outside of the destination completely.
		  (SETQ BOX-HEIGHT (- (+ DEST-Y HEIGHT) Y)))
		(WHEN (AND (PLUSP BOX-WIDTH) (PLUSP BOX-HEIGHT))
		  (SETQ BOX-X-OFFSET (- X DEST-X)
			BOX-Y-OFFSET (- Y DEST-Y))
		;; The idea is to only copy those pixels in the source which are on.  This means
		;; that we need to save a way a copy of the destination so that we can prepare
		;; it for combination with the source ahead of time.
		(BITBLT W:ALU-SETA BOX-WIDTH BOX-HEIGHT
			DEST (+ DEST-X BOX-X-OFFSET) (+ DEST-Y BOX-Y-OFFSET)
			DEST-SAVE-BITMAP 0 0)

		;; Inverse-And the mask with the destination to form an image (zeros)
		;; of those pixles which we will replace.
		(BITBLT W:ALU-ANDCA BOX-WIDTH BOX-HEIGHT
			MASK (+ MASK-X BOX-X-OFFSET) (+ MASK-Y BOX-Y-OFFSET)
			DEST (+ DEST-X BOX-X-OFFSET) (+ DEST-Y BOX-Y-OFFSET))
		
		;; Combine the source pixels with the destination pixels and store in the temporary.
		(BITBLT (AREF X-ALU-To-Explorer-ALU X-ALU) BOX-WIDTH BOX-HEIGHT
			SOURCE SOURCE-X SOURCE-Y
			DEST-SAVE-BITMAP 0 0)
		;; Only leave those pixels on which are part of the mask.
		(BITBLT W:ALU-AND BOX-WIDTH BOX-HEIGHT
			MASK (+ MASK-X BOX-X-OFFSET) (+ MASK-Y BOX-Y-OFFSET)
			DEST-SAVE-BITMAP 0 0)
		;; Merge the new pixels into the destination.
		(BITBLT W:ALU-IOR BOX-WIDTH BOX-HEIGHT
			DEST-SAVE-BITMAP 0 0
			DEST (+ DEST-X BOX-X-OFFSET) (+ DEST-Y BOX-Y-OFFSET))))))))))
;;; Test cases:
#|
(DEFPARAMETER WINDOW-ARRAY NIL)
(DEFPARAMETER WINDOW-WIDTH 500)
(DEFPARAMETER WINDOW-HEIGHT 600)

(DEFPARAMETER SOURCE-ARRAY NIL)

(DEFPARAMETER CLIP-MASK NIL)
(DEFPARAMETER OCCLUSION-STACK NIL)

(DEFUN BITBLT-TEST-SETUP ()
  (WHEN (NULL CLIP-MASK)
    (SETQ CLIP-MASK (TV:MAKE-SHEET-BIT-ARRAY TV:DEFAULT-SCREEN WINDOW-WIDTH WINDOW-HEIGHT)))
  ;; Initialize the clip mask to a gray pattern.
  (BITBLT W:ALU-SETA WINDOW-WIDTH WINDOW-HEIGHT W:50%-GRAY 0 0 CLIP-MASK 0 0)

  (WHEN (NULL WINDOW-ARRAY)
    (SETQ WINDOW-ARRAY (TV:MAKE-SHEET-BIT-ARRAY TV:DEFAULT-SCREEN WINDOW-WIDTH WINDOW-HEIGHT)))
  ;; Initialize the window array with a Zmacs window.
  (BITBLT W:ALU-SETA WINDOW-WIDTH WINDOW-HEIGHT 
          (W:SHEET-SCREEN-ARRAY (TV:FIND-WINDOW-OF-FLAVOR 'ZWEI:ZMACS-FRAME)) 0 0
          WINDOW-ARRAY 0 0)

  (WHEN (NULL SOURCE-ARRAY)
    (SETQ SOURCE-ARRAY (TV:MAKE-SHEET-BIT-ARRAY TV:DEFAULT-SCREEN WINDOW-WIDTH WINDOW-HEIGHT)))
  (BITBLT W:ALU-SETZ WINDOW-WIDTH WINDOW-HEIGHT SOURCE-ARRAY 0 0 SOURCE-ARRAY 0 0)
  (W:%DRAW-SHADED-TRIANGLE 100 100 450 200 200 450 TV:ALU-SETA T T T NIL SOURCE-ARRAY)

  (SETQ OCCLUSION-STACK NIL)
  (LET (OLD-X OLD-Y
        (WINDOW TV:SELECTED-WINDOW))
    (SETF (VALUES OLD-X OLD-Y) (SEND WINDOW :READ-CURSORPOS))
    (SEND WINDOW :SET-CURSORPOS 0 0)
    (SEND WINDOW :CLEAR-EOF)
    (SEND WINDOW :SET-CURSORPOS OLD-X OLD-Y)))

(DEFUN SHOW-BITBLT-RESULTS (ARRAY DOC-STRING)
  (LET* ((WINDOW TV:SELECTED-WINDOW)
         (OLD-LABEL (SEND WINDOW :LABEL)))
    (UNWIND-PROTECT
        (PROGN
          (SEND WINDOW :SET-LABEL (CONCATENATE 'SIMPLE-STRING DOC-STRING
                                               ".  Press any key to continue."))
          (W:SHOW-BIT-ARRAY ARRAY))
      (SEND WINDOW :SET-LABEL OLD-LABEL))))

(DEFUN BITBLT-NO-OCCLUSION ()
  (LOOP FOR ALU IN '(GX-AND GX-XOR GX-CLEAR GX-COPY GX-OR GX-AND-INVERTED)
        DO (PROGN
             (BITBLT-TEST-SETUP)
             (BITBLT-UNDER-MASK (SYMBOL-VALUE ALU) WINDOW-WIDTH WINDOW-HEIGHT CLIP-MASK 0 0
                                SOURCE-ARRAY 0 0
                                WINDOW-ARRAY 0 0
                                T)
             (SHOW-BITBLT-RESULTS WINDOW-ARRAY (FORMAT NIL "BITBLT no occlusion, alu=~A"
                                                       (SYMBOL-NAME ALU))))))

(DEFUN BITBLT-TOP-OCCLUSION ()
  (BITBLT-TEST-SETUP)
  (SETQ OCCLUSION-STACK (LIST (MAKE-BOX :LEFT 0 :TOP 0 :WIDTH WINDOW-WIDTH :HEIGHT 200)))
  (BITBLT-UNDER-MASK GX-COPY WINDOW-WIDTH WINDOW-HEIGHT CLIP-MASK 0 0
                     SOURCE-ARRAY 0 0
                     WINDOW-ARRAY 0 0
                     OCCLUSION-STACK)
  (SHOW-BITBLT-RESULTS WINDOW-ARRAY "BITBLT top occluded"))

|#


