;;; -*- Mode:Common-Lisp; Package:X11; Fonts:(MEDFNB HL12B HL12BI); Base:10 -*-

;;;			      RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;			TEXAS INSTRUMENTS INCORPORATED.
;;;				 P.O. BOX 2909
;;;			      AUSTIN, TEXAS 78769
;;;				    MS 2151
;;;
;;; Copyright (C) 1988 Texas Instruments Incorporated. All rights reserved.

#|

This file contains system-dependent mouse code.  It shuts off
the Explorer mouse handling when X is alive and vice-versa.

|#

;;;
;;; Change history:
;;;
;;;  Date      Author	Description
;;; -------------------------------------------------------------------------------------
;;; 04/04/89	WJB	Patch 1.69; Change MOUSE-OVERSEER to observe SHEET-INVSIBLE-TO-MOUSE-P.
;;; 12/06/88	DLS	Changed mouse process to be restartable so it can be reset and so
;;;			it will stay around when saved in a band.
;;; 12/05/88	WJB	Added MOUSE-WAKEUP function.
;;; 10/119*/88	1LGO*	1Don't jerk the mouse to the top-left corner the first time its moved.*
;;; 10/14/88	WJB	Replaced TV:MOUSE-INPUT with 4.1.1+patches version.
;;;  5/06/88    TWE	Modified MOUSE-OVERSEER to set the mouse process variable to NIL
;;;			if the user kills it off.
;;;  3/29/88    TWE	Changed MOUSE-PROCESS to a defvar so that it wouldn't be set
;;;			to NIL when the system is remade.
;;;  3/09/88    TWE	Fixed up the initialization of the mouse process.
;;;  2/29/88    TWE	Moved data structures to DEVICE-STRUCTURES and dispatcher code to
;;;			EVENT-DISPATCHER.
;;;  2/10/88    TWE	Initial creation.


(DEFPARAMETER MOUSE-DEVICE (MAKE-DEVICE :EVENTS MOUSE-EVENT-QUEUE))

(DEFUN INIT-KEY-BUTTONS-MODIFIER-MASKS (ARRAY)
  (LOOP FOR MASK IN (LIST BUTTON1-MASK
                          BUTTON2-MASK
                          BUTTON3-MASK
                          BUTTON4-MASK
                          BUTTON5-MASK)
        FOR INDEX FROM 1 BY 1
        DO (SETF (AREF ARRAY INDEX) MASK)))

(DEFPARAMETER DEFAULT-MOUSE-MODIFIERS (LET ((ARRAY (MAKE-KEY-MODIFIERS-LIST)))
                                        (INIT-KEY-BUTTONS-MODIFIER-MASKS ARRAY)
                                        ARRAY)
  "Default Modifier array for mouse.")

;;; New version compatible with both 4.1.1 and 4.1.1+patches.
;;; Old version has been saved in BONE-YARD.

TV:(LET ((INHIBIT-FDEFINE-WARNINGS :JUST-WARN))
     
     (DEFVAR *unglitched-mouse-x-speed* 0)
     (DEFVAR *unglitched-mouse-y-speed* 0)
     (DEFVAR *previous-unglitched-mouse-x* 0)
     (DEFVAR *previous-unglitched-mouse-y* 0)
     (DEFVAR *unglitched-mouse-x* 0)
     (DEFVAR *unglitched-mouse-y* 0)
     
     (DEFUN MOUSE-INPUT (&OPTIONAL (WAIT-FLAG T))
       "Wait until mouse moves or a button goes up or down, then report what happened.
The values are the delta-x, the delta-y (amounts of mouse motion),
buttons-newly-pushed, buttons-newly-raised, and the relevant mouse X and Y (the
current position normally, but if any buttons changed, the position then)."
       ;; Await a change in hardware status from what it was last time
       (COND (WAIT-FLAG
	      ;; This sleep makes it reasonable for the mouse process to be high priority.
	      ;; It insures that moving the mouse does not lock out lower priority
	      ;; processes for a long time.  This constant may want to be tweaked.  A
	      ;; value of 1 (1/60 of a second) does not noticably affect mouse response.
	      (PROCESS-SLEEP 1)
	      (PROCESS-WAIT "MOUSE" #'(LAMBDA ()                                              
					(IF (mac-system-p)
					    (lispm-or-mac-mouse-wakeup)
					    ;; <<new>> Added the (AND (NULL X-ENABLED) component for X.
					    (AND (NULL X11:X-ENABLED)
						 (OR MOUSE-WAKEUP MOUSE-RECONSIDER)))))))
       ;; Set some global variables used for mouse transport.  CJJ 05/06/88.
       ;;; Added by KJF for CJJ on 08/16/88 for Multiple Monitor (MMON) support.
       (SETF *unglitched-mouse-x-speed* mouse-x-speed
	     *unglitched-mouse-y-speed* mouse-y-speed
	     ;; Remember previous mouse position for mouse transport along mouse trajectory.  CJJ 08/15/88.
	     *previous-unglitched-mouse-x* *unglitched-mouse-x*
	     *previous-unglitched-mouse-y* *unglitched-mouse-y*
	     *unglitched-mouse-x* mouse-x
	     *unglitched-mouse-y* mouse-y)
       ;; Clear wakeup flag unless there are buffered mouse button transitions, since we
       ;; might not read all of them before calling MOUSE-INPUT again.
       (SETQ MOUSE-WAKEUP (NOT (= MOUSE-BUTTONS-BUFFER-IN-INDEX MOUSE-BUTTONS-BUFFER-OUT-INDEX)))
       ;; Compute delta-X and delta-Y in screen coordinates
       (LET ((DELTA-X (- MOUSE-X MOUSE-LAST-X))
	     (DELTA-Y (- MOUSE-Y MOUSE-LAST-Y))
	     (GLITCH-X NIL) (GLITCH-Y NIL) NEW-BUTTONS CHANGED-BUTTONS)
	 (INCF MOUSE-LAST-X DELTA-X)
	 (INCF MOUSE-LAST-Y DELTA-Y)
	 ;; Compute change in button status
	 (MULTIPLE-VALUE-SETQ (NEW-BUTTONS MOUSE-LAST-BUTTONS-TIME
			       MOUSE-LAST-BUTTONS-X MOUSE-LAST-BUTTONS-Y)
	   (MOUSE-BUTTONS))
	 (SETQ CHANGED-BUTTONS (LOGXOR NEW-BUTTONS MOUSE-LAST-BUTTONS)
	       MOUSE-LAST-BUTTONS NEW-BUTTONS)
	 ;; Force blinker to stay within mouse-sheet.  If the mouse moves during this
	 ;; computation, it will glitch back.  So we only SETQ the variables
	 ;; if the mouse position actually needs to be changed, rather than using
	 ;; MAX and MIN which would be more readable.
	 (IF (> 0 MOUSE-X)
	     (SETQ GLITCH-X 0))
	 (IF (<= (SHEET-WIDTH MOUSE-SHEET) MOUSE-X)
	     (SETQ GLITCH-X (1- (SHEET-WIDTH MOUSE-SHEET))))
	 (IF (> 0 MOUSE-Y)
	     (SETQ GLITCH-Y 0))
	 (IF (<= (SHEET-HEIGHT MOUSE-SHEET) MOUSE-Y)
	     (SETQ GLITCH-Y (1- (SHEET-HEIGHT MOUSE-SHEET))))
	 ;; If mouse blinker needs to be glitched, do so
	 (IF (OR GLITCH-X GLITCH-Y)
	     (WITHOUT-INTERRUPTS
	      (%OPEN-MOUSE-CURSOR)
	      (IF GLITCH-X
		  (SETQ MOUSE-LAST-X (SETQ MOUSE-X GLITCH-X)))
	      (IF GLITCH-Y
		  (SETQ MOUSE-LAST-Y (SETQ MOUSE-Y GLITCH-Y)))
	      (SETQ MOUSE-CURSOR-STATE MOUSE-CURSOR-CLOSED-STATE
		    PREPARED-SHEET NIL)))
	 (VALUES DELTA-X
		 DELTA-Y
		 (LOGAND NEW-BUTTONS CHANGED-BUTTONS)
		 (BOOLE 2 NEW-BUTTONS CHANGED-BUTTONS) ;BOOLE 2 is ANDCA
		 (IF (ZEROP CHANGED-BUTTONS) MOUSE-LAST-X MOUSE-LAST-BUTTONS-X)
		 (IF (ZEROP CHANGED-BUTTONS) MOUSE-LAST-Y MOUSE-LAST-BUTTONS-Y))))
     )


(DEFUN WHICH-MOUSE-BUTTON (BUTTON-MASK)
  (LET ((DELTA (LDB BUTTON-DELTA-BYTE-SPEC BUTTON-MASK)))
    (COND ((LOGBITP 2 DELTA) :RIGHT)
          ((LOGBITP 1 DELTA) :MIDDLE)
          ((LOGBITP 0 DELTA) :LEFT)
          (T :DELTA-X-Y))))

(DEFUN MOUSE-BUTTON-STATE (BUTTON-MASK)
  (LET ((DELTA (LDB BUTTON-DELTA-BYTE-SPEC BUTTON-MASK)))
    (COND ((LOGBITP 0 DELTA)
           (IF (LOGBITP 0 BUTTON-MASK)
               :BUTTON-DOWN
               ;;ELSE
               :BUTTON-UP))
          ((LOGBITP 1 DELTA)
           (IF (LOGBITP 1 BUTTON-MASK)
               :BUTTON-DOWN
               ;;ELSE
               :BUTTON-UP))
          ((LOGBITP 2 DELTA)
           (IF (LOGBITP 2 BUTTON-MASK)
               :BUTTON-DOWN
               ;;ELSE
               :BUTTON-UP)))))

(DEFUN BUFFER-MOUSE-EVENTS ()
  "Runs in its own process to obtain events from the Explorer mouse queue
and place them in a larger queue for X."
  ;1; This sleep makes it reasonable for the mouse process to be high priority.*
  ;1; It insures that moving the mouse does not lock out lower priority*
  ;1; processes for a long time.  This constant may want to be tweaked.  A*
  ;1; value of 1 (1/60 of a second) does not noticably affect mouse response.*
  (PROCESS-SLEEP 1)
  ;; Note that W:MOUSE-WAKEUP is a microcode register that becomes non-nil
  ;; when either the mouse moves or when the user clicks a mouse button.
  (PROCESS-WAIT "X Mouse" #'(LAMBDA () W:MOUSE-WAKEUP))
  (SETQ W:MOUSE-WAKEUP (NOT (= W:MOUSE-BUTTONS-BUFFER-IN-INDEX
                               W:MOUSE-BUTTONS-BUFFER-OUT-INDEX)))
  (LET ((DELTA-X (- W:MOUSE-X MOUSE-LAST-X))
        (DELTA-Y (- W:MOUSE-Y MOUSE-LAST-Y))
        NEW-BUTTONS
        MOUSE-LAST-BUTTONS-TIME
        GLITCH-X GLITCH-Y
        MOUSE-LAST-BUTTONS-X MOUSE-LAST-BUTTONS-Y
        CHANGED-BUTTONS)
    (MULTIPLE-VALUE-SETQ (NEW-BUTTONS MOUSE-LAST-BUTTONS-TIME
                          MOUSE-LAST-BUTTONS-X MOUSE-LAST-BUTTONS-Y)
      (W:MOUSE-BUTTONS NIL))
    (INCF MOUSE-LAST-X DELTA-X)
    (INCF MOUSE-LAST-Y DELTA-Y)
    (SETQ CHANGED-BUTTONS (LOGXOR NEW-BUTTONS MOUSE-LAST-BUTTONS)
          MOUSE-LAST-BUTTONS NEW-BUTTONS)
    ;1; Force blinker to stay within mouse-sheet.  If the mouse moves during this*
    ;1; computation, it will glitch back.  So we only SETQ the variables*
    ;1; if the mouse position actually needs to be changed, rather than using*
    ;1; MAX and MIN which would be more readable.*
    (IF (> 0 W:MOUSE-X)
        (SETQ GLITCH-X 0))
    (IF (<= (W:SHEET-WIDTH W:MOUSE-SHEET) W:MOUSE-X)
        (SETQ GLITCH-X (1- (W:SHEET-WIDTH W:MOUSE-SHEET))))
    (IF (> 0 W:MOUSE-Y)
        (SETQ GLITCH-Y 0))
    (IF (<= (W:SHEET-HEIGHT W:MOUSE-SHEET) W:MOUSE-Y)
        (SETQ GLITCH-Y (1- (W:SHEET-HEIGHT W:MOUSE-SHEET))))
    ;1; If mouse blinker needs to be glitched, do so*
    (IF (OR GLITCH-X GLITCH-Y)
        (WITHOUT-INTERRUPTS
          (W:%OPEN-MOUSE-CURSOR)
          (IF GLITCH-X
              (SETQ MOUSE-LAST-X (SETQ W:MOUSE-X GLITCH-X)))
          (IF GLITCH-Y
              (SETQ MOUSE-LAST-Y (SETQ W:MOUSE-Y GLITCH-Y)))))
    (IF (LOGAND MOUSE-LEFT-BUTTON-MASK NEW-BUTTONS)
        ;; Only record the time of the first depression of this mouse button.
        (WHEN (NULL MOUSE-LEFT-LAST-DOWN-TIME)
          (SETQ MOUSE-LEFT-LAST-DOWN-TIME (GET-TIME-IN-MILLIS)))
        ;;ELSE
        (SETQ MOUSE-LEFT-LAST-DOWN-TIME NIL))
    (IF (LOGAND MOUSE-MIDDLE-BUTTON-MASK NEW-BUTTONS)
        ;; Only record the time of the first depression of this mouse button.
        (WHEN (NULL MOUSE-MIDDLE-LAST-DOWN-TIME)
          (SETQ MOUSE-MIDDLE-LAST-DOWN-TIME (GET-TIME-IN-MILLIS)))
        ;;ELSE
        (SETQ MOUSE-MIDDLE-LAST-DOWN-TIME NIL))
    (IF (LOGAND MOUSE-RIGHT-BUTTON-MASK NEW-BUTTONS)
        ;; Only record the time of the first depression of this mouse button.
        (WHEN (NULL MOUSE-RIGHT-LAST-DOWN-TIME)
          (SETQ MOUSE-RIGHT-LAST-DOWN-TIME (GET-TIME-IN-MILLIS)))
        ;;ELSE
        (SETQ MOUSE-RIGHT-LAST-DOWN-TIME NIL))
    (IF (AND (= NEW-BUTTONS (LOGIOR MOUSE-RIGHT-BUTTON-MASK
                                    MOUSE-MIDDLE-BUTTON-MASK
                                    MOUSE-LEFT-BUTTON-MASK))
             (> MOUSE-BUTTONS-ALL-DOWN-TIME-WINDOW
                (MAX (- MOUSE-LAST-BUTTONS-TIME MOUSE-LEFT-LAST-DOWN-TIME)
                     (- MOUSE-LAST-BUTTONS-TIME MOUSE-MIDDLE-LAST-DOWN-TIME)
                     (- MOUSE-LAST-BUTTONS-TIME MOUSE-RIGHT-LAST-DOWN-TIME))))
        ;; Pressing all mouse buttons down within a specified time window means
        ;; that the user (most likely an X hacker) is trying to disable X and
        ;; reenable the Explorer window system.
	(X-SHUTDOWN)
      ;;ELSE
      (WRITE-EVENTS MOUSE-EVENT-QUEUE 
		    ;1; Merge changed buttons with current button state so we have both in the same number.*
		    (DPB CHANGED-BUTTONS BUTTON-DELTA-BYTE-SPEC NEW-BUTTONS)
		    MOUSE-LAST-BUTTONS-TIME
		    MOUSE-LAST-X MOUSE-LAST-Y)
;;;        (FORMAT TV:INITIAL-LISP-LISTENER "~D ~D, Buttons=~2R, at (~D,~D)"
;;;                W:MOUSE-BUTTONS-BUFFER-IN-INDEX
;;;                W:MOUSE-BUTTONS-BUFFER-OUT-INDEX
;;;                NEW-BUTTONS MOUSE-LAST-X MOUSE-LAST-Y)
      )))

(defun MOUSE-WAKEUP ()
  ;; Wakeup the X mouse process to re-examine mouse and window state
  (w:mouse-wakeup))

(DEFVAR MOUSE-PROCESS NIL)

(DEFUN MOUSE-OVERSEER ()
  ;1; initialize mouse position*
  (LET ((MOUSE (LOOKUP-POINTER-DEVICE))
	(WINDOW (SCREEN.DEV-PRIVATE (CAR (SCREEN-INFO.SCREENS SCREEN-INFO)))))
    (SEND WINDOW :SET-X-MOUSE-X (DEVICE.X MOUSE))
    (SEND WINDOW :SET-X-MOUSE-Y (DEVICE.Y MOUSE)))
  (SETQ MOUSE-PROCESS CURRENT-PROCESS)
  ;1; loop buffering mouse input*
  (UNWIND-PROTECT
      (LOOP
        (PROCESS-WAIT "Waiting for X" #'(LAMBDA () X-ENABLED))
        (LOOP
          (PROCESS-WAIT "Waiting for X mouse" #'(LAMBDA ()
                                                  (OR W:MOUSE-WAKEUP W:MOUSE-RECONSIDER
                                                      (NULL X-ENABLED))))
          (WHEN (NULL X-ENABLED)
            ;; Exit from the inner loop when X has been shut down.
            (RETURN NIL))

	  (WHEN W:MOUSE-RECONSIDER
	    ;; Check if the Explorer window system has made our window "invisible"
	    ;; This is indicative of a notification pending.
	    (LET ((WINDOW (AND CURRENT-SCREEN (SCREEN.DEV-PRIVATE CURRENT-SCREEN))))
	      (WHEN (AND WINDOW
			 (TYPEP WINDOW 'TV:SHEET)
			 (TV:SHEET-INVISIBLE-TO-MOUSE-P WINDOW))
		;; Unhook ourselves from the window system and exit.
		(X-DISABLE)
		(RETURN NIL)))
	    (SETQ W:MOUSE-RECONSIDER NIL))

          (BUFFER-MOUSE-EVENTS)))

    ;; If the X mouse process is killed off we set this to nil so that when
    ;; we initialize again it will be created again.
    (SETQ MOUSE-PROCESS NIL)))
 
(DEFUN INITIALIZE-MOUSE-PROCESS ()
  (if MOUSE-PROCESS
      (send mouse-process :reset)
      (PROCESS-RUN-RESTARTABLE-FUNCTION '(:NAME "X Mouse" :PRIORITY 30.)
					'MOUSE-OVERSEER)))