;;; Macros for ZWEI.   -*- Mode:Common-Lisp; Package:ZWEI; Base:8 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1980, Massachusetts Institute of Technology

;;; CHANGE HISTORY:
;;; 05/03/89    JLM     Added indentation for LETF and LETF*
;;; 04/13/89 	MAY	Changed WITH-BORROWED-FONT-MAP-OF to quiet compiler TYPEP warnings.
;;; 04/03/89  	MAY	Added (for DNG) DEFINE-INDENTATION forms for CLOS - with ticlos pkg.
;;; 01/23/89	MAY     Changed CHARMAP-PER-LINE to check for nil line. Both FORWARD-SEXP and
;;;                     CHARMAP-PER-LINE are very tightly coupled. The problem is that
;;;                     the per-line-forms of CHARMAP-PER-LINE can cause the line to 
;;;			become nil without reaching *LAST-LINE*. Although the real problem
;;;                     is in FORWARD-SEXP - who sets *THIS-IS-THE-LAST-LINE* in addition
;;;			to CHARMAP-PER-LINE - both places were changed since they are so
;;;                     tightly coupled. Fixes SPR's 6402 6158 8890.

(DEFMACRO CHARMAP ((FROM-BP-FORM TO-BP-FORM . RETURN-FORMS) &BODY BODY)
  "Execute BODY for each character from one BP to another.
Like CHARMAP-PER-LINE except that there are no LINE-FORMS.
See the documentation of CHARMAP-PER-LINE."
  `(CHARMAP-PER-LINE (,FROM-BP-FORM ,TO-BP-FORM ,@RETURN-FORMS) (NIL) ,@BODY)) 

(DEFMACRO CHARMAP-PER-LINE ((FROM-BP-FORM TO-BP-FORM . RETURN-FORMS) LINE-FORMS &BODY BODY)
  "Execute BODY for each character from one BP to another.
FROM-BP-FORM and TO-BP-FORM compute the places to start and end the loop.
On reaching the to-bp, RETURN-FORMS are evaluated and the last one's values returned.
LINE-FORMS are evaluated at the beginning of each line.

Within the BODY and LINE-FORMS, you can use
    (CHARMAP-CHAR) to get the next character (the one after the point the loop has reached),
and (CHARMAP-SET-CHAR char) to modify it.
    (CHARMAP-LINE) returns the current line.
    (CHARMAP-BP-BEFORE) returns a BP before that character
and (CHARMAP-BP-AFTER) returns a BP to after it.
    (CHARMAP-INCREMENT) skips one extra character so that the BODY will not be executed for it.
    (CHARMAP-RETURN forms) returns the values of the forms from the CHARMAP."
  
  `(LET ((*FROM-BP* ,FROM-BP-FORM)
	 (*TO-BP* ,TO-BP-FORM))
     (BLOCK *CHARMAP*
       (DO ((LINE (BP-LINE *FROM-BP*) (LINE-NEXT LINE))
	    (*FIRST-INDEX* (BP-INDEX *FROM-BP*) 0)
	    (*LAST-LINE* (BP-LINE *TO-BP*))
	    (*THIS-IS-THE-LAST-LINE*))
	   (NIL)
;;	 (SETQ *THIS-IS-THE-LAST-LINE* (EQ LINE *LAST-LINE*)) 
	 (SETQ *THIS-IS-THE-LAST-LINE* (OR (NULL line) (EQ LINE *LAST-LINE*)) ) ;; may 01/23/89 
	 ,@LINE-FORMS
	 (DO ((INDEX *FIRST-INDEX* (1+ INDEX))
	      (*LAST-INDEX* (IF *THIS-IS-THE-LAST-LINE*
				(1- (BP-INDEX *TO-BP*))
				(LINE-LENGTH LINE))))
	     ((> INDEX *LAST-INDEX*)
	      (IF *THIS-IS-THE-LAST-LINE*
		  (RETURN-FROM *CHARMAP* ,@RETURN-FORMS)))
	   ,@BODY))))) 

(DEFMACRO CHARMAP-RETURN RETURN-FORMS
  "Within the body of a CHARMAP, return from it.
The values returned are the values of the RETURN-FORMS."
  `(RETURN-FROM *CHARMAP* ,@RETURN-FORMS)) 

(DEFMACRO CHARMAP-CHAR ()
  "Within the body of a CHARMAP, the character being scanned."
  '(COND ((AND (= INDEX *LAST-INDEX*)
	       (NOT *THIS-IS-THE-LAST-LINE*))
	  #\NEWLINE)
	 (T
	  (AREF LINE INDEX)))) 

(DEFMACRO CHARMAP-CH-CHAR ()
  "Within the body of a CHARMAP, the character being scanned, ignoring bits
and font attributes."
  '(MAKE-CHAR (CHARMAP-CHAR))) 

(DEFMACRO CHARMAP-SET-CHAR (CHAR)
  "Within the body of a CHARMAP, set the character being scanned at to CHAR."
  `(PROGN
     (MUNG-BP-INTERVAL *FROM-BP*)
     (MUNG-LINE LINE)
     (COND ((AND (= INDEX *LAST-INDEX*)
		 (NOT *THIS-IS-THE-LAST-LINE*))
	    (INSERT (DELETE-INTERVAL (END-LINE LINE) (BEG-LINE LINE 1) T) ,CHAR))
	   (T
	    (SETF (AREF LINE INDEX) ,CHAR))))) 

(DEFMACRO CHARMAP-BP-BEFORE ()
  "Within the body of a CHARMAP, return a bp to before the character being scanned."
  '(CREATE-BP LINE INDEX)) 

(DEFMACRO CHARMAP-BP-AFTER ()
  "Within the body of a CHARMAP, return a bp to after the character being scanned."
  '(COND ((AND (= INDEX *LAST-INDEX*)
	       (NOT *THIS-IS-THE-LAST-LINE*))
	  (CREATE-BP (LINE-NEXT LINE) 0))
	 (T
	  (CREATE-BP LINE (1+ INDEX))))) 

(DEFMACRO CHARMAP-INCREMENT RETURN-FORMS
  "Within the body of a CHARMAP, skip over one character.
Doing this changes which character CHARMAP-CHAR, CHARMAP-SET-CHAR,
CHARMAP-BP-BEFORE and CHARMAP-BP-AFTER refer to."
  `(COND ((> (SETQ INDEX (1+ INDEX)) *LAST-INDEX*)
	  (COND (*THIS-IS-THE-LAST-LINE*
		 (RETURN-FROM *CHARMAP* ,@RETURN-FORMS))
		(T
		 (SETQ INDEX 0
		       LINE (LINE-NEXT LINE)
		       *THIS-IS-THE-LAST-LINE* (EQ LINE *LAST-LINE*)
		       *LAST-INDEX* (IF *THIS-IS-THE-LAST-LINE*
					(1- (BP-INDEX *TO-BP*))
					(LINE-LENGTH LINE)))))))) 

(DEFMACRO CHARMAP-LINE ()
  "Within the body of a CHARMAP, the line being scanned."
  'LINE) 

(DEFMACRO RCHARMAP ((FROM-BP-FORM TO-BP-FORM . RETURN-FORMS) &BODY BODY)
  "Like CHARMAP but scans backwards through the interval.
The to-bp should be earlier in the interval than the from-bp.
Instead of using CHARMAP-CHAR, CHARMAP-BP-BEFORE, etc.,
you use RCHARMAP-CHAR, RCHARMAP-BP-BEFORE, etc."
  `(RCHARMAP-PER-LINE (,FROM-BP-FORM ,TO-BP-FORM ,@RETURN-FORMS) (NIL) ,@BODY)) 

(DEFMACRO RCHARMAP-PER-LINE ((FROM-BP-FORM TO-BP-FORM . RETURN-FORMS) LINE-FORMS &BODY BODY)
  "Like CHARMAP-PER-LINE but scans backwards through the interval.
The to-bp should be earlier in the interval than the from-bp.
Instead of using CHARMAP-CHAR, CHARMAP-BP-BEFORE, etc.,
you use RCHARMAP-CHAR, RCHARMAP-BP-BEFORE, etc."
  `(LET ((*FROM-BP* ,FROM-BP-FORM)
	 (*TO-BP* ,TO-BP-FORM))
     (BLOCK *RCHARMAP*
       (DO ((LINE (BP-LINE *FROM-BP*) (LINE-PREVIOUS LINE))
	    (*FIRST-LINE-P* T NIL)
	    (*LAST-LINE* (BP-LINE *TO-BP*))
	    (*THIS-IS-THE-LAST-LINE*))
	   (NIL)
	 (SETQ *THIS-IS-THE-LAST-LINE* (EQ LINE *LAST-LINE*))
	 ,@LINE-FORMS
	 (DO ((INDEX (IF *FIRST-LINE-P*
			 (1- (BP-INDEX *FROM-BP*))
			 (LINE-LENGTH LINE))
		     (1- INDEX))
	      (*LAST-INDEX* (IF *THIS-IS-THE-LAST-LINE*
				(BP-INDEX *TO-BP*)
				0)))
	     ((< INDEX *LAST-INDEX*)
	      (IF *THIS-IS-THE-LAST-LINE*
		  (RETURN-FROM *RCHARMAP* ,@RETURN-FORMS)))
	   ,@BODY))))) 

(DEFMACRO RCHARMAP-RETURN RETURN-FORMS
  "Within the body of a RCHARMAP, return from it.
The values returned are the values of the RETURN-FORMS."
  `(RETURN-FROM *RCHARMAP* ,@RETURN-FORMS)) 

(DEFMACRO RCHARMAP-CHAR ()
  "Within the body of a RCHARMAP, the character at the scanning point."
  '(COND ((= INDEX (LINE-LENGTH LINE))
	  #\NEWLINE)
	 (T
	  (AREF LINE INDEX)))) 

(DEFMACRO RCHARMAP-CH-CHAR ()
  "Within the body of a RCHARMAP, the character at the scanning point, ignoring bits
and font attributes."
  '(MAKE-CHAR (RCHARMAP-CHAR))) 

(DEFMACRO RCHARMAP-CHAR-BEFORE ()
  "Within the body of a RCHARMAP, the character before the scanning point."
  '(IF (ZEROP INDEX)
       #\NEWLINE
       (AREF LINE (1- INDEX)))) 

(DEFMACRO RCHARMAP-CH-CHAR-BEFORE ()
  "Within the body of a RCHARMAP, the character before the scanning point, ignoring
bits and font attributes."
  '(MAKE-CHAR (RCHARMAP-CHAR-BEFORE))) 

(DEFMACRO RCHARMAP-SET-CHAR (CHAR)
  "Within the body of a RCHARMAP, set the character at the scanning point to CHAR."
  `(PROGN
     (MUNG-BP-INTERVAL *FROM-BP*)
     (MUNG-LINE LINE)
     (COND ((= INDEX (LINE-LENGTH LINE))
	    (INSERT (DELETE-INTERVAL (END-LINE LINE) (BEG-LINE LINE 1) T) ,CHAR))
	   (T
	    (SETF (AREF LINE INDEX) ,CHAR))))) 

(DEFMACRO RCHARMAP-BP-BEFORE ()
  "Within the body of a RCHARMAP, return a bp to before the character at the scanning point."
  '(CREATE-BP LINE INDEX)) 

(DEFMACRO RCHARMAP-BP-AFTER ()
  "Within the body of a RCHARMAP, return a bp to after the char at the scanning point."
  '(COND ((= INDEX (LINE-LENGTH LINE))
	  (CREATE-BP (LINE-NEXT LINE) 0))
	 (T
	  (CREATE-BP LINE (1+ INDEX))))) 

(DEFMACRO RCHARMAP-DECREMENT RETURN-FORMS
  "Within the body of a RCHARMAP, skip backward over one character.
Doing this changes which character RCHARMAP-CHAR, RCHARMAP-SET-CHAR,
RCHARMAP-BP-BEFORE and RCHARMAP-BP-AFTER refer to."
  `(COND ((< (SETQ INDEX (1- INDEX)) *LAST-INDEX*)
	  (COND (*THIS-IS-THE-LAST-LINE*
		 (RETURN-FROM *RCHARMAP* ,@RETURN-FORMS))
		(T
		 (SETQ LINE (LINE-PREVIOUS LINE)
		       *FIRST-LINE-P* NIL
		       INDEX (LINE-LENGTH LINE)
		       *THIS-IS-THE-LAST-LINE* (EQ LINE *LAST-LINE*)
		       *LAST-INDEX* (IF *THIS-IS-THE-LAST-LINE*
					(BP-INDEX *TO-BP*)
					0))))))) 

(DEFMACRO RCHARMAP-LINE ()
  "Within the body of a RCHARMAP, the line being scanned."
  'LINE) 


(DEFMACRO PUSH* (I R)
  "Push the value of I onto the list in R unless I is already a member of that list."
  `(OR (MEMBER ,I ,R :TEST #'EQ) (PUSH ,I ,R))) 

;; PUSH+ defined by patch 94.178, similar to PUSH*, but is better for methods.
;; ddd/gsl 3/4/84.
(DEFMACRO PUSH+ (I R) ;created for merge-and-eliminate-duplicates. gsl 11/27/83
  "Push the value of I onto the list in R unless I is already a member of that list."
  `(OR (MEMBER ,I ,R :TEST #'EQUAL) (PUSH ,I ,R))) 

(DEFMACRO ORDER-BPS (BP1 BP2)
  "Given two variables BP1 and BP2, ensure that BP1 is before BP2.
The values of both variables are assumed to be buffer pointers."
  `(COND ((BP-< ,BP2 ,BP1)
	  (PSETQ ,BP1 ,BP2 ,BP2 ,BP1)))) 

(DEFMACRO WORD-SYNTAX (CHAR)
  "Return the word syntax setting for character CHAR."
  `(CHAR-SYNTAX ,CHAR *MODE-WORD-SYNTAX-TABLE*)) 

(DEFMACRO ATOM-WORD-SYNTAX (CHAR)
  "Return the word syntax setting for character CHAR for Atom Word Mode."
  `(CHAR-SYNTAX ,CHAR *ATOM-WORD-SYNTAX-TABLE*)) 

(DEFMACRO LIST-SYNTAX (CHAR)
  "Return the list syntax setting for character CHAR."
  `(CHAR-SYNTAX ,CHAR *MODE-LIST-SYNTAX-TABLE*)) 

(DEFMACRO LIST-SYNTAX-OPEN-CODED (CHAR)
  "Return the list syntax setting for character CHAR."
  `(CHAR-SYNTAX-OPEN-CODED ,CHAR *MODE-LIST-SYNTAX-TABLE*)) 

(DEFSUBST CHAR-SYNTAX-OPEN-CODED (CHAR SYNTAX-TABLE)
  "Like CHAR-SYNTAX but faster."
  (IF (ARRAYP SYNTAX-TABLE)
      (AREF SYNTAX-TABLE (CHAR-CODE CHAR))
      (CHAR-SYNTAX CHAR SYNTAX-TABLE))) 

(DEFMACRO ATOM-WORD-SYNTAX-BIND (&BODY BODY)
  "Execute BODY in Atom Word mode.
The Atom Word synatx table is temporarily installed as the word syntax table."
  `(LET ((*MODE-WORD-SYNTAX-TABLE* *ATOM-WORD-SYNTAX-TABLE*))
     ,@BODY)) 

(DEFMACRO BP-CH-CHAR (BP)
  "Return the character after BP, ignoring bits and font attributes."
  `(MAKE-CHAR (BP-CHAR ,BP))) 

;;; Is it really necessary/right to count lines like this?
(DEFMACRO PRESERVE-POINT (&BODY BODY)
  "Execute BODY, moving (POINT) back afterward to where it was before BODY."
  `(LET ((NLINES (1- (COUNT-LINES (INTERVAL-FIRST-BP *INTERVAL*) (POINT) T)))
	 (NCHARS (BP-INDEX (POINT))))
     (PROGN . ,BODY)
     (MOVE-BP (POINT)
	      (FORWARD-CHAR (FORWARD-LINE (INTERVAL-FIRST-BP *INTERVAL*)
					  NLINES T)
			    NCHARS T)))) 

(DEFMACRO PRESERVE-BUFFER-POINT ((BUFFER) &BODY BODY)
  "Execute BODY, preserving BUFFER's POINT whether BUFFER is selected or not."
  `(LET (SAVED-POINT NLINES NCHARS)
     (SETQ SAVED-POINT (SEND ,BUFFER :POINT))
     (SETQ NLINES (COUNT-LINES (INTERVAL-FIRST-BP ,BUFFER) SAVED-POINT T)
	   NCHARS (BP-INDEX SAVED-POINT))
     (PROGN . ,BODY)
     (MOVE-BP (SEND ,BUFFER :POINT)
	      (LET ((*INTERVAL* ,BUFFER))
		(FORWARD-CHAR (FORWARD-LINE (INTERVAL-FIRST-BP ,BUFFER)
					    (1- NLINES) T)
			      NCHARS T))))) 

(DEFMACRO KILL-NEW-BUFFER-ON-ABORT ((BUFFER) &BODY BODY)
  "If the BODY aborts, kill the current buffer and reselect BUFFER.
Do not do either thing if the BODY returns normally."
  `(LET (.WINS.
	 (.BUFFER. ,BUFFER))
     (UNWIND-PROTECT
	 (PROG1 (PROGN . ,BODY)
		(SETQ .WINS. T))
       (OR .WINS. (EQ *INTERVAL* .BUFFER.)
	   (LET ((OBUFFER *INTERVAL*))
	     (SEND .BUFFER. :SELECT)
	     (KILL-BUFFER OBUFFER T)))))) 

(DEFMACRO WITH-REGION-OR-WHOLE-INTERVAL ((REGION-P-VAR) &BODY BODY)
  "If there is a mark now, bind *INTERVAL* to the region and REGION-P-VAR to T."
  `(MULTIPLE-VALUE-BIND (*INTERVAL* ,REGION-P-VAR)
       (REGION-OR-WHOLE-INTERVAL)
     (UNWIND-PROTECT (PROGN . ,BODY)
       (COND (,REGION-P-VAR
	      (FLUSH-BP (INTERVAL-FIRST-BP *INTERVAL*))
	      (FLUSH-BP (INTERVAL-LAST-BP *INTERVAL*))))))) 

(DEFMACRO DPRINT LIST
  (DO ((L LIST (CDR L))
       (RET NIL (CONS `(FORMAT T "~S = ~S; " ',(CAR L) ,(CAR L)) RET)))
      ((NULL L)
       `(PROGN
	  (TERPRI)
	  ,@(NREVERSE RET))))) 

(DEFMACRO CURRENT-FONT (WINDOW &OPTIONAL (NUMBER '*FONT*))
  "Return the font which is now current in the zwei-window WINDOW.
The value of *FONT* is used as the current font number.
If NUMBER is specified, it is an expression for the font number.
This should be used only on WINDOWs, not other kinds of displayers."
  `(AREF (W:SHEET-FONT-MAP (WINDOW-SHEET ,WINDOW)) ,NUMBER)) 

(DEFMACRO TYPEIN-LINE-ACTIVATE (&BODY BODY)
  "Execute BODY with a cursor blinking in the typein line."
  `(LET (OLD-SELECTED-WINDOW OLD-SUBSTITUTE)
     (UNWIND-PROTECT
	 (PROGN
	   (COND ((SEND *QUERY-IO* :OPERATION-HANDLED-P :SELECT)
		  (SEND *QUERY-IO* :SEND-IF-HANDLES :OUTPUT-HOLD-EXCEPTION)
		  (SETQ OLD-SELECTED-WINDOW (SEND *QUERY-IO* :TOP-OF-EDITOR-HIERARCHY))
		  (SETQ OLD-SUBSTITUTE (SEND OLD-SELECTED-WINDOW :SELECTION-SUBSTITUTE))
		  (SEND OLD-SELECTED-WINDOW :SET-SELECTION-SUBSTITUTE *TYPEIN-WINDOW*)))
	   ,@BODY)
       (AND OLD-SELECTED-WINDOW
	    (SEND OLD-SELECTED-WINDOW :SET-SELECTION-SUBSTITUTE OLD-SUBSTITUTE))))) 
	
(DEFMACRO TEMPORARY-WINDOW-SELECT ((ZWEI-WINDOW) &BODY BODY)
  "Temporarily select ZWEI-WINDOW, such as the one used for View File, etc.
We arrange that any attempt to use the original window for typeout
selects it again and deactivates the temporary window."
  `(LET* ((NEW-WINDOW ,ZWEI-WINDOW)
	  (*TERMINAL-IO* (SEND NEW-WINDOW :TYPEOUT-WINDOW)))
     (W:WITH-SELECTION-SUBSTITUTE
       (NEW-WINDOW (SEND *WINDOW* :TOP-OF-EDITOR-HIERARCHY))
       ,@BODY))) 

(DEFMACRO WITHOUT-IO-BUFFER-OUTPUT-FUNCTION (&BODY BODY)
  "Executes BODY inhibiting synchronous intercepted characters (Break, Abort, etc).
Asynchronous characters such as Control-Abort are not affected."
  `(LET (W:KBD-INTERCEPTED-CHARACTERS)
     ,@BODY)) 

;;; Macros used to make command easy to write.
(DEFMACRO POINT ()
  "The current editing pointer in the current window."
  '(WINDOW-POINT *WINDOW*)) 

(DEFMACRO MARK ()
  "The current window's mark, or NIL."
  '(WINDOW-MARK *WINDOW*)) 

(DEFMACRO REGION ((BP1 BP2) &BODY BODY)
  "Execute BODY with BP1, BP2 bound to bps to the ends of the region.
Barfs if there is no region now."
  `(LET ((,BP1 (POINT))
	 (,BP2 (MARK)))
     (COND ((NOT (WINDOW-MARK-P *WINDOW*))
	    (BARF "There is no region."))
	   ((BP-< ,BP2 ,BP1)
	    (PSETQ ,BP1 ,BP2 ,BP2 ,BP1)))
     ,@BODY)) 

(DEFMACRO REGION-LINES ((START-LINE STOP-LINE) &BODY BODY)
  "Execute BODY with START-LINE, END-LINE bound to the first and last lines in the region.
Barfs if there is no region now.
A line is \"in\" the region iff its beginning is included."
  `(LET ((REGION-LINES-BP1 (POINT))
	 (REGION-LINES-BP2 (MARK)))
     (COND ((NOT (WINDOW-MARK-P *WINDOW*))
	    (BARF "There is no region."))
	   ((BP-< REGION-LINES-BP2 REGION-LINES-BP1)
	    (PSETQ REGION-LINES-BP1 REGION-LINES-BP2
		   REGION-LINES-BP2 REGION-LINES-BP1)))
     (INTERVAL-LINES (REGION-LINES-BP1 REGION-LINES-BP2)
		     (,START-LINE ,STOP-LINE)
       ,@BODY))) 

(DEFMACRO INTERVAL-LINES ((BP1 BP2) (START-LINE STOP-LINE) &BODY BODY)
  "Executes BODY with START-LINE, STOP-LINE bound to the first and last lines
in the interval designated by BP1, BP2.  BP1, BP2 are inputs.
A line is \"in\" the interval iff its beginning is included."
  `(LET ((,START-LINE (COND ((ZEROP (BP-INDEX ,BP1))
			     (BP-LINE ,BP1))
			    (T
			     (LINE-NEXT (BP-LINE ,BP1)))))
	 (,STOP-LINE (COND ((ZEROP (BP-INDEX ,BP2))
			    (BP-LINE ,BP2))
			   (T
			    (LINE-NEXT (BP-LINE ,BP2))))))
     ,@BODY)) 

;;; PRESERVE-POINT

(DEFMACRO WITH-BP ((VARIABLE BP TYPE) &BODY BODY)
  "Execute BODY with VARIABLE bound to a bp copied from BP but with type TYPE.
TYPE is :NORMAL or :MOVES.  It could be NIL, but that probably defeats the purpose.
The new BP is FLUSH-BP'd when BODY is finished."
  `(LET ((,VARIABLE (COPY-BP ,BP ,TYPE)))
     (UNWIND-PROTECT (PROGN . ,BODY)
       (FLUSH-BP ,VARIABLE)))) 

(DEFMACRO BIND-MODE-LINE (LIST &BODY BODY)
  "Execute BODY with LIST as the mode-line list."
  `(LET ((*MODE-LINE-LIST* ,LIST))
     ,@BODY)) 


(DEFMACRO BIND-MOUSE-DOCUMENTATION-STRING (STRING &BODY BODY)
  "Execute BODY with STRING displayed in the who line as mouse documentation."
  `(LET ()
     (BIND (SEND *WINDOW* :MOUSE-DOCUMENTATION-STRING-LOCATION) ,STRING)
     ,@BODY)) 

;;; This is for things that take an interval as an argument, or two bps, maybe in order,
;;; it canonicalises them into two ordered bps
(DEFMACRO GET-INTERVAL (START-BP END-BP IN-ORDER-P)
  "Given variables START-BP, END-BP that specify an interval, canonicalize them.
There are two valid ways thet interval can be specified.
Either START-BP can actually be an interval, or START-BP and END-BP
can be buffer pointers.  In that case, IN-ORDER-P says it is ok
to assume that START-BP comes earlier in the buffer than END-BP.

When this macro exits, START-BP and END-BP are always buffer pointers,
and always in order (unless you lied with specifying IN-ORDER-P)."
  `(COND ((NULL ,END-BP)
	  (SETQ ,END-BP (INTERVAL-LAST-BP ,START-BP)
		,START-BP (INTERVAL-FIRST-BP ,START-BP)))
	 ((NOT ,IN-ORDER-P)
	  (ORDER-BPS ,START-BP ,END-BP)))) 

(DEFMACRO WITH-READ-ONLY-SUPPRESSED ((INTERVAL) &BODY BODY)
  "Execute BODY, allowing it to modify INTERVAL even if it is read-only."
  `(LET ((*READ-ONLY-SUPPRESSED-INTERVAL* ,INTERVAL))
     ,@BODY)) 

(DEFMACRO WITH-UNDO-SAVE ((TYPE BP1 BP2 IN-ORDER-P) &BODY BODY)
  "Record changes made by BODY to interval from BP1 to BP2 for undo.
TYPE is a string saying the kind of operation, for the user.
It should start with a capital letter."
  `(UNWIND-PROTECT
       (LET ((*BATCH-UNDO-SAVE* T))
	 (UNDO-SAVE-BEGIN ,BP1 ,BP2 ,IN-ORDER-P ,TYPE)
	 ,@BODY)
     (UNDO-SAVE-END))) 

(DEFMACRO WITH-UNDO-SAVE-IF (COND-FORM (TYPE BP1 BP2 IN-ORDER-P) &BODY BODY)
  "Record changes made by BODY to interval from BP1 to BP2 for undo, if COND-FORM evals non-NIL.
TYPE is a string saying the kind of operation, for the user.
It should start with a capital letter."
  (LET ((GENSYM (GENSYM)))
    `(LET ((,GENSYM ,COND-FORM))
       (UNWIND-PROTECT
	   (LET-IF ,GENSYM ((*BATCH-UNDO-SAVE* T))
	     (WHEN ,GENSYM (UNDO-SAVE-BEGIN ,BP1 ,BP2 ,IN-ORDER-P ,TYPE))
	     ,@BODY)
	 (WHEN ,GENSYM (UNDO-SAVE-END)))))) 

;;; Defines a command.  Form is:
;;; (DEFCOM COM-foo "Documentation." OPTIONS-LIST . BODY)
;;; Note: unlike EINE, there is no lambda-list.
(DEFMACRO DEFCOM (FN DOC OPTIONS &BODY DEF)
  "Options are:  (KM)   -- This command always preserves MARK.
              (SM)   -- Set MARK. There will be a region after this command
              (NM)   -- Reset MARK. There will be no region after this command
              (PUSH) -- Push POINT onto the point-pdl
              (R)    -- Frob recentering fraction for forward move
              (-R)   -- Frob recentering fraction for backward move."
  `(PROGN
     'COMPILE
     (COMMAND-DEFINE ',FN ',DOC ',OPTIONS)
     (DEFUN ,FN ()
       ,@(PROCESS-COMMAND-OPTIONS OPTIONS)
       ,@DEF))) 

(DEFVAR *COMMAND-ALIST* ()
   "Alist of all ZWEI command names (strings) versus the commands themselves (symbols).") 

(DEFUN COMMAND-DEFINE (COMMAND DOC IGNORE)
  (COND ((STRINGP DOC)
	 (SETF (DOCUMENTATION COMMAND 'FUNCTION) DOC))
	((OR (SYMBOLP DOC)
	     (AND (NOT (ATOM DOC))
		  (MEMBER (CAR DOC) '(FUNCTION LAMBDA) :TEST #'EQ)))
	 (SETF (GET COMMAND 'DOCUMENTATION-FUNCTION) DOC))
	(T
	 (FERROR NIL "The command ~S has invalid self-documentation ~S" COMMAND DOC)))
  (LET* ((NAME (MAKE-COMMAND-NAME COMMAND))
	 (AENTRY (ASSOC NAME *COMMAND-ALIST* :TEST 'EQUALP)))
    (SETF (GET COMMAND 'COMMAND-NAME) NAME)
    (IF AENTRY
	(UNLESS (EQ COMMAND (CDR AENTRY))
	  (WHEN (FQUERY NIL "The command name ~S is currently set up to call ~S;
redefine it to call ~S? " NAME (CDR AENTRY) COMMAND)
	    (SETF (CDR AENTRY) COMMAND)))
	(PUSH (CONS NAME COMMAND) *COMMAND-ALIST*)))) 

;No need to record COMMAND-DEFINE forms in QFASL files,
;since they accompany DEFUNs which get recorded.
(DEFPROP COMMAND-DEFINE T SI:QFASL-DONT-RECORD) 

(DEFUN PROCESS-COMMAND-OPTIONS (OPTIONS)
  (DO ((L OPTIONS (CDR L))
       (RET NIL (APPEND (CDR (ASSOC (CAR L)
				    '((NM (SETF (WINDOW-MARK-P *WINDOW*) NIL))
				      (SM (SETF (WINDOW-MARK-P *WINDOW*) T)
					  (SETQ *MARK-STAYS* T))
				      (KM (SETQ *MARK-STAYS* T))
				      (R (SETQ *CENTERING-FRACTION* (IF (PLUSP *NUMERIC-ARG*)
									*MIN-RESET-FRACTION*
									*MAX-RESET-FRACTION*)))
				      (-R (SETQ *CENTERING-FRACTION* (IF (PLUSP *NUMERIC-ARG*)
									 *MAX-RESET-FRACTION*
									 *MIN-RESET-FRACTION*)))
				      (PUSH (POINT-PDL-PUSH (POINT) *WINDOW*))
				      (OTHERWISE (FERROR NIL "Unknown DEFCOM option ~S" (CAR L))))
				    :TEST #'EQ))
			RET)))
      ((NULL L) RET))) 

;;; Convert a string into human-readable form.  Remove leading COM-, or leading
;;; and trailing *'s.  Conver hyphens into spaces, and capitalize each word.
;;; This is used both for command names and variable names.
(DEFUN MAKE-COMMAND-NAME (COMMAND)
  (SETQ COMMAND (STRING COMMAND))
  (LET ((CLEN (LENGTH COMMAND)))
    (LET ((STR (SUBSEQ COMMAND
		       (COND ((STRING-EQUAL "COM-" COMMAND :START1 0 :START2 0 :END1 4 :END2 4) 4)
			     ((STRING-EQUAL "*" COMMAND :START1 0 :START2 0 :END1 1 :END2 1) 1)
			     (T 0))
		       (COND ((CHAR-EQUAL #\* (AREF COMMAND (1- CLEN))) (1- CLEN))
			     (T CLEN)))))
      (STRING-CAPITALIZE-WORDS STR)))) 

;;; Macro to define a major mode.
(DEFMACRO DEFMAJOR (COMMAND-NAME MODE-SYMBOL MODE-NAME
		    COMMAND-DOCUMENTATION COMMAND-OPTIONS &BODY BODY)
  "Define a major mode.
COMMAND-NAME is a symbol, such as COM-MY-MODE.
MODE-SYMBOL is a symbol for the mode, such as MY-MODE.
MODE-NAME is a string for the mode, such as \"My\".
COMMAND-DOCUMENTATION is the doc string for COM-MY-MODE.
COMMAND-OPTIONS is the DEFCOM options list.
BODY is not simply a list of arbitrary forms;
only SETQ, PUSH, ASET, SET-COMTAB, SET-CHAR-SYNTAX,
 SET-SYNTAX-TABLE-INDIRECTION and COMMAND-HOOK forms may be used.
[Actually, one can enclose arbitrary forms using PROGN. No attempt
to undo them is made -- they are just evaluated when the mode is
turned on.]"
  (DEFINE-MODE-MACRO T MODE-SYMBOL MODE-NAME 0
		     COMMAND-NAME COMMAND-DOCUMENTATION COMMAND-OPTIONS BODY)) 

;;; Macro to define a minor mode.
(DEFMACRO DEFMINOR (COMMAND-NAME MODE-SYMBOL MODE-NAME
		    MODE-LINE-POSITION COMMAND-DOCUMENTATION COMMAND-OPTIONS &BODY BODY)
  "Define a minor mode.  Usage is like DEFMAJOR (which see)."
  (DEFINE-MODE-MACRO NIL MODE-SYMBOL MODE-NAME MODE-LINE-POSITION
		     COMMAND-NAME COMMAND-DOCUMENTATION COMMAND-OPTIONS BODY)) 

;;; Internal function to generate code for a major or minor mode.
(DEFUN DEFINE-MODE-MACRO (MAJOR-P MODE-SYMBOL MODE-NAME MODE-LINE-POSITION
			  COMMAND-NAME COMMAND-DOCUMENTATION COMMAND-OPTIONS BODY)
  `(PROGN 'COMPILE
	  (SETF (GET ',MODE-SYMBOL 'MODE) ',BODY) 
	  (SETF (GET ',MODE-SYMBOL 'MAJOR-MODE-P) ',MAJOR-P)
	  (SETF (GET ',MODE-SYMBOL 'MODE-LINE-POSITION) ,MODE-LINE-POSITION)
	  (AND ',MAJOR-P
	       (SETF (GET ',MODE-SYMBOL 'MODE-HOOK-SYMBOL)
		     ',(INTERN (STRING-APPEND MODE-SYMBOL "-HOOK")
			       (SYMBOL-PACKAGE 'FOO))))
	  (DEFVAR ,(INTERN (STRING-APPEND MODE-SYMBOL "-HOOK")
			   (SYMBOL-PACKAGE 'FOO)))
	  (SETQ ,MODE-SYMBOL ,(COND ((ZEROP (LENGTH MODE-NAME))
				     NIL)
				    (MAJOR-P
				     MODE-NAME)
				    (T
				     (STRING-APPEND #\SPACE MODE-NAME))))
	  (DEFCOM ,COMMAND-NAME ,COMMAND-DOCUMENTATION ,COMMAND-OPTIONS
	    ,(IF MAJOR-P
		 `(PROGN (TURN-OFF-MODE *MAJOR-MODE*)
			 (DOLIST (MODE *UNSTICKY-MINOR-MODES*)
			   (TURN-OFF-MODE MODE))
			 (TURN-ON-MODE ',MODE-SYMBOL)
			 (SEND *INTERVAL* :SET-ATTRIBUTE
			       :MODE ,(INTERN (STRING-UPCASE MODE-NAME) :KEYWORD) :QUERY))
		 `(IF (IF *NUMERIC-ARG-P*
			  (ZEROP *NUMERIC-ARG*)
			  (ASSOC ',MODE-SYMBOL *MODE-LIST* :TEST #'EQ) )
		      (TURN-OFF-MODE ',MODE-SYMBOL)
		      (TURN-ON-MODE ',MODE-SYMBOL)))
	    DIS-NONE))) 

(DEFMACRO WITH-EDITOR-STREAM ((STREAM-VAR . OPTIONS) &BODY BODY)
  "Executes BODY with STREAM-VAR bound to a stream reading and writing an editor buffer.
OPTIONS are as for OPEN-EDITOR-STREAM."
  (LET ((GENVAR (GENSYM)))
    `(LET* ((,GENVAR (OPEN-EDITOR-STREAM . ,OPTIONS))
	    (,STREAM-VAR ,GENVAR))
       (UNWIND-PROTECT (PROGN . ,BODY)
	 (SEND ,GENVAR :FORCE-REDISPLAY))))) 

;;; A variable is a symbol, whose print name starts and ends with "*".
;;; The value of the variable is the value of the symbol.
;;; It has the following  properties:
;;; VARIABLE-NAME             The name, a string derived from the print-name of the symbol.
;;; VARIABLE-INIT             The initial value.
;;; VARIABLE-TYPE             One of the type symbols below.
;;; (documentation <var> 'variable)
;;;			      A string documenting the variable.  The first line
;;;                             is the "short form.

(DEFMACRO DEFVARIABLE (VAR INIT TYPE &BODY BODY)
  "Define a ZWEI option variable, initializing it.
VAR is the variable to be defined.  INIT is an expression to initialize it with.
TYPE is a keyword saying what kind of value the variable should have:
 :BOOLEAN	    T or NIL.
 :KEYWORD	    A symbol in the keyword package.
 :STRING 	    A string.
 :CHAR   	    A character as a fixnum.
 :CHAR-LIST	    A list of characters as fixnums.
 :FIXNUM   	    A fixnum.
 :FIXNUM-OR-NIL     A fixnum or NIL.
 :PIXEL		    A fixnum which is a number of pixels, horizontally.
 :PIXEL-OR-NIL	    A fixnum which is a number of pixels, horizontally, or NIL.
 :SMALL-FRACTION    A small flonum between 0.0s0 and 1.0s0, inclusively.
 :ANYTHING          Any Lisp object.
DOC is a documentation string for the variable.
All ZWEI option variables are reinitialized from their INITs
when ZWEI is initialized."
  (DECLARE (ARGLIST VAR INIT TYPE DOC))
  `(PROGN
     'COMPILE
     (DEFINE-VARIABLE ',VAR ,INIT ',TYPE)
     (DEFPARAMETER ,VAR ,INIT ,(CAR BODY)))) 

(DEFVAR *VARIABLE-ALIST* ()
   "Alist of ZWEI option variable pretty names versus the symbols themselves.") 

(DEFUN DEFINE-VARIABLE (VAR INIT TYPE)
  (CHECK-ARG TYPE (MEMBER TYPE '(:BOOLEAN :KEYWORD :STRING :FIXNUM-OR-NIL
					  :SMALL-FRACTION :CHAR :CHAR-LIST
					  :FIXNUM :ANYTHING :PIXEL :PIXEL-OR-NIL)
			  :TEST #'EQ)
	     "a valid ZWEI variable type")
  (LET ((NAME (MAKE-COMMAND-NAME VAR)))
    (SETF (GET VAR 'VARIABLE-NAME) NAME)
    (OR (ASSOC NAME *VARIABLE-ALIST* :TEST 'EQUALP)
	(PUSH (CONS NAME VAR) *VARIABLE-ALIST*)))
  (SETF (GET VAR 'VARIABLE-INIT) INIT)
  (SETF (GET VAR 'VARIABLE-TYPE) TYPE)) 

(DEFUN SETQ-ZWEI-VARIABLES ()
  "Set all ZWEI option variables (defined with DEFVARIABLE) to their initial values."
  (DO ((L *VARIABLE-ALIST* (CDR L)))
      ((NULL L) NIL)
    (LET ((V (CDAR L)))
      (SET V (GET V 'VARIABLE-INIT)))))

(DEFMACRO DEFINE-INDENTATION (SYMBOL INDENTATION)
  "INDENTATION is a list of (FORM-COUNT OFFSET FORM-COUNT OFFSET ...)
which determines the indentation of the lines below a line containing \"(<SYMBOL>\".
OFFSET is amount to offset indentation when the start of the line is FORM-COUNT
s-expressions past SYMBOL.  For example, (DEFINE-INDENTATION LET (1 1)) means to set
the indentation for LET so that the first line after \"(LET ((FOO BAR))\" will be
offset 1 space from the standard indentation, which is the value of
*LISP-INDENT-LONE-FUNCTION-OFFSET*.  INDENTATION may also be a symbol, which is
taken as a function that may return three values;  see INDENT-PROG, the only current
use of this feature."
  `(UNLESS (GET ',SYMBOL 'LISP-INDENT-OFFSET)
     (SETF (GET ',SYMBOL 'LISP-INDENT-OFFSET) ',INDENTATION)))
;;;
;;; Add the special forms after the indentations already set up by defmacro.
(DEFINE-INDENTATION LAMBDA (1 1))
(DEFINE-INDENTATION *CATCH (1 1))
(DEFINE-INDENTATION CATCH (1 1))	; CATCH IS HERE TO HELP CLI:CATCH.
(DEFINE-INDENTATION CASE (1 1))
(DEFINE-INDENTATION LET (1 1))
(DEFINE-INDENTATION LET* (1 1))
(DEFINE-INDENTATION LETF (1 1))				; jlm 5/3/89
(DEFINE-INDENTATION LETF* (1 1))			; jlm 5/3/89
(DEFINE-INDENTATION LET-IF (2 1))
(DEFINE-INDENTATION WITH-STACK-LIST (1 1))
(DEFINE-INDENTATION WITH-STACK-LIST* (1 1))
(DEFINE-INDENTATION CATCH-CONTINUATION-IF (2 3 4 1))
(DEFINE-INDENTATION CATCH-CONTINUATION (1 3 3 1))
(DEFINE-INDENTATION CONDITION-CASE (1 3 2 1))
(DEFINE-INDENTATION CONDITION-CASE-IF (2 3 3 1))
(DEFINE-INDENTATION CONDITION-CALL (1 3 2 1))
(DEFINE-INDENTATION CONDITION-CALL-IF (2 3 3 1))
(DEFINE-INDENTATION COMPILER-LET (1 1))
(DEFINE-INDENTATION PROGV (2 1))
(DEFINE-INDENTATION PROGW (1 1))
(DEFINE-INDENTATION DO (2 1))
(DEFINE-INDENTATION DO-NAMED (3 1))
(DEFINE-INDENTATION DO* (2 1))
(DEFINE-INDENTATION DO*-NAMED (3 1))
(DEFINE-INDENTATION RETURN-FROM (1 1))
(DEFINE-INDENTATION PROG INDENT-PROG)
(DEFINE-INDENTATION PROG* INDENT-PROG)
(DEFINE-INDENTATION BLOCK (1 1))
(DEFINE-INDENTATION TAGBODY INDENT-PROG)
(DEFINE-INDENTATION MULTIPLE-VALUE (1 1))
(DEFINE-INDENTATION MULTIPLE-VALUE-BIND (1 3 2 1))
(DEFINE-INDENTATION MULTIPLE-VALUE-SETQ (1 1))
(DEFINE-INDENTATION DEFFLAVOR (1 12 3 1))
(DEFINE-INDENTATION DEFSTRUCT (1 1))
(DEFINE-INDENTATION DEFPROP (0 0))
(DEFINE-INDENTATION EVAL-WHEN (1 1))
(DEFINE-INDENTATION UNWIND-PROTECT (0 3 1 1))
(DEFINE-INDENTATION FLET (1 1))
(DEFINE-INDENTATION MACROLET (1 1))
(DEFINE-INDENTATION LABELS (1 1))
;; may 03/28/89 Added CLOS stuff below for DNG
(define-indentation ticlos:generic-function (1 1))	;; same as LET
(define-indentation ticlos:generic-flet     (1 1)) 	;; same as FLET
(define-indentation ticlos:generic-labels   (1 1)) 	;; same as LABELS



;;;
;;;  Give symbols in the CLI package the same indentation as their
;;;  GLOBAL counterparts.
(LOOP FOR (ZL . CL) IN SI:*COMMON-LISP-SYMBOL-SUBSTITUTIONS*
      AS INDENTATION = (GET ZL 'LISP-INDENT-OFFSET)
      WHEN INDENTATION
      DO (DEFINE-INDENTATION CL INDENTATION))
 
;;;
;;;  Be nice to UCL.  Push them on the front because some are macros and we
;;;  want to override defmacro's interpretation of the indentation.
(DEFPROP DEFCOMMAND (2 1 3 1) LISP-INDENT-OFFSET)
(DEFPROP MAKE-COMMAND (1 1) LISP-INDENT-OFFSET)
(DEFPROP BUILD-COMMAND-TABLE (2 1) LISP-INDENT-OFFSET)
(DEFPROP BUILD-MENU (2 1) LISP-INDENT-OFFSET)

(DEFVARIABLE *FILL-COLUMN* 1100 :PIXEL
  "Width in pixels used for filling text.") 
(DEFVARIABLE *PARAGRAPH-DELIMITER-LIST* '(#\. #\SPACE #\TAB) :CHAR-LIST
  "Characters which at beginning of line mean a new paragraph.") 
(DEFVARIABLE *PAGE-DELIMITER-LIST* '(#\PAGE) :CHAR-LIST
  "Characters which separate pages.")
;;; *DEFAULT-MAJOR-MODE* added for release 3 on 10/29/86 by rpm.
;;; It used to be in SYS:SYS;LISP-MODE for the cold load. 
(DEFVARIABLE *DEFAULT-MAJOR-MODE* :COMMON-LISP :ANYTHING
  "The major mode in which new buffers are placed by default.")
(DEFVARIABLE *STICKY-MINOR-MODES* '(ATOM-WORD-MODE WORD-ABBREV-MODE) :ANYTHING
  "Minor modes to carry from current buffer to new ones.") 
(DEFVARIABLE *UNSTICKY-MINOR-MODES*
	     '(ELECTRIC-SHIFT-LOCK-MODE ELECTRIC-FONT-LOCK-MODE RETURN-INDENTS-MODE)
	     :ANYTHING
  "Minor modes that are turned off when the mode is changed explicitly") 
(DEFVARIABLE *INITIAL-MINOR-MODES* NIL :ANYTHING
  "Minor modes turned on for either all or selected major modes.
For example, '(auto-fill-mode ((zetalisp-mode common-lisp-mode) return-indents-mode))
will turn on auto-fill-mode for all modes and return-indents-mode 
for zetalisp and common-lisp modes ONLY.") 
(DEFVARIABLE *DEFAULT-SAVE-MODE* :ASK :KEYWORD
  "Default save mode for new buffers (NIL, :ASK, :ALWAYS).") 
(DEFVARIABLE *FIND-FILE-SAVE-MODE* :ASK :KEYWORD
  "Default save mode for new buffers create by Find File (NIL, :ASK, :ALWAYS).") 
(DEFVARIABLE *DIRECTORY-LISTER* 'DEFAULT-DIRECTORY-LISTER :ANYTHING
  "Function used by Display Directory and auto directory display option.") 
(DEFVARIABLE *DIRECTORY-SINGLE-FILE-LISTER* 'DEFAULT-LIST-ONE-FILE :ANYTHING
  "Function normally called to display each file") 
(DEFVARIABLE *AUTO-PUSH-POINT-OPTION* 12 :FIXNUM-OR-NIL
  "Searches push point if it moves more than this many lines.") 
(DEFVARIABLE *AUTO-PUSH-POINT-NOTIFICATION* "Point pushed" :STRING
  "This is typed in the echo area when point is automatically pushed.") 
(DEFVARIABLE *AUTO-DIRECTORY-DISPLAY* NIL :KEYWORD
  "Tells on which kind of file commands to display directory (NIL, :READ, :WRITE, T).") 
(DEFVARIABLE *TAB-BLINKER-FLAG* T :BOOLEAN
  "If a blinker is placed over a tab, make the blinker the width of a space.") 
(DEFVARIABLE *FILL-PREFIX* "" :STRING
  "String to put before each line when filling.") 
(DEFVARIABLE *FILL-EXTRA-SPACE-LIST* '(#\. #\! #\?) :CHAR-LIST
  "Characters that should be followed by two spaces when filling is done.") 
(DEFVARIABLE *FLASH-MATCHING-PAREN* T :BOOLEAN
  "When point is to the right of a close paren, flash the matching open paren.") 
(DEFVARIABLE *FLASH-MATCHING-PAREN-MAX-LINES* 310 :FIXNUM
  "Max number of lines to scan when trying to flash the matching open paren.")
;;; Allow Prolog to specify as many lines as it wishes, independent of LISP modes.
;;; Added on 12-16-86 by rpm.
(DEFPROP *FLASH-MATCHING-PAREN-MAX-LINES* T MODE-SETTABLE-P)
(DEFVARIABLE *COMMENT-START* NIL :STRING
  "String that indicates the start of a comment.") 
(DEFVARIABLE *COMMENT-BEGIN* ";" :STRING
  "String for beginning new comments.") 
(DEFVARIABLE *COMMENT-END* "" :STRING
  "String for ending comments.") 
(DEFVARIABLE *COMMENT-COLUMN* (* 60 10) :PIXEL
  "Column (in pixels) in which to start new comments.") 
(DEFVARIABLE *CASE-REPLACE* T :BOOLEAN
  "Replacing commands try to preserve case.") 
(DEFVARIABLE *ALPHABETIC-CASE-AFFECTS-SEARCH* NIL :BOOLEAN
  "Search commands demand exact match of case of letters.") 
(DEFVARIABLE *PERMANENT-REAL-LINE-GOAL-XPOS* NIL :PIXEL-OR-NIL
  "If non-NIL, goal for Up and Down Real Line commands.") 
(DEFVARIABLE *SPACE-INDENT-FLAG* NIL :BOOLEAN
  "If true, Auto Fill mode will indent new lines.") 
(DEFVARIABLE *POINT-PDL-MAX* 10 :FIXNUM
  "The maximum number of elements on the point PDL.
The point PDL is the push-down-list of saved places in the buffer
where the POINT has been.") 
;New history stuff
(DEFVARIABLE *SEARCH-RING-MAX* 3 :FIXNUM
  "The maximum number of elements on the search ring.
The search ring is the ring buffer of default search strings.") 
(DEFVARIABLE *CENTER-FRACTION* 0.5s0 :SMALL-FRACTION
  "Where to recenter the window.
This is how far down in the window the point should be placed when ZWEI
recenters POINT in the window, as a fraction from 0.0 to 1.0.") 
(DEFVARIABLE *MIN-RESET-FRACTION* 0.8s0 :SMALL-FRACTION
  "Where to recenter the window when you go off the bottom.
This is how far down in the window the point should be placed when ZWEI
moves the text in the window because you moved off the bottom.
It should be a fraction from 0.0 to 1.0.") 
(DEFVARIABLE *MAX-RESET-FRACTION* 0.2s0 :SMALL-FRACTION
  "Where to recenter the window when you go off the top.
This is how far down in the window the point should be placed when ZWEI
moves the text in the window because you moved off the top.
It should be a fraction from 0.0 to 1.0.") 
(DEFVARIABLE *BLANKS* '(#\SPACE #\TAB #\BACKSPACE) :CHAR-LIST
  "List of characters that ZWEI thinks of as blanks.
The initial contents of this variable are the characters BLANK, TAB, and BACKSPACE.") 
(DEFVARIABLE *WHITESPACE-CHARS* '(#\SPACE #\TAB #\NEWLINE #\BACKSPACE) :CHAR-LIST
  "List of characters that ZWEI thinks of as blanks.
The initial contents of this variable are the characters BLANK, TAB, CR, and BACKSPACE.") 
(DEFVARIABLE *REGION-MARKING-MODE* :UNDERLINE :KEYWORD
  "How to mark the region.
This variable tells ZWEI how to denote the region between POINT and MARK.
It should be a symbol, either :UNDERLINE or :REVERSE-VIDEO.") 
(DEFVARIABLE *CR-AFTER-LINE-MOUSE-YANK* T :BOOLEAN
   "If true inserts #\\cr after yanks of single lines from typeout window, see mouse-yank.") 
(DEFVARIABLE *FINISHED-AFTER-ENTIRE-MOUSE-YANK* NIL :BOOLEAN
   "If true the display is refreshed after a yank of an entire node.") 
(DEFVARIABLE *REGION-RIGHT-MARGIN-MODE* NIL :BOOLEAN
  "Should region marking extend to right margin, if region extends past the line?") 
(DEFVARIABLE *DEFAULT-PACKAGE* NIL :ANYTHING
  "The package new buffers are initially in.") 
(DEFVARIABLE *INDENT-WITH-TABS* T :BOOLEAN
  "T if indentation commands should generate Tab characters.") 
(DEFVARIABLE *LISP-INDENT-OFFSET* NIL :FIXNUM-OR-NIL
  "Same as QLisp Indent Offset in EMACS.  Good luck trying to use it. - DLW & MMcM") 
(DEFVARIABLE *COMMENT-ROUND-FUNCTION* 'ROUND-FOR-COMMENT :ANYTHING
  "Function used to round up column when comments cannot be aligned to comment column.") 
(DEFVARIABLE *LISP-DEFUN-INDENTATION* '(2 1 3 1) :ANYTHING
  "Amount to indent the second line of a defun.")  
(DEFVARIABLE *LISP-INDENT-LONE-FUNCTION-OFFSET* 1 :FIXNUM
  "Amount to offset indentation of car of list.") 
(DEFVARIABLE *FILE-VERSIONS-KEPT* 2 :FIXNUM
  "Number of non-superfluous versions of a file in Dired.") 
(DEFVARIABLE *TEMP-FILE-TYPE-LIST* '("MEMO" "XGP" "@XGP" "UNFASL" "OUTPUT" "OLREC" "PRESS")
	     :ANYTHING
  "List of strings which are file types to be automatically 
marked for deletion in Dired.") 
(DEFVARIABLE *TEXT-JUSTIFIER-ESCAPE-LIST* '(#\. #\@ #\- #\\ #\') :CHAR-LIST
  "List of characters that start text justifier commands when at the start of the line.") 
(DEFVARIABLE *TEXT-JUSTIFIER-UNDERLINE-BEGIN* "" :STRING
  "String to start an underlining.") 
(DEFVARIABLE *TEXT-JUSTIFIER-UNDERLINE-END* "" :STRING
  "String to end an underlining.") 
(DEFVARIABLE *PL1-INDING-STYLE* 1 :FIXNUM
  "Pl1 indentation style.") 
(DEFVARIABLE *ELECTRIC-SHIFT-LOCK-XORS* T :BOOLEAN
  "Shift key acts as xor of shift in electric shift-lock mode") 
(DEFVARIABLE *ELECTRIC-FONT-LOCK-IF-NO-FONTS-DECLARED-P* NIL :BOOLEAN
   "Indicates what will happen when Electric Font Lock Mode is on but no fonts are declared in the attribute line.
If T, input will be done using different fonts.  NIL doesn't use fonts for these files.") 
(DEFVARIABLE *ELECTRIC-FONT-LOCK-NON-DOC-STRING-FONT* :FONT-D-OR-A :ANYTHING
   "What font to use for non-documentation strings in Electric Font Lock Mode.  Value is a fixnum or a symbol.
 0 means Font A, 1 means B, etc.  If the value is not a fixnum, Font D is used if present, else Font A is used.") 
(DEFVARIABLE *KILL-INTERVAL-SMARTS* NIL :BOOLEAN
  "Kill and yank commands try to optimize whitespace") 
(DEFVARIABLE *INDENT-NEW-LINE-NEW-LINE-FUNCTION* NIL :ANYTHING
  "Function called by Indent New Line to move to a new line.
If NIL, the function assigned to the Return key is used.") 
(DEFPROP *INDENT-NEW-LINE-NEW-LINE-FUNCTION* T MODE-SETTABLE-P) 
(DEFVARIABLE *INDENT-NEW-LINE-INDENT-FUNCTION* NIL :ANYTHING
  "Function called by Indent New Line to perform indentation.
If NIL, the function assigned to the Tab key is used.") 
(DEFPROP *INDENT-NEW-LINE-INDENT-FUNCTION* T MODE-SETTABLE-P) 

(DEFVARIABLE *AUTO-FILL-ACTIVATION-CHARACTERS* '(#\SPACE #\NEWLINE #\. #\? #\!) :CHAR-LIST
  "Characters which cause filling in auto fill mode") 

(DEFVARIABLE *VSP* 2 :FIXNUM
  "Vertical interline spacing: number of blank rows of pixels between lines.") 

(DEFVARIABLE *UNDO-SAVE-SMALL-CHANGES* T :BOOLEAN
  "Should minor editing be recorded for M-X Undo?") 

(DEFVARIABLE *ONE-WINDOW-DEFAULT* :CURRENT :KEYWORD
  "Controls which window is selected after the One Window command.
:TOP means the uppermost window.  :BOTTOM the lowermost.
:CURRENT the current window.  :OTHER some other window.") 

(DEFVARIABLE *FIND-FILE-EARLY-SELECT* NIL :BOOLEAN
  "T => Visiting a file should display the first screenful as soon as it is read in.") 

(DEFVARIABLE *FIND-FILE-NOT-FOUND-IS-AN-ERROR* NIL :BOOLEAN
  "T => Find File when file does not exist does not make an empty buffer.") 

(DEFVARIABLE *DEFAULT-BASE* NIL :FIXNUM
  "Default value for *PRINT-BASE* and *READ-BASE* when file attribute line does not specify one.") 

(ADD-INITIALIZATION 'CLEAR-*DEFAULT-BASE* '(SETQ *DEFAULT-BASE* NIL) '(LOGOUT)) 
(ADD-INITIALIZATION 'CLEAR-*DEFAULT-BASE* '(SETQ *DEFAULT-BASE* NIL) '(BEFORE-COLD)) 

(DEFVARIABLE *DISCARD-UNDO-INFO-ON-SAVING* T :BOOLEAN
  "T => when a buffer is saved, throw away its Undo information.") 

(DEFVARIABLE *CHECK-UNBALANCED-PARENTHESES-WHEN-SAVING* NIL :BOOLEAN
  "T => when saving a Lisp mode buffer, check for unbalanced parentheses.") 

(DEFVARIABLE *DEFAULT-ZMACS-BUG-TEMPLATE* NIL :ANYTHING
  "NIL or else name of template to invoke by default when sending bug reports in ZMACS.") 

(DEFVARIABLE *HISTORY-MENU-LENGTH* 24 :FIXNUM
  "Number of history elements to display by default in C-0 C-Y, etc.") 

(DEFVARIABLE *HISTORY-YANK-WRAPAROUND* T :BOOLEAN
  "T means Meta-Y will wrap from end of history to beginning, or vice versa.") 

(DEFVARIABLE *HISTORY-ROTATE-IF-NUMERIC-ARG* NIL :BOOLEAN
  "T means 1 as arg to yank command uses the history's current yank pointer.
Newer elements have negative numbers.") 

(DEFVARIABLE *TYPE-HOG-ALIST* '((:XFASL 2 :NON-CONTIG-DOESNT-MATTER)
				(:QFASL 2 :NON-CONTIG-DOESNT-MATTER)
				(:XLD 2 :NON-CONTIG-DOESNT-MATTER))
	     :ANYTHING
  "A list, each entry of which is a list (TYPE N FLAG).
TYPE is a :canonical-type, or at least a type in all caps.
N is the number of most-recent versions to keep for files of that :canonical-type.
FLAG, if :NON-CONTIG-DOESNT-MATTER, means that gaps in the sequence of versions don't
  matter:  only the last N will be kept, whether gaps must be skipped or not.") 

(DEFVARIABLE *NEXT-SCREEN-CONTEXT-LINES* 1 :FIXNUM
  "The number of lines of overlap for Next Screen and Previous Screen commands") 

(DEFVARIABLE *KILL-OR-SAVE-BUFFERS-MAX-LINES* 50 :FIXNUM
  "Maximum number of lines in the Kill Or Save Buffers menu.") 

(DEFVARIABLE *UPDATE-PLIST-ON-WRITE-OK* T :BOOLEAN
  "Non-NIL enables adding Base attributes to files missing them.") 

(DEFVARIABLE *HELP-PROMPTS-IN-OLD-STYLE* NIL :BOOLEAN
  "T    => Help prompts somewhat like the old style, in the echo area.
NIL  => Help prompts by listing the options in the typeout window.") 

(DEFVARIABLE *APROPOS-SEARCH-OTHER-COMTABS* :ASK :KEYWORD
  "Decides whether Apropos should search other comtabs for the given string.
Appropriate values are :ALWAYS, :NEVER, and :ASK.") 

(DEFVARIABLE *VIEWING-RETURN-IF-NO-MORE-DEFAULT* () :BOOLEAN
  "NIL => It is not possible to scroll forward past the end.
  T   => An attempt to scroll forward past the end while viewing will exit.") 

(DEFVARIABLE *VIEWING-IGNORE-NON-COMMAND-CHARACTERS* 'T :BOOLEAN
  "T   => Any character not a viewing command will be ignored.
NIL => Any character not a viewing command will cause the view to be exited,
           then the character will executed.") 

(DEFVARIABLE *VIEWING-SCROLL-FORWARD-SCREEN-CHARACTERS*
	     '(#\c-V #\c- #\SPACE) :ANYTHING
  "These characters will scroll forward one screenfull while viewing.") 

(DEFVARIABLE *VIEWING-SCROLL-BACKWARD-SCREEN-CHARACTERS*
	     '(#\m-V #\c- #\) :ANYTHING
  "These characters will scroll backward one screenfull while viewing.") 

(DEFVARIABLE *VIEWING-SCROLL-FORWARD-LINE-CHARACTERS*
	     '(#\c-N #\) :ANYTHING
  "These characters will scroll forward one line while viewing.") 

(DEFVARIABLE *VIEWING-SCROLL-BACKWARD-LINE-CHARACTERS*
	     '(#\c-P #\) :ANYTHING
  "These characters will scroll backward one line while viewing.") 

(DEFVARIABLE *VIEWING-EXIT-CHARACTERS* '
	     (#\RUBOUT #\ABORT #\END) :ANYTHING
  "These characters will exit viewing, but have no other effect.") 

(DEFVARIABLE *VIEWING-HELP-CHARACTERS* '(#\HELP #\?) :ANYTHING
  "Press one of these characters for help while viewing.") 

(DEFVARIABLE *FIND-PATCH-DEFINITIONS-TOO* T :BOOLEAN
  "T => In Edit Definition, find definitions in patch files as well as source files.") 

(DEFVARIABLE *EDIT-DEFINITION-IS-READ-ONLY* NIL :BOOLEAN
  "NIL => Edit Definition puts its file into an unprotected buffer.
  T => the file is put into a read-only buffer;
      prevents typos accidentally modifying source files;
      read-protection can be turned off with Toggle Read Only")

;;; This variable will be added to Profile for release 3, for the folks who do not  
;;; like line-boxing in Dired. Its value is seen by the (zwei :mouse-moves) method.
;;; Added 9-10-86 by rpm.
(DEFVARIABLE *DIRED-LINE-BOXING* T :BOOLEAN
  "T => In Dired, draw a box around the line the mouse is on.")


(DEFMACRO WITH-BORROWED-FONT-MAP-OF ((FONT-MAP-OR-WINDOW RECIPIENT) &BODY BODY)	;gsl 3-14-85
  "Execute BODY with the recipient window's font map temporarily replaced by font-map-or-window,
if it is a font-map, or the font-map of font-map-or-window if it is a window, the usual case."
  `(LET* ((OLD-FONT-MAP (SEND ,RECIPIENT :FONT-MAP))
	  (OLD-CURRENT-FONT (SEND ,RECIPIENT :CURRENT-FONT))
	  (WINDOW ,FONT-MAP-OR-WINDOW)
	  (NEW-FONT-MAP (COND ((TYPEP WINDOW 'ARRAY) ;; may 04/13/89 was :array
			       WINDOW)
			      ((AND (TYPEP WINDOW 'INSTANCE) ;; may 04/13/89 was :instance
				    (SEND WINDOW :FONT-MAP)))
			      (T
			       (BARF)))))
     (UNWIND-PROTECT
	 (PROGN (SEND ,RECIPIENT :SET-FONT-MAP NEW-FONT-MAP)
		(SEND ,RECIPIENT :SET-CURRENT-FONT (AREF NEW-FONT-MAP 0))
		,@BODY)
       (SEND ,RECIPIENT :SET-FONT-MAP OLD-FONT-MAP)
       (SEND ,RECIPIENT :SET-CURRENT-FONT OLD-CURRENT-FONT)))) 

(DEFMACRO WITH-TYPEOUT-FONT-MAP-OF ((FONT-MAP-OR-WINDOW) &BODY BODY)	;gsl 12/3/83
  "Execute BODY with the typeout window's font map temporarily replaced by font-map-or-window,
if it is a font-map, or the font-map of font-map-or-window if it is a window, the usual case."
  `(WITH-BORROWED-FONT-MAP-OF (,FONT-MAP-OR-WINDOW *TYPEOUT-WINDOW*) ,@BODY)) 

;; The next macro is from patch 94.164. ddd/gsl  3/1/84.  
;; Moved from ZWEI;COMH by Ellen  7/18/84
(DEFMACRO MULTI-BEEP (&REST TONE-DURATION-ALIST)
  "Special form, takes unquoted lists of pitch and duration or just numbers for pitch.
Duration defaults to last duration used or 2500000 if none previous this call.
Without an arg, the default beep is given.
[e.g. (multi-beep (6000. 5000.) 12000. 24000.)]"
  `(LET ((TSA ',TONE-DURATION-ALIST)
	 NEW-TSA
	 (LAST-DURATION 11422640))
     (IF (NULL TSA)
	 (BEEP)
	 (DO ((TSA TSA (CDR TSA))
	      WAVELENGTH
	      DURATION
	      PAIR)
	     ((NULL TSA))
	   (SETQ PAIR (CAR TSA))
	   (IF (ATOM PAIR)
	       (SETQ PAIR (LIST PAIR LAST-DURATION))
	       (IF (= (LENGTH PAIR) 1)
		   (RPLACA PAIR (LIST LAST-DURATION))))
	   (SETQ LAST-DURATION (CADR PAIR))
	   (SETQ WAVELENGTH (TRUNCATE 46113200 (ABS (CAR PAIR))))
	   (SETQ DURATION (* 144 (ABS (CADR PAIR))))
	   (SETQ NEW-TSA (APPEND NEW-TSA (LIST (LIST WAVELENGTH DURATION)))))
	 (DOLIST (PAIR NEW-TSA)
	   (SYSTEM:%BEEP (CAR PAIR) (CADR PAIR)))))) 

;;;
;;;  Define a command synonym with the same documentation that calls the
;;;  command it's synonymed to.
(DEFMACRO DEFINE-COMMAND-SYNONYM (SYNONYM ORIGINAL-COMMAND)
  "Define SYNONYM as a synonym for ORIGINAL-COMMAND."
  `(PROGN 'COMPILE
	  (DEFPROP ,SYNONYM ,ORIGINAL-COMMAND ORIGINAL-COMMAND)
	  (COMMAND-DEFINE ',SYNONYM
			  'DOCUMENT-SYNONYM
			  NIL)
	  (DEFUN ,SYNONYM ()
	    (,ORIGINAL-COMMAND)))) 

(DEFUN DOCUMENT-SYNONYM (COMMAND CHAR OP)
  (CASE OP
	(:NAME
	 (GET COMMAND 'COMMAND-NAME))
	(:SHORT
	 (PRINT-DOC :SHORT (GET COMMAND 'ORIGINAL-COMMAND) CHAR))
	(:FULL
	 (PRINT-DOC :FULL (GET COMMAND 'ORIGINAL-COMMAND) CHAR)
	 (FORMAT T "~&This command is a synonym for ~A.~&"
		 (COMMAND-NAME (GET COMMAND 'ORIGINAL-COMMAND))))
	(OTHERWISE
	 (FERROR NIL "Bad op ~A" OP)))) 

;;;
;;;  Define an autoloading command, a stub command that loads its full definition
;;;  from a file, then calls itself.
(DEFMACRO DEFINE-AUTOLOADING-COMMAND (NAME DOC SUPPORT-TYPE FILENAME)
  `(PROGN 'COMPILE
	  (DEFPROP ,NAME ,SUPPORT-TYPE AUTOLOADING-SUPPORT-TYPE)
	  (DEFPROP ,NAME ,FILENAME AUTOLOADING-FILENAME)
	  (DEFCOM ,NAME
		  ,(STRING-APPEND DOC
				  #\NEWLINE
				  (LET ((SUPPORT-STRING (GET SUPPORT-TYPE 'AUTOLOADING-SUPPORT-STRING)))
				    (IF (NULL SUPPORT-STRING)
					(FERROR NIL "Unknown support-type: ~A" SUPPORT-TYPE)
					SUPPORT-STRING)))
		  ()
	    (FORMAT *QUERY-IO* "~&~A"
			   (GET (GET ',NAME 'AUTOLOADING-SUPPORT-TYPE) 'AUTOLOADING-SUPPORT-STRING))
	    ,(WHEN (EQ SUPPORT-TYPE :BROKEN)
	       '(UNLESS (Y-OR-N-P "~&Load its definition anyway? ")
		  (BARF)))
	    (LET ((PATHNAME (FS:PARSE-PATHNAME (GET ',NAME 'AUTOLOADING-FILENAME)))
		  (OLD-DEFINITION (FDEFINITION ',NAME)))
	      (FORMAT *QUERY-IO* "~&Loading its definition from ~A." PATHNAME)
	      (LET ((INHIBIT-FDEFINE-WARNINGS T))
		(LOAD PATHNAME :VERBOSE NIL :SET-DEFAULT-PATHNAME NIL))
	      (IF (EQ (FDEFINITION ',NAME) OLD-DEFINITION)
		  (FERROR NIL "Loading the file ~A failed to redefine the autoloading function ~A."
			  PATHNAME ',NAME)
		  (FUNCALL (FDEFINITION ',NAME))))))) 

(DEFPROP :UNSUPPORTED
	 "This command is not currently supported by TI, but it works in most situations."
	 AUTOLOADING-SUPPORT-STRING) 
(DEFPROP :COMPATIBILITY "This command is only supported for compatibility with older systems."
	 AUTOLOADING-SUPPORT-STRING) 
(DEFPROP :BROKEN "This command is not currently supported by TI, and is probably broken."
	 AUTOLOADING-SUPPORT-STRING) 
(DEFPROP :LIBRARY
	 ;;  Per Steve Krueger's suggestion.
	 "This command is supported by TI, but to save space is not built into the band."
	 AUTOLOADING-SUPPORT-STRING) 

;;;
;;;  New macro to be used to assure the proper context for commands.
;;;  May be the single most useful new macro in zwei.
(DEFMACRO CONTEXT-CHECK (CONTEXT-INDICATOR . REST)
  "If context-indicator is a function it is run and passed any args which may follow.
If context-indicator is a symbol then boundp is run on it.
If the test returns nil, barf."
  `(LET ((CI ,CONTEXT-INDICATOR))
     (UNLESS
       (COND ((OR (FUNCTIONP CI T)
		  (INSTANCEP CI))
	      (FUNCALL CI ,@REST))
	     ((SYMBOLP CI)
	      (BOUNDP CI))
	     (T
	      (BARF "Internal context checking error using ~a." CI)))
       (BARF "This command is inappropriate in the current context."))))

;; may 02/15/89 Added doc string - still a kludge, though.
;; added for release 3.2 ft line and window must be bound
(DEFMACRO font-char-width (ind)
  "Internal macro for use in HIGHLIGHTED-STRING-UNDER-MOUSE and HIGHLIGHTED-STRING-UNDER-MOUSE-DLE"
  `(OR (AND ft (AREF ft (CHAR-CODE (AREF line ,ind))))
       (tv:font-char-width (AREF (w:sheet-font-map window)
                                 (CHAR-FONT (AREF line ,ind))))))