;;;-*- Mode:Common-Lisp; Package:ZWEI; Base:8 -*-

;;;                           RESTRICTED RIGHTS LEGEND

;;;Use, duplication, or disclosure by the Government is subject to
;;;restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
;;;Technical Data and Computer Software clause at 52.227-7013.
;;;
;;;                     TEXAS INSTRUMENTS INCORPORATED.
;;;                              P.O. BOX 2909
;;;                           AUSTIN, TEXAS 78769
;;;                                 MS 2151
;;;
;;; Copyright (C) 1984-1989 Texas Instruments Incorporated. All rights reserved.
;;; Copyright (C) 1980, Massachusetts Institute of Technology

;;; Functions and commands for doing searching

(DEFUN SEARCH (BP STRING &OPTIONAL REVERSEP FIXUP-P LINES-TO-SEARCH LIMIT-BP &AUX NLPOS)
  "Search for STRING starting at BP.  Return the bp to where it is found, or NIL.
The value points after the string found if searching forward.
REVERSEP non-NIL means search back before BP; then the value points before the occurrence.
FIXUP-P non-NIL means return beginning or end of buffer if search fails.
If LIMIT-BP is reached, that is considered failure.
LINES-TO-SEARCH non-NIL means give up after searching that many lines;
 then the first value is where to resume searching,
 and the second value is T."
  (OR LIMIT-BP (SETQ LIMIT-BP (IF REVERSEP
				  (INTERVAL-FIRST-BP *INTERVAL*)
				  (INTERVAL-LAST-BP *INTERVAL*))))
  (BLOCK SEARCH
    (RETURN-FROM SEARCH
      (COND ((CHARACTERP STRING)
	     (IF (NOT REVERSEP)
		 ;; Forward search for a character.
		 (CHARMAP-PER-LINE (BP LIMIT-BP (IF FIXUP-P (COPY-BP LIMIT-BP) NIL))
				   ((AND LINES-TO-SEARCH
					 (ZEROP (SETQ LINES-TO-SEARCH (1- LINES-TO-SEARCH)))
					 (RETURN-FROM SEARCH (CREATE-BP LINE *FIRST-INDEX*) T)))
		   (AND (IF *ALPHABETIC-CASE-AFFECTS-SEARCH*
			    (CHAR= STRING (CHARMAP-CHAR))
			    (CHAR-EQUAL STRING (CHARMAP-CHAR)))
			(CHARMAP-RETURN (CHARMAP-BP-AFTER))))
		 ;; Reverse search for a character.
		 (RCHARMAP-PER-LINE (BP LIMIT-BP (IF FIXUP-P (COPY-BP LIMIT-BP) NIL))
				    ((AND LINES-TO-SEARCH
					  (ZEROP (SETQ LINES-TO-SEARCH (1- LINES-TO-SEARCH)))
					  (RETURN-FROM SEARCH (END-OF-LINE LINE) T)))
		   (AND (IF *ALPHABETIC-CASE-AFFECTS-SEARCH*
			    (CHAR= STRING (RCHARMAP-CHAR))
			    (CHAR-EQUAL STRING (RCHARMAP-CHAR)))
			(RCHARMAP-RETURN (RCHARMAP-BP-BEFORE))))))
	    ((SETQ NLPOS (POSITION #\NEWLINE (THE STRING (STRING STRING)) :TEST #'CHAR-EQUAL))
	     (SEARCH-CR-FULL BP STRING REVERSEP FIXUP-P NLPOS LINES-TO-SEARCH LIMIT-BP))
	    (T
	     (IF (NOT REVERSEP)
		 ;; Search forward for a 1-line string.
		 (DO ((LINE (BP-LINE BP) (LINE-NEXT LINE))
		      (LEN (LENGTH STRING))
		      (FROM-INDEX (BP-INDEX BP) 0)
		      (LAST-LINE (BP-LINE LIMIT-BP)))
		     ((AND LINES-TO-SEARCH
			   (ZEROP (SETQ LINES-TO-SEARCH (1- LINES-TO-SEARCH))))
		      (RETURN-FROM SEARCH (CREATE-BP LINE FROM-INDEX) T))
		   (LET ((LASTP (EQ LINE LAST-LINE)))
		     (LET ((INDEX (AND (> (LENGTH LINE) FROM-INDEX)
				       (GLOBAL:STRING-SEARCH STRING LINE
						             FROM-INDEX NIL
                                                             0 NIL
                                                             *ALPHABETIC-CASE-AFFECTS-SEARCH*))))
		       (COND ((AND (NOT (NULL INDEX))
				   (OR (NOT LASTP)
				       (<= INDEX (- (BP-INDEX LIMIT-BP) LEN))))
			      (RETURN (CREATE-BP LINE (+ INDEX LEN))))
			     (LASTP
			      (RETURN (IF FIXUP-P (COPY-BP LIMIT-BP) NIL)))))))
		 ;; Search backward for a 1-line string.
		 (DO ((LINE (BP-LINE BP) (LINE-PREVIOUS LINE))
		      (LEN (LENGTH STRING))
		      (FROM-INDEX (BP-INDEX BP) NIL)
		      (FIRST-LINE (BP-LINE LIMIT-BP)))
		     ((AND LINES-TO-SEARCH
			   (ZEROP (SETQ LINES-TO-SEARCH (1- LINES-TO-SEARCH))))
		      (RETURN-FROM SEARCH (END-OF-LINE LINE) T))
		   (LET ((FIRSTP (EQ LINE FIRST-LINE)))
		     (LET ((INDEX (GLOBAL:STRING-REVERSE-SEARCH STRING LINE
                                                                FROM-INDEX 0
                                                                0 NIL
                                                                *ALPHABETIC-CASE-AFFECTS-SEARCH*)))
		       (COND ((AND (NOT (NULL INDEX))
				   (OR (NOT FIRSTP)
				       (>= INDEX (- (BP-INDEX LIMIT-BP) LEN))))
			      (RETURN (CREATE-BP LINE INDEX)))
			     (FIRSTP
			      (RETURN (IF FIXUP-P (COPY-BP LIMIT-BP) NIL)))))))))))
    NIL)) 

;;; Subroutine of SEARCH.  Used to search for a string containing a CR.
;;; NLPOS is the index in STRING of the first CR.
(DEFUN SEARCH-CR-FULL (BP STRING REVERSEP FIXUP-P NLPOS LINES-TO-SEARCH LIMIT-BP)
  (BLOCK SEARCH-CR-FULL
    (PROG ((STRING-LENGTH (LENGTH STRING)))
	  (RETURN
	    (IF (NOT REVERSEP)
		(LET ((CHAR1 (AREF STRING 0))
		      (CRLEADS (ZEROP NLPOS)))
		  (BLOCK LUPO
		    (DO ((LINE (IF (AND (NOT CRLEADS)
					(>= (- (LINE-LENGTH (BP-LINE BP)) (BP-INDEX BP))
					    NLPOS))
				   (BP-LINE BP)
				   (LINE-NEXT (BP-LINE BP)))
			       (LINE-NEXT LINE))
			 (LAST-LINE (BP-LINE LIMIT-BP)))
			((OR (NULL LINE)
			     (AND (EQ LINE LAST-LINE)
				  (NOT CRLEADS)))
			 ;; Since string contains a NEWLINE, it cannot start on LAST-LINE.
			 ;; Unless the NEWLINE is at the beginning.
			 (IF FIXUP-P (COPY-BP LIMIT-BP) NIL))
		      (AND LINES-TO-SEARCH
			   (ZEROP (SETQ LINES-TO-SEARCH (1- LINES-TO-SEARCH)))
			   (RETURN-FROM SEARCH-CR-FULL (BEG-OF-LINE LINE) T))
		      (LET ((START-INDEX (IF CRLEADS 0 (- (LINE-LENGTH LINE) NLPOS))))
			(COND ((OR CRLEADS
				   (AND (>= START-INDEX 0)
					(IF *ALPHABETIC-CASE-AFFECTS-SEARCH*
					    (CHAR= CHAR1 (AREF LINE START-INDEX))
					    (CHAR-EQUAL CHAR1 (AREF LINE START-INDEX)))))
			       (LET ((I (IF CRLEADS 1 0)))
				 (CHARMAP ((CREATE-BP LINE START-INDEX)
					   LIMIT-BP
					   (IF (>= I STRING-LENGTH)
					       (RETURN-FROM LUPO (CHARMAP-BP-BEFORE))))
				   (IF (>= I STRING-LENGTH)
				       (RETURN-FROM LUPO (CHARMAP-BP-BEFORE)))
				   (UNLESS (IF *ALPHABETIC-CASE-AFFECTS-SEARCH*
					       (CHAR= (CHARMAP-CHAR) (AREF STRING I))
					       (CHAR-EQUAL (CHARMAP-CHAR) (AREF STRING I)))
				     (CHARMAP-RETURN NIL))
				   (SETQ I (1+ I)))))))
		      (AND CRLEADS
			   (EQ LINE LAST-LINE)
			   (RETURN (IF FIXUP-P (COPY-BP LIMIT-BP) NIL))))))
		(PROGN
		  (SETQ NLPOS (POSITION #\NEWLINE (THE STRING (STRING STRING))
					:FROM-END T
					:TEST #'CHAR-EQUAL))
		  (LET ((CHARL (AREF STRING (1- STRING-LENGTH)))
			;; One less than number of chars after the last CR.
			(START-INDEX (- STRING-LENGTH NLPOS 2)))
		    (BLOCK LUPO
		      (DO ((LINE (IF (> (BP-INDEX BP) START-INDEX)
				     (BP-LINE BP)
				     (LINE-PREVIOUS (BP-LINE BP)))
				 (LINE-PREVIOUS LINE))
			   (CRTRAILS (CHAR-EQUAL CHARL #\NEWLINE))
			   (FIRST-LINE (BP-LINE LIMIT-BP)))
			  ((OR (NULL LINE) (EQ LINE FIRST-LINE))
			   (IF FIXUP-P (COPY-BP LIMIT-BP) NIL))
			(AND LINES-TO-SEARCH
			     (ZEROP (SETQ LINES-TO-SEARCH (1- LINES-TO-SEARCH)))
			     (RETURN-FROM SEARCH-CR-FULL (BEG-LINE (CREATE-BP LINE 0) 1) T))
			(COND ((OR CRTRAILS
				   (AND (> (LINE-LENGTH LINE) START-INDEX)
					(IF *ALPHABETIC-CASE-AFFECTS-SEARCH*
					    (CHAR= CHARL (AREF LINE START-INDEX))
					    (CHAR-EQUAL CHARL (AREF LINE START-INDEX)))))
			       (LET ((I (1- STRING-LENGTH)))
				 (RCHARMAP ((CREATE-BP LINE (1+ START-INDEX))
					    LIMIT-BP
					    (IF (< I 0)
						(RETURN-FROM LUPO (RCHARMAP-BP-AFTER))))
				   (IF (< I 0)
				       (RETURN-FROM LUPO (RCHARMAP-BP-AFTER)))
				   (IF (IF *ALPHABETIC-CASE-AFFECTS-SEARCH*
					   (CHAR= (RCHARMAP-CHAR) (AREF STRING I))
					   (CHAR-EQUAL (RCHARMAP-CHAR) (AREF STRING I)))
				       (SETQ I (1- I))
				       (RCHARMAP-RETURN NIL))))))))))))))) 

;;; Word search infernal function
(DEFUN WORD-SEARCH (BP KEY &OPTIONAL REVERSE-P FIXUP-P LIMIT-BP &AUX LEN KEY1 LEN1
		    (ALPHABETIC-CASE-AFFECTS-STRING-COMPARISON *ALPHABETIC-CASE-AFFECTS-SEARCH*))
  "Like SEARCH but does a word-wise search.
That is, each word in KEY need only be an abbreviation of a word in the buffer
for a match to occur.  KEY is a string.
The remaining args are as for SEARCH."
  (OR LIMIT-BP
      (SETQ LIMIT-BP (IF REVERSE-P
			 (INTERVAL-FIRST-BP *INTERVAL*)
			 (INTERVAL-LAST-BP *INTERVAL*))))
  ;; Discard any delimiter chars at the front of KEY.
  ;; They would make the code below blow up.
  (SETQ LEN (LENGTH KEY))
  (DO ((I 0 (1+ I)))
      ((OR (>= I LEN)
	   (= (WORD-SYNTAX (AREF KEY I)) WORD-ALPHABETIC))
       (UNLESS (ZEROP I)
	 (SETQ KEY (NSUBSTRING KEY I)))))
  (SETQ LEN (LENGTH KEY))
  ;; KEY1 gets the first word of KEY, and LEN1 gets KEY1's length:
  (DO ((I 0 (1+ I)))
      ((OR (>= I LEN)
	   (/= (WORD-SYNTAX (AREF KEY I)) WORD-ALPHABETIC))
       (SETQ LEN1 I
	     KEY1 (NSUBSTRING KEY 0 I))))
  (COND ((ZEROP LEN)
	 (WHEN FIXUP-P BP))
	((NOT REVERSE-P)
	 (BLOCK LINES
	   (DO ((LINE (BP-LINE BP) (LINE-NEXT LINE))
		(LIMIT (BP-LINE LIMIT-BP))
		(IDX (BP-INDEX BP) 0))
	       (NIL)
	     (BLOCK PER-LINE
	       (DO ((LLEN (LINE-LENGTH LINE)))
		   (NIL)
		 (OR (SETQ IDX (GLOBAL:STRING-SEARCH KEY1 LINE IDX NIL 0 NIL 	;Find first word in line
                                                     *ALPHABETIC-CASE-AFFECTS-SEARCH*))
		     (RETURN NIL))
		 (BLOCK MATCH-REMAINING-WORDS
		   (DO ((I IDX)				;I index of character in line
			(J LEN1)			;J index in search key
			(LINE1 LINE)			;Copy these in case we advance to next line
			(LLEN1 LLEN))
		       (NIL)
		     (DO NIL (NIL)
		       (AND (OR (>= I LLEN1)
				(/= (WORD-SYNTAX (AREF LINE1 I)) WORD-ALPHABETIC))
			    (RETURN NIL))
		       (SETQ I (1+ I)))
		     (DO NIL (NIL)
		       (COND ((>= J LEN)
			      (RETURN-FROM LINES (CREATE-BP LINE1 I)))	;Point after last word
			     ((= (WORD-SYNTAX (AREF KEY J)) WORD-ALPHABETIC)
			      (RETURN NIL)))
		       (SETQ J (1+ J)))
		     (DO NIL (NIL)
		       (COND ((>= I LLEN1)		;This line used up, advance to next
			      (AND (EQ LINE1 LIMIT)
				   (RETURN-FROM MATCH-REMAINING-WORDS))
			      (SETQ LINE1 (LINE-NEXT LINE1)
				    LLEN1 (LINE-LENGTH LINE1)
				    I 0))
			     ((= (WORD-SYNTAX (AREF LINE1 I)) WORD-ALPHABETIC)
			      (RETURN NIL))
			     (T
			      (SETQ I (1+ I)))))
		     (DO ((CH1 (AREF LINE1 I))
			  (CH2 (AREF KEY J)))
			 (NIL)
		       (COND
			 ((/= (WORD-SYNTAX CH2) WORD-ALPHABETIC)	;key can be shorter than data
			  (RETURN NIL))			;allowing word abbreviation
			 ((NOT (IF *ALPHABETIC-CASE-AFFECTS-SEARCH*
				   (CHAR= CH1 CH2)
				   (CHAR-EQUAL CH1 CH2)))
			  (RETURN-FROM MATCH-REMAINING-WORDS)))
		       (SETQ CH1 (IF (>= (SETQ I (1+ I)) LLEN1)
				     #\NEWLINE
				     (AREF LINE1 I)))
		       (SETQ CH2 (IF (>= (SETQ J (1+ J)) LEN)
				     #\NEWLINE
				     (AREF KEY J)))))
		   ;; Space forward in line to end of this word
		   ;; Space forward in key to start of next word
		   ;; If key exhausted, the search succeeds
		   ;; Space forward in line to start of next word.  This may actually
		   ;; be on the next line.
		   ;; Check that these two words match
		   )
		 (SETQ IDX (1+ IDX))))
	     (AND (EQ LINE LIMIT)
		  (RETURN (AND FIXUP-P (COPY-BP LIMIT-BP)))))	;search for next instance of first word
	   ))
	(T
	 (BLOCK LINES
	   (DO ((LINE (BP-LINE BP) (LINE-PREVIOUS LINE))
		(LIMIT (BP-LINE LIMIT-BP))
		(FORWARD-LIMIT (BP-LINE (INTERVAL-LAST-BP *INTERVAL*)))
		(IDX (BP-INDEX BP) NIL))
	       (NIL)
	     (BLOCK PER-LINE
	       (DO ((LLEN (LINE-LENGTH LINE)))
		   (NIL)
		 (OR (SETQ IDX (GLOBAL:STRING-REVERSE-SEARCH KEY1 LINE IDX 0 0 NIL
                                                             *ALPHABETIC-CASE-AFFECTS-SEARCH*))
		     (RETURN NIL))
		 (BLOCK MATCH-REMAINING-WORDS
		   (DO ((I IDX)				;I index of character in line
			(J LEN1)			;J index in search key
			(LINE1 LINE)			;Copy these in case we advance to next line
			(LLEN1 LLEN))
		       (NIL)
		     (DO () (NIL)
		       (AND (OR (>= I LLEN1)
				(/= (WORD-SYNTAX (AREF LINE1 I)) WORD-ALPHABETIC))
			    (RETURN NIL))
		       (SETQ I (1+ I)))
		     (DO () (NIL)
		       (COND ((>= J LEN)
			      (RETURN-FROM LINES (CREATE-BP LINE IDX)))	;Point before first word
			     ((= (WORD-SYNTAX (AREF KEY J)) WORD-ALPHABETIC)
			      (RETURN NIL)))
		       (SETQ J (1+ J)))
		     (DO () (NIL)
		       (COND ((>= I LLEN1)		;This line used up, advance to next
			      (AND (EQ LINE1 FORWARD-LIMIT)
				   (RETURN-FROM MATCH-REMAINING-WORDS))
			      (SETQ LINE1 (LINE-NEXT LINE1)
				    LLEN1 (LINE-LENGTH LINE1)
				    I 0))
			     ((= (WORD-SYNTAX (AREF LINE1 I)) WORD-ALPHABETIC)
			      (RETURN NIL))
			     (T
			      (SETQ I (1+ I)))))
		     (DO ((CH1 (AREF LINE1 I))
			  (CH2 (AREF KEY J)))
			 (NIL)
		       (COND ((/= (WORD-SYNTAX CH2) WORD-ALPHABETIC)	;key can be shorter than data
			      (RETURN NIL))		                ;allowing word abbreviation
			     ((NOT (IF *ALPHABETIC-CASE-AFFECTS-SEARCH*
				       (CHAR= CH1 CH2)
				       (CHAR-EQUAL CH1 CH2)))
			      (RETURN-FROM MATCH-REMAINING-WORDS)))
		       (SETQ CH1 (IF (>= (SETQ I (1+ I)) LLEN1)
				     #\NEWLINE
				     (AREF LINE1 I)))
		       (SETQ CH2 (IF (>= (SETQ J (1+ J)) LEN)
				     #\NEWLINE
				     (AREF KEY J)))))
		   ;; Space forward in line to end of this word
		   ;; Space forward in key to start of next word
		   ;; If key exhausted, the search succeeds
		   ;; Space forward in line to start of next word.  This may actually
		   ;; be on the next line.
		   ;; Check that these two words match
		   )
		 (SETQ IDX (+ IDX (LENGTH KEY1) -1))))
	     (AND (EQ LINE LIMIT)
		  (RETURN (AND FIXUP-P (COPY-BP LIMIT-BP))))))))) 

(DEFUN SEARCH-SET (BP LIST &OPTIONAL REVERSEP FIXUP-P LIMIT-BP &AUX CH)
  "Like SEARCH but searches for a character which is in LIST, a list of characters."
  (OR LIMIT-BP
      (SETQ LIMIT-BP (IF REVERSEP
			 (INTERVAL-FIRST-BP *INTERVAL*)
			 (INTERVAL-LAST-BP *INTERVAL*))))
  (IF (NOT REVERSEP)
      (CHARMAP (BP LIMIT-BP (AND FIXUP-P LIMIT-BP))
	(AND (MEMBER (SETQ CH (CHARMAP-CH-CHAR)) LIST :TEST #'EQ)
	     (CHARMAP-RETURN (CHARMAP-BP-AFTER) CH)))
      (RCHARMAP (BP LIMIT-BP (AND FIXUP-P LIMIT-BP))
	(AND (MEMBER (SETQ CH (RCHARMAP-CH-CHAR)) LIST :TEST #'EQ)
	     (RCHARMAP-RETURN (RCHARMAP-BP-BEFORE) CH))))) 

(DEFUN DELIMITED-SEARCH (BP STRING &OPTIONAL REVERSE-P FIXUP-P LIMIT-BP)
  "Like SEARCH but accepts only occurrences of STRING with delimiters on both sides."
  (SETQ STRING (IF (EQ STRING *LAST-DELIMITED-SEARCH-STRING*)
		   *LAST-DELIMITED-SEARCH-STRING-WITH-DELIMITERS*
		   (PROGN
		     (SETQ *LAST-DELIMITED-SEARCH-STRING* STRING)
		     (LET* ((LENGTH (LENGTH STRING))
			    (NEW-STRING (MAKE-ARRAY (+ LENGTH 2) :ELEMENT-TYPE '(UNSIGNED-BYTE 20))))
		       (SETF (AREF NEW-STRING 0) 20005)
		       (COPY-ARRAY-PORTION STRING 0 LENGTH NEW-STRING 1 (1+ LENGTH))
		       (SETF (AREF NEW-STRING (1+ LENGTH)) 20005)
		       (SETQ *LAST-DELIMITED-SEARCH-STRING-WITH-DELIMITERS* (CONS NEW-STRING NIL))))))
  (SETQ BP (FSM-SEARCH BP STRING REVERSE-P FIXUP-P NIL LIMIT-BP))
  (AND BP (IF REVERSE-P (IBP BP) (DBP BP)))) 

(DEFUN SEARCH-RING-PUSH (STRING FCN)
  "Push an entry onto the ring of saved search strings.
STRING and FCN go into the entry."
  (AND (NUMBERP (CAAR *SEARCH-RING*))
       (SETQ *SEARCH-RING* (CDR *SEARCH-RING*)))
  (PUSH (LIST STRING FCN) *SEARCH-RING*)
  (DO ()
      ((<= (LENGTH *SEARCH-RING*) *SEARCH-RING-MAX*))
    (DELETE-LAST-ELEMENT *SEARCH-RING*))) 

(DEFUN SEARCH-RING-POP (&AUX KEY FCN)
  "Pop an entry off the ring of saved search strings, returning two values.
The values are the two arguments given to SEARCH-RING-PUSH to push the entry.
The second is a function (maybe SEARCH), and the first is an argument to give it
\(maybe a string)."
  (COND ((NULL *SEARCH-RING*)
	 (BARF))
	(T
	 (SETQ KEY (CAAR *SEARCH-RING*)
	       FCN (CADAR *SEARCH-RING*))
	 (SETQ *SEARCH-RING* (NCONC (CDR *SEARCH-RING*)
				    (RPLACD *SEARCH-RING* NIL)))
	 (VALUES KEY FCN)))) 

(DEFUN REPLACE-STRING (BP FROM TO &OPTIONAL TIMES)
  "Replace the first TIMES (or all) occurrences of FROM with TO, after BP.
FROM and TO are strings; TIMES is a number or NIL.
Returns the number of occurrences replaced."
  (OR TIMES (SETQ TIMES -1))
  (DO ((I 0 (1+ I))
       (LEN (LENGTH FROM)))
      ((= I TIMES) I)
    (OR (SETQ BP (SEARCH BP FROM))
	(RETURN I))
    (SETQ BP (CASE-REPLACE (FORWARD-CHAR BP (- LEN)) BP TO T)))) 

(DEFUN CHAR-UPPERCASE-P (CHAR)
  "T if CHAR is an upper case letter."
  (UPPER-CASE-P (MAKE-CHAR CHAR))) 

(DEFUN CHAR-LOWERCASE-P (CHAR)
  "T if CHAR is a lower case letter."
  (LOWER-CASE-P (MAKE-CHAR CHAR))) 

(DEFUN CASE-REPLACE (BP1 BP2 TO &OPTIONAL IN-ORDER-P &AUX BP FIRST SECOND)
  "Replace the text from BP1 to BP2 with a copy of TO, preserving case if appropriate.
Lowercase characters in TO are converted to uppercase as needed 
to match the case pattern of the text being replaced.
Uppercase characters in TO always remain uppercase.
Returns a BP to the end of the inserted text.
IN-ORDER-P non-NIL says assume BP2 is after BP1; otherwise swap them if nec.
If *CASE-REPLACE-P* is NIL, we never preserve case."
  (OR IN-ORDER-P (ORDER-BPS BP1 BP2))
  (WITH-BP (BP3 BP1 ':NORMAL)
    (COND ((NOT *CASE-REPLACE-P*)
	   (DELETE-INTERVAL BP1 BP2 T)
	   (INSERT BP3 TO))
	  (T
	   ;; Get the first alphabetic char, and following char.
	   (SETQ BP (COPY-BP BP1))
	   (DO ()
	       ((OR (BP-= BP BP2)
		    (ALPHA-CHAR-P (BP-CH-CHAR BP))))
	     (IBP BP))
	   (SETQ FIRST (IF (BP-= BP BP2)
			   #\0
			   (BP-CHAR BP)))
	   (OR (BP-= BP BP2)
	       (SETQ BP (FORWARD-CHAR BP 1 T)))
	   (SETQ SECOND (IF (BP-= BP BP2)
			    #\0
			    (BP-CHAR BP)))
	   ;; Now do the replacement, leaving BP3 and BP2 denoting the region.
	   (DELETE-INTERVAL BP1 BP2)
	   (SETQ BP2 (INSERT BP3 (IN-CURRENT-FONT TO (CHAR-FONT FIRST))))
	   (COND ((CHAR-UPPERCASE-P FIRST)
		  (COND ((OR (CHAR-UPPERCASE-P SECOND)
			     (AND (NOT (CHAR-LOWERCASE-P SECOND))
				  (GET *MAJOR-MODE* 'ALL-UPPERCASE)))
			 ;; They are both upper case or the second one is not alphabetic
			 ;; and this mode has all uppercase.  Uppercasify the whole thing.
			 (UPCASE-INTERVAL BP3 BP2 T))
			(T
			 ;; Only the first letter is uppercase.  Capitalize on this fact.
			 (DO ()
			     ((OR (BP-= BP3 BP2)
				  (ALPHA-CHAR-P (BP-CH-CHAR BP3))))
			   (IBP BP3))
			 (OR (BP-= BP3 BP2)
			     (UPCASE-CHAR BP3))))))
	   BP2)))) 

(DEFUN YANK-AS-SEARCH-STRING (THING &OPTIONAL IGNORE IGNORE)
  "Yank THING into the buffer at point, moving point after it, leaving mark before it.
   A string is yanked as its text."
  (YANK-AS-TEXT THING (EQ *LAST-COMMAND-TYPE* *FSM-SEARCH-HISTORY*))) 

(DEFUN BUILD-FSM (STRING-LIST &OPTIONAL (CASE-DEPENDENT-P *ALPHABETIC-CASE-AFFECTS-SEARCH*)
		  &AUX NCHARN NSTATES)
  "Return a FSM that can be used to search for the strings in STRING-LIST.
These strings can contain extended search characters for pattern matching.
If CASE-DEPENDENT-P is T, the FSM is built so as not to ignore case.
See the code for FSM-SEARCH to see how to use an FSM."
  (SETQ *FSM-STRING-LIST* STRING-LIST)
  (SETQ NCHARN (BUILD-FSM-CHARACTER-SET STRING-LIST CASE-DEPENDENT-P))
  (SETQ NSTATES (BUILD-FSM-TREE STRING-LIST NCHARN))
  (CLEAN-FSM NSTATES NCHARN)) 

;;; Build up the character translation tables, returns number of character types
(DEFUN BUILD-FSM-CHARACTER-SET (STRING-LIST CASE-DEPENDENT-P &AUX (MAXCHAR 0))
  ;; First pass, get all the alphabetic characters
  (ARRAY-INITIALIZE *FSM-CHARACTER-TABLE* 0)
  (DOLIST (STRING STRING-LIST)
    (DO ((I 0 (1+ I))
	 (LEN (LENGTH STRING))
	 (CH))
	((>= I LEN))
      (OR (LDB-TEST %%FSM-SET (SETQ CH (AREF STRING I)))
	  ;; if fsm-set bit not set
	  (COND ((ZEROP (AREF *FSM-CHARACTER-TABLE* (SETQ CH (CHAR-CODE CH))))
		 (SETF (AREF *FSM-CHARACTER-TABLE* CH) (SETQ MAXCHAR (1+ MAXCHAR)))
		 (AND (NOT CASE-DEPENDENT-P)		      ;if dont care case and alpha, fill both positions
		      (ALPHA-CHAR-P (CODE-CHAR CH))
		      (SETF (AREF *FSM-CHARACTER-TABLE* (LOGXOR CH 40)) MAXCHAR)))))))
  ;; Second pass, get the character types for all the sets mentioned
  (ARRAY-INITIALIZE *FSM-CHARACTER-SET-TABLE-16* 0)
  (LET ((LEN (LENGTH *FSM-SEARCH-SET-LIST*)))
    (IF (OR (NOT (BOUNDP '*FSM-SEARCH-SET-TABLE*))
	    (< (ARRAY-TOTAL-SIZE *FSM-SEARCH-SET-TABLE*) LEN))
	(SETQ *FSM-SEARCH-SET-TABLE* (MAKE-ARRAY LEN))
	(ARRAY-INITIALIZE *FSM-SEARCH-SET-TABLE* NIL)))
  (DOLIST (STRING STRING-LIST)
    (DO ((I 0 (1+ I))
	 (LEN (LENGTH STRING))
	 (CH))
	((>= I LEN))
      (AND (LDB-TEST %%FSM-SET (SETQ CH (AREF STRING I))) ; A set character
	   (NOT (AREF *FSM-SEARCH-SET-TABLE* (SETQ CH (CHAR-CODE CH))))
	   ;;for each character in this set,
	   ;;if it is not already in *fsm-character-table*
	   ;;put a 1 in the *fsm-character-set-table* for this set and character
	   ;;else push characters type (maxchar no.) on list entry for this set in *fsm-search-set-table*
	   (LET ((*LIST* NIL))
	     (MAP-OVER-FSM-SEARCH-SET
	       CH
	       #'(LAMBDA (SET CH &AUX CHARN)
		   (IF (ZEROP (SETQ CHARN (AREF *FSM-CHARACTER-TABLE* CH)))
		       (SETF (AREF *FSM-CHARACTER-SET-TABLE* CH SET) 1)
		       (OR (MEMBER CHARN *LIST* :TEST #'EQ)
			   (PUSH CHARN *LIST*)))))
	     (SETF (AREF *FSM-SEARCH-SET-TABLE* CH) (NREVERSE *LIST*))))))
  ;; Now assign character types for all the set intersections
  (DO ((CH 0 (1+ CH))
       (SET-ALIST NIL)
       (MASK)
       (ENTRY))
      ((>= CH 400))
    (COND ((NOT (ZEROP (SETQ MASK (AREF *FSM-CHARACTER-SET-TABLE-16* CH))))
	   (COND ((NOT (SETQ ENTRY (ASSOC MASK SET-ALIST :TEST #'EQ)))
		  (PUSH (SETQ ENTRY (CONS MASK (SETQ MAXCHAR (1+ MAXCHAR))))
			SET-ALIST)
		  (DO ((SET 0 (1+ SET))
		       (BIT 1 (+ BIT 100)))
		      ((>= SET 20))
		    (AND (LDB-TEST BIT MASK)
			 (PUSH MAXCHAR (AREF *FSM-SEARCH-SET-TABLE* SET))))))
	   (SETF (AREF *FSM-CHARACTER-TABLE* CH) (CDR ENTRY)))))
  ;; Finally return the number of character types
  (1+ MAXCHAR)) 

;;; Apply FUNCTION to all members of a character set,
;;; FUNCTION is called with SET-NUMBER and character
(DEFUN MAP-OVER-FSM-SEARCH-SET (SET-NUMBER FUNCTION &AUX SET)
  "Apply FUNCTION to all members of a character set,
 FUNCTION is caled with SET-NUMBER and character"
  (SETQ SET (NTH SET-NUMBER *FSM-SEARCH-SET-LIST*))
  (IF (AND (ATOM SET) SET)
      (DOTIMES (CH 400)
	(AND (FUNCALL SET CH)
	     (FUNCALL FUNCTION SET-NUMBER CH)))
      (DOLIST (CHAR SET)
	(IF (CHARACTERP CHAR)
	    (FUNCALL FUNCTION SET-NUMBER CHAR)
	    (DO ((CH (CAR CHAR) (1+ CH))
		 (LIM (CADR CHAR)))
		((> CH LIM))
	      (FUNCALL FUNCTION SET-NUMBER CH))))))

(DEFUN FSM-WORD-DELIMITER-CHAR-P (CH)
  (OR (> CH 220) (= (WORD-SYNTAX CH) WORD-DELIMITER)))

(DEFUN FSM-ATOM-DELIMITER-CHAR-P (CH)
  (OR (> CH 220) (= (ATOM-WORD-SYNTAX CH) WORD-DELIMITER)))

(DEFUN FSM-CHAR-TRUE (IGNORE) T)

;;; Build the actual tree from the strings, NCHARN is the number of character types,
;;; Returns the number of states
(DEFUN BUILD-FSM-TREE (STRING-LIST *NCHARN* &AUX (*MAXSTATE* 0))
  (DECLARE (SPECIAL *NCHARN* *MAXSTATE*))
  (DOTIMES (I (ARRAY-DIMENSION *FSM-STATE-TABLE* 0))
    (DOTIMES (J (ARRAY-DIMENSION *FSM-STATE-TABLE* 1))
      (SETF (AREF *FSM-STATE-TABLE* I J) 0)))
  (ARRAY-INITIALIZE *FSM-WORD-TABLE* NIL)
  (DOLIST (STRING STRING-LIST)
    (BUILD-FSM-TREE-1 STRING 0 0 (LENGTH STRING)))
  (1+ *MAXSTATE*)) 

;; Handle a single character
(DEFUN BUILD-FSM-TREE-1 (STRING INDEX STATE LENGTH &AUX CHAR NOT-P STAR-P SET-P CH STAR1-P)
  "Handle a single character, and recursively handle a string thereby."
  (DECLARE (SPECIAL *NCHARN* *MAXSTATE*))
  (COND ((>= INDEX LENGTH)
	 (SETF (AREF *FSM-WORD-TABLE* STATE) STRING))	      ;End of string, save winner
	((= (SETQ CHAR (AREF STRING INDEX)) %FSM-NOOP)
	 (BUILD-FSM-TREE-1 STRING (1+ INDEX) STATE LENGTH))
	(T
	 (SETQ NOT-P (LDB-TEST %%FSM-NOT CHAR)
	       STAR-P (LDB-TEST %%FSM-STAR CHAR)
	       SET-P (LDB-TEST %%FSM-SET CHAR)
	       STAR1-P (LDB-TEST %%FSM-STAR1 CHAR)
	       CH (CHAR-CODE CHAR))
	 (IF SET-P
	     (LET ((SET (AREF *FSM-SEARCH-SET-TABLE* CH)))
	       (IF NOT-P
		   (DOTIMES (NCH *NCHARN*)
		     (OR (MEMBER NCH SET :TEST #'EQ)
			 (BUILD-FSM-TREE-2 STRING INDEX STATE LENGTH NCH STAR-P STAR1-P)))
		   (DOLIST (NCH SET)
		     (BUILD-FSM-TREE-2 STRING INDEX STATE LENGTH NCH STAR-P STAR1-P))))
	     (PROGN
	       (SETQ CH (AREF *FSM-CHARACTER-TABLE* CH))
	       (IF NOT-P
		   (DOTIMES (NCH *NCHARN*)
		     (OR (= NCH CH) (BUILD-FSM-TREE-2 STRING INDEX STATE LENGTH NCH STAR-P STAR1-P)))
		   (BUILD-FSM-TREE-2 STRING INDEX STATE LENGTH CH STAR-P STAR1-P))))))) 

;; Handle a single state transition
(DEFUN BUILD-FSM-TREE-2 (STRING INDEX STATE LENGTH CHARN STAR-P STAR1-P &AUX NEW-STATE)
  "Handle a single state transition."
  (DECLARE (SPECIAL *NCHARN* *MAXSTATE*))
  (AND (OR STAR-P STAR1-P)
       (SETF (AREF *FSM-STATE-TABLE* STATE CHARN)
	     (SETQ NEW-STATE STATE)))
  (AND (OR (NOT STAR-P) STAR1-P)			      ;The usual case is not star-p
       (AND (ZEROP (SETQ NEW-STATE (AREF *FSM-STATE-TABLE* STATE CHARN)))
	    (SETF (AREF *FSM-STATE-TABLE* STATE CHARN)
		  (SETQ NEW-STATE (SETQ *MAXSTATE* (1+ *MAXSTATE*))))
	    (WHEN (>= *MAXSTATE* *FSM-MAX-STATES*)
	      (SETQ *FSM-MAX-STATES* (+ *FSM-MAX-STATES* 764))
	      (ADJUST-ARRAY *FSM-STATE-TABLE* (LIST *FSM-MAX-STATES* 200) :INITIAL-ELEMENT 0)
	      (ADJUST-ARRAY *FSM-WORD-TABLE* *FSM-MAX-STATES*))))
  (BUILD-FSM-TREE-1 STRING (1+ INDEX) NEW-STATE LENGTH))      ;Process rest of string

;;; Clean up the fsm and build up the backpointers
(DEFUN CLEAN-FSM (NSTATES NCHARS)
  (IF (OR (NOT (BOUNDP '*FSM-BACKPOINTER-TABLE*))
	  (< (ARRAY-TOTAL-SIZE *FSM-BACKPOINTER-TABLE*) NSTATES))
      (SETQ *FSM-BACKPOINTER-TABLE* (MAKE-ARRAY NSTATES))
      (ARRAY-INITIALIZE *FSM-BACKPOINTER-TABLE* NIL))
  (DOTIMES (STATE NSTATES)
    (DOTIMES (CH NCHARS)
      (LET ((NEW-STATE (AREF *FSM-STATE-TABLE* STATE CH)))
	(AND (> NEW-STATE STATE)
	     (NULL (AREF *FSM-BACKPOINTER-TABLE* NEW-STATE))
	     (LET ((GCTAIL (LET ((BACKPTR (AREF *FSM-BACKPOINTER-TABLE* STATE)))
			     (IF BACKPTR
				 (AREF *FSM-STATE-TABLE* BACKPTR CH)
				 0))))
	       (SETF (AREF *FSM-BACKPOINTER-TABLE* NEW-STATE) GCTAIL)
	       (OR (AREF *FSM-WORD-TABLE* NEW-STATE)
		   (SETF (AREF *FSM-WORD-TABLE* NEW-STATE)
			 (AREF *FSM-WORD-TABLE* GCTAIL)))
	       (DOTIMES (NCH NCHARS)
		 (AND (ZEROP (AREF *FSM-STATE-TABLE* NEW-STATE NCH))
		      (SETF (AREF *FSM-STATE-TABLE* NEW-STATE NCH)
			    (AREF *FSM-STATE-TABLE* GCTAIL NCH))))))))
    (LET ((BACKPTR (AREF *FSM-BACKPOINTER-TABLE* STATE)))
      (AND (LET ((WORD (AND BACKPTR (AREF *FSM-WORD-TABLE* BACKPTR))))
	     (OR (NULL WORD)
		 (EQ WORD (AREF *FSM-WORD-TABLE* STATE))))
	   (SETF (AREF *FSM-BACKPOINTER-TABLE* STATE)
		 (AND BACKPTR (AREF *FSM-BACKPOINTER-TABLE* BACKPTR))))))) 

(DEFUN FSM-STRING-SEARCH (KEY STRING &OPTIONAL FROM TO
			  &AUX (LEN (LENGTH STRING)) (STRINGS (CAR KEY))
			  (EXPR (CADR KEY)) (CR-P (CADDR KEY)))
  "Search STRING or part of it for a match for KEY.
Mostly like STRING-SEARCH except:
The first element of KEY is a list of strings, and the second
is a filtering function or NIL.  If non-NIL, it is passed a list
of indices within the line, and whatever it returns, we return."
  (OR (EQ STRINGS *FSM-STRING-LIST*)
      (BUILD-FSM STRINGS))
  (OR FROM
      (SETQ FROM (IF CR-P -1 0)))
  (OR TO
      (SETQ TO (IF CR-P (1+ LEN) LEN)))
  (DO ((I FROM (1+ I))
       (LIST NIL)
       (STATE 0)
       (CH))
      ((>= I TO)
       (SETQ LIST (NREVERSE LIST))
       (IF EXPR
	   (FUNCALL EXPR LIST)
	   (CAAR LIST)))
    (SETQ CH (AREF *FSM-CHARACTER-TABLE*
		   (IF (OR (MINUSP I) (>= I LEN))
		       (CHAR-CODE #\NEWLINE)
		       (CHAR-CODE (AREF STRING I))))
	  STATE (AREF *FSM-STATE-TABLE* STATE CH))
    (DO ((STATE STATE (AREF *FSM-BACKPOINTER-TABLE* STATE))
	 (OSTATE NIL STATE)
	 (WORD))
	((OR (NULL STATE) (EQ STATE OSTATE)))
      (OR (SETQ WORD (AREF *FSM-WORD-TABLE* STATE))
	  (RETURN NIL))
      (PUSH (LIST (1+ I) WORD) LIST)))) 

;;; Do fsm search within lines
(DEFUN FSM-SEARCH-WITHIN-LINES (BP KEY &OPTIONAL REVERSE-P FIXUP-P IGNORE LIMIT-BP
				&AUX (START-LINE (BP-LINE BP))
				(START-INDEX (BP-INDEX BP)))
  "Search from BP for a match for KEY.
We look for a match for KEY within each line individually.
Mostly called like SEARCH, but the LINES-TO-SEARCH argument is ignored.
The first element of KEY is a list of strings, and the second
is a filtering function or NIL.
If non-NIL, it is passed a list of indices within the line.
If it returns NIL, there is no match; else it is the index 
 within the line of the match."
  (OR LIMIT-BP
      (SETQ LIMIT-BP (IF REVERSE-P
			 (INTERVAL-FIRST-BP *INTERVAL*)
			 (INTERVAL-LAST-BP *INTERVAL*))))
  (IF (NOT REVERSE-P)
      (DO ((LINE START-LINE (LINE-NEXT LINE))
	   (LAST-LINE (BP-LINE LIMIT-BP))
	   (LAST-INDEX (BP-INDEX LIMIT-BP))
	   (INDEX))
	  (NIL)
	(AND (SETQ INDEX (FSM-STRING-SEARCH KEY LINE
					    (AND (EQ LINE START-LINE) START-INDEX)
					    (AND (EQ LINE LAST-LINE) LAST-INDEX)))
	     (RETURN (CREATE-BP LINE INDEX)))
	(AND (EQ LINE LAST-LINE)
	     (RETURN (AND FIXUP-P LIMIT-BP))))
      (DO ((LINE START-LINE (LINE-PREVIOUS LINE)) ;was line-next gsl.
	   (FIRST-LINE (BP-LINE LIMIT-BP))
	   (FIRST-INDEX (BP-INDEX LIMIT-BP))
	   (INDEX))
	  (NIL)
	(AND (SETQ INDEX (FSM-STRING-SEARCH KEY LINE
					    (AND (EQ LINE FIRST-LINE) FIRST-INDEX)
					    (AND (EQ LINE START-LINE) START-INDEX)))
	     (RETURN (CREATE-BP LINE INDEX)))
	(AND (EQ LINE FIRST-LINE)
	     (RETURN (AND FIXUP-P LIMIT-BP)))))) 

;;; Search by characters
(DEFUN FSM-SEARCH (BP STRINGS &OPTIONAL REVERSE-P FIXUP-P IGNORE LIMIT-BP (STATE 0) &AUX WORD)
  "Search from BP for a match for one of STRINGS.
The match may cross a line boundary.
Mostly called like SEARCH, but the LINES-TO-SEARCH argument is ignored.
The first value is like that of SEARCH.
The second is the string that was found.
The third is the FSM state, which can be given back as the STATE argument
if you search for another occurrence."
  (OR LIMIT-BP
      (SETQ LIMIT-BP (IF REVERSE-P
			 (INTERVAL-FIRST-BP *INTERVAL*)
			 (INTERVAL-LAST-BP *INTERVAL*))))
  (OR (EQ STRINGS *FSM-STRING-LIST*)
      (BUILD-FSM STRINGS))
  (IF REVERSE-P						      ; this test and first clause added. gsl
      (RCHARMAP (BP LIMIT-BP (AND FIXUP-P LIMIT-BP))
	(SETQ STATE (AREF *FSM-STATE-TABLE*
			  STATE
			  (AREF *FSM-CHARACTER-TABLE*
				(IF *ALPHABETIC-CASE-AFFECTS-SEARCH*  ;this look backward. gsl.
				    (CHAR-CODE (RCHARMAP-CH-CHAR))
				    (CHAR-CODE (CHAR-UPCASE (RCHARMAP-CH-CHAR)))))))
	(AND (SETQ WORD (AREF *FSM-WORD-TABLE* STATE))
	     (RCHARMAP-RETURN (RCHARMAP-BP-AFTER) WORD STATE)))
      (CHARMAP (BP LIMIT-BP (AND FIXUP-P LIMIT-BP))
	(SETQ STATE (AREF *FSM-STATE-TABLE*
			  STATE
			  (AREF *FSM-CHARACTER-TABLE*
				(IF *ALPHABETIC-CASE-AFFECTS-SEARCH*  ;this look backward. gsl.
				    (CHAR-CODE (CHARMAP-CH-CHAR))
				    (CHAR-CODE (CHAR-UPCASE (CHARMAP-CH-CHAR)))))))
	(AND (SETQ WORD (AREF *FSM-WORD-TABLE* STATE))
	     (CHARMAP-RETURN (CHARMAP-BP-AFTER) WORD STATE))))) 

(DEFUN FSM-EXPR-SEARCH (BP KEY &OPTIONAL REVERSE-P FIXUP-P IGNORE LIMIT-BP (STATE 0)
			&AUX (STRINGS (CAR KEY)) (EXPR (CADR KEY)) LIST)
  "Search from BP for a match for KEY.
Mostly called like SEARCH, but the LINES-TO-SEARCH argument is ignored.
We look for a match for KEY, which may cross a line boundary.
The first element of KEY is a list of strings, and the second
is a filtering function or NIL.
If non-NIL, it is passed a list of indices within the line.
If it returns NIL, there is no match; else it is the index 
 within the line of the match.
The first value is like that of SEARCH.
The second is the string that was found.
The third is the FSM state, which can be given back as the STATE argument
if you search for another occurrence."
  (OR (EQ STRINGS *FSM-STRING-LIST*) (BUILD-FSM STRINGS))
  (IF REVERSE-P						      ;*** this test and first clause added. gsl
      (RCHARMAP (BP LIMIT-BP (AND FIXUP-P LIMIT-BP))
	(SETQ STATE (AREF *FSM-STATE-TABLE*
			  STATE
			  (AREF *FSM-CHARACTER-TABLE* (CHAR-CODE (RCHARMAP-CH-CHAR)))))
	(DO ((STATE STATE (AREF *FSM-BACKPOINTER-TABLE* STATE))
	     (OSTATE NIL STATE)
	     (WORD))
	    ((OR (NULL STATE) (EQ STATE OSTATE)))
	  (OR (SETQ WORD (AREF *FSM-WORD-TABLE* STATE))
	      (RETURN NIL))
	  (PUSH (LIST (RCHARMAP-BP-BEFORE) WORD) LIST)))
      (CHARMAP (BP LIMIT-BP (AND FIXUP-P LIMIT-BP))
	(SETQ STATE (AREF *FSM-STATE-TABLE*
			  STATE
			  (AREF *FSM-CHARACTER-TABLE* (CHAR-CODE (CHARMAP-CH-CHAR)))))
	(DO ((STATE STATE (AREF *FSM-BACKPOINTER-TABLE* STATE))
	     (OSTATE NIL STATE)
	     (WORD))
	    ((OR (NULL STATE) (EQ STATE OSTATE)))
	  (OR (SETQ WORD (AREF *FSM-WORD-TABLE* STATE))
	      (RETURN NIL))
	  (PUSH (LIST (CHARMAP-BP-BEFORE) WORD) LIST))))
  (SETQ LIST (NREVERSE LIST))
  (IF EXPR
      (FUNCALL EXPR LIST)
      (CAAR LIST))) 


(DEFUN INITIALIZE-EXTENDED-SEARCH ()
  (COND ((NOT (BOUNDP '*SEARCH-MINI-BUFFER-COMTAB*))
	 (SETQ *SEARCH-CONTROL-H-COMTAB*
	       (SET-COMTAB 'SEARCH-CONTROL-H-COMTAB
			   '(#\( COM-EXTENDED-SEARCH-OPEN
			     #\) COM-EXTENDED-SEARCH-CLOSE
			     #\ COM-EXTENDED-SEARCH-OR
			     #\c-O COM-EXTENDED-SEARCH-OR
			     #\ COM-EXTENDED-SEARCH-AND
			     #\& COM-EXTENDED-SEARCH-AND
			     #\c-A COM-EXTENDED-SEARCH-AND
			     #\c-N COM-EXTENDED-SEARCH-NOT
			     #\~ COM-EXTENDED-SEARCH-NOT
			     #\ COM-EXTENDED-SEARCH-NOT
			     #\SPACE COM-EXTENDED-SEARCH-WHITESPACE
			     #\- COM-EXTENDED-SEARCH-DELIMITER
			     #\9 COM-EXTENDED-SEARCH-DIGIT
			     #\1 COM-EXTENDED-SEARCH-CUSTOM-1
			     #\2 COM-EXTENDED-SEARCH-CUSTOM-2
			     #\3 COM-EXTENDED-SEARCH-CUSTOM-3
			     #\4 COM-EXTENDED-SEARCH-CUSTOM-4
			     #\5 COM-EXTENDED-SEARCH-CUSTOM-5
			     #\6 COM-EXTENDED-SEARCH-CUSTOM-6
			     #\7 COM-EXTENDED-SEARCH-CUSTOM-7
			     #\8 COM-EXTENDED-SEARCH-CUSTOM-8
			     #\A COM-EXTENDED-SEARCH-ALPHABETIC
			     #\C COM-EXTENDED-SEARCH-CAPITAL
			     #\* COM-EXTENDED-SEARCH-SOME
			     #\c-X COM-EXTENDED-SEARCH-ANY
			     #\HELP COM-DOCUMENT-CONTAINING-PREFIX-COMMAND)))
	 (SETQ *SEARCH-MINI-BUFFER-COMTAB*
	       (SET-COMTAB 'SEARCH-MINI-BUFFER-COMTAB
			   (LIST
			     #\HELP 'COM-DOCUMENT-EXTENDED-SEARCH
			     #\c-H (MAKE-EXTENDED-COMMAND *SEARCH-CONTROL-H-COMTAB*)
			     #\m-\y 'COM-EXTENDED-SEARCH-YANK)))      ;gsl
	 (SET-COMTAB-INDIRECTION *SEARCH-MINI-BUFFER-COMTAB* *MINI-BUFFER-COMTAB*)))
  (COND ((NOT (BOUNDP '*STRING-SEARCH-CONTROL-H-COMTAB*))
	 (SETQ *STRING-SEARCH-CONTROL-H-COMTAB*
	       (SET-COMTAB 'STRING-SEARCH-CONTROL-H-COMTAB
			   '(#\c-B COM-EXTENDED-SEARCH-BEGINNING
			     #\c-E COM-EXTENDED-SEARCH-END
			     #\c-F COM-EXTENDED-SEARCH-TOP-LINE
			     #\c-R COM-EXTENDED-SEARCH-REVERSE)))
	 (SET-COMTAB-INDIRECTION *STRING-SEARCH-CONTROL-H-COMTAB* *SEARCH-CONTROL-H-COMTAB*)
	 (SETQ *STRING-SEARCH-MINI-BUFFER-COMTAB*
	       (SET-COMTAB 'STRING-SEARCH-MINI-BUFFER-COMTAB
			   (LIST
			     #\ESCAPE 'COM-END-OF-MINI-BUFFER
			     #\HELP 'COM-DOCUMENT-EXTENDED-SEARCH
			     #\m-\y 'COM-EXTENDED-SEARCH-YANK ;gsl
			     #\c-H (MAKE-EXTENDED-COMMAND *STRING-SEARCH-CONTROL-H-COMTAB*))))
	 (SET-COMTAB-INDIRECTION *STRING-SEARCH-MINI-BUFFER-COMTAB* *MINI-BUFFER-MULTI-LINE-COMTAB*)))
  (COND ((NOT (BOUNDP '*STRING-SEARCH-SINGLE-LINE-COMTAB*))
	 (SETQ *STRING-SEARCH-SINGLE-LINE-COMTAB*
	       (SET-COMTAB 'STRING-SEARCH-SINGLE-LINE-COMTAB
			   '(#\NEWLINE COM-END-OF-MINIBUFFER)))
	 (SET-COMTAB-INDIRECTION *STRING-SEARCH-SINGLE-LINE-COMTAB*
				 *STRING-SEARCH-MINI-BUFFER-COMTAB*)))) 

(ADD-INITIALIZATION "INITIALIZE-EXTENDED-SEARCH"
		    '(INITIALIZE-EXTENDED-SEARCH)
		    '(:NORMAL)
		    '*EDITOR-INITIALIZATION-LIST*) 

(DEFUN GET-SEARCH-MINI-BUFFER-WINDOW ()
  (SEND *MODE-LINE-WINDOW* :SEARCH-MINI-BUFFER-WINDOW)) 

;; Change made from patch 94.215 ddd/gsl, 3/5/84.
;; Modification from patch 94.217  ddd/gsl, 3/5/84.
;; Macro for suggestions added by rpm 11-26-86.
(DEFUN GET-EXTENDED-SEARCH-STRINGS (*SEARCH-MINI-BUFFER-NAME* &AUX STR STRINGS EXPR CR-P FUNCTION)
  "Read with the mini buffer some strings to look for as substrings with FSM search.
Use this when you want to match substrings of strings
rather than search through a buffer.
*SEARCH-MINI-BUFFER-NAME* is a string to prompt with, ending in a colon.
Returns two values, a search function and an arg to give it.
The function might be STRING-SEARCH, or might just take args like it.
 The second value should be given as the first arg to the function."
  (DECLARE (VALUES FUNCTION ARG))
  (SYS:WITH-SUGGESTIONS-MENUS-FOR ZWEI:GET-EXTENDED-SEARCH-STRINGS
    (LET ((*MINI-BUFFER-WINDOW* (GET-SEARCH-MINI-BUFFER-WINDOW)))
      (WITH-BORROWED-FONT-MAP-OF (*MINI-BUFFER-WINDOW* *MODE-LINE-WINDOW*)
	(EDIT-IN-MINI-BUFFER *SEARCH-MINI-BUFFER-COMTAB* NIL NIL
			     '(*SEARCH-MINI-BUFFER-NAME*
				(:RIGHT-FLUSH " (Extended search characters: Control-H)")))))
    (SETQ STR (SEARCH-MINI-BUFFER-STRING-INTERVAL))
    (PUSH-ON-HISTORY (SEARCH-MINI-BUFFER-STRING-INTERVAL) *FSM-SEARCH-HISTORY*)
    (MULTIPLE-VALUE-SETQ (STRINGS EXPR CR-P)
      (PARSE-EXTENDED-SEARCH-STRING STR))
    (IF (OR (CONSP STRINGS) CR-P)
	(SETQ FUNCTION 'FSM-STRING-SEARCH
	      STRINGS (LIST (IF (CONSP STRINGS) STRINGS (CONS STRINGS NIL))
			    EXPR CR-P))
	(SETQ FUNCTION 'GLOBAL:STRING-SEARCH))
    (VALUES FUNCTION STRINGS STR))) 

;; Documentation changed from patch 94.215.  ddd/gsl, 3/5/84.
(DEFCOM COM-DOCUMENT-EXTENDED-SEARCH "Simple help for hairy search." ()
  (FORMAT T "~&You are typing in a search string.  ~:@C is a prefix for more commands.~
	     ~&Type ~:*~:@C ~:@C for a list of what commands are available.~
             ~&While typing the string, ~A is Extended Search Yank.~
             ~&An argument of zero to ~:*~A gives a mouseable listing of the search history.~
             ~&~A after ~:2*~A rotates through the search history."
	  #\c-H #\HELP
	  (KEY-FOR-COMMAND 'COM-EXTENDED-SEARCH-YANK *COMTAB* NIL NIL #\m-\y)
	  (KEY-FOR-COMMAND 'COM-YANK-POP *COMTAB* NIL NIL #\m-Y))
  DIS-NONE) 

(DEFCOM COM-EXTENDED-SEARCH-OPEN "Beginning of new nesting level." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 0 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-CLOSE "End of nesting level." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 1 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-OR "Infix or of two strings." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 2 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-AND "Infix and of two strings within lines." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 3 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-NOT "Negation on next character." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 4 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-SOME "Match any number of next characters." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 5 0 1))
  DIS-TEXT) 

;; standard sets

(DEFCOM COM-EXTENDED-SEARCH-WHITESPACE "Match any whitespace character." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 6 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-DELIMITER "Match any delimiter." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 7 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-ALPHABETIC "Match any alphabetic character." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 10 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-ANY "Match any character." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 11 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-DIGIT "Match any digit character." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 12 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-CAPITAL "Match any capital character." () ;;gsl
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 13 0 1))
  DIS-TEXT) 

;; custom sets

(DEFCOM COM-EXTENDED-SEARCH-CUSTOM-1 "Use custom extended search string set number 1.
Create such sets with the command Extended Search Add Custom Set." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 14 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-CUSTOM-2 "Use custom extended search string set number 2.
Create such sets with the command Extended Search Add Custom Set." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 15 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-CUSTOM-3 "Use custom extended search string set number 3.
Create such sets with the command Extended Search Add Custom Set." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 16 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-CUSTOM-4 "Use custom extended search string set number 4.
Create such sets with the command Extended Search Add Custom Set." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 17 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-CUSTOM-5 "Use custom extended search string set number 5.
Create such sets with the command Extended Search Add Custom Set." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 20 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-CUSTOM-6 "Use custom extended search string set number 6.
Create such sets with the command Extended Search Add Custom Set." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 21 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-CUSTOM-7 "Use custom extended search string set number 7.
Create such sets with the command Extended Search Add Custom Set." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 22 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-CUSTOM-8 "Use custom extended search string set number 8.
Create such sets with the command Extended Search Add Custom Set." ()
  (CONTEXT-CHECK #'EQ (GET-SEARCH-MINI-BUFFER-WINDOW) *WINDOW*)
  (INSERT-MOVING (POINT) (CODE-CHAR 23 0 1))
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-STRING-SEARCH "Search for a hairy string.
String is read in a minibuffer in which Control-H is a prefix for special characters.
A numeric arg repeats the last search." ()
  (EXTENDED-STRING-SEARCH-INTERNAL NIL)) 

(DEFCOM COM-EXTENDED-REVERSE-STRING-SEARCH "Search for a hairy string backward.
String is read in a minibuffer in which Control-H is a prefix for special characters.
A numeric arg repeats the last search." ()
  (EXTENDED-STRING-SEARCH-INTERNAL T)) 

(DEFCOM COM-CONTINUE-EXTENDED-STRING-SEARCH "Does the last extended-string-search again,
in the same direction." ()
  (LET ((*NUMERIC-ARG-P* T))
    (EXTENDED-STRING-SEARCH-INTERNAL NIL))
  DIS-BPS) 

(DEFUN EXTENDED-STRING-SEARCH-INTERNAL (REVERSE-P &AUX (BP (POINT)) BJ-P TOP-LINE-P)
  (OR *NUMERIC-ARG-P* ;;numeric arg means use the last search args again.
      (MULTIPLE-VALUE-SETQ (*EXTENDED-STRING-SEARCH-LAST-FUNCTION*
			     *EXTENDED-STRING-SEARCH-LAST-ARG*
			     REVERSE-P BJ-P TOP-LINE-P)
        (GET-EXTENDED-STRING-SEARCH-STRINGS REVERSE-P)))
  (AND BJ-P
       (SETQ BP (IF REVERSE-P
		    (INTERVAL-LAST-BP *INTERVAL*)
		    (INTERVAL-FIRST-BP *INTERVAL*))))
  (OR (SETQ BP (FUNCALL *EXTENDED-STRING-SEARCH-LAST-FUNCTION* BP
			*EXTENDED-STRING-SEARCH-LAST-ARG* REVERSE-P))
      (BARF "Search failed: ~A" *EXTENDED-STRING-SEARCH-LAST-STRING*))
  (MOVE-TO-BP BP)
  (WHEN TOP-LINE-P
    (RECENTER-WINDOW *WINDOW* :ABSOLUTE 0))  ;gsl 3-15-85
  DIS-BPS) 

;;; Read a string for string search and then return the function to use
(DEFUN GET-EXTENDED-STRING-SEARCH-STRINGS (&OPTIONAL *EXTENDED-STRING-SEARCH-REVERSE-P*
					   (*SEARCH-MINI-BUFFER-NAME* "Search:")
					   (COMTAB *STRING-SEARCH-MINI-BUFFER-COMTAB*)
					   (DEFAULT *EXTENDED-STRING-SEARCH-LAST-STRING*) ;;from brand S. gsl.
					   &AUX (*EXTENDED-STRING-SEARCH-BJ-P* NIL)
					   (*EXTENDED-STRING-SEARCH-ZJ-P* NIL)
					   (*EXTENDED-STRING-SEARCH-TOP-LINE-P* NIL)
					   EXPR CR-P)
  "Read with the mini buffer some strings to look for in the buffer with FSM search.
  ;;What is called ARG here is called the KEY in things like fsm-search-within-lines. gsl
*EXTENDED-STRING-SEARCH-REVERSE-P* should be T
 if the search is going to be backward by default.
*SEARCH-MINI-BUFFER-NAME* is a string to prompt with, ending in a colon.
COMTAB is the comtab to use in the mini buffer.
Returns five values.
The first two are a search function and an arg to give it.
The function takes args like SEARCH (but use NIL as the LINES-TO-SEARCH arg).
 The second value should be given as the first arg to the function.
The third value is T if the search really should be reverse.
Thr fourth value is T if the search should start from the beginning
 or end of the buffer rather than from point (assuming from point is your default).
The fifth value is T if you should put the thing found at the top of the window.
The last three values reflect commands the user typed while in the minibuffer."
  (DECLARE (VALUES FUNCTION ARG REVERSE-P BJ-P TOP-LINE-P))
  (LET ((*MINI-BUFFER-WINDOW* (GET-SEARCH-MINI-BUFFER-WINDOW))
	(*MINI-BUFFER-DEFAULT-STRING* DEFAULT))
    (WITH-BORROWED-FONT-MAP-OF (*MINI-BUFFER-WINDOW* *MODE-LINE-WINDOW*)
			       (EDIT-IN-MINI-BUFFER
				 COMTAB NIL NIL
				 `((*EXTENDED-STRING-SEARCH-BJ-P* "BJ ")
				   (*EXTENDED-STRING-SEARCH-ZJ-P* "ZJ ")
				   (*EXTENDED-STRING-SEARCH-REVERSE-P* "Reverse ")
				   (*EXTENDED-STRING-SEARCH-TOP-LINE-P* "Top line ")
				   (,(NOT (NULL DEFAULT)) *SEARCH-MINI-BUFFER-NAME* " (Default is \"" ,DEFAULT "\")")
				   (,(NULL DEFAULT) ,*SEARCH-MINI-BUFFER-NAME*)
				   (T ,(FORMAT NIL "  (end with ~A)"
                                               (KEY-FOR-COMMAND 'COM-END-OF-MINIBUFFER COMTAB NIL NIL #\END)))
				   (:RIGHT-FLUSH " (Extended search characters)")))))
  (LET ((STRING (SEARCH-MINI-BUFFER-STRING-INTERVAL)))
    (IF (STRING-EQUAL STRING "")
	(SETQ STRING *EXTENDED-STRING-SEARCH-LAST-STRING*))
    (UNLESS (AND (STRING-EQUAL STRING *EXTENDED-STRING-SEARCH-LAST-STRING*)
		 *EXTENDED-STRING-SEARCH-LAST-ARG*
		 *EXTENDED-STRING-SEARCH-LAST-FUNCTION*)
      (WITH-BORROWED-FONT-MAP-OF ((GET-SEARCH-MINI-BUFFER-WINDOW) *MODE-LINE-WINDOW*)
				 (SEND *MODE-LINE-WINDOW* :REDISPLAY
				       (LIST *SEARCH-MINI-BUFFER-NAME* " " STRING)
				       T))
      (SETQ *EXTENDED-STRING-SEARCH-LAST-STRING* STRING)
      (PUSH-ON-HISTORY STRING *FSM-SEARCH-HISTORY*)
      (MULTIPLE-VALUE-SETQ (*EXTENDED-STRING-SEARCH-LAST-ARG* EXPR CR-P)
	(PARSE-EXTENDED-SEARCH-STRING))
      (IF (CONSP *EXTENDED-STRING-SEARCH-LAST-ARG*)
	  (IF EXPR
	      (SETQ *EXTENDED-STRING-SEARCH-LAST-FUNCTION* 'FSM-SEARCH-WITHIN-LINES
		    *EXTENDED-STRING-SEARCH-LAST-ARG* (LIST *EXTENDED-STRING-SEARCH-LAST-ARG* EXPR CR-P))
	      (SETQ *EXTENDED-STRING-SEARCH-LAST-FUNCTION* 'FSM-SEARCH))
	  (SETQ *EXTENDED-STRING-SEARCH-LAST-FUNCTION* 'SEARCH))))
  (VALUES *EXTENDED-STRING-SEARCH-LAST-FUNCTION*
	  *EXTENDED-STRING-SEARCH-LAST-ARG*
	  *EXTENDED-STRING-SEARCH-REVERSE-P*
	  (OR *EXTENDED-STRING-SEARCH-BJ-P*
	      *EXTENDED-STRING-SEARCH-ZJ-P*)
	  *EXTENDED-STRING-SEARCH-TOP-LINE-P*)) 

(DEFCOM COM-EXTENDED-SEARCH-YANK ;;gsl 3-14-85
	"Insert last search string in minibuffer. Arg of 0 gives mouseable listing." ()
  (HISTORY-YANK *FSM-SEARCH-HISTORY*)
  (SETQ *CURRENT-COMMAND-TYPE* *FSM-SEARCH-HISTORY*)
  DIS-TEXT) 

(DEFCOM COM-EXTENDED-SEARCH-BEGINNING "Search forward from buffer's beginning." ()
  (COND ((SETQ *EXTENDED-STRING-SEARCH-BJ-P* (NOT *EXTENDED-STRING-SEARCH-BJ-P*))
	 (SETQ *EXTENDED-STRING-SEARCH-ZJ-P* NIL
	       *EXTENDED-STRING-SEARCH-REVERSE-P* NIL)))
  DIS-NONE) 

(DEFCOM COM-EXTENDED-SEARCH-END "Search backward from buffer's end." ()
  (COND ((SETQ *EXTENDED-STRING-SEARCH-ZJ-P* (NOT *EXTENDED-STRING-SEARCH-ZJ-P*))
	 (SETQ *EXTENDED-STRING-SEARCH-BJ-P* NIL
	       *EXTENDED-STRING-SEARCH-REVERSE-P* T)))
  DIS-NONE) 

(DEFCOM COM-EXTENDED-SEARCH-TOP-LINE "Put match at top of window." ()
  (SETQ *EXTENDED-STRING-SEARCH-TOP-LINE-P* (NOT *EXTENDED-STRING-SEARCH-TOP-LINE-P*))
  DIS-NONE) 

(DEFCOM COM-EXTENDED-SEARCH-REVERSE "Toggle direction of search." () ;;gsl docum.
  (IF (SETQ *EXTENDED-STRING-SEARCH-REVERSE-P* (NOT *EXTENDED-STRING-SEARCH-REVERSE-P*))
      (SETQ *EXTENDED-STRING-SEARCH-BJ-P* NIL)
      (SETQ *EXTENDED-STRING-SEARCH-ZJ-P* NIL))
  DIS-NONE) 

(DEFUN PARSE-EXTENDED-SEARCH-STRING (&OPTIONAL STRING
				     &AUX *EXTENDED-SEARCH-CR-P* STRINGS EXPR LEN)
  (OR STRING
      (SETQ STRING (STRING-INTERVAL (WINDOW-INTERVAL (GET-SEARCH-MINI-BUFFER-WINDOW)))))
  (MULTIPLE-VALUE-SETQ (STRINGS EXPR LEN)
    (PARSE-EXTENDED-SEARCH-STRING-1 STRING 0 (LENGTH STRING) 0)) ;brand s copies string with string append. gsl
  (AND (ATOM STRINGS)
       (DOTIMES (I LEN)			;if not a list and any extended search chars make it a list.
	 (AND (LDB-TEST (BYTE 10 10) (AREF STRINGS I))	;if not a list won't use extended search
	      (RETURN T)))
       (SETQ STRINGS (CONS STRINGS NIL)))
  (OR STRINGS (SETQ STRINGS ""))	;if we got nil return empty string
  (IF EXPR				;Put the expression in a function on a symbol and compile it.  gsl
      (LET ((SYMBOL (GENSYM)))
	(FDEFINE SYMBOL
		 `(LAMBDA (MATCHING-LIST)
		    ,(IF *EXTENDED-STRING-SEARCH-REVERSE-P*
			 `(LET ((FOUND-LINE-INDEX ,EXPR))
			    (AND FOUND-LINE-INDEX
				 (1- FOUND-LINE-INDEX)))
			 EXPR))
		 T NIL)
	(COMPILE SYMBOL)
	(SETQ EXPR SYMBOL))
      (AND *EXTENDED-STRING-SEARCH-REVERSE-P*	;Create a positioning funct., also need expr for right search
	   (SETQ EXPR #'(LAMBDA (X) (IF X
					(PROGN
					  (1- (CAAR (LAST X))))
					NIL)))))
  (VALUES STRINGS EXPR *EXTENDED-SEARCH-CR-P*)) 

(DEFUN PARSE-EXTENDED-SEARCH-STRING-1 (STRING INDEX LENGTH PAREN-LEVEL)
  (DO ((I INDEX (1+ I))
       (SYN)	       ;character type code from parse-extended-search-string-syntax
       (CH)	       ;real character
       (NEW-STRINGS)   ;retruned from recursive calls on tail of string arg.
       (NEW-EXPR)      ;?
       (OLD-STRINGS)   ;figured out as substring between index and i when extended char found.
       (OLD-EXPR))     ;?
      (NIL)
    (MULTIPLE-VALUE-SETQ (SYN CH I)
      (PARSE-EXTENDED-SEARCH-STRING-SYNTAX STRING I LENGTH))
    (IF (ZEROP SYN)
	(AND OLD-STRINGS
	     (IF (CONSP OLD-STRINGS)
		 (DO ((OS OLD-STRINGS (CDR OS)))
		     ((NULL OS))
		   (SETF (CAR OS) (STRING-APPEND (CAR OS) CH)))
		 (SETQ OLD-STRINGS (STRING-APPEND OLD-STRINGS CH))))
	(PROGN
	  (OR OLD-STRINGS
	      (SETQ OLD-STRINGS (AND (NOT (= INDEX I))
				     (SUBSEQ STRING INDEX I))))
	  (COND ((LOGTEST 120000 SYN)  ;EOF or CLOSE
		 (IF (PLUSP PAREN-LEVEL)
		     (AND (= SYN FSM-EOF)      ;EOF
			  (BARF "End of string inside parenthesis"))
		     (AND (= SYN FSM-CLOSE)    ;CLOSE
			  (BARF "Unmatched close")))
		 (RETURN (VALUES OLD-STRINGS OLD-EXPR I))))
	  (COND ((= SYN FSM-OPEN)      ;OPEN
		 (MULTIPLE-VALUE-SETQ (NEW-STRINGS NEW-EXPR I)
		   (PARSE-EXTENDED-SEARCH-STRING-1 STRING (1+ I) LENGTH (1+ PAREN-LEVEL)))
		 (IF (NOT OLD-STRINGS)
		     (SETQ OLD-STRINGS NEW-STRINGS)
		     (PROGN
		       (AND OLD-EXPR (BARF "I don't know how to combine these"))
		       (COND ((CONSP NEW-STRINGS)
			      (AND (CONSP OLD-STRINGS)
				   (BARF "I don't know how to combine these"))
			      (DO ((NS NEW-STRINGS (CDR NS)))
				  ((NULL NS)
				   (SETQ OLD-STRINGS NEW-STRINGS))
				(SETF (CAR NS) (STRING-APPEND OLD-STRINGS (CAR NS)))))
			     ((CONSP OLD-STRINGS)
			      (DO ((OS OLD-STRINGS (CDR OS)))
				  ((NULL OS))
				(SETF (CAR OS) (STRING-APPEND (CAR OS) NEW-STRINGS))))
			     (T
			      (SETQ OLD-STRINGS (STRING-APPEND OLD-STRINGS NEW-STRINGS))))))
		 (SETQ OLD-EXPR NEW-EXPR))
		(T ;;everything but 0 and opening paren losenge go through here.		  
		 (OR OLD-STRINGS
		     (BARF "Special token at the beginning of the string"))
		 (AND (OR (= SYN FSM-AND) NEW-EXPR)
		      (OR OLD-EXPR
			  (SETQ OLD-EXPR (IF (CONSP OLD-STRINGS)
					     (CONS 'OR ;;THIS LOOKS STRANGE but it's right
						   (MAPCAR #'EXTENDED-SEARCH-STRING-MATCHER OLD-STRINGS))
					     (EXTENDED-SEARCH-STRING-MATCHER OLD-STRINGS)))))
		 (OR (CONSP OLD-STRINGS)
		     (SETQ OLD-STRINGS (CONS OLD-STRINGS NIL)))
		 (MULTIPLE-VALUE-SETQ (NEW-STRINGS NEW-EXPR I)
		   (PARSE-EXTENDED-SEARCH-STRING-1 STRING (1+ I) LENGTH PAREN-LEVEL))
		 (OR (CONSP NEW-STRINGS)
		     (SETQ NEW-STRINGS (CONS NEW-STRINGS NIL)))
		 (IF (= SYN FSM-OR)    ;OR
		     (RETURN (VALUES (NCONC OLD-STRINGS NEW-STRINGS)
				     (AND NEW-EXPR `(OR ,OLD-EXPR ,NEW-EXPR))
				     I))
		     (PROGN
		       (OR NEW-EXPR    ;AND
			   (SETQ NEW-EXPR (CONS 'OR
						(MAPCAR #'EXTENDED-SEARCH-STRING-MATCHER NEW-STRINGS))))
		       (RETURN (VALUES (NCONC OLD-STRINGS NEW-STRINGS)
				       `(AND ,OLD-EXPR ,NEW-EXPR)
				       I)))))))))) 

(DEFUN PARSE-EXTENDED-SEARCH-STRING-SYNTAX (STRING INDEX LENGTH
					    &OPTIONAL NOT-TOP-LEVEL &AUX CH)
  "This basically takes an extended search function (e.g. com-extended-search-whitespace)
from the input string and replaces it with the appropriate bit pattern for PARSE-EXTENDED-SEARCH-STRING-1 . 
It also combines the some, not or !foo!(whatever that comes from) character with
the following character and follows that with a noop in the second position."
  (IF (>= INDEX LENGTH)
      (IF NOT-TOP-LEVEL
	  (BARF "EOF in special context")
	  (VALUES FSM-EOF NIL INDEX))
      (VALUES (IF (ZEROP (CHAR-FONT (SETQ CH (AREF STRING INDEX))))
		  (PROG1 0 (AND (= CH (CHAR-CODE #\NEWLINE))
				(SETQ *EXTENDED-SEARCH-CR-P* T)))
		  (CASE (SETQ CH (CHAR-CODE CH))
			(0 FSM-OPEN)		;OPEN
			(1 FSM-CLOSE)		;CLOSE
			(2 FSM-OR)		;OR
			(3 FSM-AND)		;AND
			((4 5 !FOO!)		;NOT,SOME
			 (MULTIPLE-VALUE-BIND (SYN NCH)
			     (PARSE-EXTENDED-SEARCH-STRING-SYNTAX STRING (1+ INDEX) LENGTH T)
			   (OR (ZEROP SYN)
			       (BARF "NOT modifier on a special character"))
			   (SETF (AREF STRING INDEX)
				 (SETQ CH (DPB 1 (CASE CH
						       (4 %%FSM-NOT)
						       (5 %%FSM-STAR1)
						       (!FOO! %%FSM-STAR))
					       NCH)))
			   (SETF (AREF STRING (SETQ INDEX (1+ INDEX))) %FSM-NOOP))
			 0)
			(OTHERWISE
			 ;;a set or an unknown.
			 (COND ((< (CHAR-CODE CH)
				   (+ *EXTENDED-STRING-SEARCH-CUSTOM-SET-MAX* *FSM-SET-H-KEYS* *FSM-NON-SET-H-KEYS*))
				(SETF (AREF STRING INDEX)
				      (SETQ CH (DPB 1 %%FSM-SET (- CH *FSM-NON-SET-H-KEYS*))))
				0)
			       (T 0)))))
	      CH INDEX))) 

(DEFUN EXTENDED-SEARCH-STRING-MATCHER (STRING)
  "Produces a form which when evaluated loops through MATCHING-LIST which presumably
contains elements of form (<index> <string>).  The first such <string> matching the
arg string to this function will cause the index assoc with it to be returned.
The form is equivalent to (car (rassoc (list string) matching-list))."
  `(DO ((L MATCHING-LIST (CDR L)))
       ((NULL L) NIL)
     (AND (EQUAL (CADAR L) ',STRING)
	  (RETURN (CAAR L))))) 

;;; Special handling for search strings
(DEFSTRUCT (16B-STRING
	     :NAMED
	     (:MAKE-ARRAY (:LENGTH 24))
	     (:SUBTYPE 'SYS:FAT-CHAR)
	     (:CONC-NAME NIL)
	     (:CALLABLE-CONSTRUCTORS NIL)
	     (:ALTERANT ALTER-16B-STRING)
	     (:PREDICATE NIL)
	     (:COPIER NIL)
	     (:TYPE :ARRAY-LEADER))
  (16B-STRING-LENGTH 0)) 

(DEFPROP 16B-STRING 16B-STRING-NAMED-STRUCTURE-INVOKE NAMED-STRUCTURE-INVOKE) 

(DEFSELECT 16B-STRING-NAMED-STRUCTURE-INVOKE
  ((:PRINT-SELF)
   (STRING &OPTIONAL (STREAM *STANDARD-OUTPUT*) IGNORE IGNORE)
   (AND *PRINT-ESCAPE*
	(GET-DISPATCH-MACRO-CHARACTER #\# #\")
	(WRITE-CHAR #\# STREAM))
   (SI:PRINT-QUOTED-STRING STRING
			   STREAM
			   (MEMBER :STRING-OUT (SEND STREAM :WHICH-OPERATIONS)
				   :TEST #'EQ)))) 


;;;Removed for release 3 by rpm on 10-10-86.
;;;(DEFUN |MY-#"-MACRO| (IGNORE STREAM)
;;;  (DO ((STRING (MAKE-16B-STRING))
;;;       (CHAR)
;;;       (SLASH-P NIL))
;;;      (NIL)
;;;    (SETQ CHAR (READ-CHAR STREAM))
;;;    (COND (SLASH-P
;;;	   (SETQ SLASH-P NIL)
;;;	   (VECTOR-PUSH-EXTEND CHAR STRING))
;;;	  ((CHAR= CHAR #\")
;;;	   (RETURN STRING))
;;;	  ((CHAR= CHAR #\/)
;;;	   (SETQ SLASH-P T))
;;;	  (T
;;;	   (VECTOR-PUSH-EXTEND CHAR STRING))))) 
;;;
;;;(DEFVAR |SPECIAL-#"-READTABLE|) 
;;;
;;;(DEFUN |INITIALIZE-SPECIAL-#"-READTABLE| ()
;;;  (COND ((NOT (BOUNDP '|SPECIAL-#"-READTABLE|))
;;;	 (SETQ |SPECIAL-#"-READTABLE| (COPY-READTABLE))
;;;	 (SET-SYNTAX-#-MACRO-CHAR #\" '|MY-#"-MACRO| |SPECIAL-#"-READTABLE|)))
;;;  |SPECIAL-#"-READTABLE|) 

(DEFUN SEARCH-MINI-BUFFER-STRING-INTERVAL (&AUX INT STRING)
  (SETQ INT (WINDOW-INTERVAL (GET-SEARCH-MINI-BUFFER-WINDOW)))
  (LET ((LEN (COUNT-CHARS INT)))
    (SETQ STRING (MAKE-16B-STRING 16B-STRING-LENGTH LEN
				  :MAKE-ARRAY (:LENGTH LEN))))
  (LET ((INT-STRING (STRING-INTERVAL INT)))
    (COPY-ARRAY-CONTENTS INT-STRING STRING)
    (RETURN-ARRAY (PROG1 INT-STRING
			 (SETF INT-STRING NIL))))
  STRING) 

(DEFUN GET-EXTENDED-SEARCH-16B-STRING (*SEARCH-MINI-BUFFER-NAME*)
  "Return a string itself, suitable for printing and reading back."
  (LET ((*MINI-BUFFER-WINDOW* (GET-SEARCH-MINI-BUFFER-WINDOW)))
    (WITH-BORROWED-FONT-MAP-OF (*MINI-BUFFER-WINDOW* *MODE-LINE-WINDOW*)
      (EDIT-IN-MINI-BUFFER *SEARCH-MINI-BUFFER-COMTAB*
			   NIL NIL
			   '(*SEARCH-MINI-BUFFER-NAME*
			      (:RIGHT-FLUSH " (Extended search characters)")))))
  (PUSH-ON-HISTORY (SEARCH-MINI-BUFFER-STRING-INTERVAL) *FSM-SEARCH-HISTORY*)
  (SEARCH-MINI-BUFFER-STRING-INTERVAL)) 

;;; Parse something read back
(DEFUN PARSE-EXTENDED-SEARCH-16B-STRING (STRING &AUX FUNCTION STRINGS EXPR)
  (DECLARE (VALUES FUNCTION KEY))
  (MULTIPLE-VALUE-SETQ (STRINGS EXPR)
    (PARSE-EXTENDED-SEARCH-STRING STRING))
  (COND ((ATOM STRINGS)
	 (SETQ FUNCTION 'SEARCH))
	(EXPR
	 (SETQ FUNCTION 'FSM-EXPR-SEARCH
	       STRINGS (LIST STRINGS EXPR)))
	(T
	 (SETQ FUNCTION 'FSM-SEARCH)))
  (VALUES FUNCTION STRINGS)) 

(DEFCOM COM-EXTENDED-SEARCH-ADD-CUSTOM-SET "Add a custom extended search set definition." ()
  (IF (OR (< *NUMERIC-ARG* 1)
	  (> *NUMERIC-ARG* *EXTENDED-STRING-SEARCH-CUSTOM-SET-MAX*))
      (BARF "ILLEGAL NUMERIC ARG, PLEASE CHOOSE BETWEEN 1 AND ~D."
	    *EXTENDED-STRING-SEARCH-CUSTOM-SET-MAX*))
  (LET (INTERVAL (SET-LIST-INDEX (+ *NUMERIC-ARG* *FSM-SET-H-KEYS* -1)))
    (UNWIND-PROTECT
	(PROGN (COMMAND-STORE 'COM-INDENT-FOR-LISP #\TAB *MINI-BUFFER-MULTI-LINE-COMTAB*)	;pardon the kludge
	       (MULTIPLE-VALUE-SETQ (NIL NIL INTERVAL)
		 (EDIT-IN-MINI-BUFFER *MINI-BUFFER-MULTI-LINE-COMTAB*
				      NIL NIL
				      `(,(FORMAT NIL
						 "Forms to test set ~D membership:   (end with End)"
						 *NUMERIC-ARG*)))))
      (COMMAND-STORE 'COM-INSERT-TAB #\TAB *MINI-BUFFER-MULTI-LINE-COMTAB*))
    (LET ((FORM-STRING (STRING-INTERVAL INTERVAL))
	  FORM
	  (EOF '(NIL)))
      (CONDITION-CASE (ERROR)
	  (MULTIPLE-VALUE-SETQ (FORM NIL)
	    (READ-FROM-STRING FORM-STRING NIL EOF))
	(SYSTEM:READ-ERROR (BARF (SEND ERROR :REPORT-STRING))))
      (IF (NOT (EQUAL FORM EOF))
	  ;;TEST THE FORM FOR VALIDITY.
	  (LET ((FUNCTION (COND ((FUNCTIONP FORM)
				 (IF (AND (CONSP FORM)
					  (EQ (CAR FORM) 'LAMBDA))
				     (LET ((NAME (GENSYM)))
				       (FDEFINE NAME FORM)
				       (SETF (GET NAME 'FSM-CUSTOM-SET) FORM)
				       (COMPILE NAME)
				       (SETQ FORM NAME)))
				 (MULTIPLE-VALUE-BIND (NIL ERRORP)
				     (CATCH-ERROR (FUNCALL FORM #\a) NIL)
				   (IF ERRORP
				       (BARF "Your function doesn't seem to work.")))
				 FORM)
				((AND (CONSP FORM)
				      (APPLY #'AND (MAPCAR #'(LAMBDA (X)
							       (AND (CONSP X)
								    (CHARACTERP (CAR X))
								    (CHARACTERP (SECOND X))))
							   FORM)))
				 FORM)
				((AND (CONSP FORM)
				      (APPLY #'AND (MAPCAR #'CHARACTERP FORM)))
				 FORM)
				(T
				 (BARF "Please enter a function of one character, a list of characters,~@
                                    or a list of two-character-list ranges.")))))
	    (RPLACA (NTHCDR SET-LIST-INDEX *FSM-SEARCH-SET-LIST*)
		    FUNCTION)))))
  DIS-NONE) 

(DEFCOM COM-EXTENDED-SEARCH-SHOW-CUSTOM-SETS "Show custom extended search set indicated by numeric arg." ()
  (DOTIMES (SET-NUMBER *EXTENDED-STRING-SEARCH-CUSTOM-SET-MAX*)
    (FORMAT T "~%CUSTOM EXTENDED SEARCH SET ~D DEFINITION:~%" (1+ SET-NUMBER))
    (LET ((DEF (NTH (+ SET-NUMBER *FSM-SET-H-KEYS*) *FSM-SEARCH-SET-LIST*)))
      (COND ((FUNCTIONP DEF)
	     (FORMAT T "~A~%" (OR (GET DEF 'FSM-CUSTOM-SET) DEF)))
	    ((AND (CONSP DEF) (CONSP (CAR DEF)))
	     (FORMAT T "(")
	     (DOLIST (PAIR DEF)
	       (FORMAT T "(~:C ~:C)" (CAR PAIR) (SECOND PAIR)))
	     (FORMAT T ")~%"))
	    ((CONSP DEF)
	     (FORMAT T "(~{~:C~^ ~})~%" DEF))
	    ((NULL DEF)
	     NIL)
	    (T
	     (FORMAT T "We have something illegal here: ~A~%" DEF)))))
  DIS-NONE) 
